:- module(pac_demo, [demo/1]).
:- use_module(pacword).
term_expansion --> expand_pac.
:- include(op).

		/*************************************************
		*     Demo :  Compile clause / DCG with regex    *
		*************************************************/

% ?- ['pac-demo'].

% ?- demo(a).
% ?- demo("zip(P,Q):-maplist(pred([X,Y,X-Y]), P, Q)").
% ?- demo("a :- call(pred([X]))").
% ?- demo("a --> b").
% ?- demo('a --> w(".*")').
% ?- demo('trim_white(A) --> wl("[ \t]*"), w(".*", A), wl("[ \t]*")').

demo(X) :- expand_pacword_display(X).

%
expand_pacword_display(S):- term_string(X, S, 
										[module(web), 
										 variable_names(Eqs)]),
	compile_pred_word(X, Eqs, H0, R0),
	smash(["\n", H0, ".\n\n", R0, "\n"]).

% a :- call(pred(['a#1']), X).
% a :- call(pred(['A']), X).
% a :- maplist(pred([X, Y, X-Y]), [1,2,3], A).
% trim_white(A) --> wl("[ \t]*"), w(".*", A), wl("[ \t]*").

% ?- ejockey:new_names([X,Y], Eqs, 1, 'A', [a1,a2,a3]).

new_names([V|Vs], [A=V|Eqs], N, Prefix, As):-
	new_name(N, As, A, Prefix, K),
	new_names(Vs, Eqs, K, Prefix, As).
new_names([], [], _, _, _).

%
new_name(N, As, B, Prx, K):- atom_concat(Prx, N, B),
						\+ memberchk(B, As),
						!,
						succ(N, K).
new_name(N, As, A, Prx, K):- succ(N, N1),
						new_name(N1, As, A, Prx, K).

% ?- pac_demo:subtractq([X,Y,X,Y,X,Y], [X], R).
subtractq([], _, []).
subtractq([A|As], B,  C):- memq(A, B), !,
	subtractq(As, B, C).
subtractq([A|As], B, [A|C]):- subtractq(As, B, C).

%
expand_clause_slim(X, Y):- 
	expand_clause(X, [], Y0),
	maplist(pred([X:-true, X] & [C, C]),  Y0, Y).
%
compile_pred_word(X-->X0, Eqs, H0, R0):-!, 
		maplist(pred([A=P, A, P]), Eqs, As, Vs),
		expand_clause_slim(X-->X0, [H|R]),
		term_variables(H, HVs),
		subtractq(HVs, Vs, SVs),
		new_names(SVs, Eqs0, 1, 'A', As),
		append(Eqs0, Eqs, Eqs1),
		term_string(H, H0, [variable_names(Eqs1), 
							quoted(true)]),
		maplist(pred(([U, [V,".\n"]] :- 
								numbervars(U, 0, _),
								term_string(U, V, [ numbervars(true), 
													quoted(true)]))),  
				R, R0).
compile_pred_word(X, Eqs, H0, R0):- 
		expand_clause_slim(X, [H|R]),
		term_string(H, H0, [variable_names(Eqs), 
							quoted(true)]),
		maplist(pred(([U, [V,".\n"]] :- 
								numbervars(U, 0, _),
								term_string(U, V, [ numbervars(true), 
													quoted(true)]))),  
				R, R0).

