:- module(querysample, []).
:- use_module(pac(engine)).
term_expansion --> expand_pac.

% ?- sqc(querysample).
% ?- calc(3, X).
% ?- calc(3+4+5, X).
% ?- numlist_to_plus(0, 100000, X),  time(Y is X).
%@ % 3 inferences, 0.004 CPU in 0.004 seconds (96% CPU, 802 Lips)
%@ X = 0+(1+(2+(3+(4+(5+(6+(7+(8+(... + ...))))))))),
%@ Y = 5000050000.
% ?- numlist_to_plus(0, 100000, X),  time(calc(X, Y)).
%@ % 300,003 inferences, 0.019 CPU in 0.025 seconds (77% CPU, 15779665 Lips)
%@ X = 0+(1+(2+(3+(4+(5+(6+(7+(8+(... + ...))))))))),
%@ Y = 5000050000.

numlist_to_plus(I, J, P):- numlist(I, J, Is),
	numlist_to_plus(Is, P).

numlist_to_plus([X], X):-!.
numlist_to_plus([X|R], X+E):- numlist_to_plus(R, E).

% ?- listing(calc).
:-bekind(calc, []).
X + Y = :plus@X@Y.
X - Y = xargs([A,B,C]:- plus(B,C,A))@X@Y.
- X  =  0 - X.
+ X  = X.
X  = `X.
:-ekind.

% ?- list_concat_xargs((([a,b]^2+[c,d]^2)^3)^5, V), length(V, L).
% ?- time(repeat(1000, list_concat((([a,b]^2+[c,d]^2)^3)^5, V))), length(V, L).
% ?- time(repeat(1000, list_concat_xargs((([a,b]^2+[c,d]^2)^3)^5, V))), length(V,L).

% ?- listing(list_concat).
:-bekind(list_concat, []).
X + Y	= :append@X@Y.
X - Y	= xargs([A,B,C]	:- append(B,C,A))@X@Y.
_X^0	= [].
X^N	= X + X^N0
	:- succ(N0, N).
X	= `X.
:-ekind.

% ?- listing(list_concat_xargs).
:-bekind(list_concat_xargs, []).
X + Y	= :append@X@Y.
X - Y	= xargs([A,B,C]
	:- append(B,C,A))@X@Y.
_X^0	= [].
X^N	= xargs([A] -> A + A^N0)@X
	:- succ(N0, N).
X	= `X.
:-ekind.

% ?- ls_recursive(X), maplist(writeln, X).

ls_skip((.)).
ls_skip((..)).

%
ls_recursive(X):- directory_files((.), Files),
		ls_recursive(Files, X, []).
%
ls_recursive([], X, X).
ls_recursive([F|R], X, Y):- ls_skip(F), !,
	   ls_recursive(R, X, Y).
ls_recursive([D|R], [D-Z|X], Y):- exists_directory(D), !,
		directory_files(D, Files),
		working_directory(_, D),
		ls_recursive(Files, Z, []),
		working_directory(_, (..)),
		ls_recursive(R, X, Y).
ls_recursive([F|R], [F|X], Y):- ls_recursive(R, X, Y).
