:- module(anti_subst, [
			  anti_subst/3,
			  anti_subst/4,
			  expand_clause/4,
	  		  expand_head/6,
			  expand_atomic_goal/6
					   ]).

:- use_module(reduce).

% :- module(anti_subst).
% some tiny for the reserved '.' for swi.

%% period_args(+X, -A, -B) is det.
% True  X is of the form A.B.
period_args(X, A, B):- compound(X),
	compound_name_arguments(X, '.', [A, B]).
%
is_role(role(X, Y), X, Y):-!.
is_role(X, A, B):- period_args(X, A, B).

% for debugging  is_odict/1.
% ?- disable_odict.
% ?- enable_odict.
% ?- disable_pac_query.
% ?- enable_pac_query.
% ?- trace, is_odict({a:1, b:{2,3}}).
% ?- is_odict({X, b:{2,3}}).
% ?- mustbe_odict({X, b:{2,3}}).
% ?- trace, is_odict({a:X, b:{2,3}}).
% ?- mustbe_odict({a:X, b:{2,3}}).

%% is_odict(+X) is det.
%  True if X is instnatiated as {k1:_, ..., kn:_}
%  for n >= 1, ground k_i (1=<k=<n), where
%  commutativity of (,) is assumed.

is_odict(X):- nonvar(X), X = {X0}, nonvar(X0),
	key_value_pairs(X0).
%
key_value_pairs((X, Y)):- nonvar(X), nonvar(Y),
	key_value_pairs(X),
	key_value_pairs(Y).
key_value_pairs(K: _):- ground(K).

%
mustbe_odict(X):- is_odict(X), !.
mustbe_odict(X):- writeln("Open dict expected found:"),  writeln(X).

% ?- module(anti_subst).
% ?- disable_odict, disable_pac_query.

% ?- expand_dict:btree_build({a:b}, R).
% ?- anti_subst(a, A, R).
% ?- anti_subst({a:b, c:d}, A, R).
% ?- anti_subst(f({a:b}), A, R).
% ?- anti_subst(f({a:g({b:c})}), A, R).
% ?- anti_subst({a:b,{d:e}, c:d}, A, R).
% ?- anti_subst(f({a:b,{d:e}, c:d}), A, R).
% ?- anti_subst(f({a:b,c:g({d:e}), c:d}), A, R).
% ?- anti_subst({a: pred([a]:-d), c:d}, A, R).
% ?- anti_subst({a: (@(b,c))}, A, R).
% ?- anti_subst(f({a: (@(b,c))}, {d:e}), A, R).

anti_subst(X, A, R):- once(anti_subst(X, A, R, [])).

anti_subst(X, X, P, P):- (var(X); atomic(X)), !.
anti_subst({X}, A, [{X0}-A|P], Q):- is_odict({X}), !, anti_subst(X, X0, P, Q).
anti_subst(quote(X), X, R, R).
anti_subst(nopac(X), X, R, R).
anti_subst((X, Y),  (X0, Y0), P, Q):-
	anti_subst(X, X0, P, P0),
	anti_subst(Y, Y0,  P0, Q).
anti_subst(M:X, M:A, P, Q):-	anti_subst(X, A, P, Q).
anti_subst(X, A, [Y-A|R], R):-	anti_subst_rule(X, Y), !.
anti_subst([X|Xs], [A|As], P, Q):- !,
	anti_subst(X, A, P, P0),
	anti_subst(Xs, As, P0, Q).
anti_subst(X, X, P, P):- compound_name_arity(X, _, 0), !.
anti_subst(X, A, P, Q):- X =..[H|Xs],
					anti_subst(Xs, As, P, Q),
					A =..[H|As].
%
% anti_subst_rule(@(X,Y), @(X, Y)).
anti_subst_rule(is(X), arith_exp(X)).
anti_subst_rule(X, pac(X0)):- normal_pac(X, X0).
anti_subst_rule(X, Y):- period_args(X, A, B),
	(	user:chk_odict ->  Y = role(A, B)
	;   Y = X
	).

%
normal_pac(pred(X),			pred(X)).
normal_pac(pred(Vs, X),		pred(Vs, X)).
normal_pac(rec(F, Cs),		rec(F,  Cs)).
normal_pac(rec(F, Vs, Cs),	rec(F, Vs, Cs)).
normal_pac(mrec(Defs),		mrec(Defs)).
normal_pac(mrec(Vs, Defs),	mrec(Vs, Defs)).

 	/****************************************
	*   Expand CIL clauses, DCG rules,		*
	*	and queries.						*
	****************************************/

% ?- anti_subst:expand_clause(a:-b, [], R, []).
% ?- anti_subst:expand_clause(a(1):-b, [], R, []).
% ?- anti_subst:expand_clause(a({x}):-b, [], R, []).
% ?- anti_subst:expand_clause(a({b:c}):-d, [], R, []).

expand_clause(M:A, M0, P, Q):- M==[], !,
		expand_clause(A, M0, P, Q).
expand_clause(M:A, _, P, Q):- !, expand_clause(A, M, P, Q).
expand_clause(:-(H, B), M, [:-(H0, B0)|P], Q):-!,
		expand_head(H, M, H0, U, P, P0),
		pac:expand_goal(B, M, V, P0, Q),
		slim_goal((U, V), B0).
expand_clause(H-->B, M, [C|P], Q):- !,
		expand_head(H, M, H0, U, P, P0),
		pac:expand_phrase(B, M, B0, P0, Q),
		dcg_translate_rule((H0 --> {U}, B0), C).
expand_clause(H, M, P, Q):- expand_clause( H:-true, M, P, Q).

%
expand_atomic_goal(G, M, H, U, P, Q):-
		anti_subst(G, G0, Aux),
		attach_prefix(M, G0, H),
		expand_aux(Aux, M, U0, P, Q),
		slim_goal(U0, U).
%
% ?- anti_subst:expand_head(a(pred([b])), [], H, G, P, []).
% ?- anti_subst:expand_head(a({b:pred([b])}), [], H, G, P, []).
% ?- anti_subst:expand_head(a(is(1+1)), [], H, G, P, []).
% ?- anti_subst:expand_head(a(is(role(X,A))), [], H, G, P, []).
% ?- anti_subst:expand_head(a(role(X,A)), [], H, G, P, []).
% ?- anti_subst:expand_head(a(is(role(X,A))), [], H, G, P, []).

expand_head(H, M, NewH, U, P, Q):-
        anti_subst(H, H0, Aux),
		expand_aux(Aux, M, U, P, Q),
		attach_prefix(M, H0, NewH).
%
expand_aux([], _, true, P, P).
expand_aux([X|Xs], M, (G0, G), P, Q):-
	expand_aux_base(X, M, G0, P, P0),
	expand_aux(Xs, M, G, P0, Q).

% ?- anti_subst:chk_odict.

expand_aux_base({X}-A, M, (G1, G0), P, Q):-
	anti_subst(X, X0, Aux),
	(	user:chk_odict
	->		expand_dict:expand_dict({X0}, A, G0)
	;		A = {X0}
	),
	expand_aux(Aux, M, G1, P, Q).
expand_aux_base(X-A, M, G, P, Q):- is_role(X, _, _), !,
	pac:expand_dict_access(X, =, A, M, G, P, Q).
expand_aux_base(pac(X)-A, M, true, P, Q):-
	pac:expand_core(X, M, A, P, Q).
expand_aux_base(arith_exp(E)-A, M, G, P, Q):-
	anti_subst(E, E0, Aux),
	expand_aux(Aux, M, G0, P, Q),
	slim_goal((G0, A is E0), G).
