:- module(arith_partial_eval,[]).

% ?- arith_partial_eval(0, X).
%@ X = 0.
% ?- arith_partial_eval(0+A+0, X).%@ A = X.
%@ A = X.
% ?- arith_partial_eval(1+(A+B)*0, X).
%@ X = 1.
% ?- arith_partial_eval(0+ (X<<((2*A*1 + 4*B*1)*0))*(1+2), R).
%@ R = X+X*2.
%@ R = X+X*2.
% ?- arith_partial_eval(0+ X<<(((((2*A*1 + 4*B*1)*0)))*(1+2)), R).
%@ X = R.
% ?- arith_partial_eval(X+Y+Z -(Y + Z), R).
%@ X = R.
% ?- arith_partial_eval(X - Y + Z -(Y + Z), R).
%@ R = X+ -Y+ -Y.
% ?- arith_partial_eval(X << (Y - Y + Z - Z), R).
%@ X = R.
% ?- arith_partial_eval(X ^ (1-1), R).

% ?- arith_partial_eval(0 ^ (1-1), R).
% ?- arith_partial_eval(X + X + X, R).
%@ R = X+X+X.
% ?- arith_partial_eval(X + X + Y + X , R).


% ?- listing(arith_rule).

elim_term(A, B, 0)	:- A==B.
elim_term(A, B, X)	:- nonvar(B, X+Y), A==Y.
elim_term(A, B, Z+Y):- nonvar(B, X+Y), elim_term(A, X, Z).
%
nonvar(X, Y):- nonvar(X), Y=X.
%
ground_compound_basic_exp(E):- compound(E),
	arg(1, E, A),
	integer(A),
	(	arg(2, E, B) -> integer(B)
	;	true
	).
%
arith_partial_eval(X, Y):- arith_rule(X, X0), !,
	arith_partial_eval(X0, Y).
arith_partial_eval(X, X).
%
arith_rule(A, _)			:- var(A), !, fail.
arith_rule(A - B, A + -(B)).
arith_rule(E*C, A*C+B*C)	:- nonvar(E, A+B).
arith_rule(C*E, C*A+C*B)	:- nonvar(E, A+B).
arith_rule(A+E, (A+B)+C)	:- nonvar(E, B+C).
arith_rule(A*E, (A*B)*C)	:- nonvar(E, B*C).
arith_rule(A+B, A0+B)		:- arith_rule(A, A0).
arith_rule(A+B, A+B0)		:- arith_rule(B, B0).
arith_rule(+(A), +(A0))		:- arith_rule(A, A0).
arith_rule(-E, -A + -B)		:- nonvar(E, A+B).
arith_rule(-E, A)			:- nonvar(E, -A).
arith_rule(-(A), -(A0))		:- arith_rule(A, A0).
arith_rule(E, V)			:- ground_compound_basic_exp(E), V is E.
arith_rule(A*B, A0*B)		:- arith_rule(A, A0).
arith_rule(A*B, A*B0)		:- arith_rule(B, B0).
arith_rule(A//B, A0//B)		:- arith_rule(A, A0).
arith_rule(A//B, A//B0)		:- arith_rule(B, B0).
arith_rule(A<<B, A0<<B)		:- arith_rule(A, A0).
arith_rule(A<<B, A<<B0)		:- arith_rule(B, B0).
arith_rule(A>>B, A0>>B)		:- arith_rule(A, A0).
arith_rule(A>>B, A>>B0)		:- arith_rule(B, B0).
arith_rule(A^B, A0^B)		:- arith_rule(A, A0).
arith_rule(A^B, A^B0)		:- arith_rule(B, B0).
arith_rule(A mod B, A0 mod B)	 :- arith_rule(A, A0).
arith_rule(A mod B, A mod B0)	 :- arith_rule(B, B0).
arith_rule(E*_, 0)		:- E == 0.
arith_rule(_*E, 0)		:- E == 0.
arith_rule(E+X, X)		:- E == 0.
arith_rule(X+E, X)		:- E == 0.
arith_rule(X>>E, X)		:- E == 0.
arith_rule(X<<E, X)		:- E == 0.
arith_rule(E//_, 0)		:- E == 0.
arith_rule(E*X, X)		:- E == 1.
arith_rule(X*E, X)		:- E == 1.
arith_rule(_^E, 1)		:- E == 0.
arith_rule(A^E, A)		:- E == 1.
arith_rule(E^_, 1)		:- E == 1.
arith_rule(A+B, B+A)	:- integer(A).
arith_rule(A*B, B*A)	:- integer(A).
arith_rule(A+B, B+A)	:- nonvar(A, A0*H), nonvar(B, B0*K),
	integer(H), integer(K),
	A0 @< B0.
arith_rule(A+B, A0*D)	:- nonvar(A, A0*H), nonvar(B, B0*K),
					A0 == B0,
					D is H+K.
arith_rule(A+B, B+A)	:- arith_mono_compare(<, A, B).
arith_rule(A+B, (U+B)+V)	:- nonvar(A, U+V),
	arith_mono_compare(<, V, B).

%
arith_mono_compare(=, A, B):-integer(A), integer(B).
arith_mono_compare(<, A, _):-integer(A).
arith_mono_compare(>, _, B):-integer(B).
arith_mono_compare(C, A, B):- nonvar(A, A0*I), nonvar(B, B0*K),
	integer(I),
	integer(K),
	compare(C, B0, A0).
arith_mono_compare(C, A, B):- nonvar(A, A0*I),
	integer(I),
	B0 = B,
	compare(C, B0, A0).
arith_mono_compare(C, A, B):- nonvar(B, B0*I),
	integer(I),
	A0 = A,
	compare(C, B0, A0).
arith_mono_compare(C, A, B):- compare(C, A, B).
