:- module(basic, [
	max/3, min/3, max_of_list/2, min_of_list/2,
	ahead_compare/4,
	associative_comma/3, comma_list/2,
	iterated_deepening/4,
	completing_options/3, kleene_star/3, kleene_star_greedy/3, kleene_plus_greedy/3,
	kleene_string_codes/3,
	term_codes/2, term_codes/3,
	concat/2, flatten/3,
	pipe_create/2, pipe_create/3, pipe_to_obj/2, obj_to_pipe/2,
	assocq/3, memq/2, clear/2,
	clear/3, collect/3, current/3, env/3,
    herbrand/2, herbrand/3, herbrand/4, herbrand0/2,
	herbrand_in_context/2, herbrand_in_context/3,
	id/2, if/2, if/3,
	listcast/2, listsubst/3,
	prepend/3,
	ordered_pair/3,
	paragraph/2,
	listp/1,
	peek/3, peek/4,
	remove/3, remove_last/3,
	smash/1, smash/2, smash_codes/2,
	term_string0/2, pred_term_smash/3, term_smash/2, term_smash0/2,
	fst/2, fst/3, snd/2, snd/3, split/2, split/3,
	split/4, split/6, split_rest/6, stack_init/0,
	split_string_by_word/3,
	scan_prefix/4, scan_prefix/5,
	remove_comment/2, remove_comment/3,
	bi_reverse/3,
	stack_pop/2, stack_push/1, stack_push/2,
	stack_top/1, stack_top/2, stack_update/1,
	residue/3,
	stack_update/2,
	union/2, flatten_dl/3,
	herbrand_opp/2, herbrand_opp/3, variant/4, cons/3
	]).


%
max(A, B, A):- A@>B, !.
max(_, B, B).

min(A, B, A):- A@<B, !.
min(_, B, B).

% ?- max_of_list([1,2,-3], M).
max_of_list([X], X):-!.
max_of_list([X, Y|Z], M):- X@<Y, !,
	max_of_list([Y|Z], M).
max_of_list([X, _|Z], M):- max_of_list([X|Z], M).

% ?- min_of_list([1,2,-3], M).
min_of_list([X], X):-!.
min_of_list([X, Y|Z], M):- Y@<X, !,
	min_of_list([Y|Z], M).
min_of_list([X, _|Z], M):- min_of_list([X|Z], M).


:- use_module(pac(op)).

%@ basic.pl qcompiled.
%%	iterated_deepening(G:pred, +D:int, +N:int, -R) is nondet
%	Iterate  at most N times call_with_depth_limit(G, D, R),
%	incrementing D to twice larger for each iteration step.

% ?- iterated_deepening(append([a,b,c,d,e,f],[g,h], X), 2, 3, R).
%@ X = [a, b, c, d, e, f, g, h],
%@ R = 7.
:- meta_predicate iterated_deepening(0, ?, ?, ?).
iterated_deepening(_,  _, 0, depth_limit_reached):-!,
	throw(depth_limit_reached).
iterated_deepening(G, D, N, S):-  N0 is N - 1,
	writeln(depth(D, N)),
	call_with_depth_limit(G, D, R),
	(  R == depth_limit_exceeded
	->	D1 is D + D,
		iterated_deepening(G, D1, N0, S)
	;	R = S
	).

%%	completing_options(+X:list, +Y:list, -Z:list) is det.
%	Update options list Y with X to Z.

% ?- completing_options([i(2)], [], X).
% ?- completing_options([i(2), j(1)], [i(1), j(A)], X).
completing_options([D|R], X, Y):-
	functor(D, F, 1),
	functor(D0,F, 1),
	(	memberchk(D0, X)
	->	completing_options(R, X, Y)
	;	completing_options(R, [D|X], Y)
	).
completing_options([], X, X).

%%	if(+X, +Y, +Z) is nondet.
%	Meta predicate for conditional senteces.

:-meta_predicate if(0, 0, 0).

if(X, Y, Z)	:-	( call(X)   ->  call(Y); call(Z) ).
if(X, Y)	:-	( call(X)   ->  call(Y); true  ).

%
cast_to_list(X, Y):- var(X) -> Y = [X]
	; (X == []; X = [_|_]) -> Y = X
	; Y = [X].

listp([]):-!.
listp([_|_]).

% refresh_term(A, X, Y):- copy_term((A,X), (A,Y)).
% fs_fresh(fs_alt(_Fs, Gs, M, Cnt), fs_alt(Gs, Gs, M, Cnt)).

fst(F) --> main(F).
snd(F) --> env(F).
main(F, (X, Y), (X0, Y)):- phrase(F, X, X0).
env(F, (X, Y), (X, Y0)):- phrase(F, Y, Y0).
pipe_init(X, (X, [])).

%%  DCG in duo pipe form 'P'(X, Y)
pipe_create(X, (X, [])).
%
pipe_create(X, Y, (Y, X)).
%
pipe_to_obj(P, Ob):- var(P), !,
	obj:obj_pull([acc(X)], Ob, Ob0),
	P = (X, Ob0).
%
pipe_to_obj((X, Y), [acc(X)|Y]).

%
obj_to_pipe(X, Y):- pipe_to_obj(Y, X).

% helpers
unary(F, A, T)	:- T =.. [F, A].
unary(F, T)	:- functor(T, F, 1).
n_ary(F, As, T)	:- T =.. [F | As].

%
list_to_comma([], true).
list_to_comma([X], X):-!.
list_to_comma([X|Y], (X,Z)):- list_to_comma(Y,Z).
%
binary_term(F, A, B, C):- C =.. [F, A, B].

% ?- binary_term(f, [a], R).
%@ R = f(a,f(b,c)).
binary_term(_, [A], A):-!.
binary_term(F, [A|B], C):- binary_term(F, B, C0),
	binary_term(F, A, C0, C).

% basic predicates
% for handling internal prolog variable preventing unexpected unification.
% narrowing(X, Y)
% ?- narrowing(X, Y).
% X = Y.
% ?- narrowing(X, (Y)).
% X = Y.
% ?- narrowing(X, f(Y)).
% false.
% ?- narrowing(f(a), f(Y)).
% Y = a.
narrowing(X, Y) :-  subsumes(Y, X).

%
memq(X, [Y|_]):- X==Y, !.
memq(X, [_|Ys]):- memq(X, Ys).

%
assocq([(A0, B)|_], A, B):- A==A0, !.
assocq([_ |Ps], A, B):- assocq(Ps, A, B).

% ?- assoc([a = b], a, R).
assoc([A0 = B |_], A, B):- A==A0, !.
assoc([_ |Ps], A, B):- assoc(Ps, A, B).

%
unlist([X], X):-!.
unlist(X, X).

% ?- member(X, Y, [a,b,c], Z).

member(X, Y, [X|Xs], [Y|Xs]).
member(X, Y, [A|Xs], [A|Ys]):- member(X, Y, Xs, Ys).

%%%
appear(X, X).
appear(X, Y):- Y=..[_|Y0], member(Z, Y0), appear(X, Z).

%
list([])    --> [].
list([X|Y]) --> [X], list(Y).

listcast(X,X) :- listp(X),!.
listcast(X,[X]).

% ?- basic:prepend(a, S, R).
%@ S = [a|R].
% ?- basic:prepend([a,[b,c]], S, R).
%@ S = [a, b, c|R] .

%%  prepend(?A:list_of_list, -X:list, -Y:list) is dete.
%	roughly works like
%		flatten(A, B), append(B, Y, X).

prepend(A, [A|X], X):- var(A), !.
prepend([], X, X).
prepend([A|R], X, Y) :- !, prepend(A, X, Z),
    prepend(R, Z, Y).
prepend(A, [A|X], X).

% a la <br>
polish_reserved(br, A, [0'\n|A]).   %'

%
save_bc(X) :- prolog_load_context(module, M),
	scan(`.\n`, A, M, X, _), !,
	herbrand(_, A, G0),
	pac:expand_arg(G0, M, G, P, []),
	maplist(assert, P),
	nb_setval(saved_bc, G).

% ?- expand_arg((split,maplist(pred([A, [A,A]]))),user, G, P, []), maplist(assert, P), phrase(G, [a], R).
%@ G = (user:split, user:maplist(user:'pred#3')),
%@ P = [ (user:'pred#3'(A, [A, A]):-true)],
%@ R = [[[a], [a]]] .

		/******************
		*     herbrand    *
		******************/

% ?- herbrand("f(X,a)", S), smash(S).
% ?- herbrand(S, f(X,a)).
% ?- herbrand(V, S, f(X,a)).
% ?- herbrand(V, "f(X,a)", T), herbrand(V, S, T).
% ?- herbrand(['X'= X], S, f(X,a)).
% ?- herbrand(V, `f(A)`, T), term_string(T, S, V).
% ?- herbrand([op(10, fx, ***)], V, `***a(X)`, R).
% ?- herbrand(V, `***a(X)`, R).

%
herbrand_in_context(Mod, X, Y):-
	(	string(X)	->	X0 = X
	;	string_codes(X0, X)
	),
	term_string(Y, X0, [variable_names(_),
						module(Mod)]).

% ?- basic: herbrand_in_context("a \\ \\ b", H).
herbrand_in_context(X, Y):-
	(	string(X)	->	X0 = X
	;	string_codes(X0, X)
	),
	term_string(Y, X0, [variable_names(_), module(pac_op)]).
%
herbrand(Mod, V_names, X, Y):-
	term_codes(Y, X, [variable_names(V_names), module(Mod)]).

%
herbrand(V_names, X, Y):-
	term_codes(Y, X, [variable_names(V_names)]).
%
herbrand(X, Y) :- herbrand(Eqs, X, Y),  maplist(call, Eqs).
%
herbrand0(X, Y) :- term_codes(Y, X, []).

%
herbrand_opp(X, Y)	:- herbrand(Y, X).
herbrand_opp(V, X, Y)	:- herbrand(V, Y, X).

%
save_current_op(op(_, A, Z), op(P0, A0, Z)):- current_op(P0, A0, Z),
	op_kind(A, Kind),
	op_kind(A0, Kind),
	!.
save_current_op(op(_, A, Z), op(0, A, Z)).

% an operator may be infix, prefix, and postfix.
op_kind(xfx,	infix).
op_kind(xfy,	infix).
op_kind(yfx,	infix).
op_kind(fx,	prefix).
op_kind(fy,	prefix).
op_kind(xf,	postfix).
op_kind(yf,	postfix).

% ?- basic:term_codes(f(a,X), C, [variable_names(['X'=X])]), basic:smash(C).

term_codes(X, Y, Options):- nonvar(Y), !,
	string_codes(S, Y),
	term_string(X, S, Options).
term_codes(X, Y, Options):- memberchk(variable_names(V), Options), !,
	equate(V),
	term_string(X, S, Options),
	string_codes(S, Y).
term_codes(X, Y, Options):-
	term_string(X, S, Options),
	string_codes(S, Y).

%
term_codes(X, Y):- term_codes(X, Y, []).

%
equate( []) :- !.
equate([X = X| R]) :- equate(R).

		/***************
		*     smash    *
		***************/

%%	smash(+X:text) is det.
%	Write all elements of text to the standard output
%	ignoring the null byte code.

% ?- smash(a).
% ?- smash(a+b).
% ?- smash(f([b,c])).
% ?- smash([b,c]).
% ?- smash([]).
% ?- smash('[]').

smash([]):- !.
smash([X|Y]):- !, smash(X), smash(Y).
% smash(0):- !.		% ignore  null bytes.
smash(X):- integer(X), !, put_code(X).
smash(X):- write(X).

%%	smashq(+X:text) is det.
%	Same as smasq, but use writeq/1 instead.

% ?- smashq(a).
% ?- smashq(a+b).
% ?- smashq(`a`).
% ?- smashq(f(`[b,c]`)).
% ?- smashq([b,c]).
% ?- smashq('[]').
% ?- smashq(`[]`).

smashq([]):- !.
smashq([X|Y]):- !, smashq(X), smashq(Y).
% smashq(0):- !.    % null byte
smashq(X):- integer(X), !, put_code(X).
smashq(X):- writeq(X).

%%	smash(+X:text, -Y:string) is det.
%	Same as smash/1, but unify Y with the output string instead.

% ?- smash(a, X).
% ?- smash("a", X).
% ?- smash(f("a",b), X).
% ?- smash([a,b,c], X).
% ?- smash(`abc`, X).

smash(X, Y):- smash_to_atomics(X, X0, []),
	atomics_to_string(X0, Y).

% ?- term_string0("ab", X).
term_string0(X, Y):-
	(	string(X) -> Y = X
	;	atom(X) -> atom_string(X, Y)
	;	term_string(X, Y)
	).

% ?- term_smash([], X).
% ?- term_smash([a], X).
% ?- term_smash([[a]], X).
% ?- term_smash(a+b, X).
% ?- term_smash("abc", X).
term_smash(X, Y):- pred_term_smash(X, Y, term_string).

% ?- term_smash0(['abc'], X).
% ?- term_smash0(["abc", def], X).
term_smash0(X, Y):- pred_term_smash(X, Y, term_string0).

%
:- meta_predicate pred_term_smash(?, ?, 2).
pred_term_smash([], "", _):-!.
pred_term_smash([X|Y], Z, F):-!, pred_term_smash(X, U, F),
	pred_term_smash(Y, V, F),
	string_concat(U, V, Z).
pred_term_smash(X, Y, F):- call(F, X, Y).

% ?- basic:smash_to_atomics([a,[],"b"+c], R,[]).
% ?- basic:smash_to_atomics([a,[97], b+c], R,[]).


smash_to_atomics(X, [X0|A], A):- var(X), !,
	term_string(X, X0).
smash_to_atomics([X|Y], A, B):- smash_to_atomics(X, A, C),
	smash_to_atomics(Y, C, B).
smash_to_atomics([], A, A).
% smash_to_atomics(0, A, A).   % unnecessary for utf8 encoding ?
smash_to_atomics(X, [Y|A], A)	:- compound(X), !, term_string(X, Y).
smash_to_atomics(C, [C0|A], A)	:- integer(C), !, char_code(C0, C).
smash_to_atomics(X, [X|A], A).

%%	smash_codes(+X:text, -Y:codes) is det.
%	Same as smash/2, but unify Y with the output codes instead.

% ?- smash_codes(a, X).
% ?- smash_codes([97,b,[c,d], f(a, b)], X), smash(X).

smash_codes(X, Y):- once(smash_to_codes(X, Y, [])).

%
smash_to_codes([], A, A).	% ground input is assumed.
smash_to_codes([X|Y], A, B):- smash_to_codes(X, A, C),
	smash_to_codes(Y, C, B).
% smash_to_codes(0, A, A).  % unnecessary for utf8 econding ?
smash_to_codes(C, [C|A], A):- integer(C).
smash_to_codes(X, A, B):- atom(X),
	atom_codes(X, X0),
	append(X0, B, A).
smash_to_codes(X, A, B):- string(X),
	string_codes(X, X0),
	append(X0, B, A).
smash_to_codes(X, A, B):- term_string(X, X0),
	string_codes(X0, X1),
	append(X1, B, A).

% ?- associative_comma((([], [a]), [b]), P, []).
% ?- associative_comma(((a,b), c), P, []).
associative_comma((X, Y), P, Q):-!, associative_comma(X, P, P0),
	associative_comma(Y, P0, Q).
associative_comma(end_of_file, P, P):-!.
associative_comma(X, [X|P], P).

% ?- comma_list((a,b), X), comma_list(Y, X).
% ?- comma_list(a, X), comma_list(Y, X).

% bi-dierectional.
comma_list(X, Y):- nonvar(X), !, associative_comma(X, Y, []).
comma_list(X, Y):- list_to_comma(Y, X).

%%	zip(?X:list, ?Y:list, ?Z:list) is det.
%	Zip X and Y into Z with comma (,).

zip([],_, []):-!.
zip(_, [], []):-!.
zip([A|X], [B|Y], [(A,B)|R]):- zip(X, Y, R).

zip(F, X, Y, Z):- maplist(F, X, Y, Z).

unzip([], [], []).
unzip([(A,B)|R], [A|X], [B|Y]):- unzip(R,X,Y).

% ?- different_terms([X,1]), X=2.
%@ X = 2.
%  declare a constraint that all terms in the list are different.
different_terms([X|Y]) :- maplist(dif(X), Y),
	different_terms(Y).
different_terms([]).

% tiny helpers
% enter(E,_,Y) :- pac:eval(E, Y).
peek(X,_,X).
peek(X,Y,X,Y).
current(X,X,X).
empty([],_).
empty_list(_, []).
car([A|_], A).
cdr([_|A], A).
cons(A, B, [A|B]).
singleton(A, [A]).
ordered_pair(A, B, (A, B)).
ordered_pair_p(A, B, p(A, B)).
return(A, B, [A|B]).		% cons <==> return !!
swap([A,B],[B,A]).
promote(A,[A]).
result(A,X) --> phrase(A), current(X). % result(A,X,Y,Y) :- act(A,Y,X).
id --> [].		%  X=Y
clear --> peek([]).
clear(X) --> peek(X,[]). %% <=> clear(X) --> current(X), clear.
fst((A,_), A).
snd((_,B), B).
pair(A, B, (A,B)).
pair(A, B, X, (A0, B0) ):- call(A, X, A0), call(B, X, B0).

			/******************
			*     splitting   *
			******************/

% ?- paragraph(`a\n\nbc`, X).
%@ X = [[97], [98, 99]] .
% ?- paragraph(`a\n\nbc\n`, X).
%@ X = [[97], [98, 99, 10]] .

paragraph(X, Y)		:-	split(plus, "\n\n", X, Y).
paragraph(E, X, Y)	:-	split(plus, E, X, Y).

% ?-basic:split_string_by_word("abadea", "a", X).
split_string_by_word(X, Y, Z):- string_length(Y, N),
		      split_string_by_word(X, Y, N, Z, []).

%
split_string_by_word(X,Y,K,[U|R],S):- sub_string(X,H,K,T,Y),!,
		      sub_string(X,0,H,_,U),
		      sub_string(X,_,T,0,V),
		      split_string_by_word(V, Y, K, R, S).
split_string_by_word(X,_,_,[X|R],R).

% ?- basic:split(" ", ` a b c `, X).
%@ X = [[], [97], [98], [99], []] .

split(X, Y)   :- split(=, `\n`, X, Y).

%
split(E, X, Y):- split(=, E, X, Y).

%
split(F, E, X, Y):- atom(E), !, atom_codes(E, E0),
	split(F, E0, Y, [], X, []).
split(F, E, X, Y):- string(E), !, string_codes(E, E0),
	split(F, E0, Y, [], X, []).
split(F, E, X, Y):- split(F, E, Y, [], X, []).

%
split(F, C, [A|X], Y) --> dot_star(A, []), split_rest(F, C, X, Y).
split(_, _, [[]|X], X, [], []).

%
split_rest(F, C, X, Y) --> delimiter(F, C), !, split(F, C, X, Y).
split_rest(_, _, X, X, [], []).

% ?- basic:delimiter(=, `a`, `a`, R).
% ?- basic:delimiter(=, `ab`, `ac`, R).
% ?- basic:delimiter(plus, `a`, `aaaac`, R).
% ?- basic:delimiter(plus, `aa`, `aaab`, R).
% ?- basic:delimiter(plus, `aa`, `ab`, R).

delimiter(=, C, X, Y):- !,  delimiter(C, X, Y).
delimiter(_, C, X, Y):- delimiter(C, X, X0), !,
	delimiter_plus(C, C, X0, Y).

%
delimiter_plus([], C, X, Y):- !, delimiter_plus(C, C, X, Y).
delimiter_plus([A|X], C, [A|Y], Z):- !, delimiter_plus(X, C, Y, Z).
delimiter_plus(_, _, X, X).

%
delimiter([A|X], [A|Y], Z):- delimiter(X, Y, Z).
delimiter([], X, X).

%
dot_star(X, X) --> [].
dot_star([A|X], Y)--> [A], dot_star(X, Y).

%%
repeat_chars(X)--> [C], {memberchk(C, X)}, repeat_chars(X).
repeat_chars(_)--> [].

remove_last(X,Y,Z) :- append(Z,[X],Y), !.
remove_last(_,Y,Y).

%
append_lists([L|R], X, Y):- append_one(L, X, X0),
	append_lists(R, X0, Y).
append_lists([], X, X).

append_one([A|R], [A|X], Y):- append_one(R, X, Y).
append_one([], X, X).

concat --> append.

fullstop  --> ".\n".

union(X, Y) :- append(X, X0), sort(X0, Y).

residue(X, Y, Z) :- append(X, Z, Y).

residue(X,Y,Z,U,V) :- append(X,V,U), append(X,Z,Y).

remove(X, Y, Z) :- delete(Y, X, Z).

% ?- basic:scan_prefix([b], [a,b,c], A, B).

scan_prefix(C, X, Y, Z):- once(scan_prefix(X, Z, C, Y, [])).
%
scan_prefix(R, S, C, A, A):- append(C, S, R).
scan_prefix([X|R], S, C, [X|A], B):-
	scan_prefix(R, S, C, A, B).
scan_prefix([], [], _, A, A).

% ?- basic:scan_escape_to(0'", `a\"bc"`, R, S, []), smash(R), nl, smash(S).
scan_escape_to(C, [C|R], R, [C|S], S).
scan_escape_to(C, [0'\\, A|X], X0, [0'\\, A|Y], Y0):-
	scan_escape_to(C, X, X0, Y, Y0).
scan_escape_to(C, [A|X], X0, [A|Y], Y0):-
	scan_escape_to(C, X, X0, Y, Y0).
scan_escape_to(_, X, X, Y, Y).

%
comment_begin_end(`%`, `\n`, `\n`).
% comment_begin_end(`//`, `\n`,`\n`).
comment_begin_end(`/*`, `*/`, []).

% ?- basic:remove_comment([a,b,c], X, []).
% ?- basic:remove_comment(`non comment%abc`, X, []), smash(X).
%@ X = [110, 111, 110, 32, 99, 111, 109, 109, 101|...] .
% ?- basic:remove_comment(`/***** abc ****/\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`/***** abc ****/xyz`, X, []), smash(X).
% ?- basic:remove_comment(`/**/\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`"/**/"\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`f("/**/")\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`f(0'\\\\, "/**/")\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`f(0'\\' , "/**/")\nxyz`, X, []), smash(X).
% ?- basic:remove_comment(`0'\\\\`, X, []), smash(X).
% ?- basic:remove_comment(`0'\\abc`, X, []), smash(X).
% ?- basic:remove_comment(`0'abc`, X, []), smash(X).
% ?- basic:remove_comment(`\\\\abc`, X, []), smash(X).

remove_comment(X,Y):- remove_comment(X, Y, []).

remove_comment([], X, X).
remove_comment(L, X, X0):-
	comment_begin_end(B, E, PushBack),
	append(B, L0, L),
	!,
	skip_to(E, L0, L1),
	append(PushBack, Y, X),
	remove_comment(L1, Y, X0).
remove_comment([Q|L], [Q|X], Y):- memq(Q,`"\`'`), !, % "
	scan_escape_to(Q, L, L0, X, X0),
	remove_comment(L0, X0, Y).
remove_comment([0'0, 0'\', A|L], [0'0, 0'\', A|X], Y):- !,
	(A == 0'\\				% '
	 ->		L=[B|L0],
			X=[B|X0],
			remove_comment(L0, X0, Y)
	;	remove_comment(L, X, Y)).
% remove_comment([0'0, 0'\', A|L], [0'0, 0'\', A|X], Y):- !,  %'
% 	remove_comment(L, X, Y).
remove_comment([A|L], [A|X], Y):-
	remove_comment(L, X, Y).

%
skip_to(E, X, Y):- append(E, Y, X).
skip_to(E, [_|X], Y):- skip_to(E, X, Y).
skip_to(_,[],[]).

:- meta_predicate delete_all(1,?,?).

delete_all(C, [X|Y], Z):- call(C, X), !, delete_all(C, Y, Z).
delete_all(C, [X|Y], [X|Z]):- delete_all(C, Y, Z).
delete_all(_, [], []).

% ?-listsubst([([a],[b])],[a,a,a], R).
%@ R = [b,b,b]

listsubst(_, [], []).
listsubst(X, Y, Z) :- member((A,B), X),
	append(A,Y1,Y),
	append(B,Z1,Z),
	listsubst(X, Y1, Z1).
listsubst(X, [A|Y], [A|Z]) :- listsubst(X,Y,Z).

% %%% Comment and Quotation

% comment(X) --> sandwich("/*", "*/", X)
% 	; sandwich("//", "\n", X)
% 	; sandwich("%", "\n", X).

%%%%
% by W. Jan [2009/02/08]
% full-stop
% ?- basic:atom_to_term('a.\n', X, Y, user).
% ?- basic:atom_to_term('1.\n', X, Y, user).

atom_to_term(Atom, Term, Bindings, Module) :-
	atom_to_memory_file(Atom, MF),
	open_memory_file(MF, read, Stream,
			 [ free_on_close(true)
			 ]),
	call_cleanup(read_term(Stream, Term,
			       [ variable_names(Bindings),
				 module(Module)
			       ]),
		     close(Stream)).

:- meta_predicate collect(:,?,?).
% ?- collect(=(1), [1,2,1,3], X).
% X = [1, 1]
% ?- collect(pred([X]:- atom_concat(_,pdf,X)), [a,b,xpdf,ypdf, c], L).
collect(F, X, Y):- collect_(X, Y, F).
%
collect_([X|R0], [X|R], F):- call(F, X), !, collect_(R0, R, F).
collect_([_|R0], R, F):- collect_(R0, R, F).
collect_([], [], _).

:- meta_predicate collect_files(1, ?, ?).

collect_files(Filter, Dir,Ls):-
	directory_files(Dir, L0),
	collect(Filter, L0, Ls).

:- meta_predicate image(3,?,?,?).
% ?- image([X,X,X], [1,2,3],[1,3,3], L).
% L = [1,3].
image(F, [X|R0], [Y|S0], U):- (call(F, X, Y, Z) ->  U=[Z|T0]; U=T0),
	image(F, R0, S0, T0).

%
dual(X, Y) :-  call(Y, X).
dual(X, Y, Z):- call(Y, X, Z).

:- meta_predicate rev(2,?,?), rev(3,?,?,?).
rev(F, X, Y) :- call(F, Y, X).
rev(F, X, Y, Z) :- call(F, X,  Z, Y).

inverse(F, X, Y) :- call(F, Y, X).
inverse(F, X, Y, Z) :- call(F, X, Z, Y).

% /**** course of values
% ?- course_of_values(n\l\n,  10, [1,1], X).
% X = [9, 8, 7, 6, 5, 4, 3, 2, 1|...].
% ?- eval(course_of_values(n\l\
%	(a\ (is:: ( a * n))@ car(l)), 10, [1,1]), X).
% X = [362880, 40320, 5040, 720, 120, 24, 6, 2, 1|...].
% ?- eval(course_of_values(n\fun([a],
%	is::( a * n)), 10, [1,1]), X).
% X = [362880, 40320, 5040, 720, 120, 24, 6, 2, 1|...].
% ****/
course_of_values(F, J, X, Y):-  cov(F, 1, J, X, Y).

course_of_values(F, I, J, X, Y):-  cov(F, I, J, X, Y).

% ?- cov(n\ fun([a], is:: (a*n)), 1, 5, [1, 1], X).
% X = [24, 6, 2, 1, 1, 1].
% ?- cov(n\ fun([], factorial(n,1)), 10, [1], X).
% X = [362880, 40320, 5040, 720, 120, 24, 6, 2, 1|...].

cov(F, J, X, Y) :- cov(F, 1, J, X, Y).

cov(_, N0, N, X, X):- N0 >= N, !.
cov(F, J,  N, X, Y):- eval:apply(F, [J, X], X0),
	J1 is J + 1,
	cov(F, J1, N, [X0|X], Y).

%%   fun/4   syntax sugar  for course of values induction
% ?- fun([x], x, [1,2,3], A).
% A = 1.
% ?- fun([x,y], set::(x+y), [[a,b], [c,d], [e]], X).
% X = [a, b, c, d].

% fun(P, F, X, A):- fresh_bind(P\F, P0\F0), append(P0, _, X), eval(F0,
% A).

% ?-  list_rec(fun([a,b], (is::(a+b))), 10, [1,1], X).
% X = [144, 89, 55, 34, 21, 13, 8, 5, 3|...].

list_rec(_, J, X, X):- J =< 0, !.
list_rec(F, J, X, Y):- eval:(F, [X], V),
	J1 is J - 1,
	list_rec(F, J1, [V|X], Y).

% /***** nat_list_rec/[4,5]
% ?- nat_list_rec([j,l]\ ([x1, x2]\
%	(is:: (j+1)*x1*x2) @ nth1(1,l) @ nth1(2,l)), 5, [1,1], X).
% X = [1440, 48, 6, 2, 1, 1, 1].
% ?- nat_list_rec([j,l]\ some([x, y, r], (l=[x, y|r],
%	is:: (j+1)*x*y)), 5, [1,1], X).
% X = [1440, 48, 6, 2, 1, 1, 1].
% *******/

nat_list_rec(F, N) --> nat_list_rec(F, 0, N).

nat_list_rec(_, N, N, X, X):- !.
nat_list_rec(F, J, N, X, Y):-
	call(F, J, X, V),
	J1 is J + 1,
	nat_list_rec(F, J1, N, [V|X], Y).


%% variant(+F:pred/3, +X:term, -Y:term, -Z:term) det
%   is true if  so is call(F, Y, X, Z).

:- meta_predicate variant(3, ?, ?, ?).
% ?- variant(append, [a,b], [c, d], X).
% X = [c, d, a, b].
variant(X, Y, Z, U):- call(X, Z, Y, U).

% ?- trace, eval(args(ev)::append(append([a,b],[c,d]), [e,f]), X).
% X = append([a, b, c, d], [e, f]).
% args(S, X, Y, [], [], true):- mapterm(eval(S), X, Y).

%
with_order(Ord, G, X, Y):- sort(Ord, Ord0),
	zip(Ord, Ord0, A),
	subst(A, X, X0),
	call(G, X0, Y0),
	zip(Ord0, Ord, B),
	subst(B, Y0, Y).

% ?- flatten(;, (a;b); c , X).
% X = [a, b, c]

flatten(F, X, Xs):- flatten(F, X, Xs, []).

flatten(F, X, A, B):- X=..[F, X1, X2], !,
	flatten(F, X1, A, A0),
	flatten(F, X2, A0, B).
flatten(_, X, [X|A], A).

% ?- basic:flatten_dl([a,[b,c]], X, []).
flatten_dl([A|B], X, Y):- flatten_dl(A, X, Z),
	flatten_dl(B, Z, Y).
flatten_dl([], X, X).
flatten_dl(A, [A|X], X).

% ?- flatten_more(f(a,[f(b,c)]), X).
% X = [f, a, f, b, c].

flatten_more(A, [A|X], X):-var(A),!.
flatten_more([], X, X):-!.
flatten_more(A,[A|X], X):-  atomic(A),!.
flatten_more([A|B], X, Y):- !, flatten_more(A, X, X0),
	flatten_more(B, X0, Y).
flatten_more(A, [F|X], Y):- A =.. [F|As],
	flatten_more(As, X, Y).

flatten_more(A, X):- once(flatten_more(A, X, [])).

% ?- tree_path_set(f(1,g(2,3)), X).
% X = [[f,1],[f,g,2],[f,g,3]].

tree_path_set(X, Y):- compound(X), !, X=..[F|A],
	maplist(tree_path_set, A, B),
	append(B, B0),
	maplist(cons(F), B0, Y).
tree_path_set(X, [[X]]).

% push & pop
% ?- stack_push(1), stack_push(2), stack_pop(X), stack_pop(Y).
% X = 2,
% Y = 1.

stack_init(S) :- nb_setval(S, []).
stack_push(X, S):- nb_getval(S, R), nb_linkval(S, [X|R]).
stack_pop(X, S):- nb_getval(S, [X|R]),  nb_linkval(S, R).
stack_update(X, S):- nb_getval(S, [_|R]),  nb_linkval(S, [X|R]).
stack_top(X, S):- nb_getval(S, [X|_]).

stack_init :- stack_init('$STACK').
stack_push(X):- stack_push(X, '$STACK').
stack_pop(X):- stack_pop(X, '$STACK').
stack_update(X):- stack_update(X, '$STACK').
stack_top(X):- stack_top(X, '$STACK').

%%	scan(+A:list/atomic, ?B:list, ?C:list, +X:list, -Y:list) is nondet.
%	True if  X =  (B-C) + A0 + Y, where A0 is the list of codes of A
%	if A is atomic, otherwise A0 = A; and
%	'+' operator means the list concatenation.
% ?- basic:scan([a,b],S,[],[1,2,a,b,3,4], R).
%@ S = [1, 2],
%@ R = [3, 4] ;

% ?- basic:scan(`.\n`,S,[],`aaaaaa.\n==\n`, R).
% ?- basic:scan(".\n",S,[],`aaaaaa.\n==\n`, R).

%@ true.
%%  first_token_codes(+L:codes, -X:codes) is det.
%	True if  X is unified with the maximum first codes block
%	that has no white codes.

scan(A, S, T, X, Y):- atomic(A), !,
	atom_codes(A, B),
	scan_codes(B, S, T, X, Y).
scan(A, S, T, X, Y):- scan_codes(A, S, T, X, Y).

%
scan_codes(A, S, S, X, Y):- append(A, Y, X).
scan_codes(Key, [A|As], U, [A|Xs], Y):- scan_codes(Key, As, U, Xs, Y).
%
fill  --> chars(` \t\r\n`).

filler(_) --> chars(` \t`). % for compatibility


% Tiny DCG
maybe_end --> chars(` \t`), peek([], end); herbrand(_).

kleene_star(_, P, P).
kleene_star([X|Xs], [X|P], Q):-  append(Xs, P0, P),
	kleene_star([X|Xs], P0, Q).

%
kleene_star_greedy([X|Xs], [X|P], Q):-  append(Xs, P0, P),
	kleene_star_greedy([X|Xs], P0, Q).
kleene_star_greedy(_, P, P).

%
kleene_plus_greedy(X, Y, Z):- append(X, Z0, Y),
	kleene_star_greedy(X, Z0, Z).

kleene_plus_greedy(X, Y, Z):- append(X, Z0, Y),
	kleene_star_greedy(X, Z0, Z).

kleene_string_codes(S, C, D):- string_codes(S, Cs),
	kleene_plus_greedy(Cs, C, D).

%
chars(_) --> [].
chars(L) --> [X], {memberchk(X, L)}, chars(L).
%
chars(L) --> [X], {memberchk(X, L)}, chars(L).
chars(_) --> [].

% Idea by @brebs at Discourse.

% ?- bi_reverse([a,b], [], X).
% ?- bi_reverse(X, Y, [b,a]).
bi_reverse(X, Y, Z):-
    bi_reverse(X, Y, Z, Z).
%
bi_reverse([], Z, _, Z).
bi_reverse([A|X], Y, [_|Z0], Z) :-
    bi_reverse(X, [A|Y], Z0, Z).

% ?- predsort(ahead_compare([a,b]), [b,a,a,b,c], X).
% ?- predsort(ahead_compare([a,c,b]), [b,a,a,b,c], X).
% ?- predsort(ahead_compare([a,c,b]), [b,a,a,b,c], X).
% ?- predsort(ahead_compare([a,b,c]), [w, u,v], X).
%
% all elements not in the first list are identified with the first one.
%
ahead_compare(_,  =, X, X):-!.
ahead_compare([], =, _, _):-!.
ahead_compare([X|_], <, X, _):-!.
ahead_compare([Y|_], >, _, Y):-!.
ahead_compare([_|List], C, X, Y):- ahead_compare(List, C, X, Y).
