:- module(iboole, [i_boole/2]).
% ?- iboole:i_normal([inf-sup, sup-inf, 3], R).
% ?- iboole:i_normal([3], R).
% ?- iboole:i_normal([3-2], R).
% syntax checking of intervals.

i_normal(X, Y):- i_normal(X, X0, []),
	predsort(i_compare, X0, X1),
	i_merge(X1, Y).

%
i_normal([], P, P).
i_normal([X|Y], P, Q):- i_check(X, P, R), i_normal(Y, R, Q).

%
i_check(X, _, _):- var(X), throw(iterval(variable(X))).
i_check(X-Y, P, Q):- nonvar(X), nonvar(Y), i_end(X), i_end(Y), !,
	i_check(X, Y, P, Q).
i_check(X, [X-X|P], P):- integer(X), !.
i_check(X, _, _):- throw(interval(unknown(X))).

%
i_check(inf, inf, P, P).
i_check(sup, sup, P, P).
i_check(X, Y, P, Q):- i_add_one(X-Y, P, Q).

%
i_end(sup).
i_end(inf).
i_end(X):- integer(X).

%
i_compare(C, X-Y, X-Z)  :- !, x_compare(C, Y, Z).
i_compare(C, X-_, Y-_)  :- x_compare(C, X, Y).

%
x_compare(=, X, X).
x_compare(<, _,sup).
x_compare(>, sup,_).
x_compare(>, _,inf).
x_compare(<, inf,_).
x_compare(C, X, Y):- compare(C, X, Y).

%
i_less(inf, X):- X\== inf.
i_less(X, sup):- X\== sup.
i_less(X, Y):- integer(X), integer(Y), X<Y.

%
i_max(_, sup, sup).
i_max(sup, _, sup).
i_max(inf, X, X).
i_max(X, inf, X).
i_max(X, Y, X):- X > Y.
i_max(_, Y, Y).

%
i_min(X, sup, X).
i_min(sup, X, X).
i_min(_, inf, inf).
i_min(inf, _, inf).
i_min(X, Y, X):- X<Y,!.
i_min(_, Y, Y).

%  ?- iboole:i_on(-2, inf- sup).

i_on(X, A-B):-   (A==inf; A =< X), !, X @=< B.


% ?- iboole:i_succ(1,X).
% ?- iboole:i_succ(X,2).

i_succ(sup, sup).
i_succ(inf, inf).
i_succ(X, Y):- plus(X, 1, Y).

%%% Boolean operations on sets of intergers represented
% as a list of characters
% and intervals of characters, e.g, [a, c-k, u, x-z].
% ?-  iboole:i_boole([^, 3-5],  R).
% ?-  iboole:i_boole(\+ ([3-5]),  R).
% ?-  iboole:i_boole([3-5];[4-8],  R).
% ?-  iboole:i_boole([1-10]&([3-5]|[4-8, 8-12]),  R).
%@ R = [3-10] .


i_boole(normal(X), Y):- i_normal(X, Y).
i_boole(dot(X), Y):- i_boole(X, Y).
i_boole(out(X), Y):- i_boole(X, X0), i_neg(X0, Y).
i_boole(^(X), Y):- i_boole(out(X), Y).
i_boole(\+(X), Y):- i_boole(out(X), Y).
i_boole((X|Y), Z):- i_boole(X, X0),
	i_boole(Y, Y0),
	i_cup(X0, Y0, Z).
i_boole((X;Y), Z):- i_boole((X|Y), Z).
i_boole(&(X,Y), Z):- i_boole(X, X0),
	i_boole(Y, Y0),
	i_cap(X0, Y0, Z).
i_boole(\(X,Y), Z):- i_boole(X, X0),
	i_boole(Y, Y0),
	i_subtract(X0, Y0, Z).
i_boole(X, X).

%
i_add([], P, P).
i_add([X|Xs], P, Q):- i_add_one(X, P, R),!,
	i_add(Xs, R, Q).

%
i_add_one(X-Y, P, P):- integer(X), integer(Y), Y < X.
i_add_one(_-inf, P, P).
i_add_one(sup-_, P, P).
i_add_one(X, [X|P], P).

% ?- i_cap([1-10],[3-6, 8-8], R).
% R = [3-6,8-8] 

i_cap(X, Y, P):- i_cap(X, Y, P, []).

%
i_cap([], _, P, P).
i_cap(_, [], P, P).
i_cap([X|Xs],[Y|Ys], P, Q):- i_cap(X, Y, NXs, Xs, NYs, Ys, P, R), !,
	i_cap(NXs, NYs, R, Q).

%
i_cap(LX-RX, LY-RY, P, Q, R, S, T, U):-
	i_max(LX, LY, Min),
	i_min(RX, RY, Max),
	i_succ(RY, Y0),
	i_succ(RX, X0),
	i_add_one(Y0-RX, P, Q),
	i_add_one(X0-RY, R, S),
	i_add_one(Min-Max, T, U).

%  union of two intervals sets.
% i_cup([1-5, 7-10], [6-6], R).
% R = [1-10] 

i_cup([], Y, Y).
i_cup(X, [], X). 
i_cup(X, Y, Z):- select_smaller(S, X, Y, X0, Y0),
	i_cup(S, X0, Y0, Z).

%
i_cup(X, [], Y, Z):- i_cup_one(X, Y, Z).
i_cup(X, Y, [], Z):- i_cup_one(X, Y, Z).
i_cup(X, As, Bs, Z):- select_smaller(Y, As, Bs, Cs, Ds),
	i_cup(X, Y, Cs, Ds, Z).

%
i_cup(X-X0, Y-Y0, As, Bs, [X-X0|Z]):-  i_off(X0, Y), !,
	i_cup(Y-Y0, As, Bs, Z).
i_cup(X-X0, _-Y0, As, Bs, Z):- i_max(X0, Y0, M),
	i_cup(X-M, As, Bs, Z).

%
i_cup_one(X, [], [X]).
i_cup_one(X-X0, [Y-Y0|Ys], [X-X0, Y-Y0|Ys]):- i_off(X0, Y), !.
i_cup_one(X-X0, [_-Y0|Ys], Z):- i_max(X0, Y0, M),
	i_cup_one(X-M, Ys, Z).

% ?- i_subtract([1-10],[3-6], R).
% R = [1-2,7-10]
%  Remark:  i_subtract(X,Y,Z):- i_neg(Y, Y0), i_cap(X, Y0, Z).
i_subtract(X, Y, Z):- i_subtract(X, Y, Z, []).

%
i_subtract([], _, P, P).
i_subtract(X, [], P, Q):- append(X, Q, P).
i_subtract([LX-RX|Xs], [LY-RY|Ys], P, Q):- 
        i_succ(RY, RY0),
	i_succ(LY0, LY),
	i_succ(RX, RX0),
	i_add_one(LX-LY0, P, R),
	i_add_one(RY0-RX, NXs, Xs),
	i_add_one(RX0-RY, NYs, Ys),
	i_subtract(NXs, NYs, R, Q).

% computing the complement of intervals.
% boole:  ?- i_neg([3-5, 7-9], R).
% R = [inf-2,6,10-sup] 

i_neg(X, Y):- i_neg(inf, X, Y, []).

i_neg(W, [], P, Q):- i_add_one(W-sup, P, Q).
i_neg(W, [X-Y|Z], P, Q):-
	i_succ(X0, X),
	i_succ(Y, Y0),
	i_add_one(W-X0, P, R),
	i_neg(Y0, Z, R, Q).


% ?- iboole:i_merge([10-sup, inf- 9], X).
%@ X = [10-sup] .
% ?- iboole:i_merge([9-9, 10-10], X).
%@ X = [9-10] .

i_merge([], []).
i_merge([X], [X]).
i_merge([X-X0, Y-Y0|Z], [X-X0|U]):- i_off(X0, Y), !, 
	i_merge([Y-Y0|Z], U).
i_merge([X-X0, _-Y0|Z], U):- i_max(X0, Y0, M),
	i_merge([X-M|Z], U).


% ?- i_partition([1-3], [2-2], X).
% X = [1-1,2-2,3-3] 
% ?- iboole:i_partition([1-3], [2-4], X).
%@ X = [1-1, 2-3] .
% ?- iboole:i_partition([1-3], [2-4], X).
% ?- iboole:i_partition([inf-sup], [-3 - -3, -2 - -2, -1 - -1],  X).
%@ X = [inf- -4, -3- -3, -2- -2, -1- -1, 0-sup] .

i_partition(X, Y, Z):- i_partition(X, Y, Z, []).

%
i_partition([], _, P, P).
i_partition(X, [], P, Q):- append(X, Q, P).
i_partition([LX-RX|Xs], [_-RY|Ys], P, Q):-  i_less(RY, LX), !,
	i_partition([LX-RX|Xs], Ys, P, Q).
i_partition([LX-RX|Xs], [LY-RY|Ys], [LX-RX|P], Q):-  i_less(RX, LY), !,
	i_partition(Xs, [LY-RY|Ys], P, Q).
i_partition([LX-RX|Xs], [LY-RY|Ys], P, Q):-  
	i_max(LX, LY, L),
	i_min(RX, RY, R),
	i_succ(L0, L),
	i_add([LX-L0, L-R], P, P0),
	i_succ(RX, RX0),
	i_add_one(RX0-RY, NYs, Ys),
	i_succ(RY, RY0),
	i_add_one(RY0-RX, NXs, Xs),
	i_partition(NXs, NYs, P0, Q).

% ?- m_partition([[1-5],[3-6]],R).
% R = [[1-2,3-5],[3-5,6-6]] 
% ?- iboole:m_partition([[1-2, 3-5],[3-6]],R).
%@ R = [[1-2, 3-5], [3-5, 6-6]] .
% ?- iboole:m_partition([[1-2, 4-5],[3-6]],R).
% ?- iboole:m_partition([[1-2, 3-5],[inf-5, 8-sup]],R).
%@ R = [[1-2, 3-5], [inf-0, 1-2, 3-5, 8-sup]] .

m_partition([], []).
m_partition([X|Y], [X0|Z]):- m_partition(Y, Z0),
	m_partition(X, Z0, X0, Z).

m_partition(X, [], X, []).
m_partition(X, [Y|Z],  X1, [Y0|V]):-
	m_partition_one(X, Y, X0, Y0),
	m_partition(X0, Z, X1, V).

%% dividing each other 
% ?- m_partition_one([1-5], [2-7], X, Y).
% X = [1-1,2-5],
% Y = [2-5,6-7] 

m_partition_one(X, Y, Z, U):- m_partition_one(X, Y, Z, [], U, []).

%
m_partition_one([], Y, P, P, Q, R):- append(Y, R, Q).
m_partition_one(X, [], P, Q, R, R):- append(X, Q, P).
m_partition_one([LX-RX|Xs], [LY-RY|Ys], P, Q, [LY-RY|U], V):- i_less(RY, LX), !,
	m_partition_one([LX-RX|Xs], Ys, P, Q, U, V).
m_partition_one([LX-RX|Xs], [LY-RY|Ys], [LX-RX|P], Q, U, V):- i_less(RX, LY), !,
	m_partition_one(Xs, [LY-RY|Ys], P, Q, U, V).
m_partition_one([LX-RX|Xs], [LY-RY|Ys], P, Q, U, V):- i_max(LX, LY, L),
	i_min(RX, RY, R),
	i_succ(L0, L),
	i_add([LX-L0, L-R], P, P0),
	i_succ(RX, RX0),
	i_add_one(RX0-RY, NYs, Ys),
	i_succ(RY, RY0),
	i_add_one(RY0-RX, NXs, Xs),
	i_add([LY-L0, L-R], U, U0),
	m_partition_one(NXs, NYs, P0, Q, U0, V).
							
%
i_off(X, Y):- integer(X), integer(Y), !, Y > X+1.
i_off(inf, X):- !, X\==inf.
i_off(X, sup):- X\==sup.

%
select_smaller(C, [A|As], [B|Bs], X, Y) :-  i_compare(P, A, B),
	  (  P == (=) -> C=A, X=As, Y=Bs
	  ;  P == (<) -> C=A, X=As, Y=[B|Bs]
	  ;  C = B, X = [A|As], Y = Bs
	  ).

%% extras:  relation to function, or making fibired structure
% ?- iboole:keymerge([a-1,a-2,b-3,c-4,c-5], R).
% R = [a-[1,2],b-[3],c-[4,5]] 

keymerge([], []).
keymerge([A-X|R],  [A-[X|P]|R0]):- keymerge(A, R, P, [], S),
	keymerge(S, R0).
%
keymerge(A, [A-X|R], [X|P], Q, S):- keymerge(A, R, P, Q,  S).
keymerge(_, R, P, P, R).

%
% ?- iboole:keymerge_r([1-a,2-a,3-b,4-c,5-c], R).
% R = [[1,2]-a,[3]-b,[4,5]-c] 

keymerge_r([], []).
keymerge_r([X-A|R],  [[X|P]-A|R0]):- keymerge_r(A, R, P, [], S),
	keymerge_r(S, R0).
%
keymerge_r(A, [X-A|R], [X|P], Q, S):- keymerge_r(A, R, P, Q,  S).
keymerge_r(_, R, P, P, R).

