:- module(pac_listing, [expand_pac_text/1,
					 expand_pac/1,
					 expand_goal/1,
					 expand_exp/1,
					string_to_terms/2,
					compile_pac/3,
					pred_grouping/2,
					clause_to_string/3,
					is_backquote_begin/2,
					is_backquote_end/2
						]).

% :- dynamic '$algebra'/2.
% :- discontiguous '$algebra'/2.
:- use_module(pac('odict-attr')).
:- use_module(pac('odict-expand')).
:- use_module(pac(reduce)).
:- use_module(pac('pac-aux')).
:- use_module(pac('expand-pac')).


term_expansion --> pac:expand_pac.

:- op(8,	fy,		user:('`')).
:- op(10,	fy,		user:(:)).
:- op(10,	fy,		user:(*)).
:- op(10,	fy,		user:(?)).
:- op(10,	fy,		user:(@)).
:- op(10,	fy,		user:(#)).
:- op(60,	yfx,	user:(@)).
:- op(750,	yfx,	user:(&)).
:- op(1200,	xfx,	user:(-->>)).
:- op(650,	xfy,	user:(::)).
:- op(1050, xfy,	user:(\)).
:- op(1105, xfy,	user:('|')).
:- op(450,	xfx,	user:(..)).
:- op(710,	fy,		user:(~)).

% (setq module-query  "qcompile(pac('pac-listing')),  module(pac_listing).")
% ?- qcompile(pac('pac-listing')),  module(pac_listing).

% ?- show(call(pred([]))).
% ?- pac_listing:(expand_pac(a(X) :- call(pred([hello]), X))).
% ?- pac_listing:expand_pac(a --> w(".*")).
% ?- expand_pac((trim_white(A) --> wl("[ \t]*"), w(".*", A), wl("[ \t]*"))).
% ?- cgi_demo(`a-->w(".*").\n`, Y).
% ?- cgi_demo(`a.\n`, Y).

memq(X, [Y|_]):- X==Y, !.
memq(X, [_|R]):- memq(X, R).

%
cgi_demo(X, Y):-
	nb_setval(pac_name_prefix, pac),
	nb_setval(nt_name_prefix, nt),
	ejockey:handle([expand, pac], X, Y0),
	flatten(Y0, Y1),
	atomics_to_string(Y1, Y).
%
expand_pac(X)	:-
	compile_pred_word(X, [], H0, R0),
	smash(["\n", H0, ".\n\n", R0, "\n"]).

%
expand_exp(X)	:- show_exp(X).
%
expand_goal(X)	:- show(X).

smash(X):- basic:smash(X).

		/*************************************************
		*     Demo :  Compile clause / DCG with regex    *
		*************************************************/

% ?- nb_setval(nt_name_prefix, nt).
% ?- nb_setval(pac_name_prefix, pac).
% ?- expand_pac_text("zip(P,Q):-maplist(pred([X,Y,X-Y]), P, Q)").
% ?- expand_pac_text("a :- call(pred([X]))").
% ?- expand_pac_text("a --> b").
% ?- expand_pac_text('a --> w(".*")').
% ?- expand_pac_text('trim_white(A) --> wl("[ \t]*"), w(".*", A), wl("[ \t]*")').
% ?- expand_pac_text('trim_white(A) --> wl("[ \t]*"), w(".*", A), wl("[ \t]*")').
% ?- expand_pac_text('a(`(b)`)').
% ?- qcompile(util(zdd)), qcompile(util('emacs-jockey')), module(pac_listing).
% ?- enter_back_quotes.
% ?- exit_back_quotes.

expand_pac_text(S):-
	term_string(X, S, [variable_names(Eqs)]),
	compile_pred_word(X, Eqs, H0, R0),
	smash(["\n", H0, ".\n\n", R0, "\n"]).
%
expand_pac_text_symbol_char(S):-
	with_backquote_symbol_char(
		(	term_string(X, S, [module(symbol_char),
					   variable_names(Eqs)]),
			compile_pred_word(X, Eqs, H0, R0),
			smash(["\n", H0, ".\n\n", R0, "\n"])
		)).

% ?- new_names([X,Y], Eqs, 1, 'A', [a1,a2,a3]).
new_names([V|Vs], [A=V|Eqs], N, Prefix, As):-
	new_name(N, As, A, Prefix, K),
	new_names(Vs, Eqs, K, Prefix, As).
new_names([], [], _, _, _).

%
new_name(N, As, B, Prx, K):- atom_concat(Prx, N, B),
						\+ memberchk(B, As),
						!,
						succ(N, K).
new_name(N, As, A, Prx, K):- succ(N, N1),
						new_name(N1, As, A, Prx, K).

% ?- subtractq([X,Y,X,Y,X,Y], [X], R).
subtractq([], _, []).
subtractq([A|As], B,  C):- memq(A, B), !,
	subtractq(As, B, C).
subtractq([A|As], B, [A|C]):- subtractq(As, B, C).

% ?- pac_listing:expand_clause_slim(a(X):-call(pred([1]),X), Y).
expand_clause_slim(X, [H|Y]):-
	anti_subst:expand_clause(X, [], Y0, []),
	maplist(pred(&([X:-true, X], [C, C])),  Y0, [H|Y1]),
	maplist(copy_term, Y1, Y).

% ?- compile_pred_word(a(X):-call(pred([1]),X), ['X'=X], H, R).
compile_pred_word(Clause, Eqs, H0, R0):-!,
		maplist(pred([A=P, A, P]), Eqs, As, Vs),
		expand_clause_slim(Clause, [H|R]),
		term_variables(H, HVs),
		subtractq(HVs, Vs, SVs),
		new_names(SVs, Eqs0, 1, 'A', As),
		append(Eqs0, Eqs, Eqs1),
		term_string(H, H0, [module(pac_op),
							variable_names(Eqs1),
							quoted(true)]),
		maplist(pred(([U, [V,"\s.\n"]] :-
								numbervars(U, 0, _, [singleons(true)]),
								term_string(U, V, [ module(pac_op),
													numbervars(true),
													quoted(true)]))),
				R, R0).

		/**********************************
		*     format clause for output    *
		**********************************/

% ?- clause_to_string(p(a(X):-b(X), ['X'=X], [b(Y)]), C, H).
clause_to_string(p(X, Eqs, H), Z, H0):-
		maplist(pred([A=P, A, P]), Eqs, As, Vs),
		term_variables([X|H], HVs),
		subtractq(HVs, Vs, SVs),
		new_names(SVs, Eqs0, 1, 'A', As),
		append(Eqs0, Eqs, Eqs1),
		(	X ==[]
		->  Z = []
		;	term_string(X, X0, [module(pac_op),
								variable_names(Eqs1),
							quoted(true)]),
			Z = [X0,"\s.\n"]
		),
		maplist(pred(([U, [V,"\s.\n"]] :-
								numbervars(U, 0, _, [singleton(true)]),
								term_string(U, V, [ module(pac_op),
													numbervars(true),
													quoted(true)]))),
				H, H0).
%
clause_to_string(X, Y, Z):- clause_to_string(p(X, [], []), Y, Z).


		/***************************
		*     compile pac block    *
		***************************/

unify(Eqs):- maplist(pred([A=A]), Eqs).

% ?- qcompile(util(zdd)), qcompile(util('emacs-jockey')), module(ejockey).
% ?- trace, pac_listing:compile_pac([(a:- maplist(=, []))-[]], P, []).



% ?- pac_listing:compile_pac([(a := [])-[], (d:=(a + a))-[]], X, []).

% ?- pac_listing:compile_pac([(f(X):-call(pred([a]), X))-[], (:-bekind(a,[]))-[], (a+b=1)-[], (a*b = c)-[],(:-ekind)-[], z-[]], P, []).
% ?- pac_listing:compile_pac([(:-bekind(a,[]))-[], (a+b=1)-[], (:-ekind)-[]], P, []).
% ?- pac_listing:compile_pac([(:-bekind(a,[]))-[], (a+b= '`'(c))-[], (:-ekind)-[]], P, []).
% ?- pac_listing:compile_pac([(:-bekind(a,[]))-[], (a+b= c@1)-[], (:-ekind)-[]], P, []).
% ?- pac_listing:compile_pac([(:-bekind(a,[]))-[], (:-ekind)-[]], P, []).
% ?- pac_listing:compile_pac([(:-betrs(a))-[], (x=y)-[], ((x=y):-z=u) - [], (:-etrs)-[]], P, []).
% ?- pac_listing:compile_pac([(:-betrs(a))-[], (x=y)-[], ((x=y):-z=u) - [], (:-etrs)-[]], P, []).
% ?- pac_listing:compile_pac([(a:=[b/2-f])-[]], P, []).
% ?- pac_listing:compile_pac([(a:=[b/0-f])-[]], P, []).

compile_pac(X, [p([],[],R)|P], Q) :- collect_sgn(X, Y, Zip), !,
	expand_sgn_defs(Zip, E, R, []),
	compile_pac(Y, E, P, Q).
%
compile_pac([X-Eqs|Xs], D, P, Q) :-
	compile_pac(X, Eqs, D, Xs, Ys, P, P0),
	!,
	compile_pac(Ys, D, P0, Q).
compile_pac([], _, P, P).

% ?- pac_listing:expand_sgn_defs([(a:=[b-c])-U], U, X, []).
% ?- pac_listing:expand_sgn_defs((a:=[a-pred([c])])-U, X).
expand_sgn_defs([], [], X, X).
expand_sgn_defs([(K:=L)-(K:=L0)|R], [K-L0|S], P, Q):-
	expand_sgn_term(L, L0, P, P0),
	expand_sgn_defs(R, S, P0, Q).

%
expand_sgn_term(L, L0, P, Q):-
	(	is_list(L)
	->	expand_sgn_term_list(L, L0, P, Q)
	;	L=..[F|As],
		expand_sgn_terms(As, Bs, P, Q),
		L0=..[F|Bs]
	).
%
expand_sgn_term_list([], [], P, P).
expand_sgn_term_list([W-U|L], [W-V|L0], P, Q):-
	expand_arg(U, [], V, P, P0),
	expand_sgn_term_list(L, L0, P0, Q).
%
expand_sgn_terms([], [], P, P).
expand_sgn_terms([A|As], [B|Bs], P, Q):-
	expand_sgn_term(A, B, P, P0),
	expand_sgn_terms(As, Bs, P0, Q).
expand_sgn_terms(A, A, P, P).
%
collect_sgn([], [], []).
collect_sgn([X-Eqs|Xs], [X0-Eqs|Y], [X-X0|Z]):- sgn_dcl_term(X), !,
	collect_sgn(Xs, Y, Z).
collect_sgn([U|Xs], [U|Y], Z):- collect_sgn(Xs, Y, Z).

%
sgn_dcl_term(_ := _).

%  'pac:' is necessary. [2017/04/12]
compile_pac(:-bekind(N, Opts), _Eqs, _D, Xs, Ys, P, Q) :-
	once(pac:kind_term(N,  N1)),
	(	memberchk(nonvar, Opts)
	->  Nonvarcheck = "(X = [] :- var(X), !, fail)",
		term_string(Ruleterm, Nonvarcheck, [variable_names(Eqs)]),
		U = [Ruleterm - Eqs| Xs]
	;	U = Xs
	),
	compile_kind_block(U, N1, Opts, Ys, P, Q).
compile_pac(:-betrs(N), Eqs, D, Xs, Ys, P, Q) :-!,
		compile_pac(:-betrs(N, []), Eqs, D, Xs, Ys, P, Q).
compile_pac(:-betrs(N, Vs), _Eqs, _D, Xs, Ys, P, Q) :-
		term_variables(Vs, Us),
		make_trs_ref(N, Us, [], N0),
		pac:new_pac_name(Sub),
		make_trs_ref(Sub, Us, [], Sub0),
		pac_aux:expand_core_rec(N0, [],
			&(([X, Y]:-	call(Sub, X, X0), !,
						call(N0, X0, Y)),
			([X, X])),
			[], _, [U1, U2], []),
		P = [p(U1,['X'=X, 'Y'=Y, 'Z'=X0],[]),
			 p(U2,['X'=X],[])|P0],
		compile_trs_block(Xs, Ys, Sub0, P0, Q).
compile_pac(:-befun, _Eqs, _, Xs, Ys, P, Q) :-
 	compile_fun_block(Xs, Ys, P, Q).
compile_pac(:-X, _Eqs, _, Xs, Xs, [p(:-X, [], [])|P], P).
compile_pac(A := Expr, Eqs, Assoc, Xs, Xs,
			[p(A := Expr, Eqs, [])|P], Q) :- !,
	rec_subst(Expr, S0, Assoc),
	pac:zip_algebra(S0, S1),
	(	Cs \== []
	->	maplist(pac:expand_sgn(A), S1,  Cs),
		pac_etc:list_to_ampersand(Cs, As),
		pac_aux:expand_core_rec(A, [], &(As, [U, U]), [], _, Ds, [])
	;	pac_aux:expand_core_rec(A, [], [U, U], [], _, Ds, [])
	),
	maplist(pred([U, p(U,[],[])]), Ds, Y0),
	append(Y0, Q, P).
compile_pac(X, Eqs, _, Xs, Xs, Q, P) :-
	expand_clause_slim(X, [C|H]),
	( C == end_of_file  ->	Q = P   % Comment or white linecgi in prolog access
	;	Q  =  [p(C, Eqs, H)|P]
	).
% ?- pac_listing:rec_subst(a+a, R, [a-b, b-[]]).
rec_subst(A+B, A0+B0, F):-
	rec_subst(A, A0,F),
	rec_subst(B, B0,F).
rec_subst(A*B, A0*B0, F):-
	rec_subst(A, A0,F),
	rec_subst(B, B0,F).
rec_subst(\(A,B), \(A0,B0), F):-
	rec_subst(A, A0, F),
	rec_subst(B, B0, F).
rec_subst(A, B, _):- (is_list(A), B = A;  A=sgn(B)), !.
rec_subst(A, B, F):- memberchk(A-A0, F), !,
	rec_subst(A0, B, F).
rec_subst(A, A, _).

%
compile_kind_block([X-Eqs|Xs], N, Opts, Ys, P, Q):-
	once(compile_kind_block(X, Eqs, Xs, N, Opts, Ys, P, Q)).
%
% compile_kind_block(:-ekind, _, Xs, _, _, Xs, P, P).
compile_kind_block(:-ekind, _, Xs, N, Opts, Xs, P, Q):-
	(	memberchk(stop, Opts)
	->  Stop = "(X = quote(X) :-true)",
		term_string(Rule, Stop, [variable_names(Eqs)]),
		pac:compile_kind_rule(N, Opts, Rule, C, H, []),
		P = [p(C, Eqs, H)|Q]
	;	Q = P
	).
compile_kind_block(end_of_file, _, Xs, N, Opts, Ys, P, Q):-
	compile_kind_block(Xs, N, Opts, Ys, P, Q).
compile_kind_block(X, Eqs, Xs, N, Opts, Ys, [p(Y, Eqs, H)|P], Q):-
	pac:compile_kind_rule(N, Opts, X, Y, H, []),
	compile_kind_block(Xs, N, Opts, Ys, P, Q).


normalilze_rule((L=R):-B, L, R, B):-!.
normalilze_rule(L=R, L, R, true).

% ?- pac_listing:make_trs_ref(t, [A,A], M, R).

make_trs_ref(T, Vs, M, R):-
	( Vs==[] -> Args = []
	;	Args = [[Vs]]
	),
	complete_args(T, Args, T0),
	attach_prefix(M, T0, R).

% ?- edit(attach_prefix).
%
compile_trs_block([(:-etrs)-_|Xs], Xs, _, P, P).
compile_trs_block([end_of_file|Xs], Ys, Ref, P, Q):-
	compile_trs_block(Xs, Ys, Ref, P, Q).
compile_trs_block([X-Eqs|Xs], Ys, Ref, [p(Y, Eqs, H)|P], Q):-
	make_trs_sub(Ref, X, Y, H, []),
	compile_trs_block(Xs, Ys, Ref, P, Q).
%
make_trs_sub(N, A = B, H, L, L):-!,
	complete_args(N, [A,B], H).
make_trs_sub(N, (A = B :- Right), H :- C, L, M):-
		complete_args(N, [A,B], H),
		once(make_trs_cond(Right, N, C, L, M)).
%
make_trs_cond((X, Y), R, (X0, Y0), P, Q):-
	make_trs_cond(X, R, X0, P, P0),
	make_trs_cond(Y, R, Y0, P0, Q).
make_trs_cond((X; Y), R, (X0; Y0), P, Q):-
	make_trs_cond(X, R, X0, P, P0),
	make_trs_cond(Y, R, Y0, P0, Q).
make_trs_cond(U=V, R, C, P, P):- complete_args(R, [U, V], C).
make_trs_cond(G, _, G0, P, Q):- expand_arg(G, [], G0, P, Q).

%
compile_fun_block([(:-efun)-_|Xs], Xs, P, P).
compile_fun_block([end_of_file|Xs], Ys, P, Q):-
	compile_fun_block(Xs, Ys, P, Q).
compile_fun_block([X-Eqs|Xs], Ys, [p(Y, Eqs, [])|P], Q):-
	pac:expand_fun(X, Y),
	compile_fun_block(Xs, Ys, P, Q).

% ?- pred_grouping([p(a,[],[]), p(a:-b,[],[])], G).
% ?- pred_grouping([p(m:a,[],[]), p(m:a:-b,[],[])], G).
pred_grouping([], []).
pred_grouping([P|R], [[P|G]|R0]):-
	pred_grouping(P, R, G, R1),
	pred_grouping(R1, R0).

%
pred_grouping(_, [], [], []).
pred_grouping(P, [Q|R], [Q|G], R0):-
	P=p(C,_,_),
	Q=p(D,_,_),
	same_predicate_arity(C, D),
	!,
	pred_grouping(P, R, G, R0).
pred_grouping(_, R, [], R).

%
same_predicate_arity(X, Y):- predicate_arity(X, S),
							 predicate_arity(Y, S).

% ?-predicate_arity(user:a(_,_), R).
predicate_arity(X, Sig):-
	strip_module(X, M, X0),
	predicate_arity(X0, M, Sig).
%
predicate_arity(X:-_, M, M0:F/N):-
	strip_module(M:X, M0, X0),
	functor(X0, F, N).
predicate_arity(X, M, M:F/N):-
	functor(X, F, N).

%
is_backquote_begin(:-bekind(X,Y), :-bekind(X, Y)).
is_backquote_begin(:-bekind(X),   :-bekind(X, [])).
is_backquote_begin(:-befun,   :-befun).

%
is_backquote_end(:-ekind, :-ekind).
is_backquote_end(:-efun, :-efun).

		/*************************
		*     string_to_terms    *
		*************************/

string_to_terms(InStr, OutStr):-
	setup_call_cleanup(
		open_string(InStr, Stream),
		string_to_terms(Stream, OutStr, []),
		close(Stream)).
%
string_to_terms(Stream, P, Q) :-
		read_term(Stream, X, [variable_names(Eqs)]),
		(	X == end_of_file	->	Q = P
		;	update_back_quotes(X, X0),
			P = [X0 - Eqs|P0],
			string_to_terms(Stream, P0, Q)
		).
%
update_back_quotes(X, Y):-
		(	is_backquote_begin(X, Y)
		->	set_prolog_flag(back_quotes, symbol_char)
		; 	is_backquote_end(X, Y)
		->	set_prolog_flag(back_quotes, codes)
		;	Y = X
		).
