:- module(setup_aux, [mk_file_search_path/1, setup_env/1
					 ]).

setup_env(X=E):- eval_concat(E, V), setenv(X, V).
%
eval_concat($(V), U):-!, getenv(V, U).
eval_concat(X+Y, V):-!, eval_concat(X, U),
	eval_concat(Y, W),
	atom_concat(U, W, V).
eval_concat(V, V).

mk_file_search_path(DirStr):-
	getenv(pac_root, Dir),
	file_directory_name(Dir, RootName),
	mk_file_search_path(RootName, DirStr),
	getenv(home, H),
	concat_atom([H, '/.config'], Configs),
	assert(user:file_search_path(configs, Configs)).
%
mk_file_search_path(Root, DirStr):-
	forest_to_paths(DirStr, Eqs),
	maplist(attach_dir_prefix(Root), Eqs, Eqs0),
	maplist(assert_search_path, Eqs0).

% Remarck: [] means empty string "" to avoid "//" in paths.
attach_dir_prefix([], E, E):-!.
attach_dir_prefix(A, P = [], P = A):-!.
attach_dir_prefix(A, P = B, P = C):-
	concat_atom([A, /, B], C).
%
assert_search_path(A = B):-
	(	string(B) -> atom_string(B0, B)
	;   B0 = B
	),
	assert(user:file_search_path(A, B0)).


%% Set context_module to the module defined at the top line
%	of the buffer-file.
user:set_context_module(File):-
	setup_call_cleanup(
		open(File, read, S, [encoding(utf8)]),
		read(S, T),
		close(S)),
	(	(T = (:- module(M));  T = (:- module(M,_)))
	->	true
	;   M = user
	),
	module(M),
	write("\n"),
	write("Context module: "),
	write(M),
	write(.).

user:log(M, X, X):- user:log(M).

user:log(X):- getenv(snapshot, Log),
	open(Log, append, S),
	writeln(S, X),
	close(S).

% Ad hoc way to get HOST and USER

% ?- apropos(split).
%% forest_to_paths(+X, -Y) is det.
%	X is a  directory structure with path alias
%	for sub directories in X.
%   Y is a set of pairs (A=B) such that B is the absolute
%	file name of A such that file_search_path(A, B) becomes true.

% ?- setup_aux:forest_to_paths([], X).
% ?- setup_aux:forest_to_paths([(a:b)-[]], X).
% ?- setup_aux:forest_to_paths([(a:b)-[(c:d)-[]]], X).
% ?- setup_aux:forest_to_paths([(a:b)-[(c:d)]], X).
% ?- setup_aux:forest_to_paths([(a:b)-[(c:d), (e:f)]], X).

forest_to_paths([], []).
forest_to_paths([(P:Dir)-L|Xs], Out):-!,
	forest_to_paths(L, D),
	maplist(attach_dir_prefix(Dir), D, D0),
	forest_to_paths(Xs, Ys),
	append(D0, Ys, Zs),
	(	P == [] -> Out = Zs
	; 	Out = [P = Dir| Zs]
	).
forest_to_paths([:(Dir)-L|Xs], Out):-!, forest_to_paths([([]:Dir)-L|Xs], Out).
forest_to_paths([A|Xs], Out):- forest_to_paths([A-[]|Xs], Out).

%
user:shot_init:- getenv(snapshot, F),
	(	exists_file(F)
	->	delete_file(F)
	;	true
	).
%
user:shot(X):- getenv(snapshot, File),
	setup_call_cleanup(
		open(File, append, S, [encoding(utf8)]),
		writeln(S, X),
		close(S)).

% Check the log file "snapshot" at Desktop.
% ?- dbg(shift(true)).
% ?- dbg((true, shift(X=1))).
% ?- dbg(shift((X=1, Y=2))).

user:dbg_init:- getenv(snapshot, F),
	(	exists_file(F)
	->	delete_file(F)
	;	true
	).
%
:- meta_predicate user:dbg(0).
%
user:dbg(G):- getenv(snapshot, File),
	setup_call_cleanup(
		open(File, append, S, [encoding(utf8)]),
		setup_aux:dbg(G, S),
		close(S)).
%
:- meta_predicate dbg(:, ?).

dbg(Goal, A):- reset(Goal, PGoal, Cont),
	(	var(PGoal) -> true
	;	dbg_trace(PGoal, A)
	),
	(	Cont == 0 -> true
	;	dbg(Cont, A)
	).

%
:- meta_predicate dbg_trace(0, ?).
dbg_trace(true, _):-!.
dbg_trace((X,Y), A):-!, dbg_trace(X, A), dbg_trace(Y, A).
dbg_trace((X;Y), A):-!, (dbg_trace(X, A); dbg_trace(Y, A)).
dbg_trace(G, A):-  call(G),
	write(A, "\n"),
	writeln(A, G).
%
user:dshot_init:- dbg_init.

:- meta_predicate user:dshot(0).
user:dshot(G):- getenv(snapshot, File),
	setup_call_cleanup(
		open(File, append, S, [encoding(utf8)]),
		setup_aux:dbg_trace(G, S),
		close(S)).
