:- module(eh, [
	       new_file/2,  new_base_name/2, counter/3,
	       assemble/2, expand_cgi_path/2,
	       boomerang/2,  getstring/1, include_text/2,
	       file_string/3,
	       getinfo_codes/2, getinfo_string/2, getinfo/2,
	       sh/1, pshell/1, qshell/1,
		   vector_term/3,
	       perform/4,
	       apply3/3,
	       choose_file/1, choose_folder/1,
	       set_dired/2, dir/1, dired/2
			  ]).

% ?- misc:sh(open("/Users/cantor/Documents/Mac\\ Fan_OCR.pdf")).
% ?- A='/Users/cantor/Documents/Mac Fan_OCR.pdf', term_string(A, B), misc:sh(open(B)).
:- use_module(pac(basic)).
:- use_module(util(obj)).
:- use_module(pac(reduce)).
:- use_module(util(file)).
:- use_module(util('prolog-elisp')).
:- use_module(util('emacs-jockey')).
:- use_module(util('emacs-jockey2')).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).
%
:- set_prolog_flag(allow_variable_name_as_functor, true).

%%  getinfo_codes(+S:atom/string, -R:codes) det
%	is True if
%  R is unified with a standard output codes of a shell command S.

% ?- eh:getinfo('echo Hello', X).
% ?- eh:getinfo("echo Hello", X).
% ?- eh:getinfo_codes('echo Hello', X).
% ?- eh:getinfo("osascript -e 'return  POSIX path of (choose folder)'", X).
% ?- eh:getinfo("date +%Y-%m-%d", X).
% ?- eh:getinfo_codes("date +%Y-%m-%d", X), smash(X).

getinfo_codes(P, X):- file(pipe(P), read, getstring(X0)),
	( X0 == [] -> X = X0; append(X, [_], X0) ).

%;; (setq module-query  "qcompile(util('emacs-handler'))), module(eh).")
% ?- qcompile(util('emacs-handler')), module(eh).

%%  getinfo_string(+S:atom/string, -R:string) is det.
%	R is unified with a string for the standard output codes
%	of a shell command S.
getinfo_string(P, X):- getinfo_codes(P, Codes),
					   string_codes(X, Codes).

%% getinfo(+S:atom/string, -R:atom/string) det
%   is True if
%  R is unified with an atom/string as standard output of the
%  shell command S.

%;; (setq module-query  "qcompile(util('emacs-hanlder')), module(eh).").
% ?- expand_file_name("~", [X]).

getinfo(P, X):- getinfo_codes(P, X0),
		atom_codes(X, X0).

%
done(_, _).

%% choose_folder(-X:string) det
%	is True if
%  X is unified with a folder name which you choose from a Finder window.

% ?- eh:choose_folder(X).
choose_folder(X):-
	Pipe = "osascript -e 'tell application \"System Events\" to POSIX path of (choose folder)'",
	getinfo_codes(Pipe, X0),
	ejockey:dakuten_convert(X0, X).

%% choosefile(-X:atom) det
%	is True if
%  X is unified with a file name which you choose from a Finder window.

% ?- eh:choose_file(X).
choose_file(X):-
	Pipe = "osascript -e 'tell application \"System Events\" to POSIX path of (choose file)'",
	getinfo_codes(Pipe, X0),
	ejockey:dakuten_convert(X0, X1),
	atom_codes(X, X1).

%% scriptstart det
%  Repeats cycles of  read / act / write on stadanrd I/O.
scriptstart :-
 	prompt(_, ''),
 	current_input(In),
 	set_stream(In, encoding(utf8)),
 	current_output(Out),
 	set_stream(Out, encoding(utf8)),
    process_loop.

%
% process_loop :- catch(process_step, E, handle_exception(E)).

process_loop :- catch(process_step, E, handle_exception(E)),
  	process_loop.

%
process_step :-	once(read_term_from_lisp(C)),
	(	phrase(C, _, R)
	->	insert_buffer(R)
	;	insert_buffer("fail")
	).

%
handle_exception(E) :- smash(["exception: ",  E], M),
			message(M),
			lisp('start-emacshandler'()).
%% filter(+F:pred/1, +A:pred/2, +X:term, -Y:term) det
%	is True if
%  Y is unified with the value of F(X) when A(X) is true;
%  otherwise  with X.

% ?-  eh:maplist(filter(atom, atom_codes), [a,f(b), c], X).
%@ X = [[97], f(b), [99]].
filter(F, A, X, Y):- call(F, X) -> call(A, X, Y); Y=X.

term(F, X, Y, Z):- Z=..[F, X, Y].


% ?- qcompile(util('emacs-handler')).
% ?- module(eh).

%% termrec(+F:pred/3, +I:term, +X:term, -Y:term)  det
%	is True if
%  Y is unified with the argument list of X folded by F with initial value I.

% ?- eh:termrec(plus, 0, f(1,2,3,5,6,7), SumOfArgs).
:- meta_predicate termrec(3, ?, ?, ?).
termrec(F, I, X, Y) :- functor(X,_,N),
	termrec(F, 0, N, X, I, Y).

termrec(_, N, N, _, V, V) :-!.
termrec(F, J, N, X, V, Y) :- J1 is J + 1,
	arg(J1, X, A),
	call(F, A, V, V1),
	termrec(F, J1, N, X, V1,Y).

%% termrec(+A:pred/4, +T:term, -T1:term, +N:term, -N1:term) det
%	is True if
%  T1 is unified with the argument list of T, and
%  N1 with N,  folded together by A.

% ?-  eh:termrec(pred([I, J, X, a(X)]:- J is I + X), f(1,2,3,5,6,7), H, 0, S).
:- meta_predicate termrec(4, ?, ?, ?,?).
termrec(A, T, T1, N, N1):- functor(T, Fun, Ar),
	functor(T1, Fun, Ar),
	termrec(A, 0, Ar, T, T1, N, N1).

%
termrec(_, M, M, _, _, N, N) :- !.
termrec(A, J, M, T, T1, N, N1) :- J1 is J + 1,
    arg(J1, T, B),
    arg(J1, T1, C),
    call(A, N, N2, B, C),
    termrec(A, J1, M, T, T1, N2, N1).
%
power(P) --> maplist(phrase(P)).

%% phrase_on_car(+F:pred/2, +X:term, -Y:term) det
%	is True if
%  Y is unified with the value of applying F as an  extended phrase to X.

phrase_on_car(X, [Y0|Z], [Y|Z]) :- once(perform([], X, Y0, Y)).

% ?- eh:perform([], (=, =, =), a, X).
% ?- perform([], seqcal: (p2q,raster,rasterx), p(right(+),[]>>[a+ (!a)],[p(right(!),[]>>[a,!a],[p(axiom,[a]>>[a])])]), X), smash(X).
% :- meta_predicate perform(?,:,?,?).
perform(_, eval, X, Y):- eval(X, Y).
perform(_, true, X, X).
perform(Ms, M:A, X, Y):-	perform([M|Ms], A, X, Y).
perform(M, (A; _), X, Y):-	perform(M, A, X, Y).
perform(M, (_; A), X, Y):-	perform(M, A, X, Y).
perform(M, (A, B), X, Y):-	perform(M, A, X, Z), perform(M, B, Z, Y).
perform([], A, X, Y):-		call(A, X, Y).
perform([M|_], A, X, Y):-	call(M:A, X, Y).

% % Prolog on emacs buffer
% /** ::prolog
%  append(X, Y,[a,b,c]), X=[_,_]
% **/

prolog --> solve_bind.

solve_bind_once --> solve_bind, !.

solve_bind -->  herbrand(Bind),
	current(X),
	{	once(pac:expand_goal(X, user, Y, P, [])),
		maplist(assert, P),
		solve(Y)
	},
	peek(Bind),
	term_codes.

%
solve((X,Y)):- solve(X), solve(Y).
solve(X):- call(X).

%
phrase(C, G, X, Y) :- call(C, call(G, X, Y)).

% %
% once(G)--> phrase(once, G).

% once(G, U, L0, L) :- once(call(G, U, L0, L)).

% % once(G, X, X0) --> phrase(once, call(G, X, X0)).
% once(G, X, X0, L0, L) :- once(call(G, X, X0, L0, L)).


% a la cd command
% ?- eh:walk_on_tree(up, [a,b,c], X).
% ?- eh:walk_on_tree(up_down([x,y]), [a,b,c], X).

walk_on_tree(up, X, Y):- !,  (append(Y, [_], X) -> true; Y=X).
walk_on_tree(down(A), X, Y):- !,  append(X, A, Y).
walk_on_tree(up_down(A), X, Y):-  walk_on_tree(up, X, X0),
	walk_on_tree(down(A), X0,  Y).

%
% ?- eh:change_unix_path(up, "/a", X).
%@ X = "".

% ?- eh:change_unix_path(up, "/ab/cd/ef", X).
% ?- eh:change_unix_path(down("x/y"), "/ab/cd/ef", X).
%@ X = "/ab/cd/ef/x/y/".
% ?- eh:change_unix_path(up_down("x/y"), "/ab/cd/ef", X).
%@ X = "/ab/cd/x/y/".

change_unix_path(up, P, Q):-  !, path_to_list(P, X),
	walk_on_tree(up, X, Y),
	path_to_list(Q, Y).
change_unix_path(down(A), P, Q):-  !,
	path_to_list(A, Z),
	path_to_list(P, X),
	walk_on_tree(down(Z), X, Y),
	path_to_list(Q, Y).
change_unix_path(up_down(A), P, Q):-
	path_to_list(A, Z),
	path_to_list(P, X),
	walk_on_tree(up_down(Z), X, Y),
	path_to_list(Q, Y).

% ?- eh:path_to_list("/", X).
% ?- eh:path_to_list(A, []).
% ?- eh:path_to_list(A, [a]).
% ?- eh:path_to_list(A, [a,b]).

path_to_list(A, X):- nonvar(A), !,
	atomics_to_string(Y, (/), A),
	remove_null(Y, X).
path_to_list(A, X):- remove_null(X, X0),
	list_to_unix_path(X0, A).

%
list_to_unix_path([], "/"):- !.
list_to_unix_path(X,  A):-  append([[""],X,[""]], X0),
	atomics_to_string(X0, "/",  A).

%
remove_null([''|A], B):-!, remove_null(A, B).
remove_null([""|A], B):-!, remove_null(A, B).
remove_null([X|A], [X|B]):- remove_null(A, B).
remove_null([], []).


% [2015/12/28]
file_string(File, Length, String):-  open(File, read, Stream, [encoding(utf8)]),
			     read_string(Stream, Length, String),
			     close(Stream).
%
file_string(File, String):- file_string(File, _, String).

%
getstring(X) :- get_code(C),
	(C == -1 -> X=[] ; X=[C|Y], eh:getstring(Y)).

putstring(X) :- maplist(put_code, X).

getline(X) :- get_code(C),
	((C == -1; C==0'\n) -> X=[] ; X=[C|Y], getline(Y)).

putline(X) :- smash(X), put_code(0'\n).

%!	assemble(L:list, F:stream) is det.
%
%	True when all elements of L have been written to F.
%	text(T) -- T as text.
%       file(G)	-- the contents of G
%	region  -- the current buffer region
%	buffer  -- the current whole buffer
%

assemble(Fs, F) :- expand_file_search_path(F, F1),
        open(F1, write, FX, [encoding(utf8)]),
        maplist(assemble_basic(FX), Fs),
        close(FX).

assemble_basic(FX, text(F)) :- !, clean_io(FX, write, basic:smash(F)).
assemble_basic(FX, file(F)) :- !, expand_file_search_path(F, F1),
        open(F1, read, FY, [encoding(utf8)]),
        clean_io(FY, read, eh:getstring(D)),
        maplist(put_code(FX), D).
assemble_basic(FX, codes(Codes)) :-!, maplist(put_code(FX), Codes).
assemble_basic(FX, region(Codes)):-!, maplist(put_code(FX), Codes).
assemble_basic(FX, buffer) :-
	call_lisp(list('point-min'(), 'point-max'()), [value(L), string(t)]),
	list_number_list(L, [Min, Max]),
	get_buffer_region(Min, Max, R),
	maplist(put_code(FX), R).

%%  Mac open file  %%%
mac_open(F):-  pshell(open(F)).

mac_open(F, F):- mac_open(F).

mac_open(P, F, F):- mac_open_prefix(P, F).

mac_open_prefix(_, F):- prefix_chk(`/`, F ), !, mac_open(F).
mac_open_prefix(_, F):- prefix_chk(`~`, F), !,	mac_open(F).
mac_open_prefix(Prefix, F):- mac_open(Prefix+ '/'+ F).

prefix_chk(Prefix, String):- append(Prefix, _, String).

:- meta_predicate wild_map(1,?).
wild_map(M, W):- expand_file_name(W, L0), maplist(M, L0).

:- meta_predicate wild_map(2,?,?).
wild_map(M, W, L):- expand_file_name(W, L0), maplist(M, L0, L).

wild_open(Wildcard):- wild_map(pred([X] :- pshell(open(X))), Wildcard).

		/******************************
		*     counter file handler    *
		******************************/

%%	counter_general(+P:pred, +X:obj, -Y:obj) is det.
%	Manage a counter file depending on  P
%		new	-- new counter
%		check	-- check existence.
%		update	-- increment the content by 1.

:- meta_predicate counter_general(1, ?, ?).

counter_general(FileProp)	-->  obj:obj_get([counter_name(C), directory(D)]),
{
	pshell(mkdir(-p, D)), !,
	setup_call_cleanup(	working_directory(Old, D),
				call(FileProp, C),
				working_directory(_, Old))
}.

%%	counter(+P, +X, -Y) is det.
%	Perform action P  on a file X to unify Y with the result.

counter(new)	-->
	counter_general(pred([C] :- file(C, write, format("~w.~n", [0])))),
	obj_put([count(0)]).
counter(update) --> counter_general(exists_file), !,
	 obj:obj_get([counter_name(C), directory(D)]),
	{	working_directory(Old, D),
		file(C, read, read(V1)),
		V is (V1 + 1) mod 100,
		file(C, write, format("~w.~n",[V])),
		working_directory(_, Old)
	},
	obj_put([count(V)]).
counter(update) --> counter(new).


%%	new_file(+X, -Y) is det.
%	Create a new file according to X, and unify Y with the result.
new_file --> obj:obj_get([ stem(R), directory(D) ] ),
        counter(update),
        obj:obj_get([count(N)]),
        {	atomic_list_concat([R,  N], B),
		atomic_list_concat([D,  / , B], F)
        },
        obj_put([base(B), file(F)]).

%%	new_file_here(+X, -Y) is det.
%	Create a new file at the current directory.

new_file_here	--> obj_get([stem(R), count(N)]),
        { atomic_list_concat([R,  N], B) },
        obj_put([base(B)]).

%%	new_base_name(+X, -Y) is det.
%	Create a new file according to X, and unify Y with the result.
%
new_base_name -->
        counter(update),
        obj_get([count(N), stem(R), directory(D)]),
        {	atomic_list_concat([R,  N], B),
		atomic_list_concat([D,  / , B], F)
        },
        obj_put([base(B), file(F)]).

%%	dir_open(+X, -Y)  is det.
%	Open the file X, and unify Y with X.

dir_open	--> obj_get([directory(D)]),
	{	pshell(open(D))	}.

%%	dir_open(+D) is det.
%	Open the directory D.

dir_open(D):-	dir_open([directory(D)], _).

%%	expand_cgi_path(+X, -Y) is det.
%	Expand a CGI path  in X, and unify Y with X.

expand_cgi_path(X, Y):-
	getenv(http_cgi_bin, CB),
    expand_path(CB, X, Y).

%%	expand_cgi_path(+P, +X, -Y) is det.
%	Expand a CGI path under the current home directory in X,
%	and unify Y with X.

expand_path(P, X, Y):-  getenv(user, Name),
        smash([`/~`, Name, `/`,  P, `/`, X], Y).

%%	include_text(+X, -Y) is det.
%	Read as string from a file X into Y.

include_text(X,Y) :- once(filepath(X,P)), file(P, read, getstring(Y)).

%%	include_text(+L, -Y) is det.
%	Read as strings from all files in X into Y.

include_text(A)--> {listp(A) -> L = A; L = [A]}, peek(L),
		   maplist(include_text).

%%	filepath(+X, -Y) is det.
%	Expand a file path X into Y.

filepath(X,X) :- atomic(X), !.
filepath(A,X) :- A=..[P|A1],
     ( P= (/) ->  Q = [ /|A1] ; dir(P, D), Q = [D, /|A1] ),
     atomic_list_concat(Q, X).
filepath(A,X) :- expand_file_search_path(A, X).

% some handy
singleton(X,[X]).
comma((X, Y), X, Y).
args(X, A, B):- arg(1, X, A), arg(2, X, B).
image(R, S) :- maplist(snd, R, S).
%
vector_term(_, [X], X).
vector_term(F, [X, Y|Z], U):- vector_term(F,[Y|Z], X0), U=..[F, X, X0].

%
wrap(X, Y, A, [X,A,Y] ).

%%	 boomerang(+Dir:<directory>, ShellCom:<shell command>) is det.
% Goto Dir,  do ShellCom, then go back to the original directory.

boomerang(Dir, ShellCom):- sh(cd(Dir); ShellCom).

same_atom(X, Y, "yes"):- atom_codes(X, Y).
same_atom(_, _, "no").

backquote_string(X) :- string_codes(X, [96]).


sh(X) :- pshell(X, [path, c, q]).

% ?- pshell(ls, [c, q, path]).
% ?- pshell(ls, [path]).
% ?- pshell(echo("$PATH") > "~/Desktop/PATH").

% ?- pshell("update-all").
% ?- pshell("update-swipl").

pshell(X, Opts):-
	(	memberchk(path(V), Opts)
	->	T0 = shell( "PATH=" + V ; X)
	;	T0 = shell(X)
	),
	(	memberchk(q, Opts)
	->  T1 = T0 + " >> /dev/null 2>&1"
	;	T1 = T0
	),
	misc:shell_string(T1, T2),
	(	memberchk(c, Opts)
	->	term_string(T2, T3),
		T4 = "/bin/sh -c " + T3
	;   T4 = T2
	),
	misc:shell_string(T4, S),
	shell(S).

% ?- pshell(ls).
% ?- pshell(pwd).
% pshell(T):- misc:shell_string(T, S), shell(S).
pshell(T):- pshell(T, []).

% pshell in quiet mode.
% ?- qshell(ls).
% ?- qshell(pwd).
qshell(X):- pshell(X,  [q]).

:-meta_predicate(directory(?, 0)).  % <= neccessary dcl.
directory(D, A):-
	working_directory(Old, D),
	call(A),
	working_directory(_, Old).
%  ?- ls_pdf_files(L).
%  ?- ls_pdf_files(_, L), smash(L).
%  ?- ls_files_suffix(['.pdf'], L).

%% ls_pdf_files(?:term, -L:list) is det
%  True when L is unified with a list of pdf files in the current folder.
ls_pdf_files(_, L)	:- ls_pdf_files(L0), insert("\n", L0, L).

%% ls_pdf_files(L:list) is det
%  True when L is unified with a list of pdf files in the current folder.
ls_pdf_files(L)		:- ls_files(suffix([".pdf"]), L).

%% ls_files_suffix(S, L:list) is det
%  True when L is unified with a list of files in the current directory wich a suffix S.
ls_files_suffix(S, L)	:- ls_files(suffix(S), L).

%%  suffix(+S:string, X:string) is det
%   True if S is a suffix  of L.
suffix(S, X):- sub_string(X, _, _, 0, S).

%%  directory_filter(+F:pred/1, -L:list) det
%	is True if
%  L is unified with the list of files filtered by F.

% ?- eh:directory_filter(pred(([X]:- sub_string(X, _, 3, 0, ".pl"))), PDFs).
:- meta_predicate directory_filter(1, ?).
directory_filter(Filter, Fs):- ls_objects(FS0), collect(Filter, FS0, Fs).

%%  ls_files(Filter:pred/1, L:list) is det.
%	True when L is unfified with names of files in the current folder
%	that  satisfies Filter.
ls_files(Filter, L):- directory_filter(exists_file, L0),
	collect(Filter, L0, L).

%%  ls_files(L:list) is det.
%	True when L is unfified with names of files in the current folder
ls_files(L):- directory_filter(exists_file, L).

%%  ls_dirs(Filter:pred/1, L:list) is det.
%	True when L is unfified with names of files in the current folder
%	that  satisfies Filter.
ls_dirs(Filter, L):- directory_filter(exists_directory, L0),
	collect(Filter, L0, L).

%%  ls_dirs(L:list) is det.
%	True when L is unfified with names of files in the current folder
ls_dirs(L):- directory_filter(exists_directory, L).

%%  ls_files_dirs(Fs:list, Ds:list) is det.
%	True when Fs and Ds are unfified with names of files
%	and directories, repectively, in the current folder
ls_files_dirs(Fs, Ds):- ls_objects(A), object_classify(A, Fs, Ds).

%% ls(-F:list, -D:list) is det
%	True if
%	F and L are unified with a list of files and directoires, respectively,
%	in the current directory.

% ?- eh:ls(X, Y).
% ?- working_directory(_, "/Users/cantor"), eh:ls(X, Y).

ls --> ls_files_dirs.
%%	ls_object(A:list) is det
%	True if
%	A is unifed with a list of files and directoires.
%	?- eh:ls_objects(A), maplist(writeln, A).
ls_objects(A):- getinfo_codes(ls, X),
	(X==[]
	-> A = []
	;  once(split(X, X0)), maplist(atom_codes, A, X0)
	).

%
object_classify([], [], []).
object_classify([A|As], [A|Xs], Ys):- exists_file(A), !,
	object_classify(As, Xs, Ys).
object_classify([A|As], Xs, [A|Ys]):- object_classify(As, Xs, Ys).

%
set_dir(X, Y):- working_directory(X, Y).
set_dir(X) :- working_directory(_, X).

%
get_dir(X) :- working_directory(X, X).

%%	excursion(A:pred/0) is det.
%	True if save the current directory,  do the action A,
%	and restore the saved directory.

excursion(A):-  get_dir(D), call(A), set_dir(D).

%% dir_tree(L:list) is det.
%	True when L is unifed with a directory structure
%	of the current directory.
%	The directory structure is a list of elements of the form f(x) or
%	d(y, z), where x is a filename, y a directory name, and z
%	a directory structure.
%
% ?- eh:dir_tree(L), maplist(writeln, L).

dir_tree(L):- ls(Fs, Ds),
	maplist(unary(f), Fs, Gs),
	maplist(excursion_dir_tree, Ds, Es),
	append(Gs, Es, L).

dir_tree(P, L):- excursion((set_dir(P), dir_tree(L))).

excursion_dir_tree(N, d(P, L)):-
	get_dir(P0),
	atomic_list_concat([P0,N,(/)], P),
	excursion((set_dir(P), dir_tree(L))).
%
unary(F, T)	:- functor(T, F, 1).
unary(F, A, T)	:- T =.. [F, A].

%%  map_directory(Act:goal) is det
%	True if Act is applied to each object  in the current directory.
% ?-  expand_file_name('~/', [H]), eh:(working_directory(_, H),  map_directory(pred([X]:- writeln(X)))).

:- meta_predicate map_directory(:).
map_directory(Act):- ls_objects(Fs), maplist(check_do(Act), Fs).

check_do(true, _):- !.
check_do(false, _):- !, fail.
check_do((A, B), X):- !,  check_do(A, X), check_do(B, X).
check_do((A; B), X):- !, once(check_do(A, X); check_do(B, X)).
check_do(\+ A , X):- !, \+ check_do(A, X).
check_do(A->B , X):- !, (check_do(A, X) -> check_do(B, X)).
check_do(A , X):- call(A, X).

%% escape_posix_file_name_char(+X:text, -Y:text) is det
%	True if Y is unified with a copy of X in which
%	special characters in X are escaped so that  Y is
%	safe for posix file name.
% ?- ejockey:escape_posix_file_name_char(`a : (b)`, R), atom_codes(A, R), smash(R).
%@ a \@ \(b\)
%@ R = [97, 32, 92, 64, 32, 92, 40, 98, 92, 41],
%@ A = 'a \\@ \\(b\\)' .
escape_posix_file_name_char(X, Y):-
  foldr(pred(	[0'(,  U, [0'\\,    0'(  | U]	] &
		[0'),  U, [0'\\,    0')  | U]	] &
		[0'\', U, [0'\\,    0'\' | U]	] &
		[0':,  U, [0'\\,    0'@  | U]	] &
		[0'/,  U, [0'\\,    0'@  | U]	] &
		[A,    U, [A|U]			]
	    ),
	X, [], Y).

%'
%%  r_act_plus(+F:pred/3, +L:list, -Y:term) is det
%	True if
%  Y is unified with the folded L by F from the right,
%  so that Y is a right branching binary tree.
% ?- eh:foldl(variant(term(+)), [b,c,d], a, Y).
%@ Y = a+b+c+d.
:- meta_predicate r_act_plus(3, ?, ?).
r_act_plus(F, Xs, V) :- r_act_plus(F, Xs, _, V).

%%  r_act_plus(+F:pred/3, +L:list, -X:term,  -Y:term) det
%	is True if
%  Y is unified with the folded L by F from the right,
%  and X with the last element of L, so that Y is a binary tree with
%  descendants  on right branches only.
%
% ?- eh:r_act_plus(term(+), [a,b,c], X, Y).
%@ X = c,
%@ Y = a+ (b+c).

% ?- eh:r_act_plus(term(+), [a,b,c,d], X, Y).
% ?- eh:r_act_plus(term(+), [1, 2, 3], X).
% ?- eh:r_act_plus(variant(term(+)), [1, 2, 3], X).
% ?- eh:r_act_plus(cons, [1, 2, 3], X).
:- meta_predicate r_act_plus(3, ?, ?, ?).
r_act_plus(_, [X], X, X):-!.
r_act_plus(F, [A|As], X, Y) :- call(F, A, Y0, Y),
	r_act_plus(F, As, X, Y0).

%%  apply(P:pred, Xs:[A1,...,An], V:term) is nondet
%	True when
%  apply(P, [A1,...,An, V]) is true.

%  ?- eh:apply(append([a,b]),[[c,d]], R).
%@ R = [a, b, c, d].

:- meta_predicate apply3(:, ?, ?).

apply3(A, Xs, V):-  apply_(Xs, A, V).

%
apply_([], A, V)		:- !, call(A, V).
apply_([X], A, V)		:- !, call(A, X, V).
apply_([X, Y], A, V)	:- !, call(A, X, Y, V).
apply_([X, Y, Z], A, V)	:- !, call(A, X, Y, Z, V).
apply_([X, Y, Z, U], A, V)	:- !, call(A, X, Y, Z, U, V).
apply_(Xs, A, V)		:- append(Xs, [V], Args),  apply(A, Args).

%% mapleaves(+F:pred/2, +X:term, -Y:term) det
%	is True if for X = f(a1,..., an),
%  Y is unified with a term f(b1,...,bn) where call(F, ai, bi) is
%  recursively true.  mapleaves/3 is a recursive version of mapterm/3.

% ?- eh:mapleaves(pred([X,[X,X]]), f(a,[u,v], b([1,2])), R).

:- meta_predicate mapleaves(2, ?, ?).
mapleaves(F, M:X, M:Y) :- !, mapleaves(F, X, Y).
mapleaves(F, X, Y) :- ala_list(X), !,
	maplist(mapleaves(F), X, Y).
mapleaves(F, X, Y) :- atomic(X),  !, call(F, X, Y).
mapleaves(F, X, Y) :- functor(X, A, N),
	functor(Y, A, N),
	mapleaves(0, N, F, X, Y).

:- meta_predicate mapleaves(?,?,2,?,?).
mapleaves(N, N, _ , _ , _) :- !.
mapleaves(I, N, F, X, Y):- J is I + 1,
	arg(J, X, Z),
	arg(J, Y, U),
	mapleaves(F, Z, U),
	mapleaves(J, N, F, X, Y).

%% mapleaves(+F:pred/1, +X:term) det
%	is True if
%  for X = f(a1,..., an), call(F, ai) is true for all ai.
%  mapleaves/2 is a recursive version of mapterm/2.

% ?-  eh:mapleaves(mapleaves(writeln), f(g(a), [c,x(j),e], h(b))).
:- meta_predicate mapleaves(1, ?).
mapleaves(F, _:X) :- !, mapleaves(F, X).
mapleaves(F, X) :- is_list(X), !, maplist(mapleaves(F), X).
mapleaves(F, X) :- atomic(X),  !, call(F, X).
mapleaves(F, X) :- functor(X, _, N), mapleaves(0, N, F, X).

%
mapleaves(N, N, _, _) :- !.
mapleaves(I, N, F, X) :- J is I + 1,
	arg(J, X, Z),
	mapleaves(F, Z),
	mapleaves(J, N, F, X).


		/****************************************
		*     accessing favorite directories    *
		****************************************/

%%	setup_candidate(+A:string, +P:pred/1) is det.
%	Collect all directory names specified by P,
%	send the list to the lisp so that it is bound
%	to A as a lisp atom.

:- meta_predicate setup_candidate(?,1).
setup_candidate(SetVar, Pred) :-
	setof(D, call(Pred, D), S),
	maplist(atom_string, S, Cs),
	call_lisp_wait(setq(SetVar, quote(Cs))),
 	atom_string(SetVar, SetVarName),
	atomics_to_string(["variable ", SetVarName, " has been set."], String),
	call_lisp_wait(message(String)).

set_dired --> { file_search_path(pac, R),
				atomics_to_string([R, /, 'dirdata.pl'], File),
				exists_file(File),
				!,
				load_files([File]),
				setup_candidate('dir-set', dir) }.
set_dired --> { lisp(setq(quote('dir-set'), nil))}.
%
dir(X):- config:dir_data(X, _).

%%	dired(?X, ?Y) is det.
%	Get a directory from lisp by completing-read with
%	candidate list 'dir-set', expand the directory name, and
%	ask lisp to open by 'dired' command.  Y is unified with X.

dired --> { call_lisp(
		prompt("directory name ",
			quote('dir-set')),
			[value(D0), string(t)]),
	  term_codes(T, D0),
	  atom_string(A, T),
	  config:expand_dir_name(A, D),
	  lisp(dired(D))
	  }.
