:- module(fol, [old_prove/2, old_prove/3]).

:- use_module(pac('expand-pac')).
:- use_module(pac(op)).
:- use_module(pac(basic)).
:- use_module(pac(basic)).
:- use_module(util(beta)).
:- use_module(util(misc)).
:- use_module(util(leantap)).

term_expansion --> pac:expand_pac.

:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<-').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<--').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<-->').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<->').

% ?- current_op(X, Y, -).
%@ X = 200,
%@ Y = fy ;
%@ X = 500,
%@ Y = yfx.
:- op(200, fy, ~).
:- op(1050, xfy, <->).

% ?- Obsolete.[2022/02/16]

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A simple but complete theorem prover based on     %
% linear  resolution using iterated  deepening      %
% search method.			            %
%						    %
%   by K.Mukai 1993 July / 2006 July, rewritten.    %
%	2012 June  front end updated to PAC-based.  %
%	2015 March main part rewritten.		    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

		/**********************************
		*     prove_by_linear_ancestor    *
		**********************************/


% ?-  F = [[append([],X,X)], [append([X|Y],Z,[X|U]), ~append(Y,Z,U)]], fol:prove(F, S, [center([[~append([a,b], U, [a,b,c,d,e])]]), initial_depth(300), clausal(true), iteration_limit(5)]).
% ?- F = ~((a;b)&(-a;b)&(a; -b)&(-a; -b)), prove(F, R, [trace(X)]).
%@ F = (~ ((a;b)& (-a;b)& (a;-b)& (-a;-b))),
%@ R = 11,
%@ X = [input_resolution([a, b], [b, ~a]), input_resolution([b], [a, ~b]), input_resolution([a], [~a, ~b]), ancestor_resolution([~b], [b])].

old_prove(X, S):- once(prove(X, S, [])).

%
old_prove(X, S, Options):-
	completing_options([center([]), initial_depth(300), iteration_limit(5)],
			   Options, Options0),
	memberchk(initial_depth(D), Options0),
	memberchk(iteration_limit(N), Options0),
	memberchk(center(SOS),  Options0),
	(  is_list(X)
	->	Y = X						% a set of clauses given.
	; 	leantap:fol_cnf(-(X), Y)	% first-order formula given.
	),
	setup_input_clauses(Y, Y0),
	once(( memberchk(trace(Trace), Options0); true )),
	once((	Y0 == [],
		S = 'The input has been reduced to empty set.',
		Trace = []
	; 	memberchk([], Y0),
		S = 'proved.',
		Trace = []
	;	is_list(SOS),
		SOS \== [],
		iterated_deepening(
			(	member(C0, SOS),
				linear_prove(C0, [], Trace, [])),
			   D, N, S)
	; 	iterated_deepening(
			(	input_clause(C0),
				linear_prove(C0, [], Trace, [])),
			   D, N, S)
	;	S = 'no proof found.',
		Trace = []
	)).

%
setup_input_clauses(Y, Y3):-
	maplist(merge_clause, Y, U),
	merge_cnf(U, Y0),
	sort(Y0, Y1),
	pure_literal_rule(Y1, Y2),
	remove_tautology_clauses(Y2, Y3),
	abolish(input_clause, 1),
	maplist(pred([C]:- assert(input_clause(C))), Y3).

% main predicates.
linear_prove(X, Y) :- linear_prove(X, Y, _, []).

%
linear_prove([], _, H, H):-!.
linear_prove([X|R], Ance, [H|P], Q):- complementary(X, X1),
	(	member(L, Ance),
		copy_term(L, L0),
		H = ancestor_resolution([X|R], L0)
	;	input_clause(L0),
		H = input_resolution([X|R], L0)
	),
	select(A, L0, L1),
	unify_with_occurs_check(A, X1),
	merge_clause_clause(L1, R, Y0),
	sort(Y0, Y1),
	has_no_complementary_pair(Y1),
	copy_term(Y1, New),
	once(add_to_ancestors(New, Ance, Ance0)),
	linear_prove(Y1, Ance0, P, Q).
%
has_no_complementary_pair(X):- \+ has_complementary_pair(X).
%
has_complementary_pair([X|Y]):- has_complementary_pair(X, Y).
has_complementary_pair([_|Y]):- has_complementary_pair(Y).
%
has_complementary_pair(X, [Y|_]):- complementary_pair(X, Y).
has_complementary_pair(X, [_|Y]):- has_complementary_pair(X, Y).
%
complementary_pair(~A, B):- A==B.
complementary_pair(A, ~B):- A==B.
%
add_to_ancestors(Y, [X|_], _):- subsumes_clause(Y, X), !, fail.
add_to_ancestors(Y, [X|R], U):- subsumes_clause(X, Y), !,
	add_to_ancestors(Y, R, U).
add_to_ancestors(Y, [X|R], [X|U]):- add_to_ancestors(Y, R, U).
add_to_ancestors(Y, [], [Y]).
%
complementary(~X, X):-!.
complementary(X, ~X).
%
remove_tautology_clauses([X|Y], Y0):- has_complementary_pair(X), !,
	remove_tautology_clauses(Y, Y0).
remove_tautology_clauses([X|Y], [X|Y0]):- remove_tautology_clauses(Y, Y0).
remove_tautology_clauses([], []).

% ?- pure_literal_rule([[a,~b], [b, c(X)], [~c(Y), ~d], [~d]], R).
%@ R = [[b, c(X)]].
% ?- pure_literal_rule([a/0, d/0], [[a,~b], [b, c(X)], [~c(Y), ~d], [~d]], R).
%@ R = [[b, c(X)]].

pure_literal_rule(X, Y):-  once(pure_literals_of_clauses(X, D)),
	once(pure_literal_rule(D, X, Y)).
%
pure_literal_rule(D, [C|X], Y):-
	(	member(A, C),
		(	A = ~(A0),
			functor(A0, F, N)
		;	functor(A, F, N)
		),
		memberchk(F/N, D),
		Y0 = Y
	;	Y = [C|Y0]
	),
	!,
	pure_literal_rule(D, X, Y0).
pure_literal_rule(_, [], []).

% ?- pure_literals_of_clauses([[a, ~b, b, ~c(Y)]], R).

pure_literals_of_clauses(X, Y):- literal_dict_list(X, [], R),
	setof(F/N, Pos^Neg^C^(member(C, X),
			      member([F, N, Pos, Neg], R),
			      (var(Pos); var(Neg))),
		Y),
	!.
pure_literals_of_clauses(_, []).

		/*********************
		*     subsumption    *
		*********************/

% ?- merge_clause([a(f(A),B), a(B, A), a(C,C)], X).
merge_cnf(X, Y):- once(merge_cnf_cnf(X, [], Y)).
%
merge_cnf_cnf([A|R], U, V):-
	merge_cnf_clause(U, A, U0),
	merge_cnf_cnf(R, U0, V).
merge_cnf_cnf([], X, X).

%
merge_cnf_clause([A|R], C, S):-
	(	subsumes_clause(C, A),
		S = [A|R]
	;	subsumes_clause(A, C),
		merge_cnf_clause(R, C, S)
	;	S = [A|S0],
		merge_cnf_clause(R, C, S0)
	).
merge_cnf_clause([], C, [C]).

%
merge_clause(X, Y):- merge_clause_clause(X, [], Y).

% ?- merge_clause_clause([a,b,c], [a,b,d], X).
merge_clause_clause([L|R], C,  S):-
	merge_clause_literal(C, L, C0),
	merge_clause_clause(R, C0, S).
merge_clause_clause([], C, C).

% ?- merge_clause_literal([a(I,J),b,c],a(K,K), R).
merge_clause_literal([A|R], C, S):-
	(	subsumes_term(C, A),
		S = [A|R]
	;	subsumes_term(A, C),
		merge_clause_literal(R, C, S)
	;	S = [A|S0],
		merge_clause_literal(R, C, S0)
	).
merge_clause_literal([], C, [C]).

%  ?- subsumes_clause([a,b], [a,b,c]).
%  ?- subsumes_clause([a,b,c], [a,b]).
subsumes_clause(X, Y):- subsumes_clause_(Y, X).

subsumes_clause_([X|R], C):- member(A, C),
	subsumes_term(A, X),
	!,
	subsumes_clause_(R, C).
subsumes_clause_([], _).

% [2015/03/15]
% ?- negation_of_cnf([[a,b,c]], R), negation_of_cnf(R, S).
% ?- negation_of_cnf([[a,b],[c,d]], R), negation_of_cnf(R, S).
% ?- negation_of_cnf([[a,b],[c,d], [e,f]], R), negation_of_cnf(R, S).
negation_of_cnf(Cs, Neg):- 	choices(Cs, Ds0),
	maplist(merge_clause, Ds0, Ds1),
	maplist(maplist(complementary), Ds1, Ds),
	merge_cnf(Ds, Neg).

% ?- literal_dict_list([[a,~b,b,c]], [], R).
literal_dict_list([X|Y], D, D0):-
	literal_dict(X, D, D1),
	literal_dict_list(Y, D1, D0).
literal_dict_list([], D, D).

% ?- literal_dict([a,~b,b,c], [], R).
literal_dict([X|Y], D, D0):-
	(	X = ~(A)
	->	Neg = on
	;	A = X,
		Pos = on
	),
	functor(A, F, N),
	(	memberchk([F, N, Pos, Neg],  D)
	->	D1 = D
	;	D1 = [[F, N, Pos, Neg]|D]
	),
	literal_dict(Y, D1, D0).
literal_dict([], D, D).

% ?- set_of_support_proof([[a,~b], [a,b], [~a,b], [~a,~b]], [], 200, 5, S, Trace).
% ?- set_of_support_proof([[a,~b], [a,b], [~a,b], [~a,~b]], [[a,b]], 200, 5, S, Trace).
% ?- set_of_support_proof([[a,~b], [a,b], [~a,b], [~a,~b]], [[xxx]], 200, 5, S, Trace).
set_of_support_proof(SOS, Centers, D, N, S, Trace):-
	abolish(fol:input_clause, 1),
	setup_input_clauses(SOS, X0),
	X0 \== [],
	maplist(pred([C]:- assert(input_clause(C))), X0),
	(	Centers \== [], member(C0, Centers)
	;	Centers  == [], input_clause(C0)
	),
	iterated_deepening( fol:linear_prove(C0, [], Trace, []),
			   D, N, S).
set_of_support_proof(_, _, _, _, fail, []).

% ?- fol:number_to_var_name(6, X).
number_to_var_name(I, X):- between(1, 6, I), memberchk(I-X, [1-x,2-y,3-z,4-u,5-v,6-w]).
number_to_var_name(I, X):- I >= 7, J is I-6, atomic_concat(x, J, X).


%%%%% example use of the prover fol.
% ?- prove_eq_leq(all(x,all(y, x=y -> same(x,y)))->all(x, same(x,x)), X).
% ?- prove(all(x,all(y, x=y -> same(x,y)))->all(x, same(x,x)), X).
% ?- prove(a, X).

axiom_equiv((all(x, x=x),
	     all(x, all(y, (x=y -> y=x))),
	     all(x, all(y, all(z, (x=y, y=z) -> x=z))))).

axiom_order((all(x, x=<x),
	     all(x, all(y, (x=<y, y=<x) ->  x=y)),
	     all(x, all(y, all(z, (x=<y, y=<z) -> x=<z))))).

prove_eq_leq(X, Y):- axiom_equiv(A), axiom_order(B), prove((A,B) -> X, Y).
