:- module(fol_prover, [prove/1, prove/2]).

% :- use_module(library(clpfd)).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-array')).

:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<-').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<--').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<-->').
:- current_op(P, T, (->)), member(T,[xfy, yfx, xfx]), op(P, T, '<->').
:- op(200, fy, ~).
:- op(400, yfx, &).

% ?- zdd((X<<pow([1,2,3]), psa(X))).
% ?- prove(all(y, some(x, f(x, y)) -> some(x, all(y, f(x, y))))). % false
% ?- prove(some(y, all(x, f(x, y)) -> all(x, some(y, f(x, y))))).
% ?- prove(all(y, some(x, f(x, y)) -> some(x, all(y, f(x, y))))). % false
% ?- prove(all(x, p(x)) -> some(z, all(y, p(y)))).
% ?- prove((some(X, p(X))->all(X,p(X)))).		% false.
% ?- refute(-(some(X, p(X))->all(X,p(X)))).		% false.
% ?- prove((some(X, p(X))->some(X,p(X)))).
% ?- prove((all(X, some(Y, p(X, Y)))->some(Y, all(X, p(X, Y))))).  % false.
% ?- prove((some(x, some(y, (p(x, y) -> all(x, all(y, p(x, y)))))))).
% ?- prove((some(x, some(y, (p(x, y) -> all(u, all(v, p(u, v)))))))).
% ?- prove(all(Y, p(Y))-> some(X, (p(X),  -p(f(X))))).   % time_limit_exceeded
% ?- prove(all(Y, p(Y))-> some(X, (p(X),  -p(f(X)))), 100).  %time_limit_exceeded
% ?- prove((p -> p) ; (q->q)).
% ?- prove(all(X, p(X))-> all(y, p(y))).
% ?- prove(all(x, p(x))-> some(y, p(y))).
% ?- prove(some(x, p(x))-> all(y, p(y))).	% invalid
% ?- prove(all(x, p(x))-> some(y, p(y))).
% ?- time(repeat(valid_formula(P), (writeln(P), once(prove(P)), writeln(proved)))).   % pass.
% ?- time(repeat(invalid_formula(P), (writeln(P), once(prove(P)), writeln(proved)))).    % pass.
% ?- time(prove(all(z, some(y, all(x, (f(x, y) == (f(x,z), -f(x,x)))))) -> -(some(z, all(x, f(x,z)))))).
% ?- time(prove(all(z, some(y, all(x, (f(x, y) == (f(x,z), -f(x,x)))))) -> -(some(z, all(x, f(x,z)))))).
% ?- prove(all(x, some(p(x, y)))->some(y, all(x, p(x,y)))).  % invalid.
% ?- prove(some(y, all(x, p(x, y)))->all(x, some(y, p(x,y)))).

% ?- repeat(valid_formula(P), (nl, writeln(P), prove(P))).
valid_formula((a * b) + (a * -b) + (-a * b) + (-a * -b)).
valid_formula((-(-a) -> a) * (a -> -(-a))).		% Double Negation
valid_formula( ((a * b) -> -((-a + -b))) * (-((-a + -b)) -> (a * b))).
%								De Morgan’s Law
valid_formula( (-b -> f) * ((b *f) -> -i) * ((i + -b) -> -f) -> b).
valid_formula(a + -a).
valid_formula(((a -> b) -> a) -> a).  % Peirce’s Law
valid_formula(-(-a)->a).
valid_formula(a -> a).
valid_formula(a -> (b -> a)).
valid_formula((a->b)->((b->c)->(a->c))).
valid_formula(a->((a->b)->b)).
valid_formula((a*(a->b))->b).
valid_formula((a->b)->(-b -> -a)).
valid_formula((a->b)->((b->c)->(a->c))).

% ?- repeat(invalid_formula(P), (nl, writeln(P), prove(P))).
invalid_formula(a).
invalid_formula((a->b)->a).
invalid_formula((a->b)->(b->a)).
invalid_formula(a -> (b -> c)).
invalid_formula(-( (-b -> f) * ((b *f) -> -i) * ((i + -b) -> -f) -> b)).

		/**************
		*     Note    *
		**************/

% Note: all(x, p== f(x)) == (p==all(x, f(x))) is provable in LK.
% Checked.

% ?- fol_cnf(all(x, p== f(x))== (p==all(x, f(x))),  X),
%	numbervars(X), maplist(writeln, X).
% ?- prove(all(x, p== f(x))== (p==all(x, f(x)))).
% ?- prove(all(x, p== f(x)) -> (p==all(x, f(x)))).
% ?- prove(all(x, p-> f(x)) <- (p==all(x, f(x)))).
% ?- prove(all(x, p<- f(x)) <- (p<-all(x, f(x)))).  % invalid
% ?- prove(all(x, p<- f(x)) <- (p==all(x, f(x)))).  % invalid
% ?- prove((p->all(x, f(x)))-> all(x, f(x)->p)).	% Invalid
% ?- prove((p->all(x, f(x)))-> all(x, p->f(x))).
% ?- prove((p==all(x, f(x)))-> all(x, f(x)->p)).	% Invalid.
% ?- prove((p->all(x, q(x) * -q(x)))-> all(x, ((q(x)* -q(x))->p))).
% ?- prove(((p* -p)->all(x, q(x) * -q(x)))-> all(x, ((q(x)* -q(x))->(p* -p)))).
% ?- prove(-(p->all(x, f(x))); all(x, f(x)->p)).	% invalid
% ?- prove(-( -p;all(x, f(x))); all(x, f(x)->p)).	% invalid
% ?- prove( p * some(x, -f(x)); all(x, -f(x)| p)).  % invalid
% ?- prove( p * some(x, -f(x)); all(x, -f(x))).		% invalid
% ?- prove( p * -f(a); all(x, -f(x))).				% invalid
% ?- prove( -f(a); all(x, -f(x))).					% invalid
% ?- prove(all(x, p==a(x))==(p==all(x, a(x)))).		% time_limit_exceeded
% ?- prove(all(y, p==f(y))==(p==all(x, f(x)))).		% time_limit_exceeded

		/**********************************************
		*     (p==all(x, f(x))) == all(x, p==f(x))    *
		*	is not a theorem.						  *
		**********************************************/
% ?- fol_cnf(-((p==all(x, f(x))) == all(x, p==f(x))), R), cnf_show(R).
% ?- fol_cnf((-p)*(p==all(x, f(x))), R), cnf_show(R).
% ?- fol_cnf(a, X).
% ?- fol_cnf(all(x, p==f(x))==(p==all(x, f(x))), R), cnf_show(R).
% ?- fol_cnf(- (all(x, p==f(x))==(p==all(x, f(x)))), R), cnf_show(R).

fol_cnf(F, CNF):-
	fol_skolemized_nnf(F, Cs),
	factoring(Cs, Csx),
	cnf_normal(Csx, Csy),
 	rm_tautology(Csy, Cs0),
	cnf_normal(Cs0, Cs1),
	cnf_variant_normal(Cs1, CNF).

q_assoc(X, Assoc, Y):- member(U-V, Assoc), X==U, !, Y = V.
%
q_subst(Assoc, A, A0):-
	(	q_assoc(A, Assoc, A0) ->  true
	;	var(A) -> A0 = A
	;	mapargs(q_subst(Assoc), A, A0)
	).

% ?- fol_skolemized_nnf(a;b, G).
% ?- fol_skolemized_nnf(all(x, b(x)*c(x)), G).
% ?- fol_skolemized_nnf(some(x, b(x);c(x)), G).
% ?- fol_skolemized_nnf(all(y, some(x, b(x);c(y))), G).
% ?- fol_skolemized_nnf(all(y, some(x, b(x);c(x,y))), G).
% ?- fol_skolemized_nnf(- (all(y, f(y)==p) == (p==all(x, f(x)))), G), maplist(writeln, G).

% [2022/03/27]
% The formula - (all(y, f(y)==p) == (p==all(x, f(x))))
% is SATISFIABLE, i.e. the formula
% (all(y, f(y)==p) == (p==all(x, f(x)))) is INVALID.
% D = {1,2,3} then
% left   p== f(1) & p==f(2) & p==f(3)
% right  p== (p(1) & p(2) & p(3))
% but left is not equivalent to right.

% Debug.
% ?-  F = (all(A, p==f(A))==(p==all(B, f(B)))), prove(F).  % time limit.
% ?-  F = ((p == all(A, f(A))) == all(B, p== f(B))), prove(F). % Debug.
% ?-  F = ((p == all(A, f(A))) -> all(B, p -> f(B))), prove(F). % valid
% ?-  F = ((p == all(A, f(A))) -> all(B, f(B) -> p)), prove(F). % Invalid
% ?-  F = ((p == all(A, f(A))) -> all(B, p -> f(B))), prove(F). % valid
% ?-  F = ((p == all(A, f(A))) -> all(B, f(B) -> p)), prove(F). % invalid

fol_skolemized_nnf(F, G):- intern(F, H),
	neg_normal(H, N),
	rename_bounded_vars(N, [], J),
	elim_quatifiers(J, K, [], Eqs, []),
	maplist(call, Eqs),
	nnf_cnf(K, G).

% ?-  rename_bounded_vars(a, [], X).
% ?-  rename_bounded_vars(A, [], X).
% ?-  rename_bounded_vars(some(y, all(x, p(x);q(y))), [], X).
% ?-  neg_normal(-some(y, all(x, p(x);q(y))), U),
%	rename_bounded_vars(U, [], X).
% ?-  neg_normal(-some(x, all(x, p(x))*q(x)), U),
%	rename_bounded_vars(U, [], X).

rename_bounded_vars(X, _, _):- var(X), !, throw(unboud_input).
rename_bounded_vars(some(X, B), Assoc, some(X0, B0)):-!,
	rename_bounded_vars(B, [X-X0|Assoc], B0).
rename_bounded_vars(some(X, D, B), Assoc, some(X0, D0, B0)):-!,
	rename_bounded_vars((B,D), [X-X0|Assoc], (B0, D0)).	% comma for pairing.
rename_bounded_vars(all(X, B), Assoc, all(X0, B0)):-!,
	rename_bounded_vars(B, [X-X0|Assoc], B0).
rename_bounded_vars(all(X, D, B), Assoc, all(X0, D0, B0)):-!,
	rename_bounded_vars((B,D), [X-X0|Assoc], (B0,D0)). % comma for pairing.
rename_bounded_vars(A*B, Assoc, A0*B0):-!,
	rename_bounded_vars(A, Assoc, A0),
	rename_bounded_vars(B, Assoc, B0).
rename_bounded_vars(A;B, Assoc, A0; B0):-!,
	rename_bounded_vars(A, Assoc, A0),
	rename_bounded_vars(B, Assoc, B0).
rename_bounded_vars(-X, Assoc, -X0):-!,
	rename_bounded_vars(X, Assoc, X0).
rename_bounded_vars(X, Assoc, Y):- X=..[F|Xs],
	maplist(q_subst(Assoc), Xs, Ys),
	Y =..[F|Ys].

% ?- neg_normal(- (a* b), X).
% ?- neg_normal(a * -B, X).
% ?- neg_normal(-(-(X)), Y).
% ?- neg_normal(- - -a, Y).
% ?- neg_normal(X, Y).
% ?- neg_normal(some(X,m(X)* -(all(A, n(A), -(Y)))), U).
% ?- neg_normal(-(x==y), U).
% ?- neg_normal(x==y, U), neg_normal(-U, V).
% ?- write_canonical(-a * -b).

neg_normal(X, _)	:- var(X), !, throw(unboud_input).
neg_normal(-X, -X)	:- var(X), !, throw(unboud_formula).
neg_normal(-(-X), Z):-!, neg_normal(X, Z).
neg_normal(-(X;Y), Z):-!, neg_normal(-X* -Y, Z).
neg_normal(-(X*Y), Z):-!, neg_normal(-X; -Y, Z).
neg_normal(-(X->Y), Z):-!, neg_normal(-(-X; Y), Z).
neg_normal(-(<-(X,Y)), Z):-!, neg_normal(-(Y->X), Z).
neg_normal(-(X==Y), Z):-!, neg_normal(- ((X->Y)*(Y->X)), Z).
neg_normal(-all(X,Y), Z):-!, neg_normal(some(X, -Y), Z).
neg_normal(-all(X, D, Y), Z):-!, neg_normal(some(X, D, -Y), Z).
neg_normal(-some(X,Y), Z):-!, neg_normal(all(X, -Y), Z).
neg_normal(-some(X, D, Y), Z):-!, neg_normal(all(X, D, -Y), Z).
neg_normal(X;Y, X0;Y0):-!, neg_normal(X, X0), neg_normal(Y, Y0).
neg_normal(X*Y, X0*Y0):-!, neg_normal(X, X0), neg_normal(Y, Y0).
neg_normal(X==Y, Z):-!, neg_normal((X->Y)*(Y->X), Z).
neg_normal(X->Y, Z):-!, neg_normal(-X;Y, Z).
neg_normal(<-(X,Y), Z):-!, neg_normal(Y->X, Z).
neg_normal(all(X, Y), all(X, Z)):-!, neg_normal(Y, Z).
neg_normal(all(X, D, Y), all(X, D, Z)):-!, neg_normal(Y, Z).
neg_normal(some(X, Y), some(X, Z)):-!, neg_normal(Y, Z).
neg_normal(some(X, D, Y), some(X, D, Z)):-!, neg_normal(Y, Z).
neg_normal(X, X).

% ?- elim_quatifiers(a, X, [], E, []).
% ?- elim_quatifiers(some(x, p(x)), X, [], E, []).
% ?- elim_quatifiers((some(x, -p(x))* some(y, p(y))), X, [], E, []).
% ?- elim_quatifiers((some(x, p(x))*some(x, p(x))), X, [], E, []).

elim_quatifiers(some(X, B), B0, Vs, [X=SKterm|Eqs], Eqs0):-!,
	gensym(sk, F),
	SKterm=..[F|Vs],
	elim_quatifiers(B, B0, Vs, Eqs, Eqs0).
elim_quatifiers(all(X, B), B0, Vs, Eqs, Eqs0):-!,
	elim_quatifiers(B, B0, [X|Vs], Eqs, Eqs0).
elim_quatifiers(A*B, A0*B0, Vs, Eqs, Eqs0):-!,
	elim_quatifiers(A, A0, Vs, Eqs, Eqs1),
	elim_quatifiers(B, B0, Vs, Eqs1, Eqs0).
elim_quatifiers(A;B, A0;B0, Vs, Eqs, Eqs0):-!,
	elim_quatifiers(A, A0, Vs, Eqs, Eqs1),
	elim_quatifiers(B, B0, Vs, Eqs1, Eqs0).
elim_quatifiers(X, X, _, Eqs, Eqs).

% ?- extern(all(X, some(Y, love(X,Y); love(Y,X))), A), print(A).
canonical_op(C, D, Cs, E):-
	member(M, Cs),
	memberchk(C, M),
	member(D, E),
	memberchk(D, M).
%
connectives([ [ ',', (*),  (&), and],
	      [;, + , '|', or],
	      [(\+), ~, not, -],
	      [(->), imply, =>],
	      [<-],
	      [iff, <->, == , =:= , <=>, <-->],
	      [every, all, forall],
	      [some, exists, ex] ]).

% ?- write((a, b|c, d)).
internal_op([*, ;, -, ->, ==, all, some]).

% ?- internal_op(and, X).
internal_op(C, D):- connectives(Cs),
	internal_op(E),
	canonical_op(C, D, Cs, E).
%
external_op([*, ;, -, ->, ==, all, some]).

% ?- external_op((or), X).
external_op(C, D):- connectives(Cs),
	external_op(E),
	canonical_op(C, D, Cs, E).

% ?- intern(some(x, all(y, a(x, y))), R).
% ?- intern(some(x, all(y, a(x, (y,z)))), R).
% ?- intern(some(x, all(y, all(x, (y,z)))), R).
% ?- intern(some(x, p(x)->q(x)), R).
% ?- intern(x=y, R).
% ?- intern((a,b), R).
% ?- intern(and(a,b), R).
% ?- intern(or(a,b), R).
% ?- intern(a == b, R).
% ?- intern(-(a == b), R).
% ?- intern(some(x, d(x), p(x)), R).
% ?- intern(-some(x, d(x), p(x)), R).
% ?- nnf_skolem(-(a == b), R).
% ?- intern(or(a,b), R), extern(R, E).
% ?- extern(imply(a,imply(b,c)), R), extern(R, E).

intern(X, X):- var(X), !.
intern(X->Y, Z):-!, intern(-(X);Y, Z).
intern(X==Y, Z):-!, intern((X*Y);(-X)*(-Y), Z).
intern(X, Y):- X =..[F|As], internal_op(F, G), !,
		maplist(intern, As, Bs),
		Y =..[G|Bs].
intern(X, X).

%
extern(X, X):- var(X), !.
extern(X, Y):- X =..[F|As], external_op(F, G), !,
		maplist(extern, As, Bs),
		Y =..[G|Bs].
extern(X, X).

		/*********************
		*     Resolution     *
		*********************/

% ?- time((valid_formula(F), prove(F), fail; true)).
% ?- time((invalid_formula(F), prove(F), fail; true)).
% ?- BarberSentence =
%	(all(x, barber(x) ==  all(y,  -shave(y, y) == shave(x, y)))),
%	time(prove(( BarberSentence -> -some(x, (barber(x), shave(x, x)))))).
% ?- Barber =
%	(all(x, barber(x) ==  all(y,  -shave(y, y) == shave(x, y)))),
%	DontShave = all(x, (barber(x)-> -shave(x,x))),
%	time(prove((Barber -> DontShave))).

% ?- prove(all(x, (p(x)->(q(x)->p(x))))).
% ?- prove((p->q)->p).	% false
% ?- prove(all(x, p(x)->p(x))).
% ?- prove(all(x, p(x)->(q(x)->p(x)))).
% ?- prove(all(x, all(y, p(x, y)))->some(x, some(y, p(x, y)))).
% ?- prove(all(x, all(y, p(x, y)))->some(x, some(y, p(x, y)))).
% ?- F = ( -((-b -> f) * ((b * f) -> -i) * ((i + -b) -> -f) -> (i * f))),
%	prove(F, 5). % invalid
% ?- prove(-some(x, all(y, (f(y,x) == -f(y,y))))).
% ?- cnf_normal([[a,a], [b,b], [c, b, a]], R).
% ?- cnf_normal([[a(X),a(Y)], [b(X),b(X)], [c(Y), b(Z), a(U)]], R).
% ?- prove((b==a)-> ((a->b) *(b->a))).
% ?- prove((b==a) == (a==b)).

cnf_normal(X, Y):-maplist(cls_sort, X, X0), cnf_sort(X0, Y).
%
cnf_variant_normal(X, Y):-maplist(cls_sort, X, X0),
	elim_variant(X0, X1),
	cnf_sort(X1, Y).

% ?- fol_cnf(all(X, (a(X)*b(X))), R).
% ?- fol_cnf(all(X, (-a(X); b(X);a(X))), R).
% ?- fol_cnf(all(X, (g,(-a(X); b(X);a(X)))), R).
% ?- fol_cnf(a; some(x, b(x)), X).
% ?- fol_cnf(all(X, (-a(X); b(X);a(X))), R).

nnf_cnf(A*B,C):-!,
	nnf_cnf(A,A0),
	nnf_cnf(B,B0),
	cnf_union(A0,B0,C0),
	cnf_sort(C0, C).
nnf_cnf(A;B, C):-!,
	nnf_cnf(A,A0),
	nnf_cnf(B,B0),
	cnf_distribute(A0,B0,C).
nnf_cnf(A, [[A]]).

%!  fol_ltr_compare(-C, +X, +Y) is det.
%   Compare literals X and Y, and C is unified with
%	<, =, > as the standard compare/3.
%   ex. a<b, a < -a,  -a < b,   -a < -b.

fol_ltr_compare(C, -X, -Y):-!, compare(C, X, Y).
fol_ltr_compare(C, X, -Y):-!, compare(C0, X, Y),
		(	C0 == (=) ->  C =(<)
		;	C = C0
		).
fol_ltr_compare(C, -X, Y):-!, compare(C0, X, Y),
		(	C0 == (=) -> C = (>)
		;	C = C0
		).
fol_ltr_compare(C, X, Y):- compare(C, X, Y).

% ?- cls_sort([-a, b, a, -b, a], R).
% ?- cls_sort([], R).
% ?- cls_sort([-a(X), a(X), -b, a], L).
cls_sort(X, Y):- predsort(fol_ltr_compare, X, Y).

% ?- cls_compare(C, [a],[b]).
% ?- cls_compare(C, [a],[]).
% ?- cls_compare(C, [],[]).
cls_compare(C, X, Y):- length(X, L), length(Y, M),
	(	L = M -> cls_compare_same_length(C, X, Y)
	;	L < M -> C = (<)
	;	C = (>)
	).
%
cls_compare_same_length(=, [], []):-!.
cls_compare_same_length(C, [X|Xs], [Y|Ys]):- fol_ltr_compare(D, X, Y),
	(  D = (=) -> cls_compare_same_length(C, Xs, Ys)
	;  C = D
	).

% ?- ltr_check(a(1), S, X, R).
% ?- ltr_check(-a(1), S, X, R).
ltr_check(-X, S, X, P/N):-!, S = (-), functor(X, P, N).
ltr_check(X, S, X, P/N):- S = (+), functor(X, P, N).

% ?- cnf_sort([[a(X), -b(A), c(A)],[-a(X), b(Y)]], X).
%@ X = [[-a(X), b(Y)], [a(X), -b(A), c(A)]].
cnf_sort(X, Y):- predsort(cls_compare, X, Y).

% ?- unord_tautology([a(X), -a(X), c, -d]).
% ?- unord_tautology([a(X), -a(Y), c, -d]).  % false
unord_tautology([-A|As]):- member(B, As), A == B, !.
unord_tautology([A|As]):- member(B, As), (-A) == B, !.
unord_tautology([_|As]):- unord_tautology(As).

tautology(X):- ord_tautology(X).

% ?- ord_tautology([a, b, -b, d]).
ord_tautology([P, -Q|_]):- P == Q, !.
ord_tautology([_|R]):- ord_tautology(R).

%
tautology_law(X, Y):- rm_tautology(X,Y).

% ?- rm_tautology([[a, -a, b], [c]], R).
rm_tautology([], []):-!.
rm_tautology([C|Cs], Ds):- tautology(C), !,
	rm_tautology(Cs, Ds).
rm_tautology([C|Cs], [C|Ds]):- rm_tautology(Cs, Ds).

%
unify(X, Y):- unify_with_occurs_check(X, Y).

		/****************
		*     prove    *
		****************/

% ?- time((invalid_formula(F), once(prove(F)), fail; true)).
% ?- invalid_formula(F), prove(F).
% ?- prove( ((-b->f)*(b*f-> -i)*(i+ -b-> -f)->i*f)).
% ?- prove( - ((-b->f)*(b*f-> -i)*(i+ -b-> -f)->i*f)).
% ?- time((invalid_formula(F), writeln(F), prove(F), fail)).
% ?- repeat(invalid_formula(F),( writeln(F), prove(F))).

% ?- Barber =
%	(all(x, barber(x) ==  all(y,  -shave(y, y) == shave(x, y)))),
%	DontShave = all(x, (barber(x)-> -shave(x,x))),
%	time(prove((Barber -> DontShave))).

% ?- Barber =
%	(all(x, barber(x) ==  all(y,  -shave(y, y) == shave(x, y)))),
%	DontShave = all(x, (barber(x)-> -shave(x,x))),
%	time(prove((Barber -> DontShave))).

% ?- prove( (p == q) -> ((p->q), (q->p))).
% ?- prove( ((p==q), (r==q)) -> (p==r)).
% ?- prove( (some(X, (p == f(X))) ->  (some(X1, (f(X1) -> p))))).  % pass
% ?- prove( (some(X, (p == f(X))) ->  (some(X1, (f(X1) -> p)), some(X2, f(X2)-> p)))).
% ?- prove( (some(X, (p == f(X))) ->  (some(X1, (p -> f(X1))) -> some(X2, (f(X2) -> p))))).
% ?- prove( (some(X, (p == f(X))) ->  (- some(X1, (p -> f(X1))) ; some(X2, (f(X2) -> p))))).
% ?- prove( and(some(X, (p == f(X))), some(X1, (p -> f(X1)))) ->  some(X2, (f(X2) -> p))).
% ?- prove( some(X, (p == f(X))) ->  some(X1, p -> f(X1))).
% ?- prove( some(X, (p == f(X))) ->  some(X1, f(X1) -> p)).
% ?- prove( some(X, (p == f(X))) ->  some(X1, f(X1) == p)).
% ?- time((valid_formula(F), writeln(F), prove(F), writeln(proved), fail; true)).

% ?- prove( all(x, p == f(x)) ->  (p == all(y, f(y)))).
% ?- prove( (p == all(x, f(x))) ->  all(x, p -> f(x))).
% ?- prove( (p == all(x, f(x))) ->  all(x, f(x)->p)).
% ?- prove( all(x, p == f(x)) ==  (p == all(y, f(y)))).
% ?- prove( all(x, p == f(x)) ==  (p -> all(y, f(y)))).
% ?- prove( (p == all(x, f(x))) ->  all(x, p == f(x))).
% ?- invalid_formula(F), writeln(F), prove(F, 1).
% ?- prove(a->b).
% ?- prove(a->a).
% ?- prove( all(x, p==f(x)) ->  (p== all(y, f(y)))).
% ?- prove((p== all(y, f(y))) -> all(x, p->f(x))).
% ?- prove( all(X, (p == f(X))) <- all(X1, f(X1) == p)).
% ?- prove((p== all(y, f(y))) <- all(x, p==f(x))).
% ?- prove((p== all(y, f(y))) -> all(x, p==f(x))).
% ?- prove((p== all(y, f(y))) -> all(y, f(y)==p)).
% ?- prove((p== all(y, f(y))) -> all(y, f(y)->p)).
% ?- prove( (p == all(x, f(x))) ->  all(y, f(y) == p)).
% ?- prove( (p == all(x, f(x))) ->  all(y, p==f(y))).
% ?- call_with_time_limit(100, prove((p== all(y, f(y))) == all(x, p==f(x)))).

cnf_show(X, Y):- writeln(X), !, cnf_show(Y).
% ?- cnf_show([[a],[b]]).
cnf_show(Cs):- forall(member(X, Cs), (numbervars(X), writeln(X))),
	writeln('---------\n').

% 1)
% ?- prove(all(x, p) == p).
% ?- prove(some(x, p) == p).
% 2)
% ?- prove(all(x, p(x)) == all(y, p(y))).
% 3)
% ?- prove(a*all(x, p(x)) == all(x, a*p(x))).
% ?- prove((a;all(x, p(x))) == all(x, a;p(x))).
% 4)
% ?- prove((a;all(x,p(x))) == all(x, a; p(x))).
% ?- prove((a*all(x,p(x))) == all(x, a*p(x))).
% 5)
% ?- prove(all(x, b(x))*all(x, c(x)) == all(x, b(x)*c(x))).
% ?- prove((some(x, b(x));some(x, c(x))) == some(x, b(x);c(x))).
% 6)
% ?- prove((all(x, b(x)); all(x, c(x))) -> all(x, b(x);c(x))).
% ?- prove((all(x, b(x)); all(x, c(x))) <- all(x, b(x);c(x))).
% ?- prove((some(x, b(x))* some(x, c(x))) <- some(x, b(x)*c(x))).

% ?- prove((some(x, b(x))* some(x, c(x))) -> some(x, b(x)*c(x))).
% 7)
% ?- prove((all(x, all(y,  b(x, y))) == all(y, all(x, b(x, y))))).
% ?- prove((some(x, some(y, b(x, y))) == some(y, some(x, b(x, y))))).
% 8)
% ?- prove((some(x, all(y, b(x, y))) -> all(y, some(x, b(x, y))))).
% ?- prove((some(x, all(y, b(x, y))) == all(y, some(x, b(x, y))))).
% 9)
% ?- prove((all(x, b(x))-> some(y, b(y)))).
% ?- prove((some(x, b(x))-> all(y, b(y)))).
% 10)
% ?- prove(-all(x, b(x)) == some(x, -b(x))).
% ?- prove(-some(x, b(x)) == all(x, -b(x))).
% 11
% ?- prove((a -> all(x, b(x))) == all(x, a->b(x))).
% ?- prove((a -> some(x, b(x))) == some(x, a->b(x))).
% 12
% ?- prove((all(x, b(x)) -> a) == some(x, b(x)->a)).
% ?- prove((some(x, b(x))->a) == all(x, b(x)->a)).
% 13
% ?- call_with_time_limit(100, prove(some(x, b(x) -> c(x)) == (all(x, b(x))->all(x, c(x))))).  % time_limit_exceeded
% 14
% ?- prove(all(x, b(x) -> c(x)) -> (all(x, b(x))->all(x, c(x)))).
% 15
% ?- prove(all(x, b(x) -> c(x)) -> (some(x, b(x))->some(x, c(x)))).
% ?- call_with_time_limit(100, prove(all(x, p == f(x)) ==  (p == all(y, f(y))))).
% ?- prove(all(x, p == f(x)) ->  (p == all(y, f(y)))).
% ?- prove(all(x, p == f(x)) <-  (p == all(y, f(y)))).

set_when_gt(I, J, S):-
	(	I =< J -> true
	;	nb_setarg(1, S, 1)
	).

apply_options([], X, X).
apply_options([P|Ps], X, Y):- call(P, X, Z),
	apply_options(Ps, Z, Y).

		/***********************
		*     Linear Solver    *
		***********************/

% ?- prove(a).
% ?- prove(a->a).
% ?- prove(a->a, 10).

prove(P):- fol_linear_prove(P, R, 30), writeln(R).
%
prove(P, Limit):- fol_linear_prove(P, R, Limit), writeln(R).

% ?- fol_linear_prove(a->a, X, 30).
% ?- fol_linear_prove(a, X, 30).
fol_linear_prove(P, R, Limit) :- fol_cnf_normal(-P, N),
		linear_solve_with_time_limit(N, R, Limit).

% ?- fol_cnf_normal((a; -a), X).
fol_cnf_normal(P, N):- fol_cnf(P, U),
		apply_options(
		[	pure_literal_law,
			factoring,
			elim_variant,
			cnf_normal
		 ], U, N).

% ?- prove(a, X).
% ?- linear_solve([[a],[-a]], R).
% ?- linear_solve([[a],[b]], R).
% ?- linear_solve([[-a],[-b], [a,b]], R).
linear_solve(Cs, R):- linear_solve_with_time_limit(Cs, R, 30).
%
linear_solve_with_time_limit(Cs, R, Limit):-
	catch(iter_deep_solve(Cs, R, Limit), E, R = E).
%
iter_deep_solve(Cs, R, Limit):-
	call_with_time_limit(Limit, linear_solve(Cs, 2, R)).

%
linear_solve(Cs, D, R):-  Flag = flag(0),
	(	solve_with_depth_limit(Cs, D, Flag) -> R = "Valid"
	;	Flag = flag(0) ->  R = "Invalid"
	;	D0 is 2 * D,
		linear_solve(Cs, D0, R)
	).
%
solve_with_depth_limit(Cs, D, Flag):- Flag = flag(0),
	select(C, Cs, Ds),
	linear_solve(C, Ds, 0, D, Flag).

%
linear_solve([], _, _, _, _):-!.
linear_solve(_, _, J,  K, Flag):- J > K, !,  nb_setarg(1, Flag, 1), fail.
linear_solve(E, Anc, J, K, Flag):- J =< K,
		E = [A|As],
		copy_term(E, E0),
	 	ltr_check(A, Sa, A0, F/N),
		member(C, Anc),
		select(H, C, C0),
		ltr_check(H, Sh, H0, F/N),
		Sa \== Sh,
		copy_term(H0 + C0, G + D0),
		unify_with_occurs_check(A0, G),
		cls_sort(As, As0),
		cls_union(D0, As0, E1),
		\+ tautology(E1),
		\+ variant_memberchk(E1, Anc),
		J0 is J + 1,
		linear_solve(E1, [E0|Anc], J0, K, Flag).

%% set-of-support as given centers plus input (but without exact definition.)
linear_solve_center(Ps, Cs, R):- catch(iterative_linear_solve_center(Ps, Cs, R), E, R = E).
%
iterative_linear_solve_center(Ps, Cs, R):- Limit = 30,
	call_with_time_limit(Limit, linear_solve_center(Ps, Cs, 2, R)).

%
linear_solve_center(Ps, Cs, D, R):-  Flag = flag(0),
	(	solve_center_with_depth_limit(Ps, Cs, D, Flag) -> R = "Valid"
	;	Flag = flag(0) ->  R = "Invalid"
	;	D0 is 2 * D,
		linear_solve_center(Ps, Cs, D0, R)
	).
%
solve_center_with_depth_limit(Ps, Cs, D, Flag):- Flag = flag(0),
	member(C, Ps),
	linear_solve(C, Cs, 0, D, Flag).

%
variant_memberchk(X, [Y|_]) :- variant(X, Y), !.
variant_memberchk(X, [_|Ys]):- variant_memberchk(X, Ys).

% ?- cls_union([a(X), a(Y)],[a(Y), b(Z)], U).
cls_union([], X, X):-!.
cls_union(X, [], X):-!.
cls_union([X|U], [Y|V], W):- fol_ltr_compare(C, X, Y),
	(	C = (=)	->
		cls_union(U, V, W0),
		W = [X|W0]
	;	C = (<)	->
		cls_union(U, [Y|V], W0),
		W = [X|W0]
	;	cls_union([X|U], V, W0),
		W = [Y|W0]
	).

% ?- cls_meet([a(Y), a(X)], [a(Y), b(Z)], M).
cls_meet([], _, []):-!.
cls_meet(_, [], []):-!.
cls_meet([X|U], [Y|V], W):- fol_ltr_compare(C, X, Y),
	(	C = (=)	->	cls_meet(U, V, W0), W =[X|W0]
	;	C = (<)	->	cls_meet(U, [Y|V], W)
	;	cls_meet([X|U], V, W)
	).

% ?- subtract([a(X), a(Y)], [a(Z)], R).
% ?- cls_subtract([a(X), a(Y)], [a(Z)], R).
% ?- cls_subtract([a(X), a(Y)], [a(Y)], R).
cls_subtract([], _, []):-!.
cls_subtract(X, [], X):-!.
cls_subtract([X|U], [Y|V], W):- fol_ltr_compare(C, X, Y),
	(	C = (=)	-> cls_subtract(U, V, W)
	;	C = (<)	->
		cls_subtract(U, [Y|V], W0),
		W = [X|W0]
	;	cls_subtract([X|U], V, W)
	).

% ?- ltr_insert(b(Z),[a(Y), b(Z)], U).
% ?- ltr_insert(c(Z),[a(Y), b(Z)], U).
ltr_insert(L, [], [L]):-!.
ltr_insert(L, [X|Xs], Y):-  fol_ltr_compare(C, L, X),
	(	C=(=) -> Y = [X|Xs]
	;	C=(<) -> Y = [L|[X|Xs]]
	;	ltr_insert(L, Xs, Us),
		Y = [X|Us]
	).

% ?- cnf_distribute([[[a]]], [[[b]],[[c]]], R).
cnf_distribute([], _, []).
cnf_distribute([A|R], X, P):- maplist(cls_union(A), X, Q),
	cnf_distribute(R, X, R0),
	cnf_union(Q, R0, P).

% ?- cnf_union([[a(X), a(Y)]],[[a(Y), b(Z)]], U), cnf_union(U, U, W).
cnf_union([], X, X):-!.
cnf_union(X, [], X):-!.
cnf_union([X|U], [Y|V], W):- cls_compare(C, X, Y),
	(	C = (=)	->
		cnf_union(U, V, W0),
		W = [X|W0]
	;	C = (<)	->
		cnf_union(U, [Y|V], W0),
		W = [X|W0]
	;	cnf_union([X|U], V, W0),
		W = [Y|W0]
	).

% ?- cnf_meet([[a(X)],[b(Y)]], [[b(Y)],[c(Y)]], U).
% ?- cnf_meet([[a(X)],[b(Y)]], [[b(Z)],[c(Y)]], U).   % U = [].
cnf_meet([], _, []):-!.
cnf_meet(_, [], []):-!.
cnf_meet([X|U], [Y|V], W):-cls_compare(C, X, Y),
	(	C = (=)	->	cnf_meet(U, V, W0), W = [X|W0]
	;	C = (<)	->	cnf_meet(U, [Y|V], W)
	;	cnf_meet([X|U], V, W)
	).

% ?- cnf_subtract([[a(X), a(Y)]], [[a(Z)]], R).
% ?- cnf_subtract([[a(X)],[a(Y)]], [[a(Y)]], R).
cnf_subtract([], _, []):-!.
cnf_subtract(X, [], X):-!.
cnf_subtract([X|U], [Y|V], W):-cls_compare(C, X, Y),
	(	C = (=)	-> cnf_subtract(U, V, W)
	;	C = (<)	->
		cnf_subtract(U, [Y|V], W0),
		W = [X|W0]
	;	cnf_subtract([X|U], V, W)
	).

% ?- cls_insert([a,b], [[b,c]], X).
% ?- cls_insert([b(X),c(Y)], [[a(Z),b(U)]], X).
cls_insert(X, [], [X]):-!.
cls_insert(X, [Y|Ys], Z):- cls_compare(C, X, Y),
	(	C = (=) -> Z = [Y|Ys]
	;	C = (<) -> Z = [X|[Y|Ys]]
	;	cls_insert(X, Ys, Z0),
		Z = [Y|Z0]
	).

		/******************************************
		*     pure literal law, unit law, etc.    *
		******************************************/

% ?-pure_literal_law([[a]], X).
% ?-pure_literal_law([[a],[b]], X).
% ?-pure_literal_law([[a(X), -b(X)],[b(Y), -a(Y)]], X).
% ?-pure_literal_law([[a(X), b(X)],[a(Y), b(Y)]], R).
% ?-pure_literal_law([[a(X)]], R).

pure_literal_law(Cs, Ds):- 	pure_literal_signs(Cs, Ps),
	pure_literal_law(Cs, Ds, Ps).
%
pure_literal_law([], [], _):-!.
pure_literal_law([C|Cs], Ds, Ps):-
	member(X, C),
	ltr_check(X, _, _, P),
	memberchk(P, Ps),
	!,
	pure_literal_law(Cs, Ds, Ps).
pure_literal_law([C|Cs], [C|Ds], Ps):-	pure_literal_law(Cs, Ds, Ps).

% ?- pure_literal_signs([[a(X), b(X)], [c, d]], Ps).
% ?- pure_literal_signs([[a, b], [-a, d]], Ps).
pure_literal_signs(Cs, Ps):-
	positive_literal_signs(Cs, P),
	sort(P, P0),
	negative_literal_signs(Cs, N),
	sort(N, N0),
	subtract(P0, N0, A),
	subtract(N0, P0, B),
	union(A, B, Ps).

% ?- positive_literal_signs([[a, a], [-b, -c]], P).
positive_literal_signs(Cs, Ps):-
	findall(P,
			(	member(C, Cs),
				member(L, C),
				ltr_check(L, +, _, P)
			),
			Ps0
		   ),
	sort(Ps0, Ps).

% ?- negative_literal_signs([[a], [-b, -c]], P).
negative_literal_signs(Cs, Ps):-
	findall(P,
			(	member(C, Cs),
				member(L, C),
				ltr_check(L, -, _, P)
			),
			Ps0
		   ),
	sort(Ps0, Ps).

% ?- elim_variant([[a(X), a(Y)], [a(A), a(B)]], R).
% ?- elim_variant([[a, a], [a, a]], U).
elim_variant([], []):-!.
elim_variant([C|Cs], [C|Ds]):- elim_variant(C, Cs, Es),
	elim_variant(Es, Ds).
%
elim_variant(_, [], []):-!.
elim_variant(C, [C0|Cs], Ds):- variant(C, C0), !,
	elim_variant(C, Cs, Ds).
elim_variant(C, [C0|Cs], [C0|Ds]):- elim_variant(C, Cs, Ds).

% ?- fol_cnf(a, R), factoring(R, F).
% ?- fol_cnf((a,b), R), factoring(R, F).
% ?- fol_cnf((a(A);a(B);a(C)), R), factoring(R, F).
%
factoring(CNF, F):-
	maplist(cls_factoring, CNF, Factors),
	maplist(unify_list_of_factors, Factors, List),
	append(List, F).

% ?- unify_list_of_factors([[[a(A), a(1)], [b(B), b(2)]], [[a(A), a(2)], [b(B), b(1)]]], R).

unify_list_of_factors([], []).
unify_list_of_factors([B|Bs], F):-
	copy_term(B, B0),
	(	unify_factors(B0, F0) ->
		F = [F0|R]
	;	F = R
	),
	unify_list_of_factors(Bs, R).

% ?- unify_factors([[a(A), a(1)], [b(B), b(2)]], X).
unify_factors([], []).
unify_factors([[A]|R], [A|R0]):-!, unify_factors(R, R0).
unify_factors([[A|As]|R], [A|R0]):-!, unify_factor(As, A),
	unify_factors(R, R0).
%
unify_factor([], _).
unify_factor([B|Bs], A):- unify_with_occurs_check(A, B),
		unify_factor(Bs, A).

% ?- list_partition([a, b], X).
% ?- time((N = 10, numlist(1, N, Ns), list_partition(Ns, X), length(X, L))).

list_partition([],[[]]).
list_partition([X|Y], Z):- list_partition(Y, Ps),
	add_one(Ps, X, Z).
%
add_one([], _, []).
add_one([P|Ps], X, [[[X]|P]|U]):-
	add_one_to_each_block(P, X, V),
	add_one(Ps, X, W),
	append(V, W, U).
%
add_one_to_each_block([], _, []).
add_one_to_each_block([B|Bs], X, [[[X|B]|Bs]|U]):-
	add_one_to_each_block(Bs, X, V),
	map_cons(V, B, U).

% ?- map_cons([[a],[b]], c, X).
map_cons(X, Y, Z):- map_cons(X, Y, Z, []).
%
map_cons([], _, U, U).
map_cons([P|Ps], X, [[X|P]|U], V):- map_cons(Ps, X, U, V).

% ?- cls_factoring([], X).
% ?- cls_factoring([a,b], X).
% ?- cls_factoring([a(1),a(2), b, c], X).
% ?- cls_factoring([a(1),a(2), -a(3), -a(4), c], X), maplist(writeln, X).

cls_factoring(X, Y):-  cls_sort(X, Z),
	ord_ltr_group(Z, U),
	maplist(list_partition, U, V),
	product_with_append(V, Y).

% ?- ord_ltr_group([], X).
% ?- ord_ltr_group([a], X).
% ?- ord_ltr_group([a,b,c], X).
% ?- ord_ltr_group([a(1),a(2), -a(1), -a(2), c], X).
ord_ltr_group([], [[]]).
ord_ltr_group([X|Y], Z):- ord_ltr_group(Y, U),
	ord_ltr_group(U, X, Z).
%
ord_ltr_group([[]|R], X, [[X]|R]):-!.
ord_ltr_group([[A|U]|R], X, [[X,A|U]|R]):-ltr_same_group(X, A), !.
ord_ltr_group(R, X, [[X]|R]).

% ?- ltr_same_group(a, -a).
ltr_same_group(X, Y):- ltr_check(X, S, _, B),
					   ltr_check(Y, S, _, B).

% ?- product_with_append([], X).
% ?- product_with_append([[[a]]], X).
% ?- product_with_append([[[a]], [[b]]], X).
% ?- product_with_append([[[a,b], [c,d]], [[b],[d]]], X).

product_with_append([], [[]]).
product_with_append([X|Y], Z):- product_with_append(Y, U),
	product_with_append(X, U, Z, []).

product_with_append([], _, Z, Z).
product_with_append([X|Xs], Y, Z, U):-
	product_with_append_base(Y, X, Z, V),
	product_with_append(Xs, Y, V, U).
%
product_with_append_base([], _, Z, Z).
product_with_append_base([Y|Ys], X, [V|Z], U):-
	append(X, Y, V),
	product_with_append_base(Ys, X, Z, U).
