:- module(bdfm, []).
% Basic Grobner Base Method without using vectors.
:-use_module(pac(basic)).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

%
gb_BDFM([P|F], G, Opt):-
	subset([order(Ord), trace(Trace), assoc(Assoc),
			fraction_free(FF), prime(Prime)],
			Opt),
	Debug = debug(Trace, Assoc),
	init_base(G0),
	add_to_base(P, G0, G1),
	poly_pairs(F, [], D, G1, G2, Ord),
	gb_spoly(D, G2, G, Debug, Ord, FF, Prime).

init_base([]).
add_to_base(P, R, [P|R]).

%
poly_pairs([], D, D, G, G, _).
poly_pairs([P|F], D0, D, G0, G, Ord):-
	update_spoly_agenda(D0, P, D1, G0, Ord),
	poly_pairs(F, D1, D, [P|G0], G, Ord).

% For Tracing method
% ?- galois_zip(3, [[1*[x^1], 5*[]]], X, Y).

galois_zip(Prime,  L, G, Zip):-
	maplist(pred(Prime, [P, Q, P-Q]:-
				poly:galois_poly(Prime, P, Q)),
		  L, G, Zip).

% ?- poly_trace([[1*[x^1], -1*[]]], [1*[x^1], 1*[a^1]], R, lexical).
poly_trace([P|Ps], X, Y, Ord):- reduce_by_poly(X, P, X0, Ord),
	poly_trace(Ps, X0, Y, Ord).

% ?- spoly_trace([1*[x^1], 1*[]], [1*[y^1], 1*[]], [], Z, total).
%@ Z = [1*[y^1], -1*[x^1]] .

spoly_trace(X, Y, Ps, Z, Ord):- poly:s_poly(X, Y, Z0, Ord),
	poly_trace(Ps, Z0, Z, Ord).

%
generic_spoly(true, I, J, Spoly, _G, Ord):- !,
	once(poly:s_poly_z(I, J, Spoly, Ord)).
generic_spoly(_, I, J, Spoly, _G, Ord):-
	once(poly:s_poly(I, J, Spoly, Ord)).

%
gb_spoly([], G, G, _, _, _, _).
gb_spoly([I-J|D], G0, G, Debug, Ord, FF, BL):-
	once(generic_spoly(FF, I, J, Spoly, G0, Ord)),
	once(gb:reduce_head_by_polyset(Spoly, G0, S0, Ord, FF, _Trace, [])),
	once(gb:generic_normal_poly(FF, BL, S0, R)),
	(  R ==[]
	->	gb_spoly(D, G0, G, Debug, Ord, FF, BL)
	;	update_spoly_agenda(D, R, D0, G0, Ord),
		gb_spoly(D0, [R|G0], G, Debug, Ord, FF, BL)
	).


debug(debug(Agenda_Trace, Assoc), D,  G0, R):-
	(	Agenda_Trace == true
	->	poly:postprocess(R, Residue, Assoc),
		length([_|D], L),
		length(G0,  L0),
		(	Residue == []
		->	W = []
		;       Residue = [_*W|_]
		),
		format("(#agenda, #gb) = (~d, ~d)  ~w\n", [L, L0, W])
	;	true
	).

%
update_spoly_agenda(D, M, D0, G, Ord):-
	update_spoly_agenda_B(D, U, [], M, Ord),
	update_spoly_agenda_DFM(M, V, G, Ord),
	sort_spoly(V, V0),
	merge_spoly_agenda(U, V0, D0).

%
update_spoly_agenda_B([], D, D, _, _).
update_spoly_agenda_B([I-J|R], P, Q, M, Ord):-  b_cond(M, I, J), !,
	update_spoly_agenda_B(R, P, Q, M, Ord).
update_spoly_agenda_B([A|R], [A|P], Q, M, Ord):-
	update_spoly_agenda_B(R, P, Q, M, Ord).

%
update_spoly_agenda_DFM(M, V, G, Ord):-
	update_spoly_agenda_DFM(G, G, M, V, [], Ord).

%
update_spoly_agenda_DFM([I|Gi], G, J, P, Q, Ord):-
	( f_cond(I, J, Gi); m_cond(I, J, G); d_cond(I, J) ),
	!,
	update_spoly_agenda_DFM(Gi,  G, J, P, Q, Ord).
update_spoly_agenda_DFM([I|Gi], G, J, [I-J|P], Q, Ord):- !,
	update_spoly_agenda_DFM(Gi, G, J, P, Q, Ord).
update_spoly_agenda_DFM(_, _,  _, P, P, _).

%
b_cond([_*Tk|_], [_*Ti|_], [_*Tj|_]):-
	poly:mono_lcm(Ti, Tj, Tij),
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Ti, Tk, Tik),
	Tik \== Tij,
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.
%
d_cond([_*Ti|_], [_*Tj|_]) :- poly: merge_mono_mono(Ti, Tj, T),
       poly:mono_lcm(Ti, Tj, T).

%
f_cond([_*Ti|_], [_*Tj|_], G) :-
	poly:mono_lcm(Ti, Tj, T),
	poly:rev_member([_*Tk|_], G),
%	member([_*Tk|_], G),
	poly:mono_lcm(Tk, Tj, T).

%
m_cond(Pi, Pj, G) :-   Pi = [_*Ti|_],
	Pj = [_*Tj|_],
	poly:mono_lcm(Ti, Tj, Tij),
	poly:rev_member(Pk, G),
%	member(Pk, G),
	Pk \== Pi,
	Pk = [_*Tk|_],
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.


% ?- merge_spoly_agenda([[1*[b^1]]-[1*[a^1]]], [[1*[a^1]]-[1*[b^1]]], R).
%@ R = [[1*[b^1]]-[1*[a^1]], [1*[a^1]]-[1*[b^1]]] .
merge_spoly_agenda([], X, X).
merge_spoly_agenda(X, [], X).
merge_spoly_agenda([P|R], [Q|S], [P|T]):-
	compare_spoly(C, P, Q),
	C = (<),
	!,
	merge_spoly_agenda(R, [Q|S], T).
merge_spoly_agenda(R, [Q|S], [Q|T]):- merge_spoly_agenda(R, S, T).


% ?- compare_spoly(C, [1*[a^2]]-[1*[b^2]], [1*[b^1]]-[1*[a^2]]).
%@ C = (>).
compare_spoly(C, [_*M|_]-[_*N|_], [_*M0|_]-[_*N0|_]):-
	poly:mono_lcm(M,  N,  A),
	poly:mono_lcm(M0, N0, A0),
	poly:compare_total_order(C0, A, A0),
	(C0 == (=) -> C = (<); C = C0 ).

% ?- sort_spoly([[1*[a^2]]-[1*[b^2]], [1*[b^1]]-[1*[a^2]]], R).
%@ R = [[1*[b^1]]-[1*[a^2]], [1*[a^2]]-[1*[b^2]]].
sort_spoly(X, Y):- predsort(compare_spoly, X, Y).
