:- module(vector, [new_vector/1, new_vector/2, v_ref/3, v_size/2, v_max/2,
				  v_member/2,  new_index/3, new_index/4]
		).

% ?- vector:(numlist(1, 10000, L), list_to_vector(L, V), list_to_vector(M, V)), M=L.

new_vector($$(Size, 0, Array), Size) :-
	functor(Array, ##, Size).

new_vector(Vector) :- new_vector(Vector, 8).

%
v_ref(I, $$(_, _, Array), R):-  arg(I, Array, R).

v_size($$(_, I, _), I).

v_max($$(M, _, _), M).

%
new_index(I, $$(Max, J, Array), $$(Max1, I, Array1)):-
	succ(J, I),
	( J < Max -> Max1 = Max, Array1 = Array
	;	plus(Max, Max, Max1),
		functor(Array1, ##, Max1),
		copy_args(Max, Array, Array1)).
%
new_index(I, E, A, B):- new_index(I, A, B),
	v_ref(I, B, E).

v_member(X, $$(_, I, Array)):-  between(1, I, J),
	arg(J, Array, X).

%
copy_args(0, _, _):-!.
copy_args(I, X, Y):- arg(I, X, A), arg(I, Y, A),
	J is I -1,
	copy_args(J, X, Y).

%
list_to_vector(L, $$(N, N, V)):- compound(L), !,
	length(L, N),
	V=..[(##)|L].
list_to_vector(L, $$(_, J, V)):-
	list_to_vector(1, J, L, [], V).

%
list_to_vector(I, J, L, L, _):- I > J, !.
list_to_vector(I, J, [A|L], M, V):-
	arg(I, V, A),
	succ(I, K),
	list_to_vector(K, J, L, M, V).
