:- module(hmodel,[]).

:- use_module(util(math)).
:- use_module(pac(op)).
:- use_module(pac(basic)).

truth(X, true):- once(X), !.
truth(_, false).


%%% groupoid %%%%
% grupoid(_, B, X, Y, [], [],true):- eval([algebra(B)], X, Y).
% grupoid(_, B, X, Y, [], [],true):- algebra(B, X, Y).


grupoid(new, D, grupoid(D, M)):- product(D, D, D2), map(D2, D, M).
grupoid(fun, grupoid(_, F), F).
grupoid(dom, grupoid(D, _), D).

%%% semigroup %%%%%%
semigroup(new, D, B):- !,  grupoid(new, D, B),
	models(grupoid::B, $associativity).
semigroup(F, X, Y):- grupoid(F, X, Y).

%%%%%% the set of DAGS and Posets %%%%
% ?- hmodel:dag(all, [a,b,c], Dags).
dag(all, L, Ds):- dags(L, Ds).
dag(new, L, dag(L,D)):- dag(all, L, Ds), member(D, Ds).
dag(dom, dag(L,_), L).
dag(rel, dag(_,R), R).
dag(succ_pair, dag(L, R), dag(L, Ps)):- succlist_pairs(R, Ps).

%% posets
% ?- poset(all, [a,b,c], X).
% ?- hmodel:new(poset::[a,b,c,d,e], X).
poset(all, L, Ds):- !, posets_in_fan(L, Ds).
poset(new, L, dag(L,D)):- !,  poset(all, L, Ds), member(D, Ds).
poset(F, X, Y):- dag(F, X, Y).

%%%%  orderd grupoid class

% ?- hmodel:og(new, [a,b], X), hmodel:rel(og::X, [_|_]).
% ?- hmodel:(og(new, [a,b], X), rel(::(og,X), [_|_])).
% ?- hmodel:(og(new, [a,b], X), og(rel, X, [_|_])).
% ?- hmodel:(new(og::[a,b], X), rel(og::X, [_|_])).

og(new, D, og(D,X,Y)):- poset(new, D, X0),
	dag(succ_pair, X0, X1),
	dag(rel, X1, X),
	grupoid(new, D, Y0),
	grupoid(fun, Y0, Y).
og(dom, og(X,_,_), X).
og(rel, og(_,X,_), X).
og(fun, og(_,_,X), X).

%%%%%%%%
% ?- hmodel:entails(grupoid::[a,b,c], true, $(associativity)).
% ?- hmodel:entails(semigroup::[a,b,c], true, $(associativity)).

entails((P::X), B, C):- new(P::X, M),
	models(P::M, B),
	\+ models(P::M, C),
        !,
	fail.
entails(_, _, _).

%;; (setq module-query  "qcompile(util(hmodel)), module(hmodel).")
% ?- qcompile(util(hmodel)), module(hmodel).
% ?- hmodel:find_model(grupoid::[a,b,c], \+ $(idempotence), X).

find_model(Class::P, C, X):- call(Class, new, P, X),
	models(Class::X, C).

%%%% model calculus
% :- index(models(0,1)).
models(_, true):-!.
models(M, (C1,C2)):- !, models(M, C1), models(M, C2).
models(M, (C1;C2)):- !,( models(M, C1); models(M, C2)).
models(M, (C1->C2)):- !, (\+ models(M, C1); models(M, C2)).
models(M, (\+ C)):- !,  \+ models(M, C).
models(M, every(Y,C)):- !, forall(in(Y,M), models(M,C)).
models(M, some(Y,C)):- !,  in(Y, M), models(M, C).
% models(M, Y=Z):- !, eval(M, Y, Y0), eval(M, Z, Y0).
models(M, Y=Z):- !, call(M, Y, Y0), call(M, Z, Y0).
models(M, $(Property)):- !, call(Property, M).
models(_, prolog(G)):- !, call(G).
models(M, forall(P,G)):- !, forall(P, models(M, G)).
models(M, exists(P,G)):- !, call(P), models(M, G).
% models(C::M, X):- eh:mapterm(eval(M), X, X0), holds(C::M, X0).
models(C::M, X):- eh:mapterm(M, X, X0), holds(C::M, X0).

%
holds(_,  A=<A).
holds(M,  A=<B):- rel(M, Fs), memberchk((A,B), Fs).

%%%%
% ?-  new(semigroup::[a,b,c], S), models(semigroup::S, $(has_subalgebra)).
% ?-  new(dag::[a,b], X).
% ?-  all(dag::[a,b], X).
% ?-  new(dag::[a,b], X), send(dag::succ_pair, X, Y).
% ?-  send(og::new, [a,b], X).

send(Class::F, X, Y):- call(Class, F, X, Y).

all(Class::X, Y):-  call(Class, all, X, Y).

new(Class::X, Y):-  call(Class, new, X, Y).

fun(Class::X, Y):-  call(Class, fun, X, Y).

dom(Class::X, Y):-  call(Class, dom, X, Y).

rel(Class::X, Y):-  call(Class, rel, X, Y).

%%%%%%%%

algebra(M, A*B, C, [A,B], [A0,B0], memberchk(((A0,B0),C), M)).
algebra(_, A, A, [], [], true).

%%%% an element of the domain.
in(X, M):- dom(M, D), member(X, D).

%%%%% a binary operation M is associative over a domain D.
% models(G, every(X,every(Y,(every(Z, ((X* Y)*Z) = (X*(Y*Z))))))).

% ?- module(pac).
% ?- hmodel:find_model(grupoid::[a,b], \+ $(associativity), X).
% ?- hmodel:find_model(grupoid::[a,b], \+ $(associativity), X).

associativity(G):- dom(G, D), fun(G, M), associativity(D, M).

associativity(D, M):-
	forall(assign([X,Y,Z],D),
	       (member(((Y,Z),YZ),M), member(((X,YZ),XYZ),M),
	    	member(((X,Y),XY),M), member(((XY,Z),XYZ),M)
	       )
	      ).

%%%%%%
assign([],_).
assign([X|R],D):- member(X, D), assign(R,D).

%%%%% has a  subalgebra
has_subalgebra(M):-  fun(M, Tbl),
	dom(M, D),
	powerset(D, P),
	member(S, P),
	S\==D,
	S\==[],
	subalgebra(S, Tbl).

%%%%% generator of ordered grupoid %%%%

succlist_pairs([],[]).
succlist_pairs([(_,[])|R], Ps):- succlist_pairs(R, Ps).
succlist_pairs([(A,L)|R], Ps):-
	succlist_pairs(A, L, Qs),
	succlist_pairs(R, P0s),
	append(Qs, P0s, Ps).

succlist_pairs(_,[],[]).
succlist_pairs(A,[X|Xs],[(X,A)|Ys]):- succlist_pairs(A,Xs,Ys).

% ?- hmodel:permutations([1,2,3], X).
%@ X = [[3, 2, 1], [2, 3, 1], [2, 1, 3], [3, 1, 2], [1, 3, 2], [1, 2, 3]] .

permutations(X, Y):- permutations(X, [[]], Y).

permutations([],X, X).
permutations([X|Xs],Z0, Z):- maplist(perm_insert(X), Z0, Z1),
	append(Z1, Z2),
	permutations(Xs, Z2, Z).

perm_insert(X, [], [[X]]).
perm_insert(X, [Y|Ys], [[X, Y|Ys]|Z]):-
	perm_insert(X, Ys, Z0),
	maplist(cons(Y), Z0, Z).

% non deterministic permutation generator
perm_nd(X, Y) :- perm_nd(X, [], Y).

perm_nd([], X, X).
perm_nd([X|Xs], Y0, Y):- insert_nd(X, Y0, Y1), perm_nd(Xs, Y1, Y).

insert_nd(X, Y, [X|Y]).
insert_nd(X,[Y|Ys], [Y|Zs]):- insert_nd(X, Ys, Zs).

diagonal([]).
diagonal([(X,X)|D]):-diagonal(D),!.

make_id([],[]).
make_id([X|Y],[(X,X)|D]):- make_id(Y,D).

field([],[]).
field([(X,Y)|R],[X,Y|F]):-field(R,F).

compose(X,Y,Z):-compose(X,Y,Z,[]).

compose([],_,X,X).
compose([(X,Y)|Z],U,V,W):- compose(X,Y,U,V,R), compose(Z,U,R,W).

compose(_,_,[],X,X).
compose(X,Y,[(Y,Z)|U],[(X,Z)|V],W):-!, compose(X,Y, U,V,W).
compose(X,Y,[_|U],V,W):- compose(X,Y, U,V,W).

%% transitive closure X of relation R

trcl(R,X):- trcl(R,R,X).

trcl(R,X,Y):- compose(R,X,RX), union(RX,X,Z), trcl(R,X,Z,Y).

trcl(_,X,Y,X):- subset(Y,X),!.
trcl(R,_,Y,Z):- trcl(R,Y,Z).

trcl_sum(X,Y,Z):-compose(X,Y,Z1),compose(Y,X,Z2),
	union(Z1,Z2,Z3),
	trcl(Z3,Z4),
	union([X,Y,Z4],Z).

%%%% test connectivity of undirectied graph (X,E).

connected(X,E):-makediscrete(X,Y),
	union_find(E,Y,Class),Class=[_].

makediscrete([],[]).
makediscrete([X|R],[[X]|R1]):-makediscrete(R,R1).

union_find([],X,X).
union_find([(X,Y)|R],C,D):-union_find(X,Y,C,C1),union_find(R,C1,D).

union_find(X,Y,Z,[Z5|Z4]):-find(X,Z,Z1,Z2),
	find(Y,Z2,Z3,Z4),
	union(Z1,Z3,Z5).

find(X,[],[X],[]):-!.
find(X,[Y|Z],Y,Z):- in(X, Y),!.
find(X,[Y|Z],U,[Y|V]):- find(X,Z,U,V).

%%%%  G is a grupoid over X.
grupoid(X,G):-
	product(X,X,XX),
	eh:maps(XX,X,XXX),
	in(G ,  XXX).

grupoidx(X,G):-  %%  backtracking version
	product(X,X,XX),
	map(XX,X,G).

map([],_,[]).
map([X|R],Y,[(X,A)|S]):- member(A, Y), map(R,Y,S).

%%%% a set X is closed
%%%% with respect to an binary operation Alg.

subalgebra(X,Alg):- in(Y, X), in(Z, X), in(((Y,Z),U), Alg),
	\+ in(U,  X), !, fail.
subalgebra(_,_).

%%  Y is the closure of X
%%  with respect to a binary operation G
binary_closed(G,X,Y):- bincl(G,X,Z), subset(Y,Z).

bincl(G,X,Y):-binresult(G,X,X,Z), bincl1(G,X,Z,Y).

bincl1(_,X,Y,X):- subset(Y,X),!.
bincl1(G,X,Y,Z):- union(X,Y,U), bincl(G,U,Z).

binresult(_,[],_,[]).
binresult(G,[X|Y],Z,U):-
	binresult1(G,X,Z,V),
	binresult(G,Y,Z,W),
	union(V,W,U).

binresult1(_,_,[],[]).
binresult1(G,X,[Y|Z],[U|V]):- in( ((X,Y),U),  G), !,  binresult1(G,X,Z,V).

%%%%%% predicates over relations
partial_order(R):-
	anti_symmetric(R),
	transitive(R),
	refexive(R).

reflexsive(R):-field(R,F),
	reflexive(F,R).

reflexive([],_).
reflexive([X|Y],R):- in((X,X), R), reflexive(Y,R).

anti_symmetric(R):- member((X,Y),R),
	X\==Y,
	member((Y,X),R),
	!,
	fail.
anti_symmetric(_).

symmetric(R):- in((X,Y), R),  \+ in((Y,X), R), !, fail.
symmetric(_).

transitive(R):- compose(R,R,RR), subset(RR,R).

%%%%%% top level call
% ?- dags([a,b,c], Dags).
% ?- posets([a,b,c,d,e,f,g], X).
% ?- posets_in_fan([a,b,c,d,e,f,g], X).

dags(Ns, Dags):- dags((peak, append), Ns, Dags).

posets(Ns, Ps):- sort(Ns, Ms), dags((anti_chain, append), Ms, Ps).

posets_in_fan(Ns, Dags):- dags((fan, union), Ns, Dags).
% 'union' is better than 'append'  at least for Ns=[a,b,c,d,e,f,g].

%%%%%%
peak(X, Dag, S, [(X,S)|Dag]).

anti_chain(X, Poset, Chain, [(X, Chain)|Poset]):- anti_chain(Chain, Poset).

anti_chain([],_).
anti_chain([N|Ns], Poset):- anti_chain(Ns, Poset),
	maplist(incomparable(N,Poset), Ns).

incomparable(N, Poset, N0):- N @> N0, !, \+reach(N0,Poset,N).
incomparable(N, Poset, N0):- \+reach(N, Poset, N0).

reach(N,_,N):-!.
reach(N0, Poset,N):- memberchk((N,L), Poset),
	member(X, L),
	reach(N0, Poset, X).

fan(X, Poset, S, [(X,S0)|Poset]):-
 	maplist(fan(Poset), S, Ls),
 	union([S|Ls], S0).

fan(P, X, L):- memberchk((X,L), P).

%%%%%% dags: universal predicate  %%%%%%
dags(F, Nodes, Dags):- once(dags(F, Nodes, [[]], [[]], _, Dags)).

dags(_, [], X, Y, X, Y).
dags(F, [N|Ns], Succs0, Dags0, Succs, Dags):-
	new_node(F, N, Succs0, Dags0, Succs1, Dags1),
        dags(F, Ns, Succs1, Dags1, Succs, Dags).

new_node(F, N, Succs, Dags, Succs1, Dags1):-
	new_node(F, N, Succs, Dags, Dags1),
	extend_succs(N, Succs, Succs1).

new_node(F, N, Succs, Dags0, Dags):-
	maplist(new_peak(F, N, Succs), Dags0, Dags1),
	F = (_, A),  % append or union
	call(A, Dags1, Dags).

new_peak(_, _, [], _, []).
new_peak(F, X, [L|Ls], Dag, [Dag0|Dags]):-
	F = (M, _),  % method to add a node
	call(M, X, Dag, L, Dag0),
	!,
	new_peak(F, X, Ls, Dag, Dags).
new_peak(F, X, [_|Ls], Dag, Dags):-
	new_peak(F, X, Ls, Dag, Dags).

extend_succs(_, [], []).
extend_succs(X, [S|R0], [S, [X|S]|R]):- extend_succs(X, R0, R).

%%%%%
% ?- find_model(grupoid::[a,b], $surj, X).

isolated(X, G):- rel(G, E), forall(in((X,Y), E); in((Y,X), E), X==Y).

maximal(X, M):-  in(X, M), rel(M, E), forall(in((X,Y), E), X=Y).

minimal(X, M):- rel(M, E), forall(in((Y,X), E), X=Y).

irreducible(L, M) :- fun(M,  G), forall(in(X, L), binary_closed(G,[X],L)).

idempotence(M):- fun(M, F),  memberchk(((A,A),A), F).

left_unit(U, G):- fun(G, M), dom(G, D),
	forall(in(X,D), in(((U,X),X), M)).

right_unit(U, G):- fun(G, M), dom(G, D),
	forall(in(X, D), in(((X,U),X), M)).

surj(M):- dom(M, D), fun(M, F), maplist(snd, F, M0),
	sort(D, D0), sort(M0, D0).

%%%% for OG
aug(OG):- models(OG,every(X,every(Y, (Y=< (X*Y))))).

slide1(M):- models(M, every(A,every(B, every(C,
		(((A*B)*C) =< (A*(B*C))))))).
mix(M):- models(M, every(A,every(B, every(C, every(D, ((A=<B, C=<D)
		-> ((B*C) =< (A*D)))))))).
