:- module(ifmap_demo, []).   % [2013/07/28]
:- use_module(pac('expand-pac')).
:- use_pac(ifmap).
:- use_pac(snippets).

%:- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).


% ?- ifmap:show_cla(cla([a,b],[1,2], [1-[a], 2-[b]])).
%@ true.

test_show_im :-
	C1 = cla([1,3],[a], [a-[1,3]]),
	C2 = cla([2,4], [b,c], [b-[2,4],c-[4]]),
	FU = [a->b],
	FD = [2->1, 4->3],
	show_IM(C1,  C2,  FU,  FD).

show_IM(cla(A,B,C), cla(D,E,F), G, H):-
	maplist(pred([X, X@[color = red]]), G, G0),
	maplist(pred([X, X@[color = green]]), H, H0),
	list_to_semicolon(G0, G1),
	list_to_semicolon(H0, H1),
	show_digraph(
	     subgraph(cluster_1, A;B;C);
	     subgraph(cluster_2, D;E;F);
		 G1;
		 H1
		    ).

%
list_to_semicolon([X], X).
list_to_semicolon([X, Y|Z], X; Y0):-
	list_to_semicolon([Y|Z], Y0).

% ?- ifmap:build_adjoint(cla([1, 2, 3, 4, 5], [], [sweet-[3, 4], apple-[4], acid-[1, 2], orange-[2]]), cla([taro], [apple, orange, sweet, acid], [sweet-[taro]]), [apple-apple, orange-orange, sweet-sweet, acid-acid], [], _G2931, _G2932).

test_ifmap(R):- classification_base(2, C2),
	classification_base(3, C3),
	C2 = cla(_, T, _),
	maplist(pred([X, X-X]), T, TT),
	extend_info_map(C3, C2, im(TT, []), R).

test(R):- sample_theory(2, Theory),
	theory_to_cla(Theory, Cla),
	Cla= cla(_, Ts, _),
	maplist(pred([X, X-X]), Ts, U),
	classification_base(2, TargetCla),
	extend_info_map(Cla, TargetCla, im(U,[]), R).

% ?- ifmap:(sample_theory(1, Th), theory_to_cla_show(Th)).
%@ Th = theory([ag, pa, ind, fs, eubd], [=>([], [ag, fs]), =>([ag, fs], []), =>([pa], [ag]), =>([ind], [ag]), =>([pa, ind], []), =>([eubd], [fs])]).
% ?- trace, ifmap:(sample_theory(2, Th), theory_to_cla_show(Th)).

theory_to_cla_show(Theory) :- once(theory_to_cla(Theory, Z)), once(show_cla(Z)).

%
tok(cla(X,_,_), X).
typ(cla(_,X,_), X).
sup(cla(_,_,X), X).

% ?- ifmap: (sample_theory(1, A), theory_to_cla(A, B), build_info_map(B, B, X)).
% ?- ifmap:test_preference(1, 1, R).
% ?- ifmap:test_preference(2, 2, R).

test_preference(I, J, R):- sample_theory(I, T_I),
	classification_base(J, C_J),
	theory_to_cla(T_I, C_I),
	typ(C_I, P),
	maplist(pred([X,X-X]), P, Id),
%	inverse_type_map(Id, C_I, C_J, R).
	invertable_type_map(Id, C_I, C_J, R).

% ERROR
% ?- ifmap:(theory_to_cla(theory([a,b],[[a,b]=>[]]), R), cla_boolean_atoms(R, R0), show_cla(R0)).
% ?- ifmap:(theory_to_cla(theory([a,b],[[a,b]=>[]]), R), cla_boolean_atoms(R, R0)).
% ?-  show_cla(cla([[1], [2], [3]], [[b, ~(a)], [~(b), a], [~(b), ~(a)]], [[~(b), a]-[2], [b, ~(a)]-[1], [~(b), ~(a)]-[3]])) .

%
print_cla(cla(X, Y, Z)):- nl, nl,
	writeln(X),
	writeln(Y),
	writeln('*** support: '),
	maplist(writeln, Z).

%  「ゴキブリの入ったコーヒーとオレンジジュースのどちらが飲みたいですか？」
%
% http://www.rease.e.u-tokyo.ac.jp/read/jp/archive/essay/index.html
% 「ゴキブリの入ったコーヒーとオレンジジュースのどちらが飲みたいですか？」
% 表象システム(３)　（坂原樹麗・佐藤崇）

% ?- ifmap:which_do_you_prefer([coffee, coffee_cockroach, orange_juice, orange_juice_cockroach]).

%@ coffee:	 please.
%@ coffee_cockroach:	 no thank you.
%@ orange_juice:	 please.
%@ orange_juice_cockroach:	 no thank you.

which_do_you_prefer(Cla_names):-
	sample_desire(0,	Desire),
	sample_theory(0,	Theory),
	maplist(pred([Desire, Theory],
		     ([Cla_name] :-
			sample_observation(Cla_name,	Observation),
			once(representation_system(Observation, Desire, Theory, Im_O, Im_D)),
%%%  check logic !!!
			(	is_function(Im_O),is_function(Im_D)
			->	true
			;	writeln("Does not form a representation system"),
				fail
			),
			(	is_preferred(Im_O, Im_D)
			->	Message = 'please'
			;	Message = 'no thank you'
			),
		        atomics_to_string([Cla_name, ':\t ', Message,  '.'], M),
			writeln(M)
		     )),
		Cla_names).

% which_do_you_prefer(Cla_names):-
% 	sample_theory(theory0,	Desire),
% 	sample_theory(theory1,	Theory),
% 	maplist(pred([Desire, Theory],
% 		     ([Cla_name] :-
% 			sample_observation(Cla_name,	Observation),
% 			once(representation_system(Observation, Desire, Theory, Im_O, Im_D)),
% %%%  check logic !!!
% 			(	is_function(Im_O),is_function(Im_D)
% 			->	true
% 			;	writeln("Does not form a representation system"),
% 				fail
% 			),
% 			(	is_preferred(Im_O, Im_D)
% 			->	Message = 'please'
% 			;	Message = 'no thank you'
% 			),
% 		        atomics_to_string([Cla_name, ':\t ', Message,  '.'], M),
% 			writeln(M)
% 		     )),
% 		Cla_names).

%%%%%test
% ?- ifmap: representation_system(cla([x],[],[]), theory([], []), theory([],[]), Im_O, Im_D).
%@ Im_O = Im_D, Im_D = [p([], [])-p([], [])].
% ?- ifmap: representation_system(cla([],[],[]), theory([], []), theory([],[]), Im_O, Im_D).
%@ **** false. ****
% ?- ifmap: representation_system(cla([],[a],[]), theory([], []), theory([],[]), Im_O, Im_D).
%@ false.
% ?- ifmap: representation_system(cla([],[a],[]), theory([a], []), theory([],[]), Im_O, Im_D).
%@ false.
% ?- ifmap: representation_system(cla([1],[a],[a-[1]]), theory([a], [[]=>[a]]), theory([a],[[]=>[a]]), Im_O, Im_D).
%@ Im_O = Im_D, Im_D = [p([a], [])-p([a], [])] .
% ?- ifmap: representation_system(cla([1],[a],[a-[1]]), theory([a], [[]=>[a]]), theory([a],[[]=>[a]]), Im_O, Im_D).
% ?- ifmap: representation_system(cla([1,2],[a],[a-[1]]), theory([a], [[]=>[a]]), theory([a],[[]=>[a]]), Im_O, Im_D).
% ?- ifmap: representation_system(cla([1,2],[a],[a-[1]]), theory([a,b], [[]=>[a],[b]=>[]]), theory([a,b],[[b]=>[], []=>[a]]), Im_O, Im_D).
% ?- ifmap: representation_system(cla([1],[a],[a-[1]]), theory([a,b], [[]=>[a],[b]=>[]]), theory([a,b],[[b]=>[], []=>[a]]), Im_O, Im_D).
% ?- ifmap: representation_system(cla([1],[a,b],[a-[1]]), theory([a,b], [[]=>[a],[b]=>[]]), theory([a,b],[[b]=>[], []=>[a]]), Im_O, Im_D).
%@ Im_O = Im_D, Im_D = [p([a], [])-p([a], [])] .
% ?- ifmap: representation_system(cla([1],[a,b],[a-[1],b-[1]]), theory([a,b], [[]=>[a],[b]=>[]]), theory([a,b],[[b]=>[], []=>[a]]), Im_O, Im_D).

% ?- ifmap:theory_to_cla(theory([x,y], [=>([], [x,y])]),  R).
% ?- ifmap:(sample_observation(-1, O), sample_desire(-1, D), sample_theory(-1, Th), representation_system(O, D, Th, Im_O, Im_D)).
% ?- spy(invertable_atom_set), ifmap:(sample_observation(0, O), sample_desire(0, D), sample_theory(0, Th), representation_system(O, D, Th, Im_O, Im_D)).

% ?- trace, spy(invertable_atom_set), ifmap:(sample_observation(0, O), sample_desire(0, D), sample_theory(0, Th), representation_system(
% ?- ifmap:(sample_observation(0, O), sample_desire(0, D), sample_theory(0, Th), representation_system(O, Th, Th, Im_O, Im_D)).
% ?- trace, ifmap:(sample_observation(0, O), sample_desire(0, D), sample_theory(0, Th), representation_system(O, D, Th, Im_O, Im_D)).
% ?- trace, ifmap:(sample_observation(0, O), sample_desire(0, D), sample_theory(0-0, Th), representation_system(O, D, Th, Im_O, Im_D)).
% ?- ifmap:(sample_theory(0, Th), theory_to_cla(Th, Cla), cla(_,Ts,_)=Cla, zip(Ts, Ts, Id), invertable_type_map(Id, Cla, Cla, R), maplist(pred([X-X]), R)).
% ?- ifmap:(sample_theory(0, Th), theory_to_cla(Th, Cla), sample_desire(0, D), theory_to_cla(D, Cla1), tok(Cla1, X), zip(X, X, Id), invertable_type_map(Id, Cla1, Cla, R)).
% ?- ifmap:(sample_theory(0, Th), theory_to_cla(Th, Cla), sample_observation(0, O), theory_to_cla(O, Cla1), tok(OD, X), zip(X, X, Id), invertable_type_map(Id, Cla1, Cla, R)).

% representation_system(Cla_O, Theo_D, Theory, F_OT, F_DT, Im_O, Im_D):-
% 	Cla_O	= cla(_, T_O, _),
% 	Theo_D	= theory(T_D, _),
% 	theory_to_cla(Theo_D, Cla_D),
% 	theory_to_cla(Theory, Cla_ch),
% 	zip(T_O, T_O, ID_O),
% 	zip(T_D, T_D, ID_D),
% 	invertable_type_map(ID_O, Cla_O, Cla_ch, Im_O),
% 	invertable_type_map(ID_D, Cla_D, Cla_ch, Im_D).

representation_system(Cla_O, Theo_D, Theory, F_OT, F_DT, Im_O, Im_D):-
	theory_to_cla(Theo_D, Cla_D),
	theory_to_cla(Theory, Cla_ch),
	( var(F_OT)
	->	typ(Cla_O, T_O),
		zip(T_O, T_O, F_OT)
	; true
	),
	( var(F_DT)
	->	typ(Cla_D, T_D),
		zip(T_D, T_D, F_DT)
	; true
	),
	invertable_type_map(F_OT, Cla_O, Cla_ch, Im_O),
	invertable_type_map(F_DT, Cla_D, Cla_ch, Im_D).

representation_system(Cla_O, Theo_D, Theory, Im_O, Im_D):- !,
	representation_system(Cla_O, Theo_D, Theory, _, _, Im_O, Im_D).

sample_theory(-1,  theory( [a], [[] => [a]])).

sample_theory(0, theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '欲しい'],
			 [ ['コーヒー', 'オレンジジュース'] => [],
			   ['ゴキブリ入り'] => []]
			 )).

sample_theory(sub(0), theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '欲しい'],
			 []
			 )).
sample_theory(sub(sub(0)), theory( ['ゴキブリ入り', 'コーヒー'],
			 []
			 )).
sample_theory(0-1, theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '欲しい'],
			 [ ['コーヒー', 'オレンジジュース'] => [],
			   ['ゴキブリ入り', '欲しい'] => []]
			 )).

% sample_theory(0,  theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '香ばしい香り', '苦味', '飲み物', '欲しい'],
% 			 [[] => ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '香ばしい香り', '苦味', '飲み物', '欲しい'],
% 			  ['ゴキブリ入り', '欲しい'] => [],
%     			  ['コーヒー', 'オレンジジュース'] => []]
% 			 )).
sample_theory(0-0,  theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース','飲み物', '欲しい'],
			 [[] => ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '飲み物', '欲しい']])).
sample_theory(1,  theory([ag,pa,ind,fs,eubd],
	   [[]		=>	[ag,fs],
	    [ag,fs]	=>	[],
	    [pa]	=>	[ag],
	    [ind]	=>	[ag],
	    [pa,ind]	=>	[],
	    [eubd]	=>	[fs]])).

sample_theory(2,  theory([apple, orange, sweet, acid],
	   [[apple]	=>	[sweet],
	    [orange]	=>	[acid],
	    [sweet, acid] => 	[]])).

%
sample_desire(-1, theory([a],	[ [] => [a]])).

% ?- ifmap:theory_to_cla_show(theory([c,o,g,h], [[h]=>[c,o], [h,g]=>[], [c,o]=>[]])).
%@ true.

% ?- ifmap:invertable_type_map(=, cla([c1, o1, g1, x], [o, c, g], [o-[o1], c-[c1],g-[g1]]), theory([c,o,g,h], [[h]=>[c,o],[c,g]=>[],  [h,g]=>[], [c,o]=>[], [g,o]=>[]])).
% ?- ifmap:invertable_type_map(=, cla([c1, o1, x], [o, c], [o-[o1], c-[c1]]), theory([c,o,g,h], [[h]=>[c,o],[c,g]=>[],  [h,g]=>[], [c,o]=>[], [g,o]=>[]])).
% ?- ifmap:invertable_type_map(=, cla([c1, o1, x], [o, c], [o-[o1], c-[c1]]), theory([c,o,h], [[h]=>[c,o], [c,o]=>[]])).

sample_desire(0,
	theory(['コーヒー', 'オレンジジュース', 'ゴキブリ入り', '欲しい'],
		[  ['欲しい'] => ['コーヒー', 'オレンジジュース'],
		   ['欲しい', 'ゴキブリ入り']  => []
		] )).

sample_desire(0-1,
	theory(['コーヒー', 'オレンジジュース', 'ゴキブリ入り', '欲しい'],
		[  ['コーヒー']		=> ['欲しい', 'ゴキブリ入り'],
		   ['オレンジジュース']	=> ['欲しい', 'ゴキブリ入り'],
		   ['欲しい', 'ゴキブリ入り']  => []
		] )).

sample_desire(1,
	theory(['コーヒー', 'オレンジジュース', '欲しい'],
		[ []		=> ['コーヒー', 'オレンジジュース'],
		['コーヒー']	=> ['欲しい'],
		['オレンジジュース']	=> ['欲しい']])).
sample_desire(2,
	theory( ['ゴキブリ入り', 'コーヒー', 'オレンジジュース', '香ばしい香り', '苦味', '飲み物', '欲しい'],
		[['ゴキブリ入り', '欲しい']	=> [],
		 ['コーヒー']			=> ['ゴキブリ入り', '欲しい'],
		 ['オレンジジュース']		=> ['ゴキブリ入り', '欲しい']
		 ])).

%
sample_observation(-1,
	cla([coffee0],
	    [a],
	    [a - [coffee0]])).
sample_observation(0,
	cla([coffee0, x],
	    ['コーヒー', 'ゴキブリ入り'],
	    ['コーヒー'   - [coffee0]])).

sample_observation(1,
	cla([coffee1],
	    ['コーヒー', 'ゴキブリ入り'],
	    ['コーヒー'		- [coffee1],
	     'ゴキブリ入り'	- [coffee1]])).
sample_observation(2,
	cla([orange_juice0],
	    ['オレンジジュース'],
	    ['オレンジジュース' - [orange_juice0]])).
sample_observation(3,
	cla([orange_juice1],
	    ['オレンジジュース', 'ゴキブリ入り'],
	    ['オレンジジュース' - [orange_juice1],
	     'ゴキブリ入り' - [orange_juice1]])).


%
is_preferred(X, Y):- member(A-[X0], X),
	member(A-[Y0], Y),
	intersection(['コーヒー', 'オレンジジュース', '欲しい'], X0, S),
	S \== [],
	intersection(['コーヒー', 'オレンジジュース', '欲しい'], Y0, T),
	T \== [].

% :- module(channel_demo,[]).

% ?- ifmap: (theory_base(1, A), theory_to_cla(A, B), build_info_map(B, B, X)).
%@ A = theory([ag, pa, ind, fs, eubd], [=>([], [ag, fs]), =>([ag, fs], []), =>([pa], [ag]), =>([ind], [ag]), =>([pa, ind], []), =>([eubd], [fs])]),
%@ B = cla([1, 2, 3, 4, 5], [ag, pa, ind, fs, eubd], [ag-[3, 4, 5], eubd-[1], fs-[1, 2], ind-[3], pa-[4]]),
%@ X = im([eubd-ag, fs-ag, ind-ag, pa-ag, ag-ag], [5-1, 4-1, 3-1, 2-1, 1-1]) .

% ?- ifmap:test_preference(1, 1, R).
% ?- spy(inverse_type_map), ifmap:test_preference(2, 2, R).
% ?- ifmap:test_preference(2, 2, R).
%@ R = [[acid, ~(sweet), ~(orange), ~(apple)]-[[acid, ~(sweet), ~(orange), ~(apple)]], [~(acid), sweet, ~(orange), ~(apple)]-[[~(acid), sweet, ~(orange), ~(apple)]]] .

% test_preference(I, J, R):- theory_base(I, T_I),
% 	classification_base(J, C_J),
% 	theory_to_cla(T_I, C_I),
% 	typ(C_I, P),
% 	maplist(pred([X,X-X]), P, Id),
% 	pre_inverse_in_boolean_atoms(Id, C_I, C_J, R).

:- meta_predicate show_graph(2,?).
% ?- ifmap:show_graph(ifmap:[A-B, (A, "", B)], [a-b,b-d]).
% ?- ifmap:show_graph((=), [(a,f,b),(b, g, c)]).

show_graph(F, DG):-  maplist(F,  DG, Arrows),
	coalgebra:autviz(ifmap:digraph_viz, Arrows).

%
digraph_viz(Arrows) :-
     format("digraph g {~n",[]),
     format("rankdir=LR;~n",[]),
     coalgebra:move_in_dot(Arrows),
     format("}~n",[]).

% ?- ifmap_demo:interoperability(X, Y).
%@ X = cla(emptyset, [[1], []], [[1]-[], []-[]]),
%@ Y = cla([sos, bca, bea, age, ins], [[5], [5, 1], [5, 2, 1], [5, 2], [5, 3, 2], [5, 3|...], [5|...], [...|...]|...], [[5]-[sos], [5, 1]-[sos, age, ins], [5, 2, 1]-[sos, age, age, ins], [5, 2]-[sos, age], [5, 3|...]-[sos, bea|...], [5|...]-[sos|...], [...|...]-[...|...], ... - ...|...]) .

% interoperability(Pow_Flip_A_UK, Pow_Flip_A_US):-
%  	Typ_UK = [fco, ho],
% 	Typ_US = [dos, doj],
% 	theory_uk(Theory_Uk),
% 	theory_us(Theory_US),
% 	theory_cla_flip_pow(Theory_UK, Pow_Flip_A_UK),
% 	theory_cla_flip_pow(Theory_US, Pow_Flip_A_US).


% iopera(P, K):- theory_uk(TUK), theory_us(TUS),
% 	       theory_to_cla(TUK,ClaUK),
% 	       theory_to_cla(TUS,ClaUS),

% ?- ifmap_demo:iopera_core_tokens(X).

% ?- ifmap_demo:(theory_uk(UK), theory_us(US),  UK=theory(P,_), US=theory(Q,_), maplist(pred([A,B,A-B]), P, Q, Zip)), ds_theory_cover_with_constraint([UK,US], [im(1,2,Zip)], [], X, Y).
%@ UK = theory([ag, pa, ind, fs, eubd], [[]=>[ag, fs], [ag, fs]=>[], [pa]=>[ag], [ind]=>[ag], [pa, ind]=>[], [eubd]=>[fs]]),
%@ US = theory([sos, bca, bea, age, ins], [[]=>[sos, age], [sos, age]=>[], [bca]=>[sos], [bea]=>[sos], [bca, bea]=>[], [ins]=>[age]]),
%@ P = [ag, pa, ind, fs, eubd],
%@ Q = [sos, bca, bea, age, ins],
%@ Zip = [ag-sos, pa-bca, ind-bea, fs-age, eubd-ins],
%@ X = [[[eubd, fs]-[ag, ind, pa], [age, ins]-[bca, bea, sos]], [[fs]-[ag, eubd, ind, pa], [age]-[bca, bea, ins, sos]], [[ag, ind]-[eubd, fs, pa], [bea, sos]-[age, bca, ins]], [[ag, pa]-[eubd, fs, ind], [bca, sos]-[age, bea, ins]], [[ag]-[eubd, fs, ind, pa], [sos]-[age, bca, bea, ins]]],
%@ Y = [[1:eubd, 2:ins], [1:fs, 2:age], [1:ind, 2:bea], [1:pa, 2:bca], [1:ag, 2:sos]] .

iopera_core_tokens(X):-  theory_uk(UK), theory_us(US),
			 theory_sum_tokens([UK,US], X).


% inter-operability demo

demo_theory(uk, theory([ag,pa,ind,fs,eubd],
			   [[]=>[ag,fs],
			    [ag,fs]=>[],
			    [pa]=>[ag],
			    [ind]=>[ag],
			    [pa,ind]=>[],
			    [eubd]=>[fs]])).
demo_theory(us, theory([sos,bca,bea,age,ins],
			   [[]=>[sos,age],
			    [sos,age]=>[],
			    [bca]=>[sos],
			    [bea]=>[sos],
			    [bca,bea]=>[],
			    [ins]=>[age]])).
% ?- import_module(engine, X).

% ?- default_module(snippets, X).
%@ X = snippets ;
%@ X = engine ;
%@ X = user ;
%@ X = system .

% ?- default_module(engine, X).
%@ X = engine ;
%@ X = user ;
%@ X = system ;
%@ false.


%@ true.

% ?- default_module('emacs-handler', X).
%@ X = 'emacs-handler' ;

%@ X = engine ;
%@ X = user ;
%@ X = system ;
%@ false.


% ?- ifmap_demo:cla_by_theory_show(uk).
%@ ERROR: snippets:show_digraph/2: Undefined procedure: snippets:new_base_name/2
%@    Exception: (23) snippets:new_base_name([out(pdf), counter_name(graph_counter), stem("dot_graph"), directory("/Users/cantor/Desktop/_iMG_TMP")], _G919) ? creep

% ?- ifmap_demo:cla_by_theory_show(us).
cla_by_theory_show(Name):- demo_theory(Name, Th),
		      theory_to_cla(Th, Cla),
		      cla(flip(Cla), Flipped),
		      show_cla(Cla),
		      show_cla(Flipped).

% ?- ifmap_demo:demo_cla(X, Y).
demo_cla(uk, cla([ag,pa,ind,fs,eubd],
	       [fco, ho],
	       [fco-[fs,eubd], ho-[ag,pa,ind]])).
demo_cla(us, cla([sos,bca,bea,age,ins],
	       [dos,doj],
	       [dos-[sos,bca,bea], doj-[age,ins]])).
demo_cla(uk_flip, A):- once(uk_flip(A)).
demo_cla(uk_flip_boole, A):- once(uk_flip_boole(A)).
demo_cla(us_flip, A):- once(us_flip(A)).
demo_cla(us_flip_boole, A):- once(us_flip_boole(A)).
demo_cla(align, A) :- once(theory_to_cla(theory([alpha,beta,gamma],[]), A)).

% ?- ifmap_demo:demo_cla(align,  A), ifmap:show_cla(A).
% ?- ifmap_demo:demo_informorphism(im(uk_flip,uk_flip_boole, X, Y)).
%@ X = [r1-[r1], r2-[r2], r3-[r3], r4-[r4], r5-[r5]],
%@ Y = [ag-ag, eubd-eubd, fs-fs, ind-ind, pa-pa] .

demo_informorphism(im(align, uk_flip, [alpha-r1, beta-r2, gamma-r4])).
demo_informorphism(im(align, us_flip, [alpha-s1, beta-s4, gamma-s2])).
demo_informorphism(im(uk_flip, uk_flip_boole, TypeMap, Id)):-
	demo_cla(uk_flip, cla(Xs,Ts,_)),
	maplist(pred([X, X-[X]]), Ts, TypeMap),
	maplist(pred([X, X-X]), Xs, Id).
demo_informorphism(im(uk, uk_flip_boole, [ho-[r1,r2,r3], fco-[r4,r5]])).
demo_informorphism(im(us_flip, us_flip_boole, TypeMap, Id)):-
	demo_cla(us_flip, cla(Xs,Ts,_)),
	maplist(pred([X, X-[X]]), Ts, TypeMap),
	maplist(pred([X, X-X]), Xs, Id).
demo_informorphism(im(us, us_flip_boole, [dos-[s1,s2,s3], doj-[s4,s5]])).


% ?- ifmap_demo:demo_token_map(X).
demo_token_map(X):-
    demo_informorphism(im(uk, uk_flip_boole, M)),
    demo_cla(uk, UK),
    demo_cla(uk_flip_boole, UKFB),
    invertable_type_map(M, UK, UKFB, X).


% ?- ifmap_demo:demo_ds(ds(X, Y, [])), maplist(writeln, X), maplist(writeln, Y).
demo_ds(ds([	uk:UK, uk_flip:UKF, uk_flip_boole:UKFB,
		align:Align,
		us:US, us_flip:USF, us_flip_boole:USFB
	   ],
	  [im(uk,uk_flip_boole, UK_to_boole),
	   im(uk_flip, uk_flip_boole, UKF_to_boole),
	   im(us, us_flip_boole, US_to_boole),
	   im(us_flip, us_flip_boole, USF_to_boole),
	   im(align, uk_flip, A_to_UKF),
	   im(align, us_flip, A_to_USF)
	  ],
	   [])):- demo_cla(uk, UK),
  		  demo_cla(us, US),
		  demo_cla(uk_flip, UKF),
  		  demo_cla(us_flip, USF),
		  demo_cla(uk_flip_boole, UKFB),
  		  demo_cla(us_flip_boole, USFB),
		  demo_cla(align, Align),
	demo_informorphism(im(align, uk_flip, A_to_UKF)),
	demo_informorphism(im(align, us_flip, A_to_USF)),
	demo_informorphism(im(uk, uk_flip_boole, UK_to_boole)),
	demo_informorphism(im(us, us_flip_boole, US_to_boole)),
	demo_informorphism(im(uk_flip, uk_flip_boole, UKF_to_boole, _)),
	demo_informorphism(im(us_flip, us_flip_boole, USF_to_boole, _)).

% ?- ifmap_demo:responsibility_uk(UK), ifmap:cla(flip(UK), R).
% ?- ifmap:cla(clean_disj(pow(flip(empty))), R), write(R).
% ?- ifmap:cla(clean_disj(pow(flip(cla([1,2],[a,b],[a-[1],b-[2]])))), R), write(R).
% ?- ifmap_demo:responsibility_us(US), ifmap:cla(cleaning(pow(flip(US))), R), write(R).
% ?- ifmap_demo:responsibility_uk(UK), ifmap:cla(cleaning(pow(flip(UK))), R), write(R).

uk_flip(A):- responsibility_uk(UK),
	     cla(flip(UK), A0),
	     cla(sort(A0), A).

% ?- ifmap_demo:uk_flip_boole(A).

uk_flip_boole(A):- uk_flip(B), cla(pow(B), A0),
	     cla(cleaning_boole(A0), A).

us_flip(A):- responsibility_us(B),
	     cla(flip(B), A0),
	     cla(sort(A0), A).
% ?- ifmap_demo:us_flip_bool(A).
us_flip_boole(A):- us_flip(B), cla(pow(B), A0),
		   	     cla(cleaning_boole(A0), A).

% ?- ifmap: invertable_type_map([a-s],cla([1],[a],[a-[1]]), cla([2],[s],[s-[2]]),R).
responsibility_uk(cla([r1,r2,r3,r4,r5],[ag, pa,ind,fs,eubd],
		      [ag	-	[r1,r2,r3],
		       pa	-	[r1],
		       ind	-	[r2],
		       fs	-	[r4,r5],
		       eubd	-	[r4]])).

responsibility_us(cla([s1,s2,s3,s4,s5],[sos,bca,bea,age,ins],
		      [sos	-	[s1,s2,s3],
		       bca	-	[s1],
		       bea	-	[s2],
		       age	-	[s4,s5],
		       ins	-	[s4]])).

%
theory_cla_flip_pow(Th, Cla):- theory_to_cla(Th, A),
			cla(flip(A), FA),
			cla(pow(FA), Cla).



% ?- ifmap:test1(A).
% ?- test1(A).
%@ A = cla([1,2,3],[bird,fly],[bird-[3],fly-[2,3]]) .

% ?- ifmap:test2(A).
%@ A = cla([1, 2, 3, 4, 5], [penguine, bird, fly], [bird-[2, 3, 4], fly-[1, 2], penguine-[4]]) .
% ?- ifmap:test3(A, B).
% ?- trace, ifmap:test4(A).
%@ A = cla([1, 2, 3, 4, 5, 6],
%	[[bird, bird], [bird, fly], [bird, penguine], [fly, bird], [fly, fly], [fly, penguine]],
%	[[bird, bird]-[2, 3, 4, 6],
%	 [bird, fly]-[2, 5, 6],
%	 [bird, penguine]-[2, 4],
%	 [fly, bird]-[1, 2, 3, 4, 6],
%	 [fly, fly]-[1, 2, 5, 6],
%	 [fly, penguine]-[1, 2, 4]]) .

% ?- [pac(ifmap)], module(ifmap).
% ?- ifmap:theory_to_cla(theory([ag,pa,ind,fs,eubd], [[]=>[ag,fs],[ag,fs]=>[],[pa]=>[ag], [ind]=>[ag], [pa,ind]=>[], [eubd]=>[fs]]), A).
%@ A = cla([1,2,3,4,5],[ag,pa,ind,fs,eubd],[ag-[1,2,3],eubd-[5],fs-[4,5],ind-[3],pa-[2]]) .
% ?- ifmap:theory_to_cla(theory([sos,bca,bea,age,ins],[[]=>[sos,age],[sos,age]=>[],[bca]=>[sos], [bea]=>[sos], [bca,bea]=>[], [ins]=>[age]]), A).
%@ A = cla([1,2,3,4,5],[sos,bca,bea,age,ins],[age-[4,5],bca-[2],bea-[3], (ins)-[5],sos-[1,2|...]]) [write]
%@ A = cla([1, 2, 3, 4, 5], [sos, bca, bea, age, ins], [age-[4, 5], bca-[2], bea-[3], (ins)-[5], sos-[1, 2, 3]]) .

% ?- module(ifmap).
%@ true.

% ?- ifmap:theory_to_cla(theory([bird,fly], [[bird] => [fly]]), A).
% ?- ifmap:theory_to_cla(theory([a,b,c], [[] => [a],[]=>[b],[]=>[c]]), A).
%@ A = cla([1], [a, b, c], [a-[1], b-[1], c-[1]]) .
% ?- ifmap:theory_to_cla(theory([a,b,c], [[a] => [b],[b]=>[c]]), A).
%@ A = cla([1,2,3,4],[a,b,c],[a-[4],b-[3,4],c-[2,3,4]]) .
% ?- ifmap:theory_to_cla(theory([a,b,c,d], [[a] => [b],[a]=>[c],[b] => [d],[c] =>[d]]), A).
%@ A = cla([1, 2, 3, 4, 5, 6], [a, b, c, d], [a-[6], b-[3, 5, 6], c-[4, 5, 6], d-[2, 3, 4, 5, 6]]) .
% ?- ifmap:theory_to_cla(theory([bird,fly,penguin], [[bird] => [fly],[penguin,fly]=>[]]), A).
%@ A = cla([1, 2, 3, 4], [bird, fly, penguin], [bird-[3], fly-[2, 3], penguin-[4]]) .
% ?- ifmap:theory_to_cla(theory([bird,fly,penguin], [[penguin]=>[bird],[bird] => [fly],[penguin,fly]=>[]]), A).
%@ A = cla([1, 2, 3], [bird, fly, penguin], [bird-[3], fly-[2, 3]]) .


test1(A) :-  ifmap:theory_to_cla(theory([bird,fly], [[bird] => [fly]]), A).
test2(A) :-  ifmap:theory_to_cla(
	theory([penguine, bird, fly],
	       [[penguine, fly] => [],
		[penguine] => [bird]]
	      ),
		A).
test3(C, M) :- test1(A),
	test2(B),
	ds_create((bird-A)+(penguine-B), DS),
	ds_cover(DS, C, M).

test4(C) :- test1(A),
	test2(B),
	cla(flip(A), A0),
	cla(flip(B), B0),
	ds_create((bird-A0) + (penguine-B0), Ds),
	ds_cover(Ds, C0, _),
	cla(flip(C0), C).

% ?- ifmap:ds_cover(
% ds(	[	bird-cla(	[bird, fly], [1, 2, 3], [2-[fly], 3-[bird, fly]]),
%		penguine-cla(	[penguine, bird, fly],
%				[1, 2, 3, 4, 5],
%				[2-[bird], 3-[bird, penguine], 4-[fly], 5-[bird, fly]])],
%	[]),
% cla(	[[bird, bird], [bird, fly], [bird, penguine], [fly, bird], [fly, fly], [fly, penguine]],
%	[1, 2, 3, 4, 5, 6],
%	[	1-[[fly, bird], [fly, fly], [fly, penguine]],
%		2-[[bird, bird], [bird, fly], [bird, penguine], [fly, bird], [fly, fly], [fly, penguine]],
%		3-[[bird, bird], [fly, bird]],
%		4-[[bird, bird], [bird, penguine], [fly, bird], [fly, penguine]],
%		5-[[bird, fly], [fly, fly]],
%		6-[[bird, bird], [bird, fly], [fly, bird], [fly, fly]]]
%   ),
% [	im(penguine,	[5-6, 4-5, 3-4, 2-3],	proj(penguine)),
%	im(bird,	[3-2, 2-1],		proj(bird))
% ]
%   ).

% ?- ifmap:test(a, b, R).
%@ R = a-b.
% test(A) --> pred(A, [X, A-X]).

% ?- ifmap:cover_test(Cover, Tmap).
%@ Cover = cla([[x, a, a, a, a, x]], [1, 2], [1-[[x, a, a, a, a, x]], 2-[[x, a, a, a, a, x]]]),
%@ Tmap = [1-[2:x, 1:a, 3:[x]], 2-[5:x, 6:a, 4:[x]]] .

% ?- [pac(ifmap)].
% ?- trace, ifmap:demo_ontology(Cover, Tmap).
%@ Cover = cla([[1, x, a, a, c, c, u], [1, x, a, a, d, d, u]], [1, 2, 3], [1-[[1, x, a, a, c, c, u], [1, x, a, a, d, d, u]], 2-[[1, x, a, a, d, d, u]], 3-[[1, x, a, a, c, c, u], [1, x, a, a, d, d, u]]]),
%@ Tmap = [1-[1:a, 1:b, 3:x, 6:u, 2:a, 4:[x], 7:c, 5:[u]], 2-[6:v, 5:[v]], 3-[7:d, 5:[u, v]]] .

% ?- [pac(ifmap)].

% ?- bagof(SequentsOut, ifmap:debug(theory([a,b], [[a]=> [b]]), left-cla([x], [a], [a-[x]]), right-cla([u,v], [c,d], [c-[u],d-[u,v]]), SequentsOut), S), length(S, L), sort(S, S0).
% ?- trace, ifmap:debug(theory([a,b], [[a]=>[b]]), left-cla([x], [a], [a-[x]]), right-cla([u,v], [c,d], [c-[u],d-[u,v]]), SequentsOut).

%@ SequentsOut = [[left:a] => [right:c], [right:c] => [left:a]] .

% ds -- distributed system, or sketch (acronym).

% debug(Theory, Left, Right, SequentsOut):-
% 	Left= (A-_),
% 	Right = (B-_),
% 	Index = [A, B],
% 	once(sample_ds(Left, Right, E, FlipLeft, FlipRight)), %
% 	ds_create(E, Ds),
% 	ds_check_info_map(Ds),
% 	theory_to_cla(Theory, M),
% 	get_cla(FlipLeft, Ds, C2),
% 	get_cla(FlipRight, Ds, D2),
% 	build_info_map(M, C2, im(LU, LL)),
% 	build_info_map(M, D2, im(RU, RL)),
% 	ds_add(im(theorycla, FlipLeft, LU, LL)
% 	      + im(theorycla, FlipRight, RU, RL)
% 	      + (theorycla- M),
% 	       Ds,  Sketch),
% 	ds_cover(Sketch, Cover, _Tmap),
% 	maplist(Sketch^[K, J]:- get_key_pos(K, Sketch, J), Index, I0),
% 	get_key_pos(theorycla, Sketch, N),
% 	Theory=theory(_, Sequents),
% 	maplist(N^[=>(L,R), =>(L0, R0)]
% 	       :- (maplist(N^[A, N:A], L, L0),
% 		   maplist(N^[A, N:A], R, R0)),
% 		Sequents, Sequents0),
% 	Cover= cla(_, Partition, _),
% %	zip(_, Partition, Tmap),
% 	partition_to_restricted_massoc(Partition, N, I0, Massoc),
% 	subst_theory_multiple(Massoc, Sequents0, Sequents1),
% 	flatten(Sequents1, Sequents2),
% 	subst_types_back(Sequents2, Sketch, Sequents3),
% 	simplify_theory(Sequents3, SequentsOut).

%
sample_ds(A, B,
	  F13 + F23 + G13 + G23 + A
	 + (c2-C2) + (c3-C3) + (d3-D3) + (d2-D2) + B,
	c2, d2)
	:-
	A=(I-C1),
	B=(J-D1),
	cla(flip(C1), C2),    % flip class
	cla(pow(C2), C3),     % power class
	cla(flip(D1), D2),
	cla(pow(D2), D3),
	power_cla_inclusion(C2, U23, L23),
	F23 = im(c2, c3, U23, L23),
	cla(supp(C1), U13),
	section(U13, L13),
	F13 = im(I, c3, U13, L13),
	power_cla_inclusion(D2, V23, M23),
	G23 = im(d2, d3, V23, M23),
	cla(supp(D1), V13),
	section(V13, M13),
	G13 = im(J, d3, V13, M13).

% ?- [pac(ifmap)].
% ?-ifmap:partition_to_restricted_massoc([[a:x, b:y, c:z]], a,[b,c], R).
%@ R = [[a:x]-[b:y, c:z]] .
% ?-ifmap:partition_to_restricted_massoc([[a:x, b:y, c:z], [a:u, a:v, c:z]],a,[b,c], R).
%@ R = [[a:x]-[b:y, c:z], [a:u, a:v]-[c:z]] .

partition_to_restricted_massoc(P, I, S, A):-
	maplist(pred([I, S], [X, Y]
	       :-  foldr(
		       pred([I,S],	([J:U, L-R, L-[J:U|R]]
						:- memberchk(J, S), ! )
					& ([I:U, L-R, [I:U|L]-R]:- ! )
					& [_, LR, LR]),
			 X, []-[], Y)),
		P, A).


% ?- ifmap:subst_types_back([[1:x] => [2:y]], ds([a,b],dummy,dummy), T).
%@ T = [[a:x]=>[b:y]].
