:- module(ifmap, [ ds_cover/3, parse_comma_list/2, parse_comma_list/3,
		  theory_to_cla/2,
		  cla/2,
		  show_cla/1, show_theory_cla/1,
		  theory_sum_tokens/2,
		  ds_theory_cover_with_constraint/5,
%		  ds_cover_with_constraint/2, ds_cover_with_constraint/3,
		  invertable_type_map/4,
		  invertable_token_map/3, invertable_token_map/4
	       ]).   % [2015/04/19], [2015/12/20]

:- style_check(-singleton).
:- use_module(library(ordsets)).
:- use_module(pac('expand-pac')).
:- use_module(util(misc)).
:- use_module(util(meta2)).
:- use_module(util(snippets)).
:- use_module(pac(basic)).
:- use_module(util(file)).
:- use_module(util(obj)).
:- use_module(util('emacs-handler')).
:- use_module(util(coalgebra)).
:- use_module(util(math)).

term_expansion --> pac:expand_pac.

:- use_module(pac(op)).

:- op(700, xfx, =>).

		/**********************
		*     tiny helpers    *
		**********************/

id_map(X):- maplist(pred([A-A]), X).

%!	show_theory_cla(+T is det.
%	View the classification of theory T in graphics using dot.

% ?- show_theory_cla(theory([a,b], [[a => b]])).
%@ true.
show_theory_cla(Theory) :- once(theory_to_cla(Theory, Z)),
			      once(show_cla(Z)).

%
tok(cla(X,_,_), X).
typ(cla(_,X,_), X).
sup(cla(_,_,X), X).

		/********************************
		*     ds programming helpers    *
		********************************/

% ?- partition_to_restricted_massoc([[a:x, b:y, c:z]], a,[b,c], R).
% ?- partition_to_restricted_massoc([[a:x, b:y, c:z], [a:u, a:v, c:z]],a,[b,c], R).

partition_to_restricted_massoc(P, I, S, A):-
	maplist(pred([I, S], [X, Y]
	       :-  foldr(
		       pred([I,S],	([J:U, L-R, L-[J:U|R]]
						:- memberchk(J, S), ! )
					& ([I:U, L-R, [I:U|L]-R]:- ! )
					& [_, LR, LR]),
			 X, []-[], Y)),
		P, A).

% ?- subst_types_back([[1:x]=>[2:y]], ds([a,b],dummy,dummy), T).
%@ T = [([a:x]=>[b:y])].

subst_types_back(Theory, ds(Keys, _, _), Theory0):- !,
	keys_to_assoc(Keys, A),
	maplist(pred(A, [=>(L, R), =>(L0, R0)]
		:- (	maplist(pred(A, [I:T, K:T]
				:- memberchk(K-I, A)),
				L, L0),
			maplist(pred(A, [I:T, K:T]
				:- memberchk(K-I, A)),
				R, R0)
		   )),
		Theory, Theory0).
subst_types_back(Theory, ds(Cs, Ms), Theory0):- zip(Keys, C0s, Cs),
	subst_types_back(Theory, ds(Keys,C0s,Ms), Theory0).

% ?- get_key_pos(a, ds([b-1, d-3, a-2], any), K).
% ?- get_key_pos(a, ds([b, d, a], [b, d, a], any), K).
get_key_pos(K, ds(A,_,_), I):- get_key_pos(K, A, 1, I).
get_key_pos(K, ds(Cs,_), I) :- get_key_assoc_pos(K, Cs, 1, I).
%
get_key_pos(K, [K|_], I, I):-!.
get_key_pos(K, [_|A], I, J):- I0 is I+1,
	get_key_pos(K, A, I0, J).
%
get_key_assoc_pos(K, [K-_|_], I, I):-!.
get_key_assoc_pos(K, [_|Cs], I, J):- I0 is I+1,
	get_key_assoc_pos(K, Cs, I0, J).
%
get_cla(K, ds(Ds,_), X):- memberchk(K-X, Ds), !.
get_cla(K, ds(L, Ds,_), X):- get_cla(K, L, Ds, X).
%
get_cla(K, [K|_], [X|_], X):-!.
get_cla(K, [_|Ks], [_|Ds], X):- get_cla(K, Ks, Ds, X).

% ?- ds_build([a:cla([x],[t], [t-[x]])], DS).
% ?- ds_build(a:cla([x],[t], [t-[x]]), DS), keyed_ds_cover(DS, CLA).
% ?- ds_build((a-cla([x],[t], [t-[x]])); (a-cla([x],[t], [t-[x]])), DS), keyed_ds_cover(DS, CLA).
% ?- ds_build((a-cla([x],[t], [t-[x]]));(b-cla([x],[t], [t-[x]])), DS), keyed_ds_cover(DS, CLA).
% ?- ds_build((a-cla([x],[t], [t-[x]]));(b-cla([x],[t], [t-[x]])); im(a,b,[t-t],[x-x]), DS), keyed_ds_cover(DS, CLA).
% ?- ds_build((a-cla([x],[t], [t-[x]])); (b-cla([x],[t], [t-[x]])); im(a,b,[t-t],[x-x]); im(b,a,[t-t],[x-x]), DS), keyed_ds_cover(DS, CLA).

ds_build(A, ds(Clfs, Morphs)):-
    once(ds_build(A, Clfs, [], Morphs, [])).
%
ds_build([], X, X, Y, Y):-!.
ds_build([A], X, X0, Y, Y0):-!, ds_build(A, X, X0, Y, Y0).
ds_build([A|B], X, X0, Y, Y0):-!, ds_build(A;B, X, X0, Y, Y0).
ds_build(A ; B, X, X0, Y, Y0):-!,
    ds_build(A, X, X1, Y, Y1),
    ds_build(B, X1, X0, Y1, Y0).
ds_build(A, X, X0, Y, Y0):- ds_add_basic(A, X, X0, Y, Y0), !.

%
ds_add_basic(Name := A, [Name:A|X], X, Y, Y).
ds_add_basic(Name = A, [Name:A|X], X, Y, Y).
ds_add_basic(Name - A, [Name:A|X], X, Y, Y).
ds_add_basic(Name : A, [Name:A|X], X, Y, Y).
ds_add_basic(B, X, X, [B|Y], Y).

% ?- restrict_index_set([a-[1:a,2:b]], [1], R).
% ?- restrict_index_set([a-[1:a,2:b], b-[2:c]], [1], R).
% ?- restrict_index_set([a-[1:a,2:b], b-[2:c, 1:d]], [1], R).
% ?- restrict_index_set([a-[1:a,2:b], b-[2:c, 1:d]], [1,2], R).
restrict_index_set(T, I, T1):-
	maplist(pred(I, [A-S, A-S0]
	       :-  foldr(
		       pred(I, ([J:X, L, [J:X|L]]:- memberchk(J, I))
				&
				([_, L, L])) ,
			 S, [], S0)),
		T, T0),
	foldr(pred([_-[], L, L] & [X, L, [X|L]]), T0, [], T1).

% ?- eval(ifmap:cla @ flip @ empty, X).
% ?- ifmap:cla(flip(cla([1,2],[a,b], [a-[1,2], b-[1,2]])), R).
% ?- ifmap:cla(empty, X).
% ?- eval(ifmap:cla::((pow @ empty)), V).
% ?- eval(ifmap:cla::((pow @  empty)), V).
% ?- ifmap:numcla(1, 10, X).

% ?- eval(ifmap:cla :: (+) @ :numcla(1,3) @ :numcla(4,5), R).
% ?- eval(ifmap:cla :: pow @ ((+) @ :numcla(1,3) @ :numcla(4,5)), R).
% ?- eval(ifmap:cla :: merge @ :numcla(1,3)@ :numcla(2,4),  X).
% ?- eval(ifmap:cla(pow :: merge @ :numcla(1,3)@ :numcla(2,4),  X).
% ?- eval(ifmap:cla(clean_disj(pow(cla([y,x,y],[a,b],[a-[x,x],b-[y]])))), Boole).

% ?- ifmap:cla(tok(empty), X).
% ?- ifmap:cla(typ(empty), X).
% ?- ifmap:cla(pow(cla([1,2],[a,b], [a-[1],b-[2]])),  X).
% ?- ifmap:cla(supp(empty), X).
% ?- ifmap:cla(merge(cla([1],[a],[a-[1]]), empty), X).
% ?- ifmap:cla(merge(cla([1],[a],[a-[1]]), cla([1],[a],[a-[1]])), X).

% ?- cla(flip(cla([1],[a],[a-[1]])), R).
% ?- cla(flip(cla([1],[a],[a-[1,2]])), R).
% ?- cla(pow(cla([1],[a, b],[a-[1], b-[2]])), R).
% ?- cla(cla([1],[a], [a-[1]]) + cla([2],[b], [b-[2]]), R).

cla(empty, cla([],[],[])):-!.
cla(singleton(X), cla([A],[],[])):-!, cla(X, A).
cla(tok(X), A):-!, cla(X, cla(A, _, _)).
cla(typ(X), A):-!, cla(X, cla(_, A, _)).
cla(supp(X), A):-!, cla(X, cla(_, _,A)).
cla(flip(X), cla(Q, P, R)):-!, cla(X, cla(P, Q, R0)),
	flip_pfn(R0, R).
cla(pow(X), cla(K,A0,S0)):-!,
	cla(X, cla(K, A, S)),
	set(pow(A), A0),
	power_cla(S, A0, S0).
cla(X + Y, A1):-!, cla(X, A2),
	cla(Y, A3),
	cla_binary_sum(A2, A3, A1).
cla(merge(A, B), cla(A1, A2, A3)):-!,
	cla(A, cla(X, S, U)),
	cla(B, cla(Y, T, V)),
	misc:set(X + Y, A1),
	misc:set(S + T, A2),
	merge_support(U, V, A3).
cla(cleaning_boole(X), A1):-!,cla(X, A2),
	disjoint_boole_sort(A2, A1).
cla(sort(X), A1):-!, cla(X, A2),
	sort_cla(A2, A1).
cla(normal(X), A1):-!, cla(X, A2),
	cla_normalize(A2, A3),
	canonical_cla(A3, A1).
cla(X, X):-!.

% ?- ifmap:support([a-[1,2]], 1, a).
% ?- ifmap:support([a-[1,2]], 1, a).
% ?- ifmap:support([], 3, a).
fn_support(S, X, A):- member(B-U, S), B==A, !,
		       memberchk(X, U).
%
support(S, X, A):- S = [_|_], !, fn_support(S, X, A).
support(P, X, A):- P\==[], call(P, X, A).

%
disjoint_boole_sort(cla(X,A,S), cla(X0,A0,S0)):-
	sort(X, X0),
	maplist(pred([P,Q]:- sort(P, Q)), A, A1),
	sort(A1, A0),
	sort_support(S, S1),
	keysort(S1, S0).

% ?- ifmap:sort_support([[a,a]-[1,2,2],[b,a]-[3,5,4]], R).
%@ R = [[a]-[1, 2], [a, b]-[3, 4, 5]].
sort_support([_-[]|R], S):- !, sort_support(R, S).
sort_support([K-A|R], [K0-A0|S]):- is_list(K), !,
	sort(K, K0),
	sort(A, A0),
	sort_support(R, S).
sort_support([K-A|R], [K-A0|S]):-
	sort(A, A0),
	sort_support(R, S).
sort_support([], []).

% ?- merge_support([a-[2]],[a-[3]], R).
% ?- merge_support([b-[2], a-[1]], [a-[3]], R).
merge_support(X, Y, Z):- sort_support(X, X0),
			 sort_support(Y, Y0),
			  merge_sorted_support(X0, Y0, Z).
%
merge_sorted_support([], X, X).
merge_sorted_support(X, [], X).
merge_sorted_support([A-U|R], [A-V|S], X):- !,
    union(U,V,W),
    merge_sorted_support([A-W|R], S, X).
merge_sorted_support([A-U|R], [B-V|S], [A-U|X]):-  A@<B, !,
	merge_sorted_support(R, [B-V|S], X).
merge_sorted_support(R, [B-V|S], [B-V|X]):-  merge_sorted_support(R, S, X).

% ?- eval((cla::(+)) @ (:numcla(1,2)) @ (:numcla(3,4)),  R ).
numcla(N0, N, cla(X, X, S)):- numlist(N0, N, X),
	maplist(pred([I, I-[I]]), X, S).

% ?- X = cla([1,2],[a,b],[a-[1]]), ifmap:cla(normal(X), Y).
% ?- X = cla([1,2],[a,b],[a-[1],b-[2]]), cla(normal(X+X+X), Y), cla(sum(Y,Y), Z).
% ?- cla(flip(cla([1,2,3], [a,b,c], [a-[1,2], b-[2,3], c-[1,3]])), F).
cla_binary_sum(cla(X,Y,Z), cla(P,Q,R), cla(U,V,W)):-
	set(X - P, U),
	maplist(pred([A, (2:A)]), Q, Q0),
	foldr(pred([A, L, [(1:A)|L]]), Y, Q0, V),
	maplist(pred(X, [A-B, (2:A)- B0]:- set(X-B, B0)), R, R0),
	foldr(pred(P, [A-B, L, [(1:A)- B0|L]]  :- set(B-P, B0)), Z, R0, W).

%
cla_normalize(cla(X, Y, Z), cla(X0, Y0, Z0)):-
	maplist(normalize_token, X, X0),
	normalize_type(Y, Y0, D),
	maplist(pred(D, [S-U, S0-U0]:- (maplist(normalize_token, U, U0),
				  fresh_index_var(S,[], P, V, S0),
				  memberchk(P-V,D))),
		Z, Z0).

% ?- normalize_token((a-(b-c)), X).
normalize_token(A, X) :-  normalize_token(A, X, []).

normalize_token(A-B, X, Y):- !, normalize_token(A, X, X0),
	normalize_token(B, X0, Y).
normalize_token(A, [A|X], X).

%
normalize_type(X, Y, Assoc):-
	maplist(pred([A, B, P-V]:- fresh_index_var(A, [], P, V, B)), X, Y, Assoc0),
	keysort(Assoc0, Assoc1),
	consecutive_merge(Assoc1, Assoc),
	length(Assoc, L),
	numlist(1, L, S),
	maplist(pred([X-Y, Y]), Assoc, S).
%
consecutive_merge([], []).
consecutive_merge([X,X|Y], Z):- !, consecutive_merge([X|Y],Z).
consecutive_merge([X|Y], [X|Z]):- consecutive_merge(Y, Z).

%%	rel_pfn_flip(?X:rel, ?Y:pfn) is det.
%   True if for all X(a, k) <==> a in Y(k).

% ?- rel_pfn_flip([a-1], R).
% ?- rel_pfn_flip([a-1, b-1], R).
% ?- rel_pfn_flip([a-1, b-2, c-1], R).
rel_pfn_flip(X, Y):- nonvar(X), !,
	flip_rel(X, Y0),
	sort(Y0, Y1),
	rel_to_pfn(Y1, Y).
rel_pfn_flip(X, Y):- rel_to_pfn(X0, Y),
	flip_rel(X0, X1),
	sort(X1, X).

%%	fn_to_pfn_flip(?X:fn, ?Y:pfn) is det.
%	True if X is a function and Y is a set function
%	such that X(a) = k <==>  a in Y(k), i.e.,
%	Y represents the inverse of X.
%   This predicate is bidrectional, and X is checked
%	to be a function when X is instantiated.

% ?- fn_to_pfn_flip([1-b, 2-c, 3-a], R).
% ?- fn_to_pfn_flip([1-b, 2-c, 3-a], R), ifmap:fn_to_pfn_flip(F, R).
% ?- fn_to_pfn_flip([1-b, 1-c, 3-a], R), ifmap:fn_to_pfn_flip(F, R). % false
%@ false;  because [1-b, 1-c, 3-a] is not a function, but a relation.

rel_pfn_flip(X, Y):-
	(	nonvar(X) -> rel_to_pfn_flip(X, Y)
	;	rel_to_pfn_flip(X, Y),
		is_function(X)
	).
%
is_function([A-_, B-_|R]):- A\==B, !, is_function([B-_|R]).
is_function([_]).
is_function([]).

% ?- fn_to_qset([a-1, b-2, c-1], R).
fn_to_qset(X, Y):- fn_to_pfn_flip(X, Y0),
	zip(_, Y, Y0).

% ?- fn_to_zip(=, [1,2,3], R).
:- meta_predicate fn_to_pfn_flip(2,?,?).

fn_to_zip(F, X, Z):- maplist(pred(F, [A, A-V]:- call(F, A, V)), X, Z).

:- meta_predicate map(2,2,?,?).
% ?- map(=, =, =, a, X).

map(F, G, H, X, Y):- call(F, X, X0),
	call(G, X0, Y0),
	call(H, Y0, Y),
	!.

sort_rel(X,Y) :- sort(X, Y).

% [2013/07/18]
%%	 flip_pfn(?S:fn, ?F:fn) is det.
%	 flip the support relation S to a fn F so that
%	 y in F(x) <==>  x in  S(y).
%    Property: if F is a sorted fn then so is S, and
%	 vice versa.
% ?- H=[a-[1,2], b-[2,3], c-[1,3]], flip_pfn(H, F), flip_pfn(F, G).
% ?- H=[a-[1,2], b-[2,3], c-[1,3]], flip_pfn(H, F), flip_pfn(G, F).
% ?- H=[a-[1,2], b-[2,3], c-[1,3]], flip_pfn(H, F), flip_pfn(G, F), flip_pfn(F, G).

flip_pfn(X, Y):-
	(	nonvar(X) ->
		X = FnA,
		Y = FnB
	;	X = FnB,
		Y = FnA
	),
	pfn_to_rel(FnA, R, []),
	flip_rel(R, R0),
	sort(R0, R1),
	rel_to_pfn(R1, FnB, []).

%
flip_rel([], []).
flip_rel([A-B|R], [B-A|S]):- flip_rel(R, S).



% ?- N = 100,  findall(I-I, between(1, N, I), Zip),
%	time(repeat(1000, rel_pfn_by_findall(Zip, R))).

% %
% rel_pfn_by_findall([1-a, 1-b, 2-b],R).
% %@% rel_pfn_by_findall(X, Y):- findall(A-L, (member(A-_, X), bagof(U, member(A-U, X), L)), Y).
% rel_pfn_by_findall(X, Y):- findall(A-L, (member(A-_, X), bagof(U, member(A-U, X), L)), Y0),
% 						   sort(Y0, Y).

% ?- N = 1000,  findall(I-I, between(1, N, I), Zip),
%	time(repeat(1000, rel_pfn_by_findall(Zip, R))).
%@ % 1,016,015,001 inferences, 79.852 CPU in 80.332 seconds (99% CPU, 12723746 Lips)
%@ N = 1000,
%@ Zip = [1-1, 2-2, 3-3, 4-4, 5-5, 6-6, 7-7, 8-8, ... - ...|...].

% ?- N = 1000,  findall(I-I, between(1, N, I), Zip),
%	time(repeat(1000, rel_pfn(Zip, R))).
%@ % 2,006,001 inferences, 0.162 CPU in 0.162 seconds (100% CPU, 12371954 Lips)
%@ N = 1000,
%@ Zip = [1-1, 2-2, 3-3, 4-4, 5-5, 6-6, 7-7, 8-8, ... - ...|...].

%%	rel_pfn(?X:rel, ?Y:fn) is det.
%	Y is a coalgebra version of a relation X, i.e.,
%	X(a, b) <==> b in Y(a).
%   Property: If X is sorted then so is Y, and vice versa.
% ?- rel_pfn([1-a, 1-b, 2-b],R).
% ?- S = [1-a, 1-b, 2-b], rel_pfn(S, R), ifmap:rel_pfn(S0, R).

rel_pfn(X, Y):-
	(	var(X) -> pfn_to_rel(Y, X)
	;	rel_to_pfn(X, Y, [])
	).
%
rel_to_pfn([], Y, Y).
rel_to_pfn([A-B|R], [A-[B|P]|Y], Z):-
	rel_to_pfn(R, A, R0, P, []),
	rel_to_pfn(R0, Y, Z).
%
rel_to_pfn([], _, [], P, P).
rel_to_pfn([A-B|R], A, U, [B|P], Q):-!,
	rel_to_pfn(R, A, U, P, Q).
rel_to_pfn(R, _, R, P, P).

%%	fn_to_rel(+X:fn, -Y:rel) is det.
%	Y is a relation version of X, i.e.,
%   b in X(a)  <==> Y(a, b).
%   Property;  if X is sorded  then so is Y.

% ?- rel_to_pfn([a-b, a-c, b-d], X), ifmap:fn_to_rel(X, Y).
pfn_to_rel(X, Y):- pfn_to_rel(X, Y0, []),
				   keysort(Y0, Y).

%
pfn_to_rel([], Y, Y).
pfn_to_rel([U-L|X], Y, Z):-
	pfn_to_rel(L, U, Y, Y0),
	pfn_to_rel(X, Y0, Z).
%
pfn_to_rel([], _, Z, Z).
pfn_to_rel([V|Vs], U, [U-V|Y], Z):-
	pfn_to_rel(Vs, U, Y, Z).



%
fresh_index_var(J:A, P, [J|Q], V, B):- !, fresh_index_var(A, P, Q, V, B).
fresh_index_var(A, P, [0|P], V, V:A).

% ?- numcla(1,2,A), ifmap:numcla(3,4,B), ifmap:cla_list_sum([A,B], C).
% ?- cla_list_sum([cla([],[a,b],[])], X).
cla_list_sum(R0, cla(X, Y, Z)):-
	maplist(pred([cla(A,_,_), A]), R0, R1),
	maplist(pred([cla(_,A,_), A]), R0, R2),
	maplist(pred([cla(_,_,A), A]), R0, R3),
	choices(R1, X),
	type_sum(R2, 1, [], Y),
	fresh_support_index(R3, 1, [], Z0),
	reverse(R1, P),
	expand_support(Z0, P, [[]], [], Z).

%
cla_list(A+B, X, Y):- !, cla_list(A, X, Y0), cla_list(B, Y0, Y).
cla_list(A, [A|X], X).


		/*************************************************
		*     ds_core: colimit of distributed system.    *
		*************************************************/


% ?- ds_core([cla([x],[a], [a-[x]])], [im(1->1, [a-a], [x-x])], Cla).
ds_core(Cs, Is, cla(Tok, Typ, Sup)):- initial_type_sum(Cs, 1, Singletons, []),
	ds_type_colimit(Is, Singletons, Typ),
	ds_token_limit(Cs, Is, Tok),
	core_support_rel_as_fun(Cs, Typ, Tok, Sup).

% ?- initial_type_sum([cla([], [a,b], [])], 1, R, []).
% ?- initial_type_sum([cla([], [a,b], []), cla([], [a,b], [])], 1, R, []).
initial_type_sum([], _, X, X).
initial_type_sum([cla(_,As,_)|Cs], I, X, Y):- raise_to_singleton(As, I, X, Z),
	J is I + 1,
	initial_type_sum(Cs, J, Z, Y).
%
raise_to_singleton([], _I, X, X).
raise_to_singleton([A|As], I, [[I:A]|X], Y):- raise_to_singleton(As, I, X, Y).

% ?- indexed_union_find_by_map([a-b], 1, 2, [], X).
% ?- indexed_union_find_by_map([a-b, c-b], 1, 2, [], X).
% ?- indexed_union_find_by_map([a-b, c-b, d-e], 1, 2, [], X).
indexed_union_find_by_map([], _, _, X, X).
indexed_union_find_by_map([A-B|Rest], I, J, X, Y):-
	union_find(I:A, J:B, X, Z),
	indexed_union_find_by_map(Rest, I, J, Z, Y).

% ?- ds_type_colimit([im(1->2, [a-a], [])], [], X).
% ?- ds_type_colimit([im(1->2, [a-a, b-a], [])], [], X).
% ?- ds_type_colimit([im(1->2, [a-a, b-a, c-d], [])], [], X).
% ?- ds_type_colimit([im(1->2, [a-a, b-a, c-d], []), im(2->1, [a-a, b-a, c-d], [])], [], X).
ds_type_colimit([], X, X).
ds_type_colimit([im(I->J, F, _)|R], X, Y):-
	indexed_union_find_by_map(F, I, J, X, Z),
	ds_type_colimit(R, Z, Y).

% ?- ds_type_sum([cla([], [a,b], [])], S).
% ?- ds_type_sum([cla([], [a,b], []), cla([], [a,b,c], [])], S).
ds_type_sum(DS, S):- length(DS, N),
	numlist(1, N, Ns),
	ds_type_sum(DS, Ns, S, []).
%
ds_type_sum([], [], S, S).
ds_type_sum([Cla|R], [J|Js], S, S0):- typ(Cla, T),
	 attach_index(T, J, S, S1),
	 ds_type_sum(R, Js, S1, S0).
%
attach_index([], _, S, S).
attach_index([X|Xs], J, [J:X|S], S0):-	attach_index(Xs, J, S, S0).

% ?- ds_token_limit([cla([x],[a], [a-[x]])], [im(1->1, [a-a], [x-x])], Limit).
%@ Limit = [[x]].
% ?- ds_token_limit([cla([x],[a], [a-[x]]), cla([x],[a], [a-[x]])], [im(1->2, [a-a], [x-x]),
% im(2->1, [a-a], [x-x])], Limit).
%@ Limit = [[x, x]].
ds_token_limit(Cla_list, Ims, Limit):-token_choices(Cla_list, Tok_choices),
	filter_choices_by_ims(Tok_choices, Ims, Limit).

% ?- token_choices([], R).
% ?- token_choices([cla([a,b], [], [])], R).
% ?- token_choices([cla([a,b], [], []), cla([a,b], [], [])], R).
token_choices([], [[]]).
token_choices([Cla|As], X):- token_choices(As, X0),
	tok(Cla, Tok),
	choices(Tok, X0, X, []).
%
choices([], _, Y, Y).
choices([A|As], Ch, X, Y):-
	cons_choice(Ch, A, X, X0),
	choices(As, Ch, X0, Y).
%
cons_choice([], _A, X, X).
cons_choice([B|Bs], A, [[A|B]|X], Y):- cons_choice(Bs, A, X, Y).

% ?- filter_choices_by_ims([[a, b, c],[b, c, a]], [im(2->1, [], [a-b]), im(3->2, [], [b-c])], R).
%@ R = [[a, b, c]].
% ?- filter_choices_by_ims([[b, a]], [im(1->2, [], [a-b]), im(2->1, [], [b-a])], R).
%@ R = [[b, a]].
filter_choices_by_ims([], _, []).
filter_choices_by_ims([X|R], Ims, Z):-
	(	ims_compatible(Ims, X)  ->  Z = [X|U]
	;	Z = U
	),
	filter_choices_by_ims(R, Ims, U).

% ?- ims_compatible([im(1->2, [], [b-a]), im(2->1, [], [a-b])], [a, b]).
ims_compatible([], _).
ims_compatible([im(I->J, _, F)|R], X):-  nth1(I, X, Xi),
	nth1(J, X, Xj),
	memberchk(Xj-Xi, F),
	ims_compatible(R, X).

% ?- ds_core_support([cla(_,_,[a-[b]]), cla(_,_,[a-[b]])], [b, b], [1:a, 2:a]).
% ?- ds_core_support([cla(_,_,[a-[b]]), cla(_,_,[a-[b]])], [b, b], []).
ds_core_support(Cs, Choice, Ts):- core_support(Ts, Choice, Cs).
%
core_support([], _, _).
core_support([J:T|R], Choice, Cs):-
	nth1(J, Cs, C),
	sup(C, S),
	nth1(J, V, X),
	memberchk(T-U, S),
	memberchk(X, U),
	core_support(R, Choice, Cs).

% ?- core_support_rel_as_fun([cla([x],[a], [a-[x]])], [[1:a]], [[x]], Fun).
%@ Fun = [[1:a]-[[x]]].
core_support_rel_as_fun(Cs, Colimit_types, Choices, Fun):-
	support_as_fun(Colimit_types, Choices, Fun, Cs).
%
support_as_fun([], _Choices, [], _).
support_as_fun([T|Ts], Choices, [T-S|R], Cs):-
	core_support(Choices, T, S, Cs),
	support_as_fun(Ts, Choices, R, Cs).
%
core_support([], _T, [], _Cs).
core_support([Vec|R], T, Y, Cs):-
	(	core_support(T, Vec, Cs) ->  Y = [Vec|U]
	;	Y = U
	),
	core_support(R, T, U, Cs).

% ?- choices([[a,c],[b,d]], X).
% ?- type_sum([[a], [a]], 1, [], _50848)
% ?- type_sum([[a,b],[c,d,e]], 1, [], X).

type_sum([], _, C, C).
type_sum([Y|R], I, C, D):- J is I + 1,
	maplist(pred(I, [A, I:A]), Y, Y0),
	append(Y0, C, C0),
	type_sum(R, J, C0, D).

%
fresh_support_index([], _, C, C).
fresh_support_index([S|R], I, C, D):- J is I + 1,
	maplist(pred(I, [A-U, (I:A)-U]), S, S0),
	fresh_support_index(R, J, [S0|C], D).

%
expand_support([], _, _, C, C).
expand_support([S|R], [U|V], M, C, D):-
	maplist(pred(U, [S-X,S-Y]
	       :- scons(U,X,Y)), C, C0),
	foldr(pred([U,M], [W-L, Z, [W-N|Z]]
	     :- scons(L, M, N)),
	      S, C0, C1),
	scons(U, M, M0),
	expand_support(R, V, M0, C1, D).

%
refresh_cla(X, cla(A, B, C), Tmap):- canonical_cla(X,  cla(A, _, S)),
	once(rename_types(S, C, Tmap)),
	zip(B, _, C).
%
canonical_cla(cla(X,Y,Z), cla(X0, Y0, Z0)):-
	sort(X, X0),
	sort(Y, Y0),
	maplist(pred([A-B, A-B0] :- sort(B, B0)), Z, Z1),
	sort(Z1, Z0).

% ?- eval(#(pred([[A-B],[B-A]]))@[1-a], R).
% ?- eval(pred([[A-B],[B-A]])@[1-a], R).
% ?- eval(ifmap:rel(-[1-a,2-b,3-c]), R).
% ?- eval(ifmap:rel(-[1-a,2-b,3-c,4-a]), R).
% ?- eval(ifmap:fn@(quote(choice)@(misc:set(pow([a,b]) \ singleton([])))), X).
% ?- eval(ifmap:fn@(quote(choice)@ [[a,b],[c,d]]), R).
% ?- eval(fn::[], X).
% ?- eval(fn::[1-a,2-b], X).
% ?- eval(ifmap:fn::dom([1-a,2-b]), X).    % unknow error fn
% ?- eval(ifmap:fn:: -([1-a,2-b]), X).
% ?- eval(misc:set(pow([a,b])), V).
% ?- eval(ifmap:rel(-[a-1, b-2]),V).

%% def of fn is off.

% ?- fn(id([a,b]), R).
% ?- fn([a-a, b-b], F), fn((F;F;F;F), G).
% ?- fn(-([a-1, b-2]), R).
% ?- fn(-([a-1, b-2, c-1]), R).
% ?- fn(-[a-1, b-2], V).
% ?- fn(-[a-1, b-2, c-1], V).

fn([], fn([])):-!.
fn(fn(A, B, C), fn(A, B, C)):-!.
fn(fn(A, B), fn(A, B)):-!.
fn(fn(A), fn(A)):-!.
fn([A|B], fn(D, R, [A|B])):-!,
	zip(D, R0, [A|B]),
	sort(R0, R).
fn(dom(fn(D,_,_)), D):-!.
fn(dom(fn(D,_)), D):-!.
fn(dom(fn(D)), D):-!.
fn(dom(X), A1):-!, fn(X, A2), fn(dom(A2), A1).
fn(ran(fn(_, R, _)), R):-!.
fn(ran(fn(_, G)), R):-!, fn_range(G, R).
fn(ran(fn(G)), R):-!, fn_range(G, R).
fn(ran(X), A1):-!, fn(X, A2), fn(ran(A2), A1).
fn(fld(fn(D, R, _)), F):-!, ord_union(D, R, F).
fn(fld(fn(_, G)), F):-!, fn_field(G, F).
fn(fld(fn(G)), F):-!, fn_field(G, F).
fn(fld(X), A1):-!, fn(X, A2), fn(fld(A2), A1).
fn(graph(fn(A1, A2, G)), G):-!.
fn(graph(fn(A1, G)), G):-!.
fn(graph(fn(G)), G):-!.
fn(graph(X), A1):-!,fn(X, A2), fn(graph(A2), A1).
fn(id(A), A1):-!, misc:set(A, A2), id_fn(A2, A1).
fn((F;G), A1):-!, fn(F, A2),
	fn(G, A3),
	fn_compose(A2, A3, A1).
fn(F+G, A1):-!,
	fn(F, A2),
	fn(G, A3),
	fn_sum(A2, A3, A1).
fn(F*G, A1):-!,
	fn(F, A2),
	fn(G, A3),
	fn_prod(A2, A3, A1).
fn(-fn(D, R, G), fn(R, D, A1)):-!,
	fn_to_pfn_flip(G, A1).
fn(-fn(_, G), fn(A2)):-!,
	fn_to_pfn_flip(G, A2).
fn(-fn(G), fn(A1)):-!,
	fn_to_pfn_flip(G, A1).
fn(-X, A1):-!,
	fn(X, A2),
	fn(-A2, A1).
fn(choice(F), A1):-!,
	misc:set(F, A2),
	maplist(pred([[A|B], [A|B]-A]), A2, A1).
fn(quote(X), X):-!.			% A bug !

% ?- rel(-[a-1, b-2], V).
% ?- rel(id(pow([a,b])), R).
rel(X, X):-is_list(X),!.
rel(id(S), A1):-!,
	misc:set(S, A2),
	maplist(pred([A, A-A]), A2, A1).
rel(dom(F), A1):-!,
	misc:set(F, A2),
	rel_domain(A2, A1).
rel(ran(F), A1):-!,
	misc:set(F, A2),
	rel_range(A2, A1).
rel(fld(F), A1):-!,
	misc:set(F, A2),
	rel_field(A2, A1).
rel((F;G), A1):-!,
	rel(F, A2),
	rel(G, A3),
	rel_compose(A2, A3, A1).
rel(flip(F), A1):-!,
	rel(F, A2),
	maplist(pred([A-B, B-A]), A2, A1).
rel(coa(F), A1):-!,
	rel(F, A2),
	sort(A2, A3),
	rel_to_pfn(A3, A1).
rel(-F, A1):-!,
	misc:set(F, A2),
	rel_to_pfn_flip(A2, A1).
%
ap(fn(_, F, _), X, Y):- call(F, X, Y).

%
id_fn(X, Y):- maplist(pred([A, A-A]), X, Y).

%
power_flip_map(cla(_, _, R), R, L):- section(R,  L).

section(X, Y):- foldr(pred(	[A-[X|_], L, [A-X|L]]
				&
				[_, L, L]
			  ),
		      X, [], Y).
%
ifmap_to_flip_inclusion(cla(X, A, _), Upper, ID):-
	maplist(pred([P, P-[P]]), A, Upper),
	maplist(pred([Q, Q-Q]), X, ID).

%
dom_of(F, D)	:- zip(D0, _, F), sort(D0, D).
ran_of(F, R)	:- zip(_, R0, F), sort(R0, R).
fld_of(F, Field):- zip(A, B, F),
	sort(B, B0),
	ord_union(A, B0, Field).

%
rel_domain(rel(G), D) :- dom_of(G, D).
rel_domain(rel(D, _), D).
rel_domain(rel(D, _, _), D).
rel_domain(X, Y):- dom_of(X, Y).
%
rel_range(rel(G), R) :- ran_of(G, R).
rel_range(rel(_, G), R) :- ran_of(G, R).
rel_range(rel(_, R, _), R).
rel_range(G, R):- ran_of(G, R).
%
rel_field(rel(G), F):-  fld_of(G, F).
rel_field(rel(_, G), F):-  fld_of(G, F).
rel_field(rel(_, _, G), F):-  fld_of(G, F).
rel_field(G, F):-  fld_of(G, F).
%
rel_graph(rel(G), G).
rel_graph(rel(_, G), G).
rel_graph(rel(_, _, G), G).
rel_graph(G, G).

%
fn_domain(fn(G), D) :- dom_of(G, D).
fn_domain(fn(D, _), D).
fn_domain(fn(D, _, _), D).
fn_domain(X, Y):- dom_of(X, Y).
%
fn_range(fn(G), R) :- ran_of(G, R).
fn_range(fn(_, G), R) :- ran_of(G, R).
fn_range(fn(_, R, _), R).
fn_range(G, R):- ran_of(G, R).
%
fn_field(fn(G), F)	:-  fld_of(G, F).
fn_field(fn(_, G), F)	:-  fld_of(G, F).
fn_field(fn(_, _, G), F):-  fld_of(G, F).
fn_field(G, F)			:-  fld_of(G, F).
%
fn_graph(fn(G), G).
fn_graph(fn(_, G), G).
fn_graph(fn(_, _, G), G).
fn_graph(G, G).

%
zip([], [], []).
zip([A|A0], [B|B0], [A-B|R]):- !, zip(A0, B0, R).
zip([A|A0], [B|B0], [(A,B)|R]):- zip(A0, B0, R).

comma_zip([], [], []).
comma_zip([A|A0], [B|B0], [(A,B)|R]):-
	comma_zip(A0, B0, R).

% ?-  ifmap_to_flip_inclusion(cla([1],[a,b], []), U, L).
% ?-  fn(dom([a-b,c-d]), D).
% ?-  fn(ran([a-b,c-d]), D).
% ?-  fn(fld([a-b,c-d]), D).
% ?-  rel(dom([a-b,c-d]), R).
% ?-  fn(fld([a-b,c-d]), R).
% ?-  rel(-[a-1, b-1, c-2], R).
% ?-  rel_to_pfn_flip([a-1, b-1, c-2], R).
% ?-  coarsest_partition([a,b,c,d], [[a,b,c],[a,b,d]], P0),
%	maplist(sort, P0, P1),	sort(P1, P).

% The coarsest partition of a set by a family of sets.
coarsest_partition(X, S, P):- length(S, L),
	numlist(1, L, N),
	zip(S, N, Z),
	maplist(pred(Z, [A, I-A]:- indexes(A, Z, I)), X, Y),
	keysort(Y, Y0),
	index_merge(Y0, P).

% ?- indexes(a, [[a]-1,[a]-2], X).
indexes(_, [], []).
indexes(A, [S-J|R], [J|P]):- memberchk(A, S), !,
	indexes(A, R, P).
indexes(A, [_|R], P):- indexes(A, R, P).

% ?- index_merge([a-1,a-2,b-4, b-5], R).
index_merge([], []).
index_merge([A-M|X], Y):- index_merge(A, X, [M], Y).

%
index_merge(_, [], S,  [S]).
index_merge(A, [A-I|R], S, Y):- !,
	index_merge(A, R, [I|S], Y).
index_merge(_, [B-I|R], S, [S|Y]):-
	index_merge(B, R, [I], Y).

% ?-  power_type_cla([a,b], C).
% ?-  power_type_cla([a,b,c], C).
power_type_cla(T, cla(Ks, T, S)):-
	set(pow(T), PowT),
	length(PowT, N),
	numlist(1, N, Ks),
	zip(Ks, PowT, A),
	power_type_support(T, A, S).

power_type_support([], _, []).
power_type_support([P|P0], A, [P-Ks|R]):-
	power_type_extension(P, A, Ks),
	power_type_support(P0, A, R).

power_type_extension(_, [], []).
power_type_extension(P, [K-U|R], [K|Ks]):-
	memberchk(P, U), !,
	power_type_extension(P, R, Ks).
power_type_extension(P, [_|R], Ks):- power_type_extension(P, R, Ks).

% ?- power_cla([x-[a], y-[b]], [[x], [x,y]], R).
power_cla(S, P, S0):-
	maplist(
	pred(S, ([A, A-U] :- maplist(
		pred(S, ([B, U0] :- memberchk(B-U0, S); U0 =[])),
			   A, V),
		   append(V, U))),
		P, S0).

%%
% ?- fn_sum([a-c, d-e],[b-c, d-e], R).
% ?- fn_sum([a-c, b-c],[a-c, d-c], R).
fn_sum(F, G, H):-
	maplist(pred([X-Y, (0,X)-Y]), F, F0),
	maplist(pred([U-V, (1,U)-V]), G, G0),
	append(F0, G0, H).

% ?- fn_prod([2-a,1-b], [1-c, 2-d], R).
fn_prod(F, G, H):-
	sort(F, F0),
	sort(G, G0),
	maplist(pred([X-Y, _-Z, X-(Y-Z)]), F0, G0, H).

%%  fn_compose(+A:fn, +B:fn, -C:fn(D, R, F)) is det.
%   True if C is the composition of two function A and B.
% ?- fn_compose([a-1, b-2], [1-x, 2-y], R).
fn_compose(A, B, fn(D, R, F)):-
	fn_domain(A, D),
	fn_range(B, R),
	fn_graph(A, F1),
	fn_graph(B, F2),
	fn_compose(F1, F2, F, []).
%
fn_compose([X-Y|F], G, [X-Z|H], H0):-
	member(Y-Z, G),
	!,
	fn_compose(F, G, H, H0).
fn_compose([], _, H, H).

% ?- info_map_compose(im(1, 2, [a-b, c-d], [y-x, v-u]), im(2, 3, [b-e,d-f],[l-y, m-v]), R).
% ?- info_map_compose(im(1, 2, [a-b], [y-x]), im(2, 3, [b-e],[l-y]), R).

info_map_compose(im(A, B, F0, G0), im(B, C, F1, G1), im(A, C, F, G)):-
	fn_compose(F0, F1, F),
	fn_compose(G1, G0, G).

% ?- ds_cover([cla([x],[a], [a-[x]])], [], R).
% ?- ds_cover([cla([x,y],[a,b], []), cla([x,y], [a, b],[])], [], R).
% ?- ds_cover([cla([x,y],[a,b], []), cla([x,y], [a, b],[])], [im(1,2,[a-a,b-b],[x-x,y-y])], R).
%@ R = cla([[x, x], [y, y]], [[1:b, 2:b], [1:a, 2:a]], [[1:b, 2:b]-[], [1:a, 2:a]-[]]) .

ds_cover(ds(C, I), CLA):-!, ds_cover(C, I, CLA).
%
ds_cover(Cla_list, Imorph_list, cla(X, A, R)):-
	cla_list_sum(Cla_list, Sum),
	Sum = cla(X0, A0, R0),	% like open bases of the product topology.
    colimit_cover(Imorph_list, A0, A),
	limit_cover(Imorph_list, X0, X),
	merge_types(A, R0, R1),
	maplist(pred(X, [T-S, T-S0]:- intersection(S, X, S0)), R1, R).
%
zip_colon([], [], []).
zip_colon([X|Xs], [Y|Ys], [X:Y|Zs]):- zip_colon(Xs, Ys, Zs).

% ?- keyed_ds_cover([k1:cla([x],[a], [a-[x]])], [], R).
keyed_ds_cover(ds(X,Y), Z):-!, keyed_ds_cover(X, Y, Z).
%
keyed_ds_cover(Keyed_cla_list, Imorph_list, CLA):-
	zip_colon(Keys, Cla_list, Keyed_cla_list),
	length(Keys, Len),
	numlist(1, Len, Ns),
	maplist(pred([K, I, K-I]), Keys, Ns, KeyAssoc),
	maplist(pred(KeyAssoc, ( [im(K, I, U, L), im(K0, I0, U, L)]
			:-	memberchk(K-K0, KeyAssoc),
				 memberchk(I-I0, KeyAssoc))),
		   Imorph_list, Imorph_list0),
	ds_cover(Cla_list, Imorph_list0, Center_cla),
	back_to_keyed_cla(Center_cla, CLA, KeyAssoc).

%
back_to_keyed_cla(cla(X, A, R), cla(X, A0, R0), KeyAssoc):-
	maplist(
		pred(KeyAssoc,
			 [U, U0]:-
				 maplist(
					pred([U, U0, KeyAssoc],
						  [I:W, J:W]:- memberchk(J-I, KeyAssoc)
						),
					U, U0)
			),
		A, A0),
	maplist(
		pred(KeyAssoc,
			 [U-Z, V-Z]:-
				maplist(
					pred([U, V, KeyAssoc],
						 [W:Y, W0:Y] :- memberchk(W0-W, KeyAssoc)
						),
				U, V)
			),
		R, R0).


% ?- ifmap:index_cla_name([a,b], [1-x,2-y], R).
index_cla_name(A, Ms, Ns):-
	maplist(pred(A, [I-X,  um(K, X)]:-  nth1(I, A, K)),
		     Ms, Ns).

% ?- cover_imap([a-[1:x]], [], R).
% ?- cover_imap([a-[1:x,2:y]],[],R).
% ?- cover_imap([a-[1:x,2:y], b-[1:u, 2:v]], [], R).

cover_imap([], X, X).
cover_imap([N-A|B], X, Y):- cover_imap(A, N, X, X0),
	cover_imap(B, X0, Y).
%
cover_imap([], _, X, X).
cover_imap([J:A|R], N, X, Y):- select(J:S, X, X0), !,
	cover_imap(R, N, [J:[A:N|S]|X0], Y).
cover_imap([J:A|R], N, X, Y):-
	cover_imap(R, N, [J:[A:N]|X], Y).

% ?- keys_to_assoc([a, b, c], Z).
%@ Z = [a-1, b-2, c-3].
keys_to_assoc(K, Z):-
	foldr(pred([A, (I, U), (J, [A-J|U])] :- I is J + 1), K, (_, []), (1, Z)).

% ?- rename_types([[a, b]- x, [c,d]-y], Y, Z).
rename_types([], [], []):-!.
rename_types(X, Y, Z):-
	maplist(pred([A-_, A] & [A, A]), X, X0),
	length(X0, Len),
	numlist(1, Len, Ns),
	maplist(pred([N, A-B, N-A, N-B] & [N, A, N-A, N]), Ns, X, Y, Z).

% ?- inverse_partition_map([a-[1,2], b-[3,4]], Y).
inverse_partition_map(X, Y):-
	foldr(pred([A-S, U, V]
			:- foldr(pred(A, [P, L,[P-A|L]]), S, U, V)),
	      X, [], Y).

% ?- merge_types([[a,b]], [a-[1,2], b-[3,4]], R).
% ?- merge_types([[a,b, c]], [a-[1,2], b-[3,4]], R).
merge_types([], _, []).
merge_types([A|As], R, [A-M|Bs]):-
	merge_types(A, R, Xs, R0),
	append(Xs, M),
	merge_types(As, R0, Bs).

%
merge_types([], R, [], R).
merge_types([T|Ts], R, [X|Xs], Q):-
	select(T-X, R, R0),
	!,
	merge_types(Ts, R0, Xs, Q).
merge_types([_|Ts], R, Xs, Q):- merge_types(Ts, R, Xs, Q).

% ?- colimit_cover([], [1:a], L).
colimit_cover(Ms, A, Limit):-
	maplist(pred(
	([im(I, J, Ps, _), Qs]:- maplist(pred([I,J], [X-Y, (I:X)-(J:Y)]), Ps, Qs))
		    &
	([im(I, J, Ps), Qs]:- maplist(pred([I,J], [X-Y, (I:X)-(J:Y)]), Ps, Qs))),
		Ms, List_of_lists),
	append(List_of_lists, Pairs),
	maplist(pred([C, [C]]), A, A0),
	union_find(Pairs, A0, Limit).

%
limit_cover([], X, X).
limit_cover([im(I, J, _, Qs)|Ms], X, Y):-
	filter(J, I, Qs, X, X0),
	limit_cover(Ms, X0, Y).

%
filter(_, _, [], X, X).
filter(I, J, [A-B|R], X, Y):-
	filter(I, J, A, B, X, X0),
	filter(I, J, R, X0, Y).

filter(_, _, _, _, [], []).
filter(I, J, A, B, [V|X], Y):-  nth1(I, V, A), !,
	nth1(J, V, C),
	( B==C
	->	Y=[V|Y0],
		filter(I, J, A, B, X, Y0)
	;	filter(I, J, A, B, X, Y)
	).
filter(I, J, A, B, [V|X], [V|Y]):- filter(I, J, A, B, X, Y).

		/*********************************************************
		*     Parsing a distributed system in a suface syntax    *
		*********************************************************/

% ?- ds_parse_cover_with_constraint([a,b,c], Y).
ds_parse_cover_with_constraint(X, X).

%   ;;  >  ; >  \s\t
% ?- parse_comma_list("", Y).
% ?- parse_comma_list("a", Y).
% ?- parse_comma_list("a,  b\n", Y).
% ?- parse_comma_list("a", Y).
% ?- parse_comma_list("a,b", Y).
% ?- parse_comma_list("a,b\n", Y).
% ?- parse_comma_list("a,' ' ,c", Y).
% ?- parse_comma_list("a,b\n\n\n", Y).

parse_comma_list(codes, X, Y):-!, string_codes(Z, X),
	parse_comma_list(Z, Y).
parse_comma_list(string, X, Y):- parse_comma_list(X, Y).
%
parse_comma_list(X, Y):- term_string(T, X),
	(	T== end_of_file ->  Y = []
	;	comma_list(T, Y)
	).

% ?-  parse_classifications(";", Y).
% ?-  parse_classifications("a:b, c, d, k", Y).

parse_classifications(X, Y):- parse_comma_list(X, Z),
	maplist(sort_cla, Z, Y).
%
sort_cla(X:cla(Y, Z, U), X:cla(Y0, Z0, U0)) :-
	sort(Y, Y0),
	sort(Z, Z0),
	maplist(pred([P-Q, P-Q0]:- sort(Q, Q0)), U, U1),
	sort(U1, U0).
%
parse_informorphisms(X, Y):- parse_comma_list(X, Y).
%
parse_constraints(X, Y):- parse_comma_list(X, Y).

		/*********************************
		*     ds_cover_with_constraint   *
		*********************************/

%
codes_string(X, Y):- string_codes(Y, X).
%
string_term(X, Y):- term_string(Y, X).

% ?- maplist(parse_list(string), ["a,b", "x,y,z", "1,2,3"],  X).
% ?- maplist(parse_list, ["a,b", "\s\t\n", "1,2,3"],  X).
% ?- maplist(parse_list(codes), [`a,b`, `x,y,z`, `1,2,3`],  X).
parse_list(X, Y):- parse_list(codes, X, Y).
%
parse_list(string, X, Y):-!, string_term(X, Z), comma_list(Z, Y).
parse_list(codes, X, Y):- string_codes(Z, X),
						  string_term(Z, U),
						  comma_list(U, Y).

% ds_cover_with_constraint(_, "hello\nworld").

ds_cover_with_constraint([X, Y, Z], A):-
 	ds_cover_with_constraint(X, Y, Z, A).

%
ds_cover_with_constraint(Cla_list0, Imorph_list0, Con0, OutString):-
	ds_keys(Cla_list0, Keys, Cla_list),
	ds_prepare(Imorph_list0, Con0, Imorph_list, Con, Keys),
	ds_check(Cla_list, Imorph_list, Con),
	ds_cover_with_constraint(Cla_list, Imorph_list, Con, Tokens, Types),
	maplist(map_token_vector(Keys), Tokens, Tokens0),
	maplist(map_type(Keys), Types, Types0),
	term_string(Tokens0, U),
	term_string(Types0, V),
	atomics_to_string([U,"\n",V], OutString).

% ?- ifmap:ds_cover_with_constraint([[cla([],[a,b],[])], [], []], X, Y).
% ?- ifmap:ds_cover_with_constraint([[cla([],[a,b],[]),cla([], [x,y],[])], [], []], X, Y).
% ds_cover_with_constraint([Cla_list0, Imorph_list0, Con0], Types0, Tokens0):-
% 	ds_keys(Cla_list0, Keys, Cla_list),
% 	catch((
% 		ds_prepare([Imorph_list0, Con0], [Imorph_list, Con], Keys),
% 		ds_check(Cla_list, Imorph_list, Con),
% 		ds_cover_with_constraint(Cla_list, Imorph_list,
% 				       Con, Tokens, Types),
% 		maplist(map_token_vector(Keys), Tokens, Tokens0),
% 		maplist(map_type(Keys), Types0, Types)),
% 		Error,
% 		(	ds_error_atomics(Keys, Error, M),
% 			atomics_to_string(M, OutString),
% 			writeln(OutString),
% 			fail
% 		)).

% ?- indexed_cla_intern([a:cla([],[],[])], X, Y, Z, U).
%@ X = [a],
%@ Y = [cla([], [], [])],
%@ Z = v(cla([], [], [])),
%@ U = [a-1].
% ?- indexed_cla_intern([a:cla([],[],[]), b:cla([],[],[])], X, Y, Z, U).
%@ X = [a, b],
%@ Y = [cla([], [], []), cla([], [], [])],
%@ Z = v(cla([], [], []), cla([], [], [])),
%@ U = [a-1, b-2].
% ?- indexed_cla_intern([a:cla([],[],[]), a:cla([],[],[])], X, Y, Z, U).

indexed_cla_intern(Indexed_cla, Index_list, Cla_list, Vec_cla, Zip_index_num):-
	zip_colon(Index_list, Cla_list, Indexed_cla),
	Vec_cla =..[v|Cla_list],
	length(Indexed_cla, N),
	numlist(1, N, Ns),
	zip(Index_list, Ns,  Zip_index_num),
	functor(Vec_cla, _,  L),
	sort(Index_list, S),
	length(S, K),
	(	K \== Ns  -> throw("multiple index found.")
	;	true
	).

% ?- ifmap:ds_keys([a,b,c], R, V).
% ?- ifmap:ds_keys([a,b:c,c], R, V).
ds_keys(X, Y, Z):- ds_keys(X, Y, Z, 1).

%
ds_keys([A:X|R], [A|U], [X|V], I):- J is I + 1,
	ds_keys(R, U, V, J).
ds_keys([X|R], [I|U], [X|V], I):- J is I + 1,
	ds_keys(R, U, V, J).
ds_keys([], [], [], _).

% ?- ifmap:key_index_prefix([a,b], X, 1:c).
% ?- ifmap:key_index_prefix([a,b], b:c, X).
key_index_prefix(Keys, K:X, I:X):- var(K), !, integer(I),  nth1(I, Keys, K).
key_index_prefix(Keys, K:X, J:X):- key_index(Keys, K, 1, J).
key_index_prefix(_, X, X).

% ?- key_index([a,b,c], b, I).
key_index(Key_list, Key, Index):- key_index(Key_list, Key, 1, Index).
%
key_index([K|_], K,  I, I):-!.
key_index([_|R], K, I, J):- I0 is I + 1, key_index(R, K, I0, J).

% ?- ifmap:map_type([a,b], [a:x, b:y], R), ifmap:map_type([a,b], S, R).
map_type(Keys, Type, Type0):- maplist(key_index_prefix(Keys), Type, Type0).

% ?- map_token_vector([1,2], [a, b], X).
map_token_vector(Keys, Vec, Vec0):- maplist(pred([K, T, K:T]), Keys, Vec, Vec0).

% ?- ds_prepare([], [], X, Y, []).
% ?- ds_prepare([im(u->v,[a-b], [])], [[u:a] => [v:b]], X, Y, [u,v]).
ds_prepare(Im, Con, Im0, Con0, Keys):-
	check_duplicate_free(Keys),
	maplist(pred(Keys, ([im(A->B, U, L), im(I->J, U, L)]
			    :- key_index(Keys, A, I),
			       key_index(Keys, B, J))
			&  ([im(A->B, U), im(I->J, U)]
			    :- key_index(Keys, A, I),
			       key_index(Keys, B, J))
				),
		Im, Im0),
	maplist(pred(Keys, ([L => R, L0 => R0]:-
						maplist(pred(Keys,
								 [U:X, U0:X]:-key_index(Keys, U, U0)), L, L0),
							maplist(pred(Keys,
										 [U:X, U0:X]:-key_index(Keys, U, U0)), R, R0))),
		Con, Con0).

%
check_duplicate_free(X):- sort(X, X0),
	length(X, N),
	length(X0, N0),
	(	N == N0 -> true
	;	throw(multiple_index(A))
	).


		/*****************************
		*     Check ds & build core  *
		*****************************/

% ?- check_ds_build_core([], [], Core).
% ?- check_ds_build_core([cla([x],[a], [a-[x]])], [im(1->1, [a-a], [x-x])], Core).
% ?- check_ds_build_core([], [im(1->1, [a-a], [x-x])], Core).  % << throw
check_ds_build_core(Cs, Ims, Core):-
	ds_check(Cs, Ims),
	ds_core(Cs, Ims, Core).

% ?- ds_check([], []).
% ?- ds_check([cla([x],[a], [a-[x]])], [im(1->1, [a-a], [x-x])]).
ds_check(Cla_list, Imorph_list):-
     maplist(cla_check, Cla_list),
     maplist(im_check(Cla_list), Imorph_list).

%
% ds_check(Cla_list, Imorph_list, Con):-
%      maplist(cla_check, Cla_list),
%      maplist(im_check(Cla_list), Imorph_list),
%      maplist(con_check(Cla_list), Con).
%
cla_check(cla(X, A, S)):- forall( member(U-L, S),
				  ( memberchk(U, A),
				    subset(L, X) -> true
				  ;	throw('an unspecified token appears a supporter'(U-L))
				  )).

%
im_check(Cla_list, im(I->J, FA, FX)):-
    is_map(FA),
    is_map(FX),
    nth1(I, Cla_list, Ci),
    nth1(J, Cla_list, Cj),
    Ci=cla(Xi, Ai,_),
    Cj=cla(Xj, Aj,_),
    is_map(FA, Ai, Aj),
    is_map(FX, Xj, Xi),
    check_imorph(Ci, Cj, FA, FX),
    !.
im_check(Cla_list, im(I->J, FA)):-
    nth1(I, Cla_list, cla(_, Ti, _)),
    nth1(J, Cla_list, cla(_, Tj, _)),
    is_map(FA),
    is_map(FA, Ti, Tj),
    !.
im_check(_, IM):- throw(failure(im_check(IM))).

%
con_check(Cla_list, L=>R):-
	length(Cla_list, N),
	forall( ( member(A, L) ;  member(A, R)),
		( A = J:T,
		  integer(J),
		  1=<J,
		  J=<N,
		  nth1(J, Cla_list, cla(_, Typs, _)),
		  memberchk(T, Typs))),
	!.
con_check(_, Seq):- throw(failure(con_check(Seq))).


% Error Messages
ds_error_atomics(_, unknown(X),  X).
ds_error_atomics(_, cla_check(X),
		 ["Classification check fail: ", X0,  "\n[]"]):- term_string(X, X0).
ds_error_atomics(Keys, im_check(X),
		 ["Informorphism check fail: ", X0, "\n[]"]):-
	X =.. [IM,I,J|R],
	maplist(key_index_prefix(Keys), [I0, J0], [I, J]),
	X1 =..[IM,I0, J0|R],
	term_string(X1, X0).
ds_error_atomics(Keys, con_check(L-R),
		 ["Constraint check fail: ", X0, "\n[]"]):-
	maplist(key_index_prefix(Keys), L0, L),
	maplist(key_index_prefix(Keys), R0, R),
	term_string(L0-R0, X0).
ds_error_atomics(_, cla_not_with_key(X),
	["Missing Key: ", X0, "\n[]"])	:- term_string(X, X0).
ds_error_atomics(_, duplicate_key(X),
	["Duplicate Key: ", X0, "\n[]"]):- term_string(X, X0).
ds_error_atomics(_, X,  ["other error: ", X0, "\n[]"]):- term_string(X, X0).

% ?- ifmap:sum_by_constraint([[1:a,2:a]-[]],[cla([x],[a],[a-[x]]), cla([x],[a],[a-[x]])], Core).
% ?- ifmap:sum_by_constraint([[1:a]-[2:a], [1:a]-[2:a]],[cla([x],[a],[a-[x]]), cla([x],[a],[a-[x]])], Core).
% ?- ifmap:ds_cover_with_constraint([cla([x,y],[a],[a-[x,y]]),cla([x,y],[a],[a-[x,y]])], [im(1,2, [a-a],[x-x])], [[1:a,2:a]-[]], U, V).
% ?- ifmap:ds_cover_with_constraint([cla([x,y],[a],[a-[x,y]]),cla([x,y],[a],[a-[x,y]])], [im(1,2, [a-a],[x-x])], [[1:a]-[2:a]], U, V).
% ?- ifmap:ds_cover_with_constraint([cla([x,y],[a],[a-[x,y]]),cla([x,y],[a],[a-[x

% ?- ifmap:ds_cover_with_constraint([cla([x,y],[a,b],[b-[y], a-[x]]),cla([x,y],[a],[a-[x,y]])], [im(1,2, [a-a,b-a],[x-x])], [[1:a]-[2:a],[2:a]-[1:a]], U, V).
%@ U = [],
% ?- ifmap:ds_cover_with_constraint([cla([x,y],[a,b],[b-[y], a-[x]]),cla([x,y],[a],[a-[x,y]])], [im(1,2, [a-a,b-a],[x-x])], [[1:a]-[2:a],[2:a]-[1:a]], U, V).
ds_cover_with_constraint(Cla_list, Imorph_list, Con, Tokens, Types):-
	sum_by_constraint(Con, Cla_list, CoreTokens),
	maplist(pred([cla(_,A,_), A]), Cla_list, TypS),
	type_sum(TypS, 1, [], IndexedTypes),
    colimit_cover(Imorph_list, IndexedTypes, Types),
	limit_cover_by_invertability(Imorph_list, CoreTokens, Tokens).

% ?- ifmap:ds_theory_cover_with_constraint([theory([a],[[a]=>[a]])], [], [], X, Y).
% ?- ifmap:ds_theory_cover_with_constraint([theory([a],[[a]=>[a]]), theory([a],[[a]=>[a]])], [], [], X, Y).
% ?- ifmap:ds_theory_cover_with_constraint([theory([a],[[a]=>[a]]), theory([a],[[a]=>[a]])]), [im(1,2,[a-a])], [], X, Y).
% ?- ifmap:ds_theory_cover_with_constraint([theory([a],[a=>a]), theory([a],[a=>a])], [im(1,2,[a-a])], [[1:a]-[2:a]], X, Y).
ds_theory_cover_with_constraint(Theories, Imorph_list, Con, Tokens, Types):-
	theory_sum_tokens(Theories, CoreTokens),
	filter_with_constraints(CoreTokens, Con, CoreTokens0),
	maplist(pred([theory(A,_), A]), Theories, TypS),
	type_sum(TypS, 1, [], IndexedTypes),
    colimit_cover(Imorph_list, IndexedTypes, Types),
	limit_cover_by_invertability(Imorph_list, CoreTokens0, Tokens).

% ?- ifmap:check_info_map(cla([x],[a], [a-[x]]), cla([y],[b],[b-[y]]), [a-b],[y-x]).
limit_cover_by_invertability([im(I, J, Zip, _)|Ms], X, Y):-!,
    limit_cover_by_invertability(X, Zip, I, J, X0),
    limit_cover_by_invertability(Ms, X0, Y).
limit_cover_by_invertability([im(I, J, Zip)|Ms], X, Y):-!,
    limit_cover_by_invertability(X, Zip, I, J, X0),
    limit_cover_by_invertability(Ms, X0, Y).
limit_cover_by_invertability([], X, X).
%
limit_cover_by_invertability([V|R], Z, I, J, [V|R0]):-
    limit_cover_by_invertability(Z, V, I, J),
    !,
    limit_cover_by_invertability(R, Z, I, J, R0).
limit_cover_by_invertability([_|R], Z, I, J, R0):-!,
    limit_cover_by_invertability(R, Z, I, J, R0).
limit_cover_by_invertability([], _, _, _, []).

%
limit_cover_by_invertability(Zip, Vec, I, J):-
    nth1(I, Vec, Li-Ri),
    nth1(J, Vec, Lj-Rj),
    map_from_to(Li, Zip, Lj),
    map_from_to(Ri, Zip, Rj).

% ?- ifmap:map_from_to([a,b], [a-1,b-2], [1,2]).
map_from_to([X|Y], Z, T):- memberchk(X-U, Z),
			   memberchk(U, T),
			   map_from_to(Y, Z, T).
map_from_to([], _, _).


% ?- ifmap:check_info_map(cla([x],[a], [a-[x]]), cla([y],[b],[b-[y]]), [a-b],[y-x]).
% ?- ifmap:check_info_map(cla([x],[a,b], [a-[x]]), cla([y],[b],[b-[y]]), [a-b],[y-x]).
ds_check_info_map(ds(FCs, IM)):-
	maplist(pred(FCs, [M] :- check_info_map(M, FCs)), IM).

%
check_info_map(im(I, J, U, L), Cs):- !,
	memberchk(I-A, Cs),
	memberchk(J-B, Cs),
	check_info_map(A, B, U, L).
%
check_info_map(cla(_,A0, R0), cla(X1,_,R1), Upper_map, Lower_map):-
	forall((member(P, X1),
		member(T, A0),
		memberchk(P-Q, Lower_map),
		memberchk(T-S, Upper_map)),
	       iff(support(R0, Q, T), support(R1, P, S))).

% ?- set_equal([a,a,b], [b,a,b]).
set_equal(X, Y):- subset(X, Y), subset(Y, X).

% ?- check_informorphism( cla([x],[a], [a-[x]]), cla([x],[a],[a-[x]]),
% [a-a],[x-x]).
% ?- check_informorphism( cla([x],[a], [a-[y]]), cla([x],[a],[a-[x]]),
% [a-a],[x-x]).   % << false

check_informorphism(C, D, U, L):- check_imorph(C, D, U, L).

% check_informorphism(cla(X0, A0, R0), cla(X1, A1, R1), Upper_map, Lower_map):-
% 	zip(U, U0, Upper_map),
% 	set_equal(U, A0),
% 	subset(U0, A1),
% 	zip(V, V0, Lower_map),
% 	set_equal(V, X1),
% 	subset(V0, X0),
% 	forall( (member(P, X1), member(T, A0)),
% 		forall( (memberchk(P-Q, Lower_map), memberchk(T-S, Upper_map)),
% 				iff(support(R0, Q, T), support(R1, P, S)))).


% ?- check_imorph( cla([x],[a], [a-[x]]), cla([x],[a],[a-[x]]),
% [a-a],[x-x]).
% ?- check_imorph( cla([x],[a], [a-[y]]), cla([x],[a],[a-[x]]),
% [a-a],[x-x]).   % << false

% ?- time(repeat(10000, (check_imorph( cla([x],[a], [a-[x]]), cla([x],[a],[a-[x]]), [a-a],[x-x])))).
%@ % 220,001 inferences, 0.012 CPU in 0.012 seconds (98% CPU, 18585875 Lips)
%@ true.

check_imorph(cla(X0, A0, R0), cla(X1, A1, R1), Upper_map, Lower_map):-
	forall(	(	member(U-V, Upper_map),
				memberchk(U-Su, R0),
				memberchk(V-Sv, R1)),
			(	fun_inverse_image(Lower_map, Su, G),
				set_equal(G, Sv)
			)).

%
check_adjoint_upper(X, Y, Upper, Ms, Mt):-
	forall(member(P-Q, Upper),
	      iff(support(Ms, Y, P), support(Mt, X, Q))).

check_adjoint_lower(X, Y, Lower, Ms, Mt):-
	forall(member(P-Q, Lower),
	        iff(support(Ms, Q, X), support(Mt, P, Y))).

% ?- X=cla([1,2],[a,b], [a-[1,2],b-[2]]), Y = cla([3,4],[c,d],[c-[3,4],d-[3]]),  ifmap: build_info_map(X, Y, I).
% ?- X=cla([1,2],[a,b], [a-[1,2],b-[2]]), Y = cla([3,4],[c,d],[c-[3,4],d-[3]]),  setof(I, ifmap: build_info_map(X, Y, I), S), length(S, N).
% ?- X=cla([1,2],[a,b], [a-[1,2],b-[2]]), Y = cla([3,4],[c,d],[c-[3,4],d-[3]]),  ifmap: extend_info_map(X, Y, im([], []), IM).
% ?- X=cla([1,2],[a,b], [a-[1,2],b-[2]]), Y = cla([3,4],[c,d],[c-[3,4],d-[3]]),  ifmap: extend_info_map(X, Y, im([b-d], [4-2]), Extended).
% ?- ifmap: extend_info_map(cla([x,z],[a,b],[a-[x], b-[z]]), cla([y,z],[c,d],[c-[y], d-[z]]), im([a-b],[]), R).

extend_info_map(cla(X,A,S), cla(Y,B,T), im(U,L), im(U0,L0)):-
	zip(P, _, U),
	zip(Q, _, L),
	subtract(A, P, A0),
	subtract(Y, Q, Y0),
	build_adjoint(cla(X, A0, S), cla(Y0, B, T), U, L, U0, L0).

%
build_info_map(C,D,im(U, L)):- build_adjoint(C,D,[],[],U,L).

%
build_adjoint(cla(X, [A|As], R), cla([V|Vs], S, T),  U, L, U0, L0):- !,
	member(B, S),
	check_adjoint_upper(A, B, U, R, T),
	member(W, X),
	check_adjoint_lower(V, W, L, R, T),
        build_adjoint(cla(X, As, R), cla(Vs, S, T), [A-B|U], [V-W|L], U0, L0).
build_adjoint(cla(X, [A|As],R), cla([], S, T), U, L, U0, L0):-  !,
      	member(B, S),
	check_adjoint_upper(A, B, U, R, T),
        build_adjoint(cla(X, As, R), cla([], S, T), [A-B|U], L, U0, L0).
build_adjoint(cla(X,[],R), cla([V|Vs], S, T), U, L, U0, L0):- !,
      	member(W, X),
	check_adjoint_lower(V, W, L, R, T),
        build_adjoint(cla(X, [], R), cla(Vs, S, T), U, [V-W|L], U0, L0).
build_adjoint(cla(_,[],_), cla([], _,_),  U, L, U, L).

% ?- qcompile(util(ifmap)), module(ifmap).
% ?- ifmap:rel_compose([a-1, a-2], [2-x, 2-y], R).
rel_compose(F, G, H):- rel_compose(F, G, H, []).

rel_compose([], _, H, H).
rel_compose([X|R], G, H, K):- rel_compose_one(X, G, H, H0),
	rel_compose(R, G, H0, K).

%
rel_compose_one(_,[], H, H).
rel_compose_one(X-Y, [Y-Z|G], [X-Z|H], K):- !,
	rel_compose_one(X-Y, G, H, K).
rel_compose_one(U, [_|G], H, K):- rel_compose_one(U, G, H, K).

% ?- fun_inverse_image([a-1, b-2, c-2], [1], R).
% ?- fun_inverse_image([a-1, b-2, c-2], [1,2], R).
fun_inverse_image([], _, []).
fun_inverse_image([X-Y|R], A, [X|M]):-
	memberchk(Y, A), !,
	fun_inverse_image(R, A, M).
fun_inverse_image([_|R], A, M):-
	fun_inverse_image(R, A, M).

% ?- ifmap:image_of_map([a-b,b-a,c-a], R).
image_of_map(Fn, Image):- zip(_,Vals, Fn), sort(Vals, Image).

% ?- ifmap:find_map([a,b], [1,2,3], F).
find_map([A|As], R, [A-B|F]):- member(B, R), find_map(As, R, F).
find_map([], _, []).

%%	invertable_type_map(+M: map, +S: cla, +T: cla) is det.
%       True if there exists a token map f  such that the pair
%	(M, f) is an infomorphism from classifications S to T.

% ?- invertable_type_map([],cla([],[],[]), cla([],[],[]), R).
% ?- invertable_type_map([],cla([],[a],[]), cla([],[],[]), R).
% ?- invertable_type_map([a-s],cla([1],[a],[a-[1]]), cla([2],[s],[s-[2]])).
% ?- invertable_type_map([a-s,b-t],cla([1],[a,b],[a-[1],b-[1]]), cla([2,3],[s,t],[s-[2],t-[3]])).
% ?- invertable_type_map([a-s,b-t],cla([1],[a,b],[a-[1],b-[2]]), cla([3],[s,t],[s-[3],t-[3]])).
% ?- invertable_type_map([a-s],cla([1],[a],[a-[1]]), cla([2],[s],[s-[2]]),R).
% ?- invertable_type_map([a-s],cla([1],[a],[a-[1]]), cla([2],[s],[s-[2]]),R).

invertable_type_map(Map, SourceCla, TargetCla):-
	invertable_type_map(Map, SourceCla, TargetCla, _).

% ?- is_map([a-b],[a],[b,c]).
% ?- is_map([a-b],[a,d],[b,c]).
% ?- is_map([a-b],[a],[]).
% ?- is_map([],[],[]).
is_map(Map, D, R) :- zip(D0, R0, Map),
		     subset(D0, D),
		     subset(D, D0),
		     subset(R0, R).

%
is_map(Zip):- sort(Zip, Zip0),
	check_unique_key(Zip0).

% ?- check_unique_key([a-b,c-d]).
% ?- check_unique_key([a-b,a-d]).
check_unique_key([A-_, A-_|_]):- !,fail.
check_unique_key([_,X|M]):- check_unique_key([X|M]).
check_unique_key([_]).
check_unique_key([]).

%
invertable_type_map(Map, theory(X,Y), TargetCla, AtomMap):- !,
	theory_to_cla(theory(X, Y), Z),
	invertable_type_map(Map, Z, TargetCla, AtomMap).
invertable_type_map(Map, SourceCla, theory(X, Y), AtomMap):- !,
     	theory_to_cla(theory(X, Y), Z),
	invertable_type_map(Map, SourceCla, Z, AtomMap).
invertable_type_map(=, SourceCla, TargetCla, AtomMap):- !,
	typ(SourceCla, T),
	zip(T, T, Id_T),
	invertable_type_map(Id_T, SourceCla, TargetCla, AtomMap).
invertable_type_map(Map, SourceCla, TargetCla, AtomMap):-
	typ(SourceCla, U),
	typ(TargetCla, V),
	is_map(Map, U, V),
       	image_of_map(Map, IoM),
	cla_restrict(IoM, TargetCla, TargetCla0),
	cla_boolean_atoms(SourceCla, cla(_,_,A)),
	cla_boolean_atoms(TargetCla0, cla(_,_,B)),
	zip(AtomsA,_,A),
	zip(AtomsB,_,B),
	sort_atoms(AtomsA, AtomsA0),
	sort_atoms(AtomsB, AtomsB0),
	exists_atom_map(Map, AtomsA0, AtomsB0, AtomMap).

% ?- ifmap:invertable_token_map([1-3,2-4], cla([3,4],[a,b],[a-[3], b-[4]]), cla([1,2],[a,b],[a-[1], b-[2]]), R).
% ?- ifmap:invertable_token_map([1-3,2-3], cla([3,4],[a,b],[a-[3], b-[4]]), cla([1,2],[a,b],[a-[1], b-[2]]), R).

invertable_token_map(Map, cla(Xa, Aa, Sa), cla(Xb, _Ab, Sb), TypeMap):-
	is_map(Map, Xb, Xa),
	maplist(invert_token_map_one(Map, Sa, Sb), Aa, TypeMap).

invert_token_map_one(TokMap, Sa, Sb, A, A-B):-
	memberchk(A-U, Sa),
	collect_types_by_adjoint(Sb, TokMap, U, B, []).

invertable_token_map(M, C1, C2) :- invertable_token_map(M, C1, C2, _).

% ?- collect_types_by_adjoint([a-[1,2]],[1-3,2-4], [3,4], R, []).
collect_types_by_adjoint([], _, _, B, B).
collect_types_by_adjoint([A-V|S], TokMap, U, [A|R], R0):-
    member(K, V),
    memberchk(K-U0, TokMap),
    memberchk(U0, U),
    !,		% checking adjointness.
    forall(member(P, V), (memberchk(P-Q, TokMap), memberchk(Q, U))),
    collect_types_by_adjoint(S, TokMap, U, R, R0).
collect_types_by_adjoint([_|S], TokMap, U, R, R0):-
    collect_types_by_adjoint(S, TokMap, U, R, R0).

%%	sort_atoms(+X:atoms, -Y:atoms) is det.
%	Sort X as atoms set, and unify Y with the result.

% ?- sort_atoms([p([b,a],[2,1])], R).
sort_atoms(X, Y) :- maplist(pred(([p(U,V), p(U0,V0)]
			     :- sort(U, U0),
				sort(V, V0))),
			    X, Y).

%%	exists_atom_map(+M: map, +S:atoms, +T:atoms, +AtomMap) is det.
%	True if there exists an infomorphism (M, f) from S' to T'
%	such that S is  the quotient classification of S', and
%	T is so of T'. AtomMap is unified with the function from
%	T to S.

% ?- exists_atom_map([a-b], [p([a],[])], [p([b],[])], R).
% ?- exists_atom_map([a-x,b-x], [p([a,b],[])], [p([x],[])], R).
exists_atom_map(Map, SourceAtoms, TargetAtoms, AtomMap):-
    fn_to_pfn_flip(Map, SetDict),
    invertable_atom_set(TargetAtoms, SetDict, SourceAtoms, AtomMap).

% ?- invertable_atom_set([p([a],[])], [a-[b]], [p([b],[])], R).
invertable_atom_set([Q|As], Dict, Bs, [Q-P|AtomMap]):-
	Q = p(U, _),
	union_set_dict(U, Dict, U0),
	P = p(U0, _),
	memberchk(P, Bs),
	invertable_atom_set(As, Dict, Bs, AtomMap).
invertable_atom_set([], _, _, []).

% ?- ifmap:union_set_dict([a,b], [a-[y, u],b-[z, v]], R).
union_set_dict(A, Dict, Set) :- proj_dict(A, Dict, S),
				append(S, S0),
				sort(S0, Set).

% ?- proj_dict([a,b], [a-b,c-d], R).
% ?- proj_dict([a,b, x], [a-b,c-d], R).
% ?- proj_dict([c], [a-b,c-d], R).
proj_dict([A|R], Dict, [U|P]):- memberchk(A-U, Dict),
			     !,
			     proj_dict(R, Dict, P).
proj_dict([_|R], Dict, P) :- proj_dict(R, Dict, P).
proj_dict([], _, []).

% ?- subst_multiple([a-[1,2], b-[3,4]], f(g(a,b),a), R).
%@ R = [f(g(1, 3), 1), f(g(1, 3), 2), f(g(1, 4), 1), f(g(1, 4), 2), f(g(2, 3), 1), f(g(2, 3), 2), f(g(2, 4), 1), f(g(..., ...), 2)].
% ?- subst_multiple([a-[1,2], b-[3,4]], [a,b]->[b], R).
% ?- subst_multiple_massoc([[a1,a2]-[1,2], [b1,b2]-[3,4]], f(g(a1,b2),a2), R).
% ?- subst_multiple_massoc([[a1,a2]-[1,2], [b1,b2]-[3,4]], f(g(a1,b2),a2), R),
%	subst_multiple([a-[1,2], b-[3,4]], f(g(a,b),a), R0), R=R0.

subst_multiple(M, X, Y):- memberchk(X-Y, M), !.
subst_multiple(M, X, Y):- X=..[F|Xs],
	maplist(subst_multiple(M), Xs, Zs),
	choices(Zs, Us),
	maplist(pred(F, [A, B]:- B=..[F|A]), Us, Y).

subst_multiple_massoc(M, X, Y):- member(K-Y, M), memberchk(X, K), !.
subst_multiple_massoc(M, X, Y):- X=..[F|Xs],
	maplist(subst_multiple_massoc(M), Xs, Zs),
	choices(Zs, Us),
	maplist(pred(F, [A, B]:- B=..[F|A]), Us, Y).

% ?- subst_theory_multiple([[a]-[b,c]], [[a]=>[a]], R).
subst_theory_multiple(M, X, Y):-
	maplist(subst_sequent_multiple(M), X, X0),
	append(X0, Y).

subst_sequent_multiple(M, =>(L, R), S):-
	maplist(pred(M, [X, U]:- (member(V-U, M), memberchk(X, V))), L, L0),
	maplist(pred(M, [X, U]:- (member(V-U, M), memberchk(X, V))), R, R0),
	choices(L0, C),
	choices(R0, D),
	choices([C,D], E),
	maplist(pred([[X, Y], =>(X, Y)]), E, S).

% Moving theory forward / backward along a function.
%
% ?- move_theory_backward([x-a, y-a], [[a]=>[a], [a]=>[a]], R).
move_theory_forward(M, Theory, Theory0):-
	maplist(pred(M, ( [=>(L, R), =>(L0, R0)]
				:-
				maplist(   pred(M, ([A, B]:- memberchk(A-B, M))
							&
						    [A, A]),
					L, L0),
				maplist(   pred(M, ([A, B]:- memberchk(A-B, M))
							&
						    [A, A]),
					R, R0))),
		Theory, Theory1),
	simplify_theory(Theory1, Theory0).

%
move_theory_backward(M, X, Y):-
	fn_to_pfn_flip(M, M0),
	maplist(move_sequent_backward(M0), X, X0),
	append(X0, Y0),
	simplify_theory(Y0, Y).
%
move_sequent_backward(M, =>(L, R), S):-
	maplist(pred(M, [X, U]:- memberchk(X-U, M)), L, L0),
	maplist(pred(M, [X, U]:- memberchk(X-U, M)), R, R0),
	choices(L0, C),
	choices(R0, D),
	choices([C,D], E),
	maplist(pred([[X, Y], =>(X, Y)]), E, S).

% % Clustering; divide a  graph into  blocks.
union_find(X, Y):- union_find(X, [], Y).

% ?- union_find([(a,b),(x,y), (x,x), (y, z), (b,c)], R).
% ?- union_find([(a-b),(x-y), (x-x), (y-z), (b-c)], R).
% ?- union_find([(a=b),(x=y), (x=x), (y=z), (b=c)], R).

union_find([], X, X).
union_find([P|R], C, D):- (P = (X, Y); P = (X-Y); P = ( X = Y )), !,
	union_find(X, Y, C, C1),
	union_find(R, C1, D).
%
union_find(X, Y, Z, U):-select_cluster(X, Z, C, Z0),
	(	memberchk(Y, C) -> U=[C|Z0]
	;	select_cluster(Y, Z0, C0, Z1),
		append(C,C0, C1),
		U=[C1|Z1]
	).

% ?- select_cluster(a, [[a,b],[c,d]], C, X).

select_cluster(X, [], [X], []):-!.
select_cluster(X, [Y|Z], Y, Z):- memberchk(X, Y), !.
select_cluster(X, [Y|Z], U, [Y|V]):- select_cluster(X, Z, U, V).


% ?- theory_to_cla(theory([a,b], [[b]=>[a], [a]=>[b]]), M).
% ?- theory_to_cla(theory([a,b], [[a,b]=>[]]), M).
% ?- theory_to_cla(theory([a,b], [[a,b]=>[],[a]=>[b]]), M).
% ?- theory_to_cla(theory([bird, penguine, fly], [[bird]=>[fly], [penguine]=>[bird], []=>[penguin], [fly, penguine]=>[]]), C).
% ?- theory_to_cla(theory([bird, fly, penguine], [[penguine]=>[bird], [fly, penguine]=>[]]), C2), C1 = cla([a, b], [bird, fly], [bird-[a], fly-[a,b]]), ifmap:build_info_map(C2, C1, R).
% ?- theory_to_cla(theory([a], [[a]=>[], []=>[a]]), C).
% ?- theory_to_cla(theory([a,b,c], [[]=>[a], []=>[b], []=>[c]]), C).
% ?- theory_to_cla(theory([bird, fly, penguine], [[penguine]=>[bird], [fly, penguine]=>[]]), C2).
% ?- theory_to_cla(theory([bird, fly, penguine], [[penguine]=>[bird], [fly, penguine]=>[], [bird]=>[fly]]), C2).

% ?- theory_to_cla(
%	theory([bird, fly, penguine],
% 		  [[penguine]=>[bird], [fly, penguine]=>[]]),
%		  C2),
%   C1 = cla([a, b], [bird, fly], [bird-[a], fly-[a,b]]),
%   build_info_map(C1, C2, im(U, L)),
%	ds_cover([C1, C2], [im(1, 2, U, L)], P).

% ?- theory_to_cla(
%	theory([bird, fly, penguine],
% 		  [[penguine]=>[bird], [fly, penguine]=>[]]),
%		  C2),
%   C1 = cla([a, b], [bird, fly], [bird-[a], fly-[a,b]]),
%   build_info_map(C1, C2, im(U, L)),
%	keyed_ds_cover([cla1:C1, cla2:C2], [im(cla1, cla2, U, L)], P).

theory_to_cla(theory(S, SeqL), cla(T, S, Supp)):- set(pow(S), P), !,
	collect_token(P, SeqL, M),
	fresh_token(M, T, Supp).

% ?- theory_sum_tokens([theory([a], [[a]=>[], []=>[a]])], C).
% ?- theory_sum_tokens([theory([a], [[a]=>[a]]),
%		theory([a], [[a]=>[a]])], C).
% ?- theory_sum_tokens([theory([a], [[a]=>[a]]),
%		theory([a], [[a]=>[a]]), theory([a], [[a]=>[a]])], C).

theory_sum_tokens(Theories, Tokens):-
    maplist(theory_to_seqs, Theories, ListofTokenSets),
    choices(ListofTokenSets, Tokens).

% ?- ifmap:theory_to_seqs(theory([bird, fly, penguine], [[penguine]=>[bird], [fly, penguine]=>[], [bird]=>[fly]]), C2).
theory_to_seqs(theory(S, Con), Seqs):-
	set(pow(S), P), !,
	collect_token(P, Con, Atoms),
	maplist(atom_to_seq(S), Atoms,  Seqs).

%
atom_to_seq(S, Atom0, Atom-Comp):-  sort(Atom0, Atom),
	subtract(S, Atom, Comp0),
	sort(Comp0, Comp).

%
collect_token([], _, []).
collect_token([P|R], L, [P|S]):- satisfy_all(P, L), !,
	collect_token(R, L, S).
collect_token([_|R], L, S):- collect_token(R, L, S).

% ?- satisfy([a,b], [a]=>[b]).
satisfy(X, =>(Y, Z)):- subset(Y, X), !,
	member(A, Z),
	member(A, X).
satisfy(_, _).
%
satisfy_all(X, L):- maplist(satisfy(X), L).

%
fresh_token(P, X, S):- length(P, N),
	( N == 0
	->	X=[], S=[]
	;	numlist(1, N, X),
		zip(X, P, Z),
		flip_pfn(Z, S)
	).

% ?- simplify_theory([[a,b]=>[c,d], [a]=>[d], [b]=>[c]], R).
simplify_theory(Sequents, Sequents0):-
	maplist(pred([=>(L,R), =>(L0, R0)]
		:- (sort(L, L0), sort(R, R0))),
		Sequents, Sequents1),
	sort(Sequents1, Sequents2),
	simplify_theory(Sequents2, [], Sequents3),
	reverse(Sequents3, Sequents0).

%
simplify_theory([], X, X).
simplify_theory([Q|S], X, Y):- weakened_identity(Q), !,
	simplify_theory(S, X, Y).
simplify_theory([Q|S], X, Y):- member(P, X), sequent_stronger(P, Q), !,
	simplify_theory(S, X, Y).
simplify_theory([Q|S], X, Y):- sequent_remove_weaker(Q, X, X0),
	simplify_theory(S, [Q|X0], Y).

%
sequent_remove_weaker(_, [], []).
sequent_remove_weaker(Q, [P|X], Y) :- sequent_stronger(Q, P), !,
	sequent_remove_weaker(Q, X, Y).
sequent_remove_weaker(Q, [P|X], [P|Y]) :- sequent_remove_weaker(Q, X, Y).

%
sequent_stronger(=>(L, R), =>(L0, R0)):- subset(L, L0), subset(R, R0).

%
weakened_identity(=>(L, R)):- member(A, L), member(A, R).

% ?- ifmap:cla_boolean_atoms([a],[x], [x-[]], A,B,C).
% ?- cla_boolean_atoms([a,b],[x], [x-[a]], A,B,C).
% ?- cla_boolean_atoms([a,b], [x,y], [x-[a],y-[b]], A,B,C).
% ?- cla_boolean_atoms([a,b], [x,y], [], A,B,C).
% ?- ifmap:cla_boolean_atoms([a,b], [x,y], [x-[a],y-[b]], A,B,C).

cla_boolean_atoms(cla(Tok, Typ, Sup), cla(Tok0, Typ0, Sup0)):-
	cla_boolean_atoms(Tok, Typ, Sup, Tok0, Typ0, Sup0).

%
cla_boolean_atoms(Tok, Typ, Sup, Tok0, Typ0, Sup0):-
	cla_boolean_atoms(Typ, Sup, [p([],[])-Tok], Sup0),
	maplist(pred([A, B, A-B]), Typ1, Tok1, Sup0),
	sort(Typ1, Typ0),
	sort(Tok1, Tok0).

%
cla_complete_boolean_atoms(cla(Tok, Typ, Sup),
			   cla(Tok, Typ0, Sup0)):-
	complete_support(Typ, Sup, Sup1),
	refine_boolean_atoms(Tok, Sup1, Typ0, Sup0).

%
cla_boolean_atoms([A|As], Sup, R0, R):-
	( memberchk(A-X, Sup);  X = []),
	!,
	cla_boolean_atoms_(R0, A, X, [], R1),
	cla_boolean_atoms(As, Sup, R1, R).
cla_boolean_atoms([], _, R, R).

%
cla_boolean_atoms_([p(L,R)-X|P], A, Y, U0, U1):-
	intersection(X, Y, X0),
	subtract(X, Y, X1),
	add_non_empty([p([A|L], R)-X0, p(L,[A|R])-X1], U0, U2),
	cla_boolean_atoms_(P, A, Y, U2, U1).
cla_boolean_atoms_([], _, _, R, R).
%

% ?- ifmap:add_non_empty([a-[], b-[c]], R, R0).
add_non_empty([_-[]|A], R, R0) :- !, add_non_empty(A, R, R0).
add_non_empty([X|A], R, R0):- !, add_non_empty(A, [X|R], R0).
add_non_empty([], R, R).

% ?- ifmap:cla_restrict([b], cla(_,_,[a-[1],b-[2]]), R).
% ?- ifmap:(numcla(1,3, A), numcla(1, 3, B), cla(A+B, C), cla_restrict([1:1, 2:1], C, D)).
cla_restrict(S0, cla(_, _, P), cla(X0, S0, P0)):-
	collect(pred(S0, [A-_]:- memberchk(A, S0)), P, P0),
	zip(_, Xs, P0),
	union(Xs, X0).

% ?- ifmap:rename_literals([a,a,~(a)], [a-b], R).
% ?- ifmap:rename_literals([a,a,~(a), b, ~(c)], [a-b], R).

is_super(X-_, S):- !, subset(S, X).
is_super(X, S):- subset(S, X).

% ?- ifmap:inverse_type_map([a-1, b-2], cla(_, _, [[a]-a,[b]-b]), cla(_,_, [[1]-1, [2]-2]), G).
% ?- ifmap:inverse_type_map([a-1, b-2], [[a],[b]], [[1], [2]], G).
inverse_type_map(F, cla(_, _, Sup), Trgt, G):- !,
	zip(Src,_,Sup),
	inverse_type_map(F, Src, Trgt, G).
inverse_type_map(F, Src, cla(_, _, Sup), G):- !,
	zip(Trgt, _,Sup),
	inverse_type_map(F, Src, Trgt, G).
inverse_type_map(F, Src, Trgt, G):-
	maplist(pred(F, [X, X-Y]:-
			 rename_literals_with_complementary_rule(X, F, Y)),
		Src, Zip),
	maplist(pred(Trgt, ( [A-B, A-C]:-
			 collect_supersets(Trgt, B, C))),
		Zip, Zip0),
	flip_pfn(Zip0, G).


% ?-  ifmap:type_support_by_tok_vec(1:a, [[a]-[]]).
% ?-  ifmap:type_support_by_tok_vec(1:b, [[a]-[]]).
% ?-  ifmap:type_support_by_tok_vec(2:b, [[a]-[], [b]-[]]).
% ?-  ifmap:type_support_by_tok_vec(2:b, [[a]-[], [a, c]-[]]).

type_support_by_tok_vec(N:A, Svec):- type_support_by_tok_vec(N, A, Svec).

type_support_by_tok_vec(1, A, [L-_|_]):- !, memberchk(A, L).
type_support_by_tok_vec(2, A, [_,L-_|_]):- !, memberchk(A, L).
type_support_by_tok_vec(3, A, [_,_,L-_|_]):- !,  memberchk(A, L).
type_support_by_tok_vec(N, A, Svec):-  nth1(N, Svec, L-_),
				       memberchk(A, L).

%
types_support_by_tok_vec([A|As], V):-
    type_support_by_tok_vec(A, V),
    types_support_by_tok_vec(As, V).
types_support_by_tok_vec([], _).

% ?- ifmap:tok_vec_respect_seq([[a]-[b]], [1:a]-[1:b]).
% ?- ifmap:tok_vec_respect_seq([[a, b]-[1:c]], [1:a]-[1:b]).
% ?- ifmap:tok_vec_respect_seq([[a]-[b]], [1:b]-[1:b]).
% ?- ifmap:tok_vec_respect_seq([[a]-[b]], []-[1:a]).
% ?- ifmap:tok_vec_respect_seq([[a]-[b]], []-[1:c]).

tok_vec_respect_seq(Svec, L-R):- types_support_by_tok_vec(L, Svec), !,
	member(A, R),
	type_support_by_tok_vec(A, Svec),
	!.
tok_vec_respect_seq(_, _).

%
full_respect_by_tok_vec([A|Con], Svec):- !, tok_vec_respect_seq(Svec, A),
				      full_respect_by_tok_vec(Con, Svec).
full_respect_by_tok_vec([], _).

% ?- ifmap:filter_with_constraints([], [], R).
% ?- ifmap:filter_with_constraints([[[a]-[]]], [[1:a]-[1:a]], R).
% ?- ifmap:filter_with_constraints([[[a]-[]]], [[1:a]-[1:b]], R).

filter_with_constraints([Vec|Vecs], Con, [Vec|Vecs0]):-
    full_respect_by_tok_vec(Con, Vec),
    filter_with_constraints(Vecs, Con, Vecs0).
filter_with_constraints([_|Vecs], Con, Vecs0):-
    filter_with_constraints(Vecs, Con, Vecs0).
filter_with_constraints([], _, []).

% ?-ifmap:subst_key([a-1],[a:t]-[], R).
subst_key(Zip, L-R, L0-R0):-
	maplist(pred(Zip, [A:U, B:U] :- memberchk(A-B, Zip)), L, L0),
	maplist(pred(Zip, [A:U, B:U] :- memberchk(A-B, Zip)), R, R0).

% ?- ifmap:channel_by_constraint([[k:a]-[k:a]],[k-cla([x],[a],[a-[x]])], Core, KeyZip).
% ?- ifmap:channel_by_constraint([[m:a]-[k:a], [k:a]-[m:a]],[k-cla([x],[a],[a-[x]]), m-cla([x],[a],[a-[x]])], Core, KeyZip).
% ?- ifmap:channel_by_constraint([[k:a,m:a]-[]],[k-cla([x],[a],[a-[x]]), m-cla([x],[a],[a-[x]])], Core, KeyZip).

channel_by_constraint(Con, KeyedCs, Core, KeyZip):-
    zip(Keys, _Cs, KeyedCs),
    length(Keys, N),
    numlist(1, N, Ns),
    zip(Keys, Ns, KeyZip),
    maplist(subst_key(KeyZip), Con, Con0),
    sum_by_constraint(Con0, _ListOfSeqSet, Core).

% ?- ifmap:sum_by_constraint([[1:a,2:a]-[]],[cla([x],[a],[a-[x]]), cla([x],[a],[a-[x]])], Core).
% ?- ifmap:sum_by_constraint([[1:a]-[2:a], [1:a]-[2:a]],[cla([x],[a],[a-[x]]), cla([x],[a],[a-[x]])], Core).

sum_by_constraint(Con, Cs, Core):-
    maplist(convert_seq_set, Cs, ListOfSeqSet),
    misc:choices(ListOfSeqSet, SeqVecSet),
    filter_with_constraints(SeqVecSet, Con, Core).

% ?- ifmap:convert_seq_set(cla([x],[a],[a-[x]]), R).
% ?- ifmap:convert_seq_set(cla([x],[a],[a-[x],b-[x]]), R).

convert_seq_set(Cla, SeqSet):-
    ifmap:cla_complete_boolean_atoms(Cla, cla(_, Pairs, _)),
    maplist(pred([p(U,V), U-V]), Pairs, SeqSet).

%% 	refine_boolean_atoms(+Tok:tokens, +Sup:support, -Typ:types, -Sup0:support).
%
%	For given Sup = [T1 - X1, T2 - X2, ..., Tn - Xn] and
%	Tok, which is the union of X1, ..., Xn, then Typ and Sup0 are
%	unified as follows:
%
%		Typ  = [p(U1,V1), ..., p(Um, Vm)],  and
%
%		Sup0 = [p(U1,V1) - P1, ..., p(Um, Vm) - Pm]
%
%	such as follows: P1, ..., Pm are the atoms obtained by partitioning Tok
%	with given support Sup (cf. Channel theory), and the extension of p(Ui, Vi)
%	is Pi (i = 1, ..., m), 	where the notion of the extension is defined as follows.
%
%	The extension of Ti is Xi  (i=1,..,n).  The extension of p(U, V) is
%	the set of tokens in Tok which is a member of the extension
%	of any element of U, and on the other hand is not an elemenent
%	of the extension of W for any W in V.

% ?- ifmap:refine_boolean_atoms([a,b], [], A,B).
% ?- ifmap:refine_boolean_atoms([], [x], A,B).
% ?- ifmap:refine_boolean_atoms([a,b], [x-[a],y-[b]], A,B).
% ?- ifmap:refine_boolean_atoms([a], [x-[a]], A,B).
% ?- ifmap:refine_boolean_atoms([a,b], [x-[a],y-[b]], A,B).

refine_boolean_atoms([], _, [], []):- !.
refine_boolean_atoms(Tok, Sup, Typ0, Sup0):-
	refine_boolean_atoms(Sup, [p([],[])-Tok], Sup0),
	zip(Typ0, _, Sup0).

%
refine_boolean_atoms([A-X|As], R0, R):-
	refine_boolean_atoms_update(R0, A, X, R1),
	refine_boolean_atoms(As, R1, R).
refine_boolean_atoms([], R, R).

%
refine_boolean_atoms_update([p(U,V)-X|R], A, Y, S):-
	intersection(X, Y, X1),
	subtract(X, Y, X2),
	(	X1 ==[]
	->	S = S1
	;	S = [p([A|U], V)-X1|S1]
	),
	(	X2 == []
	->	S1 = S2
	;	S1 = [p(U,[A|V])-X2| S2]
	),
	refine_boolean_atoms_update(R, A, Y, S2).
refine_boolean_atoms_update([], _, _, []).

% ?- ifmap: (numcla(1,3, A), numcla(1, 3, B), cla(A+B, C), cla_restrict(C, [1:1, 2:1], D)).

%% eval_boolean_type(+C:cla, +E:boolen_exp, -V:list) det
%	is True if
%  V is unified with the list of tokens of C which satisfies E.

% ?- ifmap:eval_boolean_type(a, cla(_,_, [a-[1]]), R).
% ?- ifmap:eval_boolean_type(a+a & a, cla(_,_, [a-[1]]), R).
% An example use of flip options.
% eval_boolean_type(A&B,A1,A2):-!,(eval_boolean_type(A,A1,A3),eval_boolean_type(B,A1,A4)),misc:set(A3&A4,A2) .
% eval_boolean_type(A+B,A1,A2):-!,(eval_boolean_type(A,A1,A3),eval_boolean_type(B,A1,A4)),misc:set(A3+A4,A2) .
% eval_boolean_type(A,cla(A1,A2,Sup),X):-!,memberchk(A-X,Sup) .

%
rename_literals_with_complementary_rule(X, F, Y):-
	rename_literals(X, F, Y0),
	complementary_rule(Y0, Y1),
	sort(Y1, Y).

% ?- ifmap: rename_literals([a,a,~(a)], [a-b], R).
% ?- ifmap: rename_literals([a,a,~(a), b, ~(c)], [a-b], R).
rename_literals([X|R], F, [Y|R0]):-
	(  X = ~(X0), Y = ~(V) ;  X0 = X, Y = V	),
	(  memberchk(X0-V, F)  ;  V = X0 ),
	rename_literals(R, F, R0).
rename_literals([], _, []).

%
complementary_rule(Y, []):- member(~(X), Y), member(X, Y), !.
complementary_rule(Y, Y).

% ?- ifmap:collect_supersets([[a,b,c]], [b,a], X).
% ?- ifmap:collect_supersets([[a,b,c]-[], [b]-c], [b,a], X).
collect_supersets([X|Y], S,  [X|Y0]):- is_super(X, S), !,
	collect_supersets(Y, S, Y0).
collect_supersets([_|Y], S,  Y0):- collect_supersets(Y, S, Y0).
collect_supersets([], _,  []).

% ?- ifmap:complete_support([a,b], [c-[]], C).
%@ C = [b-[], a-[], c-[]].
complete_support([A|R], S, S0) :-
	( memberchk(A-_, S)
	->	S1 = S
	;	S1 = [A-[]|S]
	),
	complete_support(R, S1, S0).
complete_support([], X, X).
%

		/**********************
		*     view channel    *
		**********************/

% ?- qcompile(util(ifmap)), module(ifmap).
% ?- show_cla(subgraph(sub1, label="sub1";a ->b); subgraph(sub2, label="sub2"; sub1->sub2; [c(d)]); [a(c), d(b)]).
% ?- show_graph(pred([A-B, (A, f, B)]), [a-b,b-d]).
% ?- show_graph((=), [(a,f,b),(b, g, c)]).

:- meta_predicate show_graph(2,?).
show_graph(F, DG):- maplist(F,  DG, Arrows),
 	coalgebra:autviz(ifmap:digraph_viz, Arrows).
%
digraph_viz(Arrows) :-
     format("digraph g {~n",[]),
     format("rankdir=LR;~n",[]),
     coalgebra:move_in_dot(Arrows),
     format("}~n",[]).

% The codes in the rest is to be checked for working as before.
% ?- show_cla(subgraph(sub1, label="sub1";a ->b); subgraph(sub2, label="sub2"; sub1->sub2; [c(d)]); [a(c), d(b)]).
% ?- show_cla(cla([ag,b],[1,2], [1-[a], 2-[b]])).
% ?- show_cla(cla([a,b],[1,2], [1-[a,b], 2-[b]])).
% ?- view_cla(cla([a,b],[1,2], [1-[a], 2-[b]]), [out(pdf)]).

% ?- ifmap_config_basic(X).
ifmap_counter_config(
	[	dir_name(ifmap),
		stem(ifmap),
		counter_name(ifmap_counter)
	]).

% ?- cd('/Users/cantor/Desktop').
% ?- trace, update_counter_file(testfile, Count, 2), integer(Count).
% Assuming the counter file is at current working directory.
update_counter_file(FileName, Count):-
	update_counter_file(FileName, Count, 99).
%
update_counter_file(FileName, Count, Max):-
	(	exists_file(FileName) ->
		open(FileName, read, S),
		read(S, U),
		close(S),
		V is U + 1
	;	V  = 0
	),
	(	V > Max -> Count =  0
	;	Count = V
	),
	open(FileName, write, W),
	write(W, Count),
	write(W, ".\n"),
	close(W).

read_counter_file(Counter, Count):-
	open(Counter, read, S),
	read(S, Count),
	close(S).


% ?- completing_counter_config([],X).
% ?- ifmap_counter_config(C), completing_counter_config(C, C0).
%
completing_counter_config(X, Y):-
	check_directory_prefix(X, X0),
	memberchk(directory(D), X0),
	working_directory(Old, D),
	check_counter_directory_name(X0, X1),
	memberchk(dir_name(E), X1),
	cd(E),
	check_counter_name(X1, Y),
	cd(Old).
%
check_directory_prefix(X,  Y):-
	(	memberchk(directory(_), X) -> Y = X
	;	getenv(home_html_root, D),
		Y = [directory(D)|X]
	).
%
check_counter_directory_name(X,  Y):-
	(	memberchk(dir_name(N), X) -> Y = X
	;	N = anonymous,
		Y = [dir_name(N)|X]
	),
	pshell(mkdir(-p, N)).
%
check_counter_name(X, Y):-
	(	memberchk(counter_name(_), X) -> Y = X
	;	Y = [counter_name(counter)|X]
	).

% ?- theory_to_cla(
%  theory(	[tweety, penguin, bird, fly],
% [	[penguin]	=>	[bird],
% [bird]		=>	[fly]
%  ]	), CLA).
% ?- theory_to_cla_html(theory([a],[[]=>[a]]), pdf, A),
%	pshell(open("http://localhost"+A)).
% ?- theory_to_cla_html(theory([a,b],[[b]=>[a], [a]=>[b]]), pdf, A),
%	pshell(open("http://localhost"+A)).
% ?- theory_to_cla_html(theory([a],[[]=>[a]]), pdf, A).
theory_to_cla_html(X, T, URL):-
    theory_to_cla(X, Cla),
	ifmap_counter_config(Conf),
	completing_counter_config([out(T)|Conf], Conf0),
	obj_get([counter_name(CN), stem(Stem), directory(D), dir_name(M)],Conf0),
	working_directory(Old, D),
	cd(M),
	update_counter_file(CN, Count),
	atomic_list_concat([Stem, Count, '.html'], Targethtml),
    create_cla_html(Cla, Conf0, Targethtml),
	getenv(host_html_root, H),
	cd(Old),
	atomic_list_concat([H, / , M , / ,Targethtml], URL).

% ?- obj([a(2), b(3)], [a(2), b(3)], Y).
% ?- obj_get([b(X), a(Y), c(5)], [a(2), b(3)], Y).
% ?- obj_get([b(X), a(Y)], [a(2), b(3)], _).
% ?- obj_get([b(X), a(Y)], [a(2), b(3)]).

% For trace debuggin theory_to_cla_html in Prolog  without brouser.
% ?- Text= [command=`theory([bird, fly], [[bird] =>	[fly]])`, radioButton=`pdf`],
%  web:run_cgi( `(peek([@command, @radioButton]), maplist(herbrand_in_context),(eh:apply3(ifmap:theory_to_cla_html)), (basic:smash), through)`,
%	Text).

% ?- ifmap_counter_config(Opts).
% ?- ifmap_counter_config(Opts),
%	create_cla_html(cla([a,b],[1,2], [1-[a], 2-[b], 2-[a], 1-[b]]), Opts, A),
%	pshell(open("http://localhost" + A)).

create_cla_html(Cla, Conf, Target):-
	cla_to_new_dot(Cla, Conf, Conf0),
    divtag(Conf0, Dvitag),
	HTML = [	"<html><body>\n",
				Dvitag,
				"</body></html>\n"
	       ],
	file(Target, write, smash(HTML)).

%
cla_to_new_dot(cla(L, M, N)) --> !,
	{ A = subgraph(cluster_token, L)	},
	{ B = subgraph(cluster_type, M)		},
	graph_to_new_dot(A; B; [L,M,N]).
cla_to_new_dot(G) --> graph_to_new_dot(G).

%
graph_to_new_dot(G, Conf, Conf0):-
	obj_get([counter_name(CN), stem(ST)], Conf),
	read_counter_file(CN, Count),
	atom_concat(ST, Count, Base),
    graph_to_dot_write(G, [base(Base)|Conf], Conf0).
%
graph_to_dot_write(G, Opts, Opts) :- % for DCG use.
	once(term_to_graph(G, H, E, N)),
	once(graph_to_dot(digraph(g, H, E, N), Digraph)),
	obj_get([base(B), out(T)], Opts),
	atomic_list_concat([ B, (.), dot], Dot),
	atomic_list_concat([ B, (.), T], Tout),
	smash(Digraph, DigraphStr),
	file(Dot, write, write(DigraphStr)),
	sh(dot(-'T'(T), -o(Tout), Dot)).
%
divtag(Opts, DivTag) :-
   	obj_get([base(Base), out(T)], Opts),
	atomics_to_string([Base, ".", T], Loc),
	DivTag = [
		"<p><div ",
		"id='diagram' ",
%		"style='border : solid 2px #ff0000; ",
		"style='border : solid 1px red; ",
			"width : 1200px; ",			% was 600px
			"height : 700px; ",			% was 300px
			"overflow : auto; '><br/>\n",
		"<img src=\"",   Loc, "\"/>\n",
		"</div></p>\n"
	 ].

%!	show_cla(+Cla) is det.
%	View classification Cla in graphics using dot.

% ?- show_cla(cla([a,b],[1,2], [1-[a], 2-[b]])).
% ?- show_cla(cla([a,b],[1,2], [1-[a,b], 2-[b]])).
% ?- show_cla(cla([1],[a], [a-[1]])).
% ?- show_cla(cla([1,2],[a], [a-[1,2]])).

show_cla(Cla):- cla_to_dot(Cla, DotStr),
	T = pdf,
	tmp_file_stream(DOT, F0, [encoding(utf8), extension(dot)]),
	close(F0),
	tmp_file_stream(PDF, F1, [encoding(utf8), extension(pdf)]),
	close(F1),
	file(DOT, write, write(DotStr)),
	pshell(dot(-'T'(T), -o(PDF), DOT); open(PDF)).

% old_show_cla(Cla):- cla_to_dot(Cla, DotStr),
% 	T = pdf,
% 	expand_file_name("~/tmp/TEMP.dot", [DOT]),
% 	expand_file_name("~/tmp/TEMP.pdf", [PDF]),
% 	file(DOT, write, write(DotStr)),
% 	pshell(dot(-'T'(T), -o(PDF), DOT); open(PDF)).

cla_to_dot(cla(L, M, N), Dot):-
	A = subgraph(cluster_token, L),
	B = subgraph(cluster_type, M),
	graph_to_dot_string(A; B; [L,M,N], Dot).
%
graph_to_dot_string(G, Dot):-
	once(term_to_graph(G, H, E, N)),
	once(graph_to_dot(digraph(g, H, E, N), Digraph)),
	smash(Digraph, Dot).
