:-module(lambda,[pac_calc/2, pac_lisp/2]).
:- use_module(library(apply)).
:- use_module(library(gensym)).
:- use_module(pac(op)).

% pac_calc/2
% ?- pac_calc(X\X, V).
% ?- pac_calc((X\X)@1, V).
% ?- pac_calc((X\X)@(Y\Y), V).
% ?- pac_calc((X\X)@(Y\Y)@(Z\Z), V).
% ?- pac_calc((X\X)@(Y\Y)@(Z\Z)@hello, V).
% ?- pac_calc(X^((X\X)@1), V).
% ?- pac_calc((X\X)@(X\X)@hello, V).
% ?- pac_calc((X\X)@(X\X), V).

pac_calc(X, Y):- pac_calc(X, [], Y, _).

pac_calc(X, F, Y, G):- pac_calc_one(X, F, X0, F0), !,
	pac_calc(X0, F0, Y, G).
pac_calc(X, F, X, F).

pac_calc_one(X, F, Y, G):-
	subtree([(@), (\)/2, (^)/2], X, Y, S, T),
	clause(pac_calc_rule(S0, F, T, G), Body),
	subsumes_term(S0, S),
	S=S0,
	call(Body).

%%
pac_calc_rule(F^M, F0, M, G):- term_variables((F, F0), G).
pac_calc_rule((P\M)@A, G, N, G):- copy_term((G,P,M), (G,A,N)).


%% lisp in copy_term/2

% ?- pac_lisp([append, [a], []], V).
% ?- pac_lisp([append, [a,b], [c,d]], V).
% ?- pac_lisp([not, t], R).
% ?- pac_lisp([not, []], R).

pac_lisp(E, V):- pac_lisp(E, V, []).

pac_lisp([], [], _):-!.
pac_lisp([quote, A|_], A, _):-!.
pac_lisp(A^E, V, B):- !, term_variables((A, B), C), pac_lisp(E, V, C).
pac_lisp([(Xs\E)|L], V, A):- !, pac_lisp_(L, L0, A),
	copy_term((A, Xs, E), (A, L0, F)),
	pac_lisp(F, V, A).
pac_lisp([=, X, Y], R, A):- !, pac_lisp(X, U, A),
	pac_lisp(Y, V, A),
	( U==V -> R=t; R=[]).
pac_lisp([if, C, X|Ys], V, A):- !, pac_lisp(C, B, A),
	(	B\==[]
	->	pac_lisp(X, V, A)
	;	pac_lisp_(Ys, Zs, A),
		last(Zs, V)
	).
pac_lisp([and|L], V, A):-!,pac_lisp_and(L,V,A).

pac_lisp([F|L], V, A):- defun(F, G), !, pac_lisp([G|L], V, A).
pac_lisp(X, X, _):- !.

%
pac_lisp_([], [], _).
pac_lisp_([E|Es], [V|Vs], A):- pac_lisp(E, V, A),
	pac_lisp_(Es, Vs, A).

%
pac_lisp_and([], t, _):- !.
pac_lisp_and([H|M], V, A):- pac_lisp(H, U, A), U==t, !,
	pac_lisp_and(M, V, A).
pac_lisp_and(_, [], _).

%%%% basics for list processing
defun(car, [[X|_]]\ X).
defun(cdr, [[_|X]]\ X).
defun(cons, [X,Y]\ [X|Y]).
defun(not, [X]\[if, [=, X,[]], t, []]).
defun(append, [X,Y]\[if, [=, X, []], Y,
		     [cons, [car, X], [append, [cdr, X], Y]]]).


% ?- length(L, 3), maplist(st([X], X=1), L).
% ?- call(st([X], X=1), A).
% ?- call(st([X,Y], (X=1, Y=X)), A, B).
% ?- maplist(st([X], [X], X=1), [A,B]).
% ?- maplist(st([a(X)], [X], X=1), [A, B]).
% ?- length(L, 3), maplist(st([a(X)], [X], X=1), L).
% ?- length(L, 3), Y=5, maplist(st([a(X, Y)], [Y], X=1), L).
% ?- call(st([B], [], call(st([A], [], A=1), B)), Z), Z).
% ?- maplist(st([B], [], call(st([A], [], A=1), B)), [Z1, Z2, Z3]).
% ?- maplist(st([B], [], maplist(st([B0, B1], [], B0=B1), [1,2,3], B)), [Z1, Z2, Z3]).
% ?- maplist(st([A], [], maplist(st([B0, B1], [], B0=B1), [1,2,3], A)), Cs).
% ?- length(Cs, 2), P = hi, maplist(st([A], [P], maplist(st([B0, B1], [P], B0=(B1, P)), A, [1,2,3])), Cs).
% ?- time(1000000, maplist(lambda:(X\Y\Z\append(X, Y, Z)), [[1]],[[2]], _), T).
% ?- time(1000000, eval(ap::maplist(append, [[1]],[[2]]), _), T).

% internal_lambda(X\\Y, st(Xs, Ps, Y0)):- !, eval:cast_to_list(X, Xs),
% 	internal_lambda(Y, st(_, Ps, Y0)).
% internal_lambda(^^(X,Y), st([], Ps, Y)):- !, eval:cast_to_list(X, Ps).
% internal_lambda(X\Y, st(Xs, Ps, Y0)):- !, internal_lambda(X\Y, Xs, Ps, Y0).
% internal_lambda(X^Y, st(Xs, Ps, Y0)):- !, internal_lambda(X^Y, Xs, Ps, Y0).

% internal_lambda(Y, st([], [], Y)).

% internal_lambda(A, Xs, Ps, Y):- internal_lambda(A, [], [], X0s, Ps, Y),
% 	reverse(X0s, Xs).

% internal_lambda(X\Y, Xs, Ps, Ys, Qs, Y0):-!,
% 	internal_lambda(Y, [X|Xs], Ps, Ys, Qs, Y0).
% internal_lambda(X^Y, Xs, Ps, Ys, Qs, Y0):- !,
% 	internal_lambda(Y, Xs, [X|Ps], Ys, Qs, Y0).
% internal_lambda(Y, Xs, Ys, Xs, Ys, Y).

% % lambda_expansion(X, _):- writeln(X), fail.
% lambda_expansion(X, X):- (var(X); atomic(X)), !.
% lambda_expansion(X\\Y, Z):- internal_lambda(X\\Y, S), lambda_expansion(S, Z).
% lambda_expansion(^^(X,Y), Z):- internal_lambda(^^(X, Y), S), lambda_expansion(S, Z).
% lambda_expansion(X\Y, Z):- !, internal_lambda(X\Y, Xs, Ps, Y0),
% 	lambda_expansion(st(Xs, Ps, Y0), Z).
% lambda_expansion(X^Y, Z):- !, internal_lambda(X^Y, Xs, Ps, Y0),
% 	lambda_expansion(st(Xs, Ps, Y0), Z).
% lambda_expansion(st(X, Body), Y):-!,
% 	lambda_expansion(st(X, [], Body), Y).
% lambda_expansion(st(Xs, Ps, Body), Y):-!,
% 	lambda_expansion(Body, Body0),
% 	gensym('$st', Pred),
% 	H=..[Pred, Ps|Xs],
%  	assert(H:-Body0),
% 	Y=..[Pred, Ps].
% lambda_expansion(M, M0):- M=..[F|As],
% 	maplist(lambda_expansion, As, Bs),
% 	M0=..[F|Bs].

% goal_expansion(X, Y):- lambda_expansion(X, Y).

% % term_expansion(X, Y):- lambda_expansion(X, Y).

% % sample uses of st/2, st/3.
% % ?- eval(maplist(l\ maplist(i\ nth1(i,l), [2,1,1]), [[a,b],[b,c]]), X).
% %  Not work because of another meaning of \
% % ?- csv_column_extract([3, 1], [[a,b,c],[c,d,e],[f,g,h]],  Rs).

% csv_column_extract_ulrich(Indexes, Raws, New_raws):-
% 	maplist(Raw\New_raw\
% 		(Indexes^maplist(Index\Elem\
% 				(Raw^nth1(Index, Raw, Elem)),
% 				Indexes, New_raw)),
% 		Raws, New_raws).

% csv_column_extract(Indexes, Raws, New_raws):-
% 	maplist(st([Raw, New_raw], [Indexes],
% 		   maplist(st([Index, Elem], [Raw],
% 			      nth1(Index, Raw, Elem)),
% 			   Indexes, New_raw)), Raws, New_raws).

% csv_column_extract_me(Indexes, Raws, New_raws):-
% 	maplist([Raw, New_raw]\\
% 		       (Indexes^maplist(
% 			[Index, Elem]\\
% 			       (Raw^nth1(Index, Raw, Elem)),
% 				Indexes, New_raw)),
% 		Raws, New_raws).

% %% some tiny
% list_lambda :- predicate_property(X, dynamic),
% 	functor(X, F, _),
% 	atom_concat('$st', _, F),
% 	listing(F),
% 	fail; true.
