:- module(misc, [
	 compile_pred_word/4
  ,  random_choice/2
  ,  collect_subterm/3
  ,  collect_subterm/4
  ,  completing_optional_args/3
  ,  cs/2
  ,  desugaring/2
  ,  dir_minus/3
  ,  dir_plus/3
  ,  directory_files/3
  ,  directory_remove_ymhms/1
  ,  drop3zw/2
  ,  find_exportables_from_calls/2
  ,  ignore/2
  ,  insert/3
  ,  insert_pause/2
  ,  kanji/3
  ,  list/2
  ,  map_directory_files/2
  ,  matrix_paths/2
  ,  parse_time/3
  ,  parse_utf8/2
  ,  predicate_arity/2
  ,  predicate_arity_for_export/2
  ,  rename_directory_suffix/3
  ,  set/2
  ,  skip_filler/2
  ,  split_by_filler/3
  ,  string/2
  ,  shell_string/2
  ,  texadjust/2
  ,  texuncomment/2
  ,  token_and_split/3
  ,  try2problem/2
  ,  expand_sgn_brace/2			]).

:- use_module(pac(basic)).
:- use_module(util(file)).
:- use_module(util('env-dict')).
:- use_module(util('meta2')).
:- use_module(util(math)).

term_expansion --> pac:expand_pac.
:- use_module(pac(op)).
%
:- use_module(pac('expand-pac')).

		/***********************************
		* Compile clause / DCG with regex  *
		***********************************/

new_names([V|Vs], [A=V|Eqs], N, Prefix, As):-
	new_name(N, As, A, Prefix, K),
	new_names(Vs, Eqs, K, Prefix, As).
new_names([], [], _, _, _).

%
new_name(N, As, B, Prx, K):- atom_concat(Prx, N, B),
						\+ memberchk(B, As),
						!,
						succ(N, K).
new_name(N, As, A, Prx, K):- succ(N, N1),
						new_name(N1, As, A, Prx, K).

% ?- ejockey:subtractq([X,Y,X,Y,X,Y], [X], R).
subtractq([], _, []).
subtractq([A|As], B,  C):- memq(A, B), !,
	subtractq(As, B, C).
subtractq([A|As], B, [A|C]):- subtractq(As, B, C).

%
expand_clause_slim(X, Y):-
	expand_clause(X, [], Y0),
	maplist(pred([X:-true, X] & [C, C]),  Y0, Y).
%
compile_pred_word(X-->X0, Eqs, H0, R0):-!,
		maplist(pred([A=P, A, P]), Eqs, As, Vs),
		expand_clause_slim(X-->X0, [H|R]),
		term_variables(H, HVs),
		subtractq(HVs, Vs, SVs),
		new_names(SVs, Eqs0, 1, 'A', As),
		append(Eqs0, Eqs, Eqs1),
		term_string(H, H0, [variable_names(Eqs1),
							quoted(true)]),
		maplist(pred(([U, [V,".\n"]] :-
								numbervars(U, 0, _),
								term_string(U, V, [ numbervars(true),
													quoted(true)]))),
				R, R0).
compile_pred_word(X, Eqs, H0, R0):-
		expand_clause_slim(X, [H|R]),
		term_string(H, H0, [variable_names(Eqs),
							quoted(true)]),
		maplist(pred(([U, [V,".\n"]] :-
								numbervars(U, 0, _),
								term_string(U, V, [ numbervars(true),
													quoted(true)]))),
				R, R0).


			/********************
			*     shell string  *
			********************/

%%	shell_string(+X:term, -Y:string) is det.
%	Expand  a shell command term X into
%	a string that can be passed to shell/1.

% ?- shell_string(open("1"), X).
% ?- shell_string(open(1), X).
% ?- shell_string(open(abc), X).
% ?- shell_string(open(-l, -m, abc), X).
% ?- shell_string(open(-f(a)), X).
% ?- shell_string(open(-f(a)), X).

shell_string(X, Y):- shell_codes(X, Y0), basic:smash(Y0, Y).

%%	shell_codes(+X:term, -Y:text) is det.
%	Expand  a shell command term X into
%	a text that can be passed to shell/1.

shell_codes(++(A, B), [A1, " ", A2]):-!,
	shell_codes(A,A1),
	shell_codes(B,A2).
shell_codes(A+B, [A1, A2]):-!,
	shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes(&(A, B), [A1, " & ",  A2]):-!,
	shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes((A;B), [A1," ; ",A2]):-!,
	shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes((A|B), [A1," | ",A2]):-!,
	shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes(A>B, [A1, " > ",A2]):-!, shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes(A>>B, [A1," >> ", A2]):-!,
	shell_codes(A, A1),
	shell_codes(B, A2).
shell_codes(shell(A), ["( ", A1," )"]):-!, shell_codes(A, A1).
shell_codes(A, A):- (is_list(A); atomic(A)), !.
shell_codes(A, B):- compound(A),
	A =.. [F|As],
	maplist(shell_codes_option, As, Bs),
	insert(" ", [F|Bs], B).

% ?- misc:shell_codes(ls, X).
%%	shell_codes_option(+X:term, -Y:text) is det.
%	Expand a shell command option term X into a text.
shell_codes_option(++(A, C), [B, " ", D]) :-!,
     shell_codes_option(A, B),
     shell_codes_option(C, D).
shell_codes_option(A+C, [B, D]) :-!,
     shell_codes_option(A, B),
     shell_codes_option(C, D).
shell_codes_option(-A, ["-", B, " "|D]) :- A=..[B|C], !,
     maplist(shell_codes_plus, C, D).
shell_codes_option(--(A), ["--", B, " "|D]) :- A=..[B|C], !,
     maplist(shell_codes_plus, C, D).
shell_codes_option(A, B) :- number(A), !,
     number_codes(A, B).
shell_codes_option(A, A).

%
shell_codes_plus(++(A, C), [B, " ", D]) :-!,
     shell_codes_plus(A, B),
     shell_codes_plus(C, D).
shell_codes_plus(A+C, [B, D]) :-!,
     shell_codes_plus(A, B),
     shell_codes_plus(C, D).
shell_codes_plus(A, B) :-number(A), !,
     number_codes(A, B).
shell_codes_plus(A, A).

		/**********************************
		*     compleging optional args    *
		**********************************/

% ?- completing_optional_args([a(hello)], [], R).
% ?- completing_optional_args([a(X):hello], [], R).
% ?- completing_optional_args([a(X): !(X=hello)], [], R).
% ?- completing_optional_args([a(X):hello], [], R).
% ?- completing_optional_args([a(X):hello], [a(world)], R).

%% completing_optional_args(+X, +Y, -Z) is det.
%	option Z is unified with Y completed with
%	default option X.
completing_optional_args(X, Y, Z):-
	completing_optional_args(X, Y, Z, []).

% ?- zdd_array:completing_optional_args([a(X)], [a(Y)], R, S).
% ?- zdd_array:completing_optional_args([a(world)], [a(hello)], R, S).
% ?- zdd_array:completing_optional_args([a(X):b], [a(Y)], R, S).
% ?- zdd_array:completing_optional_args([a(X):10], [a(Y)], R, S).

completing_optional_args([], R, R, []):-!.
completing_optional_args([A|As], Bs, [B|R], R0):- (A = A0:_ ; A0= A), !,
	functor(A0, F, 1),
	functor(B, F, 1),
	( select(B, Bs, Cs); Cs = Bs ), !,
	unify_one(A, B),
	completing_optional_args(As, Cs,  R,  R0).

% For preventing expansion brace {..} by open_dict as pac macro.
% ( not used )
:- set_prolog_flag(open_dict, false).
unify_default({G}, _):-!, once(G).
unify_default(X, X).
:- set_prolog_flag(open_dict, true).

%
unify_one(A:Val, B):- !, functor(A, F, 1),
	functor(B, F, 1),
	arg(1, A, U),
	arg(1, B, V),
	(	var(V)	->	U = Val,	% unify_default(Val, U),
					V = U
	;	var(U)	->  U = V
	;	true
	).
unify_one(A, A):-!.
unify_one(_, _).

			/************************
			*       Set kind        *
			************************/

%%	set(+X:term, -Y:list) is det.
%	Evaluate a set expression X into a list Y.

% ?- set([a,b]-[1,2], R).
% ?- set(singleton(pow([1,2])), X), length(X, L).

set(emptyset,[]):-! .
set(X,X):-listp(X),! .
set(singleton(A),[B]):-!,set(A,B) .
set(A+B,A1):-!,(set(A,A2),set(B,A3)),union(A2,A3,A1) .
set(plus(A,B),A1):-!,set(A+B,A1) .
set(cup(A),A1):-!,(set(A,A2),set(append(A2),A3)),sort(A3,A1) .
set(cap(A),A1):-!,set(A,A2),math:bigcap(A2,A1) .
set(++(A,B),A1):-!,(set(A,A2),set(B,A3)),math:direct_sum(A2,A3,A1) .
set(A*B,A1):-!,(set(A,A2),set(B,A3)),math:product(A2,A3,A1) .
set(A-B,A1):-!,(set(A,A2),set(B,A3)),
	pac_meta:pac_product(misc:set_aux,A2,A3,A1) .
set(\(A,B),A1):-!,(set(A,A2),set(B,A3)),subtract(A2,A3,A1) .
set(#(A,B),A1):-!,(set(A,A2),set(B,A3)),scramble_cons(A2,A3,A1) .
set(&(A,B),A1):-!,(set(A,A2),set(B,A3)),math:intersection(A2,A3,A1) .
set(pow(A),A1):-!,set(A,A2), math:powerset(A2,A1) .
set(zip(A,B),A1):-!,(set(A,A2),set(B,A3)),zip(A2,A3,A1) .
set(sort(A),A1):-!,set(A,A2),sort(A2,A1) .
set((A->B),A1):-!,(set(A,A2),set(B,A3)),math:mapset(A2,A3,A1) .
set(..(I,J),A1):-(I0 is I,J0 is J),!,numlist(I0,J0,A1) .
set(in(X,Y),A1):-!,(set(X,A2),set(Y,A3)),truth(memberchk,A2,A3,A1) .
set(X=Y,A1):-!,(set(X,A2),set(Y,A3)),truth(==,A2,A3,A1) .
set(X=<Y,A1):-!,(set(X,A2),set(Y,A3)),truth(subset,A2,A3,A1) .
set(X>=Y,A1):-!,set(Y=<X,A1) .
set(X<Y,A1):-!,(set(X,A2),set(Y,A3)),truth(math:proper_subset,A2,A3,A1) .
set(X>Y,A1):-set(Y<X,A1) .

%
set_aux(A,B,A-B).

%%	insert(+M:term, +X:list, -Y:list) is det.
%	Insert M beteen every successive elements of X,
%	and unify Y with the result.

% ?- misc:insert(a, [1,2,3], R).

insert(M, [A|P], [A|Q]):- foldl(pred(M, [B, [M, B| U], U]), P, Q, []).
insert(_, [], []).

%
truth(X, true) :- call(X), !.
truth(_, false).

truth(X, Y, true) :- call(X, Y), !.
truth(_, _, false).

truth(X, Y, Z, true) :- call(X, Y, Z), !.
truth(_, _, _, false).

%%	desugaring(+X:term, -Y:term) is det.
%	Eliminate infix '-' and  '/', and  prefix '+',  '-',
%	and ufify Y with the result.

% ?- desugaring( -(a-b-c), Y).

desugaring(X-Y,A1+ -1*A2):-!,desugaring(X,A1),desugaring(Y,A2).
desugaring(+X,A1):-!,desugaring(X,A1).
desugaring(-X,-1*A1):-!,desugaring(X,A1).
desugaring(X/Y,A1 rdiv A2):-!,desugaring(X,A1),desugaring(Y,A2).
desugaring(X,X).

insert_pause(cs(item),[cs(pause),cs(item)]):-! .
insert_pause(X,A1):-(listp(X),maplist(insert_pause,X,Y)),!,insert_pause(Y,A1) .
insert_pause(X,A1):-(X=..[F|As],maplist(insert_pause,As,Bs),Z=..[F|Bs]),!,
		insert_pause(Z,A1).

texadjust([cs(item)|X],A1):-texadjust(X,Y),!,texadjust([10,cs(item)|Y],A1) .

cs(cs(N),[N]):-!.
cs(A,A1):-listp(A),!,maplist(cs,A,A1) .
cs(A,A1):-A=..[_A2|As], maplist(cs,As,A1) .

texuncomment(comment(_A1),[]).

try2problem(env(try,Body),A1):-try2problem(Body,A2),try2problem(env(problem,A2),A1).

drop3zw(ddol(E),[cs(noindent),cs(skip),"3zw",env(coronamath,[dol([cs(displaystyle)," ",E])])]).



% product(A, B, C):-
% 	foldr( pred(B, [X,P,Q] :-
% 		foldr(pred(X, [Y,U,[X-Y|U]]), B, P, Q)), A, [], C).

% product_cons(A, B, C):-
% 	foldr( pred(B, [X,P,Q] :- foldr(pred(X, [Y,U,[[X|Y]|U]]), B, P, Q)), A, [], C).

% ?- product_cons([1,2],[[a,b],[c,d,e]], X).
%@ X = [[1,a,b],[1,c,d,e],[2,a,b],[2,c,d,e]] .

% product3(A, B, C, D):-
% 	foldr( pred([B,C], [X,P,Q] :-
% 		foldr(pred([X,C], [Y,U, V]:-
% 			 foldr(pred([X,Y], [Z, M, [X-Y-Z|M]]), C, U, V)),
% 				B, P, Q)), A, [], D).

% product4(A, B, C, D, E):-
% 	foldr( pred([B,C,D],
% 		    [X, P, Q] :-
% 		foldr(pred([X,C,D],
% 			   [Y, U, V]:-
% 			 foldr(pred([X,Y,D],
% 				    [Z, M, N]:-
% 			      foldr(pred([X,Y,Z],
% 					 [W, R,[X-Y-Z-W|R]]), D, M, N)),
% 			       C, U, V)),
% 		      B, P, Q)),
% 	        A, [], E).

matrix_paths([], [[]]).
matrix_paths([X|Y], Z):- matrix_paths(Y,  PY),
	foldr(pred(PY, [J, P, Q]:-
	     foldr(pred(J, [K, N, [[J|K]|N]]), PY, P, Q)),
	      X, [], Z).


% ?- time(repeat(10^8, choices_by_fold([[1,2],[a,b,c],[3,4]], R))).
%@ % 7,900,000,001 inferences, 394.599 CPU in 394.833 seconds (100% CPU, 20020346 Lips)
%@ true.

% ?- time(repeat(10^8, choices([[1,2],[a,b,c],[3,4]], R))).
%@ % 4,300,000,001 inferences, 138.387 CPU in 138.479 seconds (100% CPU, 31072358 Lips)
%@ true.


%%	choices by foldl/foldr version. Compiled source is listed
%	by [compile, pac, region] of emacs-jockey.pl
%
% choices_by_fold(X, Y):- reverse(X, X0),
% 	foldl(pred([A,B,C] :-
% 	     foldr( pred(B,
% 			 [L,P,Q] :-
% 		  foldr(pred(L, [M,U,[[L|M]|U]]),
% 			B, P, Q)),
% 		    A, [], C)),
% 	      X0, [[]], Y).

% ?- length(L, 10), maplist(misc:random_choice([[a,b],[c,d,e]]), L), maplist(writeln, L).

random_choice([L|R], [A|Q]):- length(L, N), I is random(N),
			      nth0(I, L, A),
			      random_choice(R, Q).
random_choice([], []).


%%	list(+X:term, -Y:list) is det.
%	Expand a list term X into a list Y.

% ?- list([1,2]\[1,2,3,4], R).
% ?- list([a,b]^2, R).
% ?- list(^^([a,b], 2), R).
% ?- list([a,b]^2+[c,d]^3+[e,f]^4, R), length(R, L).
% ?- list([a,b]^0+[c,d]^1+[e,f]^2, R), length(R, L).
% ?- list([a,b]^^3, R), length(R, L).

list(A+B,A1):-!, list(A,A2),
	list(B,A3),
	append(A2,A3,A1).
list(\(A,B),Z):-!, list(A,X),list(B,Y), append(X,Z,Y) .
list(A/B,Z):-!, list(A,X),list(B,Y), append(Z,Y,X) .
list(E^L,A1):-N is L,!,list(E,A2),times(N,A2,A1) .
list(^^(E,L),C):-N is L,!,list(E,A),math:nlist(A,N,C) .
list(X,X):-listp(X),! .
list(+A,A1):-!,list(A,A2), append(A2,A1) .
list(flat(A),A1):-list(A,A2),flatten(A2,A1) .

%%	string(+X:term, -Y:string) is det.
%	Expand a  string term X into a string Y.

% ?- string("ab"+"cd"+"ef", X).
% ?- string("abcd"/"cd", X).
% ?- string("ab"\"abcd", X).
% ?- string(("ab"\"abcd")^3, X).
% ?- string(("ab"\"abcd")^3, X).
% ?- val(string((+) @ (list@ (+(["ab", "cd"], ["ef", "gf"])))), Y).
% ?- string(reverse("abcd"), X).
% ?- string(reverse(reverse("uvw") + reverse("ab" + "12")), X).

string(A+B,A1):-!, string(A,A2),
	string(B,A3),
	string_concat(A2,A3,A1).
string(\(A,B),Z):-!, string(A,X),
	string(B,Y),
	string_concat(X,Z,Y).
string(A/B,Z):-!, string(A,X),
	string(B,Y),
	string_concat(Z,Y,X).
string(E^L,A1):-N is L,!,string(E,A2),
	string_times(N,A2,A1) .
string(reverse(X),B):-!, string(X,A1),
	string_codes(A1,A2),
	reverse(A2,A),
	string_codes(B,A).
string(+A,A1):-!, string(A,A2),
	list(A2,A3),
	string_list_concat(A3,A1).
string(X,X):-(string(X);atom(X)),!.


% split by variable length filler.
split_by_filler(X) --> (skip_filler; []), token_and_split(X).

%
token_and_split([]) --> elisp:end_of_list.
token_and_split([A|X]) --> wl("[^\s\t\r\n]+", A, []),  split_by_filler(X).

skip_filler --> wl("[\s\t\r\n]+").

% ?- listing(token_and_split).
% ?- listing(skip_filler).
% ?- listing('nt#6').
% ?- listing('nt#5').
% ?- skip_filler(`   abc`, X).

% ?- trace, string_chars("ab", L), split_by_filler(X, L, Y).
% ?- trace, string_chars("    ", L), split_by_filler(X, L, Y).
% ?- trace, string_chars(" ab  c\n \t   ", L), split_by_filler(X, L, Y).
% ?- trace, string_chars(" ab  c\nd\t ef   ", L), split_by_filler(X, L, Y).

% Apply set operation on directories as a set of files.

:- meta_predicate directory_files(2,?,?).
directory_files(E, Ds, L):-
	maplist(pred([D, D0]:-expand_file_name(D, [D0])), Ds, Es),
	maplist(directory_files, Es, Ls),
	call(E, Ls, L).

:- meta_predicate map_directory_files(1, ?).
map_directory_files(F, D):- directory_files(D, Fs),
	working_directory(D0, D),
	maplist(ignore(F), Fs),
	working_directory(_, D0).

ignore(F, X):- ignore(call(F, X)).

% ?- directory_remove_ymhms('/Volumes/sparrow/raw-scan-master').
% ?- directory_remove_ymhms('/Volumes/shark/searchable-master').
directory_remove_ymhms(D):- expand_file_name(D, [D0]),
	map_directory_files(pred(([X]:-
					atom_codes(X, X0),
					parse_time_me(".pdf", X0, Y0),
					atom_codes(Y,Y0),
					rename_file(X, Y) )),
					D0).

% :- listing(parse_time).
% ?-  misc:parse_time(`pdf`, `2007年12月21日21時35分18秒.pdf`, P), atom_codes(Y, P).

parse_time(Ext) -->
	w(*(char(digit)), Year),	"年",
	w(*(char(digit)), Month),	"月",
	w(*(char(digit)), Day),		"日",
	w(*(char(digit)), Hour),	"時",
	w(*(char(digit)), Minute),	"分",
	w(*(char(digit)), Second),	"秒",
	".",
	w(*(.), Ext),
	current([]),
	{ append([Year, Month, Day, Hour, Minute, Second, `.`,  Ext], Y) },
	peek(Y).

% ?- dir_minus('~/Desktop', '~/Desktop', Z).
% ?- dir_minus('~', '~/Desktop', Z).
% ?- dir_minus('~', '~', Z).
dir_minus(X,Y,Z):- directory_files(fun([[A,B]]-> (set::(A\B))), [X, Y], Z).
dir_plus(X,Y,Z):- directory_files(fun([[A,B]]-> (set::(A+B))), [X, Y], Z).

%
rename_directory_suffix(Suffix0, Suffix, Dir):-
	expand_file_name(Dir, [Dir0]),
	map_directory_files(
 	 pred([Suffix0,Suffix], [X]:- (	atom_concat(X0, Suffix0, X),
					atom_concat(X0, Suffix, Y),
					rename_file(X, Y)
				)),
	Dir0).

:- meta_predicate collect_subterm(1,?,?).

collect_subterm(F, X, Y):- collect_subterm(X, Y0, [], F), sort(Y0, Y).

%
collect_subterm(X, [X|V], V, F):- call(F, X), !.
collect_subterm(X, V, W, F):- compound(X), !,
	X=..[_|As],
	foldr(	pred(F, [A, P, Q] :- collect_subterm(A, P, Q, F), As, V, W)).
collect_subterm(_, V, V, _).

% :- dynamic calls/2.
% % from swi library.
% assert_call_graph :-
%         retractall(calls(_, _)),
%         prolog_walk_code([ trace_reference(_),
%                            on_trace(assert_edge),
%                            source(false)
%                          ]),
%         predicate_property(calls(_,_), number_of_clauses(N)),
%         format('Got ~D edges~n', [N]).

% assert_edge(Callee, Caller, _Where) :-
%         calls(Caller, Callee), !.
% assert_edge(Callee, Caller, _Where) :-
%         assertz(calls(Caller, Callee)).

%
predicate_arity(F, L):-
	setof(P/N,
		A^X^(	predicate_property(A:X, file(F)),
			functor(X, P, N)
		  ),
	      L),
	insert(", ", L, L0),
	smash(["[", L0, "]"]).

%
predicate_arity_for_export(F,  L):-
	writeln('start ...'),
	assert_call_graph,
	writeln('call graph done'),
	find_exportables_from_calls(F, L).

% ?- qcompile(zdd(zdd)), module(zdd).


% ?- find_exportables_from_calls('/Users/cantor/local/lib/pacpl7/engine.pl', Expo), insert(",\n", Expo, E), maplist(write, E).

find_exportables_from_calls(Loc,  Exp):-
	predicate_arity(Loc,  L0),
	setof(F/N, P^X^(
			member(F/N, L0),
			functor(P, F, N),
			once(calls(X,  user:P)),
			\+ predicate_property(X, file(Loc))
		       ),
	      Exp),
	insert(",\n", Exp, E),
	maplist(write, E).


% ?- calls(X, user:assocq(_,_,_)), predicate_property(X, file(A)).
% ?- misc:residue(([a,b,c,d,e]-[a,b]-[c]), R).

residue(X-Y,V):-!, residue(X,A),
	residue(Y,U),
	append(U,V,A).
residue(X,X).

% ?-  parse_utf8(`My name is 向井 国昭.`, R), smash(R).
parse_utf8 --> sed(kanji(A), =(A)),
	maplist(pred(([X, Y] :- listp(X), string_codes(Y, X))
	       &
	       ([X, Y] :- char_code(Y, X)))).

kanji(A)  --> w(char(utf8b), A, B), wl(*char(utf8c), B).

% ?- misc:cgi_bin_name_edit.
%@ true.

% For test.
% cgi_bin_name_edit :- cgi_bin_name_edit(
%      		"/Users/cantor/Desktop/http_request.js",
% 		"/Users/cantor/Desktop/http_request1.js").
% for test.
cgi_bin_name_edit :- cgi_bin_name_edit(
	 	"/Users/cantor/local/cgi-bin/http_request.js",
		"/Users/cantor/Desktop/http_request.js").

cgi_bin_name_edit(In_js_name, Out_js_name) :-  getenv(cgi_bin, U),
	atomics_to_string([U, /], V),
	term_string(V, V0),
	atomics_to_string([V0, ";"], V1),
	string_codes(V1, V2),
	setup_call_cleanup(
	 (	open(In_js_name, read, In),
		open(Out_js_name, write, Out) ),
		repeat_line(V2, In, Out),
	(	close(In),
		close(Out))).

cgi_bin_name_edit(V, A, B):-
	head_sed(sed([V],
		 (wl("[\s\t]*var[\s\t]+CGIBIN[\s\t]*=[\s\t]*", X), wl(".*")),
		 (X + V)), A, B).
%
repeat_line(V, In, Out):- read_line_to_codes(In, A),
	A \== end_of_file,
	cgi_bin_name_edit(V, A, B),
	maplist(put_code(Out), B),
	put_code(Out, 0'\n),			%%%% '
	!,
	repeat_line(V, In, Out).
repeat_line(_,_,_).


% ?- expand_sgn_brace(sgn([a,b]), X).
% ?- expand_sgn_brace((sgn([a,b]), sgn([c,d])), X).
expand_sgn_brace(sgn([A,A0|B]),(A1;A2)):-!,
	expand_sgn_brace(sgn([A]),A1),
	expand_sgn_brace(sgn([A0|B]),A2).
expand_sgn_brace(sgn([A]),A1):-!,expand_sgn_brace(A,A1).
expand_sgn_brace((A,B),(A1,A2)):-!,expand_sgn_brace(A,A1),
	expand_sgn_brace(B,A2).
expand_sgn_brace((A;B),(A1;A2)):-!,expand_sgn_brace(A,A1),
	expand_sgn_brace(B,A2).
expand_sgn_brace(\+A,\+A1):-!,expand_sgn_brace(A,A1).
expand_sgn_brace(A/N,A0/N):-!,expand_sgn_brace(A,A0).
expand_sgn_brace(X,A1):-is_list(X),!,
	maplist(expand_sgn_brace,X,A1).
expand_sgn_brace(X,X).

% Samples defining functions on a sort of expressions.

% ?- listing(list_plus).
% ?- misc:list_plus([a,b,c]+[d,e,f]+[g,h,i], R).

list_concat(X+Y,A1):-list_concat(X,A2),
	list_cocnat(Y,A3),
	append(A2,A3,A1) .
list_concat(X,X).


% ?- listing(my_add).
% ?- misc:my_add(1+2+3+4+5, R).
my_add(X+Y,A1):- my_add(X,A2),
	my_add(Y,A3),
	plus(A2,A3,A1).
my_add(X,X).

% ?- misc:string_rev(rev("abc"), R).
%@ R = "cba" .
% ?- write_canonical([rev/1-misc:string_reverse]).
%@ [:(-(/(rev,1),misc),string_reverse)]

% list_plus		:= [(+)/2-append].
% list_rev		:= [rev/1-reverse].
% list_plus_rev	:= (list_plus + list_rev).
% list_plus_only	:= (list_plus_rev \ list_rev).

% % string sort.
% % ?- misc:string_plus_rev(rev(rev("abc")+rev("def")), R).
% %@ R = "defabc" .

% string_plus		:= [(+)/2- string_concat].
% string_rev		:= [rev/1- (misc:pred(([X, Y]:-
% 								string_codes(X, X0),
% 								reverse(X0, Y0),
% 								string_codes(Y, Y0))))].
% string_plus_rev	:= (string_plus + string_rev).
% string_plus_rev_meet := (string_plus * string_rev).
% string_plus_only	:= (string_plus_rev \ string_rev).

% empty_sgn := [].

% Example uses of @
% uuu(A,B) --> A@B.
% vvv(A,B) --> #(A)@B.
% www(A, B) --> fun([X,Y,Z]->(X@Y@Z))@A@B.
%?- misc:www(append, [a,b], [c,d], R).
%@ R = [a, b, c, d].
% ?- call(fun([F,X]-> F@X), append([a,b]), [c,d], R).
%@ R = [a, b, c, d].

% DISCOURSE by @brebs

% ?- split_list_at_nth1(N, [a,b,c], [X|T], End).
% ?- split_list_at_nth1(N, [a,b,c], St, [a,b,c|T]).
% ?- trace, split_list_at_nth1(N, [a,b], Start, End).
split_list_at_nth1(Nth1, Long, Start, End) :-
    (   nonvar(Nth1) -> must_be(nonneg, Nth1), Once = true
    ;   is_list(Long), once(is_list(Start) ; is_list(End)) -> Once = true
    ;   is_list(End), is_list(Start) -> Once = true
    ;   Once = false
    ),
    split_list_at_nth1_(Long, 0, Nth1, Once, Start, End).

split_list_at_nth1_(L, N, N, Once, [], L) :-
    (Once == true -> ! ; true).
split_list_at_nth1_([H|T], N, Nth1, Once, [H|Upto], End) :-
    N1 is N + 1,
    split_list_at_nth1_(T, N1, Nth1, Once, Upto, End).
