:- module(model,[]).
:- use_module(pac('expand-pac')).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

entails(X, B, C):- generate(X, M), models(M, B), \+ models(M, C), !, fail.
entails(_, _, _).

% ?- hmodel:find_model( grupoid::[a,b], \+ $(idempotence), R).

find_model(F, C, X):- call(F, generate, X), models(X, C).

models(_, true):-!.
models(M, (C1,C2)):- !, models(M, C1), models(M, C2).
models(M, (C1;C2)):- !,( models(M, C1); models(M, C2)).
models(M, (C1->C2)):- !, (\+ models(M, C1) ; models(M, C2)).
models(M, (\+ C)):- !,  \+ models(M, C).
models(M, every(Y,C)):- !, forall(belong(Y,M), models(M,C)).
models(M, some(Y,C)):- !,  belong(Y, M), models(M, C).
models(M, Y=Z):- !, call(M, Y, V), call(M, Z, V).
models(M, $(Property)):- !, call(Property, M).
models(_, prolog(G)):- !, once(G).
models(M, forall(P, G)):- !, forall(P, models(M, G)).
models(M, exists(P, G)):- !, call(P), models(M, G).
models(M, X):- 	mapterm(M, X, X0), holds(M, X0).

%%%%%
holds(_,  A=<A).
holds(M,  A=<B):- world(M, Fs), memberchk((A,B), Fs).

% ?- listing(algebra).
% ?- listing('pac#52').

algebra(M,A*B,A1):-!,(algebra(M,A,A2),algebra(M,B,A3)),'pac#431'(M,A2,A3,A1) .
algebra(A1,A,A):-! .

'pac#431'(A,B,C,D):-memberchk(((B,C),D),A) .

%%%% Plugin; an element of the domain.
belong(X,og(L,_,_)):-!, in(X, L).
belong(X,int(K,J)):- !,  between(K,J,X).
belong(X,[Y|Z]):-!, (X=Y; in(X, Z)).
belong(X,lin(N,_)):-!,  between(1,N,X).
belong(X,grupoid(L,_)):- in(X, L).

%%%% Plugin: eval (interface) %%%%%%

% :-bekind(og//3).
% X = with(_, _, H, model:algebra(H)::X).
% :-ekind.

% :-bekind(grupoid//2).
% X = with(_, B, algebra(B)::X).
% :-ekind.

%%%%% some tiny
promote_to_mode(M, [model:M]).

world(og(_,X,_), X).

%%%%  miscellaneous properties of models

aug(OG):- models(OG,every(X,every(Y, (Y=< (X*Y))))).

associative(G):-
	models(G, every(X,every(Y,(every(Z, ((X* Y)*Z) = (X*(Y*Z))))))).

slide1(M):-
	models(M, every(A,every(B, every(C, (((A*B)*C) =< (A*(B*C))))))).

mix(M):- models(M, every(A,every(B, every(C, every(D,
	((A=<B, C=<D) -> ((B*C) =< (A*D)))
	))))).

isolated(X, graph(_,E)):- forall((in((X,Y),  E); in((Y,X),  E)), X==Y).

maximal(X, graph(_,E)):- forall(in((X,Y), E), X=Y).

minimal(X, graph(_,E)):- forall(in((Y,X),  E), X=Y).

irreducible(L, grupoid(_,G)):- forall(in(X, L), binary_closed(G,[X],L)).

idempotent(A, grupoid(_,H)):- in(((A,A),A),  H).

idempotent(M):- grupoid_table(M, Table), memberchk(((A,A),A), Table).

left_unit(U, grupoid(D,M)):- forall(in(X, D), in(((U,X),X),  M)).

right_unit(U, grupoid(D,M)):- forall(in(X, D), in(((X,U),X), M)).

rich(M):-!, M=lin(N,Asg),
	models(M, every(K, (setofint(K,1,N,S), subalgebra(S,Asg)))).

surj(og(D,_,M)):- maplist(snd, M, M0), sort(M0, D).


in(X,[X|_]).
in(X,[_|Y]):- in(X,Y).

belong_list(_, []).
belong_list(A, [X|R]):- belong(X,A), belong_list(A,R).

grupoid_table(og(_,_,H), H).

%%%% Plugin; generator for a class of structures (models)

og(D, generate, og(D0,X,Y)):-!, sort(D, D0), ordered_grupoid(D0, X, Y).

og0(D, _, _, domain, D).
og0(_, R, _, order, R).
og0(_, _, M, op, M).

lin(N, generate, lin(N,X)):-!, linear_grupoid(N, X).
grupoid(D, generate, grupoid(D,X)):-!, grupoidx(D, X).
int(K, generate,  int(1, K)):-!.
int(K,J, generate,int(K, J)).

generate(og(D),og(D0,X,Y)):-!, sort(D, D0), ordered_grupoid(D0, X, Y).
generate(lin(N),lin(N,X)):-!, linear_grupoid(N, X).
generate(grupoid(D),grupoid(D,X)):-!, grupoidx(D, X).
generate(int(K),  int(1, K)):-!.
generate(int(K,J),int(K, J)).

perm_group(X, Y):- permutations(X, PermX), maplist(zip(X), PermX, Y).

% bi-directional list comparison: less-than-or-equal:
% ?- leq_int_seq(3,X,Y).
% X = [1, 1, 1], Y = [1, 1, 1] ; ....
% ?- linear_grupoid(2,X).
% X = [ ((2, 1), 1), ((2, 2), 1), ((1, 1), 1), ((1, 2), 1)] ; ...

linear_grupoid(N,X):-up_sequence(N,Y), list2grupoid(N,Y,X).

leq_int_seq(N,X,Y):-leq_int_seq(N,N,X,Y).

leq_int_seq(_,0,[],[]).
leq_int_seq(N,J,[X|Y],[Z|U]):- J>0,
	J1 is J-1,
	leq_int_seq(N,J1,Y,U),
	between(1,N,X),
	between(X,N,Z),
	lower_bound(N,Z,U),
	lower_bound(N,X,Y).

% ?- up_sequence(3,X).
% X = [[1, 1, 1], [1, 1, 1], [1, 1, 1]] ; ....

up_sequence(N,X):- up_sequence(N,N,X).

up_sequence(_,0,[]).
up_sequence(N,J,[X|Y]):- J>0, J1 is J-1,
	up_sequence(N,J1,Y),
	up_sequence1(N,X,Y).

up_sequence1(N,X,[]):-int_up_seq(N,N,X).
up_sequence1(N,X,[Y|_]):-leq_int_seq(N,X,Y).

int_up_seq(_,0,[]).
int_up_seq(N,J,[X|Y]):- J>0, J1 is J-1,
	int_up_seq(N,J1,Y),
	lower_bound(N, X,Y).

lower_bound(N,Z,[]):- between(1,N,Z).
lower_bound(_,Z,[H|_]):-between(1,H,Z).

list2map(N,X,Y):- list2map(N,1,X,Y).

list2map(_,_,[],[]).
list2map(J,K,[X|Y],[((J,K),X)|Z]):-
	K1 is K+1,
	list2map(J,K1,Y,Z).

list2grupoid(_,[],[]).
list2grupoid(N,[X|Y],Z):-
	list2map(N,X,X1),
	N1 is N-1,
	list2grupoid(N1,Y,Y1),
	lists:append(X1,Y1,Z).


%%%%% eval_grupoid %%%%%
eval_grupoid(M,A*B,V):-!,
	eval_grupoid(M,A,A1),
	eval_grupoid(M,B,B1),
	in(((A1,B1),V), M).
eval_grupoid(_,A,A).

%%%%%%%%%%%%% set of integers %%%%%%
setofint(0,_,_,[]).
setofint(J,K,L,[K|R]):- J>0, K=<L,
	J1 is J-1,
	K1 is K+1,
	setofint(J1,K1,L,R).
setofint(J,K,L,S):-J>0, K=<L,
	K1 is K+1,
	setofint(J,K1,L,S).

%%%% rich alegbra %%%%%
rich_algebra(X,Alg):- powerset(X,P), length(X,N), N1 is N-1,
	rich_algebra(N1,P,Alg).

rich_algebra(1,_,_):-!.
rich_algebra(K,P,Alg):- in(S, P), length(S,K), subalgebra(S,Alg),
	K1 is K-1, rich_algebra(K1,P,Alg).

%%%%% generator of ordered grupoid %%%%
ordered_grupoid(X,E,G):-
	posets_in_fan(X,Ps),
	member(E0, Ps),
	succlist_pairs(E0, E),
	grupoid(X,G).

succlist_pairs([],[]).
succlist_pairs([(_,[])|R], Ps):- succlist_pairs(R, Ps).
succlist_pairs([(A,L)|R], Ps):-
	succlist_pairs(A, L, Qs),
	succlist_pairs(R, P0s),
	append(Qs, P0s, Ps).

succlist_pairs(_,[],[]).
succlist_pairs(A,[X|Xs],[(X,A)|Ys]):- succlist_pairs(A,Xs,Ys).

%%%% misc
leq_poset(_,A,A).
leq_poset(E,A,B):- in((A,A1),  E), A\==A1, leq_poset(E,A1,B).

% for debugging;  usage:
show(lin(N,G), Vs, L, R):- belong_list(lin(N,G), Vs),
       eval_grupoid(G,L,L1),
       eval_grupoid(G,R,R1),
       format("~w ~w :  ~w~n", [Vs, L1, R1]),
       fail; true.

% permutations([1,2,3], X).
% X = [[3, 2, 1], [2, 3, 1], [2, 1, 3], [3, 1, 2], [1, 3, 2], [1, 2, 3]]

permutations(X, Y):- permutations(X, [[]], Y).

permutations([],X, X).
permutations([X|Xs],Z0, Z):- maplist(perm_insert(X), Z0, Z1),
	append(Z1, Z2),
	permutations(Xs, Z2, Z).

perm_insert(X, [], [[X]]).
perm_insert(X, [Y|Ys], [[X, Y|Ys]|Z]):-
	perm_insert(X, Ys, Z0),
	maplist(cons(Y), Z0, Z).

% non deterministic permutation generator
perm_nd(X, Y) :- perm_nd(X, [], Y).

perm_nd([], X, X).
perm_nd([X|Xs], Y0, Y):- insert_nd(X, Y0, Y1), perm_nd(Xs, Y1, Y).

insert_nd(X, Y, [X|Y]).
insert_nd(X,[Y|Ys], [Y|Zs]):- insert_nd(X, Ys, Zs).

diagonal([]).
diagonal([(X,X)|D]):-diagonal(D),!.

make_id([],[]).
make_id([X|Y],[(X,X)|D]):- make_id(Y,D).

field([],[]).
field([(X,Y)|R],[X,Y|F]):-field(R,F).

compose(X,Y,Z):-compose(X,Y,Z,[]).

compose([],_,X,X).
compose([(X,Y)|Z],U,V,W):- compose(X,Y,U,V,R), compose(Z,U,R,W).

compose(_,_,[],X,X).
compose(X,Y,[(Y,Z)|U],[(X,Z)|V],W):-!, compose(X,Y, U,V,W).
compose(X,Y,[_|U],V,W):- compose(X,Y, U,V,W).

%% transitive closure X of relation R

trcl(R,X):- trcl(R,R,X).

trcl(R,X,Y):- compose(R,X,RX), union(RX,X,Z), trcl(R,X,Z,Y).

trcl(_,X,Y,X):- subset(Y,X),!.
trcl(R,_,Y,Z):- trcl(R,Y,Z).

trcl_sum(X,Y,Z):-compose(X,Y,Z1),compose(Y,X,Z2),
	union(Z1,Z2,Z3),
	trcl(Z3,Z4),
	union([X,Y,Z4],Z).

%%%% test connectivity of undirectied graph (X,E).

connected(X,E):-makediscrete(X,Y),
	union_find(E,Y,Class),Class=[_].

makediscrete([],[]).
makediscrete([X|R],[[X]|R1]):-makediscrete(R,R1).

%%%%  G is a grupoid over X.
grupoid(X,G):-
	cartesian(X,X,XX),
	maps(XX,X,XXX),
	in(G, XXX).

grupoidx(X,G):-  %%  backtracking version
	cartesian(X,X,XX),
	map(XX,X,G).

map([],_,[]).
map([X|R],Y,[(X,A)|S]):- in( A,  Y), map(R,Y,S).

%%
maps(Xs, Ys, Fs):-  foldl(maps(Ys), Xs, [[]], Fs).

maps(Ys, X,  F0s, Fs):- foldl(maps(F0s, X), Ys, Fs, []).

maps([], _, _, Gs, Gs).
maps([F|Fs], X,Y, [[(X,Y)|F]|G0s], Gs):- maps(Fs, X, Y, G0s, Gs).


%%%% a set X is closed
%%%% with respect to an binary operation Alg.

subalgebra(X,Alg):- in(Y, X), in(Z, X), in(((Y,Z),U), Alg), \+ in(U, X), !, fail.
subalgebra(_,_).

%%%%% a binary operation M is associative over a domain D.
associative(D,M):-assign([X,Y,Z],D),
	member(((Y,Z),YZ),M),
	member(((X,YZ),XYZ),M),
	member(((X,Y),XY),M),
	member(((XY,Z),XYZ),M),
	!,
	fail.
associative(_,_).

assign([],_).
assign([X|R],D):- in(X , D), assign(R,D).

%%  Y is the closure of X
%%  with respect to a binary operation G
binary_closed(G,X,Y):- bincl(G,X,Z), subset(Y,Z).

bincl(G,X,Y):-binresult(G,X,X,Z), bincl1(G,X,Z,Y).

bincl1(_,X,Y,X):- subset(Y,X),!.
bincl1(G,X,Y,Z):- union(X,Y,U), bincl(G,U,Z).

binresult(_,[],_,[]).
binresult(G,[X|Y],Z,U):-
	binresult1(G,X,Z,V),
	binresult(G,Y,Z,W),
	union(V,W,U).

binresult1(_,_,[],[]).
binresult1(G,X,[Y|Z],[U|V]):- in(((X,Y),U),  G), !,  binresult1(G,X,Z,V).

%%%%%% predicates over relations
partial_order(R):-
	anti_symmetric(R),
	transitive(R),
	refexive(R).

reflexsive(R):-field(R,F),
	reflexive(F,R).

reflexive([],_).
reflexive([X|Y],R):- in((X,X), R), reflexive(Y,R).

anti_symmetric(R):- member((X,Y),R),
	X\==Y,
	member((Y,X),R),
	!,
	fail.
anti_symmetric(_).

symmetric(R):- in((X,Y), R),  \+ in((Y,X), R), !, fail.
symmetric(_).

transitive(R):- compose(R,R,RR), subset(RR,R).

%%%%%% the set of DAGS and Posets %%%%
% ?- dags([a,b,c], Dags).
% ?- posets([a,b,c,d,e,f,g], X).
% ?- posets_in_fan([a,b,c,d,e,f,g], X).

%%%%%% top level call

dags(Ns, Dags):- dags((peak, append), Ns, Dags).

posets(Ns, Ps):- sort(Ns, Ms), dags((anti_chain, append), Ms, Ps).

posets_in_fan(Ns, Dags):- dags((fan, union), Ns, Dags).

%%%%%%
peak(X, Dag, S, [(X,S)|Dag]).

anti_chain(X, Poset, Chain, [(X, Chain)|Poset]):- anti_chain(Chain, Poset).

anti_chain([],_).
anti_chain([N|Ns], Poset):- anti_chain(Ns, Poset),
	maplist(incomparable(N,Poset), Ns).

incomparable(N, Poset, N0):- N @> N0, !, \+reach(N0,Poset,N).
incomparable(N, Poset, N0):- \+reach(N, Poset, N0).

reach(N,_,N):-!.
reach(N0, Poset,N):- memberchk((N,L), Poset),
	member(X, L),
	reach(N0, Poset, X).

fan(X, Poset, S, [(X,S0)|Poset]):-
 	maplist(fan(Poset), S, Ls),
 	union([S|Ls], S0).

fan(P, X, L):- memberchk((X,L), P).

%%%%%% dags: universal predicate  %%%%%%

dags(F, Nodes, Dags):- once(dags(F, Nodes, [[]], [[]], _, Dags)).

dags(_, [], X, Y, X, Y).
dags(F, [N|Ns], Succs0, Dags0, Succs, Dags):-
	new_node(F, N, Succs0, Dags0, Succs1, Dags1),
        dags(F, Ns, Succs1, Dags1, Succs, Dags).

new_node(F, N, Succs, Dags, Succs1, Dags1):-
	new_node(F, N, Succs, Dags, Dags1),
	extend_succs(N, Succs, Succs1).

new_node(F, N, Succs, Dags0, Dags):-
	maplist(new_peak(F, N, Succs), Dags0, Dags1),
	F = (_, A),  % append or union
	call(A, Dags1, Dags).

new_peak(_, _, [], _, []).
new_peak(F, X, [L|Ls], Dag, [Dag0|Dags]):-
	F = (M, _),  % method to add a node
	call(M, X, Dag, L, Dag0),
	!,
	new_peak(F, X, Ls, Dag, Dags).
new_peak(F, X, [_|Ls], Dag, Dags):-
	new_peak(F, X, Ls, Dag, Dags).

extend_succs(_, [], []).
extend_succs(X, [S|R0], [S, [X|S]|R]):- extend_succs(X, R0, R).
