:- module(poly, [bits/3, poly_in/2, poly_out/2
				]).

% Euler - Mersenne prime number  2147483647
:- use_module(pac(basic)).
:- use_module(util(meta2)).
:- use_module(pac('expand-pac')).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).
:- style_check(-singleton).

% gcd by Euclidean method of mutual division.
% greatest common diviser
% gcd(121,33,X) => X = 11
% ?- poly:gcd(123, 33, X).
% ?- poly:gcd(3, 10, X).
% ?- poly:gcd(6, 8, X).
% ?- poly:gcd(12, 8, X).
% ?- X is  gcd(10, 10).
% ?- 10^10 =:= gcd(10^10, 10^10).

gcd(X, Y, Z):- Z is gcd(X, Y).

% ?- poly:lcm(8, 12, X).
% ?- poly:lcm(-3, 4, X).
% ?- poly:lcm(6, -4, X).
% ?- poly:lcm(6, -4, X).

lcm(X, Y, Z):- Z is X*Y // gcd(X, Y).

% ?- poly:lcm(6, -4, X, A, B).
%@ X = -12,
%@ A = -2,
%@ B = 3.
lcm(X, Y, Z, U, V):- Z is X*Y // gcd(X, Y), U is Z//X, V is Z//Y.

% ?- divmod(-10, 3, X, Y).
% ?- divmod(-10, -3, X, Y).

% TODO
% euclid(X, Y, Q, R).

% %  given x >= y => (a, b), s.t a*x +b*y= gcd(x, y))
% % nat_gcd(x, y, a, b, d) <=> a*x +b*y= d (= gcd(x, y))
% % ?- nat_gcd(12, 8, A, B, D).
% % A = 1,
% % B = -1,
% % D = 4.
% %
% nat_gcd(X, Y, A, B, D):- X < Y, !, nat_gcd_(Y, X, B, A, D).
% nat_gcd(X, Y, A, B, D):- nat_gcd_(X, Y, A, B, D).

% nat_gcd_(X, Y, A, B, D):- Z is X mod Y,
% 	(Z==0 -> A = 0, B = 1, D = Y
% 	; nat_gcd_(Y, Z, K, L, D), A0 is X//Y,
% 	    A = L, B is K- L*A0
% 	).

next_prime(X, N, P):-  member(A, X), 0 =:= N rem A, !,
	N0 is N + 1,
	next_prime(X, N0, P).
next_prime(_, N, N).

% ?- trace, poly:make_poly_homogeneous(t, [1*[a^1]], PH).
% ?- poly:make_poly_homogeneous(t, [1*[a^1], 1*[b^1]], PH).
% ?- poly:make_poly_homogeneous(t, [1*[a^2], 1*[b^1]], PH).

make_poly_homogeneous(Hvar, P,  PH):-
	maplist(pred([C*M, C*M0, T, D]:- mono_tail(M, M0, T, D)),
	       P, PH, Ts, Ds),
	list_max(Ds, N),
	maplist(pred([N, Hvar],
				 ([N, []]
					&
				 ([J, [Hvar^K]]:- K is N-J))),
		Ds, Ts).
%
make_poly_non_homogeneous(SW, X, Y):-
	maplist(pred([I*Vs, I*Us]:-
		    foldl(pred([0^_, U, U]  & [A, [A|U], U]), Vs, Us, [])),
		     X, Y0),
	sort_poly(SW, Y0, Y1),
	merge_poly(Y1, Y2),
	canonical_form(Y2, Y).

% ?- poly: mono_tail([a^1, b^1], X, T, D).
mono_tail(M, X, T, D):- mono_tail(M, X, T, 0, D).
%
mono_tail([], U, U, D, D).
mono_tail([X^I|R], [X^I|U], T, D0, D):-
	D1 is D0 + I,
	mono_tail(R, U, T, D1, D).

% ?- poly:list_max([1,2,3], X).
list_max(L,  M):- list_max(L, 0, M).

%
list_max([], M, M).
list_max([I|R], M, M0):- I> M, !, list_max(R, I, M0).
list_max([_|R], M, M0):- list_max(R, M, M0).

% ?-  poly:int_div(-13, -3, Q, R).
%@ Q = 4,
%@ R = -1.

int_div(X, D, Q, R):- divmod(X, D, Q, R).

%
is_sorted_spoly([], _).
is_sorted_spoly([_],_).
is_sorted_spoly([P, Q|Z], F):-
	P = [_*X|_]-[_*Y|_],
	Q = [_*U|_]-[_*V|_],
	mono_lcm(X,Y,XY),
	mono_lcm(U,V,UV),
	call(F, C, XY, UV),
	C\== (>),
	!,
	is_sorted_spoly([Q|Z], F).
is_sorted_spoly([P, Q|_], _):-
	P = [_*X|_]-[_*Y|_],
	Q = [_*U|_]-[_*V|_],
	mono_lcm(X,Y,XY),
	mono_lcm(U,V,UV),
	writeln(not_ordered(XY=X-Y, UV=U-V)).

% ?- poly:sum_list_abs_two_halves([-1,2,-3,4,-5,6], S1, S2).
%@ S1 = 6,
%@ S2 = 15.

sum_list_abs_two_halves(L, Sum1, Sum2):-
	length(L, N),
	N0 is N//2,
	sum_list_abs_count(N0, L, L0, 0, Sum1),
	sum_list_abs(L0, 0, Sum2).

% Elegant way for sums  of abs value of integers in the list

% sum_list_abs_two_halves(L, Sum1, Sum2):- maplist(abs, L, L0),
% 	length(L, N),
% 	N0 is N//2,
% 	length(Snd_half, N0),
% 	append(Fst_half, Snd_half, L0),
% 	sum_list(Fst_half, Sum1),
% 	sum_list(Snd_half, Sum2).

% ?- poly:sum_list_abs_count(3, [-1, -3, -5, -6], L, 0, S).
%@ L = [-6],
%@ S = 9.

sum_list_abs_count(0, L, L, X, X):-!.
sum_list_abs_count(N, [A|R], L, X, Y):- X0 is abs(A) + X,
	N0 is N-1,
	sum_list_abs_count(N0, R, L, X0, Y).

% ?- poly:sum_list_abs([-1, -3, -5, -6], 0, S).
%@ S = 15.
sum_list_abs([], X, X).
sum_list_abs([A|R], X, Y):- X0 is abs(A)+ X,
	sum_list_abs(R, X0, Y).

% ?- poly:remove_content_poly([-6*_, 4*_, -8*_], Cs).
% ?- poly:remove_content_poly([], Cs).
% ?- poly:remove_content_poly([1*_], Cs).

remove_content_poly(_, [], []).
remove_content_poly(BL, [C*M|P], R):- 0 =:= (  C>>BL), !,
	once(remove_content_poly([C*M|P], R)).
remove_content_poly(_, P, P).

%
remove_content_poly(L,  L0):- maplist(pred([C*M, C]), L, Cs),
	once(remove_content(Cs, _, Ds)),
	maplist(pred([_*M, C, C*M]), L, Ds, L0).

% ?- poly:remove_content([-6, 4, -8], G, Cs).
%@ G = 2,
%@ Cs = [-3, 2, -4].
% ?- poly:remove_content([1, -6, 4, -8], G, Cs).
%@ Cs = [1, -6, 4, -8] .
% ?- poly:remove_content([-1, -6, -4, -8], G, Cs).
%@ G = 1,
%@ Cs = [-1, -6, -4, -8] .

remove_content([], _, []).
remove_content(L, G, Cs):-
	sum_list_abs_two_halves(L, Sum1, Sum2),
	G1 is gcd(Sum1, Sum2),
	maplist(pred(G1, [C, Q, R]:- divmod(C, G1, Q, R)),
		L, Qs, Rs),
	gcd_list(Rs, G2),
	G is gcd(G1, G2),
	K is G1//G,
	maplist(pred([K,G], [Q, R, C]:- C is Q*K + R//G),
		Qs, Rs, Cs).


% ?- poly:gcd_list([6, 4, 8], R).
gcd_list([A|R], G):- gcd_list(R, A, G).

gcd_list([], X, X).
gcd_list([A|B], X, Y):-   X0 is gcd(A, X),
	gcd_list(B, X0, Y).

%
% sum_abs_half_half(0, L, B, B, A):- sum_abs_half_half(L, 0, A).
% sum_abs_half_half(N, [I|Is], B0, B, A):-  B1 is B0 + abs(I),
% 	N0 is N -1,
% 	abs_sum_half_half(N0, Is, B1, B, A).

%
% sum_list_abs([], S, S).
% sum_list_abs[A|R], S0, S):- S1 is S0 + abs(A),
% 	sum_list_abs(R, S1, S).

% ?- poly: ( poly_in(a, Y),  poly_out(Y, Z)).
% ?- poly: ( poly_in(a^2;b, Y), sort_polyset(total, Y, Y1),  poly_out([], Y1, Z)).
% ?- poly:exp_poly(a^2, P).
% ?- poly: ( poly_in(a^2;b, Y), sort_polyset(rev(total), Y, Y1),  poly_out([], Y1, Z)).

poly_in --> flatten(;),
	maplist(exp_poly),
	remove([]).

%
poly_in([], []) --> poly_in.

poly_in(Order, Assoc) -->
	pred([Order, Assoc], [X, Y]:- preprocess(X, Y, Order, Assoc)),
	poly_in.

%
poly_out	--> poly_out([]).

poly_out(Assoc) -->
	maplist(pred(Assoc, [X, Y]:- postprocess(X, Y, Assoc))),
	maplist(poly_exp).

%
print_polyset(X):- maplist(writeln, X).

% ?- poly:s_poly([1*[x^1], 1*[a^1]],  [1*[y^1], 1*[b^1]], S, total).
%@ S = [1*[y^1, a^1], -1*[x^1, b^1]] .

s_poly([C0*V0|P0], [C1*V1|P1], S, F):-
      mono_lcm(V0, V1, _, U0, U1),
      mul_mono_poly(C1*U0, P0, Q0),
      mul_mono_poly(C0*U1, P1, Q1),
      subtract_poly_poly(Q0, Q1, S0, F),
      canonical_form(S0, S).

% ?- poly:s_poly_z([1*[x^1], 1*[a^1]],  [1*[y^1], 1*[b^1]], S, total).
%@ S = [1*[y^1, a^1], -1*[x^1, b^1]] .
% ?- poly:s_poly_z([4*[x^1], 1*[a^1]],  [6*[y^1], 1*[b^1]], S, total).
%@ S = [3*[y^1, a^1], -2*[x^1, b^1]] .

s_poly_z([M0|P0], [M1|P1], S, F):-
      mono_lcm_z(M0, M1, _, U0, U1),
      mul_mono_poly(U0, P0, Q0),
      mul_mono_poly(U1, P1, Q1),
      subtract_poly_poly(Q0, Q1, S, F).

% [2013/09/23]
% some tiny
canonical_form([]).
canonical_form([1*_|_]).

%
canonical_form([],[]).
canonical_form([1*A|X], [1*A|X]):- !.
canonical_form([C*M|Ps], [1*M|Qs]):- C0 is 1 rdiv C,
	maplist(pred(C0, [D*M, D0*M]:- D0 is C0*D), Ps, Qs).

% ?- poly:join_with_semicolon([a,b], R).
%@ R = (a;b).
% ?- poly:join_with_semicolon([0,0], R).
%@ R = 0.

join_with_semicolon(X, Y):- remove(0, X, X0), join_with_semicolon_(X0, Y).

join_with_semicolon_([], 0):-!.
join_with_semicolon_([X], X):-!.
join_with_semicolon_([X, Y], (X;Y)):-!.
join_with_semicolon_([X, Y|Z], (X;U)):- join_with_semicolon_([Y|Z], U).

%
zero_poly([]).
unit_poly([1*[]]).
constant_poly([_*[]]).

% ?- poly:exp_poly((a-b)^3, X), poly:poly_exp(X, Y).
% ?- poly:(poly_in((a-1)^100, X), poly_out(X, Y)).
% ?- poly:(poly_in((a-1)^3, X), sort_poly(total, X, Y0), poly_out(Y0, Y)).
% ?- poly:(poly_in((a-b)^3, X), sort_poly(total, X, Y0), poly_out(Y0, Y)).
% ?- poly:poly_exp([1*[a],1*[b],1*[c]], X).
% ?- poly:poly_exp([1*[a],1*[b],1*[c], -2*[]], X).
% ?- poly:poly_exp([(-3)*[x^1, y^3, z^2], (-3)*[x^1, y^3, z^2]], R).
% ?- A is -1 rdiv 3, poly:poly_exp([(-3)*[x^1, y^3], A*[x^1, y^2, z^1]], R).

exp_rat(T, R):- rat(T, R0), rat_rat(R0, R).

rat_poly(r(X,[1*[]]), X):-!.
rat_poly(r(X,[C*[]]), Y):- C\==0, C0 is 1 rdiv C,
	mul_scalar_poly(C0, X, Y).

%
exp_poly(L=R, P):- !, exp_poly(L-R, P).
exp_poly([], []):- !.
exp_poly(T, P):- exp_rat(T, R), rat_poly(R, P).

% ?- poly:exp_polyset(a;b;c, X).
%@ X = [[1*[a^1]], [1*[b^1]], [1*[c^1]]] .
exp_polyset --> exp_polyset(lexical).

exp_polyset(Ord)  --> flatten(;),
	maplist(exp_poly),
	maplist(canonical_form),
	remove([]),
	sort_polyset(Ord).

%
rat_rat(r(P, Q), r(P0, Q0)):- poly(P, P0, lexical),
	poly(Q, Q0, lexical).

%
poly_exp([], 0):-!.
poly_exp([A|R], T) :-  mono_exp(A, A0),
	slim_mono_exp(A0, A1),
	foldl(pred(([M, U, V]:- mono_exp(M, M0),
			poly_exp(M0, U, V))),
	  R, A1, T).

%
poly_exp(0*_, U, U).
poly_exp(1*M, U, U + M).
poly_exp(-1*M, U, U - M).
poly_exp(C*M, U, U + C0*M):- C>0, rdiv_slash(C, C0).
poly_exp(C*M, U, U - C1*M):- C0 is -C,  rdiv_slash(C0, C1).
poly_exp(0, U, U).
poly_exp(C, U, V):- number(C), !,
	(	C> 0
	->	rdiv_slash(C, C0),
		V =  U + C0
	;	C0 is -C,
		rdiv_slash(C0, C1),
		V = U -C1
	).
poly_exp(C, U, U + C).

% ?- poly:mono_exp((-3)*[x^1, y^3, z^2], R).
%@ R = -3* (x*y^3*z^2) .

mono_exp(C*M,  E):- maplist(pred([X^1, X] & [X, X]), M, M0),
	mono_exp_fold(C*M0, E).

%
mono_exp_fold(C*[], C).
mono_exp_fold(C*[A|R], C*T) :- foldl(pred([U, X, X*U]), R, A, T).

%
slim_mono_exp(1*M, M).
slim_mono_exp(M, M).
%
rdiv_slash(rdiv(X,Y), X/Y).
rdiv_slash(X, X).

%%%%
rat_exp(R, T):- rat_poly(R, P), !, poly_exp(P, T).
rat_exp(r(X, Y), /(X0, Y0)):- poly_exp(X, X0), poly_exp(Y, Y0).

% ?- poly: rat_pow( -1 rdiv 3, -3, R).
rat_pow(X, J, Q) :- J >= 0, !, rat_pow(X, J, 1, Q).
rat_pow(X, J, Q) :- J0 is -J, rat_pow(X, J0, 1, Q0),
	Q is 1 rdiv Q0.

rat_pow(_, 0, P, P):-!.
rat_pow(X, J, P, Q):- J0 is J-1, P0 is P*X,
	rat_pow(X, J0, P0, Q).

%
exp_exp --> exp_rat, rat_exp.


%  ?- poly:poly([1*[x^1]] - [1*[y^1]], P, lexical).


basic_poly(0,[]):-! .
basic_poly(X,[X*[]]):-rational(X),! .
basic_poly(X,[1*[X^1]]):-! .


% ?- poly:poly(3*x*y, X, total).
% ?- poly:poly(3*x*y, X).
% ?- poly:poly((3*x*y)^2, X).
% ?- poly:poly((x-y+z)^2, X).

poly(X, Y) :- poly(X, Y, lexical).

poly(0, [], _).
poly(X, X, _):- listp(X).
poly(X, [X*[]], _):- number(X).
poly(X - Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
	subtract_poly_poly(X0, Y0, Z, F).
poly(X + Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
	add_poly_poly(X0, Y0, Z, F).
poly(X * Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
	mul_poly_poly(X0, Y0, Z, F).
poly(X ^ N, Z, F):- poly(X, X0, F), power_poly(N, X0, Z, F).
poly('$Var'(X), [1*[X^1]], _).
poly(X, [1*[X^1]], _).

% ?- poly:poly_stack([1+(2*3)], [], X, lexical).
% ?- poly:poly_stack([(1+2)*3], [], X, lexical).
% ?- poly:poly_stack([a -b +c ], [], X, lexical).

% ?- once(time(poly:poly_stack([(a -b +c)^50 ], [], X, lexical))), fail; true.
%@ % 4,350,956 inferences, 0.785 CPU in 0.848 seconds (93% CPU, 5542196 Lips)
%@ true.
% ?- once(time(poly:non_modular_poly( (a - b + c)^50, R, lexical))), fail; true.
% ?- poly:poly_stack([(a -b +c)^10 ], [], X, lexical).

poly_stack([],  D, D, _).
poly_stack(['$OPR'(F/N)|R],  D,  E, SW):- !, apply_stack(N, F, D, D0, SW),
	poly_stack(R, D0, E, SW).
poly_stack([X|R],  D, E, SW):-
	poly_stack(X, R, D, R0, D0),
	poly_stack(R0, D0, E, SW).

%
poly_stack(X+Y, R, D, [X, Y, '$OPR'((+)/2)|R], D).
poly_stack(X-Y, R, D, [X, Y, '$OPR'((-)/2)|R], D).
poly_stack(X*Y, R, D, [X, Y, '$OPR'((*)/2)|R], D).
poly_stack(+X, R, D, [X|R], D).
poly_stack(-X, R, D, [X, '$OPR'((-)/1)|R], D).
poly_stack(X^N, R, D, [X, '$OPR'(pow(N)/1)|R], D).
poly_stack('$Var'(N), R, D, R, [[V]|D]):- !, make_var(N, V).
poly_stack(I, R, D, R, [[C]|D]):- integer(I), !, make_constant(I, C).
poly_stack(X, R, D, R, [[V]|D]):- make_var(X, V).

%
make_var(X, 1*[X^1]).

make_constant(I, I*[]).

%
apply_stack(2, F, [A, B|D], [C|D], SW):- !, apply_poly(F, B, A, C, SW).
apply_stack(1, F, [A|D], [C|D], SW):- apply_poly(F, A, C, SW).

%
apply_poly(+, A, B, C, SW):- add_poly_poly(A, B, C, SW).
apply_poly(*, A, B, C, SW):- mul_poly_poly(A, B, C, SW).
apply_poly(-, A, B, C, SW):- subtract_poly_poly(A, B, C, SW).

%
apply_poly(-, A, C, _):- prefix_minus_poly(A, C).
apply_poly(pow(N), A, C, SW):- power_poly(N, A, C, SW).

% poly(0, [], _).
% poly(X, X, _):- listp(X).
% poly(X, [X*[]], _):- number(X).
% poly(X - Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
% 	subtract_poly_poly(X0, Y0, Z, F).
% poly(X + Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
% 	add_poly_poly(X0, Y0, Z, F).
% poly(X * Y, Z, F):- poly(X, X0, F), poly(Y, Y0, F),
% 	mul_poly_poly(X0, Y0, Z, F).
% poly(X ^ N, Z, F):- poly(X, X0, F), power_poly(N, X0, Z, F).
% poly('$Var'(X), [1*[X^1]], _).
% poly(X, [1*[X^1]], _).

% ?- poly:modular_poly( 1, R,  lexical, 5).
% ?- poly:modular_poly( 2*3*7, R,  lexical, 5).
% ?- poly:modular_poly( (x+y)^3, R,  lexical, 5).
% ?- poly:modular_poly( (x+y)^100, R,  lexical, 5).
% ?- poly:modular_poly( (x+y)^100, R, total, 10).
% ?- poly:modular_poly( (x-y)^10, R, total, 3).
% ?- poly:poly(x-y+z, R, total), poly:poly(R^2, S, total).
% ?- poly:poly(x+y+z, R, total), poly:power_poly(R^100, S, total).
% ?- poly:poly(x+y+z, R, total), poly:power_poly(70, R, S, total).
% ?- poly:poly(x+y+z, R, total), poly:modular_poly(R^100, S, total, 3).
% ?- poly:poly(x+y+z, R, total), poly:modular_poly(R^200, S, total, 11).
% ?- poly:modular_poly(x^3, Y, total, 3).
% ?- time(poly:modular_poly( (x-y+z)^100, R, total, 3)).
% ?- time(poly:modular_poly( (x-y+z)^100, R, lexical, 2)).
% ?- time(poly:modular_poly( (x-y+z)^500, R, lexical, 2)).
% ?- time(poly:modular_poly( (5*x-7*y)^100, R, total, 3)).
% ?- listing(poly:modular_poly).


modular_poly(X-Y,C,Ord,G):-!,(modular_poly(X,A,Ord,G),modular_poly(Y,B,Ord,G)),subtract_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly(X+Y,C,Ord,G):-!,(modular_poly(X,A,Ord,G),modular_poly(Y,B,Ord,G)),add_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly(X*Y,C,Ord,G):-!,(modular_poly(X,A,Ord,G),modular_poly(Y,B,Ord,G)),mul_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly(X^N,C,Ord,G):-!,modular_poly(X,A,Ord,G),galois_power_poly(N,A,B,Ord,G),poly_galois(G,B,C) .
modular_poly(-X,A1,A2,A3):-!,modular_poly(X,A4,A2,A3),prefix_minus_poly(A4,A1) .
modular_poly('$Var'(I),[1*[I^1]],A1,A2):-! .
modular_poly(0,[],A1,A2):-! .
modular_poly(X,X,A1,A2):-is_list(X),! .
modular_poly(X,[Y*[]],A1,G):-(integer(X),galois(G,X,Y)),! .
modular_poly(X,[1*[X^1]],A1,A2):-! .


% ?- poly:modular_poly_calc(1, Y, lexical, 3).
% ?- poly:modular_poly_calc(5^10000, Y, lexical, 3).
% ?- poly:modular_poly_calc((a-b)^5, Y, lexical, 2).
% ?- poly:modular_poly_calc((a-b-c)^50, Y, lexical, 3).

modular_poly_calc(X, Y, Ord, Prim):- exp_poly(X, Poly),
	sort_poly(Ord, Poly, Poly0),
	poly_galois(Prim, Poly0, Poly1),
	modular_poly_list(Poly1, Poly2, Ord, Prim),
	poly_exp(Poly2, Y).

% ?- poly:poly_calc(1, Y, lexical).
% ?- poly:poly_calc(5^10000, Y, lexical).
% ?- poly:poly_calc((a-b)^5, Y, lexical).
% ?- poly:poly_calc((a-b-c)^50, Y, lexical).

poly_calc(X, Y, Ord):- non_modular_poly(X, Y0, Ord), poly_exp(Y0, Y).

% ?- poly:modular_poly_list([1,2], X, lexical, 3).
% ?- listing(modular_poly_list).

modular_poly_list([],[],A1,A2):-! .
modular_poly_list([X|Y],[X|Y],A1,A2):-! .
modular_poly_list(X-Y,C,Ord,G):-!,(modular_poly_list(X,A,Ord,G),modular_poly_list(Y,B,Ord,G)),subtract_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly_list(X+Y,C,Ord,G):-!,(modular_poly_list(X,A,Ord,G),modular_poly_list(Y,B,Ord,G)),add_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly_list(X*Y,C,Ord,G):-!,(modular_poly_list(X,A,Ord,G),modular_poly_list(Y,B,Ord,G)),mul_poly_poly(A,B,AB,Ord),poly_galois(G,AB,C) .
modular_poly_list(X^N,C,Ord,G):-!,modular_poly_list(X,A,Ord,G),galois_power_poly(N,A,B,Ord,G),poly_galois(G,B,C) .
modular_poly_list(-X,A1,A2,A3):-!,modular_poly_list(X,A4,A2,A3),prefix_minus_poly(A4,A1) .


% ?- poly:poly_galois(3, [4*[x^1], 5*[]], R).
poly_galois(G, B, C):-
	foldl(pred(G, (	[C*X, [D*X|U], U]:- galois(G, C, D), D\==0)
				&	[_, U, U]), B,  C, []).

% ?- time(poly:non_modular_poly( (5*x-7*y)^100, R, total)).
% ?- time(poly:non_modular_poly( (5*x-7*y)^100, R, lexical)).
% ?- time(poly:non_modular_poly( (-(x) + -y)^2, R, lexical)).


non_modular_poly(X-Y,C,Ord):-!,(non_modular_poly(X,A,Ord),non_modular_poly(Y,B,Ord)),subtract_poly_poly(A,B,C,Ord) .
non_modular_poly(X+Y,C,Ord):-!,(non_modular_poly(X,A,Ord),non_modular_poly(Y,B,Ord)),add_poly_poly(A,B,C,Ord) .
non_modular_poly(X*Y,C,Ord):-!,(non_modular_poly(X,A,Ord),non_modular_poly(Y,B,Ord)),mul_poly_poly(A,B,C,Ord) .
non_modular_poly(X^N,C,Ord):-!,non_modular_poly(X,A,Ord),power_poly(N,A,C,Ord) .
non_modular_poly(-X,A1,A2):-!,non_modular_poly(X,A3,A2),prefix_minus_poly(A3,A1) .
non_modular_poly(0,[],A1):-! .
non_modular_poly('$Var'(N),[1*[N^1]],A1):-! .
non_modular_poly(X,[X*[]],A1):-integer(X),! .
non_modular_poly(X,X,A1):-listp(X),! .
non_modular_poly(X,[1*[X^1]],A1):-! .


% ?-gblex:rat((1+x)/y, R).
%@ R = r((1*1+x*1)*1, 1*1*y).
% ?- poly:rat((a/b)+1, X).
%@ X = r(a*1*1+1* (1*b), 1*b*1).


rat(+X,A1):-!,rat(X,A1) .
rat(-X,A1):-!,rat(-1*X,A1) .
rat(X-Y,A1):-!,rat(X+ -1*Y,A1) .
rat(X+Y,r(X0*Y1+Y0*X1,X1*Y1)):-!,rat(X,r(X0,X1)),rat(Y,r(Y0,Y1)) .
rat(X*Y,r(X0*Y0,X1*Y1)):-!,rat(X,r(X0,X1)),rat(Y,r(Y0,Y1)) .
rat(X/Y,r(X0*Y1,X1*Y0)):-!,rat(X,r(X0,X1)),rat(Y,r(Y0,Y1)) .
rat(X^N,r(X0^N,X1^N)):-!,rat(X,r(X0,X1)) .
rat('$Var'(X),r([1*[X^1]],1)):-! .
rat(X,r(X,1)):-! .


indeterminates(X-Y, X0-Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X+Y, X0+Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X*Y, X0*Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X/Y, X0/Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X=Y, X0=Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X;Y, X0;Y0, A, B):-
	indeterminates(X, X0, A, A0),
	indeterminates(Y, Y0, A0, B).
indeterminates(X^Y, X0^Y, A, B):-
	indeterminates(X, X0, A, B).
indeterminates(-X, -X0, A, B):-
	indeterminates(X, X0, A, B).
indeterminates(X, X, A, A):- number(X).
indeterminates(X, '$Var'(N), [X-N|A], A).


% ?- poly:(preprocess(1+x, X, A), poly(X, Y), postprocess(Y, Z, A)).
% ?- poly:(preprocess(1+x, X, A), poly(X, Y), postprocess(Y, Z, A)).
% ?- poly:(preprocess((1+x)^100, X, A), poly(X, Y), postprocess(Y, Z, A)).
% ?- poly:preprocess(x, X, R).

preprocess(X, Y, Assoc):- once(preprocess(X, Y, [], Assoc)).

% ?-  poly:preprocess(x+y+a, X, [y,x], R).
%@ X = '$Var'(3)+'$Var'(2)+'$Var'(1),
%@ R = [a-1, x-3, y-2] .

preprocess(X, Y, Order, DefaultAssoc):-
	indeterminates(X, Y, B, []),
	sort(B, B0),
	merge_assoc(B0, DefaultAssoc),
	merge_order_to_assoc(Order, DefaultAssoc).

%
assoc_numbering(A):- length(A, N),
	numlist(1, N, L),
	zip_hyphen(_, L, A).

% ?- poly:number_vars(a+b, [a-1, b-2], Y).
%@ Y = '$Var'(1)+'$Var'(2) .
% ?- poly:number_vars(a+b+c, [a-1, b-2], Y).
%@ Y = '$Var'(1)+'$Var'(2)+'$Var'(c) .

number_vars(X, Assoc, Y):-
	indeterminates(X, Y, B, []),
	subtract_assoc(B, Assoc, B0),
	maplist(pred([X-X]), B0).    % pass through unknown new variabls.

% ?- poly:subtract_assoc([a-X, b-Y, a-Z, c-U], [a-1], R).
%@ X = Z, Z = 1,
%@ R = [b-Y, c-U] .
subtract_assoc(X, Y, Z):-
	foldl(pred(Y,	  (	[A, U, U]  :- memberchk(A, Y))
			& (	[A, [A|U], U]) ),
		X, Z, []).

%
postprocess(X, X, []).
postprocess(X, Y, Assoc):-
	maplist(pred(Assoc, [I*Vs, I*Us]:-
		    maplist(pred(Assoc, ( [P^J, Q^J]:- memberchk(Q-P, Assoc), !)
					&  [U, U] ),
				 Vs, Us)),
		     X, Y).

%
merge_assoc([], []).
merge_assoc([X-N|R], [X-N|A]):- merge_assoc(R, R0, X, N),
	merge_assoc(R0, A).
%
merge_assoc([X-N|R], R0, X, N):- !, merge_assoc(R, R0, X, N).
merge_assoc(R, R,  _, _).

%
zip_hyphen([], [], []).
zip_hyphen([A|B], [C|D], [A-C|R]):- zip_hyphen(B, D, R).

% ?- poly:merge_order_to_assoc([d, c, a], [a-X, b-Y,c-Z]).
%@ X = 4,
%@ Y = 1,
%@ Z = 3.

merge_order_to_assoc([], Assoc):- !, assoc_numbering(Assoc).
merge_order_to_assoc(L, Assoc):-  zip_hyphen(L, _, Bs),
	union(Assoc, Bs, Cs),
	assoc_numbering(Cs).

%
list_size(X, L):- list_size(X, 0, L).

list_size([], N, N).
list_size([_|R], N0, N):-  N1  is N0 + 1,
	list_size(R, N1, N).

% ?- poly:sort_mono([a^1, c^1, b^1, c^1], R).
%@ R = [c^2, b^1, a^1].

sort_mono(L, R) :- length(L, N),
	sort_mono(N, L, _, R).

%
sort_mono(2, [X1, X2|L], L, R) :- !, sort2(X1, X2, R).
sort_mono(1, [X|L], L, [X]) :- !.
sort_mono(0, L, L, []) :- !.
sort_mono(N, L1, L3, R) :-
	N1 is N // 2,
	plus(N1, N2, N),
	sort_mono(N1, L1, L2, R1),
	sort_mono(N2, L2, L3, R2),
	merge_mono_mono(R1, R2, R).

%
sort2(X^I, X^J, [X^K]):- !,  K is I+J.
sort2(X^I, Y^J, [X^I, Y^J]):- X @> Y, !.
sort2(U, V, [V, U]).

% ?- poly: dim_mono_term([a^1, b^2], X).
%@ X = 3.

dim_mono_term([], 0).
dim_mono_term([_^I|R], N):- dim_mono_term(R, N0), N is I+N0.

% ?- poly: compare_total_order(C, [b^1, a^1], [c^1]).
%@ C = (>).
% ?- poly: compare_total_order(C, [b^1, a^1], [b^1, a^1]).
%@ C = (=) .

compare_total_order(C, X, Y):- dim_mono_term(X, D0),
	dim_mono_term(Y, D1),
	compare(C, D0, D1),
	C \== (=),
	!.
compare_total_order(C, X, Y):- compare_mono_term(C, X, Y).

% ?- poly: compare_mono_term(C, [b^1, a^1], [b^1, a^1]).
%@ C = (=) .
compare_mono_term(C, [X^J|R], [X^K|S]):- !, compare(C0, J, K),
	(	C0 == (=)
	->	compare_mono_term(C, R, S)
	;	C = C0
	).
compare_mono_term( C , [X^_|_], [Y^_|_]):- !,
	(	X @> Y
	->	C = (>)
	;	C = (<)
	).
compare_mono_term( = , [], []).
compare_mono_term( < , [], _).
compare_mono_term( > , _, []).

%
compare_mono(lexical, C, _*M, _*N):- !, compare_mono_term(C, M, N).
compare_mono(total, C, _*M, _*N):- !, compare_total_order(C, M, N).
% compare_mono(rev(Order), C, X, Y):- compare_mono(Order, D, X, Y),
%  	( D == (<), C = (>);  D == (>), C = (<);  true ),
% 	!.

%
merge_mono([X^I|R], [X^J|S]):-  merge_mono(R, R0, X, I, J),
	merge_mono(R0, S).
merge_mono([], []).
%
merge_mono([X^K|R], R0, X, I, J):- !,  I0 is I + K,
	merge_mono(R, R0, X, I0, J).
merge_mono(R, R, _, I, I).
%
merge_mono_mono(X, [], X):- !.
merge_mono_mono([], X, X):- !.
merge_mono_mono([X^I|R], [X^J|S], [X^K|T]):- !, K is I + J,
	merge_mono_mono(R, S, T).
merge_mono_mono([X^I|R], [Y^J|S], [Y^J|T]):-  X @< Y, !,
	merge_mono_mono([X^I|R], S, T).
merge_mono_mono([U|R], B, [U|T]):-  merge_mono_mono(R, B, T).

% ?- poly:mul_mono_mono(3*[c^1, a^1], 2*[c^1, b^2], X).
%@ X = 6*[c^2, b^2, a^1].
mul_mono_mono(C*M, D*N, E*L):- merge_mono_mono(M, N, L), E is C*D.

% ?- gblex:mul_mono_poly(1*[c^1], [1*[b^1], 1*[a^1]], X).
mul_mono_poly(M, P, P0):- maplist(mul_mono_mono(M),P, P0).

% ?- poly:compare_poly(lexical, X, [1*[y^2, x^2]], [1*[y^1, x^2]]).
%@ X = (>).
% ?- poly:compare_poly(rev(lexical), X, [1*[y^2, x^2]], [1*[y^1, x^2]]).
%@ X = (<).
% ?- poly:compare_poly(total, X, [1*[y^2, x^2]],[1*[y^2, x^2]]).
%@ X = (=) .
% ?- poly:compare_poly(rev(total), X, [1*[y^2, x^1]],[1*[y^2, x^2]]).
%@ X = (>).
%
sort_polyset(F, X, Y):- predsort(compare_poly(F), X, Y).

%
compare_poly(_, =, [], []).
compare_poly(_, <, [], _).
compare_poly(_, >, _, []).
compare_poly(F, C, [M|_], [N|_]):- compare_mono(F, C, M, N),
	C\== (=).
compare_poly(_, <, _, _).

% ?- poly:merge_mono([c^1, a^1, a^1], X).
% ?- poly:merge_mono_mono([c^1, a^1], [d^1, b^2, a^1], X).
% ?- poly:sort_poly(total, [1*[y^1, x^1], 1*[x^3]], X).
% ?- poly:sort_poly(lexical, [1*[x^2], 1*[y^1]], X).
% ?- poly:sort_poly(total, [1*[y^1, x^1], 1*[x^3]], X).
%@ X = [1*[x^3], 1*[y^1, x^1]] .

sort_poly(F, L, R) :- length(L, N),
	sort_poly(N, L, _, R, F).

%
sort_poly(2, [X1, X2|L], L, R, F) :- !,
	sort2_poly(X1, X2, R, F).
sort_poly(1, [X|L], L, [X], _) :- !.
sort_poly(0, L, L, [], _) :- !.
sort_poly(N, L1, L3, R, F) :-
	N1 is N // 2,
	plus(N1, N2, N),
	sort_poly(N1, L1, L2, R1, F),
	sort_poly(N2, L2, L3, R2, F),
	merge_poly_poly(R1, R2, R, F).

%
sort2_poly(U, V, R, F):- compare_mono(F, A, U, V),
	(	A == (<),  R = [V, U]
	;	A == (>),  R = [U, V]
	;	U = C*M,
		V = D*_,
		E is C + D,
		R = [E*M]
	).


% ?- poly:merge_polyset_polyset([],[], [], total).
% ?- poly:merge_polyset_polyset([[2*[]]],[[1*[]]], X, total).
% ?- poly:merge_polyset_polyset([[2*[a^2]]],[[1*[b^1]]], X, total).
% ?- poly:merge_polyset_polyset([[2*[a^2]]],[[1*[b^1]]], X, lexical).
%  is this pred necessary ?
merge_polyset_polyset(X, Y, Z, F):- merge_polyset_polyset(X, Y, Z, [], F).

%
merge_polyset_polyset([], Y, Y, [], _).
merge_polyset_polyset(X, [], X, [], _).
merge_polyset_polyset([A|X], [A|Y], [A|Z], U, F):- !,
	merge_polyset_polyset(X, Y, Z, U, F).
merge_polyset_polyset([A|X], [B|Y], [A|Z], U, F):-
	compare_poly(F, C, A, B), C == (>), !,
	merge_polyset_polyset(X, [B|Y], Z, U, F).
merge_polyset_polyset([A|X], [B|Y], [B|Z], U, F):-
	merge_polyset_polyset([A|X], Y, Z, U, F).


% ?- poly:sort_poly([1*[a^1], 1*[b^1], 1*[a^1]], X, total).
%@ X = [1*[b^1], 2*[a^1]] .
% ?- poly:sort_poly([1*[a^2], 1*[b^1], 1*[a^1]], X, compare_total_order).
%@ X = [1*[a^2], 1*[b^1], 1*[a^1]] .
% ?- poly:sort_poly([1*[a^2], 1*[b^1], 1*[a^1]], X, compare_mono_term).
%@ X = [1*[b^1], 1*[a^2], 1*[a^1]] .
% ?- poly:sort_poly([1*[a^2], 1*[b^1], 1*[a^1]], X, compare_total_order).
%@ X = [1*[a^2], 1*[b^1], 1*[a^1]] .
% ?- poly:merge_poly([1*[b^1], 1*[b^1], 1*[a^1]], X).
% ?- gblex:merge_poly([1*[b^1], 1*[a^1]], X).
% ?- gblex:merge_poly([1*[b^1], (-1)*[b^1], 1*[a^1]], X).
%@ X = [1*[a^1]].

merge_poly([], []).
merge_poly([C*M|R], S):- merge_poly(R, R0, M, C, C0),
	(	C0\==0
	->	S = [C0*M|S0]
	;	S0 = S
	),
	merge_poly(R0, S0).
%
merge_poly([C*M|R], R0, M, D, E):- !, D0 is D+C,
	merge_poly(R, R0, M, D0, E).
merge_poly(R, R, _, C, C).

% ?- poly:merge_poly_poly([1*[b^1], 1*[a^1]],[1*[b^1], 1*[a^1]], X, lexical).
%@ X = [2*[b^1], 2*[a^1]] .

merge_poly_poly([], X, X, _).
merge_poly_poly(X, [], X, _).
merge_poly_poly([M|U], X, P, F):- merge_poly_poly(M, X, X0, P, Q, F),
	merge_poly_poly(U, X0, Q, F).

%
merge_poly_poly(M, [N|V], V0, P0, P, F):- compare_mono(F, A, M, N),
	merge_poly_poly(A, M, N, V, V0, P0, P, F).
merge_poly_poly(M, [], [], [M|P], P, _).

%
merge_poly_poly(=, I*X, J*_, V, V, [K*X|P], P, _):-  K is I+J, K\== 0, !.
merge_poly_poly(=, _  , _  , V, V, P, P, _):- !.
merge_poly_poly(>, M  , N  , V, [N|V], [M|P], P, _).
merge_poly_poly(_, M  , N  , V, V0, [N|P], Q, F):-
	merge_poly_poly(M, V, V0, P, Q, F).

% ?- poly:mul_mono_mono(1*[c^1, a^1], 2*[c^1, b^2], X).
%@ X = 2*[c^2, b^2, a^1].
prefix_minus_poly(X, Y):- maplist(pred([I*X, J*X]:- J is 0 - I), X, Y).

mul_scalar_poly(C, P, Q):- maplist(mul_scalar_mono(C), P, Q).

mul_scalar_mono(C, C0*M, C1*M) :- C1 is C0*C.

% ?- poly:div_mono_mono(-1*[b^2, c^1, a^3], 1*[b^1, a^1], D).
%@ D = -1*[b^1, c^1, a^2].
% ?- poly:div_mono_mono_term([b^2, c^1, a^3], [b^1, a^1]).

%
div_mono_mono_term(X, Y):- div_mono_mono_term(X, Y, _).
%
div_mono_mono_term(M, [], M):-!.
div_mono_mono_term([X^I|Ys], [X^J|Ys0], Zs):- !,
	( I == J
	->	div_mono_mono_term(Ys, Ys0, Zs)
	;	I> J,
		K is I-J,
		Zs=[X^K|Zs0],
		div_mono_mono_term(Ys, Ys0, Zs0)
	).
div_mono_mono_term([X^I|Ys], [X0^J|Ys0], [X^I|Zs]):- compare((>), X, X0),
	div_mono_mono_term(Ys, [X0^J|Ys0], Zs).

%
div_mono_mono(C*M, C0*M0, C1*M1):- div_mono_mono_term(M, M0, M1),
	C1 is C rdiv C0.


% ?- poly:div_poly_mono([-1*[b^2, c^1, a^3]], 1*[b^1, a^1], Q, D).
%@ Q = [-1*[b^1, c^1, a^2]],
%@ D = [].
% ?- poly:div_poly_mono([-1*[b^2, c^1, a^3], 1*[b-2]], 1*[b^1, a^1], Q, D).
%@ Q = [-1*[b^1, c^1, a^2]],
%@ D = [1*[b-2]].

div_poly_mono([], _, [], []).
div_poly_mono([H|P], M, [Q0|Qs], R):- div_mono_mono(H, M, Q0), !,
	div_poly_mono(P, M, Qs, R).
div_poly_mono([H|P], M, Q, [H|R]):- div_poly_mono(P, M, Q, R).


% ?- poly:merge_poly_poly([2*[a^1]], [1*[a^1]], X, total).
% ?- poly:add_poly_poly([2*[a^1]], [1*[a^1]], X, lexical).

add_poly_poly(X, Y, Z, F):- merge_poly_poly(X, Y, Z, F).

% ?- poly:subtract_poly_poly([3*[x^1]], [1*[y^1], 2*[x^1]], R, total).
%@ R = [-1*[y^1], 1*[x^1]] .
% ?- poly:subtract_poly_poly([1*[]], [1*[]], R, lexical).
%@ R = [] .
% more efficient then " minus  then  add ".  [2014/07/13]

subtract_poly_poly([], X, Y, _):- !, prefix_minus_poly(X, Y).
subtract_poly_poly(X, [], X, _):- !.
subtract_poly_poly([M|U], X, P, F):-
	subtract_poly_poly(M, X, X0, P, Q, F),
	subtract_poly_poly(U, X0, Q, F).

%
subtract_poly_poly(M, [N|V], V0, P0, P, F):- compare_mono(F, A, M, N),
	subtract_poly_poly(A, M, N, V, V0, P0, P, F).
subtract_poly_poly(M, [], [], [M|P], P, _).

%
subtract_poly_poly(=, I*X, J*_, V, V, [K*X|P], P, _):-  K is I-J, K\== 0, !.
subtract_poly_poly(=, _,   _, V, V, P, P, _):- !.
subtract_poly_poly(>, M,   N, V, [N|V], [M|P], P, _):-!.
subtract_poly_poly(_, M,   I*Y, V, V0, [J*Y|P], Q, F):-  J is 0 - I,
	subtract_poly_poly(M, V, V0, P, Q, F).


		/*****************************************************************
		*    Smart multiplication on polynomials with multi-variables    *
		*     reducing add (sort/merge) operations.                      *
		*****************************************************************/

% ?- poly: (mul_poly_poly([1*[b^1], 1*[a^1]],[1*[b^1], 1*[a^1]], X, total), mul_poly_poly(X, X, Y, total)).
%@ X = [1*[b^2], 2*[b^1, a^1], 1*[a^2]],
%@ Y = [1*[b^4], 4*[b^3, a^1], 6*[b^2, a^2], 4*[b^1, a^3], 1*[a^4]] .

mul_poly_poly([], _, [], _).
mul_poly_poly([A|As], Bs, X, Ord):-
	mul_mono_poly(A, Bs, U),
	mul_poly_poly(As, Bs, U, V, X, V, Ord).

%
mul_poly_poly([], _, U, U, X, X, _).
mul_poly_poly([A|As], Bs, U, V, X, Y, Ord):-
	mul_mono_poly_head(A, Bs, U, U0, X,  X0, Ord),
	mul_poly_poly(As, Bs, U0, V, X0, Y, Ord).

%
mul_mono_poly_head(_, [], U, U, X, X, _):-!.
mul_mono_poly_head(A, Bs, U, V, X, Y, Ord):-
	mul_mono_poly(A, Bs, [C|Cs]),
	sweep_higher_terms(C, U, U0, X, Y, Ord),
	merge_poly_poly(Cs, U0, V, Ord).

%
sweep_higher_terms(C, [R|Rs], Ps, X, Y, Ord):-
	compare_mono(Ord, A, C, R),
	(	A == (=)	->	add_coeff(C, R, CR),
						(  CR == [] -> Ps = Rs ; Ps = [CR|Rs] ),
						X = Y
	;	A == (>)	->	Ps = [C, R|Rs],
						X = Y
	;	X = [R|X0],
		sweep_higher_terms(C, Rs, Ps, X0, Y, Ord)
	).
sweep_higher_terms(C, [], [C], X, X, _).

%
add_coeff(C*M, D*_, E*M):- E is C+D, E\==0, !.
add_coeff(_, _, []).

% ?- poly:galois_power_poly(1, [1*[a^1]], R, total, 3).
% ?- poly:mul_poly_poly([1*[]], [1*[]], X, total).
% ?- poly:galois_power_poly(3, [1*[a^3]], R, total, 3).
%@ R = [1*[a^9]] .
% ?- poly:galois_power_poly(3, [1*[a^3], 1*[a^2]], R, total, 3).
% ?- poly:galois_power_poly(3, [1*[a^3], 1*[a^2]], R, total, 3).

galois_power_poly(N, X, Y, SW, G):- bits(N, [], Bs),
	unit_poly(U),
	galois_power_poly(Bs, X, U, Y, SW, G).

%
galois_power_poly([], _, X, X, _, _):-!.
galois_power_poly([0|Bs], A, X, Y, SW, G):- !,
	mul_poly_poly(X, X, X2, SW),
	galois_poly(G, X2, GX2),
	galois_power_poly(Bs, A, GX2, Y, SW, G).
galois_power_poly([_|Bs], A, X, Y, SW, G):-
	mul_poly_poly(X, X, X2, SW),
	mul_poly_poly(A, X2, AX2, SW),
	galois_poly(G, AX2, Y0),
	galois_power_poly(Bs, A, Y0, Y, SW, G).


%   ?- poly:galois_poly(3, [5*_, 6*_, 7*_], R).
galois_poly(G, P, Q):- rem_poly(P, Q, G).

%  ?- poly:rem_poly([5*_, 7*_], R, 3).
%@ R = [2*_G11, 1*_G12].

rem_poly([A*M|R], [B*M|R0], P):-  B is  A rem P, B\==0, !,
 	rem_poly(R, R0, P).
rem_poly([_|R], R0, P):- !, rem_poly(R, R0, P).
rem_poly([], [], _).

% ?- time(poly:power_poly(200, [1*[b^1], 1*[a^1]], R, lexical)).
%@ % 245,396 inferences, 0.064 CPU in 0.074 seconds (86% CPU, 3857821 Lips)
%@ R = [1*[b^200], 200*[b^199, a^1], 19900*[b^198, a^2], 1313400*[b^197, a^3], 64684950*[b^196, a^4], 2535650040*[b^195, ... ^ ...], 82408626300*[... ^ ...|...], 2283896214600*[...|...], ... * ...|...] .
%@ % 311,590 inferences, 0.094 CPU in 0.109 seconds (86% CPU, 3305223 Lips)

power_poly(N, X, Y, SW):- bits(N, [], Bs),
	unit_poly(U),
	power_poly(Bs, X, U, Y, SW).

%
power_poly([], _, X, X, _):-!.
power_poly([0|Bs], A, X, Y, SW):- !,
	mul_poly_poly(X, X, X2, SW),
	power_poly(Bs, A, X2, Y, SW).
power_poly([_|Bs], A, X, Y, SW):-
	mul_poly_poly(X, X, X2, SW),
	mul_poly_poly(A, X2, AX2, SW),
	power_poly(Bs, A, AX2, Y, SW).

% ?- poly: mono_lcm_z(1*[a^1], 1*[a^1], A, B, C).
% ?- poly: mono_lcm_z(6*[a^1], 8*[a^1], A, B, C).
% ?- poly: mono_lcm_z(-6*[b^2, a^1], 8*[b^1, a^2], A, B, C).
% ?- poly: mono_lcm([b^2, a^1], [b^1, a^2], LCM, Y, Z).
% ?- poly: mono_lcm([b^2, a^1], [b^1, a^2], LCM).
% ?- poly: mono_lcm([b^2, a^1], [b^1, a^2], LCM).

%
mono_lcm(X, Y, Z):- mono_lcm(X, Y, Z, _, _).

mono_lcm([], Y, Y, Y, []):-!.
mono_lcm(X, [], X, [], X):-!.
mono_lcm([V^I|X],[V^J|Y], [V^M|Z], P, Q):- !, compare(C, I, J),
	( C == (=) -> M=I, mono_lcm(X, Y, Z, P, Q)
	; C == (>) -> M=I, K is I-J, Q=[V^K|Q0], mono_lcm(X, Y, Z, P, Q0)
	; M =J, K is J-I, P=[V^K|P0], mono_lcm(X, Y, Z, P0, Q)
	).
mono_lcm([U^I|X],[V^J|Y], [U^I|Z], P, [U^I|Q]):- V @< U,!,
	mono_lcm(X, [V^J|Y], Z, P, Q).
mono_lcm(X,[B|Y], [B|Z], [B|P], Q):- mono_lcm(X, Y, Z, P, Q).


% ?- poly:mono_lcm_z(4*[], 6*[], A, B, C).
%@ A = 12*[],
%@ B = 3*[],
%@ C = 2*[].
mono_lcm_z(X, Y, U, V):- mono_lcm_z(X, Y, _, U, V).

%
mono_lcm_z(A*X, B*Y, C*Z, D*U, E*V):- mono_lcm(X, Y, Z, U, V),
	lcm(A, B, C, D, E).

% ?- poly:rev_member(X, [a,b,c]).
rev_member(X, [_|R]):- rev_member(X, R).
rev_member(X, [X|_]).

% ?- poly:mod((1/2^21000)*2^2100, X, 2147483647).
%@ X = 1024 .
% ?- poly:mod((1 rdiv 2)^10000000, R, 2147483647).
%@ R = 1 rdiv 1048576 .

% ?- poly:galois(5, -rdiv(-3, -2), Z).
% ?- poly:galois(5, -4, Z).
% P = 2147483647.
galois(P, rdiv(X, Y), Z):- !,
	X0 is (X rem  P),
	Y0 is (Y rem  P),
	Z  is  X0 rdiv Y0.
galois(P, -rdiv(X, Y), Z):- !, galois(P, rdiv(-X, Y), Z).
galois(P, X, X0):-  X0 is (X rem P).

% ?- poly:bits(3, [], X).
%@ X = [1, 1] .
bits(0, X, X).
bits(1, X, [1|X]).
bits(N, X, Y):- B is  N /\ 1,  N0 is N>>1,
	bits(N0, [B|X], Y).

% ?- poly:mod_exp(2147483647, 3, 2, X).
% ?- poly:mod_exp(2147483647, 1000000000000000000000000000, 2, X).
mod_exp(P, N, X, Y):- bits(N, [], Bs), !,  mod_exp(P, Bs, X, 1, Y).

%
mod_exp(P, [B|Bs], X, Y, Z):-!, Ysquare is Y*Y,
	(	B == 0
	->	Y0 is Ysquare
	;	Y0 is X*Ysquare
	),
	galois(P, Y0, Y1),
	mod_exp(P, Bs, X, Y1, Z).
mod_exp(_, [], _,  X, X).
