:- module(primrecfast, []).
:- use_module(engine).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(op).
:- op(1200, xfy, :=).


% ?- run_cfgi(`(peek([@query1, @defs]), maplist(herbrand), apply(primrecfast:eval_primrec_demo))`, [query1=`fact(3)`]).

%  ?- eval_primrec(f(4), [f(x):= if(x, x*f(x-1), 1)], V).
%  V = 24.

eval_primrec_demo(Q, Def, V):- semicolon_to_list(Def, Def0), eval_primrec(Q, Def0, V).

% semicolon_to_list --> eval([primrecfast:semicolon_to_list]).


semicolon_to_list((X;Y),A1):-!,(semicolon_to_list(X,A2),semicolon_to_list(Y,A3)),append(A2,A3,A1) .
semicolon_to_list(X,[X]):-! .


eval_primrec(Q, Defs, V):- eval_primrec(Q, Defs, V, module_kleene).

eval_primrec(Q, Defs, V, P):-
	Ds = [('$VALUE'(0):= Q)| Defs],   % to  keep '$VALUE' as a  predicate name.
	maplist(primrec_horn, Ds, Es),
	Hs=[ (-(X,Y,Z):- primrecfast: -(X,Y,Z)), % builtin
	     (*(X,Y,Z):- primrecfast: *(X,Y,Z)), % builtin
	     (+(X,Y,Z):- primrecfast: +(X,Y,Z))  % builtin
	   |Es],
	maplist(retract_module(P), Hs),  % clear  conflicting predicates
        maplist(P:assert, Hs),
	call(P:'$VALUE'(_, V)).

primrec_horn --> {b_setval(out_arg_last, true)},
	fresh_variables,
%	eval([primrecfast:code]),
	primrecfast:code,
	current(clause(X, Y)),
	{reduce_horn_body(Y, Y0)},
	peek(X:- Y0).

reduce_horn_body -->
	reduce:reduce(primrecfast:horn,
			   subtree((:-) ; (,) ; (;) ; (->))).

retract_module(P, A:-_):- !, retract_module_(P, A).
retract_module(P, A):- retract_module_(P, A).

retract_module_(P, A):- functor(A, F, N), functor(B, F, N),
	retractall(P:B).


%%  Translate primitive recursive functions into Prolog %%%%%
% ?- primrecfast:compile_primrec_horn(test, 'testout.pl').
% ?- primrecfast:primrec_horn(a(X):= 2, Y).
% Y = a(X, 2).
% ?- test_compile.

% symbols--> eval([primrecfast:symbols]).


symbols(X,[X]):-var(X),! .
symbols([],[]):-! .
symbols(X,[X]):-atom(X),! .
symbols(X,[]):-atomic(X),! .
symbols(T,A1):-T=..[A2|As],!,(symbols(As,A3),maplist(symbols,A3,A4)),union(A4,A1) .


test_compile:- compile_primrec_horn(lib('primrecfast-example'),
				    lib('primrecfast-example.out')).

compile_primrec_horn --> translate_file(primrec_horn).

fresh_variables(X, Y):- symbols(X, L), fresh_bind(\(L, X), \(_, Y)).

read_terms(Ts) :- read(X),
	(   X \== end_of_file -> Ts=[X|Ts0], read_terms(Ts0)
	; Ts=[]
	).

:- meta_predicate translate_file(2, ?, ?).

translate_file(A, X, Y):-
	file(X, read, read_terms(X0)),
	maplist(A, X0, Y0),
	file(Y, write, maplist(write_fullstop, Y0)).

write_fullstop(X):- numbervars(X, 0, _, [singletons(true)]),
	print(X), write('.'), nl.

%%%%%%%% simplifying horn clauses

horn(X, Y):- var(X), !, Y=X.
horn(X, Y):- listp(X), list_goal(X, Y).
horn((X, Y), Y):- X==true.
horn((X, Y), X):- Y==true.
horn(clause(X, Y), X:-Y).
horn(X:- Y, X):- Y==true.

list_goal([], true).
list_goal([X|Y], (X, Y0)):- list_goal(Y, Y0).

subcode(X, Y):- nonvar(X), listp(Y).

subcode0(X, Y):- nonvar(X), nonvar(Y).


% % ?-  eval(m(ap, [code], primrecfast), a, Y).
% code(X, (true,X), [], [], true):- var(X); number(X).
% code(if(C, T, E), ((C0, D), V), [C, rec(B, E, T)], [(C0, B), (D, V)], true).
% code(rec(V, T, E), ((V==0 -> T0, R=R1; E0, R=R2), R), [T, E], [(T0, R1), (E0, R2)], true).
% code(H:=Body, clause(H0, Body0), [head(H), Body], [(H0,V), (Body0, V)], true).
% code(head(H), (H0, V), [], [], fun_rel(F, As, H0, V)):- H=..[F|As].
% code(T, Code,  As, Bs, primrecfast:term_code(F, Bs, Code)):- T=..[F|As].

% code(X, (true,X)) :- var(X); number(X).
% code(if(C, T, E), (T0, V=V0; E(C0, D), V)) :-
% code(if(C, T, E), ((C0, D), V)) :-
% 	code(C, rec(B, E, T)),
% 	code((C0, B), (D, V)).
% code(rec(V, T, E), ((V==0 -> T0, R=R1; E0, R=R2), R)):-
% 	code(T, E),
% 	code((T0, R1), (E0, R2)).
% code(H:=Body, clause(H0, Body0)):-
% 	code(head(H), Body),
% 	code((H0,V), (Body0, V)).
% code(head(H), (H0, V)):-
% 	H=..[F|As],
% 	fun_rel(F, As, H0, V).
% code(T, Code):- T=..[F|As].
% 	maplist(code,  As, Bs),
% 	term_code(F, Bs, Code).

%
fun_rel(F, As, G, V):- nb_getval(out_arg_last, true)
	-> append(As, [V], Bs), G=..[F|Bs]
	; G=..[F,V|As].

term_code(F, As, ((Hs, G), V)):- unzip(As, Gs, Vs),
	period_to_comma(Gs, Hs),
	fun_rel(F, Vs, G, V).

period_to_comma([], true):-!.
period_to_comma([X|Y], (X0, Y0)):-!,
	period_to_comma(X, X0),
	period_to_comma(Y, Y0).
period_to_comma(X, X).

%% runtime library
+(X, Y, Z):- Z is X+Y.
-(X, Y, Z):- X>=Y-> Z is X-Y; Z=0.
*(X, Y, Z):- Z is X*Y.

% +(X, Y, Z):- X is Y+Z.
% -(X, Y, Z):- Y>=Z-> X is Y-Z; X=0.
% *(X, Y, Z):- X is Y*Z.

% ?- codes_with(user:eval, "user:append([a,b],[c,d])", X), smash(X).
% :- meta_predicate codes_with(2, ?, ?).
codes_with(A, X, Y) :-  herbrand(X, X0), phrase(A, X0, X1),
	 numbervars(X1, 0, _, [singletons(true)]),
	 with_output_to(codes(Y), print(X1)).

% %% 定義例:
% &(N, X)   ::= if(N, X, 0).
% or(N, X)  ::= if(N, 1, X).

% %% 階乗関数
% fact(N) := rec(1, N * fact(N-1)).

% %% 素数判定
% zero(N)   := (N=0).
% not(N)    := zero(N).
% prime(X)  := &(X>1, not(has_factor(X, X))).
% positive(N) := N>0.
% has_factor(N, Y) := rec(0, &((N-1)>1, or(mod(Y, N-1)=0, has_factor(N-1, Y)))).

% %% eval/2
% % ?- eval_(3+3, X).
% % X = 6
% % ?- eval_(prime(5), X).
% % X = 1

% eval_(X, X):- var(X), !.
% eval_(X, X):- integer(X), !, X>=0.
% eval_(if(X, Y, Z), V):- eval_(X, V0),
% 	(V0==0 -> eval_(Z, V); eval_(Y, V)).
% eval_(X, V):- (X::= Body), eval_(Body, V).
% eval_(T, V):- mapterm(primrecfast:eval_, T, T0), pr_apply(T0, V).

% % pr_apply/2
% pr_apply(X > Y, V):- !, X > Y -> V=1; V=0.
% pr_apply(X = Y, V):- !, X==Y -> V=1; V=0.
% pr_apply(X // Y, V):- !, V is X // Y.
% pr_apply(X mod Y, V):- !, V is X mod Y.
% pr_apply(X*Y, V):- !, V is X*Y.
% pr_apply(X^Y, V):- !, V is X^Y.
% pr_apply(X+Y, V):- !, V is X+Y.
% pr_apply(X-Y, V):- !, X>Y -> V is X-Y; V=0.
% pr_apply(X, Y):-  :=(X, Body), !, pr_apply(X, Body, Y).
% pr_apply(X, X).

% % pr_apply/3
% pr_apply(X, rec(A, B), V):- !, arg(1, X, N), (N==0 -> eval_(A, V); eval_(B, V)).
% pr_apply(_, rec(N, A, B), V):- !,  (N==0 -> eval_(A, V); eval_(B, V)).
% pr_apply(_, Body, V):- eval_(Body, V).


% subst_var(X, Y):- subst_var(X, Y, [], _).

% subst_var(X, X, A, A):- (integer(X); var(X)), !.
% subst_var(X, V, A, B):- atom(X), variable(X), !, assoc_var(X, V, A, B).
% subst_var(X, Y, A, B):- X=..[X0|Xs], subst_var_list(Xs, Ys, A, B),
%          Y =..[X0|Ys].

% subst_var_list([], [], A, A).
% subst_var_list([X|Xs], [Y|Ys], A, B):- subst_var(X, Y, A, A0),
% 	subst_var_list(Xs, Ys, A0, B).

% variable(X):- atom_codes(X, C), expr((code(lower), *code(digit)), C, []).

%%%%%%  sample translation   %%%%
% input (primitive recursive defition)
% fact(n) := rec(n, 1, n * fact(n-1)).
% zero(n) := rec(n, 1, 0).
% not(n) :=  rec(n, 1, 0).
% &(n, x) := rec(n, 0, not(zero(x))).
% or(x, y) :=  not(&(not(x), not(y))).
% positive(n) := rec(n, 0, 1).
% (x > y) := positive(x - y).
% (x=y) := &(zero(x-y), zero(y-x)).
% mod_aux(m, r) := if(m-r>1, r+1, 0).
% mod(x, m) := rec(x, 0, mod_aux(m, mod(x-1, m))).
% prime(x) := &(x>1, not(has_a_factor(x, x))).
% has_a_factor(n, y) := rec(n, 0, &((n-1) >1, or(mod(y,(n-1))=0, has_a_factor(n-1, y)))).

%% output (prolog program)
% fact(B, C):-B==0->C=1; (-(B, 1, D), fact(D, E), *(B, E, F)), C=F.
% zero(B, C):-B==0->C=1;C=0.
% not(B, C):-B==0->C=1;C=0.
% &(B, D, C):-B==0->C=0; (zero(D, E), not(E, F)), C=F.
% or(B, C, G):-not(B, D), not(C, E), &(D, E, F), not(F, G).
% positive(B, C):-B==0->C=0;C=1.
% >(B, C, E):- -(B, C, D), positive(D, E).
% =(B, C, H):- -(B, C, D), zero(D, F), -(C, B, E), zero(E, G), &(F, G, H).
% mod_aux(B, C, F):- -(B, C, D), >(D, 1, E), (E==0->F=0;+(C, 1, G), F=G).
% mod(B, E, C):-B==0->C=0; (-(B, 1, D), mod(D, E, F), mod_aux(E, F, G)), C=G.
% prime(B, F):- >(B, 1, D), has_a_factor(B, B, C), not(C, E), &(D, E, F).
% has_a_factor(B, E, C):-B==0->C=0; (-(B, 1, D), >(D, 1, Ko), -(B, 1, F), mod(E, F, G), =(G, 0, I), -(B, 1, H), has_a_factor(H, E, J), or(I, J, L), &(K, L, M)), C=M.

%% sample run
% ?- prime(5, X).
% X = 1
