:- module(prog_in_prog, [solve/1]).

% Prolog in Prolog.
% Available codes in targets codes.
%
% true/0, fail/0, !/0, var/1, nonvar/1, ;/2,
% clause/2, call/1, predicate_property/2,
% SWI standard builtins

% The codes relies on these primitives:
% true/0, fail/0, !/0, var/1, nonvar/1, ;/2,
% clause/2, call/1, predicate_property/2,

% ?- solve(true).
% ?- solve(false).
% ?- solve(false -> true).   % Prolog behaviour.
% ?- solve(1=2 -> 1=2; 2=2).
% ?- solve(((X=1, X=2) -> (Y = 1; Y=2)); Y =3; Y=4).
% ?- solve(((X=1; X=2) -> (Y = 1; Y=2)); Y =3; Y=4).
% ?- solve(solve(((X=1, X=2) -> (Y = 1; Y=2)); Y =3; Y=4)).
% ?- solve(solve(((X=1; X=2) -> (Y = 1; Y=2)); Y =3; Y=4)).

% test usual Prolog goals.
% ?- x(X).
% ?- solve(x(X)).
% ?- solve(solve(x(X))).  % Idempotency.
% ?- solve(solve(solve(x(X)))).
% ?- solve(solve(solve(solve(x(X))))).

x(1).
x(2).
x(3):-!.
x(4).

% test the cut (!) in toplevel query.
% ?- X=1; X=2; ! ; X=3; X=4.
% ?- solve(X=1; X=2; ! ; X=3; X=4).
% ?- solve((member(X, [1,2,3]), !, member(Y, [a,b,c]), !, member(Z, [u,v,w]))).
% ?- solve(solve((member(X, [1,2,3]), !, member(Y, [a,b,c]), !, member(Z, [u,v,w])))).

% More samples which use clauses  a/1, b/1, c/1, d/2, e/1 defined below.
% ?- a(X).
% ?- solve(a(X)).
% ?- solve(solve(a(X))).
% ?- solve(solve(solve(a(X)))).

% sample clauses.
a(X):- b(X).
%
b(Y):- (Y= 1 ; c(X), !, d(X, Y) ; Y =3).
%
c(2).
c(4).
%
d(X, f(X)).
d(X, g(X)).
%
e(X):- a(X).

%
:- meta_predicate solve(0).
solve(G):- solve(G, Cut, user),
		(	var(Cut)
		;	nonvar(Cut), !, fail
		).
%
solve(true, _, _):-!.
solve(fail, _, _):-!, fail.
solve(!, Cut, _):- var(Cut).	% crucial.
solve(!, !, _):-!.				% crucial.
solve(M:A, Cut, _):-!, solve(A, Cut, M).
solve((A, B), Cut, M):-!,
	( solve(A, Cut, M), solve(B, Cut, M) ).
solve((A; B), Cut, M):-!,
	( solve(A, Cut, M)
	; solve(B, Cut, M)
	).
solve(_, Cut, _):- Cut==!, !.
solve(solve(G), _, M):-!, solve(G, Cut, M),
	( var(Cut);  nonvar(Cut), !, fail).
% solve(solve(G), _, M):-!, solve(M:G).   % Also, of course, OK!
solve(A->B, Cut, M):-!, solve((A, !, B), Cut, M).
solve(call(G), _, M):-!, call(M:G).
solve(A, _, M):- predicate_property(M:A, built_in),	!,
	call(M:A).
solve(A, _, M):- predicate_property(M:A, imported_from(_)), !,
	call(M:A).
solve(A, _, M):- predicate_property(M:A, interpreted), !,
	clause(M:A, G),
	solve(G, Cut, M),
	( var(Cut); nonvar(Cut), !,	fail ).		% crucial.
solve(G, _, M):- call(M:G).  % Unexpected_type_of_call, though.
