:- module(prog_in_prog2, []).

% Prolog in Prolog without builtin ! for deep cut, but
%	with use of nb_setarg trm for keeping track of !.

% true/0, fail/0, !/0 ony for "shallow cut", var/1, nonvar/1,
% ;/2, ->/2,
% clause/2, call/1, predicate_property/2, nb_setarg/3

% test the cut (!) in toplevel query.
% ?- solve(X=1; X=2; ! ; X=3; X=4).
% ?- solve((member(X, [1,2,3]), !, member(Y, [a,b,c]), !, member(Z, [u,v,w]))).
% ?- solve(true).
% ?- solve(((X=1; X=2), !)).
% ?- solve(((X=1; X=2), !, (Y=a; Y=b))).
% ?- solve((member(X, [1,2,3]), !, member(Y, [a,b,c]), !, member(Z, [u,v,w]))).

% Idempotency.  solve(solve(G)) <=> solve(G) <=> G.
% ?- solve(solve(solve(((X=1; X=2)->(Y=a; Y=b))))).
% ?- solve(solve((member(X, [1,2,3]), !, member(Y, [a,b,c]), !, member(Z, [u,v,w])))).

% ?- solve(a(X)).
% ?- solve(b(X)).
% ?- solve((Y= 1 ; c(Y), ! ; Y =3)).

a(X):- b(X).
b(Y):- Y= 1 ; c(Y), ! ; Y =3.
c(2).
c(4).

% ?- x(X).
% ?- solve(x(X)).
% ?- solve(solve(x(X))).			% Idempotency.
% ?- solve(solve(solve(x(X)))).		% Idempotency.

x(1).
x(2).
x(3):-!.
x(4).

%
:- meta_predicate solve(:).
%
solve(G):- Port=port(true),
	solve(G, Cut, user, Port),
	(	var(Cut)
	;  	nonvar(Cut),
		nb_setarg(1, Port, false),
		fail
	).

%
solve(_, _, _, port(false)):-!, fail.
solve(true, _, _, _):-!.
solve(fail, _, _, _):-!, fail.
solve(!, Cut, _, _):- var(Cut).		% crucial.
solve(!, !, _, _):-!.				% crucial.
solve(M:A, Cut, _, P):-!, solve(A, Cut, M, P).
solve((A, B), Cut, M, P):-!,
	(	solve(A, Cut, M, P)
	,	solve(B, Cut, M, P)
	).
solve((A; B), Cut, M, P):-!,
	(	solve(A, Cut, M, P)
	;	solve(B, Cut, M, P)
	).
solve(_, Cut, _, _):- Cut==!, !.
solve(solve(G), Cut, M, P):-!, solve(G, Cut, M, P).
solve(A->B, Cut, M, P):-!, solve((A, !, B), Cut, M, P).
solve(call(G), _, M, _):-!, call(M:G).
solve(A, _, M, _):- predicate_property(M:A, built_in),	!,
	call(M:A).
solve(A, _, M, _):- predicate_property(M:A, imported_from(_)), !,
	call(M:A).
solve(A, _, M, _):- predicate_property(M:A, interpreted), !,
	Port = port(true),		% set first as backtrackable.
	clause(M:A, G),
	Port == port(true),		% backtrackable as if without cut.
	solve(G, Cut, M, Port),
	(	var(Cut)
	;	nonvar(Cut),
		nb_setarg(1, Port, false),  % cut alternatives.
		fail
	).
solve(G, _, M, _):- call(M:G).  % Unexpected type of atomic call.
