:- module(ptqfrag, []).
term_expansion --> pac:expand_pac.
:- use_module(library(clpfd)).

role(X, Y, Z):- user:role(X, Y, Z).

% Note 1: `pred' is a special symbol in the pac.
% Note 2:  quote is reserved in pac quote. So asis is used instead.
% Note 3:  open dicts are not expanded in pred.

% sample queries.

% ?- module(ptqfrag).
% ?- run_samples.
% ?- disable_odict.
% ?- disable_pac_query.
% ?- enable_odict.
% ?- enable_pac_query.

% ?- X={a:{}}.
% ?- X={a:1}, Y={a:1}.
% ?- X={a:1}, Y={a:1}, X=Y.
% ?- X={a:1}, X={b:2}.
% ?- X={a:1}, X={b:2}, Ans=X.a.
% ?- X={a:1}, X.a=Ans, X={a:1}, X={a:2}.  % <= shoud be false.
% ?- 1=X.a.b, X={a:{b:A}}.
% ?- {}=X.a.b, X={a:{b:A}}, Ans=X.a.b.
% ?- X.a - X.b #= 1, X.a+X.b #= 3, [X.a, X.b] ins 1..10, write(['X.a'=X.a, 'X.b'=X.b]).
% ?- A - B #= 1, A + B #= 3, [A,B] ins 1..10.
% ?- X.a - X.b #= 1, X.a+X.b #= 3, [X.a, X.b] ins 1..10, writeln(X.a), writeln(X.b).
% ?- X.a - X.b #= 1, X.a + X.b #= 3, [X.a, X.b] ins 1..10, writeln(X.a), writeln(X.b).
% ?- X.a = Y.b.

% Example use of the iterm
% ?-  U= {a:X,  b:f(X, X)}, X = hello, Ans = U.b.
% ?-  U= {a:X,  b:X.c.d}, X={c:{d:hello}}, Ans= U.b.
% ?-  U= {a:f(X.a(I))}, X={a(1):1, a(2):hello, a(3):3}, Ans=X.a(I), I=2.
% ?-  U= {a:f(X.a, X.b)}, X={a:hello, b:world}, Ans=U.a.
% ?-  U= {a:f(X.a)}, X={a:hello, b:world}, Ans=U.a.
% ?-  U= {a:f(X.a, X.b)}, X={a:hello, b:world}, Ans=U.a, Ans0=X.b.
% ?-  U= {a:f(X.a, X.b)}, X={a:hello, b:world}, Ans=U.a, Ans0=X.b, btree_to_odict(U, D).
% ?-  A = {a:1}, call(pred(A, [X]:- X= A), U), writeln(U.a).

put_attr(X,A):- put_attr(X, cil, A).
get_attr(X,A):- get_attr(X, cil, A).

	/*******************************
	*         PTQ sample runs      *
	*******************************/


% ?- ptqfrag:run_samples.

run_samples :- sample(S), format("~w.\n",[S]),
       once(call(ptqfrag:S, X)),
	   once(call(ptqfrag:X, V)),
       format("Ans = ~w.~n",[V]), fail.
run_samples:- nl.

%
sample(ptq(s, [john, is, a, man], [man(j),find(j,j)])).
sample(ptq(s, [every, man, is, john], [man(j),find(j,j)])).
sample(ptq(s, [every, man, is, john], [man(j),man(k)])).
sample(ptq(s, [every, man, finds, every, man], [man(j), man(k), find(j,j)])).
sample(ptq(pn, [john], [])).
sample(ptq(np, [a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(ptq(s, [a, unicorn, walks], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(ptq(vp, [find, a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(ptq(s, [john,finds, a, unicorn], [male(j),male(b),female(m),unicorn(u), find(m,u),walk(j),walk(b),walk(m)])).
sample(ptq(tv, [find], [man(a),find(j,a),walk(j)])).
sample(ptq(itv, [walk], [man(a),find(j,a),walk(j)])).
sample(ptq(vp, [find, a, unicorn], [unicorn(a),find(j,a),walk(j)])).
sample(ptq(s, [john, walks], [man(a),find(j,a),walk(j)])).
sample(ptq(s, [every, man,  walks], [man(a),find(j,a),walk(j)])).
sample(ptq(s, [every, man,  walks], [man(j),find(j,a),walk(j)])).

	/************************************
	*           For ptq testing.	    *
	************************************/
%
ptq(S, F)  :- ptq(s, S, F).
%
ptq(C, S, F) :- ptq(C, S, F, V),
				call(V, R),
			    writeln(V),
			    format("Ans = ~w.\n", [R]).
%
ptq(P, S, F, Fun):- call(P, E, S, []),
				 individuals(F, Inds),
				 Fun = eval_ptq(E.sem, world(Inds, F)).
ptq(_,_,_,'** syntax error ').

% ?- ptq(s, [every, man,  is, every, man], [man(j)]).
% ?- ptq(s, [every, man,  is, every, man], [man(j), man(k)]).
% ?- ptq(s, [a, man,  is, every, man], [man(j), man(k)]).
% ?- ptq(s, [a, man, finds, every, man], [man(j), man(k), find(j, k)]).
% ?- ptq(s, [a, man,  walks], [man(j), walk(j)]).
% ?- ptq(s, [every, man, finds, every, man], [man(j), man(k), find(j, j), find(j, k)]).
% ?- ptq(s, [john, walks], [walk(j)]).
% ?- ptq(s, [john, is, a, man], [man(j), is(j, j)]).
% ?- ptq(s, [john, finds, john], [find(j, m)]).
% ?- ptq(s, [john, finds, a,  unicorn], [find(j, u), unicorn(u), man(m), walk(m)]).
% ?- ptq(s, [every, man, walks], [man(j), walk(j)]).
% ?- ptq(s, [every, man, finds, a,  unicorn], [find(m, u), unicorn(u), man(m), walk(m)]).
% ?- ptq(vp, [finds, a,  man], [man(k), man(m), find(m, k), find(k, k)]).
% ?- ptq(vp, [finds, john], [find(j, j)]).
% ?- ptq(vp, [finds, every, man], [man(j), man(k), find(j, j), find(j, k)]).
% ?- ptq(vp, [is, a, man], [man(j)])
% ?- ptq(vp, [is, john], [man(j)])
% ?- ptq(np, [a, man], [man(j), man(k), man(l)]).
% ?- ptq(np, [every, man], [man(k), man(m), find(m, k), find(k, k)]).

% ?- ptq(np, [every, man], [man(k)]).
% ?- ptq(determiner, [a], [man(j), walk(j)]).
% ?- ptq(tv, [find], [find(j, k), find(l, m)]).
% ?- ptq(pn, [john], [man(x),find(j,x),walk(j)]).
% ?- ptq(cn, [unicorn], [unicorn(u)]).

	/******************************************
	*            A simple subset of PTQ       *
	******************************************/

% Sentence
s({ sem:truth(in(VP.sem, NP.sem)) }) -->
	np(NP),  vp(VP), { NP.agree = VP.agree }.

% Noun phrase
np({ sem:app(rel_to_fun(Det.sem), CN.sem),
	agree:CN.agree }) -->	determiner(Det), cn(CN).
np({ sem:principal_filter(PN.sem),
	agree:PN.agree,
	cat: PN.cat }
   ) --> pn(PN).

% Verb phrase

% ?- listing(vp).

vp(ITV)  --> itv(ITV).
vp({ sem:inverse_image(rel_to_fun(TV.sem), NP.sem),
    agree:TV.agree }) --> tv(TV), np(NP).

% Intransitive verb
itv(A) --> dict(itv, A).
% Transitive verb
tv(A) --> dict(tv, A).
% Common noun
cn(A) --> dict(cn, A).
% Proper name
pn(A) --> dict(pn, A).
%
determiner(A) --> dict(det, A).

%
dict(Cat, A) --> [X], { dict(X, A), A.cat=Cat }.

% Agreement feature 1
agree_3s({ per:3, num:s }).		% ex. walks

% Agreement feature 2
agree_n3s({ per:X, num:Y }):-	% ex. walk
	when((nonvar(X), nonvar(Y)),
		 member(X-Y, [	1-s, 1-p,
						2-s, 2-p,
							 3-p ])).

% Dict
dict(walk, { sem:predicate(walk/1),
			agree:X,
			cat:itv }):- agree_n3s(X).
dict(walks, { sem:predicate(walk/1),
			 agree:X,
			 cat:itv }):- agree_3s(X).
dict(is, { sem:predicate(is/2),
			agree:X,
			cat:tv }):- agree_3s(X).
dict(find, { sem:predicate(find/2),
			agree:X,
			cat:tv }):- agree_n3s(X).
dict(finds, { sem:predicate(find/2),
			 agree:X,
			 cat:tv }):- agree_3s(X).
dict(kick, { sem:predicate(kick/2),
			agree:X,
			cat:tv }):- agree_n3s(X).
dict(kicks, { sem:predicate(kick/2),
			 agree:X,
			 cat:tv }):- agree_3s(X).

% pronoun. Not used yet.
dict(i, { agree:{ per:1, num:s },
		 cat:prn }).
dict(we,{ agree:{ per:1, num:p },
		 cat:prn }).
dict(you, { agree:{ per:2, num:X },
		   cat:prn }):-
	when(nonvar(X), member(X, [s, p])).
dict(he, { agree:{ per:3, num:s },
		  cat:prn }).
dict(she, { agree:{ per:3, num:s },
		   cat:prn }).
dict(they, { agree:{ per:3, num:p },
			cat:prn }).

%
dict(john, { sem: ind(j),
			agree:{ per:3, num:s },
			cat:pn }).
dict(bill, { sem:ind(b),
			agree:{ per:3, num:s },
			cat:pn }).
dict(mary, { sem: ind(m),
			agree:{ per:3, num:s },
			cat:pn }).
%
dict(unicorn, { sem:predicate(unicorn/1),
			   agree:{ per:3, num:s },
			   cat:cn }).
dict(man,	{ sem:predicate(man/1),
		     agree:{ per:3, num:s },
			 cat:cn }).
dict(woman, { sem:predicate(woman/1),
			agree:{ per:3, num:s },
			cat:cn }).
%
dict(a,		{ sem:quant(a),
			 cat:det }).
dict(every, {  sem:quant(every),
			 cat:det }).

	/**************************************************
	*        Interpreting semantic expressions.       *
	**************************************************/

% ?- module(ptqfrag).
% ?- [util('ptq-fragment')].
% ?- listing(eval_ptq).

% ?- eval_ptq(predicate(walk/1), world([i, j], [walk(i), walk(j)]),  S).

eval_ptq(truth(X), W, S) :-  eval_boole(X, W, S).
eval_ptq(asis(X), _, X) :- !.  % quote is reserved.
eval_ptq(if(X, Y, Z), W, S):-
	eval_boole(X, W, B),
	( B==true ->  eval_ptq(Y, W, S)
	;	eval_ptq(Z, W, S)
	).
eval_ptq(app(F, A), W, V):- !,
	eval_ptq(F, W, F0),
	eval_ptq(A, W, A0),
	memberchk(A0-V, F0).
eval_ptq(call(X), _, _):- !, once(X).
eval_ptq(X, W, S)  :- is_boole(X), !,
	eval_boole(X, W, S).
eval_ptq(L, _, L):- (L==[]; L=[_|_]), !.
eval_ptq(X, W, S)  :- eval_atom(X, W, S), !.

%
eval_atom(predicate(X), W, S) :- !, basic_ext(predicate(X), W, S).
eval_atom(ind(X), _, [X]) :- !.
eval_atom(filter(S), W, V) :- !, eval_ptq(S, W, V0),
 		filter(W, V0, V).
eval_atom(principal_filter(S), W, V) :- !,
	( S = ind(J) ->  Ind=J
	; Ind = S
	),
	filter(W, [Ind], V).
eval_atom(quant(Q), W, V) :- !, eval_quant(Q, W, V).
eval_atom(X, W, Y):- X=..[F|As],
			 maplist(eval_arg(W), As, Bs),
			G=..[F|Bs],
			call(G, Y).

%
eval_arg(_, X, X):- var(X),!.
eval_arg(W, X, Y):- eval_ptq(X, W, Y).

%
is_boole(truth(_)).
is_boole(true).
is_boole(false).
is_boole(and(_,_)).
is_boole(or(_,_)).
is_boole(implry(_,_)).
is_boole(not(_,_)).
is_boole(in(_,_)).
is_boole(=(_,_)).

%
ind(I, _, I).

% ?- eval_boole(truth(not(true)), _, X).
% ?- eval_boole(truth(not(true)), _, X).
% ?- eval_boole(or(1=2, 2=1), _, X).
% ?- eval_boole(imply(1=2, 2=1), _, X).

eval_boole(true, _, true).
eval_boole(false, _, false).
eval_boole(and(X,Y), M, V):-eval_and(X, Y, M, V).
eval_boole(or(X,Y), M, V):-eval_boole(not(and(not(X), not(Y))), M, V).
eval_boole(not(X), M, V):- eval_not(X, M, V).
eval_boole(imply(X,Y), M, V):-eval_boole(or(not(X), Y), M, V).
eval_boole(in(X,Y), M, V):- eval_ptq(X, M, X0),
							eval_ptq(Y, M, Y0),
							check_truth(member(X0, Y0), V).
eval_boole(truth(X), W, V):- eval_boole(X, W, V).
eval_boole(X, _,  V):- check_truth(X, V).

%
eval_not(X, M, V):- eval_boole(X, M, U),
					   ( U== true -> V = false
					   ; V = true
					   ).

%
eval_and(X, Y, M, V):- eval_boole(X, M, U),
						  ( U == false -> V=false
						  ; eval_boole(Y, M, V)
						  ).

%
check_truth(X, true) :-  call(X), !.
check_truth(_, false).

% ?- basic_ext(predicate(love/2),world(_, [love(c,b), love(a,b)]),  X).
% ?- basic_ext(predicate(walk/1), world(_, [walk(a)]),  X).
% ?- basic_ext(predicate(is/2), world([a,b], [walk(a)]),  X).

basic_ext(predicate(P/N), DB, E):-!, basic_ext(P/N, DB, E).
basic_ext(P/1, world(_,F), E):-!,
	T =..[P, X],
	( setof(X, member(T, F), E) -> true
	;  E = []
	).
basic_ext(is/2, world(D, _), E):-!,
	( setof(X-X, member(X, D), E) -> true
	;  E = []
	).
basic_ext(P/2, world(_,F), E):-!,
	T =..[P, X, Y],
	( setof(X-Y, member(T, F), E) -> true
	;  E = []
	).

% ?- eval_quant(a, world([1,2],_),  R).
% ?- eval_quant(every, world([1,2], _),  R).
eval_quant(Q, world(D, _), R):- eval_quant_(Q, D, R0), sort_pairs(R0, R).

%
sort_pairs([X-Y|R], [X0-Y0|R0]):- sort(X, X0), sort(Y, Y0),
								 sort_pairs(R, R0).
sort_pairs([],[]).

eval_quant_(a, D, R):- !,
	(   powerset(D, D0),
		maplist(sort, D0, PowD),
	    setof(X-Y, (member(X,PowD), member(Y,PowD), meet(X,Y)), R)
	->  true
	;   R=[]
	).
eval_quant_(every, D, R):- !,
	(   powerset(D,D0),
		maplist(sort, D0, PowD),
	    setof(X-Y, (member(X,PowD), member(Y,PowD), subset(X,Y)), R)
	->  true
	;   R=[]
	).

% ?- individuals([f(a);g(b)], S).

individuals(F, S):- maplist(atoms,F,F1),
	append(F1, F2),
	sort(F2,S).

%
atoms(X,[X]):-atomic(X),!.
atoms(X,Y):- is_list(X),!,
	maplist(atoms, X, Z),
	append(Z, Y).
atoms(X,Y):- X=..[_|A],
	maplist(atoms, A, B),
	append(B, Y).

	/*****************************************
	*                Helper predicates       *
	*****************************************/

pair(A-B, A, B).
pair(A=B, A, B).
pair(A:B, A, B).

% ?- rel_to_fun([a-b, x-y, a-c, x-z], R).
% ?- rel_to_fun([a=b, x-y, a=c, x-z], R).
% ?- rel_to_fun([a-c, x-y, a-b, x-z], R).
% ?- rel_to_fun([a:c, x-y, a=b, x:z], R).

rel_to_fun(X, Y):- rel_to_fun(X, Y, sort_right).
%
rel_to_fun(X, Y, []):-!, rel_to_fun_(X, [], Y).
rel_to_fun(X, Y, G):- rel_to_fun_(X, [], Y0),
					  call(G, Y0, Y).

%
rel_to_fun_([], X, X).
rel_to_fun_([P|R], X, Y):- pair(P, A, B),
		( select(A-M, X, X0)
		->	rel_to_fun_(R, [A-[B|M]|X0], Y)
		;	rel_to_fun_(R, [A-[B]|X], Y)
		).
%
sort_right([], []).
sort_right([L-R|M], [L-R0|M0]):-
	sort(R, R0),
	sort_right(M, M0).

% ?-  powerset([a,b], X).
powerset(X, Y):- powerset(X, [[]], Y).

powerset([], X, X).
powerset([A|R], X, Y):-
	powerset(X, A, X, X0),
	powerset(R, X0, Y).

%
powerset([], _, X, X).
powerset([X|R], A, S, Y):- powerset(R, A, [[A|X]|S], Y).

% ?- filter(world([a,b,c,d], _), [c], R, mapsort).
% ?- filter(world([a,b,c,d], _), [c], F,  []).
% ?- filter(world([a,b,c,d], _), [c], R).

filter(W, D,  F):- filter(W, D, F, mapsort).

%
filter(W, D, F, []):- !, filter_(W, D, F).
filter(W, D, F, G):-  filter_(W, D, F0),
	  call(G, F0, F).
%
filter_(world(X, _), D, F):- subtract(X, D, Y),
	powerset(Y, P),
	maplist(append(D), P, F).

%
mapsort(X, Y):- maplist(sort, X, Y0),
				sort(Y0, Y).

% ?- principal_filter([a,b,c], b, X).
% ?- principal_filter([a,b,c], b, X, []).

principal_filter(D, A, PF):- principal_filter(D, A, PF, mapsort).
%
principal_filter(D, A, PF, []):- !, principal_filter_(D, A, PF).
principal_filter(D, A, PF, G):- principal_filter_(D, A, PF0),
								call(G, PF0, PF).
%
principal_filter_(D, A, PF):- select(A, D, D0), !,
	powerset(D0,  PD),
	maplist(cons(A), PD, PF0),
	maplist(sort, PF0, PF).
principal_filter_(_, _, []).

%
cons(X,Y,[X|Y]).

%
meet(X, Y):- member(A, X), member(A, Y).

% ?- image([a-v, b-u], [a,b], S).
image(F, X, S):- fun_image(X, F, S0, []),
				 sort(S0, S).

%
fun_image([], _, S, S).
fun_image([X|Y], F, [X0|S], T):- memberchk(X-X0, F), !,
	fun_image(Y, F, S, T).
fun_image([_|Y], F, S, T):- fun_image(Y, F, S, T).

% ?- inverse([a-x, b-y], R).
inverse([], []).
inverse([X-Y|R], [Y-X|R0]):-
	inverse(R, R0).

% ?- inverse_image([a-x, b-y, c-y, d-z], [x, y], U).
inverse_image(F, Y, U):-
	inverse_image(F, Y, V, []),
	sort(V, U).
%
inverse_image([], _, U, U).
inverse_image([X-Y|Fs], P, [X|U], V):-	memberchk(Y, P), !,
	inverse_image(Fs, P, U, V).
inverse_image([_|Fs], P, U, V):-
	inverse_image(Fs, P, U, V).

% Not mine but from SWI-Prolog mailing list.
term_size(Term, Size) :-
    setup_call_cleanup(
        (
            current_prolog_flag(gc, Bool),
            set_prolog_flag(gc, false)
        ),
        (
            statistics(globalused, Used0),
            duplicate_term(Term, _TermCp),
            statistics(globalused, Used1),
            Size is Used1 - Used0
        ),
        set_prolog_flag(gc, Bool)
    ).
