:- module(pureprimrec, []).
:- dynamic  def/2.

% ?- eval(0,X).
% X = 0
% ?- eval(decr(s(0)),X).
% X = 0
% ?- eval(s(decr(0)),X).
% X = s(0)
% ?- eval(s(0)*s(0), X).
% X = s(0)
% ?- eval(s(0)*s(s(0)), X).
% X = s(s(0))
% ?- eval(s(s(0))*s(s(0)), X).
% X = s(s(s(s(0))))
% ?- eval(fact(s(s(0))), X).
% X = s(s(0))
% ?- eval(fact(s(s(s(0)))), X).
% X = s(s(s(s(s(s(0))))))
% ?- eval(decr(s(s(s(0)))), X).
% X = s(s(0))
% ?- eval(decr(decr(s(s(s(0))))), X).
% X = s(0)
% ?- eval(s(s(0))> s(0), X).
% X = s(0)
% ?- eval(s(s(0)) = s(0), X).
% X = 0

def(
   fact(0)   = s(0) ; 
   fact(s(n)) = s(n) * fact(n) ;
    
   0 * x     =  0 ;
   s(n)* x   =  x + n * x ;
    
   0 + x     = x ;
   s(n) + x  = s(n + x) ;

   decr(0)   = 0 ;
   decr(s(x)) = x ;

  (m - 0)    = m ;
  (m - s(n)) = decr(m- n) ;

   zero(0)   = s(0) ;
   zero(s(x))=0;

   not(0)    = s(0) ;
   not(s(n)) = 0;

   &(0, x)   = 0;
   &(s(n), x) = not(zero(x)) ;

   or(x, y)  =  not(&(not(x), not(y))) ;

   if(0, x, y) = y;
   if(s(n), x, y) = x;

   positive(0) = 0;
   positive(s(x)) =  s(0) ;
    
   (x > y)    =  positive(x - y) ;
    
   (x = y)    =  &(zero(x-y), zero(y-x))
   ).

assoc_var(X, V, A, A) :- memberchk((X,V), A), !.
assoc_var(X, V, A, [(X,V)|A]).

subst_var(X, Y) :- subst_var(X, Y, [], _).

subst_var(X, X, A, A):- integer(X), !.
subst_var(X, V, A, B):- atom(X), !, assoc_var(X, V, A, B).
subst_var(X, Y, A, B):- X=..[X0|Xs], subst_var_list(Xs, Ys, A, B),
         Y =..[X0|Ys].

subst_var_list([], [], A, A).
subst_var_list([X|Xs], [Y|Ys], A, B):- subst_var(X, Y, A, A0),
	subst_var_list(Xs, Ys, A0, B).

eval_basic(0, 0):- !.
eval_basic(T, V):- mapterm(eval, T, T0), apply_def(T0, V).

apply_def(X, Y):- def(X, Body), !, eval_basic(Body, Y).
apply_def(X, X).

eval(Expr, V):- def(Eqs), eval_basic(Expr, Eqs, V).

eval_basic(Expr, Eqs, V):- abolish(def/2),
	assert_pr(def, Eqs),
	eval_basic(Expr, V).

assert_pr(D, A; B):- !, assert_pr(D, A), assert_pr(D, B).
assert_pr(D, A = B):- subst_var(A=B, A0=B0),
	C =.. [D, A0, B0],
	assert(C).
