:- module(snippets,
	  [show_digraph/1, show_digraph/2,
	   term_to_graph/4, graph_to_dot/2,
	   clean_paragraph/2,
        al/2, ans/2, ans/3, anseq/2, asy/2, ax/2, bc/2,
	bcopy/2, bv/2, c/2, cl/2, clr/2, cp/2, cp_sample/2, csvcols/2,
	csvcols/3, d/2,
	dcl/3, detex/2, df/2, ds/2, e/2,
	e/3, e_/2, e_/3,
	environment/3,
	lin_sol/2, lip/2, lqa/2, lsample/2, ltx/2, m/2, marriage/3, mc/2,
	mm/2, mmm/2, msmp/3, mt/2, mtex/2, mv/2, nax/2, ndf/2, nlatitem/2,
	npr/2, o/2, oh/2, ol/2, open_pdf/2, pcopy/2, pdo/2, pdot/2,
	pf/2, pipe/3, pm/2, pp/2, pr/2, prf/2, prove_full/2, prove_monadic/2,
	pt/2, ptex/2, qalist/2, qar/2, query_ans_buffer_html/3,
	query_ans_buffer_html_/3, r/2, r_term_dot/2, r_term_dot/3,
	region_lines/3, s/2, s_/2, sa/2, sa_/2, sbc/2, script_size/2, seth/2,
	setp/2, show_set_in_tex/2,
	tabcols/3, tc/2, term_dot/1, term_dot/2, tex_kanji/2, tk/2, try/2,
	ttb/2, tth/2, ttl/2, ttm/2, tto/2, ttq/2, tts/2, ttt/2, ttx/2, ul/2,
	uptex/2, verb/2, x/2, xetex/2]).
%
:- use_module(util(misc)).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
% :- use_module(pac('expand-pac')).
% term_expansion --> expand_pac.

:- use_module(pac(basic)).
:- use_module(pac(op)).

clean_paragraph --> paragraph,
	maplist(phrase((split, remove([])))),
	remove([]).

% ?- environment(itemize, `a\n`, R), smash(R).
% ?- environment(itemize, `\n\na\n\n\n`, R), smash(R).
environment(E) --> clean_paragraph,
	maplist(pred([X, ["\\item ", Y]]:- insert("\n", X, Y))),
	insert("\n"),
        peek(X, ["\\begin{", E, "}\n", X, "\n\\end{", E, "}\n"]).

%
enum	-->	region, environment(enumerate), overwrite.

eit	-->	region, environment(itemize), overwrite.

%
lip --> region, paragraph, remove([]),
	maplist(pred([X, ["<li> ", X, "</li>"]])),
	insert("\n\n"),
	overwrite.

li --> region, split, remove([]),
	maplist(pred([X, ["<li> ", X, "</li>"]])),
	insert("\n\n"),
	overwrite.

ul --> region, peek(X, ["<ul>\n", X, "</ul>\n"]), overwrite.

ol --> region, peek(X, ["<ol>\n", X, "</ol>\n"]), overwrite.

% ?- trace, eval_markup_text(`@ (=).\n hello`, R).

%  [2010/09/27]
ep --> region, parse_bind_context, clear.
ep0 --> region, parse_bind_context.
%  [2013/05]
epr --> region, eval_markup_text, clear.
epr0 --> region, eval_markup_text.

%
sbc --> region, pred([X, X] :- save_bc(X)), clear.
sbc --> peek("parse fail.\n").

bc --> region,
	pac([D, D]:-
		(	nb_getval(saved_bc, G),
			bind_context(([com(G, D)], []), (_, _))
		)
	),
	clear.
bc --> peek("fail.\n").

% [2011/05/06]
ds -->  region, wrap_sort_def, overwrite.

% [2010/12/05]
pm --> region, parse_mimetex([mimetex_name(mimetex)]).

open_pdf --> region, mac_open("~/Cabinet/chronicle/2009/scansnap"), clear.


% dbg --> split,
%	remove([]),
%	maplist(phrase((herbrand(_), eval, herbrand_opp))),
%	insert(`\n`).

c --> elisp:wdf.

pp(X/N, L):- functor(P, X, N), source_file(P, L).

fp(X)-->elisp:find_pred(X).

marriage(N) --> marriage:marriage(N).

%%
x --> {elisp:line_exec}.

pdo --> {platex_buffer}.

% math viewer
mv --> region, latex:math_view, clear.

ptex --> {latex_control:latex_control(ptex, latex_control:jarticle)}.
uptex --> {latex_control:latex_control(uptex, latex_control:uarticle)}.
xetex --> {latex_control:latex_control(xetex, latex_control:xarticle)}.
pdot --> {latex_control:latex_control(ptex, latex_control:powerdot)}.

%
pf --> region, proof_figure.
pt --> region, prooftree:proof_platex_region.

%
cl --> elisp:mark_whole_buffer, elisp:check_length_region.
clr --> elisp:check_length_region.

ltx --> {pdflatex_current_buffer_file}, clear.

%% asymptote
asy --> {asy_current_buffer_file}, clear.

% help(X) :- system:help(X).
% apr(X) :- system:apropos(X).
% listing :- system:listing.
% listing(X) :- system:listing(X).

% ?- make_stable_marriage_problem(3, Y, Z).
msmp(X, Y, Z):- make_stable_marriage_problem(X, Y, Z),
	maplist(writeln, Y), nl,nl,
	maplist(writeln, Z).

% lsample --> e(fun(prover:refute_with_support)).
lsample --> e(fun(fol)).

prove_monadic --> region, herbrand, folm, smash.
prove_full --> region, herbrand, fol, smash.

script_size --> region, wrap(`{\\scriptsize\n`, `}`), overwrite.

'M' --> eit.

tab_to_spaces --> region, listsubst([(`\t`, `    `)]), overwrite.

bv --> region, verb, overwrite.
verb --> peek(X, [`\\verb|`, X, `|`]).
pcopy --> {copy_mirror_pdf(map_mirror)}.
bcopy --> {copy_mirror_buffer(map_mirror)}.

%
pipe(F)--> pipe_to_obj, phrase(F), obj_to_pipe.
dcl(D) --> snd(obj_put(D)).
mtex --> pipe(mimetex).
hopen--> pipe(browse).
qalist--> pipe(question_button_list).

%%%% create html
%% logic-exercise バッファにて::  e(fun(fol))  ＝＞　logic-exercise.html
%% logic-exercise バッファにて::  eho(fun(fol))  ＝＞　deldel.thml
query_ans_buffer_html(X) --> region, query_ans_buffer_html_(X).

query_ans_buffer_html_(X) --> buffer_html(query_ans_html(X, ol)), clear.

e --> query_ans_buffer_html(eval).
e(S) --> query_ans_buffer_html(eval(S)).
s --> query_ans_buffer_html(solve).
sa --> query_ans_buffer_html(solve_all).

e_ --> query_ans_buffer_html_(eval).
e_(S) --> query_ans_buffer_html_(eval(S)).
s_ --> query_ans_buffer_html_(solve).
sa_ --> query_ans_buffer_html_(solve_all).

%%%%
%   ex.  f(rubic:rubic)
%   f(id) ::  1 <cr> 2
%   eho ::  append([1,2,3], [4,5,6])
%   ex_infomath --> query_ans_html(infomath('deldel.html'),
%      eval(fun(id)), ol).
%
% f(F) --> query_ans_html(test('deldel.html'), eval(fun(F)), ol).
% eho    --> query_ans_html(test('deldel.html'), eval, ol).
% eho(S) --> query_ans_html(test('deldel.html'), eval(S), ol).
% solve  --> query_ans_html(test('deldel.html'), solve, ol).
% solve_all --> query_ans_html(test('deldel.html'), solve_all, ol).

% f_(F) --> query_ans_html_(test('deldel.html'), eval(fun(F)), ol).
% eho_    --> query_ans_html_(test('deldel.html'), eval, ol).
% eho_(S) --> query_ans_html_(test('deldel.html'), eval(S), ol).
% solve_  --> query_ans_html_(test('deldel.html'), solve, ol).
% solve_all_ --> query_ans_html_(test('deldel.html'), solve_all, ol).

%%%%%
% ttb:: [1,2]+[3,4]
ttb --> region, set_calc_html.
% ttt:: [1,2]+[3,4]
ttt --> region, herbrand, term_value, peek(X, [`$`, X, `$`]), overwrite.
% ttm:: [1,2]+[3,4]
ttm --> region, herbrand, term_value(set_tex), overwrite.
% ttl:: [1,2]+[3,4]
ttl --> region, query_ans_html(eval(set), ol), html_show. % [//]
% tto:: [1,2]+[3,4]
tto --> region, query_ans_html(eval(set), ol), overwrite.
% tth:: [1,2]+[3,4]
tth --> region, herbrand, set_mimetex, html_show.
% ttx: [1,2]+[3,4]
ttx --> region, herbrand, set_mimetex, overwrite.
% ttq: [1,2]+[3,4]
ttq --> region, herbrand, set_pred, overwrite.
% tts: [1,2]+[3,4]
tts --> region, herbrand, ans_button_set, overwrite.
% seth :: pow([1,2,3])
seth --> region, show_set_in_tex, html_show.
% setp :: pow([1,2,3])
setp --> region, show_set_in_tex, overwrite.

show_set_in_tex --> split,
	remove([]),
	maplist(phrase((herbrand, set_pred, tag(li)))),
	tag(ol).

hc --> region,	peek(A, [`<center>\n`, A, `<br>\n</center>\n`]), overwrite.

% % sweep ::  x+y=1; x+2*y=5/****
% /**  :: er
% #  flatten, herbrand, matrix:solve_equations
% x+y=1;
% x-y=1
% [y, x]=[0, 1]  ***/

lin_sol --> sweep.
sweep --> region, herbrand, matrix:solve_equations, smash.

eq -->  region, env_equation(``),  overwrite.
eql --> region, env_equation(`\\label{eq:}`), smash, current(X),
       overwrite, { elisp:goto_char(`\\label{eq:`, X) }.
go --> region, herbrand(_), phrase, clear.
cp_sample--> {filepath(test('deldel.html'), A),
	 filepath(mathcgi('fpop-sample.html'), B),
	 sh(cp(A, B))}.

cp --> region, herbrand, filepath,
	{filepath(test('deldel.html'), A) },
	#(B, sh(cp(A, B))).

%%%%%%% mimetex for question and answer HTML %%%%%%
oh  --> { html:open_current_html }, clear.

%%% html file
hs --> region,
	obj_init(text, [file_name(test('deldel.html'))]),
	browse,
	clear.

mt  --> html(test('temporal.html'), id, (question(mimetex), tag(p))).
lim --> region, question(mimetex), peek(A, [`<li>`, A, `</li>`]), overwrite.

%% question part
% mm  --> region, question(mimetex), overwrite.
mm --> region, parse_mimetex([]), overwrite.
mmm --> question(mimetex).
mc --> region,
	current(X),
	{ parse_mimetex((X, [mimetex_name(mimetexcode)]), (Y, _)) },
	peek(Y),
	overwrite.

%%% answer part
ans        --> ans([]).
anseq      --> ans([`=`]).
ans(Options) --> region, mm_answer(ans, Options), overwrite.

qar --> region,
	paragraph,
	remove([]),
	sed(qamimetex(mimetex, ans, tag(li), [])),
	tag(ul).

m --> message.
o --> overwrite.
r --> region.
d --> dired.

detex --> region, (tex:tex_parse), detex_rule, overwrite.

tk --> {ensure_loaded('convert-dcg')}, tex_kanji.
tex_kanji --> region, texparse,
	eval([tex_kanji_rule, detex_rule, ps]),
	overwrite.

corona --> tex(drop3zw).

enclose_env(E) --> peek(X, ["\\begin{", E, "}\n", X, "\\end{", E, "}"]).

nlatitem --> texparse, eval(texadjust), tex_codes.
ip       --> region, insert_pause, overwrite.

t2p      --> texparse, eval(try2problem), tex_codes.
nax      --> named_env(ax).
ndf      --> named_env(df).
npr      --> named_env(prop).
try      --> enclose_env(try).
ax       --> enclose_env(ax).
ex       --> enclose_env(ex).
df       --> enclose_env(df).
pr       --> enclose_env(prop).
prf      --> enclose_env(proof).
en       --> enclose_env(enumerate).
al       --> enclose_env(align).
it       --> enclose_env(itemize).
lqa      --> qa(longans).

js      --> region, peek(X, [`<script>`,X,`</script>`]), overwrite.
tc      --> region, peek(X, [`<code>`,X,`</code>`]), overwrite.

% ?- edit(enum).
sol --> region, herbrand, matrix:solve_equations.

%% prove lines  of formulae.

region_lines(X)-->region, split, remove([]), maplist(phrase(X)).

%
csvcols --> csv:columns([1,6], `,`, `\n`).

csvcols(X) --> csv:columns(X, `,`, `\n`).

tabcols --> csv:columns([1,4], `\t`, `\n`).

tabcols(X) --> csv:columns(X, `\t`, `\n`).

%% Display a term using Graphviz
% ?- term_dot(f([a,b,c])).
% ?- term_dot(f(a, g(b1,b2,b3), c)).

%% dot -Tjpg -oterm.jpg term.dot
%% open term.jpg
% ?- graphviz:term2.
%@ true.% ?- graphviz:term2.
%@ true.
% ?- sld2.

%%%%%%%
r_term_dot --> r_term_dot(pdf).
r_term_dot(Toption) --> region, herbrand, current(X),
	{term_dot(Toption, X)},
	clear.

term_dot(X):- term_dot(pdf, X).
%
term_dot(Toption, X):- use_module(graphviz),
          graphviz:term(X),
	  sh(dot( -'T'(Toption), -o('__term.'+ Toption), 'term.dot');
	       open('__term.'+Toption)).


		/**********************
		*     show digraph    *
		**********************/

% ?- snippets:show_digraph(subgraph(g, graph([rankdir = "LR"]); a-[b,c];x-y;u->v)).
% ?- snippets:show_digraph(subgraph(g, graph([rankdir = "LR"]); {rank=same; a; u}; a-[b,c];x-y;u->v)).
% ?- snippets:show_digraph(subgraph(g, graph([rankdir = "LR"]); a-[b,c]; x-y; u->v)).
% ?- snippets:show_digraph(subgraph(g, [a(b,c),x(y)]); y->a; x->c).
%@ true.
% ?- snippets:show_digraph(subgraph(g, [a(b,c)]); subgraph(h,[x(y)]); y->a; x->c).
% ?- snippets:show_digraph(subgraph(cluster_g, [a(b,c)]); subgraph(cluster_h,[x(y)]); y->a; x->c).
% ?- snippets:show_digraph(rankdir="LR"; node([shape=oval]); a->b).
% ?- snippets:show_digraph(a->a->b->c->d).
% ?- snippets:show_digraph([x, y, a->b->c->d]).
% ?- snippets:show_digraph([x,y]).
% ?- snippets:show_digraph([x->y@[color = red]]).  %% << error
% ?- snippets:show_digraph(x->y@[color = red]).
% ?- snippets:show_digraph(x@[color = red, style=filled]).
% ?- snippets:show_digraph(node([shape = box; color = red]); [x, y, a->b->c->d]).
% ?- snippets:show_digraph([a(a)]).
% ?- snippets:show_digraph([a(a,b,c)]).
% ?- snippets:show_digraph([s(y,m), k(y,m)]; rankdir = "LR").
% ?- snippets:show_digraph([s(y,m), k(y,m)]; node([shape=point]); rankdir = "LR").
% ?- snippets:show_digraph(subgraph("clusterABC", [s(y,m), k(y,m)])).
% ?- snippets:show_digraph(a->b->c->d; subgraph("ABC", [s(y,m), k(y,m)])).
%@ true.

show_digraph_default(
	[ out(pdf),
	  counter_name(graph_counter),
	  stem("dot_graph"),
	  directory(D)
	]):- getenv(home, H),
	     atomics_to_string([H,/,'Desktop', /, '_iMG_TMP'], D).
%

%
show_digraph(X) :-  show_digraph_default(Opts),
    once(show_digraph(X, Opts)).

%
show_digraph(G, Opts) :-
	once(term_to_graph(G, H, E, N)),
	once(graph_to_dot(digraph(g, H, E, N), Digraph)),
	new_base_name(Opts, OX),
	obj(get([file(U)]), OX, _),
	atomics_to_string([U, ".dot"], DOT),
	atomics_to_string([U, ".pdf"], PDF),
	file(DOT, write, smash(Digraph)),
	sh(dot(-'T'(pdf), -o(PDF), DOT)),
	sh(open(PDF)).


		/****************************
		*     Graphviz interface    *
		****************************/

% [2015/04/16]
% ?- snippets:term_to_graph([s(m,y), k(m, y)], H, E, N).
% ?- snippets:term_to_graph([a(1),b(1)], H, E, N).
% ?- snippets:term_to_graph(a->a, H, E, N).
% ?- snippets:term_to_graph(a->a->b->c->d,H, E, N).
% ?- snippets:term_to_graph(x-[a,b,c], H, E, N).
% ?- snippets:term_to_graph([x-[a,b,c]], H, E, N).
% ?- snippets:term_to_graph(a=b; c=d, H, E, N).
% ?- snippets:term_to_graph(subgraph(s, a=b; c=d); u->v, H, E, N).
% ?- snippets:term_to_graph(a@b; c=d, H, E, N).
% ?- snippets:term_to_graph(subgraph(cluster1, [1-[a]]),  H, E, N).

term_to_graph(G, H, E, N):- term_to_graph(G, H,[], E,[], N, []).

%
term_to_graph(digraph(G, Body),
	      [digraph(G, Head, E0, N0)|H], H, E, E, N, N):-!,
	term_to_graph(Body, Head, [], E1, [], N1, []),
	sort(E1, E0),
	sort(N1, N0).
term_to_graph(subgraph(G, Body),
	      [subgraph(G, Head, E0, N0)|H], H, E, E, N, N):-!,
	term_to_graph(Body, Head, [], E1, [], N1, []),
	sort(E1, E0),
	sort(N1, N0).
term_to_graph(nopac({G}), H, H0, E, E0, N, N0):- !,
	term_to_graph(G, H, H0, E, E0, N, N0).
term_to_graph((G;G0), H, H0, E, E0, N, N0):- !,
	term_to_graph(G, H, H1, E, E1, N, N1),
	term_to_graph(G0, H1, H0, E1, E0, N1, N0).
term_to_graph((G,G0), H, H0, E, E0, N, N0):- !,
	term_to_graph(G, H, H1, E, E1, N, N1),
	term_to_graph(G0, H1, H0, E1, E0, N1, N0).
term_to_graph(X@A, [X@A|H], H, E, E, N, N):-!.
term_to_graph(X->(Y@A), [(X->Y)@A|H], H, E, E, N, N):-!.
term_to_graph(X->Y, H, H, E, E0, [X|N], N0):-!,
	flatten(->, X->Y, Ns),
	append(Ns, N0, N),
	list_to_links(Ns, E, E0).
term_to_graph(X-Y, H, H, E, E0, [X|N], N0):-!,
	term_to_arrows(Y, X, E, E0, N, N0).
term_to_graph(X, H, H, E, E, [X|N], N):- atomic(X), !.
term_to_graph(X, H, H, E, E0, N, N0):- is_list(X), !,
	term_to_arrows(X, E, E0, N, N0).
term_to_graph(X, [X|H], H, E, E, N, N).

% ?- trace, snippets:term_to_graph(subgraph(cluster1, [a-[1]]), H, E, N).
%@ H = [subgraph(cluster1, [], [(a->1)], [1, a])],
%@ E = N, N = [].


%
list_to_links([A, B|R], [A->B|E], E0):-
	list_to_links([B|R], E, E0).
list_to_links([_], E, E).

%% term_to_graph([], _, E, E, N, N) det

term_to_arrows(A-B, F, [F->A|E], E0, [A|N], N0):- !,
	term_to_arrows(B, A, E, E0, N, N0).
term_to_arrows(A->B, F, [F->A|E], E0, [A|N], N0):- !,
	term_to_arrows(B, A, E, E0, N, N0).
term_to_arrows([A|As], F, E, E0, N, N0):- !,
	term_to_arrows(A, F, E, E1, N, N1),
	term_to_arrows(As, F, E1, E0, N1, N0).
term_to_arrows([], _, E, E, N, N):- !.
term_to_arrows(A, F, [F->G|E], E0, [G|N], N0):-
	A =..[G|As],
	term_to_arrows(As, G, E, E0, N, N0).

%
term_to_arrows(A-B, E, E0, [A|N], N0):- !,
	term_to_arrows(B, A, E, E0, N, N0).
term_to_arrows(A->B, E, E0,[A|N], N0):- !,
	term_to_arrows(B, A, E, E0, N, N0).
term_to_arrows([X|R], E, E0, N, N0):- !,
	term_to_arrows(X, E, E1, N, N1),
	term_to_arrows(R, E1, E0, N1, N0).
term_to_arrows([], E, E, N, N):-!.
term_to_arrows(X, E, E0, [F|N], N0):-!, X =.. [F|As],
	term_to_arrows(As, F, E, E0, N, N0).

%
graph_to_dot(digraph(G, H, E, N),
		      [ "digraph ", G0, " {\n", Body, "\n}\n"]):-
	    term_string(G, G0),
	    maplist(graph_to_dot, [H, N, E], Body).
graph_to_dot(subgraph(G, H, E, N),
		      [ "subgraph ", G0, " {\n", Body, " };\n"]):-
	    term_string(G, G0),
	    maplist(graph_to_dot, [H, N, E], Body).
graph_to_dot(X,  Y):- is_list(X), !, maplist(graph_to_dot, X, Y).
graph_to_dot(@(X, A), D):- !, graph_to_dot(X, A, D).
graph_to_dot(X, [ F0, " [", A0,  "];\n" ]):- X =.. [F, A], !,
	term_string(F, F0),
	flatten_term_string(A, A0).
graph_to_dot(X, [Y, ";\n"]):- term_string(X, Y).

%
graph_to_dot(rank, [A|D], ["{rank=", A, "; ", D0, "};\n"]):- !,
	flatten_list(D, D1, [], ;),
	maplist(term_string, D1, D2),
	insert("; ", D2, D0).
graph_to_dot(X,  D, [ X0, " [", D0,  "];\n" ]):-
	term_string(X, X0),
	flatten_term_string(D, D0).

% ?- snippets:flatten_list([a;b,b;c],X,[],;).
%@ X = [a, b, b, c].

flatten_term_string(X, Y):- flatten_list(X, X0, [], ;),
	maplist(term_string, X0, X1),
	insert(", ", X1, Y).

%
flatten_list([], A, A, _).
flatten_list([X;Y|Z], A, B, ;):- !,
	flatten_list([X], A, C, S),
	flatten_list([Y], C, D, S),
	flatten_list(Z, D, B, S).
flatten_list([X|R], [X|A], B, S):- flatten_list(R, A, B, S).
