:- module(swap_args, [swap_args_of/5]).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% ?- module(swap_args).
smash(X):- basic:smash(X).

% ?- help(snap).

		/***************************************
		*     swap arguments of a predicate    *
		***************************************/

% ?- swap_args_of(f, 1, 2,` f([], [])`, X), basic:smash(X).
% ?- swap_args_of(f, 1, 2,` f([a,b], [[]])`, X), smash(X).
% ?- swap_args_of(f, 1, 2,`f(a, g(b))`, X), smash(X).
% ?- swap_args_of(f, 1, 2,`f([], [])`, X), smash(X).
% ?- swap_args_of(f, 2, 3,`f(a, b, c)`, X), smash(X).
% ?- swap_args_of(f, 2, 3,`f(a, b, c) + x`, X), smash(X).
% ?- swap_args_of(f, 2, 3,` f(a, b, c, d`, X), smash(X).
% ?- swap_args_of(f, 2, 10,`f(a, b, c)`, X), smash(X).
% ?- swap_args_of(f, 1, 2,` f([], [])`, X), smash(X).

swap_args_of(_, I, I, X, X).
swap_args_of(F, I, J, X, Y):-  J < I, swap_args_of(F, J, I, X, Y).
swap_args_of(F, I, J, X, Y):-  atom_codes(F, F0),
       	append(F0, `(`, F1),
        once(swap_args_of(F1, I, J, Y, X, [])).

%
swap_args_of(X, I, J, [A, X, NewArgs, `)` |R]) -->
	skip_to_args_of(X, A),
	full_args_rest(Args),
	!,
	{ length(Args, N),
		(	N >= J
		->   	once(swap_args_by_index(I, J, Args, Args0))
		;	Args0 = Args
		),
	 insert(`,`, Args0, NewArgs)
	},
	swap_args_of(X, I, J, R).
swap_args_of(_, _, _, X) --> is_rest(X).

%
is_rest(X, X, []).

%
is_prefix([C|R])-->[C], is_prefix(R).
is_prefix([])-->[].

% ?- trim_left_white(`abc `, X), smash(X).
% ?- trim_left_white(`\tabc `, X).
trim_left_white --> wl("[ \t]*").

% ?- coalgebra:show_am("(.*[^1-9a-zA-Z_])?").
% ?- coalgebra:show_am(".*").

skip_to_args_of(X, A) --> w("(.*[^1-9a-zA-Z_])?", A, []),
	is_prefix(X).

% ?- full_args_rest(X, `a, [])`, R).
full_args_rest([A|X]) --> argument(A, []),
	args_rest(X, []).

%
args_rest(X, X) --> ")".
args_rest([A|X], Y) --> ",", argument(A, []),
	args_rest(X, Y).

%
argument(X, X) --> look_ahead(`,)`).
argument(X, Y) --> part_of_arg(X, Z),
	argument(Z, Y).

%
part_of_arg(X, Y) --> group(X, Y).
part_of_arg(X, Y) --> quoted_codes(X, Y).
part_of_arg([C|X], X) --> [C].

%
look_ahead(L, R, R):- R = [C|_], memberchk(C, L).

%
group_mark(0'[, 0']).
group_mark(0'(, 0')).
group_mark(0'{, 0'}).

%
group([C|X], Y) --> [C], {group_mark(C, End)},
	group_rest(End, X, Y).
%
group_rest(E, [E|X], X) --> [E].
group_rest(E, X, Y) --> part_of_group(X, Z),
	group_rest(E, Z, Y).
%
part_of_group(X, X) --> look_ahead(`)]}`).
part_of_group(X, Y) --> group(X, Y).
part_of_group(X, Y) --> quoted_codes(X, Y).
part_of_group(X, Y) --> zero_quote_prefix(X, Y).
part_of_group([C|X], Y) --> [C], part_of_group(X, Y).
%
zero_quote_prefix([0'0, 0'\'|X], X) --> "0'".

% Remark: keeping the syntax rule for not spacing for
% the first argument of predicate call like this:  f(a, b, c).

% ?- swap_args_by_index(2, 3, [`a`,` b`,` c`], R).
% ?- swap_args_by_index(1, 2, [`a`,` b`,` c`], R).
% ?- swap_args_by_index(1, 2, [`a`,`b`], R), smash(R).

swap_args_by_index(I, J, X, Y):-  D is J - I,
	once(scan_args(I, D, X, Y, Assocs)),
	edit_args(I, Assocs).

%
scan_args(1, D, [A|X], [B|Y], [A-B, P-Q]):-  scan_args(D, X, Y, P-Q).
scan_args(N, D, [U|X], [U|Y], R):-  succ(N0, N),
	scan_args(N0, D, X, Y, R).

%
scan_args(1, [A|X], [B|X], A-B).
scan_args(N, [U|X], [U|Y], R):- succ(N0, N),
	scan_args(N0, X, Y, R).
%
edit_args(1, [A-B0,  B-[S, A]]):- space_code(S), !,
	trim_left_white(B, B0).
edit_args(_, [A-B,  B-A]).

% ?- quoted_codes(X, [], `'abc'`, R).
% ?- quoted_codes(X, [], `"abc" `, R).

space_code(0'\s).		%'
%
escape_code(0'\\).	%'
%
quotation_code(0'\').	%'
quotation_code(0'\").	%'
quotation_code(0'\`).	%'

quoted_codes([C|X],  Y) --> [C], {quotation_code(C)}, !,
	verb(C, X, Y).
%
verb(C, [C, C|X], X)	--> [C, C].
verb(C, [C|X], X)	--> [C].
verb(C, [E, C|X], X)	--> [E, C], {escape_code(E)}.
verb(C, [A|X], Y)	--> [A], verb(C, X, Y).

		/*************************
		*     insert argument    *
		*************************/
% ?- elem_list(X, [], `abc[`, R).
% ?- elem_list(X, [], `abc]`, R).
% ?- elem_list(X, [], `abc ef)`, R).
% ?- elem_list(X, [], `(a,b)`, R).
% ?- elem_list(X, [], `(f(a),b)`, R).
% ?- elem_list(X, [], `f(a,b )`, []).
% ?- elem_list(X, [], `a(1),b)`, R).
% ?- elem_list(X, [], `(a(1),b)`, R).
% ?- elem_list(X, [], `c(a(1),b).`, R).
% ?- elem_list(X, [], `c(a(1),b)d(c(2), e).`, R).
% ?- block_elem(X, [], `{f(a)=b}`, R).
% ?- elem_list(X, [], `{f(a)=b}`, R).
% ?- elem_list(X, [], `f(a,b),`, R), collect_functors(X, S, []).
% ?- elem_list(X, [], `f((a,b)),`, R), collect_functors(X, S, []).
% ?- elem_list(X, [], `f((a)),`, R).

elem_list(X, Y)	--> elem(X, X0), elem_list(X0, Y).
elem_list(X, X) --> [].

%
elem(X, Y)	--> compound_elem(X, Y)
	|	block_elem(X, Y)
	|	zero_quote_code(X, Y)
	|	radix_number(X, Y)
	|	comment(X, Y)
	|	filler(X, Y)
	|	back_quote_code(X, Y)
	|	string(X, Y)
	|	w('.', X, Y).

% ?- compound_elem(X, [], `f(1)`, R).
% ?- compound_elem(X, [], `f(g( 1 ),h( 2 ))`, R).
compound_elem([$(Functor/N, Args)|X], X) -->
	atom(A, []),	"(",
	{ atom_codes(Functor, A) },
	arg_list(0, N, Args, []).
%
arg_list(I, I, X, X)	--> ")", !.
arg_list(I, J, X, Y)	--> ",", arg_list(I, J, X, Y).
arg_list(I, J, [A|X], Y) --> elem_arg(A, []),
	{ I0 is I+1 },
	arg_list(I0, J, X, Y).
%
elem_arg(X, X) --> look_ahead(`,)`), !.
elem_arg(X, Y) --> elem(X, X0), elem_arg(X0, Y).

% ?- atom(X, [], `abc(`, R).
atom(X, Y) --> wl("[a-z][_a-zA-Z0-9]*", X, Y).

%
block_elem([group(Open, Close, Body)|X], X)--> [Open],
	{ group_mark(Open, Close) },
	block_content(Body, [], Close).

%
block_content(X, X, Close)-->[Close].
block_content(X, Y, Close)--> w(",", X, X0),
	block_content(X0, Y, Close).
block_content(X, Y, Close)--> elem(X, X0),
	block_content(X0, Y, Close).
%
escape_code(X,  Y) --> w("\\", X, [A|Y]), [A].

% ?- string(X, [], `"ab"`, R).
% ?- string(X, [], `"""ab"`, R).
% ?- string(X, [], `"""""ab"`, R).
% ?- string(X, [], `"""a""b"c`, R).
% ?- string(X, [], `"\"\"a\"\"b"c`, R).
string(X, Y) --> w("\"", X, X0),
		   string_content(X0, X1),
		   w("\"", X1, Y).
%
string_content(X, Y) -->  % keep as it is: do not contract anything.
		(	escape_code(X, X0)
		|	w("\"\"", X, X0)
		|	w("[^\"]", X, X0)
		),
		string_content(X0, Y).
string_content(X, X) --> [].

% ?- zero_quote_code(X, [], `0'\\a`, R), smash(X).
zero_quote_code(X, Y) --> w("0'\\\\.", X, Y).

% ?- radix_number(X, [], `12'a1B2C3!`, R), smash(X).
radix_number(X, Y)--> wl("[0-9]+'[0-9a-zA-Z]+", X, Y).

% ?- filler(X, [], `.  .. abc(`, R).
filler(X, Y) --> wl("[\n\r\s\t]+", X, Y).

% ?- comment(X, [], `% comment\t\n\n`, R), smash(X).
% ?- comment(X, [], `% comment\t`, R), smash(X).
% ?- comment(X, [], `/* comment\t\n\n end***/xyz`, R), smash(X).
comment(X, Y) --> line_comment(X, Y) | block_comment(X, Y).

%
line_comment(X, Y) --> wl("%[^\n]*", X, X0),
	( w("\n", X0, Y) | end_of_line_comment(X0, Y) ).
%
end_of_line_comment(X, X, [], []).

% ?- block_comment(X, [], `/*a/*b/*c*/def`, R).
% ?- block_comment(X, [], `/*(a/*b/*c)*/(def`, R).
block_comment(X, Y) --> w("/\\*", X, X0),
		w(".*", X0, X1),
		( w("\\*/", X1, Y)
		| end_of_block_comment(X1, Y)
		).
%
end_of_block_comment([0'*|X], X, [0'*], []).
end_of_block_comment(X, X, [], []).

% ?- back_quote_code(X, [], ```a b c```, R).
% ?- back_quote_code(X, [], ```a ````b c```, R).
back_quote_code(X, Y) --> w("\`", X, X0),
		back_quote_code_content(X0, X1),
		w("\`", X1, Y).
%
back_quote_code_content(X, Y) -->  % Similarly to string/4.
	(	escape_code(X, X0)
	|	w("\`\`", X, X0)
	|	w("[^\`]+", X, X0)
	),
	back_quote_code_content(X0, Y).
back_quote_code_content(X, X) --> [].

% ?- elem_list(X, [], `c(a(1),b) (f(a), g(b)) d(c(2), e).`, R),
%	collect_functors(X, Fs, []), smash(Fs).

% ?- elem_list(X, [], `(c(a(1),b)d(c(2), e)).`, R),
%	collect_functors(X, Fs, []), smash(Fs).
collect_functors(X, Y):- collect_functors(X,  Y0, []),
	sort(Y0, Y).
%
collect_functors([], X, X).
collect_functors([A|B], X, Y):-	!, collect_functors(A, X, X0),
	collect_functors(B, X0, Y).
collect_functors($(F,Args), [F|X], Y):-!, collect_functors(Args, X, Y).
collect_functors(group(_,_, Body), X, Y):-!,
	collect_functors(Body, X, Y).
collect_functors(_, X, X).

% ?- elem_list(X, [], `c(a(1),b), d(c(2), e).`, R),
%	edit_elem_list(swap_args:insert_last_arg, [c/2, c/1], " State", X, U),
%	smash(U).

% ?- elem_list(X, [], `c(a(1),/*c(3)*/b), d(c(2), e).`, R),
%	edit_elem_list(swap_args:insert_last_arg, [c/2, c/1], " State", X, U),
%	smash(U).

edit_elem_list(_, _, _, A, A):- atomic(A), !.
edit_elem_list(F, Sgn, Varname, [X|Y], [X0|Y0]):-!,
	edit_elem_list(F, Sgn, Varname, X, X0),
	edit_elem_list(F, Sgn, Varname, Y, Y0).
edit_elem_list(F, Sgn, Varname, X, Y):-
	call(F, Sgn, Varname, X, Y).
%
insert_last_arg(Sgn, Varname, $(Arity, As), Y):-
	maplist(edit_elem_list(swap_args:insert_last_arg, Sgn, Varname),
			As, Bs),
	insert_to_last(Sgn, Varname, Arity, Bs, Y).
insert_last_arg(Sgn, Varname, group(Open, Close, Body),
				[Open, Body0, Close]):-
	edit_elem_list(swap_args:insert_last_arg, Sgn, Varname,
		Body, Body0).

%
insert_to_last(Sgn, Varname, F/N, As, [F,"(", Bs, ")"]):-
	(	memberchk(F/N, Sgn)
	->	append(As, [Varname], Cs)
	;	Cs = As
	),
	insert(",", Cs, Bs).
