:- module(tex, [
	texparse/2, tex_tokens/2, tex_env/2, filler/2]).

:- encoding(utf8).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.

% :- use_module(pac('expand-pac')).
% term_expansion --> expand_pac.
:- use_module(pac(op)).


		/*************
		*     TeX    *
		*************/

% ?- tex:texparse(`$a$`, X).
% ?- tex:texparse(`Let $x$ be \sqrt{a^2+b^2}.$`, X), writeln(X).
%@ [76,101,116,32,dol([120]),32,98,101,32,32,113,114,116,group([97,94,50,43,98,94,50]),46,36]
%@ X = [76, 101, 116, 32, dol([120]), 32, 98, 101, 32|...].

texparse --> tex_tokens, !, tex_env.

tex_tokens(X,Y) :- tex_tokens(Y,X,[]).

tex_env(X,Y) :- tex_env(Y,[],X,[]), !.

tex_tokens([]) --> "".
tex_tokens(X)  --> "\\\n", !,
		   tex_tokens(X).
tex_tokens(X)  --> tex_token(X,Y), !,
		   tex_tokens(Y).

tex_token(X,Y)	   --> tex_cs(X,Y).
tex_token(X,Y)	   --> tex_group(X,Y).
tex_token(X,Y)	   --> tex_math(X,Y).
tex_token(X,Y)	   --> tex_comment(X,Y).
tex_token([C|X],X) --> [C].

%
tex_cs([cs(Name)|X],X) --> "\\",
			   wl("[a-zA-Z]+", Name1),
			{compound(Name1)},
			{atom_codes(Name, Name1)}.
tex_cs([cs(C)|X],X)    --> "\\" ,
			   [C1],
			   {atom_codes(C, [C1])}.
%
tex_group([group(X)|Y],Y) --> "{", tex_tokens(X), "}".

%
tex_math([ddol(X)|Y],Y) --> "$$", tex_tokens(X), "$$".
tex_math([dol(X)|Y],Y)  --> "$", tex_tokens(X), "$".

%%%%% TeX environments
tex_env(X,X) --> "".
tex_env([group(X)|Y],Z)   --> [group(G)],
			      {tex_env(G,X)},
			      tex_env(Y,Z).
tex_env([env(Name,X)|Y],Z) --> [cs(begin)],
			       filler,
			       [group(T)],
				tex_env(X,[]),
				[cs(end)],
				filler,
				[group(T)],
				{atom_codes(Name,T)},
				tex_env(Y,Z).
tex_env([ddol(X)|Y], Z)	   --> [ddol(M)],
			       {tex_env(M,X)},
			       tex_env(Y,Z).
tex_env([dol(X)|Y], Z)	   --> [dol(M)],
			       {tex_env(M,X)},
			       tex_env(Y,Z).
tex_env([X|Y], Z)	   --> [X],
			       tex_env(Y,Z).

% ?- tex:tex_comment(X, [], `%abc\n`, R).
% ?- tex:tex_comment(X, [], `%abc`, R).
tex_comment([comment(X)|Y],Y) --> "%", wl("[^\n]*$", X).

atomname(N) --> filler,
		wl("[a-zA-Z]*", T),
		filler,
		{atom_codes(N, T)}.

filler --> wl("[\s\t]*").
