:- module(tex, []).
:- style_check(-singleton).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
% :- use_module(pac('expand-pac')).
% term_expansion --> expand_pac.
:- use_module(pac(op)).

:- meta_predicate once(2, ?, ?).
once(P, X, Y):- once(phrase(P, X, Y)).

			/****************************************
			*     listing tex command sequences.    *
			****************************************/

list_texcs_file(File, R):- read_file_to_codes(File, R0, []),
	tex_parse(R0, R1),
	list_texcs(R2, [], R1, []),
	sort(R2, R).

%
list_texcs(X, Y):-  list_texcs(Y, [], X, []).

%
list_texcs([A|X], Y)	--> [cs(A)], !, list_texcs(X, Y).
list_texcs([F|X], Y)	--> [env(F, B)], !,
	{ list_texcs(X, X0, B, []) },
	 list_texcs(X0, Y).
list_texcs(X, Y)	--> [L], { listp(L) } , !,
	{ list_texcs(X, X0, L, []) },
	list_texcs(X0, Y).
list_texcs(X, Y)	--> [_], !, list_texcs(X, Y).
list_texcs(X, X)	--> [].

% some tiny
rest_of_line(X)		-->  prefix(X),	end_of_line(_).
rest_of_line(X, Y)	-->  prefix(X, Y), end_of_line(_).

% ?- tex: filler(X, `   aaa `, R).

filler(X) --> wl(*char(white), X).

%
prefix(X, Y, Z) :-  append(X, Z, Y).
prefix(P, P,  U, U).
prefix([X|P], Q,  [X|U], V):- prefix(P, Q, U, V).

% ?-tex: end_of_line(X, `\nabc`, R).
% ?-tex: end_of_line(X, [], Y).
%@ X = Y, Y = [].

end_of_line(`\n`) --> "\n".
end_of_line([], [], []).

% ?- tex:tex_parse(`$$\\abc + u + \n/* comment */ \n\\xyz$$`, R), tex:flat_tex(R, S), smash(S).
% ?- tex:tex_cs(X, Y, `\\+abc`, R).


			/*****************************
			*     parse text in latex    *
			*****************************/

tex_parse --> tex_tokens, tex_env.

tex_tokens(X, Y) :- once(tex_tokens(Y, [], X,[])).

tex_env(X, Y) :- once(tex_env(Y, [], X, [])).

%
tex_tokens(X, X) --> [].
tex_tokens(X, Y)  --> tex_token(X, X0), !, tex_tokens(X0, Y).

%
tex_token(X, Y)	   --> tex_cs(X, Y).
tex_token(X, Y)	   --> tex_group(X, Y).
tex_token(X, Y)	   --> tex_math(X, Y).
tex_token([comment(C)|X], X) --> comment(C).
tex_token([C|X], X) --> [C].

%
tex_cs([cs(Name)|X],X) --> "\\", wl(+(char(alpha)), Y),
	{atom_codes(Name, Y)}.
tex_cs([cs(C)|X],X)    --> "\\" , w(char(\(alnum)), [C0]), {char_code(C, C0)}.

%
tex_group([group(X)|Y],Y) --> "{", tex_tokens(X), "}".

%
tex_math([dmath(X)|Y], Y) --> "$$", tex_tokens(X), "$$".
tex_math([math(X)|Y], Y)  --> "$", tex_tokens(X), "$".


			/***********************
			*     parse comment    *
			***********************/

%%% Comment and Quotation
% ?-  tex: comment(X, `%comment\nabc`, R), smash(X).
%@ %comment
% ?-  tex: comment(X, `//%comment\nabc`, R), smash(X).
%@ //%comment
% ?-  tex: comment(X, `/********\n  comment  \n********/\nabc`, R), smash(X).
%@ /********
%@   comment
%@ ********/

% [2014/08/29]
% ?- X = (("/\\*" + *(\("\\*/")) + "\\*/") | (("//" | "%") +  "[^\n]*" + "\n")), coalgebra:show_am(X).
% ?- X = (("/\\*" + *(.) + "\\*/") | (("//" | "%") +  "[^\n]*" + "\n")), coalgebra:show_am(X).
%@ X = ("/\\*"+ * ('.')+"\\*/"| ("//"| "%")+"[^\n]*"+"\n")
%
% (("/\\*" + *(.) + "\\*/") | (("//" | "%") +  "[^\n]*" + "\n" )) :   (("/\\*" +  *(\ ("\\*/")) + "\\*/") | (("//" | "%") +  "[^\n]*" + "\n"))

%uncommetn !!!! [2014/09/26]
%%%%% comment(X) --> 	 w( ("/\\*" + *(.) + "\\*/") | (("//" | "%") +  "[^\n]*" + "\n"),    X).

% comment(X) --> w( (("/\\*" +  *(\ ("\\*/")) + "\\*/")
%  		| (("//" | "%") +  "[^\n]*" + "\n")), X).

			/**************************
			*     TeX environments    *
			**************************/
tex_env(X, X) --> [].
tex_env([group(X)|Y],Z)   --> [group(G)], {tex_env(G,X)},  tex_env(Y,Z).
tex_env([env(Name,X)|Y],Z) --> [cs(begin)],
	[group(T)],
	tex_env(X,[]),
	[cs(end)],
	[group(T)],
	{atom_codes(Name,T)},
	tex_env(Y,Z).
tex_env([dmath(X)|Y], Z)   --> [dmath(M)], {tex_env(M, X)}, tex_env(Y, Z).
tex_env([math(X)|Y], Z)	   --> [math(M)],  {tex_env(M, X)}, tex_env(Y, Z).
tex_env([X|Y], Z)	   --> [X], tex_env(Y,Z).


			/*************************************
			*     structured tex to  flat tex    *
			*************************************/

flat_tex(X, Y):- flat_tex(X, Y0, []), flatten(Y0, Y).

%
flat_tex([], X, X).
flat_tex([A|B], X, Y):- flat_tex(A, X, X0), flat_tex(B, X0, Y).
flat_tex(env(Name, A), ["\\being{", Name, "}"|X],  Y):- flat_tex(A, X, X0),
	X0 = ["\n\\end{", Name, "}"|Y].
flat_tex(group(A), ["{"|X],  Y):- flat_tex(A, X, X0), X0 = ["}"|Y].
flat_tex(cs(A), ["\\", A, " "|X],  X).
flat_tex(math(E), ["$"|X], Y):- flat_tex(E, X, X0), X0 = ["$"|Y].
flat_tex(dmath(E), ["$$"|X], Y):- flat_tex(E, X, X0), X0 = ["$$"|Y].
flat_tex(comment(A), [A|X], X).
flat_tex(A, [A|X], X).


% ?- libsort:detex_rule(math([math(a), dmath([b, env(frame, [c]),c])]), R).
% R = [a,[b,[c],c]]


detex_rule(cs(A1),[]):-! .
detex_rule(group(X),A1):-!,detex_rule(X,A1) .
detex_rule(env(A1,X),A2):-!,detex_rule(X,A2) .
detex_rule(math(X),A1):-!,detex_rule(X,A1) .
detex_rule(dmath(X),A1):-!,detex_rule(X,A1) .
detex_rule(X,A1):-listp(X),!,maplist(detex_rule,X,A1) .
detex_rule(X,X):-! .
tex_kanji_rule(cs(X),A1):-!,ts_atom(X,A1) .
tex_kanji_rule(A1,[]):-! .
