:- module(orth, []).

:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('minato-r5')).
:- use_module(pac(op)).
:- use_module(zdd('mate-indirect')).

		/*******************
		*	 helpers       *
		*******************/

% ?- setup_ctrl_opt([rect(2,2)], X), write(X).
% ?- setup_ctrl_opt([rect(2,2), gc(all), a-b], X), write(X).
setup_ctrl_opt(X, ctrl(GC, S, T, Rect)):-
	memberchk(rect(W, H), X),
	Rect = rect(W, H),
	(	memberchk(gc(GC), X) -> true
	;	GC = 0
	),
	(	memberchk(S-T, X) -> true
	;	S = p(0,0),
		T = p(W, H)
	).

% get ctrl option.
get_ctrl(X, end, Y)	:-!, arg(3, X, Y).
get_ctrl(X, gc, Y)	:-!, arg(1, X, Y).
get_ctrl(X, start, Y):-!, arg(2, X, Y).
get_ctrl(X, rect, Y):- arg(4, X, Y).

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
arrow_symbol( _ -> _).
%
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).

% One of the most basic helpers.
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(A-B, U-V):-!, ( B @< A -> U=B, V=A; U=A, V=B ).
normal_pair(A->B, U->V):- ( B @< A -> U=B, V=A; U=A, V=B ).

% ?- frontier_node(p(0, 0)-[p(0,1), p(1,0)], X).
% ?- frontier_node(p(2, 2)-[], X).
% ?- frontier_node(p(1, 1)-[p(2,1)], X).
% ?- frontier_node(p(1, 1)-[p(1,2)], X).

% For rectangular grid graph
frontier_node(_-[_, P], P):-!.
frontier_node(p(_, J)-[p(K, J)], p(K, J)):-!.
frontier_node(p(I, J)-[p(I, _)], p(K, J)):-!, K is I+1.
frontier_node(p(I, J)-[], p(K, J)):- K is I + 1.

% ?- rect_nodes(rect(0,2), Ns).
% ?- rect_nodes(rect(10,10), Ns), length(Ns, L).
rect_nodes(rect(W, H), Ns):-
	findall(p(I,J),
			 (	between(0, W, I),
				between(0, H, J)
			 ),
			 Ns).

% ?- rect_udg(rect(1,1), UDG).
% ?- rect_udg(rect(10,10), UDG),length(UDG, L).
rect_udg(rect(W, H), UDG):-
	findall( p(I,J)-Suc,
				 (	between(0, W, I),
					between(0, H, J),
					findall( p(K,L),
						(  L=J, K is I + 1, K =< W
						;  K=I, L is J + 1, L =< H
						),
						Suc0),
					sort(Suc0, Suc)
				 ),
				 UDG).


		/************************
		*     sample queries    *
		************************/

% ?- rect_mate(rect(0,0), X, S), card(X, C, S).
% ?- rect_mate(rect(1,0), X, S), card(X, C, S).
% ?- rect_mate(rect(1,1), X, S), card(X, C, S).
% ?- rect_mate(rect(2,2), X, S), card(X, C, S).
% ?- rect_mate(rect(3,3), X, S), card(X, C, S).
% ?- rect_mate(rect(4,4), X, S), card(X, C, S).
% ?- rect_mate(rect(5,5), X, S), card(X, C, S).
% ?- time((rect_mate(rect(6,6), X, S), card(X, C, S))).
%@ % 67,880,838 inferences, 5.383 CPU in 5.438 seconds (99% CPU, 12610288 Lips)
%@ X = 8730,
%@ S = ..,
%@ C = 575780564 .

%  ?- time((rect_mate(rect(7,7), X, S), card(X, C, S))).
%@ % 386,937,184 inferences, 35.667 CPU in 35.846 seconds (100% CPU, 10848642 Lips)
%@ X = 112010,
%@ S = ..,
%@ C = 789360053252 .

% ?- time((rect_mate(rect(8,8), X, S), card(X, C, S))).
%@ % 1,590,410,546 inferences, 161.991 CPU in 162.594 seconds (100% CPU, 9817886 Lips)
%@ X = 178466,
%@ S = ..,
%@ C = 3266598486981642 .

% ?- call_with_time_limit(2000, time((rect_mate(rect(9,9), X, S), card(X, C, S)))).
%@ % 6,692,550,225 inferences, 699.389 CPU in 702.117 seconds (100% CPU, 9569144 Lips)
%@ X = 377107,
%@ S = ..,
%@ C = 41044208702632496804.

% ?- call_with_time_limit(36000, time((rect_mate(rect(10,10), X, S), card(X, C, S)))).
%@ % 29,396,165,381 inferences, 3825.651 CPU in 3841.534 seconds (100% CPU, 7683964 Lips)
%@ X = 1268341,
%@ S = ..,
%@ C = 1568758030464750013214100.

% [2022/10/01]
% ?- call_with_time_limit(36000, time((rect_mate(rect(11,11), X, S), card(X, C, S)))).
%@ % 129,787,391,715 inferences, 14036.125 CPU in 14092.162 seconds (100% CPU, 9246668 Lips)
%@ X = 4207535,
%@ S = ..,
%@ C = 182413291514248049241470885236.

% ?- call_with_time_limit(360000, time((rect_mate(rect(12,12), X, [gc(all)], S), card(X, C, S)))).
% ?- call_with_time_limit(360000, time((rect_mate(rect(13,13), X, [gc(0)], S), card(X, C, S)))).

		/**************
		*     main    *
		**************/
%
rect_mate(Rect, X, S):-	udg_mate([gc(0), Rect], X, S).
%
rect_mate(Rect, Opts, X, S):- udg_mate([Rect|Opts], X, S).
%
udg_mate(Opts, X, S):-
	setup_ctrl_opt(Opts, Ctrl),
	get_ctrl(Ctrl, rect, Rect),
	rect_udg(Rect, Dg0),
	reverse(Dg0, Dg),
	open_state(S),
	udg_mate(Ctrl, Dg, 1, X0, S),
	get_ctrl(Ctrl, start, Start),
	get_ctrl(Ctrl, end, End),
	prune_final(Start, End, X0, X, S).
%
udg_mate(_, [], X, X, _).
udg_mate(Ctrl, [N-Ps|UDG], X, Y, S):-
	get_ctrl(Ctrl, end, End),
	frontier_node(N-Ps, F),
	add_node(F-End, N, Ps, X, X0, S),
	prune_by_classify_link(F, End, X0, X1, S),
	ctrl_gc(Ctrl, N, X1, X2, S),
	udg_mate(Ctrl, UDG, X2, Y, S).

		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

add_node(_, N, [], X, Y, S):-!, zdd_insert(N-N, X, Y, S).
add_node(FE, p(0,0), [P], X, Y, S):-!, N=p(0,0),
	subst_node(FE, [N->P], P, N, X, Y, S).
add_node(FE, N, [P], X, Y, S):-!,
	subst_node(FE, [N->P], P, N, X, X0, S),
	zdd_insert(N-N, X, X1, S),
	zdd_join(X0, X1, Y, S).
add_node(FE, p(0,0), [P, Q], X, Y, S):-!, N=p(0,0),
	subst_node(FE, [N->P], P, N, X, X0, S),
	subst_node(FE, [N->Q], Q, N, X, X1, S),
	zdd_join(X0, X1, Y, S).
add_node(FE, N, [P, Q], X, Y, S):-
	subst_node(FE, [N->P], P, N, X, X0, S),
	subst_node(FE, [N->Q], Q, N, X, X1, S),
 	bypass(FE, [N->P, N->Q], P-Q, X, X2, S),
	zdd_insert(N-N, X, X3, S),
	zdd_join_list([X0, X1, X2, X3], Y, S).

% ?- spy(subst_node).
% ?- zdd((X<< +[[c-d]], @(subst_node(c-a, [a->c], c, a, X, Y)), psa(X), psa(Y))).
% ?- #(S), X<< +[[c-d]], subst_node(a-x, [u->v], c, b, X, Y, S), psa(X, S), psa(Y, S).
subst_node(_, _, _, _, X, 0, _):- X < 2, !.
subst_node(FE, Es, A, P, X, Y, S):-	 FE = Fr-End,		% replace A with P
	cofact(X, t(U, L, R), S),
 	subst_node(FE, Es, A, P, L, L0, S),
	classify_link(on_frontier_for_subst, Fr, End, U, Case),
	arrow_symbol(U, _, Lu, Ru),
	(	( Case = 0 ; Case = arrow ) -> R0 = 0
	;	Case = ignore ->
		subst_node(FE, Es, A, P, R, R0, S)
	;	(	Ru = A	->
			normal_pair(Lu-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(P-Ru, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(FE, Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		)
	),
	zdd_join(L0, R0, Y, S).
%
bypass(_, _, _, X, 0, _):- X<2, !.
bypass(FE, Bypass, U, X, Y, S):- FE = Fr-E,
	cofact(X, t(A, L, R), S),
	bypass(FE, Bypass, U, L, L0, S),
	classify_link(on_frontier, Fr, E, A, Case),
	(	( Case = 0; Case = arrow ) ->	R0 = 0			% many hits.
	;	Case = ignore -> bypass(FE, Bypass, U, R, R0, S)		% no hits.
	;	(	A = U -> R0 = 0		%,  write(.)			% so so hits
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node(FE, Bypass, U0, V0, R, R0, S)
		;	bypass(FE, Bypass, U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		)
	),
	zdd_join(L0, R0, Y, S).

% NOT USED HERE. But as a prototype.
add_link(_, _, X, 0, _):- X<2, !.
add_link(FE, U, X, Y, S):- FE = F-E,
	cofact(X, t(A, L, R), S),
	add_link(FE, U, L, L0, S),
	classify_link(on_frontier, F, E, A, Case),
	(	( Case = 0; Case = arrow ) ->	R0 = 0			% many hits.
	;	Case = ignore -> add_link(FE, U, R, R0, S)		% no hits.
	;	U = Ul-Ur,
		(	A = U -> R0 = 0		%,  write(.)			% so so hits
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node(FE, [Ul->Ur], U0, V0, R, R0, S)
		;	add_link(FE, U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		)
	),
	zdd_join(L0, R0, Y, S).

		/***************
		*     prune    *
		***************/

% ?- #(S), X<< +[[a-b, a->b]], prune_final(a, b, X, Y, S).
prune_final(P, P, _, 1, _):-!.
prune_final(_, _, X, 0, _):- X<2, !.
prune_final(P, Q, X, Y, S):- cofact(X, t(A, L, R), S),
	prune_final(P, Q, L, L0, S),
	(	A = (_->_) -> R0 = 0
	;  	A = P-Q -> prune_final0(R, R0, S)
	;	R0 = 0
	),
	zdd_join(L0, R0, Y, S).
%
prune_final0(X, X, _):- X<2, !.
prune_final0(X, Y, S):- cofact(X, t(A, L, R), S),
	prune_final0(L, L0, S),
	(	A = (_->_) -> zdd_insert(A, R, R0, S)
  	;	A = (B-B) -> prune_final0(R, R0, S)
	;	R0 = 0
	),
	zdd_join(L0, R0, Y, S).
%
ctrl_gc(Ctrl, P, X, Y, S):- get_ctrl(Ctrl, gc, GC),
	ctrl_gc_case(GC, P, X, Y, S).
%
ctrl_gc_case(all, P, X, Y, S):-!,
	format("GC at ~w \n", [P]),
	zdd_slim(X, Y, S),
	garbage_collect.
ctrl_gc_case(K, P, X, Y, S):-!,
	P = p(_, J),
	(	J = K ->
		format("GC at ~w \n", [P]),
		zdd_slim(X, Y, S),
		garbage_collect
	;	Y = X
	).
ctrl_gc_case(_, _, X, X, _).

		/***********************************
		*     classify_link by frontier    *
		***********************************/

%  "A node P is on frontier" means that P may be touched by a new link in the future.

% ?- on_frontier(p(0,3), p(1,1)). % false.
% ?- on_frontier(p(1,0), p(1,1)).

%  Left most columnn aware frontier.
on_frontier(P, p(1, K)):-!,   % for pruning after adding nodes.
	(	P = p(1,J) -> J < K
	;	P = p(0,J),
		J =< K+1
	).
% Also works.
% on_frontier(P, p(1, K)):-!,   % for pruning after adding nodes.
% 	(	P = p(1,J) -> J =< K
% 	;	P = p(0,J),
% 		J =< K + 1
% 	).
on_frontier(P, F):- P @< F.
%
on_frontier_for_subst(P, p(1, K)):-!,   % for during substituting nodes.
	(	P=p(1,J) -> J =< K
	;	P=p(0,J), J =< K+1
	).
on_frontier_for_subst(P, F):- P @=< F.

:- meta_predicate classify_link(2,  ?, ?, ?, ?).
classify_link(_, _, _, _->_, arrow):-!.
classify_link(Pred, F, End, A-B,  Case):- call(Pred, A, F), !,
   	(	call(Pred, B, F) -> Case = keep
	;	B = End -> Case = keep
	;	Case = 0
	).
classify_link(_, _, E, E-E, 0):-!.		% added. Many hits, though only at beginning
classify_link(_, _, _, A-A, ignore):-!.
classify_link(_, _, _, _, 0).

% ?- classify_link(p(1,1), p(3,3), p(0,2)-p(3,3), X).  % keep
% ?- classify_link(p(0,0), p(3,3), p(1,1)-p(1,1), X).  % ignore
% ?- classify_link(p(0,0), p(3,3), p(1,0)-p(1,1), X).  % 0
%
classify_link_for_bypass(F, E, A, Case):- classify_link(on_frontier, F, E, A, Case).
%
classify_link_for_subst(X, Y, Z, U):- classify_link(on_frontier_for_subst, X, Y, Z, U).

%
prune_by_classify_link(_, _, X, X, _):- X<2, !.
prune_by_classify_link(F, End, X, Y, S):- cofact(X, t(A, L, R), S),
	prune_by_classify_link(F, End, L, L0, S),
	classify_link(on_frontier, F, End, A, Case),    %  better than classify_link0 as for End-End
	(	Case = arrow -> zdd_insert(A, R, R0, S)
	;	Case = keep ->					% many hits.
		prune_by_classify_link(F, End, R, R1, S),
		zdd_insert(A, R1, R0, S)
	;	Case = ignore ->				% many hits.
		prune_by_classify_link(F, End, R, R0, S)
	;	R0 = 0							% many bits.
	),
	zdd_join(L0, R0, Y, S).

		/***************
		*     debug    *
		***************/

% ?- #(S), rect_mate(rect(1,1), X, S), pmate(X, S).
pmate(X, S):- setup_call_cleanup(
				open_state(M),
				(	drop_path(X, Y, S, M),
					psa(Y, M)
				),
				close_state(M)).

%
drop_path(X, Y, _, _):- X<2, !, Y=X.
drop_path(X, Y, S, M):- cofact(X, t(A, L, R), S),
	drop_path(L, L0, S, M),
	(	A=(_-_) ->
		drop_path(R, R1, S, M),
		zdd_insert(A, R1, R0, M)
	;	R0 = 1
	),
	zdd_join(L0, R0, Y, M).
