:- module(add_node_one_by_one_indirect, []).

:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('minato-r5')).
:- use_module(pac(op)).
:- use_module(zdd('mate-indirect')).

		/*******************
		*   Tiny helpers   *
		*******************/
%
mate_less_than(_ - A, B -_):- A @< B.

% One of the most basic helpers.
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(A-B, U-V):-!, ( B @< A -> U=B, V=A; U=A, V=B ).
normal_pair(A->B, U->V):- ( B @< A -> U=B, V=A; U=A, V=B ).


		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

% ?- zdd((X<< +[[c-d]], @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _, _):- X<2, !.
subst_node(Es, A, P, X, Y, S, W):-			% replace A with P
	cofact(X, t(U, L, R), S),
	(	U= @(_) -> Y = 0
	;   U = Lu-_, A @< Lu -> Y = 0
	;	U = Lu-Ru,
		subst_node(Es, A, P, L, L0, S, W),
		(	Ru = A	->
			normal_pair(Lu-P, V),
			mate_ord_insert([V|Es], R, R0, S, W)
		;	Lu = A	->
			normal_pair(Ru-P, V),
			mate_ord_insert([V|Es], R, R0, S, W)
		;	subst_node(Es, A, P, R, R1, S, W),
			mate_insert(U, R1, R0, S, W)
		),
		mate_join(L0, R0, Y, S, W)
	).

% ?- rect_nodes(rect(0,2), Ns).
% ?- rect_nodes(rect(10,10), Ns), length(Ns, L).
rect_nodes(rect(W, H), Ns):-
	findall(p(I,J),
			 (	between(0, W, I),
				between(0, H, J)
			 ),
			 Ns).

% ?- rect_udg(rect(1,1), UDG).
% ?- rect_udg(rect(10,10), UDG),length(UDG, L).
rect_udg(rect(W, H), UDG):-
	findall( p(I,J)-Suc,
				 (	between(0, W, I),
					between(0, H, J),
					findall( p(K,L),
						(  L=J, K is I + 1, K =< W
						;  K=I, L is J + 1, L =< H
						),
						Suc)
				 ),
				 UDG).

% ?- rect_mate(rect(0,0), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(1,0), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(0,1), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(1,1), X, [gc(all)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(3,3), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(4,4), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- time((rect_mate(rect(4,4), X, [gc(0)], S, W), mate_card(X, C, S, W))).
% ?- time((rect_mate(rect(5,5), X, [gc(0)], S, W), mate_card(X, C, S, W))).
% ?- time((rect_mate(rect(7,7), X, [gc(0)], S, W), mate_card(X, C, S, W))).
%@ C = 789360053252 .
% ?- call_with_time_limit(1000, time((rect_mate(rect(7,7), X, [gc(0)], S, W), mate_card(X, C, S, W)))).
% ?- call_with_time_limit(1000, rect_mate(rect(13,13), X, [gc(all)], S, W)), mate_card(X, C, S, W).
% ?- time((rect_mate(rect(13,13), X, [gc(all)], S, W)), mate_card(X, C, S, W)).
% ?- rect_mate(rect(20,20), X, [gc(0)], S, W), mate_card(X, C, S, W).
% ?- rect_mate(rect(30,30), X, [gc(no)], S, W), mate_card(X, C, S, W), psa(20, W).

% [2022/10/05] terminal (swipl-zdd)
% ?- call_with_time_limit(36000, rect_mate(rect(10,10), X, [gc(all)], S, W)), mate_card(X, C, S, W).
% W = ..,
% C = 1568758030464750013214100.

reverse_successors(X-A, X-B):- reverse(A, B).

% slim(K);  slim(no); slim(all).
rect_mate(Rect, X, S, W):-rect_mate(Rect, X, [slim(0)], S, W).

%
rect_mate(Rect, X, InOpts, S, SW):-
	(	memberchk(gc(CtrlSlim), InOpts) -> true
	;	CtrlSlim= 0
	),
	Rect = rect(W,H),
	rect_udg(Rect, Dg0),
	maplist(reverse_successors, Dg0, Dg1),
	reverse(Dg1, Dg),
	open_mate(S, SW),
	set_key(max_node, p(W, H), S),
	udg_mate(CtrlSlim, Dg, 1, X0, S, SW),
	get_key(max_node, Max, S),
	prune_final(p(0,0), Max, X0, X, S, SW).
%
udg_mate(_, [], X, X, _, _).
udg_mate(Ctrl, [N-Ps|UDG], X, Y, S, W):-
	add_node(N, Ps, X, X0, S, W),
	get_key(max_node, Max, S),
	prune_by_classify_link(N, Max, X0, X1, S, W),
	ctrl_slim(Ctrl, N, X1, X2, S),
	udg_mate(Ctrl, UDG, X2, Y, S, W).

%
ctrl_slim(all, P, X, Y, S):-!,
	format("GC at ~w \n", [P]),
	zdd_slim(X, Y, S),
	garbage_collect.
ctrl_slim(no, _, X, X, _):-!.
ctrl_slim(K, P, X, Y, S):-integer(K),
	P = p(_, J),
	(	J = K ->
		format("GC at ~w \n", [P]),
		zdd_slim(X, Y, S),
		garbage_collect
	;	Y = X
	).

%
add_node(N, [], X, Y, S, _):-!, zdd_insert(N-N, X, Y, S).
add_node(N, [P|Ps], X, Y, S, W):- normal_pair(N->P, Q),
	subst_node([Q], P, N, X, X0, S, W),
	zdd_insert(N-N, X, X1, S),
	mate_join(X0, X1, X2, S, W),
	add_neighbor_links(N, Ps, X2, Y, S, W).
%
add_neighbor_links(_, [], X, X, _, _):-!.
add_neighbor_links(N, [P|Ps], X, Y, S, W):- normal_pair(N-P, Link),
	add_link(Link, X, X0, S, W),
	mate_join(X0, X, X1, S, W),
	add_neighbor_links(N, Ps, X1, Y, S, W).

% ?- open_mate(S, W),
%	zdd_eval(+[[a-b, c-d]], X, S), ipsa(X, S, W),
%	add_link(b-c, X, Y, S,W), ipsa(Y, S, W).

% ?- open_mate(S, W),
%	zdd_eval(+[[a-b, c-d], [e-f]], X, S), ipsa(X, S, W),
%	add_link(b-c, X, Y, S, W), ipsa(Y, S, W).

% ?- open_mate(S, W),
%	zdd_eval(+[[a-b, c-d], [a-a, b-b]], X, S),
%	ipsa(X, S, W),
%	add_link(b-c, X, Y, S, W), ipsa(Y, S, W).

% ?- open_mate(S, W),
%	zdd_eval(+[[a-a, b-b, c-c]], X, S),
%	ipsa(X, S, W),
%	add_link(b-c, X, Y, S, W), ipsa(Y, S, W).

add_link(_, X, 0, _, _):- X<2, !.
add_link(U, X, Y, S, W):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = @(_) -> Y = 0
	;	add_link(U, L, L0, S, W),
		U = Ul-Ur,
		(	mate_less_than(U, A) -> R0 = 0
		; 	A = U -> R0 = 0
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node([(Ul->Ur)], U0, V0, R, R0, S, W)
		;	add_link(U, R, R1, S, W),
			mate_insert(A, R1, R0, S, W)
		),
		mate_join(L0, R0, Y, S, W)
	).

% ?- open_mate(S, W),
%  zdd_eval(+[[p(1,1)-p(3,3)]], X, S),
%	prune_final(p(0,0), p(3,3), X, Y, S, W).

% ?- open_mate(S, W),
%  zdd_eval(+[[p(0,0)-p(3,3)]], X, S),
%	mate_insert(p(2,2)->p(2,3), X, X0, S, W),
%	prune_final(p(0,0), p(3,3), X0, Y, S, W),
%	ipsa(Y, S, W).

prune_final(_, _, X, X, _, _):- X<2, !.
prune_final(P, Q, X, Y, S, W):- cofact(X, t(A, L, R), S),
	prune_final(P, Q, L, L0, S, W),
	(	A = @(_) -> cofact(R0, t(A, 0, 1), S)
	;	A = A1-A2,
		(	 A1 = P ->
			 (	A2 = Q
			 -> prune_final(P, Q, R, R0, S, W)
			 ;	R0 = 0
			 )
		;	A1 = A2 ->
			prune_final(P, Q, R, R0, S, W)
		;	R0 = 0
		)
	),
	mate_join(L0, R0, Y, S, W).

% ?- classify_link(p(0,1), p(3,3), p(1,0)-p(3,3), X).
% ?- classify_link(p(0,1), p(3,3), p(0,2)-p(3,3), X).
% ?- classify_link(p(0,1), p(3,3), p(0,1)-p(3,3), X).

classify_link(_, _, @(_), id):-!.
classify_link(P, End, A-B,  Case):- on_frontier(P, A), !,
   	(	on_frontier(P, B) -> Case = keep
	;	B = End -> Case = keep
	;	Case = 0
	).
classify_link(_, _, A-A, ignore):-!.
classify_link(_, _, _, 0).
%
on_frontier(p(0, J), p(0, J)):-!.
on_frontier(p(0, J), p(K, L)):-!, K = 1, L< J.
on_frontier(p(I, J), p(I, K)):-  K >= J, !.
on_frontier(p(I, J), p(I0, K)):- I0 is I + 1, K < J.
%
prune_by_classify_link(_, _, X, X, _, _):- X<2, !.
prune_by_classify_link(P, End, X, Y, S, W):- cofact(X, t(A, L, R), S),
	classify_link(P, End, A, Case),
	(	Case = id ->   Y = X
	;  	prune_by_classify_link(P, End, L, L0, S, W),
		(	Case = keep ->
			prune_by_classify_link(P, End, R, R1, S, W),
			mate_insert(A, R1, R0, S, W)
		;	Case = ignore ->
			prune_by_classify_link(P, End, R, R0, S, W)
		;	R0 = 0
		),
		mate_join(L0, R0, Y, S, W)
	).

% ?- open_mate(S, W),
%	mate_ord_insert([a-b], 1, X, S, W),
%	mate_ord_insert([c-d, e->f], 1, Y, S, W),
%	mate_join(X, Y, Z, S, W),
%	mate_card(Z, C, S, W), ipsa(Z, S, W).
