:- module(count_loops, []).

% (start at copy of minato-r4)

:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).
:- op(1060, xfy, #).		% exor
:- op(1000, xfy, &).

% term_expansion --> pac:expand_pac.
term_expansion --> expand_pac.

		/******************
		*     Switches    *
		******************/
% ORTHOGONAl
% active_node(X, Y):- active_node_orth(X, Y).
% rect_nodes(X, Y):- rect_nodes_orth(X, Y).

% SKEW.
active_node(X, Y):- active_node_skew(X, Y).
rect_nodes(X, Y):- rect_nodes_skew(X, Y).

% some tiny.
w(Opt):- memberchk(out(X), Opt), memberchk(state(S), Opt), psa(X, S).

% ?- count_rect_cycles(rect(0,0), C, _).
% ?- count_rect_cycles(rect(1,0), C, _).
% ?- count_rect_cycles(rect(0,1), C, _).
% ?- count_rect_cycles(rect(1,1), C, U).
% ?- count_rect_cycles(rect(1,2), C, _).
% ?- count_rect_cycles(rect(2,1), C, _).
% ?- count_rect_cycles(rect(2,2), C, _).
% ?- count_rect_cycles(rect(1,10), C, _).
% ?- count_rect_cycles(rect(10,1), C, _).
% ?- count_rect_cycles(rect(100,1), C, _).
% ?- count_rect_cycles(rect(1,100), C, _).
% ?- repeat(between(1, 6, N), (check(N, C), writeln( count_for(N) =  C))).
check(N, C):- findall((I,J), (between(0, N, I), between(I, N, J), I < J),  S),
	 length(S, C).

% ?- repeat(between(1, 6, N), (count_rect_cycles(rect(N, 1), C), writeln(count_rect_for(N) =  C))).

% ?- N = 3, findall((I,J), (between(0, N, I), between(I, N, J), I < J),  S),
%	length(S, C).

% ?- time(count_rect_cycles(rect(6,6), C, _)).
%@ % 25,642,240 inferences, 3.558 CPU in 4.465 seconds (80% CPU, 7207373 Lips)
%@ C = 575780564.
% ?- time(count_rect_cycles(rect(7,7), C, _)).
%@ % 106,041,498 inferences, 12.348 CPU in 12.454 seconds (99% CPU, 8587850 Lips)
%@ C = 789360053252.
% ?- time(count_rect_cycles(rect(8,8), C, _)).
%% [2022/08/10] change of rehash.  faster, but correct ?
%@ % 332,129,654 inferences, 33.514 CPU in 33.728 seconds (99% CPU, 9910274 Lips)
%@ C = 2318527339461265.
%@ % 436,367,556 inferences, 52.415 CPU in 52.981 seconds (99% CPU, 8325235 Lips)
%@ C = 3266598486981642.
% ?- time(count_rect_cycles(rect(9,9), C, _)).
%@ % 1,467,957,762 inferences, 167.674 CPU in 168.988 seconds (99% CPU, 8754851 Lips)
%@ C = 27359264067916806101.
%@ % 1,793,021,796 inferences, 224.894 CPU in 227.462 seconds (99% CPU, 7972753 Lips)
%@ C = 41044208702632496804.
% ?- call_with_time_limit(1800, time(count_rect_cycles(rect(10,10), C, S))).
%@ % 12,782,653,306 inferences, 1164.030 CPU in 1174.499 seconds (99% CPU, 10981375 Lips)
%@ C = 1568758030464750013214100,
% ?- call_with_time_limit(36000, count_rect_cycles(rect(11,11), C, _)).
%@ C = 182413291514248049241470885236.
% ?- convert_skew(p(1,1), Q), convert_skew(P, Q), convert_skew(P, R).
%@ Q = R, R = p(2, 1),
%@ P = p(1, 1).
orth_skew(p(I, J), Y):- var(Y), !,
	Y = p(K, I),
	K is I + J.
orth_skew(p(I, J), p(K, I)):- J is K - I.

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
change_symbol(A-B, A->B).
arrow_symbol( _ -> _).
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).
%
is_link(->(_, _)).
%
convert_normal_pair(P, Q, S):- orth_skew_pair(P, Q0),
	normal_pair(Q0, Q, S).
%
normal_arrow(X, Y):- @(normal_arrow(X, Y)).
%
normal_arrow(A->B, Arr, S):-!,
	(	zdd_compare(>, A, B, S) -> Arr = (B->A)
	;	Arr = (A->B)
	).
normal_arrow(A-B, Arr, S):-normal_arrow(A->B, Arr, S).
%

lt_pair(A - _, B -_, S):- zdd_compare(<, A, B, S).


		/*************************************
		*     basics of incrementing DG      *
		*************************************/

composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(P, U):- @(normal_pair(P, U)).
%
normal_pair(A-B, B-A, S):- zdd_compare(>, A, B, S), !.
normal_pair(A->B, B->A, S):- zdd_compare(>, A, B, S), !.
normal_pair(P, P, _).

% ?- orth_skew_pair(p(1,1)-p(2,2), X), orth_skew_pair(Y, X).
%@ X = p(2, 1)-p(4, 2),
%@ Y = p(1, 1)-p(2, 2).
orth_skew_pair(P, Q):-
	(	P = (A-B) -> Q = (A0-B0)
	;	P = (A->B),
		Q = (A0->B0)
	),
	orth_skew(A, A0),
	orth_skew(B, B0).

% ?- link_sort_opp([loop, a, loop], X).
link_sort_opp(X, Y):- sort(X, X0), reverse(X0, Y).

% ?- node_ends(p(0,0), En, rect(1,1)).
% ?- node_ends(p(0,0), En, rect(0,0)).
% ?- node_ends(p(1,1), En, rect(2,2)).

node_ends(P, En, rect(W,H)):- P = p(I,J),
	(	I < W ->
		K is I + 1,
		En0 = [p(K,J)]
	;	En0 = []
	),
	(	J < H ->
		L is J + 1,
		En = [p(I, L)|En0]
	;   En = En0
	).


% ?- node_ends_skew(p(0,0), En, rect(1,1)).
% ?- node_ends_skew(p(0,0), En, rect(0,0)).
% ?- node_ends_skew(p(1,1), En, rect(2,2)).
% ?- node_ends_skew(p(1,1), En, rect(1,2)).
node_ends_skew(P, En, rect(W,H)):- P = p(I,J),
	(	J < W ->
		K is J + 1,
		M is I + 1,
		En0 = [p(M, K)]
	;	En0 = []
	),
	(	I < H + J ->
		L is I + 1,
		En = [p(L, J)|En0]
	;   En = En0
	).

% ?- R=rect(2,2), repeat(in_rect(P, R),
%	(node_ends_shadows(P, En, Sh, R), writeln(P-En-Sh))).
in_rect(p(I, J), rect(W, H)):- between(0, W, I), between(0, H, J).

% ?- rect_nodes_orth(rect(1,1), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(3,3), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(10,10), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(30,30), PES), length(PES, L).
% ?- rect_nodes_orth(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_orth(rect(1000,1000), PES), length(PES, L))).

rect_nodes_orth(R, PES):-
	findall(P-E, (in_rect(P, R), node_ends(P, E, R)), PES).

% ?- rect_nodes_skew(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_skew(rect(1000,1000), PES), length(PES, L))).
rect_nodes_skew(R, PES_skew):- rect_nodes_orth(R, PES_orth),
	convert_orth_skew(PES_orth, PES_skew).
%
convert_orth_skew([], []).
convert_orth_skew([A-B|C], [A0-B0|C0]):-
	orth_skew(A, A0),
	maplist(orth_skew, B, B0),
	convert_orth_skew(C, C0).

		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

select_cycles(P, [A, B], X, Cyc, S):-!,
	select_cycles_([P->A, P->B], A-B, X, Cyc, S).
select_cycles(_, _, _, 0, _).

% ?- zdd((X<<{[a-b, c-c, a->d, b->d], [a-b, c-c, c->f, b->f], [u-v]}, psa(X), @(select_cycles_([1->a, 1->b], a-b, X, Y)), psa(Y))).
select_cycles_(_, _, X, 0, _):- X<2, !.
select_cycles_(B, A, X, Y, S):- cofact(X, t(U, L, R), S),
	select_cycles_(B, A, L, L0, S),
 	select_cycles_(B, A, U, R, R0, S),
	zdd_join(L0, R0, Y, S).

%
select_cycles_(B, A, A, X, Y, S):-!,
	collect_link_free(X, Y0, S),
	zdd_ord_insert(B, Y0, Y, S).
select_cycles_(B, A, U-U, X, Y, S):-!, select_cycles_(B, A, X, Y, S).
select_cycles_(_, _, _, _, 0, _).


% ?- zdd((X<<{[a-b, c->d], [e->f]}, @(collect_link_free(X, Y)), psa(Y))).
% ?- zdd((X<<{[a-b, c->d], [a-a, e->f]}, @(collect_link_free(X, Y)), psa(Y))).
collect_link_free(X, X, _S):- X<2, !.
collect_link_free(X, Y, S):- cofact(X, t(A, L, R), S),
	collect_link_free(L, L0, S),
	(	A = (U-U) -> collect_link_free(R, R0, S)
	;	A = (_-_) -> R0 = 0
	;	zdd_insert(A, R, R0, S)
	),
	zdd_join(L0, R0, Y, S).

		/*************************
		*     count_rect_path    *
		*************************/

% ?- test(rect(1,1), Y).
test(X, Y):- count_rect_cycles(X, Y, _, _).


% ?- N = 3,  count_rect_cycles(rect(0, N), C).
% ?- N = 3, count_rect_cycles(rect(1, N), C).
% ?- N = 10, count_rect_cycles(rect(1, N), C).
%% Qcompile: /Users/cantor/devel/zdd/prolog/zdd/count-loops.pl
%@ true.
%@ Context module: count_loops.
%@ true.
% ?- spy(node_one_by_one).
% ?- spy(psa).
% ?- count_rect_cycles(rect(0,0), C, S), w(S).
% ?- count_rect_cycles(rect(1,0), C, S), w(S).
% ?- count_rect_cycles(rect(0,1), C, S), w(S).
% ?- count_rect_cycles(rect(1,1), C, S).
% ?- count_rect_cycles(rect(2,1), C, S).
% ?- count_rect_cycles(rect(1,2), C, S).
% ?- count_rect_cycles(rect(3,1), C, S).
% ?- count_rect_cycles(rect(2,2), C, S).
% ?- count_rect_cycles(rect(2,1), C, S).
% ?- count_rect_cycles(rect(3,1), C, S).
% ?- count_rect_cycles(rect(3,3), C, S).
% ?- count_rect_cycles(rect(4,4), C, S).

% [2022/08/08]
% ?- time(count_rect_cycles(rect(10,10), C, S)).
%@ % 6,626,002,510 inferences, 613.462 CPU in 618.651 seconds (99% CPU, 10801003 Lips)
%@ C = 988808811046283595068099,
%@ S = [state(..), out(2473748)|_].
% [2022/08/07]
% ?- time(count_rect_cycles(rect(10,10), C, S)).
%@ % 6,626,002,510 inferences, 639.629 CPU in 644.225 seconds (99% CPU, 10359131 Lips)
%@ C = 988808811046283595068099,
%@ S = [state(..), out(2473748)|_].


% ?- zdd((X<<pow([1,2]), card(X, C))).
% ?- time(count_rect_cycles(rect(8,8), C, S)).
%@ % 436,367,556 inferences, 35.323 CPU in 36.801 seconds (96% CPU, 12353625 Lip
%@ C = 3266598486981642,
%@ S = [state(..), out(4921865)|_].
% ?- call_with_time_limit(200, time(count_rect_cycles(rect(8,8), C, _))).
%@ % 908,479,618 inferences, 89.804 CPU in 91.150 seconds (99% CPU, 10116254 Lips)
%@ C = 3266598486981642.
% ?- count_rect_cycles(rect(10,10), C, _).
%@ C = 988808811046283595068099.

count_rect_cycles(Rect, C):- count_rect_cycles(Rect, C, _).
%
count_rect_cycles(Rect, C, Opts):-
	memberchk(state(S), Opts),
	memberchk(out(X), Opts),
	open_state(S),
	set_key(rect, Rect, S),
	rect_nodes(Rect, A),
	sort(A, A1),
	reverse(A1, B),
	B =[Max-_|_],
	arg(1, Max, K),
	set_key(prune_index, K, S),
	node_one_by_one(B, 1-_, 0-X, S),
%	psa(X, S),
	card(X, C, S).

% ?- zdd(@(node_one_by_one([], 1-X, 0-Y))).
node_one_by_one([], X-X, Y-Y, _).
node_one_by_one([P-En|R], X-X0, Y-Y0, S):-
	writeln(P),
	select_cycles(P, En, X, Cyc, S),
	zdd_join(Cyc, Y, Y1, S),
	add_node_one(P, En, X, X1, S),
	prune_inactive_node(P, X1, X2, S),
	get_key(prune_index, I, S),
	arg(1, P, J),
	(	J = I ->  X3 = X2,  Y2 = Y1
	;	set_key(prune_index, J, S),
		writeln("Slimming..."),
		zdd_slim([X2, Y1], [X3, Y2], S),
		garbage_collect
	),
	node_one_by_one(R, X3-X0, Y2-Y0, S).


% ?- zdd((X<<{[b-c, b->c]}, psa(X),  @(add_node_one(a, [b], X, Y)), psa(Y))).
add_node_one(P, [], X, Y, S):-!, zdd_insert(P-P, X, Y, S).
add_node_one(P, [A], X, Y, S):-!,
	subst_node([P->A], A, P, X, Y0, S),
	zdd_insert(P-P, X, Y1, S),
	zdd_join(Y0, Y1, Y, S).
add_node_one(P, [A, B], X, Y, S):-
	subst_node([P->A], A, P, X, Y0, S),
	subst_node([P->B], B, P, X, Y1, S),
	bypass_via_node([P->A, P->B], B-A, X, Y2, S),
	zdd_insert(P-P, X, Y3, S),
	zdd_join_list([Y0, Y1, Y2, Y3], Y, S).


% ?- zdd((X<<{[c-d]}, @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _):- X < 2, !.
subst_node(Es, A, P, X, Y, S):- % memo is useless here
	/* replace A with P */
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	zdd_compare(<, A, Lu, S) ->	Y = 0
	;	subst_node(Es, A, P, L, L0, S),
		(	Lu = A	->
			normal_pair(Ru-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Ru = A	->
			normal_pair(Lu-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[a-b, c-d]}, @(bypass_via_node(b-c, X, Y)), psa(X), psa(Y))).
bypass_via_node(E, X, Y, S):- bypass_via_node([], E, X, Y, S).

% ?- zdd((X<<{[a-b, c-d]}, @(bypass_via_node([b->c], b-c, X, Y)), psa(X), psa(Y))).
bypass_via_node(_, _, X, 0, _):- X < 2,!.
bypass_via_node(Bypass, E, X, Y, S):- % memo is useless here
	cofact(X, t(P, L, R),  S),
	(	P = (_->_)  ->  Y = 0
	;	lt_pair(E, P, S) -> Y = 0
	;  	bypass_via_node(Bypass, E, L, L0, S),
		(	E = P ->	R0 = 0
		;  	composable_pairs(E, P, U, V) ->
			subst_node(Bypass, U, V, R, R0, S)
		;	bypass_via_node(Bypass, E, R, R1, S),
			zdd_insert(P, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(1), X, Y, p(2))))).
% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(0), X, Y, p(3))))).
% ?- zdd((X<<{[p(1)-p(2), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5)], [p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(5), p(2)-p(2)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4), p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(5), X, Y, p(5))), psa(Y))).

% ?- count_rect_cycles(rect(10, 1), C, S).
% ?- count_rect_cycles(rect(1,1), C, S).
% ?- R=rect(1,1), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R), active_node(P, Q)), S),
%  maplist(writeln, S).

% ?- R=rect(10,10), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R),
%	active_node(P, Q)), S),
%   maplist(writeln, S), fail; true.

% ?- active_node_orth(p(0,0),p(0,0)).  % true (conentionally)
% ?- active_node_orth(p(0,0),p(0,1)).
% ?- active_node_orth(p(0,0),p(1,0)).
% ?- active_node_orth(p(0,0),p(1,1)).
% ?- active_node_orth(p(0,1),p(1,0)).  % true.
% ?- active_node_orth(p(0,1),p(1,1)).
% ?- active_node_orth(p(1,0),p(1,1)).  % true

% active_node_orth(P, Q) is true if Q is directly linked from some P' < P.
active_node_orth(p(0, J), p(0, L)):-!, L = J.
active_node_orth(p(I, J), p(I, L)):-!, J =< L.
active_node_orth(p(I, J), p(K, L)):- K is I+1, !, L < J.

% ?- rect_nodes(rect(10,10), Ns), findall(Q, ( member(P-_, Ns),
%	orth_skew(P, Q)), Sk), sort(Sk, SSk),
%	findall(active(U, V), (member(U, SSk), member(V, SSk), active_node_skew(U, V)), W), maplist(writeln, W), fail; true.

% ?- active_node_skew(p(0,0), p(0, 0)).  % true
% ?- active_node_skew(p(0,0), p(0, 1)).  % false (vacuously)
% ?- active_node_skew(p(1,1), p(2, 1)).  % true
% ?- active_node_skew(p(1,0), p(1, 1)).  % true
% ?- active_node_skew(p(1,1), p(2, 2)).  % false
% ?- active_node_skew(p(1,1), p(2, 1)).  % true.
% ?- active_node_skew(p(1,0), p(2, 1)).  % false
% ?- active_node_skew(p(1,1), p(2, 0)).  % true
active_node_skew(p(I, J), p(I, L)):- !, J =< L. % Actually not ncessary clause ?
active_node_skew(p(I, J), p(K, L)):- K is I + 1,
			(	 L = J ->  J > 0
			;	 L < J
			).
%
prune_inactive_node(_, X, X, _):- X<2, !.
prune_inactive_node(P, X, Y, S):-
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	; 	prune_inactive_node(P, L, L0, S),
		A = (U - V),
		(	active_node(P, U) ->
			(	active_node(P, V) ->
				prune_inactive_node(P, R, R1, S),
				zdd_insert(A, R1, R0, S)
			;	R0 = 0
			)
		;	U = V ->
			prune_inactive_node(P, R, R0, S)
		;	R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).
