:- module(minato_r5, [rect_udg/3]).

% (start at copy of minato-r4)

:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).
:- op(1060, xfy, #).		% exor
:- op(1000, xfy, &).

% term_expansion --> pac:expand_pac.
term_expansion --> expand_pac.

		/******************
		*     Switches    *
		******************/
% ORTHOGONAl
% active_node(X, Y):- active_node_orth(X, Y).
% rect_nodes(X, Y):- rect_nodes_orth(X, Y).

% SKEW.
active_node(X, Y):- active_node_skew(X, Y).
rect_nodes(X, Y):- rect_nodes_skew(X, Y).

% ?- minato_r5(rect(0,0), C, _).
% ?- minato_r5(rect(1,0), C, _).
% ?- minato_r5(rect(0,1), C, _).
% ?- minato_r5(rect(1,1), C, _).
% ?- minato_r5(rect(1,2), C, _).
% ?- minato_r5(rect(2,1), C, _).
% ?- minato_r5(rect(2,2), C, _).
% ?- minato_r5(rect(1,10), C, _).
% ?- minato_r5(rect(10,1), C, _).
% ?- minato_r5(rect(100,1), C, _).
% ?- minato_r5(rect(1,100), C, _).

% ?- time(minato_r5(rect(6,6), C, _)).
%@ % 25,642,240 inferences, 3.558 CPU in 4.465 seconds (80% CPU, 7207373 Lips)
%@ C = 575780564.
% ?- time(minato_r5(rect(7,7), C, _)).
%@ % 106,041,498 inferences, 12.348 CPU in 12.454 seconds (99% CPU, 8587850 Lips)
%@ C = 789360053252.
% ?- time(minato_r5(rect(8,8), C, _)).
%@ % 877,350,286 inferences, 78.741 CPU in 79.304 seconds (99% CPU, 11142281 Lips)
%@ C = 3266598486981642.
%@ % 436,367,556 inferences, 52.415 CPU in 52.981 seconds (99% CPU, 8325235 Lips)
%@ C = 3266598486981642.
% ?- time(minato_r5(rect(9,9), C, _)).
%@ % 1,793,021,796 inferences, 224.894 CPU in 227.462 seconds (99% CPU, 7972753 Lips)
%@ C = 41044208702632496804.

% ?- call_with_time_limit(1800, time(minato_r5(rect(10,10), C, S))).
%@ % 12,782,653,306 inferences, 1164.030 CPU in 1174.499 seconds (99% CPU, 10981375 Lips)
%@ C = 1568758030464750013214100,

% ?- call_with_time_limit(36000, minato_r5(rect(11,11), C, _)).
%@ C = 182413291514248049241470885236.

% ?- convert_skew(p(1,1), Q), convert_skew(P, Q), convert_skew(P, R).
%@ Q = R, R = p(2, 1),
%@ P = p(1, 1).
orth_skew(p(I, J), Y):- var(Y), !,
	Y = p(K, I),
	K is I + J.
orth_skew(p(I, J), p(K, I)):- J is K - I.

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
change_symbol(A-B, A->B).
arrow_symbol( _ -> _).
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).
%
is_link(->(_, _)).
%
convert_normal_pair(P, Q, S):- orth_skew_pair(P, Q0),
	normal_pair(Q0, Q, S).
%
normal_arrow(X, Y):- @(normal_arrow(X, Y)).
%
normal_arrow(A->B, Arr, S):-!,
	(	zdd_compare(>, A, B, S) -> Arr = (B->A)
	;	Arr = (A->B)
	).
normal_arrow(A-B, Arr, S):-normal_arrow(A->B, Arr, S).
%

lt_pair(_ - A, B -_, S):- zdd_compare(<, A, B, S).

		/*************************************
		*     basics of incrementing DG      *
		*************************************/

composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(P, U):- @(normal_pair(P, U)).
%
normal_pair(A-B, B-A, S):- zdd_compare(>, A, B, S), !.
normal_pair(A->B, B->A, S):- zdd_compare(>, A, B, S), !.
normal_pair(P, P, _).

% ?- orth_skew_pair(p(1,1)-p(2,2), X), orth_skew_pair(Y, X).
%@ X = p(2, 1)-p(4, 2),
%@ Y = p(1, 1)-p(2, 2).
orth_skew_pair(P, Q):-
	(	P = (A-B) -> Q = (A0-B0)
	;	P = (A->B),
		Q = (A0->B0)
	),
	orth_skew(A, A0),
	orth_skew(B, B0).
% ?- link_sort_opp([loop, a, loop], X).
link_sort_opp(X, Y):- sort(X, X0), reverse(X0, Y).

% ?- node_ends(p(0,0), En, rect(1,1)).
% ?- node_ends(p(0,0), En, rect(0,0)).
% ?- node_ends(p(1,1), En, rect(2,2)).

node_ends(P, En, rect(W,H)):- P = p(I,J),
	(	I < W ->
		K is I + 1,
		En0 = [p(K,J)]
	;	En0 = []
	),
	(	J < H ->
		L is J + 1,
		En = [p(I, L)|En0]
	;   En = En0
	).

% ?- node_ends_skew(p(0,0), En, rect(1,1)).
% ?- node_ends_skew(p(0,0), En, rect(0,0)).
% ?- node_ends_skew(p(1,1), En, rect(2,2)).
% ?- node_ends_skew(p(1,1), En, rect(1,2)).
node_ends_skew(P, En, rect(W,H)):- P = p(I,J),
	(	J < W ->
		K is J + 1,
		M is I + 1,
		En0 = [p(M, K)]
	;	En0 = []
	),
	(	I < H + J ->
		L is I + 1,
		En = [p(L, J)|En0]
	;   En = En0
	).

% ?- R=rect(2,2), repeat(in_rect(P, R),
%	(node_ends_shadows(P, En, Sh, R), writeln(P-En-Sh))).
in_rect(p(I, J), rect(W, H)):- between(0, W, I), between(0, H, J).

% ?- rect_nodes_orth(rect(1,1), PES), maplist(writeln, PES).
%@ p(0,0)-[p(0,1),p(1,0)]
%@ p(0,1)-[p(1,1)]
%@ p(1,0)-[p(1,1)]
%@ p(1,1)-[]
%@ PES = [p(0, 0)-[p(0, 1), p(1, 0)], p(0, 1)-[p(1, 1)], p(1, 0)-[p(1, 1)], p(1, 1)-[]].

% ?- rect_udg(rect(1,1), Links, P).
%@ Links = [p(0, 0)-p(0, 1), p(0, 0)-p(1, 0), p(0, 1)-p(1, 1), p(1, 0)-p(1, 1)],
%@ P = p(0, 0)-p(1, 1).

rect_udg(rect(W, H), Links, p(0,0)-p(W,H)):-
	rect_nodes_orth(rect(W, H), PES),
	findall(U-V, ( member(U-L, PES), member(V, L), U @< V ), Links).

% ?- rect_nodes_orth(rect(3,3), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(10,10), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(30,30), PES), length(PES, L).
% ?- rect_nodes_orth(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_orth(rect(1000,1000), PES), length(PES, L))).

rect_nodes_orth(R, PES):-
	findall(P-E, (in_rect(P, R), node_ends(P, E, R)), PES).

% ?- rect_nodes_skew(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_skew(rect(1000,1000), PES), length(PES, L))).
rect_nodes_skew(R, PES_skew):- rect_nodes_orth(R, PES_orth),
	convert_orth_skew(PES_orth, PES_skew).
%
convert_orth_skew([], []).
convert_orth_skew([A-B|C], [A0-B0|C0]):-
	orth_skew(A, A0),
	maplist(orth_skew, B, B0),
	convert_orth_skew(C, C0).

		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

% ?- zdd((X<<{[c-d]}, @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _):- X < 2, !.
subst_node(Es, A, P, X, Y, S):-  	/* replace A with P */
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	zdd_compare(<, A, Lu, S) ->	Y = 0
	;	subst_node(Es, A, P, L, L0, S),
		(	Ru = A	->
			normal_pair(Lu-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(Ru-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[a-b, c-d]}, psa(X), @(add_links([b-c], X-X0, 0-Y)), psa(Y))).
% ?- zdd((X<<{[a-b, c-d]}, psa(X), @(add_links([b-c], X-X0, 0-Y)), psa(Y))).
% ?- zdd((X<<{[a-b, c-c, a->x, b->x]}, psa(X), @(add_links([a-b], X-X0, 0-Y)), psa(Y))).
add_links([], X, X, _).
add_links([L|Ls], X, X0, S):-
	add_link(L, X, X1, S),
	zdd_join(X, X1, X2, S),
	add_links(Ls, X2, X0, S).

% ?- zdd((X<<{[a-b, c-d]}, psa(X), @(add_link(b-c, X, Y)), psa(Y))).
add_link(_, X, 0, _):- X<2, !.
add_link(U, X, Y, S):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = 0
	;	add_link(U, L, L0, S),
		U = Ul-Ur,
		(	lt_pair(U, A, S) -> R0 = 0
		; 	A = U -> R0 = 0
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node([Ul->Ur], U0, V0, R, R0, S)
		;	add_link(U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

		/*************************
		*     count_rect_path    *
		*************************/

% ?- N = 3,  minato_r5(rect(0, N), C).
% ?- N = 3, minato_r5(rect(1, N), C).
% ?- N = 10, minato_r5(rect(1, N), C).
% ?- minato_r5(rect(1,1), C, S).

% ?- minato_r5(rect(1,1), C, S).
% ?- minato_r5(rect(3,1), C, S).
% ?- minato_r5(rect(2,2), C, S).
% ?- minato_r5(rect(2,1), C, S).
% ?- minato_r5(rect(3,1), C, S).
% ?- minato_r5(rect(3,3), C, S).
% ?- minato_r5(rect(4,4), C, S).
% ?- call_with_time_limit(200, time(minato_r5(rect(8,8), C, _))).
%@ % 908,479,618 inferences, 89.804 CPU in 91.150 seconds (99% CPU, 10116254 Lips)
%@ C = 3266598486981642.
% ?- trace.
% ?- spy(node_one_by_one).
% ?- minato_r5(rect(0,0), C).
% ?- minato_r5(rect(1,0), C).
% ?- minato_r5(rect(0,1), C).
% ?- minato_r5(rect(1,1), C).
% ?- minato_r5(rect(1,1), C, X).
% ?- minato_r5(rect(1,1), C, _).
% ?- minato_r5(rect(2,2), C, _).
% ?- minato_r5(rect(3,3), C, _).
% ?- minato_r5(rect(4,4), C, _).
% ?- minato_r5(rect(5,5), C, _).
% ?- minato_r5(rect(6,6), C, _).
% ?- time(minato_r5(rect(2,2), C, _)).
% ?- time(minato_r5(rect(9,9), C, _)).
%@ % 3,992,214,479 inferences, 377.403 CPU in 379.840 seconds (99% CPU, 10578111 Lips)
%@ C = 41044208702632496804.
%@ % 3,992,214,479 inferences, 391.036 CPU in 394.074 seconds (99% CPU, 10209317 Lips)
%@ C = 41044208702632496804.

% ?- time(minato_r5(rect(10,10), C, _)).
%@ % 18,246,128,207 inferences, 1826.781 CPU in 1842.135 seconds (99% CPU, 9988130 Lips)
%@ C = 1568758030464750013214100.
%@ % 18,246,106,192 inferences, 1871.297 CPU in 1889.274 seconds (99% CPU, 9750514 Lips)
%@ C = 1568758030464750013214100.

minato_r5(Rect, C):- minato_r5(Rect, C, _).

%
minato_r5(Rect, C, Opts):-
	memberchk(state(S), Opts),
	memberchk(out(X), Opts),
	open_state(S),
	set_compare(compare, S),
	set_key(rect, Rect, S),
	rect_nodes(Rect, A),
	sort(A, A1),
	reverse(A1, B),
	B =[Max-_|_],
	arg(1, Max, K),
	set_key(max, Max, S),
	set_key(prune_index, K, S),
	node_one_by_one(B, 1, X, S),
	card(X, C, S).

p(Opts):- subset([out(X), state(S)], Opts),
		  psa(X, S).

% ?- spy(add_node_one).
% ?- spy(add_link).
% ?- spy(prune_inactive_node).

% ?- minato_r5(rect(0,0), C, S).
% ?- minato_r5(rect(0,1), C, S), p(S).
% ?- minato_r5(rect(1,0), C, S).
% ?- minato_r5(rect(1,1), C, S), p(S).
% ?- minato_r5(rect(1,2), C, S).
% ?- minato_r5(rect(2,1), C, S).
% ?- minato_r5(rect(2,2), C, S).
% ?- minato_r5(rect(2,3), C, S).
% ?- minato_r5(rect(3,2), C, S).
% ?- minato_r5(rect(3,3), C, S).
% ?- minato_r5(rect(4,4), C, S).
% ?- minato_r5(rect(5,5), C, S).
% ?- minato_r5(rect(6,6), C, S).
% ?- minato_r5(rect(7,7), C, S).
%@ C = 789360053252,
%@ S = [state(..), out(57444)|_].
% ?- zdd(@(node_one_by_one([], 1, X))).
node_one_by_one([], X, X, _).
node_one_by_one([P-En|R], X, X0, S):-
	zdd_insert(P-P, X, XP, S),
	add_node_one(P, En, XP, X1, S),
	prune_inactive_node(P, X1, X2, S),
	get_key(prune_index, I, S),
	arg(1, P, J),
	(	J = I ->  X3 = X2
	;	set_key(prune_index, J, S),
		format("Slimming at ~w.\n", [P]),
		zdd_slim(X2, X3, S),
		garbage_collect
	),
	node_one_by_one(R, X3, X0, S).
%
is_max_node(P, S):- get_key(max, P, S).


% ?- zdd((X<<{[b-c, b->c]}, psa(X),
%  @(add_node_one(a, [b], X, Y)), psa(Y))).
% ?- time(minato_r5(rect(8,8), C, S)).
%@ % 436,367,556 inferences, 35.323 CPU in 36.801 seconds (96% CPU, 12353625 Lip
%@ C = 3266598486981642,
% ?- time(minato_r5(rect(1,1), C, S)).

add_node_one(_, [], X, X, _):-!.
add_node_one(P, [A], U, V, S):-!, add_links([P-A], U, V, S).
add_node_one(P, [A, B], U, V, S):- add_links([P-A, P-B], U, V, S).

% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(1), X, Y, p(2))))).
% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(0), X, Y, p(3))))).
% ?- zdd((X<<{[p(1)-p(2), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5)], [p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(5), p(2)-p(2)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4), p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(5), X, Y, p(5))), psa(Y))).

% ?- minato_r5(rect(10, 1), C, S).
% ?- minato_r5(rect(1,1), C, S).
% ?- R=rect(1,1), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R), active_node(P, Q)), S),
%  maplist(writeln, S).

% ?- R=rect(10,10), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R),
%	active_node(P, Q)), S),
%   maplist(writeln, S), fail; true.

% ?- active_node_orth(p(0,0),p(0,0)).  % true (conentionally)
% ?- active_node_orth(p(0,0),p(0,1)).
% ?- active_node_orth(p(0,0),p(1,0)).
% ?- active_node_orth(p(0,0),p(1,1)).
% ?- active_node_orth(p(0,1),p(1,0)).  % true.
% ?- active_node_orth(p(0,1),p(1,1)).
% ?- active_node_orth(p(1,0),p(1,1)).  % true
%% active_node_orth(P, Q) is true if Q is directly linked from some P' < P.

active_node_orth(p(0, J), p(0, L)):-!, L = J.
active_node_orth(p(I, J), p(I, L)):-!, J =< L.
active_node_orth(p(I, J), p(K, L)):- K is I+1, !, L < J.

% ?- rect_nodes(rect(10,10), Ns), findall(Q, ( member(P-_, Ns),
%	orth_skew(P, Q)), Sk), sort(Sk, SSk),
%	findall(active(U, V), (member(U, SSk), member(V, SSk), active_node_skew(U, V)), W), maplist(writeln, W), fail; true.

% ?- active_node_skew(p(0,0), p(0, 0)).  % true
% ?- active_node_skew(p(0,0), p(0, 1)).  % false (vacuously)
% ?- active_node_skew(p(1,1), p(2, 1)).  % true
% ?- active_node_skew(p(1,0), p(1, 1)).  % true
% ?- active_node_skew(p(1,1), p(2, 2)).  % false
% ?- active_node_skew(p(1,1), p(2, 1)).  % true.
% ?- active_node_skew(p(1,0), p(2, 1)).  % false
% ?- active_node_skew(p(1,1), p(2, 0)).  % true
active_node_skew(p(I, J), p(I, L)):- !, J =< L. % Actually not ncessary clause ?
active_node_skew(p(I, J), p(K, L)):- K is I + 1,
			(	 L = J ->  J > 0
			;	 L < J
			).
%
prune_inactive_node(P, X, Y, S):-  get_key(max, Max, S),
	prune_inactive_node(P, X, Y, Max, S).

%
prune_inactive_node(_, X, X, _, _):- X<2, !.
prune_inactive_node(P, X, Y, Max, S):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	; 	prune_inactive_node(P, L, L0, Max, S),
		A = (U - V),
		(	active_node(P, U) ->
			(	( active_node(P, V); V = Max )	 ->
				prune_inactive_node(P, R, R1, Max, S),
				zdd_insert(A, R1, R0, S)
			;	R0 = 0
			)
		;  U = Max -> R0 = 0
		;  U = V -> prune_inactive_node(P, R, R0, Max, S)
		;  R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).
