:- module(path_count7, [zdd0/1, co_compare/3]).

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
- use_module(util(meta2).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(zdd(zdd)).
:- use_module(zdd('update-link')).
:- use_module(pac(op)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

%@ true.
% ?- zdd0((true)).
% ?- zdd0((X<<pow([a,b]),  psa(X))).

:- meta_predicate zdd0(:).
zdd0(X):- zdd((	set_compare(co_compare), X)).

% ?- co_compare(C, c-d, c->d).
% ?- co_compare(C, c->d, c->d).
% ?- co_compare(C, c->d, c-d).
% ?- co_compare(C, c-d, c-d).
% ?- predsort(co_compare, [a->b, b->a], R).
% ?- predsort(co_compare, [a-b, b-a], R).
% ?- predsort(co_compare, [b->a, a->b, a-b, a->b, c-b, a-b, b-a], R).
% ?- zdd((set_compare(co_compare), zdd_compare(C, a-b, c-d))).
% ?- zdd((set_compare(co_compare), zdd_compare(C, c-d, c->d))).
% ?- zdd((set_compare(co_compare), zdd_compare(C, c->d, c-d))).
% ?- zdd((set_compare(co_compare), zdd_insert_atoms([a-b, c-d], 1, X))).
% ?- zdd((set_compare(co_compare), zdd_insert_atoms([a->b, c-d], 1, X), psa(X))).

co_compare(C, X, Y):- compare(C0, X, Y),
	(	C0 = (=)  -> C = (=)
	;	functor(X, Fx, _),
		functor(Y, Fy, _),
		(	Fx == Fy
		-> 	(	Fx = (-)
			->  ( C0 = (<) -> C = (>) ; C = (<) )
			;	C = C0
			)
		;	(	Fx = (-) -> C = (<) ; C = (>) )
		)
	).

% ?- predsort(co_compare, [a->b, b->a, a-b, a->b, c-b, b-a, a-b ], R).
% ?- predsort(co_compare, [a->b, b->a], R).

% some tiny.
% x(p(X, _), X).
% y(p(_, Y), Y).
% link_symbol(_->_).
arrow_symbol( _ -> _).
change_symbol(A - B, A -> B).
%
touch(A - _, _ - A):-!.
touch(_ - A, A - _).
% ?- connect_pairs(a-b, [b-c, d-a], W).
% ?- connect_pairs(a-b, [a-c, a-b], W).  % false
connect_pairs(X-Y, [U-X, Y-V], U-V):-!.
connect_pairs(X-Y, [Y-V, U-X], U-V).

%
prune_frontier_by_level(F, K, P, F0):-
	@(prune_frontier_by_level(F, K, P, F0)).

% For rect only.
prune_frontier_by_level(F, _, _, F, _):- F<2, !.
prune_frontier_by_level(F, K, P, F0, S):- s_cofact(F, t(A, L, R), S),
	prune_frontier_by_level(L, K, P, L0, S),
	(	arrow_symbol(A)
	->	U = R
	;	(	(	A = (X-X)
			;	A = P - p(K,_)
			; 	A = p(K, _)-p(K, _)
			)
		->	prune_frontier_by_level(R, K, P, U, S)
		;	U = 0
		)
	),
	s_zdd_insert(A, U, R0, S),
	s_zdd_join(L0, R0, F0, S).

card_path(X, ST, Y):- @(card_path(X, ST, Y)).
%

card_path(X, ST, Y, S):- open_state(M, [hash_size(128)]),
	card_path(X, ST, Y, S, M),
	close_state(M).

%
card_path(I, _, 0, _, _):- I < 2, !.
card_path(I, ST, C, S, M):- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_path(L, ST, Cl, S, M),
		(	A = X-Y
		->	(	X = Y	-> card_path(R, ST, Cr, S, M)
			;  	A = ST	-> card_path1(R, Cr, S, M)
			;	Cr = 0
			)
		;  Cr = 0
		),
		C is Cr + Cl
	).

card_path1(I, I, _, _):- I<2, !.
card_path1(I, C, S, M):- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_path1(L, Cl, S, M),
		(  	A = X-Y
		->  (	X = Y ->  card_path1(R, Cr, S, M)
			;	Cr = 0
			)
		;	card(R, Cr, S, M)
		),
		C is Cr + Cl
	).

% ?- bridge_shift(h, [p(1,1)-p(1,1)], X).
% ?- bridge_shift(v, [p(1,1)-p(1,1)], X).

bridge_shift(F, X, Y):- maplist(shift_link(F), X, Y).

% ?- zdd((convert_to_col([[1-2], [3-4]], Col, 0),
%	map_h_shift(Col, Col0))).
map_h_shift(X, Y):- @(map_h_shift(X, Y)).
%
map_h_shift([], [], _).
map_h_shift([X|Y], [X0|Y0], S):- h_shift(X, X0, S),
	map_h_shift(Y, Y0, S).

% ?- zdd((convert_to_row([[1-2], [3-4]], Row, 0),
%	map_v_shift(Row, Row0))).
map_v_shift(X, Y):- @(map_v_shift(X, Y)).
%
map_v_shift([], [], _).
map_v_shift([X|Y], [X0|Y0], S):- v_shift(X, X0, S),
	map_v_shift(Y, Y0, S).

%
h_shift(X, Y):- @(h_shift(X, Y)).
%
h_shift(X, X, _):- X < 2, !.
h_shift(X, Y, S):- cofact(X, t(p(I, J)-p(I, K), L, R), S),
	I0 is I + 1,
	h_shift(L, L0, S),
	h_shift(R, R0, S),
	cofact(Y, t(p(I0, J)-p(I0, K), L0, R0), S).

v_shift(X, Y):- @(v_shift(X, Y)).
%
v_shift(X, X, _):- X < 2, !.
v_shift(X, Y, S):- cofact(X, t(p(I, J)-p(K, J), L, R), S),
	v_shift(L, L0, S),
	v_shift(R, R0, S),
	J0 is J + 1,
	cofact(Y, t(p(I, J0)-p(K, J0), L0, R0), S).

% ?- gogo.
% ?- zdd0((solve_rect(rect(1,1), Z, C))).
% ?- zdd0((solve_rect(rect(2,2), Z, C))).
%@ C = 12 .
% ?- time(zdd0((solve_rect(rect(3,3), Z, C)))).
%@ % 37,629,160 inferences, 5.016 CPU in 5.895
%@ Z = 455351,
%@ C = 184 .
%@ % 37,629,160 inferences, 3.744 CPU in 3.945 seconds (95% CPU, 10049933 Lips)
%@ Z = 455351,
%@ C = 184 .
% ?- time(zdd0((solve_rect(rect(4,4), Z, C)))).
%@ % 15,026,838,147 inferences, 2173.210 CPU in 14110.223
%@ Z = 154551488,
%@ C = 8512 .
%@ % 6,527,546,920 inferences, 915.198 CPU in 956.075
%@ Z = 70016262,
%@ C = 7888 .
%@ path-count7.pl qcompiled.

% ?- zdd0((solve_rect(rect(0,0), Z, C))).
% ?- zdd0((solve_rect(rect(0,1), Z, C))).
% ?- zdd0((solve_rect(rect(0,2), Z, C))).
% ?- zdd0((solve_rect(rect(0,3), Z, C))).
% ?- zdd0((solve_rect(rect(0,15), Z, C))).
% ?- zdd0((solve_rect(rect(1, 0), Z, C))).
% ?- zdd0((solve_rect(rect(2, 0), Z, C))).
% ?- zdd0((solve_rect(rect(3, 0), Z, C))).
% ?- zdd0((solve_rect(rect(10, 0), Z, C))).
% ?- zdd0((solve_rect(rect(15, 0), Z, C))).
% ?- zdd0((solve_rect(rect(20, 0), Z, C))).

% ?- zdd0((solve_rect(rect(1,1), Z, C))).
% ?- zdd0((solve_rect(rect(2,1), Z, C))).
%@ C = 4 .
% ?- zdd0((solve_rect(rect(3,1), Z, C))).
% ?- zdd0((solve_rect(rect(4,1), Z, C))).
% ?- zdd0((solve_rect(rect(5,1), Z, C))).
% ?- zdd0((solve_rect(rect(10,1), Z, C))).
% ?- zdd0((solve_rect(rect(16,1), Z, C))).
% ?- zdd0((solve_rect(rect(20,1), Z, C))).
%@ Z = 116399365,
%@ C = 1048576 .

% ?- zdd0((solve_rect(rect(3,3), Z, C))).
%@ Z = 802789,
%@ C = 184 .
%@ Z = 726823,
%@ C = 184 .

% ?- time(zdd0((solve_rect(rect(4,4), Z, C)))).
%@ % 15,026,838,146 inferences, 2128.657 CPU in 2589.153
%@ Z = 154551488,
%@ C = 8512 .
%@ % 13,320,352,776 inferences, 1920.237 CPU in 2125.245
%@ Z = 146748055,
%@ C = 8512 .

% ?- time(zdd0((solve_rect(rect(5,5), Z, C)))).

% ?- zdd((X<<{[c,b,a]}, Y<<{[],[c],[b,c],[a,b,c]}, zdd_merge(X, Y, Z), psa(Z))).

% gogogogo
% ?- zdd0((solve_rect(rect(1, 1), Z, C))).
%@ Z = 184,
%@ C = 2 .
% ?- zdd0((solve_rect(rect(1, 2), Z, C))).
%@ Z = 1886,
%@ C = 4 .
% ?- zdd0((solve_rect(rect(1, 3), Z, C))).
%@ Z = 16497,
%@ C = 8 .
% ?- zdd0((solve_rect(rect(1, 4), Z, C))).
%@ Z = 138089,
%@ C = 16 .
% ?- zdd0((solve_rect(rect(1, 5), Z, C))).
%@ Z = 1152578,
%@ C = 32 .
% ?- zdd0((solve_rect(rect(1, 6), Z, C))).
%@ Z = 12510705,
%@ C = 64 .
%@ Z = 12447245,
%@ C = 64 .
%@ Z = 9695864,
%@ C = 64 .

% ?- call_with_time_limit(600, zdd0((solve_rect(rect(2,2), Z, C)))).
%@ Z = 10622,
%@ C = 12.
%@ Z = 7273,
%@ C = 12.

% ?- call_with_time_limit(600, zdd0((solve_rect(rect(3,3), Z, C)))).
% ?- call_with_time_limit(3600, zdd0((solve_rect(rect(4,4), Z, C)))).
% ?- call_with_time_limit(5400, zdd0((solve_rect(rect(5,5), Z, C)))).

solve_rect(R, C):- solve_rect(R, _, C).
solve_rect(rect(W, H), Z, C):- solve_rect(p(0,0), p(W, H), Z, C).
%
solve_rect(X, Y, Z, C):- @(solve_rect(X, Y, Z, C)).
%
solve_rect(p(I, J), p(I, L), Z, 1, S):-!, initial_column(I, J, L, Z, S).
solve_rect(p(I, J), p(K, L), Z, C, S):-
	initial_col_bridge(I, J, L, Col, B0, S),
	reverse(B0, B),
	N is (K-I),
	repeat_bridge(N, Col, I, p(I,J), Col, B, Z, S),
	card_path(Z, p(I,J)-p(K, L), C, S).

%@ true.
%@ path-count7.pl qcompiled.


% ?- dg_path_count_naive(dg([a,b],[a-b]), a-b, C).
% ?- dg_path_count_naive(dg([a,b],[b-a]), a-b, C).
% ?- dg_path_count_naive(dg([a, b, c],[a-b, b-c]), a-c, C).
% ?- dg_path_count_naive(dg([a, b, c],[a-b, b-c, a-c]), a-c, C).
% ?- dg_path_count_naive(dg([a, b, c, d],[a-b, a-c, b-a, b-d, c-a, c-d, d-c, d-b]), a-d, C).
% ?- dg_path_count_naive(dg([a, b, c, d],[b-c, c-b, a-b, a-d, a-c, b-a, b-d, c-a, c-d, d-c, d-b, d-a]), a-d, C).
%@ C = 5 .

% ?- rect_dg(rect(1,1), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(1,1), C).
% ?- rect_dg(rect(2,2), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(2,2), C).
%@ C = 12 .
% ?- rect_dg(rect(1,2), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(1,2), C).
% ?- rect_dg(rect(1,3), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(1,3), C).
% ?- rect_dg(rect(1,5), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(1,5), C).
% ?- call_with_time_limit(100, (rect_dg(rect(3,3), X, Y), dg_path_count_naive(dg(X, Y), p(0,0)-p(3,3), C))).  % timeout.


% ?- rect_dg(rect(1,2), X, Y), dg_path_naive(dg(X, Y)).
% ?- rect_path_count_naive(rect(1,2), C).
% ?- rect_path_count_naive(rect(1,3), C).
% ?- rect_path_count_naive(rect(3,1), C).
% ?- rect_path_count_naive(rect(2,2), C).
% ?- rect_path_count_naive(rect(2,3), C).
% ?- rect_path_count_naive(rect(3,2), C).
% ?- rect_path_count_naive(rect(3,3), C).


rect_path_count_naive(rect(W, H), C):-
	rect_dg(rect(W, H), X, Y),
	dg_path_count_naive(dg(X, Y), p(0,0)-p(W, H), C).

dg_path_count_naive(G, ST, C):- dg_path_count_naive(G, ST, _, C).

dg_path_count_naive(dg(Ns, Es), ST, Fin, C):- zdd0(
	(
		zdd_sort(Es, E0s),
		findall(X-X, member(X, Ns), X_Xs),
		zdd_insert_atoms(X_Xs, 1, Initial),
		update_links(E0s, Initial, Fin),
		card_path(Fin, ST, C)
	)).

% ?- dg_path_naive(dg([a, b, c, d],[a-b, a-c, b-a, b-d, c-a, c-d, d-c, d-b])).
dg_path_naive(dg(Ns, Es)):- zdd0(
	(	zdd_sort(Es, E0s),
		findall(X-X, member(X, Ns), Id),
		zdd_insert_atoms(Id, 1, Initial),
		update_links(E0s, Initial, Fin),
		psa(Fin)
	)).

% ?- zdd0(zdd_sort([a, b, c], X)).
% ?- zdd0((solve_rect(rect(1, 1), Z, C))).
% ?- zdd0((solve_rect(rect(1, 2), Z, C))).
% ?- zdd0((solve_rect(rect(2, 1), Z, C))).
% ?- zdd0((solve_rect(rect(3, 1), Z, C))).
% ?- zdd0((solve_rect(rect(8, 1), Z, C))).
% ?- zdd0((solve_rect(rect(16, 1), Z, C))).
% ?- zdd0((solve_rect(rect(1, 2), Z, C))).
% ?- zdd0((solve_rect(rect(1, 3), Z, C))).
% ?- listing(s_zdd_insert).
% ?- listing(zdd_insert).


% ?- zdd0((solve_rect(rect(2, 2), Z, C))).
% ?- time(zdd0((solve_rect(rect(3, 3), Z, C)))).  % links_across.
%@ % 81,838,566 inferences, 10.673 CPU in 12.942 seconds (82% CPU, 7668071 Lips)
%@ Z = 726823,
%@ C = 184 .

% ?- time(zdd0((solve_rect(rect(3, 3), Z, C)))).   % glue_frontiers.
%%@ % 114,953,539 inferences, 13.781 CPU in 16.134 seconds (85% CPU, 8341355 Lips)
%@ Z = 802789,
%@ C = 184

% ?- time(zdd0((solve_rect(rect(4, 4), Z, C)))).
% ?- zdd0((solve_rect(rect(1, 2), Z, C))).
repeat_bridge(0, X, _, _, _, _, X, _):-!, writeln(bridge(0)).
repeat_bridge(N, X, K, P, C, B, Y, S):- writeln(bridge(N)),
	mqp_shift(h, C, C0, S),
	links_across(B, X, C0, X0, S),
%	glue_frontiers(B, X, C0, X0, S),
	maplist(shift_link(h), B, B0),
	K0 is K + 1,
	N0 is N - 1,
	prune_frontier_by_level(X0, K0, P, X1, S),
	repeat_bridge(N0, X1, K0, P, C0, B0, Y, S).

% ?- time(zdd0((initial_col_bridge(0, 0, 15, C, B)))).
% ?- time(zdd0((initial_col_bridge(0, 0, 2, C, B), psa(C)))).

initial_col_bridge(I, L, H, C, B):-	@(initial_col_bridge(I, L, H, C, B)).
%
initial_col_bridge(I, L, H, C, B, S):- initial_column(I, L, H, C, S),
	initial_bridge(I, L, H, B).

% ?- zdd0((initial_column(1, 2, 4, Col), s_psa(Col))).
% ?- zdd0((initial_column(1, 0, 10, Col), card(Col, C))).
% ?- zdd0((initial_column(1, 0, 3, Col), card(Col, C))).  % seems good.
% ?- zdd0((initial_column(1, 0, 3, Col),
%	mqp_shift(h, Col, Col1),
%	mqp_shift(v, Col1, Col2),
%	card(Col, C),
%	card(Col1, C1),
%	card(Col2, C2))).
% ?- zdd0((initial_col_bridge(0, 0, 3, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 4, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 5, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 15, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 2, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 15, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 20, Col, B), card(Col, C))).
% ?- zdd0((initial_col_bridge(0, 0, 10, C, B), card(C, Count))), length(B, Len).

initial_column(I, Low, Hi, Col):- @(initial_column(I, Low, Hi, Col)).
%
initial_column(I, Low, Hi, Col, S):-
	itr_linear_grid(Low, Hi, X, S),
	mqp_lift(x(I), X, Col, S).

% initial_column(I, Low, Hi, Col, S):-
% 	s_mqp_linear_grid(Low, Hi, X, S),
% 	mqp_lift(x(I), X, Col, S).


% ?- initial_bridge(0, 1, 4, Bridge), maplist(writeln, Bridge).
initial_bridge(I, Low, Hi, Bridge):- J is I + 1,
	findall(A,
			(	member(A, [p(I, V)-p(J, V), p(J, V)-p(I, V)]),
				between(Low, Hi, V)
			),
			Bridge0),
	sort(Bridge0, Bridge).


% ?- spy(mqp_linear_grid).
% ?- I = 0, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 1, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 2, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% Checked by hand.

% ?- forall(between(0, 20, I), zdd0((mqp_linear_grid(0,I,X), card(X, C), format("count(~w) = ~w\n", [I, C])))).

% ?- zdd0((mqp_linear_grid(0,1,X))).
% ?- zdd0((mqp_linear_grid(0,2,X))).
% ?- zdd0((mqp_linear_grid(0,3,X), card(X, C))).
% ?- zdd0((mqp_linear_grid(0,4,X), card(X, C))).

% ?- zdd0((initial_col_bridge(0, 0, 20, C, B), card(C, Count))), length(B, Len).
%@ C = 295985945,
%@ Count = 54608393,
%@ Len = 42.


% ?- zdd0((initial_col_bridge(0, 0, 20, C, B), card(C, Count))), length(B, Len).
%@ C = 26385389,
%@ Count = 4479227,
%@ Len = 42 .

% ?- zdd0((initial_col_bridge(0, 0, 4, C, B), card(C, Count))), length(B, Len).

		/*********************************************************
		*     mqp_linear_grid is better than itr_linear_grid.    *
		*********************************************************/



% ?- time(zdd0((mqp_linear_grid(0,15,X), card(X, C)))).
%@ % 204,734,142 inferences, 17.969 CPU in 18.20
%@ X = 1837608,
%@ C = 665857 .
% ?- time(zdd0((mqp_linear_grid(0,16,X), card(X, C)))).
%@ % 567,780,697 inferences, 112.569 CPU in 158.862
%@ X = 4391322,
%@ C = 1607521
% ?- time(zdd0((mqp_linear_grid(0,16,X), card(X, C)))).
%@ % 567,780,697 inferences, 161.364 CPU in 327.436
%@ X = 4391322,
%@ C = 1607521 .
% ?- time(zdd0((mqp_linear_grid(0,16,X), card(X, C)))).
%@ % 567,780,697 inferences, 94.204 CPU in 135.880
%@ X = 4391322,
%@ C = 1607521 .
% ?- time(zdd0((mqp_linear_grid(0,18,X), card(X, C)))).
%@ % 2,946,152,712 inferences, 388.382 CPU in 402.191
%@ X = 25469487,
%@ C = 9369319

% ?- time(zdd0((itr_linear_grid(0,15,X), card(X, C)))).
%@ % 211,111,055 inferences, 20.500 CPU in 21.370
%@ X = 1802950,
%@ C = 665857 .
% ?- time(zdd0((itr_linear_grid(0,16,X), card(X, C)))).
%@ % 586,094,784 inferences, 68.215 CPU in 87.623 seconds (78% CPU, 8591868 Lips)
%@ X = 4352186,
%@ C = 1607521 .
% ?- time(zdd0((itr_linear_grid(0,18,X), card(X, C)))).
%@ % 3,062,986,110 inferences, 378.321 CPU in 403.198
%@ X = 25364313,
%@ C = 9369319 .

% simple line quasi path
mqp_linear_grid(X, Y, Z):-
	@(mqp_linear_grid(X, Y, Z)).

mqp_linear_grid(E, X, Y, S):-atmark(S),!,
	s_mqp_linear_grid(E, X, Y, S).
mqp_linear_grid(E, X, Y, S):-
	open_state(M, [hash_size(128)]),
	s_mqp_linear_grid(E, X, Y, @(S, M)),
	close_state(M).
%
s_mqp_linear_grid(I, I, P, S):-!,
	s_zdd_singleton(I-I, P, S).
s_mqp_linear_grid(I, J, Q, S):-
	M is (I+J)//2,
	M0 is M+1,
	s_mqp_linear_grid(I, M, R, S),
	s_mqp_linear_grid(M0, J, R0, S),
%	s_zdd_merge(R, R0, U, S),
%	update_links([M-M0, M0-M], U, Q, S).
	links_across([M-M0, M0-M], R, R0, 0, Q0, S),
	s_zdd_merge(R, R0, U, S),
	s_zdd_join(U, Q0, Q, S).

		/*******************************
		*     iterative linear grid    *
		*******************************/
% ?- spy(update_links).
% ?- spy(zdd0).
% ?- zdd0((itr_linear_grid(0, 0, Z), psa(Z))).
% ?- zdd0((itr_linear_grid(0, 1, Z), psa(Z))).
% ?- zdd0((itr_linear_grid(0, 2, Z), psa(Z))).
% ?- zdd0((itr_linear_grid(0, 3, Z), card(Z, C))).
% ?- zdd0((itr_linear_grid(0, 15, Z), card(Z, C))).
%@ Z = 3823906,
%@ C = 665857
% ?- zdd0((itr_linear_grid(0, 5, Z), mqp_linear_grid(0,5,X))).
%@ Z = X, X = 312 .
%@ Z = X, X = 582 .
% ?- zdd0((itr_linear_grid(0, 15, Z), mqp_linear_grid(0,15,X))).
%@ Z = X, X = 1802905 .
%@ Z = X, X = 3823862 .
%@ Z = X, X = 3823906 .

itr_linear_grid(X, Y, Z):- @(itr_linear_grid(X, Y, Z)).

itr_linear_grid(X, Y, Z, S):- atmark(S), !,
	s_itr_linear_grid(X, Y, Z, S).
itr_linear_grid(X, Y, Z, S):-
	open_state(M, [hash_size(128)]),
	s_itr_linear_grid(X, Y, Z, @(S, M)),
	close_state(M).

% Best linear grid.
s_itr_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
s_itr_linear_grid(I, J, P, S):-J0 is J-1,
	s_itr_linear_grid(I, J0, Q, S),
	s_zdd_singleton(J-J, R, S),
	s_zdd_insert(J-J, Q, Q0, S),  %	equivalently, s_zdd_merge(R, Q, Q0, S),
	links_across([J0-J, J-J0], Q, R, Q0, P, S).

% %
% s_itr_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
% s_itr_linear_grid(I, J, P, S):-J0 is J-1,
% 	s_itr_linear_grid(I, J0, Q, S),
% 	s_zdd_singleton(J-J, U, S),
% 	links_across([J0-J, J-J0], Q, U, P, S).

% s_itr_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
% s_itr_linear_grid(I, J, P, S):-J0 is J-1,
% 	s_itr_linear_grid(I, J0, Q, S),
% 	s_zdd_insert(J-J, Q, Q0, S),
% 	links_across([J0-J, J-J0], Q0, P, S).


% works best
% s_itr_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
% s_itr_linear_grid(I, J, P, S):-J0 is J-1,
% 	s_itr_linear_grid(I, J0, Q, S),
% 	s_zdd_singleton(J-J, R, S),
% 	links_across([J0-J, J-J0], Q, R, 0, P, S).

% woks!
% s_itr_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
% s_itr_linear_grid(I, J, P, S):-J0 is J-1,
% 	s_itr_linear_grid(I, J0, Q, S),
% 	s_zdd_singleton(J-J, R, S),
% 	s_zdd_merge(R, Q, U, S),
% 	links_across([J0-J, J-J0], Q, R, U, P, S).


		/********************************
		*     lift/shift  quasi path    *
		********************************/

% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), card(X, D),  mqp_lift(y(0), X, Y), card(Y, C))).
%
mqp_lift(F, X, Y):- @(mqp_lift(F, X, Y)).
%
mqp_lift(F, X, Y, S):-open_state(M, [hash_size(128)]),
	mqp_lift(F, X, Y, S, M),
	close_state(M).
%
mqp_lift(_, X, X, _, _):- X<2, !.
mqp_lift(F, X, Y, S, M):- memo(mqp_lift(X)-Y, M),  % F dropped.
	(	nonvar(Y)-> true
	;	s_cofact(X, t(A, L, R), S),
		mqp_lift(F, L, L0, S, M),
		mqp_lift(F, R, R0, S, M),
		lift_link(F, A, B),
		s_cofact(Y, t(B, L0, R0), S)
	).

% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y), card(Y, C))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y), card(Y, C),
%	mqp_shift(v, Y, Z), psa(Z), card(Z, D))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y),
%	mqp_shift(v, Y, Z), mqp_shift(v, Z, U), psa(U), card(U, C))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(x(0), X, Y),
%	mqp_shift(h, Y, Z), mqp_shift(h, Z, U), psa(U), card(U, C))).
%
mqp_shift(F, X, Y):- @(mqp_shift(F, X, Y)).
%
mqp_shift(F, X, Y, S):-
	open_state(M, [hash_size(128)]),
	mqp_shift(F, X, Y, S, M),
	close_state(M).

%
mqp_shift(_, X, X, _, _):- X < 2, !.
mqp_shift(F, X, Y, S, M):- memo(mqp_shift(X)-Y, M),  % F dropped.
	(	nonvar(Y)-> true
	;	s_cofact(X, t(A, L, R), S),
		mqp_shift(F, L, L0, S, M),
		mqp_shift(F, R, R0, S, M),
		shift_link(F, A, B),
		s_cofact(Y, t(B, L0, R0), S)
	).

% ?- lift_point(3, x(2), R).
% ?- lift_point(3, y(2), R).
%
lift_point(J, x(I), p(I, J)):-!.
lift_point(J, y(I), p(J, I)).

% ?- lift_link(x(2), 4-5, R).
lift_link(Ctr, A-B, P-Q):-!, lift_point(A, Ctr, P),
	lift_point(B, Ctr, Q).
% ?- lift_link(x(2), 4->5, R).
lift_link(Ctr, A->B, P->Q):- lift_point(A, Ctr, P),
	lift_point(B, Ctr, Q).
%
shift_link(h, A, B):-!, h_shift_link(A, B).
shift_link(v, A, B):- v_shift_link(A, B).

% ?- h_shift_point(p(1,1), R).
% ?- h_shift_point(-1, p(1,1), R).
h_shift_point(p(I, J), p(I0, J)):- I0 is I + 1.
h_shift_point(K, p(I, J), p(I0, J)):- I0 is I + K.
%
v_shift_point(p(I, J), p(I, J0)):- J0 is J + 1.
v_shift_point(K, p(I, J), p(I, J0)):- J0 is J + K.

%
h_shift_link(P-Q, P0-Q0):-!, h_shift_point(P, P0),
	h_shift_point(Q, Q0).
h_shift_link(P->Q, P0->Q0):- h_shift_point(P, P0),
	h_shift_point(Q, Q0).
%
v_shift_link(P-Q, P0-Q0):-!, v_shift_point(P, P0),
	v_shift_point(Q, Q0).
v_shift_link(P->Q, P0->Q0):- v_shift_point(P, P0),
	v_shift_point(Q, Q0).

%% test basics.
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).

% ?- rect_dg(rect(0,1), X, Y).
% ?- rect_dg(rect(0,0), X, Y).
% ?- rect_dg(rect(1,1), X, Y).

rect_dg(rect(W, H), Nodes, Links):-
	rect_nodes(W, H, Nodes),
	rect_links(W, H, Nodes, Links).

% ?- rect_nodes(4,5,Ns), length(Ns, L).
rect_nodes(W, H, Nodes):-
	findall(p(I, J),
			(	between(0, W, I),
				between(0, H, J)
			),
			Nodes0),
	sort(Nodes0, Nodes).

%
rect_links(W, H, Ns, Links):-
	findall(P-Q,
			(	member(P, Ns),
				P = p(I, J),
				(	Q = p(I, J1),
					(	J1 is J-1,
						J1 >= 0
					;	J1 is J+1,
						J1 =< H
					)
				;   Q = p(I1, J),
					(	I1 is I-1,
						I1 >= 0
					;	I1 is I+1,
						I1 =< W
					)
				)
			),
			Links0),
	sort(Links0, Links).
