:- module(update_link, [
		links_across/4, links_across/5, links_across/6
	,	link_across_out/6, link_across_in/6
	,	merge_blocks/5
	,	card_hybrid/3, card_hybrid/4
	,	mortal_paths/3, mortal_paths/4, mortal/2
	,	replace_right/6, replace_left/6
	,   atmark/1, qp_connect/4
	,	repeat_update_mqp/3, repeat_update_mqp/4
	,	u_repeat_update_mqp/3, u_repeat_update_mqp/4
	,	update_link/3, update_link/4
	,	update_mqp/3, update_mqp/4, update_qp/3, update_qp/4
	,	u_update_qp/3, u_update_qp/4
	,	normal_pair/2, normal_pair/3
	,	update_links/3, update_links/4
	,	update_link_slim/3, update_link_slim/4
	,	update_links_slim/3, update_links_slim/4
	,	slim_mqp/3, slim_mqp/4
						]).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).

:- op(1060, xfy, #).		% exor
:- op(1000, xfy, &).

term_expansion --> pac:expand_pac.

% ?- zdd0((X<<pow([a,b]),  psa(X))).
% ?- q_compare(C, p(1, 1), p(0, 0)).
% ?- predsort(q_compare, [a->b, c-c, b], X).
% ?- predsort(q_compare, [a->b, c-d], X).
% ?- zdd0(zdd_sort([p(0,0), p(1,1)], X)).
% ?- zdd(zdd_sort([p(0,0), p(1,1)], X)).
% ?- zdd((set_compare(opp_compare), zdd_sort([p(1,0)->p(0,0), p(1,1)->p(0,1)], X))).
is_normal_order(S):- get_compare(F, S),	!, call(F, <, 1, 2).

q_compare_layer(C, I-_, J-_):- compare(C, I, J).
%
q_compare_layer_opp(C, I-_, J-_):- compare(C, J, I).
%
succ_opp(I, J):- succ(J, I).

% ?- predsort(q_compare, [a->b, 1, c-b, b->a, 2, b-c], X).
q_compare(C, X->Y, U->V):-!, compare(C, (X,Y), (U, V)).
q_compare(>, _->_, _-_):-!.
q_compare(<, _-_, _->_):-!.
q_compare(C, A, B):-compare(C, B, A).

% ?- t_compare(C, p(0, 1), p(1,0)).
% ?- predsort(t_compare, [p(1,1), p(0,0), p(1,0), p(0,1), p(0,0)], L).
t_compare(=,  X, X):-!.
t_compare(C, p(I, J), p(L, M)):-  A is I+J,
	B is L + M,
	compare(C0, A, B),
	(	C0 = (=) -> compare(C, I, L)
	;	C = C0
	).

% ?- convert_skew(p(1,2), X).
convert_skew(p(I, J), p(K, I)):- K is I + J.

% ?- X = p(1, 2), ortho_skew(X, Y), ortho_skew(Z, Y), ortho_skew(Z, U).
%@ X = Z, Z = p(1, 2),
%@ Y = U, U = p(3, 1).
ortho_skew(P, Q):-
	(	nonvar(P)
	->  P = p(I, J),
		L is I + J,
		Q = p(L, I)
	;	Q = p(L, I),
		J is L - I,
		P = p(I, J)
	).

% ?- udg(normal_pair(p(1,1)-p(0,1), X)).
% ?- rect_dg(rect(1,1), X, Y), convert_skew_udg(X, Y, Z, U),
%	udg((zdd_sort(Z, Z0),
%		maplist(normal_pair, U, U0), zdd_sort(U0, U1))),
%	writeln(U1), length(Y, LY),
%	length(U, LU), length(U0, LU0).

% ?- rect_dg(rect(1,1), X, Y), convert_skew_udg(X, Y, Z, U),
%	zdd((zdd_sort(Z, Z0),
%		maplist(normal_pair, U, U0), zdd_sort(U0, U1))),
%	writeln(U1), length(Y, LY),
%	length(U, LU), length(U0, LU0).

% ?- zdd((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- zdd0((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- zdd((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(0,0), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(1,0), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(1,1), N, L))).
% ?- udg((rect_dg_in_skew_axis(rect(10,10), N, L))).
% ?- time(udg((rect_dg_in_skew_axis(rect(100,100), N, L)))).

rect_dg_in_skew_axis(R, Ns, Ls):- @(rect_dg_in_skew_axis(R, Ns, Ls)).
%
rect_dg_in_skew_axis(R, Ns, Ls, S):-
	rect_dg(R, X, Y),
	convert_skew_udg(X, Y, Z, U),
	zdd_sort(Z, Ns, S),
	maplist(pred(S, [A, B]:- normal_pair(A, B, S)), U, U0),
	zdd_sort(U0, Ls, S).
%
convert_skew_udg(X, Y, Z, U):-
	maplist(convert_skew, X, Z),
	maplist(pred(([A-B, C-D]:- convert_skew(A, C), convert_skew(B, D))),
				 Y, U).

% ?- v_compare(C, p(0, 1), p(1,0)).
% ?- predsort(t_compare, [p(1,1), p(0,0), p(1,0), p(0,1), p(0,0)], L).
v_compare(C, p(I, J), p(L, M)):- compare(C0, I, L),
	(	C0 = (=) -> compare(C, J, M)
	;	C = C0
	).
% ?- ht_compare(C, p(0,0)-p(1,0),p(0,0)-p(0,1)).
% ?- predsort(ht_compare, [p(0,0)-p(1,0),p(0,0)-p(0,1)], L).
ht_compare(C, P-Q, U-V):- t_compare(C0, P, U),
	(	C0 = (=) -> t_compare(C, Q, V)
	;	C = C0
	).

touch(A - _, _-A).
touch(_ - A, A-_).

qp_connect(A-B, U-A, B-V, U-V):-!.
qp_connect(A-B, B-V, U-A, U-V).

change_symbol(A-B, A->B).

minus_symbol(_-_).

% ?- check_node_boole(2, [1,2,3]).
% ?- check_node_boole(b, b-c; d-b).
% ?- check_node_boole(a, \+[b,c]).
% ?- check_node_boole(a, [b,a,c]).
% ?- check_node_boole(a, \+[b,a,c]).  % false
% ?- check_node_boole(a, (\+[b,c]; [b,c])).
% ?- check_node_boole(a, ([b,a,c], [d, a, e])).
% ?- check_node_boole(a, \+(([b,a,c], [d, a, e]))). % false
% ?- time(repeat(1000000, check_node_boole(a, \+(([b,c]; [d, e]))))).
% ?- time((between(1, 1000000,_), check_node_boole(a, \+(([b,c]; [d, e]))))).
% ?- time(repeat(1000000, check_boole_node(\+(([b,c]; [d, e])), a))).
% ?- time(repeat(1000000, check_node_boole_(a, \+(([b,c]; [d, e]))))).

% ?- repeat(10000, write(.)).
check_node_boole(X, Boole):- check_boole_node(Boole, X).

%
check_boole_node(>=(I), X):-!, I @=< X.
check_boole_node(=<(I), X):-!, X @=< I.
check_boole_node(I-J, X):-!, I @=< X, X @=< J.
check_boole_node((A, B), X):-!,
	check_boole_node(A, X),
	check_boole_node(B, X).
check_boole_node((A; B), X):-!,
	( check_boole_node(A, X) -> true
	; check_boole_node(B, X)
	).
check_boole_node(\+A, X):-!,
	( check_boole_node(A, X) -> fail
	; true
	).
check_boole_node([A|As], X):-
	(	X = A -> true
	;	memberchk(X, As)
	).

%
check_node_boole_(X, >=(I)):-!, I @=< X.
check_node_boole_(X, =<(I)):-!, X @=< I.
check_node_boole_(X, I-J):-!, I @=< X, X @=< J.
check_node_boole_(X, (A, B)):-!,
	check_node_boole_(X, A),
	check_node_boole_(X, B).
check_node_boole_(X, (A; B)):-!,
	( check_node_boole_(X, A) -> true
	; check_node_boole_(X, B)
	).
check_node_boole_(X, \+A):-!,
	( check_node_boole_(X, A) -> fail
	; true
	).
check_node_boole_(X, [A|As]):-
	(	X = A -> true
	;	memberchk(X, As)
	).
%
arrow_symbol( _ -> _).
% ?- arrow_symbol(_->_, F).
arrow_symbol(A, A0):- functor(A, A0, 2).
% ?- arrow_symbol(a->b, F, X, Y).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).
%
card_hybrid(X, ST, Y) :- @(card_hybrid(X, ST, Y)).
%
crad_hybrid(_, S-S, 1, _) :-!.
card_hybrid(X, ST, Y, S) :- open_state(M, [hash_size(128)]),
	card_hybrid(X, ST, Y, S, M),
	close_state(M).
%
card_hybrid(I, _, 0, _, _)	:- I < 2, !.
card_hybrid(I, ST, C, S, M)	:- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_hybrid(L, ST, Cl, S, M),
		(	A = U-U ->	card_hybrid(R, ST, Cr, S, M)
		;	A = ST 	->	card_hybrid1(R, Cr, S, M)
		; 	Cr = 0
		),
	C is Cr + Cl
	).
%
card_hybrid1(I, I, _, _) :- I<2, !.
card_hybrid1(I, C, S, M) :- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_hybrid1(L, Cl, S, M),
		(  	A = U-U ->  card_hybrid1(R, Cr, S, M)
		;	A = (_-_) ->	Cr = 0
		;	card(R, Cr, S, M)
		),
		C is Cr + Cl
	).

%%%
card_path(X, ST, Y) :- @(card_path(X, ST, Y)).
%
card_path(_, S-S, 1, _) :-!.
card_path(X, ST, Y, S) :- open_substate(S),
	card_path_xmemo(X, ST, Y, S),
	close_substate(S).
%
card_path_xmemo(I, _, 0, _)	:- I < 2, !.
card_path_xmemo(I, ST, C, S)	:- memo(card(I)-C, S),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_path_xmemo(L, ST, Cl, S),
		(	A = U-U ->	card_path_xmemo(R, ST, Cr, S)
		;	A = ST 	->	card_path_xmemo(R, Cr, S)
		; 	Cr = 0
		),
	C is Cr + Cl
	).
%
card_path_xmemo(I, I, _) :- I<2, !.
card_path_xmemo(I, C, S) :- memo(card(I)-C, S),
	(	nonvar(C) -> true
	;	cofact(I, t(A, L, R), S),
		card_path_xmemo(L, Cl, S),
		(  	A = U-U ->  card_path_xmemo(R, Cr, S)
		;	A = (_-_) ->	Cr = 0
		;	card(R, Cr, S)
		),
		C is Cr + Cl
	).

% ?- rect_dg(rect(0,1), X, Y).
% ?- rect_dg(rect(0,0), X, Y).
% ?- rect_dg(rect(1,1), X, Y).
rect_dg(rect(W, H), Nodes, Links):-
	rect_nodes(W, H, Nodes),
	rect_links(W, H, Nodes, Links).

% ?- zdd((rect_udg(rect(1,1), X, Y, ST))).
% ?- zdd((rect_udg(rect(20,20), X, Y, ST))).
% ?- zdd0((rect_udg(rect(1,1), X, Y, ST))).
% ?- q_zdd((rect_udg(rect(1,1), X, Y, ST))).

rect_udg(A, B, C, ST):- @(rect_udg(A, B, C, ST)).
%
rect_udg(Rect, Nodes, Links, ST, S):-
	rect_dg(Rect, Ns, Ls),
	zdd_sort(Ns, Nodes, S),
	maplist(pred(S, [U,V]:- normal_pair(U, V, S)), Ls, Ls0),
	zdd_sort(Ls0, Links, S),
	Rect=rect(W, H),
	normal_pair(p(0,0)-p(W,H), ST, S).

% ?- zdd((rect_skew(rect(1,1), X, Y, ST))).
% ?- zdd0((rect_skew(rect(1,1), X, Y, ST))).
% ?- zdd((rect_skew(rect(20,20), X, Y, ST))).
rect_skew(A, B, C, D):- @(rect_skew(A, B, C, D)).
%
rect_skew(Rect, Nodes, Links, ST, S):-
	rect_dg_in_skew_axis(Rect, Nodes, Links, S),
	Rect=rect(W, H),
	convert_skew(p(W, H), P),
	normal_pair(p(0,0)-P, ST, S).

% ?- rect_htdg(rect(0,0), N, L).
% ?- rect_htdg(rect(0,1), N, L).
% ?- rect_htdg(rect(1,1), N, L).
% ?- rect_htdg(rect(2,2), N, L), maplist(writeln, L).
% ?- rect_htdg(rect(3,3), N, L), maplist(writeln, L).
% ?- time(rect_htdg(rect(30,30), N, L)).

rect_htdg(Rect, Nodes, Sorted_Links):-  Rect = rect(W, H),
	rect_dg(Rect, Nodes, Links),
	findall(A-B, (	member(A-B, Links),
					A @< B
				  ),
			 Canonical_links),
	Max is W+H,
	findall(I-L, (	between(0, Max, I),
					findall(P-Q, ( member(P-Q, Canonical_links),
								   P = p(P1,P2),
								   I =:= P1 + P2
								 ),
							L)
				 ),
			Sorted_Links).

% ?- zdd((q_problem(rect(1,1), L, I, ST))).

% ?- zdd((q_problem(rect(1,1), L, I, ST),
%		@(elim_links(L, I, M)),
%		@(card_hybrid(M, ST, C))
% )).

q_problem(Rect, Links, ID, p(0,0)-p(W, H)):-
	Rect=rect(W, H),
	@(rect_udg(Rect, Nodes, Links)),
	findall(A-A, member(A, Nodes), ID).

% ?- zdd0((skew_problem(rect(1,1), L, I, ST))).
% ?- zdd((skew_problem(rect(1,1), L, I, ST))).
% ?- zdd0((skew_problem(rect(1,1), L, I, ST))).
skew_problem(Rect, Links, ID, ST):- @(skew_problem(Rect, Links, ID, ST)).
%
skew_problem(Rect, Links, ID, ST, S):-
	rect_dg_in_skew_axis(Rect, Nodes, Links, S),
	findall(A-A, member(A, Nodes), ID0),
	zdd_sort(ID0, ID, S),
	Rect=rect(W, H),
	convert_skew(p(W, H), P),
	set_key(start, P, S),
	Level is W + H,
	set_key(level, Level, S),
	normal_pair(p(0,0)-P, ST, S).

%
% rect_skew(Rect, Links, ID, ST, S):-
% 	rect_dg_in_skew_axis(Rect, Nodes, Links),
% 	findall(A-A, member(A, Nodes), ID0),
% 	zdd_sort(ID0, ID, S),
% 	Rect=rect(W, H),
% 	convert_skew(p(W, H), P),
% 	normal_pair(p(0,0)-P, ST, S).

% ?- rect_nodes(4,5,Ns), length(Ns, L).
rect_nodes(W, H, Nodes):-
	findall(p(I, J),
			(	between(0, W, I),
				between(0, H, J)
			),
			Nodes0),
	sort(Nodes0, Nodes).
%
rect_links(W, H, Ns, Links):-
	findall(P-Q,
			(	member(P, Ns),
				P = p(I, J),
				(	Q = p(I, J1),
					(	J1 is J-1,
						J1 >= 0
					;	J1 is J+1,
						J1 =< H
					)
				;   Q = p(I1, J),
					(	I1 is I-1,
						I1 >= 0
					;	I1 is I+1,
						I1 =< W
					)
				)
			),
			Links0),
	sort(Links0, Links).

%
atmark(X):- functor(X, @, 2).

		/*******************************
		*     rect_path_count_naive    *
		*******************************/
% ?- zdd0(rect_path_count_naive(rect(1,0), C)).
% ?- zdd0(rect_path_count_naive(rect(2,0), C)).
% ?- zdd0(rect_path_count_naive(rect(1,2), C)).
% ?- zdd0(rect_path_count_naive(rect(1,3), C)).
% ?- zdd0(rect_path_count_naive(rect(3,1), C)).
% ?- zdd0(rect_path_count_naive(rect(2,2), C)).
% ?- zdd0(rect_path_count_naive(rect(2,2), C)).
% ?- zdd0(rect_path_count_naive(rect(2,2), C)).
% ?- time(zdd0(rect_path_count_naive(rect(3,2), C))).
% ?- time(zdd0(rect_path_count_naive(rect(2,3), C))).
% ?- time(zdd0(rect_path_count_naive(rect(3,3), C))).

rect_path_count_naive(rect(W, H), C):-
	rect_dg(rect(W, H), X, Y),
	dg_path_count_naive(dg(X, Y), p(0,0)-p(W, H), C).

		/*****************************
		*     dg_path_count_naive    *
		*****************************/

% ?-zdd0(dg_path_count_naive(dg([a, b],[a-b]), a-b, C)).
% ?-zdd0(dg_path_count_naive(dg([a, b],[a-b, b-a]), a-b, C)).
% ?-zdd0(dg_path_count_naive(dg([a, b, c],[a-b, b-c, a-c]), a-c, C)).
% ?-zdd0(dg_path_count_naive(dg([a, b, c],[a-b, b-c, c-b, c-a, a-c]), a-c, C)).
dg_path_count_naive(D, ST, C):- @(dg_path_count_naive(D, ST, C)).
%
dg_path_count_naive(dg(Ns, Es), ST, C, S):-
		zdd_sort(Es, Fs, S),
		reverse(Fs, Gs),
		findall(X-X, member(X, Ns), Id),
		zdd_insert_atoms(Id, 1, Initial, S),
		repeat_update_mqp(Gs, Initial, Out, S),
		card_hybrid(Out, ST, C, S).

		/***************************
		*     repeat_update_mqp    *
		***************************/

repeat_update_mqp(Es, X, Y):- @(repeat_update_mqp(Es, X, Y)).
%
repeat_update_mqp([], X, X, _).
repeat_update_mqp([E|Es], X, Y, S):- update_mqp(E, X, X0, S),
	zdd_join(X0, X, Y0, S),
	repeat_update_mqp(Es, Y0, Y, S).

% ?- zdd0((X<<{[a-a, b-b]}, apply_links_on_mqp([a-b], X, Y))).
% ?- zdd0((X<<{[a-a, b-b, c-c]},
%	apply_links_on_mqp([a-b, b-c, c-a, b-a, a-c], X, Y),
%	card_hybrid(Y, a-c, C))).

apply_links_on_mqp(Es, X, Y):-@(apply_links_on_mqp(Es, X, Y)).
%
apply_links_on_mqp(Es, X, Y, S):- zdd_sort(Es, Fs, S),
	zdd_eval(pow(Fs), Pow, S),
	power_on_mqp(Pow, X, Y, S).
%
power_on_mqp(0, _, 0, _):-!.
power_on_mqp(1, X, X, _):-!.
power_on_mqp(_, X, 0, _):- X<2, !.
power_on_mqp(P, X, Y, S):- memo(power_on_mqp(P,X)-Y, S),
	(	nonvar(Y) -> true,	write(.)
	;	cofact(P, t(E, L, R), S),
		power_on_mqp(L, X, L0, S),
		update_mqp(E, X, EM, S),
		power_on_mqp(R, EM, R0, S),
		zdd_join(L0, R0, Y, S)
	).

		/********************
		*     update_mqp    *
		********************/

% ?- zdd0((X<<{[a-a, b-b, c-c]}, psa(X), update_mqp(a-b, X, Y), zdd_join(X, Y, Y0),
%	update_mqp(b-c, Y0, Z), zdd_join(Y0, Z, Z0),
%	update_mqp(a-c, Z0, U), zdd_join(Z0, U, U0), psa(U0),
%	card_hybrid(U0, a-c, C))).

% ?- zdd0((X<<{[a-a, b-b, c-c]}, psa(X), update_mqp(a-b, X, Y), zdd_join(X, Y, Y0),
%	update_mqp(a-c, Y0, Z), zdd_join(Y0, Z, Z0), psa(Z0),
%	update_mqp(b-c, Z0, U), zdd_join(Z0, U, U0), psa(U0),
%	card_hybrid(U0, a-c, C))).

update_mqp(E, X, Y):- @(update_mqp(E, X, Y)).

%
update_mqp(_, X, 0, _):- X < 2,!.
%
update_mqp(E, X, Y, S):- cofact(X, t(P, L, R),  S),
	(	arrow_symbol(P) -> Y = 0
	;  	update_mqp(E, L, L0, S),
		E = (A-B),
		P = (C-D),
		(	zdd_compare(<, B, C, S) -> R0 = 0	% , write(.)  % Many hits.
		;	B = C -> replace_right([A->B], A, D, R, R0, S)
		;	A = D -> replace_left([A->B], B, C, R, R0, S)
		;	update_mqp(E, R, R1, S),
			zdd_insert(P, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

		/**********************
		*     elim_links    *
		**********************/

% ?- mk_sorted_pair(a, b, U).
% ?- mk_sorted_pair(b, a, U).
mk_sorted_pair(A, B, A-B):- A@<B, !.
mk_sorted_pair(A, B, B-A).
%
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).


%
normal_pair(P, U):- @(normal_pair(P, U)).
%
normal_pair(A-B, B-A, S):- zdd_compare(>, A, B, S), !.
normal_pair(A->B, B->A, S):- zdd_compare(>, A, B, S), !.
normal_pair(P, P, _).
%
convert_skew_pair(P, Q):-
	(	P = (A-B) -> Q = (A0-B0)
	;	P = (A->B), Q = (A0->B0)
	),
	convert_skew(A, A0),
	convert_skew(B, B0).
% ?- open_state(S), set_compare(compare, S),
%	convert_normal_pair(p(1,0)-p(1,1), Q, S).
% ?- open_state(S), set_compare(compare, S),
%	convert_normal_pair(p(1,0)->p(1,1), Q, S).

convert_normal_pair(P, Q, S):- convert_skew_pair(P, Q0),
	normal_pair(Q0, Q, S).

%
normal_arrow(X, Y):- @(normal_arrow(X, Y)).
%
normal_arrow(A->B, Arr, S):-!,
	(	zdd_compare(>, A, B, S) -> Arr = (B->A)
	;	Arr = (A->B)
	).
normal_arrow(A-B, Arr, S):-
	(	zdd_compare(>, A, B, S) -> Arr = (B->A)
	;	Arr = (A->B)
	).

%
lt_pair(_ - A, B -_, S):- zdd_compare(<, A, B, S).

		/*********************************
		*     pruning rectangular mqp    *
		*********************************/

to_be_pruned(X, Y, Z):- @(to_be_pruned(X, Y, Z)).

to_be_pruned(Min, A-B, P, S):- A \== B,
	(	zdd_compare(<, B, P, S)
	;	A \== Min,
		zdd_compare(<, A, P, S)
	).

%
prune_for_frontier(_, _, X, 0, _):- X<2, !.
prune_for_frontier(F, M, X, Y, S):- cofact(X, t(Ar, L, R), S),
		arrow_symbol(Ar, Arrow, P, Q),
		(	Arrow = (->)  ->  Y = 0
		;
		prune_for_frontier(F, M, L, L0, S),
		(	on_frontier(F, M, P)
		->	(	on_frontier(F, M, Q) ->	prune_for_frontier(F, M, R, R0, S)
			;	R0 = R
			)
		;	(	on_frontier(F, M, Q) ->	R0 = R
			;	R0 = 0
			)
		),
		cofact(Y, t(P-Q, L0, R0), S)
		).
%
on_frontier(F, _, p(F, _)).
on_frontier(_, M, M).
%
u_prune(P, X, Y, S):- get_key(prune, F, S), !,
	zdd_shift(u_prune(F, P, X, Y, S)).
u_prune(_, X, X, _).
%
u_prune(F, P, X, Y, S, M):- apply_prune(F, P, X, X0, S, M),
	zdd_subtr(X, X0, Y, S).
%
apply_prune(_, _, X, 0, _, _):- X<2, !.
apply_prune(F, P, X, Y, S, M):- memo(apply_prune(X)-Y,  M),
	(	nonvar(Y) -> true	% , write(/)	% many hits.
	;	cofact(X, t(Ar, L, R), S),
		apply_prune(F, P, L, L0, S, M),
		(	call(F, Ar, P, S) -> R0 = R
		;	lt_pair(Ar, P, S) -> R0 = 0
		;	apply_prune(F, P, R, R0, S, M)
		),
		cofact(Y, t(Ar, L0, R0), S)
	).

%
:- meta_predicate udg(:).
udg(X):- zdd((set_compare(opp_compare), X)).

:- meta_predicate q_zdd(:).
q_zdd(X):- zdd((set_compare(q_compare), X)).
% ?- q_zdd(normal_pair(a-b, X)).
% ?- q_zdd(normal_pair(1-2, X)).
% ?- udg((zdd_sort([p(0,0),p(1,1)], R))).
% ?- spy(q_problem).
% ?- w_count(rect(1,1), C).
w_count(Rect, C):- 	zdd((q_problem(Rect, Links, ID, ST),
		  MQP<<set(ID),
		  @(elim_links(Links, MQP, MQP1),
		  card_hybrid(MQP1, ST, C)))).

%%%uuuu
% ?- zdd(q_count(rect(0, 1), C)).
%@ p(1,0)-p(0,0)
%@ C = 1 .
% ?- zdd(q_count(rect(1, 0), C)).
%@ p(1,1)-p(0,0)
%@ C = 1 .
% ?- zdd(q_count(rect(1, 1), C)).
% ?- zdd(q_count(rect(1, 2), C)).
% ?- zdd(q_count(rect(2, 2), C)).
% ?- time(zdd(q_count(rect(2, 2), C))).
% ?- time(zdd(q_count(rect(3, 3), C))).
% ?- time(zdd(q_count(rect(3, 4), C))).
% ?- time(zdd0(q_count(rect(4, 4), C))).
% ?- time(zdd0(q_count(rect(5, 5), C))).
%@ % 12,276,144,277 inferences, 1672.961 CPU in 1693.367 seconds (99% CPU, 7337975 Lips)
%@ C = 1262816.

q_count(Rect, C):- @(q_count(Rect, C)).
%
q_count(Rect, C, S):- set_compare(q_compare, S),
		(	is_normal_order(S)
		->	set_pred(compare_layer, q_compare_layer, S),
			set_pred(succ, succ, S)
		;	set_pred(compare_layer, q_compare_layer_opp, S),
			set_pred(succ, succ_opp, S)
		),
		skew_problem(Rect, Links, ID, ST, S),
		layer_by_level(ID, IDLayers, S),
		set_key(idlayers, IDLayers, S),
		layer_by_level(Links, Layers, S),
		get_key(start, Start, S),
		set_pred(prune, to_be_pruned(Start), S),
		zdd_ord_insert(ID, 1, MQP, S),
		elim_layers(Layers, MQP, MQP1, S),
		card_hybrid(MQP1, ST, C, S).
%
layer_by_level(Links, OutLinks, S):-
	get_key(level, MaxLevel, S),
	findall(I-G, (	between(0, MaxLevel, I),
					findall(Ar, (	member(Ar, Links),
									Ar = p(I,_)-_
								),
							G0),
					zdd_sort(G0, G, S)
				 ),
			Leveled_Links),
	get_key(compare_layer, Compare_Layer, S),
	predsort(Compare_Layer, Leveled_Links, OutLinks).

%
elim_layers(X, Y, Z):- @(elim_layers(X, Y, Z)).
%
elim_layers([], X, X, _).
elim_layers([Level-Es | Layers], X, Y, S):-
	get_key(start, Min, S),
	prune_for_frontier(Level, Min, X, X0, S),
	zdd_subtr(X, X0, X1, S),
	get_key(succ, SUCC, S),
	call(SUCC, Level1, Level),
	prune_for_cut_path(Level1, X1, X2, S),
	zdd_subtr(X1, X2, X3, S),
	elim_links(Es, X3, Z, S),
	elim_layers(Layers, Z, Y, S).
%
prune_for_cut_path(Lev, X, Y, S):-
	get_key(idlayers, IDLayers, S),
	(	memberchk(Lev-ID, IDLayers)
	->  zdd_sub_for_ord_list(ID, X, Y, S)
	;	Y = 0
	).

		/**********************************
		*     basic operation on links    *
		**********************************/
:- meta_predicate q_count_plain(:, :, ?, ?).
:- meta_predicate q_count_plain(:, :, ?, ?, ?).

% ?- zdd(q_count_plain(compare, rect_skew, rect(1, 0), C)).
% ?- zdd(q_count_plain(compare, rect_skew, rect(1, 2), C)).
% ?- zdd(q_count_plain(compare, rect_skew, rect(3, 2), C)).
% ?- zdd(q_count_plain(q_compare, rect_udg, rect(2, 2), C)).
% ?- zdd(q_count_plain(opp_compare, rect_udg, rect(2, 2), C)).
% ?- zdd(q_count_plain(opp_compare, rect_udg, rect(3, 3), C)).
% ?- zdd(q_count_plain(compare, rect_udg, rect(1, 0), C)).
% ?- zdd(q_count_plain(compare, rect_udg, rect(2, 0), C)).
% ?- zdd(q_count_plain(compare, rect_udg, rect(2, 2), C)).

q_count_plain(X, Y, Z, U):-@(q_count_plain(X, Y, Z, U)).
%
q_count_plain(Compare, F, Rect, C, S):-
		set_compare(Compare, S),
		call(F, Rect, Nodes, Links, ST, S),
		ST = Start-_,
		set_pred(prune, to_be_pruned(Start), S),
		findall(A-A, member(A, Nodes), 	ID),
		zdd_insert_atoms(ID, 1, MQP, S),
		elim_links(Links, MQP, MQP1, S),
		card_hybrid(MQP1, ST, C, S).
%
elim_links(A, B, C):- @(elim_links(A, B, C)).
%
elim_links([], X, X, _).
elim_links([E|Es], X, Y, S):-
	writeln(E),
	elim_link(E, X, X0, S),
	zdd_join(X, X0, X1, S),
	E  = P-_,
	u_prune(P, X1, X2, S),
	elim_links(Es, X2, Y, S).
%
elim_link(E, X, Y):- @(elim_link(E, X, Y)).
% ?- time(zdd0(q_count(rect(3, 4), C))).

elim_link(_, X, 0, _):- X < 2,!.
elim_link(E, X, Y, S):-	cofact(X, t(P, L, R),  S),
	(	P = (_->_)  ->  Y = 0   %Y = X
	;	lt_pair(E, P, S) -> Y = 0
	;  	elim_link(E, L, L0, S),
		(	composable_pairs(E, P, U, V)
		-> 	normal_arrow(E, Arrow, S),
			replace_end([Arrow], U, V, R, R0, S)
		;	elim_link(E, R, R1, S),
			zdd_insert(P, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% elim_link with memo
elim_link(_, X, 0, _, _):- X < 2,!.
elim_link(E, X, Y, S, M):-	cofact(X, t(P, L, R),  S),
	(	P = (_->_)  ->  Y = 0   %Y = X
	;	lt_pair(E, P, S) -> Y = 0
	;  	elim_link(E, L, L0, S, M),
		(	composable_pairs(E, P, U, V)
		-> 	normal_arrow(E, Arrow, S),
			replace_end([Arrow], U, V, R, R0, S, M)
		;	elim_link(E, R, R1, S, M),
			x_insert(P, R1, R0, S, M)
		),
		x_join(L0, R0, Y, S, M)
	).

% Most basic operation.
replace_end(_, _, _, X, 0, _):- X < 2, !.
replace_end(Es, A, P, X, Y, S):-  	/* replace A with P */
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	replace_end(Es, A, P, L, L0, S),
		U = Lu-Ru,
		(	zdd_compare(<, A, Lu, S)
		->	Y = 0
		;	Ru = A
		->	normal_pair(Lu-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A
		->  normal_pair(Ru-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	replace_end(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).


		/********************************
		*     u_repeat_update_mqp		*
		********************************/

u_repeat_update_mqp(Es, X, Y):-
	@(u_repeat_update_mqp(Es, X, Y)).
%
u_repeat_update_mqp([], X, X, _).
u_repeat_update_mqp([E|Es], X, Y, S):- u_update_mqp(E, X, X0, S),
	zdd_join(X0, X, X1, S),
	u_obsolete_mqp(X1, Ob, S),
	zdd_subtr(X1, Ob,  Y0, S),
	u_repeat_update_mqp(Es, Y0, Y, S).
%
u_obsolete_mqp(X, Y, S):-
	(	get_key(obsolete_checker, F, S)
	->	call(F, X, Y, S)
	;	Y = 0
	).
%
u_update_mqp_aux(A-B, C-D, U, V):-
		(	B = C -> U = A, V = D
		;	B = D -> U = A, V = C
		;	A = C -> U = B, V = D
		;	A = D -> U = B, V = C
		).
%
u_update_mqp(E, X, Y):- @(u_update_mqp(E, X, Y)).
%
u_update_mqp(_, X, 0, _):- X < 2,!.
u_update_mqp(E, X, Y, S):- E = A-B,
	cofact(X, t(P, L, R),  S),
	(	arrow_symbol(P)  ->  Y = 0
	;  	u_update_mqp(E, L, L0, S),
		(	u_update_mqp_aux(E, P, U, V)
		-> 	u_replace_mqp(A->B, U, V, R, R0, S)
		;	u_update_mqp(E, R, R1, S),
			zdd_insert(P, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

%
u_replace_mqp(_, _, _, X, 0, _):- X < 2, !.
u_replace_mqp(E, A, P, X, Y, S):- cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	u_replace_mqp(E, A, P, L, L0, S),
		U = Lu-Ru,
		(	zdd_compare(<, A, Lu, S),
			zdd_compare(<, P, Lu, S)
		->	Y = 0
		;	Ru = A
		->	normal_pair(Lu-P, V, S),
			zdd_insert_atoms([V, E], R, R0, S)
		;	Lu = A
		->  normal_pair(Ru-P, V, S),
			zdd_insert_atoms([V, E], R, R0, S)
		;	u_replace_mqp(E, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).



% ?- zdd((X<<{[a-a,b-b,c-c]}, update_qp(a-b, X, Y), psa(Y), update_qp(b-c, Y, Z), psa(Z))).
% ?- zdd0((X<<{[a-a,b-b,c-c]}, update_qp(a-b, X, Y), psa(Y), update_qp(c-a, Y, Z), psa(Z))).
% ?- zdd0((X<<{[a-a,b-b,c-c]}, update_qp(a-b, X, Y), psa(Y), update_qp(b-c, Y, Z), psa(Z), update_qp(b-a, Z, U), psa(U))).

update_qp(E, X, Y):- @(update_qp(E, X, Y)).

%
update_qp(_, X, 0, _):- X < 2,!.
update_qp(E, X, Y, S):- cofact0(X, P, R, S),
		E = (A-B),
		P = (C-D),
		(	zdd_compare(<, A, C, S)
		->  Y = 0		% , write(.)  % Many hits.
		;	B = C -> qp_replace_right(A, D, R, Y, S)
		;	A = D -> qp_replace_left(B, C, R, Y, S)
		;	update_qp(E, R, R0, S),
			zdd_insert_as_list(P, R0, Y, S)
		).

%  udg: unordered pair   a-b = b-a
u_update_qp(E, X, Y):- @(u_update_qp(E, X, Y)).

%
u_update_qp(_, X, 0, _):- X < 2,!.
u_update_qp(E, X, Y, S):- cofact0(X, P, R, S),
		E = (A-B),
		P = (C-D),
		( 	zdd_compare(<, B, C, S)	-> Y = 0 %, write(.)		% Great many hits.
		;	B = C -> u_qp_replace(A, D, R, Y, S)
		;	B = D -> u_qp_replace(A, C, R, Y, S)
		;	A = C -> u_qp_replace(B, D, R, Y, S)
		;	A = D -> u_qp_replace(B, C, R, Y, S)
		;	u_update_qp(E, R, R0, S),
			u_insert(P, R0, Y, S)
		).
%
u_insert(_, 0, 0, _):-!.
u_insert(P, 1, Y, S):-!, zdd_singleton(P, Y, S).
u_insert(P, X, Y, S):- cofact0(X, Q, R, S),
	zdd_compare(C, P, Q, S),
	(	C = (<) ->	cofact0(Y, P, X, S)
	;	C = (=) ->	Y = X
	;	u_insert(P, R, R0, S),
		cofact0(Y, Q, R0, S)
	).

% ?- zdd0((set_key(min, d), X << {[d-e]}, slim_mqp(d, X, Y), psa(Y))).
% ?- zdd0((X << {[d-d]}, slim_mqp(d, X, Y), psa(Y))).
% ?- zdd0((X << {[]}, slim_mqp(d, X, Y), psa(Y))).
% ?- zdd0((X << {[d-e]}, slim_mqp(e, X, Y), psa(Y))).

slim_mqp(P, X, Y):- @(slim_mqp(P, X, Y)).
%
slim_mqp(P, X, Y, S):- get_key(min, Q, S),
	obsolete_mqp(P, Q, X, U, S),
	zdd_subtr(X, U, Y, S).

% ?- spy(obsolete_mqp).
% ?- zdd0((X << {[d-d]}, obsolete_mqp(d, d, X, Y), psa(Y))).
% ?- zdd0((X << {[d-d]}, obsolete_mqp(e, X, Y), psa(Y))).
% ?- zdd0((X << {[d-d]}, obsolete_mqp(a, X, Y), psa(Y))).
% ?- zdd0((X << {[d-e]}, obsolete_mqp(d, X, Y), psa(Y))).
% ?- zdd0((X << {[x-z, d-e]}, obsolete_mqp(z, X, Y), psa(Y), psa(X))).

obsolete_mqp(P, Q, X, Y):- @(obsolete_mqp(P, Q, X, Y)).
%
obsolete_mqp(_, _, X, 0, _):- X<2, !.
obsolete_mqp(P, Q, X, Y, S):- cofact(X, t(E, L, R), S),
	arrow_symbol(E, E0, A, B),
	(	E0=(-)
	->	obsolete_mqp(P, Q, L, L0, S),
		zdd_compare(C, A, P, S),
		(	C = (<) ->	R0 = R
		;	C = (=) ->	R0 = R
		;	B = Q	->  obsolete_mqp(P, Q, R, R0, S)
		;	B = P	->  R0 = R
		),
		cofact(Y, t(E, L0, R0), S)
	;	Y = 0
	).

% f(p(I0, J0), p(I, J)):- I0 is I-1, J0 is J-1, I0>=0, J0>=0.
% g(p(I0, J0), p(I, J)):-

		/*********************
		*     update_link    *
		*********************/

update_links(Es, X, Y):- @(update_links(Es, X, Y)).
%
update_links([], X, X, _).
update_links([E|Es], X, Y, S):- update_link(E, X, X0, S),
	zdd_join(X0, X, Y0, S),
	update_links(Es, Y0, Y, S).

% ?- zdd0((X<<{[a-a, b-b]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[p(0,0)-p(0,0), p(1,0)-p(1,0)]}, psa(X), update_link(p(0,0)-p(1,0), X, Y), psa(Y))).
% ?- zdd0((X<<{[p(1,0)-p(1,0), p(0,0)-p(0,0)]}, psa(X), update_link(p(0,0)-p(1,0), X, Y), psa(Y),
% ?- zdd0((X<<{[a-a, b-b, c-c]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[c-c, b-b, a-a]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[b-b, a-a]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[a-a, b-b, c-c]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[c-c, b-b, a-a]}, psa(X), update_link(a-b, X, Y), psa(Y))).
% ?- zdd0((X<<{[1-2, 3-4, 5-6]}, update_link(3-4, X, Y), psa(Y))).
% ?- zdd0((X<<{[4-5]}, psa(X), update_link(2-3, X, Y), psa(Y))).
% ?- zdd0((X<<{[4-5,3-3]}, psa(X), update_link(3-4, X, Y), psa(Y))).
% ?- zdd0((X<<{[4-5,3-3]}, psa(X), update_link(5-6, X, Y), psa(Y))).
update_link(E, X, Y):- @(update_link(E, X, Y)).

update_link(_, X, 0, _):- X < 2,!.
update_link(E, X, Y, S):- memo(update_link(E, X)-Y, S),
	(	nonvar(Y) -> true	% many hits !
	;  	cofact(X, t(P, L, R),  S),
		update_link(E, L, L0, S),
		(	touch(E, P)
		->	update_link(P, E, R, R0, S)
		;	update_link(E, R, U, S),
			zdd_insert(P, U, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

%
update_link(_, _, X, 0, _):- X<2, !.
update_link(P, E, X, Y, S):- memo(update_link(P, E, X)-Y, S),
	(	nonvar(Y)	-> true		% many hits !
	;	cofact(X, t(Q, L, R), S),
		update_link(P, E, L, L0, S),
		(	touch(E, Q)
		->  qp_connect(E, P, Q, U),
			change_symbol(E, E0),
			zdd_insert_atoms([E0, U], R, R0, S)
		;	update_link(P, E, R, U, S),
			zdd_insert(Q, U, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

		/**********************
		*     mortal paths    *
		**********************/

% ?- zdd0((X<<{[3-4]}, mortal_paths(5-2, X, Y))).
% ?- zdd0((X<<{[3-4]}, mortal_paths(5-2, X, Y))).
% ?- zdd0((X<<{[3-4]}, mortal_paths(2-3, X, Y))).   % Y = 0

mortal_paths(E, X, Y):- @(mortal_paths(E, X, Y)).
%

mortal_paths(_, X, 0, _):- X < 2,!.
mortal_paths(E, X, Y, S):-
	  	cofact(X, t(P, L, R),  S),
		mortal_paths(E, L, L0, S),
		(	arrow_symbol(P, ->) ->	R0 = 0
		;	mortal(E, P) ->	zdd_insert(P, R, R0, S)
		; 	mortal_paths(E, R, U, S),
			zdd_insert(P, U, R0, S)
		),
	zdd_join(L0, R0, Y, S).

%
mortal(A-_, X-Y):- X\==Y,  A @> Y.   % As are monotone increarsing

%
update_link_slim(E, X, Y):- @(update_link_slim(E, X, Y)).
%
update_link_slim(E, X, Y, S):- update_link(E, X, X0, S),
	zdd_join(X, X0, X1, S),
	mortal_paths(E, X1, MP, S),
	zdd_subtr(X1, MP, Y, S).

%
update_links_slim(E, X, Y):- @(update_links_slim(E, X, Y)).

update_links_slim([], X, X, _).
update_links_slim([E|Es], X, Y, S):-  update_link_slim(E, X, X0, S),
	update_links_slim(Es, X0, Y, S).

		/*********************
		*     merge_blocks   *
		*********************/

% ?- zdd0((X<<{[a-a]}, Y<<{[b-b]},
%	merge_blocks([a-b], X, Y, Z),
%	psa(Z))).

merge_blocks(Es, X, Y, Z):- @(merge_blocks(Es, X, Y, Z)).
%
merge_blocks(Es, X, Y, Z, S):- merge_blocks(Es, X, Y, 0, Z, S).
%
merge_blocks([], X, Y, Z, W, S):- zdd_merge(X, Y, U, S),
	zdd_join(U, Z, W, S).
merge_blocks([E|Es], X, Y, Z0, Z, S):- merge_blocks_step(E, X, Y, U, S),
	update_mqp(E, Z0, V, S),
%	update_link(E, Z0, V, S),
	zdd_join_list([U, V], Z0, Z1, S),
	merge_blocks(Es, X, Y, Z1, Z, S).

merge_blocks_step(A-B, X, Y, Z, S):-
	(	A @< B
	-> 	link_across_out(A, B, X, Y, Z, S)
%	;	link_across_in(B, A, X, Y, Z, S)
	;	link_across_in(A, B, X, Y, Z, S)
	).

%
link_across_out(_, _, R, _, 0, _):- R < 2, !.
link_across_out(M, M0, R, R0, Q0, S):- cofact(R, t(A, Left, Right), S),
	arrow_symbol(A, A0, U, V),
	(	A0 = (->)  -> Q0 = 0
	; 	link_across_out(M, M0, Left, R0, Q1, S),
		(	M = V ->
			replace_left([M->M0], M0, U, R0, R1, S),
			zdd_merge(Right, R1, R2, S)
		;	M @> V ->  R2 = 0	% , write(.) % small hits.
		;	link_across_out(M, M0, Right, R0, R1, S),
			zdd_insert(U-V, R1, R2, S)
		),
	zdd_join(Q1, R2, Q0, S)
	).

link_across_in(_, _, R, _, 0, _):- R < 2, !.
link_across_in(M, M0, R, R0, Q0, S):- cofact(R, t(A, Left, Right), S),
	arrow_symbol(A, A0, U, V),
	(	A0 = (->) -> Q0 = 0
	;  	link_across_in(M, M0, Left, R0, L0, S),
		(	M = U ->
			replace_right([M0->M], M0, V, R0, R1, S),
			zdd_merge(Right, R1, R2, S)
		;	M @> U ->  R2 = 0	% , write(+)  % small hits.
		;	link_across_in(M, M0, Right, R0, R1, S),
			zdd_insert(U-V, R1, R2, S)
		),
		zdd_join(L0, R2, Q0, S)
	).



		/**************************
		*     u_merge_blocks    *
		**************************/

on_the_same_column(X, Y):- arg(1, X, A), arg(1, Y, A).
%
u_merge_blocks(Es, X, Y, Z):- @(u_merge_blocks(Es, X, Y, Z)).
%
u_merge_blocks(Es, X, Y, Z, S):- u_merge_blocks(Es, X, Y, 0, Z, S).
%
u_merge_blocks([], X, Y, Z, W, S):- zdd_merge(X, Y, U, S),
	zdd_join(U, Z, W, S).
u_merge_blocks([E|Es], X, Y, Z0, Z, S):- u_merge_blocks_step(E, X, Y, U, S),
	u_update_mqp(E, Z0, V, S),
	zdd_join_list([U, V], Z0, Z1, S),
	u_merge_blocks(Es, X, Y, Z1, Z, S).
%
u_merge_blocks_step(_, X, _, 0, _):- X<2, !.
u_merge_blocks_step(E, X, Y, Z, S):- E = A-B,
	cofact(X, t(P, L, R), S),
	(	arrow_symbol(P) -> Z = 0
	;	u_merge_blocks_step(E, L, Y, L0, S),
		(	u_update_mqp_aux(E, P, U, V)
		->  (	on_the_same_column(U, P)
			-> 	u_replace_mqp(U, V, R, R1, S),
				zdd_merge(R1, Y, R0, S)
			;	u_replace_mqp(U, V, Y, Y0, S),
				zdd_merge(R, Y0, R0, S)
			)
		;	u_merge_blocks_step(E, R, Y, R1, S),
			zdd_insert(A->B, R1, R0, S)
		),
		zdd_join(L0, R0, Z, S)
	).

		/****************************
		*    bridge across blocks   *
		****************************/

links_across(Es, X, Y, Z):- @(links_across(Es, X, Y, Z)).
%
links_across(Es, X, Y, Z, S):- links_across(Es, X, Y, 0, Z, S).
%
links_across([], X, Y, Z, W, S):- zdd_merge(X, Y, U, S),
	zdd_join(U, Z, W, S).
links_across([E|Es], X, Y, Z0, Z, S):- link_across(E, X, Y, U, S),
	update_link(E, Z0, V, S),
	zdd_join_list([U, V], Z0, Z1, S),
	links_across(Es, X, Y, Z1, Z, S).
%
link_across(_, X, _, 0, _):- X < 2, !.
link_across(E, X, Y, Z, S):- cofact(X, t(A, L, R), S),
	E = LoE - RoE,
	arrow_symbol(A, A0, LoA, RoA),
	(	A0 = (->) ->  Z = 0
	;	link_across(E, L, Y, LY, S),
		(	(LoE = RoA ;  RoE = LoA) ->
			link_across(A, E, R, Y, RY, S)
		;  	link_across(E, R, Y, U, S),
			zdd_insert(A, U, RY, S)
		),
		zdd_join(LY, RY, Z, S)
	).

link_across(_, _, _, Y, 0, _):- Y<2, !.
link_across(A, E, X, Y, Z, S):- cofact(Y, t(B, L, R), S),
	(	arrow_symbol(B, ->) ->  Z = 0
	;	link_across(A, E, X, L, XL, S),
		(	touch(E, B)
		->	qp_connect(E, A, B, C),
			zdd_merge(X, R, U, S),
			change_symbol(E, E0),
			zdd_insert_atoms([E0, C], U, XR, S)
		;	link_across(A, E, X, R, U, S),
			zdd_insert(B, U, XR, S)
		),
		zdd_join(XL, XR, Z, S)
	).

		/*****************************
		*     glue-block-frontier    *
		*****************************/
%
glue_block_column(Es, X, Y, Z):- @(glue_block_column(Es, X, Y, Z)).

glue_block_column(Es, X, Y, Z, S):- glue_block_column(Es, X, Y, 0, Z, S).
%
glue_block_column([], _, _, Z, Z, _).
glue_block_column([E|Es], X, Y, Z0, Z, S):-
	glue_columns(E, X, Y, U, S),
	update_link(E, Z0, V, S),
	zdd_join_list([U, V], Z0, Z1, S),
	glue_block_column(Es, X, Y, Z1, Z, S).

%
glue_columns(_, X, _, 0, _):- X < 2, !.
glue_columns(E, X, Y, Z, S):- E = U-V,
	cofact(X, t(A, L, R), S),
	arrow_symbol(A, A0, A1, A2),
	(	A0 = (->) -> Z = 0
	;   glue_columns(E, L, Y, L0, S),
		(	U @< V ->
			(	U = A2 ->
				replace_left_column(A1, V, Y, W, S),
				zdd_merge_list([E, R], W, R0, S)
			;	glue_columns(E, R, Y, W, S),
				zdd_insert(A, W, R0, S)
			)
		;	(	V = A1 ->
				replace_right_column(A2, U, Y, W, S),
				zdd_merge_list([E, R], W, R0, S)
			;	V @< A1  ->  R0 = 0
			;	glue_columns(E, R, Y, W, S),
				zdd_insert(A, W, R0, S)
			)
		),
		zdd_join(L0, R0, Z, S)
	).

%
replace_right_column(_, _, X, 0, _):- X < 2, !.
replace_right_column(A, P, X, Y, S):- cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	replace_right_column(A, P, L, L0, S),
		(	Ru = A ->
			cofact(Y, t(Lu - P, L0, R), S)
		;	replace_right_column(A, P, R, R0, S),
			cofact(Y, t(U, L0, R0), S)
		)
	).

%
replace_left_column(_, _, X, 0, _):- X < 2, !.
replace_left_column(A, P, X, Y, S):- cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) -> Y = 0
	;	replace_left_column(A, P, L, L0, S),
		compare(C, A, Lu),
		(	C = (>) ->	R0 = 0
		; 	C = (=)	->
			zdd_insert(P - Ru, R, R0, S)
		;	replace_left_column(A, P, R, W, S),
			zdd_insert(U, W, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

%
slim_block_column(X, 0, _, _):- X<2, !.
slim_block_column(X, Y, K, S):- cofact(X, t(A, L, R), S),
	arrow_symbol(A, A0, A1, A2),
	(	A0 = (->) -> Y = 0
	;	slim_block_column(L, L0, K, S),
		(	A1 = p(K, _) ->
			(	( A2 = p(K, _); A2 = p(0, 0))  ->
				slim_block_column(R, R0, K, S)
			;	R0 = 0
			)
		;	A1 = A2 -> slim_block_column(R, R0, S)
		;	R0 = 0
		),
		cofact(Y, t(A, L0, R0), S)
	).
%
slim_block_column(X, X, _):- X < 2, !.
slim_block_column(X, Y, S):- cofact(X, t(A, L, R), S),
	arrow_symbol(A, A0, A1, A2),
	(	A0 = (->) -> Y = X
	;	slim_block_column(L, L0, S),
		(	A1 = A2 ->
			slim_block_column(R, R0, S)
		;	R0 = 0
		),
		cofact(Y, t(A, L0, R0), S)
	).
%
compare_rev(C, X, Y):- compare(C, Y, X).

%
replace_right(A, P, X, Y, S):- replace_right([A->P], A, P, X, Y, S).
%
replace_right(_, _, _, X, 0, _):- X < 2, !.
replace_right(Es, A, P, X, Y, S):- cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	replace_right(Es, A, P, L, L0, S),
		(	Ru = A ->
			zdd_insert_atoms(Es, R, R0, S),
			cofact(Y, t(Lu - P, L0, R0), S)
		;	replace_right(Es, A, P, R, R0, S),
			cofact(Y, t(U, L0, R0), S)
		)
	).

%
replace_left(A, P, X, Y, S):- replace_left([P->A], A, P, X, Y, S).
%
replace_left(_, _, _, X, 0, _):- X < 2, !.
replace_left(Es, A, P, X, Y, S):-
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) -> Y = 0
	;	replace_left(Es, A, P, L, L0, S),
		zdd_compare(C, A, Lu, S),
		(	C = (<) ->	R0 = 0
		; 	C = (=)	->
			zdd_insert_atoms(Es, R, W, S),
			zdd_insert(P - Ru, W, R0, S)
		;	replace_left(Es, A, P, R, W, S),
			zdd_insert(U, W, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

%
qp_replace_right(_, _, X, 0, _):- X < 2, !.
qp_replace_right(A, P, X, Y, S):- cofact0(X, U, R, S),
		U = (Lu-Ru),
		(	Ru = A -> cofact0(Y, Lu - P, R, S)
		;	qp_replace_right(A, P, R, R0, S),
			cofact0(Y, U, R0, S)
		).
%
qp_replace_left(_, _, X, 0, _):- X < 2, !.
qp_replace_left(A, P, X, Y, S):- cofact0(X, U, R, S),
		U = Lu-Ru,
		zdd_compare(C, A, Lu, S),
		(	C = (<) ->	Y = 0
		; 	C = (=)
		->  normal_pair(P-Ru, B, S),
			zdd_insert_as_list(B, R, Y, S)
		;	qp_replace_left(A, P, R, W, S),
			zdd_insert_as_list(U, W, Y, S)
		).

%
qp_replace(_, _, X, X, _):- X < 2, !.
qp_replace(A, P, X, Y, S):- cofact0(X, U, R, S),
		U = Lu-Ru,
		(	A = Lu
		->  normal_pair(P-Ru, B, S),
			zdd_insert(B, R, Y, S)
		;	A = Ru
		->  normal_pair(P-Lu, B, S),
			zdd_insert(B, R, Y, S)
		;	qp_replace(A, P, R, R0, S),
			cofact0(Y, U, R0, S)
		).
%
u_qp_replace(_, _, X, 0, _):- X < 2, !.
u_qp_replace(A, P, X, Y, S):- cofact0(X, U, R, S),
		U = Lu-Ru,
		(	zdd_compare(<, A, Lu, S),
			zdd_compare(<, P, Lu, S)
		->  Y = 0		%, write(.) % Many hits.
		;	(	A = Lu
			->  normal_pair(P-Ru, B, S),
				u_insert(B, R, Y, S)
			;	A = Ru
			->  normal_pair(P-Lu, B, S),
				u_insert(B, R, Y, S)
			;	u_qp_replace(A, P, R, R0, S),
				u_insert(U, R0, Y, S)
			)
		).
