:- module(zdd_misc, [subst_atomic/3]).
:- use_module(zdd('zdd-array')).

% :- use_module(pac(basic)).
% From util(math).
% ?- bdd_bench:subst_atomic([1-true, 0-false], f(1,0,1), R).
%@ R = f(true, false, true).
% ?- bdd_bench:subst_atomic([1-true, 0-false], g([1,0,1], f(0)), R).

% ?- rel_to_fun([1-2,1-3,2-3],R).
%%	rel_to_fun(+R, -F) is det.
% Unify F with a function from dom(R) to
% the pow(ran(R)) which assigns to each A in dom(R)
% the set of B in ran(R) such that A and B stands in the realtion R.

rel_to_fun(X, Y):- rel_to_fun(X, Y, []).

rel_to_fun([], Y, Y).
rel_to_fun([A-B|R], [A-[B|P]|Y], Z):-
	rel_to_fun(R, A, R0, P, []),
	rel_to_fun(R0, Y, Z).
%
rel_to_fun([], _, [], P, P).
rel_to_fun([A-B|R], A, U, [B|P], Q):-!,
	rel_to_fun(R, A, U, P, Q).
rel_to_fun(R, _, R, P, P).

% ?- zdd_misc:key_val_check([a-b], a, X).
% ?- zdd_misc:key_val_check([a-b], c, X).

key_val_check([X-Y|_], X, Y) :-!.
key_val_check([P|_], X, Y) :- arg(1, P, X), !, arg(2, P, Y).
key_val_check([_|A], X, Y) :- key_val_check(A, X, Y).

%%	subst_atomic(+M, +X, -Y) is det.
%  ( A naive version of substitution.)
%	With M being an assoc list as an assignment, but whose domain
%	may consist of any terms, Y is unified with M'(X) where M' is the induced
%   substitution by the assignment M.
%
% ?- zdd_misc:subst_atomic([a-b, b-a], f(a, b), F).
% ?- zdd_misc:subst_atomic([a-b, b-a], f(c, d), F).

subst_atomic(M, X, Y)	:-
	(	atomic(X), ( key_val_check(M, X, Y); Y = X	), !
	;   functor(X, F, N),
		functor(Y, F, N),
        subst_atomic_args(N, M, X, Y)
	).

%
subst_atomic_args(0, _, _, _) :- !.
subst_atomic_args(J, M, X, Y) :-
        arg(J, X, A),
        arg(J, Y, B),
        subst_atomic(M, A, B),
        K is J-1,
        subst_atomic_args(K, M, X, Y).

%%	pred_subst(+P, +M, +X, -Y) is det.
%	With given P a predicate and M an assoc list as an assignment,
%   Y is unified with M'(X),  where M' is the substitution induced
%	by the assignment M restricted to the domain
%		{ A in dom(M) | P(A) is true}.
%	That is, the substitution M' on the terms is defined as follows:
%	1. if P(X) is true and X in dom(M), then M'(X) = M(X).
%	2. Otherwise, with X = f(X1,...,Xn), M'(X) = f(M'(X1), ..., M'(Xn)).
%
% ?- zdd_misc:pred_subst(atomic, [a-b, b-a], f(a, b), F).
% ?- zdd_misc:pred_subst(integer, [a-b, b-a], f(a, b), F).
% ?- zdd_misc:pred_subst(atom, [a-b, b-a, 1-c], [a, 1,[a,b], f(a,b), b,c], F).
%@ F = [b, 1, [b, a], f(b, a), a, c].
% ?- zdd_misc:pred_subst(atomic, [a-b, b-a, 1-c], [a, 1,[a,b], f(a,b), b,c], F).
%@ F = [b, c, [b, a], f(b, a), a, c].
% ?- zdd_misc:pred_subst(atomic, [a-b, b-a, a(b)-c], a(b), F).
%@ F = a(a).

:- meta_predicate pred_subst(1, +, ?, ?).
pred_subst(Pred, M, X, Y)	:-
	(	call(Pred, X), key_val_check(M, X, Y) ->  true
	;	atomic(X)	->	Y = X
	;	functor(X, F, N),
		functor(Y, F, N),
		pred_subst_args(N, Pred, M, X, Y)
	).

%
pred_subst_args(0, _,  _, _, _)		:- !.
pred_subst_args(J, Pred, M, X, Y)	:-
        arg(J, X, A),
        arg(J, Y, B),
        pred_subst(Pred, M, A, B),
        K is J-1,
        pred_subst_args(K, Pred, M, X, Y).


% ?- module(zdd_misc).
% ?- empty_assoc(E).
% ?- empty_assoc(E), assoc(E, a, b, R), assoc(R, a, A, R0),
%	assoc(R0, b, c, R1), assoc(R1, a, A0, R2),
%	assoc(R2, b, B, R3).
%@ E =  (-),
%@ R = R0, R0 = t(a, b, -, -),
%@ A = A0, A0 = b,
%@ R1 = R2, R2 = R3, R3 = t(a, b, -, t(b, c, -, -)),
%@ B = c.

empty_assoc(_).

% no update version.
%% assoc(+T, +X, -Y)
%  Unify Y with X such that Y is associated with X
%  in T.

assoc(T, X, Y):- var(T), !, T = t(X, Y, _, _).
assoc(t(X, Y, _, _), X, Y):- !.
assoc(t(U, _, L, R), X, Y):-
	(	X @< U
	->	assoc(L, X, Y)
	;	assoc(R, X, Y)
	).

% normal (update) version.
%% assoc(+T, +X, -Y, -U) is det.
%  An assoc list U is unified with a minimum extension of an assoc list T,
%  and  Y with a term such that Y is associated with X in U.

assoc(-, X, Y, t(X, Y, -, -)):-!.
assoc(t(X, Y, L, R), X, Y, t(X, Y, L, R)):- !.
assoc(t(U, V, L, R), X, Y, t(U, V, L0, R0)):-
	(	X @< U
	->	R0 = R,
		assoc(L, X, Y, L0)
	;	L0 = L,
		assoc(R, X, Y, R0)
).

%% zip(?X, ?Y, ?Z) is det.
% Unify lists X, Y, and Z so that they are of the same length
% and that each ordered pair of A-B is i-th element of of Z
% if and only if  A and B are
% the i-th element of X and Y respectively.

zip([], [], []).
zip([X|Xs], [Y|Ys], [Z|Zs]) :-
	(	Z = X-Y
	;	Z = (X:Y)
	;	functor(Z, X, 1),
		arg(1, Z, Y)
	), !,
	zip(Xs, Ys, Zs).

% % Complementary conjunction
% slim_rule(A, true):- literal_in(+, A, not(X)),
% 					 literal_in(+, A, X).
% %
% literal_in(+, X+_, Y):- literal_in(+, X, Y).
% literal_in(+, _+X, Y):- literal_in(+, X, Y).
% literal_in(_, X, X):- (atomic(X); X = not(_)).

		/**************
		*     math    *
		**************/
% ?- Z is gcd(10, 11).
% ?- A is gcd(0, 3).
% ?- gcd([12, 6, 4], Z).
% ?- lcm([1,2,3,4,5,6,7,8], Z).
% ?- mul_list([1,2,3,4,5,6,7,8,9,10], Z).
% ?- factorial(10, F), numlist(1, 10, L), mul_list(L, F).

mul_list([], 1).
mul_list([A|As], M):- mul_list(As, M0), M is A * M0.

% ?- nCr(4,2,Z), factorial(4, R), factorial(2, S), factorial(2, T), Z is div(R, S*T).
nCr_list(M, R):- sum_list(M, S), factorial(S, FS),
	mul_fact_list(M, 1, PF),
	R is div(FS, PF).
%
mul_fact_list([], X, X).
mul_fact_list([A|As], X, Y):- factorial(A, FA), X0 is X*FA,
	mul_fact_list(As, X0, Y).

lcm(X, Y, L):- L is div(X*Y, gcd(X, Y)).

lcm([],1).
lcm([A|As], L):- lcm(As, L0), lcm(A, L0, L).

gcd([], 0):-!.
gcd([A|As], G):- gcd(As, G0), G is gcd(A, G0).

%
sum_prod([], _, N, N).
sum_prod([P|Ps], Ws, U, V):-
	prod(P, Ws, 1, Q),
	multi_combi(P, MC),
	U0 is Q*MC + U,
	sum_prod(Ps, Ws, U0, V).

prod([], _, V, V).
prod([A|As], [W|Ws], U, V):-
	U0 is (W**A)*U,
	prod(As, Ws, U0, V).


% ?- module(zdd_misc).
% ?- N=2, setof(P->Q, move(N, P->Q), Moves), zdd:univ_to_binary(->(Moves), Log).

move(N, p(I, J)->p(K, L)):-
	between(0, N, I),
	between(0, N, J),
	between(0, N, K),
	between(0, N, L),
	one_step(p(I,J), p(K,L)).
%
one_step(p(I,J),p(I0, J)):- succ(I, I0).
one_step(p(I,J),p(I, J0)):- succ(J, J0).


% ?- reverse_dag(3,f(a-[2,1,3], a-[1], a-[1,2]), R).
%@ R = f([1, 2, 3], [1, 3], [1]) .

reverse_dag(N, D, R):- functor(D, F, K),
	    functor(R, F, K),
		initial_args(R, []),
		reverse_dag_step(N, D, R).
%
reverse_dag_step(0, _, _):-!.
reverse_dag_step(I, A, B):-arg(I, A, E),
		E = _- S,
		reverse_direction(S, I, B),
		succ(J, I),
		reverse_dag_step(J, A, B).
%
reverse_direction([], _, _).
reverse_direction([J|Js], I, A):- arg(J, A, S),
		setarg(J, A, [I|S]),
		reverse_direction(Js, I, A).


%
product_with_atom(_, true, true):- !.
product_with_atom(_, false, false):-!.
product_with_atom(A, X, Y):- cofact(X, if(I, L, R)),
	product_with_atom(A, L, L0),
	product_with_atom(A, R, R0),
	make_pair(A, I, Pair),
	cofact(Y, if(Pair, L0, R0)).

%
succ_list_to_zdd(Initial, Dag, Initial0, Index):-
		internalize(Initial-Dag, U),
		U = Initial0-Dag0,
		succ_list_to_zdd(Dag0, Index).
%
succ_list_to_zdd([], _).
succ_list_to_zdd([X-As|G], IDX):-
	succ_list_to_zdd(G, IDX),
	elem(X, IDX, V),
	succ_list_to_zdd_with_index(As, X, V0, IDX),
	V = V0.
%
succ_list_to_zdd_with_index(false, _, false, _):-!.
succ_list_to_zdd_with_index(true, X, U, IDX):-!,
	succ_list_to_zdd_with_index([], X, U, IDX).
succ_list_to_zdd_with_index([], X, U, _):-!,
	cofact(U, if(X, true, false)).
succ_list_to_zdd_with_index([A], X, U, IDX):-!,
	elem(A, IDX, R),
	cofact(U, if(X, R, false)).
succ_list_to_zdd_with_index([A|As], X, U, IDX):-
	succ_list_to_zdd_with_index(As, X, V, IDX),
	elem(A, IDX, R),
	(	nonvar(R)
	-> 	cofact(W,if(X, R, false)),
		zdd([join(W, V, U)])
	;	R = false
%	;	throw(forward_reference_bug(A))
	).
% ?- module(zdd_misc).
% ?- time(zdd_misc:usual_n_queens(8, C)).

%@ % 68,761 inferences, 0.008 CPU in 0.008 seconds (98% CPU, 8581181 Lips)
%@ C = 92 .
% ?- time(zdd_misc:usual_n_queens(9, C)).
%@ % 304,612 inferences, 0.032 CPU in 0.033 seconds (99% CPU, 9450315 Lips)
%@ C = 352 .

%@ % 304,481 inferences, 0.033 CPU in 0.033 seconds (99% CPU, 9203827 Lips)
%@ C = 352 .
%@ % 304,612 inferences, 0.034 CPU in 0.034 seconds (99% CPU, 9056130 Lips)
% ?- time(zdd_misc:usual_n_queens(10, C)).
%@ % 1,394,360 inferences, 0.155 CPU in 0.156 seconds (100% CPU, 9002318 Lips)
%@ C = 724 .

% ?- time(zdd_misc:usual_n_queens(13, C)).
% ?- time(zdd_misc:usual_n_queens(14, C)).
%@ % 1,360,772,993 inferences, 175.021 CPU in 175.629 seconds (100% CPU, 7774894 Lips)
%@ C = 365596 .
%@ C = 365596.
% ?- time(zdd_misc:usual_n_queens(15, C)).

%
usual_n_queens(N, Count):- zdd:n_queen_matrix(N, M),
	usual_n_queens(M, [], [], [], S),
	length(S, Count).
%
usual_n_queens([], _, _, _, [[]]).
usual_n_queens([C|R], Po, Ne, Ho, U):-
	usual_n_queens(C, R, Po, Ne, Ho, U).
%
usual_n_queens([], _, _, _, _, []).
usual_n_queens([_I-J|D], R, Po, Ne, Ho, U):- memberchk(J, Ho), !,
		usual_n_queens(D, R, Po, Ne, Ho, U).
usual_n_queens([I-J|D], R, Po, Ne, Ho, U):- K is I+J, L is I-J,
	(	zdd:check_safe(K, L, Po, Ne)
	->	usual_n_queens(R, [K|Po], [L|Ne], [J|Ho], V),
		maplist(zdd:cons(K-L), V, V0),
		usual_n_queens(D, R, Po, Ne, Ho, U0),
		append(V0, U0, U)
	;   usual_n_queens(D, R, Po, Ne, Ho, U)
	).
