:- module(zdd_samples, [
   ]).

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(zdd(zdd)).
:- use_module(pac(op)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).
% :- use_module(zdd('zdd-misc')).
% :- use_module(zdd('zdd-graphviz')).
% :- use_module(zdd('zdd-euler')).
% :- use_module(zdd('zdd-reset-shift')).
attr_unify_hook(V, Y):-
	(	get_attr(Y, zdd, A)
	->	zdd_unify(V, A)
	; 	zdd_unify(V, Y)
	).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).


% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

% ?- zdd((P<<pow([1,2,3]), use_zdd(fibo_with_memo(10, Fibo)))).
% ?- zdd(zdd(zdd(fibo_with_memo(100, F)))).
%@ F = 354224848179261915075 .
% ?- time(zdd(use_zdd(fibo_with_memo(10000, F)))).
%@ % 633,519 inferences, 0.057 CPU in 0.062 seconds (92% CPU, 11169235 Lips)
%@ % 575,144 inferences, 0.057 CPU in 0.063 seconds (91% CPU, 10036716 Lips)

fibo_with_memo(0, 0):-!.
fibo_with_memo(1, 1):-!.
fibo_with_memo(N, F):- memo(N-F),
	(	nonvar(F) -> true
	;	N0 is N-1,
		N1 is N-2,
		fibo_with_memo(N0, F0),
		fibo_with_memo(N1, F1),
		F is F0 + F1
	).

% ?- time(fast_fibo(10000, F)).
%@ % 20,000 inferences, 0.018 CPU in 0.032 seconds (58% CPU, 1096852 Lips)
% ?- time(fast_fibo(20000, F)).
%@ % 39,999 inferences, 0.024 CPU in 0.039 seconds (63% CPU, 1637894 Lips)

fast_fibo(N, F):- fast_fibo(N, _, _, F).
%
fast_fibo(0, 0, 0, 0):-!.
fast_fibo(1, 0, 0, 1):-!.
fast_fibo(N, X, Y, F):- N0 is N-1,
	fast_fibo(N0, Y, _, X),
	F is X + Y.

% ?- zdd((P<<pow([1,2,3]), use_zdd(fibo_with_memo(100, Fibo), 1280), card(P,C))).
%%	fibonacci(+N, -F) is de<t.
% Unify F with the value of the Fibonacci function for N.
% :-table naive_fibo/2.
% ?- time(naive_fibo(10, F)).
% ?- call_with_time_limit(60, naive_fibo(1000, _)).
%@ ERROR: Unhandled exception: Time limit exceeded   <== without table.
% ?- time(use_zdd(fibo_with_memo(1000, X))).
%@ % 51,652 inferences, 0.005 CPU in 0.005 seconds (93% CPU, 10338671 Lips)
% ?- time(use_zdd(fibo_with_memo(10000, X))).
%@ % 573,228 inferences, 0.054 CPU in 0.058 seconds (92% CPU, 10662525 Lips)
% ?- time(naive_fibo(1000, _)).
% ?- time(naive_fibo(10000, _)).
% ?- time(naive_fibo(100 000, _)).

:- table naive_fibo/2.
naive_fibo(0, 0):-!.
naive_fibo(1, 1):-!.
naive_fibo(X, S):- X0 is X-1,
	X1 is X0-1,
	naive_fibo(X0, S0),
	naive_fibo(X1, S1),
	S is S0 + S1.

		/******************
		*     N-queens    *
		******************/

% ?- time(zdd(n_queens(8, C))).
% ?- time(zdd(n_queens(9, C))).
% ?- time(zdd(n_queens(10, C))).
% ?- time(zdd(n_queens(11, C))).
%@ % 11,708,109 inferences, 1.087 CPU in 1.103 seconds (99% CPU, 10769276 Lips)
%@ C = 2680 .
% ?- time(zdd(n_queens(12, C))).  % [2021/04/21]
%@ % 61,028,680 inferences, 6.153 CPU in 6.217 seconds (99% CPU, 9918445 Lips)
%@ C = 14200 .

% ?- time(zdd(n_queens(12, C))).  % [2020/10/16]
%@ % 60,919,433 inferences, 5.954 CPU in 5.968 seconds (100% CPU, 10232153 Lips)
%@ C = 14200 .
%@ % 11,708,109 inferences, 1.087 CPU in 1.103 seconds (99% CPU, 10769276 Lips)
%@ C = 2680 .
% ?- profile(zdd(n_queens(12, C))).  % [2018/10/25]
%@ C = 14200.
% ?- time(zdd(n_queens(12, C))).  % [2018/11/04]
%@ % 54,199,597 inferences, 6.360 CPU
%@ C = 14200 .
% ?- time(zdd(n_queens(13, C))). % [2021/04/21]
%@ % 352,183,186 inferences, 37.683 CPU in 37.882 seconds (99% CPU, 9345887 Lips)
%@ C = 73712 .

% ?- time(zdd(n_queens(13, C))).  % [2018/11/01]
%@ % 288,612,429 inferences, 36.548 CPU
% ?- time(zdd(n_queens(14, C))).	% [2018/10/25]
%@ % 1,743,638,761 inferences, 256.849 CPU
%@ C = 365596 .
% ?- time(zdd(n_queens(14, C))).	% [2020/10/16]
%@ % 2,003,469,945 inferences, 221.555 CPU in 223.231 seconds (99% CPU, 9042752 Lips)
%@ C = 365596 .

%%	n_queens(+N, -C) is det.
% Unify C with the number of solutions of N-queen puzzle.

n_queens(N, Count):- @(n_queens(N, Count)).

n_queens(N, Count, State):- n_queen_matrix(N, M),
	n_queens(M, [], [], [], S, State),
	card(S, Count, State).
%
n_queens([], _, _, _, 1, _).
n_queens([C|R], Po, Ne, Ho, U, S):-
	n_queens(C, R, Po, Ne, Ho, U, S).

%
n_queens([], _, _, _, _, 0, _).
n_queens([_I-J|D], R, Po, Ne, Ho, U, S):- memberchk(J, Ho), !,
		n_queens(D, R, Po, Ne, Ho, U, S).
n_queens([I-J|D], R, Po, Ne, Ho, U, S):- K is I+J, L is I-J,
	(	check_safe(K, L, Po, Ne)
	->	n_queens(R, [K|Po], [L|Ne], [J|Ho], V, S),
		zdd_insert(K-L, V, V0, S),
		n_queens(D, R, Po, Ne, Ho, U0, S),
		zdd_join(V0, U0, U, S)
	;   n_queens(D, R, Po, Ne, Ho, U, S)
	).

% ?- matrix([1,2],[a,b], C).
% ?- matrix([1,2],[], C).
% ?- matrix([],[a,b], C).

%%	matrix(+X, +Y, -Z) is det.
% Unify Z with the cartesian product of X and Y.
matrix(X, Y, Z):- matrix(X, Y, Z, []).
%
matrix([], _, L, L).
matrix([A|As], Bs, [C|L], M):-
	matrix(Bs, A, C, []),
	matrix(As, Bs, L, M).
%
matrix([], _, L, L).
matrix([B|Bs], A, [A-B|L], M):- matrix(Bs, A, L, M).

% ?- n_queen_matrix(10, M).
n_queen_matrix(N, M):- N0 is N-1,
	numlist(0, N0, L),
	matrix(L, L, M).
%
check_safe(K, L, Po, Ne):-
	(	(memberchk(K, Po); memberchk(L, Ne))
	->	false
	;	true
	).
