
:- module(zdd,
  [		card/2, card/3
	,	ltr_cofact/2, ltr_cofact/3
	,	ltr_join/3,	ltr_join/4
  	,	ltr_merge/3, ltr_merge/4, ltr_card/3, ltr_card/4
  	,	sat/1, sat_count/1, sat_count/2
    ,	sat0/1, sat_count0/1, sat_count0/2
    ,	fos/1, fos_count/0, fos_count/1, fos_count/2
	,	make_boole_canonical/2, simple_boole/2
	,	boole_to_dnf/2, ltr_card/3
	,	dnf/2, dnf/3, dnf0/2, dnf0/3
	, 	cnf/2, cnf/3, cnf0/2, cnf0/3
  	,	zdd_funs/3, zdd_funs/4
	,	cofact0/3, cofact0/4
	,	zdd/1
	, 	zdd_append/3,  zdd_append/4, zdd_insert/4
	, 	zdd_insert_atoms/3, zdd_insert_atoms/4
	,	zdd_has_1/2
	,	zdd_join/3, zdd_join/4
  	,	zdd_singleton/2, zdd_singleton/3
	,	zdd_merge/3, zdd_merge/4
	,	zdd_meet/3,  zdd_meet/4
	,	zdd_subtr/3, zdd_subtr/4
  	,	zdd_divide/3,  zdd_divide/4, zdd_residue/3, zdd_residue/4
	,	zdd_div_by_list/3,  zdd_div_by_list/4
  	,	zdd_res_by_list/3,  zdd_res_by_list/4
	,	zdd_power/2, zdd_power/3
  	,	get_extra/1, get_extra/2
	,	get_key/2, get_key/3, set_key/2, set_key/3
	,	delete_key/1, delete_key/2
	,	intern/3
	,	make_boole_canonical/2
	,	use_zdd/1, use_zdd/2
	,   zdd_shift/1
    ,	psa/1, psa/2, sets/2, sets/3, ppoly/1, ppoly/2, poly_term/2, poly_term/3
	,	@/1, significant_length/3
	,	charlist/2, charlist/3, atomlist/2
  ]).

% ?- fos(pow(1..3)//(1+2)), fos_count.
% ?- fos(pow(1..3)//(1*2)), fos_count.

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).
% :- use_module(zdd('zdd-misc')).
% :- use_module(zdd('zdd-graphviz')).
% :- use_module(zdd('zdd-euler')).
% :- use_module(zdd('zdd-reset-shift')).
attr_unify_hook(V, Y):-
	(	get_attr(Y, zdd, A)
	->	zdd_unify(V, A)
	; 	zdd_unify(V, Y)
	).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).


% ?- zdd((X<<(pow([a,b])-pow([a])), card(X, C))).

term_expansion --> pac:expand_pac.

:- meta_predicate @(:).
@(X):- shift(X).

% ?- zdd({X=1}).
% ?- zdd(X=1).
% ?- zdd((append([], [], X))).
% ?- zdd((true, true)).
% ?- zdd((Y<<cnf((a+b)*(-(a)+c)), ltr_remove(-(a), Y, U),  sets(U, S))).
% ?- numlist(1, 16, L), zdd((X<< pow(L), card(X, C))).
% ?- numlist(1, 16, L), zdd((X<< pow(L))).
% ?- numlist(1, 10000, L), zdd((X<< pow(L), card(X, C))).
% ?- numlist(1, 16, L),
% zdd((X<< pow(L), Y<<pow(L), Z << (X+Y),  U<< (X*Y), card(X, C))).
% ?- time((numlist(1, 40, L),
% zdd((X << pow(L), Y<<pow(L), Z << merge(X, Y), card(Z, C))))).
% ?- zdd((setmemo(a-1), setmemo(a-2), memo(a-V))).
% ?- zdd(X<<{[1,1]}).
% ?- zdd((X<<pow([1,2,3]), prz(X))).
% ?- zdd((X<< [1,2,3,1], prz(X))).
% ?- zdd((X<< [[1],[1]], prz(X))).
% ?- zdd((X<< [[a-b, b-c]], prz(X))).
% ?- zdd( use_zdd((memo(a-1), memo(a-X)))).
% ?- zdd(use_zdd(zdd(use_zdd((memo(a-1), memo(a-X)))))).
% ?- zdd((X<<pow([1,2]), {zdd((Y<<pow([a,b]), sets(Y, Sy)))}, sets(X, Sx))).
% ?- zdd((X<<pow([1,2]), {zdd((Y<<pow([a,b]), sets(Y, Sy)))},
%	sets(X, Sx), sets(Y, Sy1))).
% ?- zdd((set_key(root, hello), get_key(root, R))).
% ?- zdd((intern(a, X), show_state)).
% ?- zdd((intern(a, X), intern(b, Y))).
% ?- zdd((intern(A, X), intern(B, Y))).
% ?- zdd((zdd_funs([a],[b], F), psa(F))).
% ?- zdd((zdd_funs([],[b], F), psa(F))).
% ?- zdd((zdd_funs([a],[], F), psa(F))).
% ?- zdd((zdd_funs([a,b],[c,d], F), psa(F), card(F, C))).
% ?- N=3,  numlist(1, N, Ns), zdd((zdd_funs(Ns, Ns, F), psa(F), card(F, C))).
% ?- N=3, M=5,  numlist(1, N, Ns), numlist(1, M, Ms),
%	zdd((zdd_funs(Ns, Ms, F), card(F, C))).
% ?- time((N=100, M=100,  numlist(1, N, Ns), numlist(1, M, Ms),
%	zdd((zdd_funs(Ns, Ms, F), card(F, C))))).

% Cardinality # of Y^X = { f | f: X-> Y }.
%   ({1,...,K}^L)^({1,...,I}^J) = (K^L)^(I^J).

% ?- I=1, J=1, K=1, L=1,
%	numlist(1, I, X), numlist(1, K, Y),
%	raise_list(X, J, Xj),
%	raise_list(Y, L, Yl),
%	time( ( call_with_time_limit(200, (
%		zdd((zdd_funs(Xj, Yl, F, default_pair), card(F, C))))),
%		C =:= (K^L)^(I^J),
%	    significant_length(C, Keta, 10))).

% ?- I=3, J=5, K=3, L=5,
%	numlist(1, I, X), numlist(1, K, Y),
%	raise_list(X, J, Xj),
%	raise_list(Y, L, Yl),
%	time( ( call_with_time_limit(200, (
%		zdd((zdd_funs(Xj, Yl, F), card(F, C))))),
%		C =:= (K^L)^(I^J),
%		significant_length(C, Keta, 10))).

significant_length(0, 0, _):-!.
significant_length(X, 1, Radix):- X < Radix, !.
significant_length(X, L, Radix):- Y is X // Radix,
	significant_length(Y, L0, Radix),
	L is L0+1.
%

default_pair(X, Y, X-Y).

equational_mapsto([A,B], C, A+B = C).

:- meta_predicate zdd_funs(?,?,?,3).
zdd_funs(D, R, F):- zdd_funs(D, R, F, default_pair).

:- meta_predicate zdd_funs(?,?,?,3).
zdd_funs(D, R, F, Pred):- shift(zdd_funs(D, R, F, Pred)).
%
zdd_funs([], _, 1, _, _).
zdd_funs([A|As], Bs, F, Pred, S):-
	zdd_funs(As, Bs, F0, Pred, S),
	extend_funs_by_one(A, Bs, F0, 0, F, Pred, S).
%
extend_funs_by_one(_, [], _, F, F, _, _):-!.
extend_funs_by_one(A, [B|Bs], H, F0, F, Pred, S):- call(Pred, A, B, Pair), !,
	zdd_insert(Pair, H, H0, S),
	zdd_join(H0, F0, F1, S),
	extend_funs_by_one(A, Bs, H, F1, F, Pred, S).
extend_funs_by_one(A, [_|Bs], H, F0, F, Pred, S):-
	extend_funs_by_one(A, Bs, H, F0, F, Pred, S).

% ?- N = 1000, numlist(1, N, Ns),
%	time(zdd((X<<pow(Ns), rank_family_by_card(X, P),
%	memo(family_with_card(500)-L), card(L, C)))).
%@ % 90,174,772 inferences, 10.355 CPU in 10.540 seconds (98% CPU, 8708585 Lips)

% ?- N = 1000, numlist(1, N, Ns),
%	time(zdd(( X<<pow(Ns),
%			  get_family_of_rank(500, X, Y),
%			  card(Y, C)))).

%% get_family_of_rank(Order, X, F) is det.
%  Collect sets of a given cardinality.

get_family_of_rank(Order, X, F):- shift(get_family_of_rank(Order, X, F)).
%
get_family_of_rank(Order, X, F, S):- rank_family_by_card(X, _, S),
	memo(family_with_card(Order)-F, S).


% ?- zdd((X<<set([a]), rank_family_by_card(X, P),
%	memo(family_with_card(0)-L), sets(L, Sl),
%   memo(family_with_card(1)-M), sets(M, Sm))).

%% rank_family_by_card(X, Y[, S]) is det.
%
% Classifying sets in a family by size.
rank_family_by_card(X, Y):- shift(rank_family_by_card(X, Y)).
%
rank_family_by_card(0, 0, _):-!.
rank_family_by_card(1, 0, S):-!, memo(family_with_card(0)-1, S).
rank_family_by_card(I, P, S):- memo(rank_family_by_card_done(I)-B, S),
	(	nonvar(B) ->  true
	; 	cofact(I, t(A, L, R), S),
		rank_family_by_card(L, Pl, S),
		rank_family_by_card(R, Pr, S),
		max(Pl, Pr, P0),
		insert_one_to_family(A, P0, New, S),
		P is P0 + 1,
		memo(family_with_card(P)-New, S),
		B = true
	).
%
insert_one_to_family(A, 0, G, S):-!, memo(family_with_card(0)-F, S),
	zdd_insert(A, F, G, S).
insert_one_to_family(A, P, G, S):- P0 is P-1,
	insert_one_to_family(A, P0, G0, S),
	memo(family_with_card(P)-Fp, S),
	zdd_insert(A, Fp, G, S),
	zdd_join(Fp, G0, Gp, S),
	setmemo(family_with_card(P)-Gp, S).

%
max(A, B, A):- A@>B, !.
max(_, B, B).


		/**********************
		*     State access    *
		**********************/

get_extra(Extra):- shift(get_extra(Extra)).
get_extra(Extra, s(_, Extra)).

%
set_extra(Extra):- shift(set_extra(Extra)).
set_extra(Extra, S):- setarg(2, S, Extra).
%
get_key(K, V):- shift(get_key(K, V)).
get_key(K, V, s(_, Assoc)):- memberchk(K-V, Assoc).

set_key(K, V):- shift(set_key(K, V)).
set_key(K, V, S) :- S = s(_, Assoc),
	(	select(K-_, Assoc, Assoc0)
	->	setarg(2, S, [K-V|Assoc0])
	;	setarg(2, S, [K-V|Assoc])
	).

delete_key(K):- shift(delete_key(K)).
%
delete_key(K, S):- S = s(_, Extra),
	(	select(K-_, Extra, Extra0) -> setarg(2, S, Extra0)
	;	true
	).

%
get_state(X):- shift(=(X)).

%
set_state(X):- shift(set_state(X)).
set_state(X, Y):- arg(1, X, A),
	setarg(1, Y, A),
	arg(2, X, B),
	setarg(2, Y, B).
%
show_state :- get_state(s(#(N, #(K,_),_), Extra)),
	format("array max = ~w~n", [N]),
	format("hash keys count = ~w~n", [K]),
	format("extra = ~w~n", [Extra]).
%
get_vector(X):- shift(get_vector(X)).
get_vector(Vec, s(#(_,_,Vec),_)).
%
vector(X):- get_vector(X).

% ?- zdd((X<<pow([a,b,c,d]), zdd_dump, card(X, C))).
zdd_dump:- get_state(s(#(N, _, Vec), _)),
	forall(	between(2, N, I),
			(	arg(I, Vec, X),
				format("~w = ~w\n", [I, X])
			)).

% ?- time((numlist(1, 10000, Ns), zdd((X<<pow(Ns),
%	zdd_export(X, Y, S), card(X, Mx),  card(Y, My, S))))),
%	Mx = My.
% ?- zdd((zdd((X<<pow([a,b]), zdd_export(X, Y, S), card(X, Sx))), card(Y, Sy, S))).
% ?- zdd((zdd((X<<pow([a,b]), zdd_export(X, Y, S))), card(X, Sx), card(Y, Sy, S))).  % false

zdd_export(I, I0, S0):- shift(zdd_export(I, I0, S0)).
%
zdd_export(I, I0, S0, S):-  open_state(S0),
	zdd_shift(zdd_export(I, S, I0, S0)).
%
zdd_export(I, _, I, _, _):- I<2, !.
zdd_export(I,  S, I0, S0,  M):- memo(I-I0, M),
	(	nonvar(I0)	-> true
	;	cofact(I, t(A, L, R), S),
		zdd_export(L, S, L0, S0, M),
		zdd_export(R, S, R0, S0, M),
		cofact(I0, t(A, L0, R0), S0)
	).

% ?- zdd((varnum(D), varnum(E))).

%%	varnum(?D) is det.
% update a global variable varnum with D.
%
varnum(D):- shift(get_key(varnum, D)).

% ?- zdd((cofact0(X, a, 1), psa(X), cofact0(X, A, B))).
% ?- zdd((cofact0(X, 1, a), psa(X), cofact0(X, A, B))).
% ?- zdd((cofact0(X, 1, 0), psa(X), cofact0(X, A, B))).

%% cofact0(X, A, Y[, G]) is det.
%	 Restriction of cofact/[2,3] to a singleton family.

cofact0(X, A, Y):- shift(cofact0(X, A, Y)).
%
cofact0(X, A, Y, S):- var(X), !, cofact(X, t(A, 0, Y), S).
cofact0(X, A, Y, S):- X > 1, cofact(X, t(A, 0, Y), S).

% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([a,c], X))).
% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([a,c,d], X))). % false
% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([b,c], X))).
% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([], X))).

%% zdd_memberchk(L, Z[, G]) is det.
%	A list L as a set is a member of a family Z of sets.
zdd_memberchk(L, Z):- shift(zdd_memberchk(L, Z)).
%
zdd_memberchk(L, Z, S):- sort(L, L0),
	zdd_ord_memberchk(L0, Z, S).
%
zdd_ord_memberchk(L, Z):- shift(zdd_ord_memberchk(L, Z)).
%
zdd_ord_memberchk([], Z, S):- !, zdd_has_1(Z, S).
zdd_ord_memberchk([A|As], Z, S):- Z>1,
	cofact(Z, t(B, L, R), S),
	(	A == B
	->	zdd_ord_memberchk(As, R, S)
	;	A @> B,
		zdd_ord_memberchk([A|As], L, S)
	).

% ?- zdd((X=1, Y =2, Z is X + Y)).
% ?- zdd((X<<[a], sets(X, S))).
% ?- zdd((X<<[a], S<<sets(X))).
% ?- zdd((X<<pow([a,b]), S<<sets(X))).
% ?- zdd((X<<pow([a,b,c,d,e,f]), Y<<pow([a,b,c,d,e,f]), U<<(X-Y), S<<sets(U))).
% ?- zdd((X<< ([a,b]+[c,d]), S<<sets(X))).
% ?- zdd((X<<([a,b]&[c,d]), S<<sets(X))).
% ?- zdd((X<<{[a],[b],[c]}, sets(X,S))).
% ?- zdd((X<<{[a],[b, c]}, sets(X,S))).
% ?- zdd((X<<{[a],[b, c]}, Y<<{[c, b, c]}, Z<<(X-Y), sets(Z,S))).
% ?- zdd((X<<{[a],[b, c]}, Y<<{[b, c], [a], [a]}, Z<< (X-Y), U<<sets(Z))).
% ?-  I =1000, J =2000,
%	time(zdd(( numlist(1, I, L), numlist(1, J, M), X << pow(L), Y << pow(M),
%	 Z << (Y - X), card(Z, C), C =:= 2^J-2^I))).
% ?- I =1000, J is I + 1, numlist(1, I, L), prepare_ltr_list(L, L0),
%	time(zdd(( X<<ltr_ord_power(L0), ltr_has_set([-(1), 3, 3,4, 6, I],X)))).

% ?- zdd((X<<cnf(a), Y<<sets(X))).
% ?- zdd((X<<cnf(-a), Y<<sets(X))).
% ?- zdd((X<<cnf(a+b), Y<<sets(X))).
% ?- zdd((X<<cnf(-(a+b)), Y<<sets(X))).
% ?- zdd((X<<cnf(-(a+b+b+a)), Y<<sets(X))).
% ?- zdd((X<<cnf(-(-(a+b+b+a))), Y<<sets(X))).
% ?- zdd((X<<cnf(-(-a + -b)), Y<<sets(X))).
% ?- zdd((X<<dnf(a), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-a)), Y<<sets(X))).
% ?- zdd((X<<dnf(a+b), Y<<sets(X))).
% ?- zdd((X<<dnf(-(a+b)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(a+b+b+a)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-(a+b+b+a))), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-a + -b)), Y<<sets(X))).
% ?- zdd((X<<dnf((-a + a)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-a + a)), Y<<sets(X))).
% ?- zdd(X << append([a,b], [c,d])).
% ?- zdd((X << (([a,b]+[c,d])*[c,d]), sets(X, S))).
% ?- zdd((X << (([a,b]+[c,d])*[c, d, d]), sets(X, S))).
% ?- zdd((X << dnf(a*b+c*d+c*d*d), sets(X, S))).
% ?- zdd((X << dnf(a*b+c*d+c*d*d))).
% ?- zdd((zdd_list_to_singletons([], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([a], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([a,b], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([c, a, b], X), prz(X))).

%% zdd_list_to_singletons(+As, -X) is det.
%  X is unified with a ZDD I such that I is the family of
%  singletons of an element of X.

%  For example, if  X=[a,b,c] then ZDD I is [[a], [b], [c]]
%  as a family of sets in ZDD.

zdd_list_to_singletons(As, X):- shift(zdd_list_to_singletons(As, X)).
%
zdd_list_to_singletons([], 1, _).
zdd_list_to_singletons([A|As], X, S):-zdd_list_to_singletons(As, Y, S),
		zdd_singleton(A, Y0, S),
		zdd_join(Y0, Y, X, S).

% ?- zdd((zdd_partial_choices([], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a], [b]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a1], [b, b1], [c, c1]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a], [a, b], [a, c]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[b, a], [a, b], [a, a, b, b]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[c, c], [b, b, b], [a, a]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a], [], [a, c]], X), prz(X)).)
% ?- zdd((zdd_partial_choices([[a], [b], [c], [a], [d]], X), prz(X), card(X, C))).
% ?- zdd((zdd_partial_choices([[a,a1], [b,b1]], X), prz(X), card(X, C))).

% ?- findall([I], between(1, 10000, I),  Ls),
%	zdd((zdd_partial_choices(Ls, X),
%	card(X, C))), !, C =:= 2^10000.

%%  zdd_partial_choices(+X, Y[, G]) is dete.
%    For a list X of lists. Y is the ZDD of all possible partial
%	 choices from X.
%   Example:   if X=[[a],[b]]  then Y =[[], [a], [b], [a, b]]
%			as a ZDD.

zdd_partial_choices(X, Y):- shift(zdd_partial_choices(X, Y)).
%
zdd_partial_choices([], 1, _):-!.
zdd_partial_choices([L|Ls], X, S):-  zdd_partial_choices(Ls, X0, S),
	zdd_fold_insert(L, X0, 0, X1, S),
	zdd_join(X0, X1, X, S).

% ?- zdd((dnf((A+B)*(C+D)*(E+F), F), sets(F, G))), maplist(writeln, G).

% ?- zdd((dnf((A+B)*(C+D)*(E+F), F1), sets(F1, G))), maplist(writeln, G).
% ?- zdd((dnf((A+B)*(C+D)*(E+F), G), zdd_dump)).
% ?- zdd:zdd((P<<pow([1,2,3]), use_zdd((X=1;X=2)), card(P,C))).
% ?- zdd((P << (set([1,2,3,2,3])+set([4, 2,4])), prz(P))).
% ?- zdd({append([a,b], [c,d], X)}).
% ?- zdd(X << append([a,b], [c,d])).
% ?- zdd(zdd(X << append([a,b], [c,d]))).
% ?- zdd(zdd(zdd((X << append([a,b], [c,d]), psa(X))))).
% ?- zdd(zdd(zdd((X<<pow([1,2]), true, true)))).
% ?- zdd(zdd(zdd((X<<pow([1,2]), true, true, card(X, C))))).

%% zdd(+E) is det.
%   E is interpreted as commands on ZDD.

:- meta_predicate zdd(:).
zdd(Exp):- use_zdd(Exp).

% ?- listing(zdd).

:- meta_predicate zdd(0, ?).
zdd(M:A, S):- zdd(A, S, M).

%% zdd0(E) is det.
%  Same as zdd(E), without closing the state.
:- meta_predicate zdd0(:).
zdd0(Exp):- open_state(S), zdd(Exp, S).

% ?- open_state(M), zdd_reset_continue((memo(a-1), memo(a-R)), M), close_state(M).
% ?- open_state(M), zdd_reset_continue(( zdd((X<<pow([1,2]),
%	memo(a-X), card(X, C)))), M).

%% zdd_reset_continue(+G) is det.
%  Reset goal G and coninue the rest.

:- meta_predicate zdd_reset_continue(:, ?).
zdd_reset_continue(Goal, S):- reset(Goal, Ball, Rest),
	(	var(Ball) -> true
	;	call(Ball, S)
	),
	zdd(Rest, S).

%% use_zdd(X[, HashSize]) is det.
%   Create a state S of a given hash size, and act command X on S.
%	On exit, the state is closed.
:- meta_predicate  use_zdd(:).
use_zdd(X):- use_zdd(X, [hash_size(128)]).

:- meta_predicate  use_zdd(:, ?).
use_zdd(X, Options):-
	setup_call_cleanup(
		open_state(S, Options),
		zdd(X, S),
		close_state(S)).

%% zdd_shift(G) is det.
%  Perform shift-wrapped command.  Short hand for used_zdd(shift(G))
%  for convenience.
:- meta_predicate zdd_shift(:).
zdd_shift(G):- use_zdd(shift(G)).


% ?- zdd((X<<pow([1,2]), memo(a-X), card(X, C))).

%%   zdd(E, S, M) is det.
%   E is command expression,  S a state, M a module name.
%	E is interpred as a command on S in module M.
%	(A, B)		(A, B).sequentiual (and)
%	X<<E		evaluate E, then unify X with it.
%	M:G			evaluate G under module M
%	true		true
%	0		    treated as true
%	{G}			plain prolog goal.
%	(A ; B)		(A; B) (or)
%	(A -> B)	A->B   (condional)
%	Other		Reset, and handle the ball and continuation.

zdd((A,B), S, M):-!, zdd(A, S, M), zdd(B, S, M).
zdd(X<<E, S, _):-!, zdd_eval(E, X, S).
zdd(M:G, S, _):-!, zdd(G, S, M).
zdd(0, _, _):-!.
zdd(true, _, _):-!.
zdd({G}, _, _):-!, call(G).
zdd((A;B), S, M):-!, (zdd(A, S, M); zdd(B, S, M)).
zdd((A->B), S, M):-!, (zdd(A, S, M)->zdd(B, S, M)).
zdd(Cont, S, M):- zdd_reset_continue(M:Cont, S).

% ?- zdd((X<<cofact(a, 1, 1), psa(X), Y<<cofact(X), card(X, C))).

%% zdd_eval(+E, -V, +S) is det.
%   V is the value of an expression E with working on a state S.

%	x		x  (integer)
%	x       x  (list)
%	set		a singleton family.
%	{a,b,..}  a family of sets.
%   +		join
%	-		subtract
%	\		subtract
%	*		intersection
%	merge   merge
%	&		merge
%	prod	product
%	**		product
%	pow		powerset
%	power	powerset
%	sets	convert to a usual list of lists
%	cnf		CNF
%	dnf		DNF
%	exp(E)  user-defined expression
%   cofact/3	cofact  (for tutorial purpose, to be dropped)
%   cofact/1	cofact  (for tutorial purpose), to be dropped)
%	prolog	call with last argument as output
%   E     evaluate E recursively.

% ?- zdd((X << card(pow([a,b])))).
% ?- zdd((X << pow([a,b]))).
% ?- zdd((X << [a, b, c])).
% ?- zdd((X << set([a, b, c]))).
% ?- zdd((X << (set([a, b, c]) + set([1,2,3])), psa(X))).
% ?- zdd((C << card(pow([a,b,c,1,2,3])))).
% ?- zdd((C << card(pow(prolog(append([a,b,c], [1,2,3])))))).

%
zdd_eval(X, X, _)	:- integer(X), !.  % for any x >= 0.
zdd_eval(L, Y, _)	:- is_list(L), !, sort(L, Y).
zdd_eval(set(L), Y, S)	:- zdd_eval(L, L0, S),
	zdd_append(L0, 1, Y, S).
zdd_eval(ltr_set(L), Y, S)	:- zdd_eval(L, L0, S),
	ltr_append(L0, 1, Y, S).
zdd_eval({X}, Y, S)	:-!, associative_comma(X, U, []), zdd_family(U, Y, S).
zdd_eval(dnf(A), X, S):-!, dnf(A, X, S).
zdd_eval(cnf(A), X, S):-!, cnf(A, X, S).
zdd_eval(gf2(A), X, S):-!, gf2(A, X, S).
zdd_eval(fos(A), X, S):-!, fos(A, X, S).
zdd_eval(X + Y, A1, S):-!, zdd_eval(X, A2, S),
	zdd_eval(Y, A3, S),
	zdd_join(A2, A3, A1, S).
zdd_eval(X * Y, A1, S):-!, zdd_eval(X, A2, S),
	zdd_eval(Y, A3, S),
	zdd_meet(A2, A3, A1, S).
zdd_eval(X - Y, A1, S):-!, zdd_eval(X, A2, S),
	zdd_eval(Y, A3, S),
	zdd_subtr(A2, A3, A1, S).
zdd_eval(\(X,Y), A, S):-!, zdd_eval(X-Y, A, S).
zdd_eval(merge(X, Y), A1, S):-!, zdd_eval(X, A2, S),
	zdd_eval(Y, A3, S),
	zdd_merge(A2, A3, A1, S).
zdd_eval(&(X, Y), A, S):-!, zdd_eval(merge(X, Y), A, S).
zdd_eval(prod(X, Y), A1, S):-!, zdd_eval(X, A2, S),
	zdd_eval(Y, A3, S),
	zdd_product(A2, A3, A1, S).
zdd_eval(**(X, Y), A, S):-!, zdd_eval(prod(X, Y), A, S).
zdd_eval(pow(X), A, S) :-!, zdd_eval(X, X0, S),
	sort(X0, X1),
	zdd_power(X1, A, S).
zdd_eval(power(X), A, S):-!, zdd_eval(pow(X), A, S).
zdd_eval(sets(X), A, S)	:-!, zdd_eval(X, Y, S),
	sets(Y, A, S).
zdd_eval(cofact(A, L, R), X, S):-!, cofact(X, t(A, L, R), S).
zdd_eval(cofact(X), Y, S):-!, cofact(X, Y, S).
zdd_eval(prolog(E), Y, _):-!, call(E, Y).
zdd_eval(E, A, S) :- E =.. [F|As],
	maplist(zdd_eval_flip(S), As, Bs),
	E0 =.. [F|Bs],
	zdd(call(E0, A), S).

%
zdd_eval_flip(S, A, B):- zdd_eval(A, B, S).

% ?- zdd(C<<card(pow([a,b]))).
% ?- zdd((X<<pow([a, b, c]), zdd_super_power([a, b], X, Y), psa(Y))).

% ?- N = 2, numlist(1, N, Ns), cartesian(Ns, Ns, P),
%	zip(Ns, Ns, Z),  maplist(pred([A-B, (A, B)]), Z, Z0),
%	zdd((X<<pow(P),
%		zdd_super_power(Z0, X, Q),
%		card(Q, C))), C=:= 2^(N^2-N).

% ?- N = 100, numlist(1, N, Ns), cartesian(Ns, Ns, P),
%	zip(Ns, Ns, Z),  maplist(pred([A-B, (A, B)]), Z, Z0),
%	time(zdd((X<<pow(P), zdd_super_power(Z0, X, Q),
%		 card(Q, C)))), C=:= 2^(N^2-N).


zdd_super_power(X, Y, Z):- shift(zdd_super_power(X, Y, Z)).

zdd_super_power([], P, P, _).
zdd_super_power([A|As], P, Q, S):-
	zdd_super_power(As, P, Q0, S),
	zdd_insert(A, Q0, Q, S).

%
family(X, Y):- shift(family(X, Y)).

family(X, Y, S):- zdd_family(X, Y, S).
% ?- zdd((zdd_family([[a],[a,b]], X), card(X, C), psa(X))).


zdd_family(X, Y):- shift(zdd_family(X, Y)).
%
zdd_family(X, Y, S):- zdd_family(X, 0, Y, S).

% ?- zdd((zdd_family([[a],[a,b]], X), card(X, C), psa(X))).
% ?- zdd((zdd_family([[], [a],[a,b]], X), psa(X))).
%
zdd_family([], U, U, _).
zdd_family([X|Xs], U, V, S):-
	zdd_append(X, 1, X0, S),
	zdd_join(X0, U, U0, S),
	zdd_family(Xs, U0, V, S).

% ?- zdd((ltr_family([[a, b], [c, d]], X), sets(X,S))).
% ?- zdd((ltr_family([[a, b], [c, -c]], X), sets(X,S))).
% ?- zdd((ltr_family([[a,-b, a],[b, a, -b]], X), sets(X,S))).
% ?- zdd((ltr_family([[a],[a,b]], X), ltr_card(3, X, C), psa(X))).
% ?- zdd((ltr_family([[], [a],[a,b]], X), ltr_card(2, X, C), psa(X))).
% ?- zdd((X << cnf((a+(-b)+a)*(b+a+(-b))), S<<sets(X))).
% ?- zdd((X << dnf((a+(-b)+a)), psa(X))).
% ?- zdd((X << dnf((a+(-b)+a)* (b + a+(-b))), psa(X))).

%
ltr_family(X, Y):- shift(ltr_family(X, Y)).
%
ltr_family(X, Y, S):- ltr_family(X, 0, Y, S).

%
ltr_family([], U, U, _).
ltr_family([X|Xs], U, V, S):-
	ltr_append(X, 1, X0, S),
	ltr_join(X0, U, U0, S),
	ltr_family(Xs, U0, V, S).


% ?- associative_comma((([], [a]), [b]), P, []).
associative_comma((X, Y), P, Q):-!, associative_comma(X, P, P0),
	associative_comma(Y, P0, Q).
associative_comma(X, [X|P], P).

% ?- zdd((X<<{[], [a]}, psa(X), zdd_remove_1(X, Y), psa(Y))).
% ?- zdd((X<<{[a]}, zdd_remove_1(X, U))).  % false
zdd_remove_1(X, Y):- shift(zdd_remove_1(X, Y)).
%
zdd_remove_1(1, 0, _):-!.
zdd_remove_1(X, Y, S):- X>2,
	cofact(X, t(A, L, R), S),
	zdd_remove_1(L, L0, S),
	cofact(Y, t(A, L0, R), S).


% ?- zdd(( (X<< {[a,b,c]}, zdd_choices(X, Y), psa(Y)))).
% ?- zdd(( (X<< [a,b,c], zdd_choices(X, Y), psa(Y)))).
% ?- zdd(( (X<< {[a, b], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[], [a, b], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[a], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b], [c, d], [e,f]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd((X<<([a,b]+[1,2]), zdd_choices(X, Y), psa(Y), card(Y, C))).
% ?- zdd((A<<charlist(a,z), B<<numlist(1,10), X<<(A+B),
%	zdd_choices(X, Y), card(Y, C))).
% ?- charlist(a,z, AZ), zdd((A<< AZ, ltr_choices(A, Y), card(Y, C))).
% ?- zdd((A<<numlist(1, 2), psa(A), zdd_choices(A, Y), card(Y, C))).

zdd_choices(X, Y):- shift(zdd_choices(X, Y)).
%
zdd_choices(1, 0, _):-!.
zdd_choices(0, 1, _):-!.
zdd_choices(X, Y, S):- memo(zdd_choices(X)-Y, S),
	(	nonvar(Y)	->	true
	;	zdd_choices(X, [], 1, Y, S)
	).
%
zdd_choices(0, _, Y, Y, _):-!.
zdd_choices(1, As, Y0, Y, S):-!, zdd_fold_insert(As, Y0, 0, Y, S).
zdd_choices(X, As, Y0, Y, S):- cofact(X, t(A, L, R), S),
	(	L = 0
	->	zdd_choices(R, [A|As], Y0, Y, S)
	; 	L = 1
	->	(	As = []
		->  Y = 0
		;	zdd_fold_insert(As, Y0, 0, Y1, S),
			zdd_choices(R, [A|As], Y1, Y, S)
		)
	;	zdd_choices(L, As, Y0, Y1, S),
		zdd_choices(R, [A|As], Y1, Y, S)
	).
%
zdd_fold_insert([], _, Y, Y, _).
zdd_fold_insert([A|As], Y0, U, Y, S):-
	zdd_insert(A, Y0, V, S),
	zdd_join(U, V, U0, S),
	zdd_fold_insert(As, Y0, U0, Y, S).

% ?- zdd((X<<pow([a,b]), zdd_fold_cons([1,2,3], X, Y), sets(Y, U))), writeln(U).
% ?- zdd((zdd_singleton([], X), zdd_fold_cons([1,2,3], X, Y), sets(Y, U))), writeln(U).
% ?- zdd((zdd_list_length(0, [a, b], X), psa(X))).
% ?- spy(zdd_list_length).
% ?- zdd((zdd_list_length([a, b], 1, X), psa(X))).
% ?- zdd((zdd_list_length([a, b, c], 3, X), psa(X))).
% ?- N=100, numlist(1, N, Ns), zdd((zdd_list_length(Ns, 3, X), card(X, C))).
% ?- time(( N=50, numlist(1, N, Ns), zdd((zdd_list_length(Ns, 3, X), card(X, C))))).
%@ % 251,894,470 inferences, 39.500 CPU in 41.508 seconds (95% CPU, 6377136 Lips)
%@ C = 125000.

zdd_list_length(L, N, X):- shift(zdd_list_length(L, N, X)).
%
zdd_list_length(_, 0, X, S):-!, zdd_singleton([], X, S).
zdd_list_length(L, N, X, S):- N0 is N-1,
	zdd_list_length(L, N0, X0, S),
	zdd_fold_cons(L, X0, 0, X, S).
%
zdd_fold_cons(L, X, Y):- shift(zdd_fold_cons(L, X, 0, Y)).
%
zdd_fold_cons([], _, Y, Y, _).
zdd_fold_cons([A|As], Y0, U, Y, S):-
	map_zdd(cons(A), Y0, V, S),
	zdd_join(U, V, U0, S),
	zdd_fold_cons(As, Y0, U0, Y, S).


% ?- zdd((X<<{[a], [a,b], [b]}, psa(X),
%	select_singleton(X, Y, U), sets(U, SU),
%   select_singleton(U, V, W), sets(W, SW),
%   select_singleton(W, L, M), sets(M, SM))).

% ?- zdd((X<<{[a,b]}, psa(X),
%	select_singleton(X, Y, U))).

% ?- numlist(1, 500, L), numlist(1, 250, M),  maplist(pred([I, J]:-
%	J is I+I), M, M0), 	zdd((X<<pow(M0), zdd_join(X, X, Y))).

% ?- numlist(1, 5, L), zdd((X<<pow(L), card(X, C))).

% ?- numlist(1, 500, L), numlist(1, 250, M),  maplist(pred([I, J]:-
%	J is I+I), M, M0), zdd((X<<pow(M0), Y<<pow(L),
%	zdd_join(X, Y, Z), card(X, Cx), card(Y, Cy),  card(Z, Cz))).


%%	zdd_join(+X, +Y, -Z[, +G]) is det.
% Z is the union of X and Y.

zdd_join(X, Y, Z):- shift(zdd_join(X, Y, Z)).
%
zdd_join(0, X, X, _):-!.
zdd_join(X, 0, X, _):-!.
zdd_join(X, X, X, _):-!.
zdd_join(1, X, Y, S):-!, zdd_join_1(X, Y, S).
zdd_join(X, 1, Y, S):-!, zdd_join_1(X, Y, S).
zdd_join(X, Y, Z, S):-
	(	X@<Y -> memo(zdd_join(X,Y)-Z, S)
	;	memo(zdd_join(Y, X)-Z, S)
	),
	(	nonvar(Z) -> true  %, write(.)
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (<)
		-> 	zdd_join(L, Y, U, S),
			cofact(Z, t(A, U, R), S)
		;	C = (=)
		->	zdd_join(L, L0, U, S),
			zdd_join(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_join(L0, X, U, S),
			cofact(Z, t(A0, U, R0), S)
		)
	).

% ?- zdd((X<<([a]+[b]+[]), psa(X))).
% ?- zdd((A<<{[a]},  psa(A), zdd_join_1(A, X), psa(X))).

%%	zdd_join_1(+X, -Y[, +G]) is det.
%  Y is the union of X and the singleton {{}}.
zdd_join_1(X, Y):- shift(zdd_join_1(X, Y)).
%
zdd_join_1(X, Y, S):-
	(	X<2  -> Y = 1
	; 	cofact(X, t(A, L, R), S),
		zdd_join_1(L, L0, S),
		cofact(Y, t(A, L0, R), S)
	).



% ?- zdd((X<<[a, b], Y<<[a, b], Z << (X*Y), sets(Z, S))).
% ?- zdd((X<<[a, a], Y<<[a, b], Z << (X*Y), sets(Z, S))).
% ?- zdd((X<<[a, b], Y<<[a, b], Z << (X&Y), sets(Z, S))).
% ?- zdd((X<<([a, b]+[c,d]), Y<<([x,y]+[u,v]), Z << (X&Y), sets(Z, S))).
% ?- trace, zdd(X=[b, a]).
% ?- zdd((X<<([a, b]+[c,d]), Y<<([x,y]+[u,v]), Z << (X**Y), sets(Z, S), card(Z, C))).
% ?- zdd((X<<[b, a], Y<<[a, b], Z <<(X+Y), sets(Z, S))).

%%	zdd_meet(+X, +Y, -Z[, +G]) is det.
%  Z is the intersection of X and Y.

zdd_meet(X, Y, Z):- shift(zdd_meet(X, Y, Z)).

%
zdd_meet(0, _, 0, _):-!.
zdd_meet(_, 0, 0, _):-!.
zdd_meet(1, X, Y, S):-!, zdd_meet_1(X, Y, S).
zdd_meet(X, 1, Y, S):-!, zdd_meet_1(X, Y, S).
zdd_meet(X, X, X, _):-!.
zdd_meet(X, Y, Z, S):-
	(	X @< Y -> memo(zdd_meet(X,Y)-Z, S)
	;	memo(zdd_meet(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (<)
		-> zdd_meet(L, Y, Z, S)
		;	C = (=)
		->	zdd_meet(L, L0, U, S),
			zdd_meet(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_meet(R0, X, Z, S)
		)
	).

%%	zdd_meet_1(+X, -Y[, +G]) is det.
% Y is the intersection of X and the singleton of the emptyset {{}}.
%
zdd_meet_1(X, Y):- shift(zdd_meet_1(X, Y)).
%
zdd_meet_1(X, X, _):- X<2, !.
zdd_meet_1(X, Y, S):- cofact(X, t(_,L,_), S),
	zdd_meet_1(L, Y, S).

%%	zdd_subtr(+X, +Y, -Z[, +G]) is det.
% Z is the subtraction of Y from X: Z = Y \ Z.

% ?- zdd((X<<{[a,b],[b,c]}, Y<<{[b,c]}, psa(Y), zdd_subtr(X, Y, Z), psa(Z))).
zdd_subtr(X, Y, Z):-shift(zdd_subtr(X, Y, Z)).
%
zdd_subtr(X, X, 0, _):-!.
zdd_subtr(0, _, 0, _):-!.
zdd_subtr(X, 0, X, _):-!.
zdd_subtr(X, 1, Y, S):-!, zdd_subtr_1(X, Y, S).
zdd_subtr(1, X, Y, S):-!,
	(	zdd_has_1(X, S)
	->	Y = 0
	;	Y = 1
	).
zdd_subtr(X, Y, Z, S):- memo(zdd_subtr(X,Y)-Z, S),
	(	nonvar(Z) -> true
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (<)
		->	zdd_subtr(L, Y, U, S),
			cofact(Z, t(A, U, R), S)
		;	C = (=)
		->	zdd_subtr(L, L0, U, S),
			zdd_subtr(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_subtr(L, L0, U, S),
			cofact(Y, t(A, U, R), S)
		)
	).

%%	zdd_subtr_1(+X, -Y[, +G]) is det.
% Y is the subtraction of the set {{}} from X: Y = X \ {{}}.

% ?- zdd((X<<([]+[a]), zdd_subtr_1(X, Y), psa(Y))).
% ?- zdd((X<<{[a], [a,b,c]}, zdd_subtr_1(X, Y), psa(Y))).
% ?- zdd((X<<{[a], [a,b,c], []}, zdd_subtr_1(X, Y), psa(Y))).
zdd_subtr_1(X, Y):-shift(zdd_subtr_1(X, Y)).
%
zdd_subtr_1(X, 0, _):- X < 2, !. % empty family or singleton of the empty.
zdd_subtr_1(X, Y, S):- cofact(X, t(A, L, R), S),
	zdd_subtr_1(L, L0, S),
	cofact(Y, t(A, L0, R), S).

		/***************
		*     merge    *
		***************/

%%	zdd_merge(X, Y, Z[, G]) is det.
%	Z is the merge of X and Y, that is,
%   Z = { U | U is the union of A in X and B in Y}.
%	Remark. The merge is associative and commutative, but not idempotent.

% ?- zdd(( X<<{[a], [b]}, Y<<{[c]}, zdd_merge(X, Y, Z), psa(Z))).
zdd_merge(X, Y, Z)	:- shift(zdd_merge(X, Y, Z)).

%
zdd_merge(0, _, 0, _):-!.
zdd_merge(_, 0, 0, _):-!.
zdd_merge(1, X, X, _):-!.
zdd_merge(X, 1, X, _):-!.
zdd_merge(X, Y, Z, S):-
	(	X@<Y -> memo(zdd_merge(X,Y)-Z, S)
	;	memo(zdd_merge(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(Y, t(A, L, R), S),
		zdd_merge(X, R, U, S),
		zdd_merge(X, L, V, S),
		zdd_insert(A, U, U0, S),
		zdd_join(U0, V, Z, S)
	).

		/****************
		*     divide    *
		****************/

%%	zdd_divide(X, Y, Z[, G]) is det.
%	Z is the quotient of X divided by Y, that is

% ?- spy(zdd_divide).
% ?- zdd(( X<< 0, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<< 0, Y<< 0, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<1, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<1, Y<<1, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [b, c]}, Y<<{[a],[b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a, b, c]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b], [a, c]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[c]}, Y<<{[a, c]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a, b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b]}, Y<<{[a, b, c]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b,c]}, Y<<{[a,b], [c]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a], [b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[b]}, zdd_divide(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[b]}, zdd_divide(X, Y, Z), psa(Z))).

zdd_divide(X, Y, Z)	:- shift(zdd_divide(X, Y, Z)).

%
zdd_divide(0, _, 0, _):-!.
zdd_divide(_, 0, 0, _):-!.
zdd_divide(1, X, Y, S):-!,
	(	zdd_has_1(X, S)
	->	Y = 1
	;	Y = 0
	).
zdd_divide(X, 1, X, _):-!.
zdd_divide(X, Y, Z, S):-
	memo(zdd_divide(X, Y)- Z, S),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (=)
		->	zdd_divide(R, R0, Z0, S),
			zdd_divide(R, L0, Z1, S),
			zdd_insert(A, Z1, Z2, S),
			zdd_divide(L, L0, Z3, S),
			zdd_join(Z0, Z2, Z4, S),
			zdd_join(Z3, Z4, Z, S)
		;   C = (<)
		->	zdd_divide(R, Y, Z0, S),
			zdd_insert(A, Z0, Z1, S),
			zdd_divide(L, Y, Z2, S),
			zdd_join(Z1, Z2, Z, S)
		; 	zdd_divide(X, L0, Z, S)
		)
	).

% ?- zdd(( X<<{[a, b], [a, c]}, zdd_div_by_list(X, [b], Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [c, d]}, zdd_div_by_list(X, [b], Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, b, d]}, zdd_div_by_list(X, [b], Z), psa(Z))).
%
zdd_div_by_list(X, Y, Z):- shift(zdd_div_by_list(X, Y, Z)).
%
zdd_div_by_list(X, As, Y, S):- zdd_append(As, 1, D, S),
	zdd_divide(X, D, Y, S).



		/*****************
		*     residue    *
		*****************/

% ?- zdd(( X<< 0, Y<<{[a]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<< 0, Y<< 0, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a, c]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [b, c]}, Y<<{[a, c]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [b, c]}, Y<<{[a, c]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<1, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<1, Y<<1, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a], [b]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b]}, zdd_residue(X, Y, Z), psa(Z))).



% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b], [a, c]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a, c]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a], [b]}, Y<<{[b]}, zdd_residue(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, c], [b]}, Y<<{[b]}, zdd_residue(X, Y, Z), psa(Z))).
%@ Z = 3.

%@ true.
%@ zdd.pl qcompiled.

zdd_residue(X, Y, Z)	:- shift(zdd_residue(X, Y, Z)).

%
zdd_residue(0, _, 0, _):-!.
zdd_residue(_, 1, 0, _):-!.
zdd_residue(X, 0, X, _):-!.
zdd_residue(X, X, 0, _):-!.
zdd_residue(1, _, 1, _):-!.
zdd_residue(X, Y, Z, S):-
	memo(zdd_residue(X, Y)- Z, S),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (=)
		->	zdd_residue(R, R0, Z0, S),
			zdd_insert(A, Z0, Z1, S),
			zdd_join(L, Z1, Z2, S),
			zdd_residue(Z2, L0, Z, S)
		;   C = (<)
		->	zdd_residue(R, Y, R1, S),
			zdd_insert(A, R1, Z0, S),
			zdd_residue(L, Y, Z1, S),
			zdd_join(Z0, Z1, Z, S)
		; 	zdd_residue(X, L0, Z, S)
		)
	).

zdd_res_by_list(X, Y, Z):- shift(zdd_res_by_list(X, Y, Z)).
%
zdd_res_by_list(X, As, Y, S):- zdd_append(As, 1, D, S),
	zdd_residue(X, D, Y, S).


		/*********************
		*     meta on ZDD    *
		*********************/

% ?- zdd((X<< {[1],[2]}, map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- zdd((X<<pow([1,2,4,5,6]), map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=2, numlist(1, N, L), zdd((X<<pow(L), map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=10, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=10, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:- B is A + 1), X, Y), card(Y, C))), C=:= 2^N.
% ?- N=1000, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:- B is A + 1), X, Y), card(Y, C))), C=:= 2^N.

%% map_zdd(+F, +X, -Y[, +G]) is det.
%  Y is the set of all F_S with S running in X,
%  where f_S = { F(a) | a in S}.
%
%  map_zdd(F, X, Y)  works roughly like
%  maplist(maplist(F), X, Y)) with a list X of lists.

:- meta_predicate map_zdd(2,?,?).
map_zdd(F, X, Y):- shift(map_zdd(F, X, Y)).
:- meta_predicate map_zdd(2,?,?,?).
map_zdd(F, X, Y, S):- zdd_shift(map_zdd(F, X, Y, S)).

:- meta_predicate map_zdd(2,?,?,?,?).
map_zdd(_, X, X, _, _):- X<2, !.
map_zdd(F, X, Y, S, M):- memo(map(X,F)-Y, M),
	(	nonvar(Y)-> true
	;	cofact(X, t(A, L, R), S),
		map_zdd(F, L, L0, S, M),
		map_zdd(F, R, R0, S, M),
		call(F, A, B),
		zdd_insert(B, R0, R1, S),
		zdd_join(L0, R1, Y, S)
	).

% ?- zdd((X<<pow([a,b]), map_zdd0(=, X,Y))).
% ?- zdd((X<<pow([a,b]), map_zdd0(==, X,Y), psa(X))).

%% map_zdd0(+F, +X, -Y[, +G]) is det.
%  A variant of map_zdd/3,4. Differece is that map_zdd0 uses 'memo'.

:- meta_predicate map_zdd0(2,?,?).
map_zdd0(F, X, Y):- shift(map_zdd0(F, X, Y)).
:- meta_predicate map_zdd0(2,?,?,?).
map_zdd0(F, X, Y, S):-zdd_shift(map_zdd0(F, X, Y, S)).

:- meta_predicate map_zdd(2,?,?,?,?).
map_zdd0(_, X, X, _, _):- X<2, !.
map_zdd0(F, X, Y, S, M):- memo(map(X,F)-Y, M),
	(	nonvar(Y)-> true
	;	cofact(X, t(A, L, R), S),
		map_zdd0(F, L, L0, S, M),
		(	call(F, A, B)
		-> 	map_zdd0(F, R, R0, S, M),
			zdd_insert(B, R0, R1, S)
		;	R1 = 0
		),
		zdd_join(L0, R1, Y, S)
	).



% ?- zdd((memo(a-1), setmemo(a-2), memo(a-V))).
% ?- zdd((X<<pow([a,b,c,d]), prz(X))).
% ?- zdd((X<<pow([a,b,c]), psa(X))).
% ?- zdd((prz(1))).
% ?- zdd((prz(0))).

prz(X):- print_set_at(X).
pzm(X):- print_zdd_matrix(X).

%
psa(X):- print_set_at(X).
psa(X, S):- print_set_at(X, S).
%
print_set_at(X):- shift(print_set_at(X)).
print_set_at(0, _):-!, nl, writeln(0).
print_set_at(1, _):-!, nl, writeln([]).
print_set_at(X, S):- format("\nzdd ~w:", [X]),
	forall(set_at(P, X, S), format("\n\t~w", [P])).
%

% ?- zdd((X<<(set([1-2, 2-3, 3-4])+ set([1-2,2-3])), set_at(U, X))).
% ?- zdd((X<<pow([1,2]), set_at(U, X))).
set_at(P, X):- shift(set_at(P, X)).
%
set_at([], 1, _):-!.
set_at(P, X, S):-
	cofact(X, t(A, L, R), S),
	(	set_at(P, L, S)
	;	set_at(As, R, S),
		P=[A|As]
	).
%
sets(X, P):- shift(sets(X, P)).

sets(X, Y, S):- zdd_sets(X, Y, S).
%

zdd_sets(1, [[]], _):-!.
zdd_sets(0, [], _):-!.
zdd_sets(X, P, S):-
	cofact(X, t(A,L,R), S),
	zdd_sets(L, Y, S),
	zdd_sets(R, Z, S),
	maplist(cons(A), Z, AZ),
	maplist(sort, AZ, AZ0),
	sort(AZ0, AZ1),
	union(Y, AZ1, P).
%
ppoly(X):- @(ppoly(X)).

ppoly(X, S):- poly_term(X, Poly, S),
	writeln(Poly).

%@ true.
%@ zdd.pl qcompiled.

% ?- zdd((X<<pow([a,b]), poly_term(X, P))).
%
poly_term(X, Poly):- @(poly_term(X, Poly)).
%
poly_term(X, Poly, S):-zdd_sets(X, Sets, S),
	maplist(sort, Sets, Sets0),
	sort(Sets0, Sets1),
	poly_term0(Sets1, Poly).

% ?- poly_term0([], Y).
% ?- poly_term0([[]], Y).
% ?- poly_term0([[a], [b]], Y).
% ?- poly_term0([[a*b], [c]], Y).

poly_term0(X, Y):- 	maplist(mul_term, X, X0),
	sum_term(X0, Y).
%
mul_term([], [[]]):-!.
mul_term([X], X):-!.
mul_term([X, Y|Xs], Z):-mul_term([X*Y|Xs], Z).
%
sum_term([], []):-!.
sum_term([X], X):-!.
sum_term([X, Y|Xs], Z):-sum_term([X+Y|Xs], Z).

%%	famset(+X, -S) is det.
% Unify a list S with the family of sets that is represented
% in a ZDD X.

famset(X, P):- shift(zdd_sets(X, P)).

%%	zdd_path is det.
%   Print all paths in the ZDD.
% ?- zdd((X<< pow([a,b,c]), zdd_path(X))).
% ?- zdd((prepare_sat(4), memo(filter(4)-X), zdd_path(X))).
zdd_path(I)	:- shift(zdd_path(I)).
%
zdd_path(I, S):- get_vector(Vec, S),
	(	zdd_path(I, P, [], Vec),
		writeln(P),
		fail
	;	true
	).
%
zdd_path(true, X, X, _):-!.
zdd_path(I, X, Y, Vec):- integer(I),
	arg(I, Vec, if(A, L, R)),
	(	X = [A|X0],
		zdd_path(L, X0, Y, Vec)
	;	zdd_path(R, X, Y, Vec)
	).

% ?- zdd((X << [a, b, c], zdd_rand_path(X))).
% ?- zdd((X << pow([a, b, c]), zdd_rand_path(X))).
% ?- zdd((X << {[a], [b], [c]}, zdd_rand_path(X))).
% ?- zdd((X << pow([a, b, c, d, e, f, g]), zdd_rand_path(X))).

%%	zdd_rand_path(+I) is det.
% Print a path at radom starting at a node I in the ZDD.
zdd_rand_path(I):- shift(zdd_rand_path(I)).
%
zdd_rand_path(I, S):- zdd_rand_path(I, P, [], S), writeln(P).
%
zdd_rand_path(1, X, X, _):-!.
zdd_rand_path(I, X, Y, S):- I>1,
	cofact(I, t(A, L, R), S),
	(	L == 0
	->	X = [A|X0],
		zdd_rand_path(R, X0, Y, S)
	;	random(2) =:= 0
	-> 	X = [A|X0],
		zdd_rand_path(R, X0, Y, S)
	; 	zdd_rand_path(L, X, Y, S)
	).

% ?- zdd((X<<[a], zdd_nodes(X, N))).
% ?- N=10, numlist(1, N, Ns), zdd((X<<pow(Ns), zdd_nodes(X, U))).
% ?- N=10000, numlist(1, N, Ns), time(zdd((X<<pow(Ns), zdd_nodes(X, U)))), length(U, Len).
%@ % 350,941,340 inferences, 18.703 CPU in 19.647 seconds (95% CPU, 18763903 Lips)

%%	zdd_nodes(+X, -Y[, +G]) is det.
%  Y is the list of nodes of a ZDD X, which excludes
%  1/0.  Memo is used.

zdd_nodes(X, P):- shift(zdd_nodes(X, P)).
%
zdd_nodes(X, P, S):- zdd_shift(zdd_nodes(X, P, S)).

%
zdd_nodes(X, [], _, _):- X<2, !.
zdd_nodes(X, P, S, M):- memo(nodes(X)-P, M),
	(	nonvar(P) -> true
	;	cofact(X, t(_, L, R), S),
		zdd_nodes(L, Nl, S, M),
		zdd_nodes(R, Nr, S, M),
		ord_union([Nl, [X], Nr], P)
	).

% ?- zdd((zdd_atoms(1, S))).
% ?- zdd((family([[a,b],[b,c]], X), zdd_atoms(X, S))).
% ?- N = 10000, time(zdd(({numlist(1, N, L)}, X<<power(L), zdd_atoms(X, S)))).
%@ % 200,887,717 inferences, 9.620 CPU in 10.312 seconds (93% CPU, 20883111 Lips)

%%	zdd_atoms(+X, -Y]) is det.
%  Y is the set of  ZDD atoms appearing on ZDD X.
%
zdd_atoms(X, P):- shift(zdd_atoms(X, P)).

zdd_atoms(X, P, S):-zdd_shift(zdd_atoms(X, P, S)).

%
zdd_atoms(X, [], _, _):- X<2, !.
zdd_atoms(X, P, S, M):- memo(atoms(X)-P, M),
	(	nonvar(P) -> true
	; 	cofact(X, t(A, L, R), S),
		zdd_atoms(L, Al, S, M),
		zdd_atoms(R, Ar, S, M),
		ord_union(Al,Ar, P0),
		P=[A|P0]
	).

% ?- collect(pred([a-_]), [a-b, c-d, a-e], L).

% ?- delete(pred([a-_]), [a-b, c-d, a-e], L).
:- meta_predicate delete(1, ?, ?).
delete(X, Y, Z):- delete(X, Y, Z, []).

:- meta_predicate delete(1, ?, ?, ?).
delete(_, [], L, L).
delete(Pred, [X|Xs], L, L0):-
	(	call(Pred, X)
	->	delete(Pred, Xs, L, L0)
	;	L = [X|L1],
		delete(Pred, Xs, L1, L0)
	).

% ?- zdd((zdd_power([a,b], R), sets(R, U))).
% ?- zdd((zdd_power(charlist(a-c), R), sets(R, U))).
% ?- zdd((R<< pow(charlist(a-d)), card(R, C))).
zdd_power(X, Y):- shift(zdd_power(X, Y)).
%
zdd_power(X, Y, S):- is_list(X), !,
	sort(X, X0),
	zdd_ord_power(X0, Y, S).
%
zdd_power(X, Y, S):- callable(X), !, call(X, X0),
	zdd_power(X0, Y, S).

% ?- concat_atom([a, b], C).
% ?- atom_concat(a, b, C).
% ?- atomlist(ax(1+1, 3+1), As).
% ?- atomlist(ax(1, 3), As).
% ?- atomlist(ax(1..3), As).

atomlist(X, As):- X=..[A|B],
	( B = [I..J];  B = [I, J] ), !,
	I0 is I,
	J0 is J,
	findall(Y, (	between(I0, J0, K),
					atom_concat(A, K, Y)
			   ),
			As).

% ?- charlist(a-c, X).
charlist(A-B, I):-
	findall(X, (	char_type(X, alnum),
				    X @>= A,
					X @=< B
			   ),
			I).
% ?- charlist(a, c, X).
charlist(A, B, I):- charlist(A-B, I).

%
zdd_ord_power([], 1, _).
zdd_ord_power([X|Xs], P, S):- zdd_ord_power(Xs, Q, S),
	cofact(P, t(X, Q, Q), S).

% ?- zdd((ltr_sort([-a, a, -b, b], L), ltr_power(L, P), card(P, C))).
% ?- zdd((X<<ltr_pow([a,b]), S<<sets(X), C<<card(X))).
% ?- zdd((L<<ltr_pow([-a, -b]), C<<card(L))).
% ?- zdd((L<<ltr_pow([-a, a, -b, b]), U<<sets(L))).
% ?- zdd((L<<ltr_pow([-a, a, -b, b]))).
% ?- zdd((L<<ltr_pow([-a, b]), psa(L))).
% ?- zdd((L<<ltr_pow([]), psa(L))).

% ?- zdd((X<<pow([a,b]), powerzdd(X, Y), powerzdd(Y, Z), card(Z, C))).
%%	powerzdd(+X, -Y) is det.
% Unify Y with the powerZDD of a ZDD X, i.e.,
% the powerset of nodes of a ZDD X.
% powerzdd(X, Y):- powerzdd(X, Y, rozdd).
%
powerzdd(X, Y):- shift(powerzdd(X, Y)).

powerzdd(X, Y, S):- cofact(False, if(p(false), true, true), S),
	cofact(A, if(p(true), False, False), S),
	powerzdd(X, A, Y, S).

%
powerzdd(true, Pow, Pow, _):-!.
powerzdd(false, Pow, Pow, _):-!.
powrzdd(I, Pow, P, S):- memo(pow(I)-P, S),
	(	nonvar(P) -> true
	;	cofact(I, if(_, L, R), S),
		powerzdd(L, Pow, Pow0, S),
		powerzdd(R, Pow0, Q, S),
		cofact(P, if(p(I), Q, Q), S)
	).

% ?- zdd((dnf(A+B, X), get_key(varnum, N), extern(X, Y), X==Y)).
% ?- zdd((dnf(X->Y, U), sets(U, V), extern(V, A))).

%%	extern(+X, -Y) is det.
% Convert an internal form X into an external one
% in order to unify Y with the result.
extern(X, Y):-shift(extern(X,Y)).

extern(X, Y, S):- get_key(varzip, Zip, S),
	subst_opp(X, Y, Zip).

% ?- count_atomic_prop(a+b, 0-C, atomic_prop).
% ?- count_atomic_prop(a+b+c+a+A, 0-C, atomic_prop).

:- meta_predicate count_atomic_prop(?, ?, 1).
count_atomic_prop(X, Y, Z):- zdd_shift(count_atomic_prop(X, Y, Z)).
%
count_atomic_prop(X, C-C, _, _):- var(X), !.
count_atomic_prop(X, C-D, Pred, S):- call(Pred, X), !,
	memo(atomic_prop(X)-B, S),
	(	nonvar(B)
	->	D = C
	;	D is C+1,
		B = true
	).
count_atomic_prop([], C-C, _, _):-!.
count_atomic_prop([X|Y], C-D, Pred, S):-!,
	count_atomic_prop(X, C-C0, Pred, S),
	count_atomic_prop(Y, C0-D, Pred, S).
count_atomic_prop(X, P, Pred, S):- compound(X), !,
	X =..[_|As],
	count_atomic_prop(As, P, Pred, S).
count_atomic_prop(_, C-C, _, _).

% ?- term_atoms(a+b=c, L).
% ?- term_atoms(a+b, L).
% ?- term_atoms(f(a+a), L).

%% term_atoms(+X, -L) is det.
% Unify L with a sorted list of atomic boolean
% subterms of X.

term_atoms(X, L):- term_atoms(X, L0, []),
	sort(L0, L).

%
term_atoms(X, L, L):- var(X), !.
term_atoms(X, [X|L], L):- atom(X), !.
term_atoms([X|Y], L, L0):-!,
	term_atoms(X, L, L1),
	term_atoms(Y, L1, L0).
term_atoms(X, L, L0):- compound(X), !,
	X =..[_|As],
	term_atoms(As, L, L0).
term_atoms(_, L, L).

% ?- subst_opp(f(a,b,c,f(a)), Y, [b-a]).
% ?- subst_opp(a, X, [b-a]).
% ?- subst_opp([a], X, [b-a]).

subst_opp(X, Y, L):-  memberchk(Y-X, L).
subst_opp([X|Y], [X0|Y0], L):-!,
	subst_opp(X, X0, L),
	subst_opp(Y, Y0, L).
subst_opp(X, Y, L):- compound(X), !,
	X =..[F|As],
	subst_opp(As, Bs, L),
	Y =..[F|Bs].
subst_opp(X, X, _).

% ?- subst_term(f(a), R, [a-b]).
% ?- subst_term(g(f(a),g(a)), R, [f(a)-b]).
% ?- subst_term(f(a), R, [a-X]).
% ?- subst_term(g(f(a),g(f(a))), R, [f(a)-X]).
% ?- subst_term(g(f(a),g(f(a))), R, [f(a)-X]).
% ?- subst_term(g(f(X),g(f(Y))), R, [f(X)-a, f(Y)-b]).

%%	subst_term(+X, -Y, +S) is det.
% Apply an assoc list X as substitution S to X to unify Y with
% the result.

subst_term(X, Y, S):-  member(X0-Y, S), X0==X, !.
subst_term(A, B, S):-  compound(A), !,
	(	A = [X|Y]
	->	B = [X0|Y0],
		subst_term(X, X0, S),
		subst_term(Y, Y0, S)
	; 	functor(A, F, N),
		functor(B, F, N),
		subst_term(N, A, B, S)
	).
subst_term(X, X, _).
%
subst_term(0, _, _, _):-!.
subst_term(I, X, Y, S):-
	arg(I, X, A),
	arg(I, Y, B),
	subst_term(A, B, S),
	J is I - 1,
	subst_term(J, X, Y, S).

% ?- all_instances([a,b], [0,1], a+b=b, R).
% ?- all_instances([x,y], [0,1], x+y=x, R).
%%	all_instances(+As, +Vs, +X, -Y) is det.
% Unify Y with the list of possible variations P of X,
% where P is a variation of X if for each A in As with some
% V in Vs which depends on A,  P is obtained from X by replacing
% all occurrences of A appearing in X with V for A in As.

all_instances(Ks, Ts, X, Y):- 	all_maps(Ks, Ts, Fs),
	apply_maps(Fs, X, Y, []).
%
apply_maps([], _, Y, Y).
apply_maps([F|Fs], X, [X0|Y], Z):-
	subst_term(X, X0, F),
	apply_maps(Fs, X, Y, Z).

% ?- mod_table(3, [a,b], a+b, Table), maplist(writeln, Table).
% ?- mod_table(3, [a,b,c], a+b+c, Table), maplist(writeln, Table).
% ?- mod_table(3, [a,b], a*b, Table), maplist(writeln, Table).
% ?- mod_table(2, [a], a*a, Table), maplist(writeln, Table).
%%	mod_table(+M, +X, +E, -T) is det.
% Unify T with a set of form E' = V where
% E' is a possible ground instance  of an integer expression E
% and V is the value of E' with modulo M, where X is a set of
% parameters appearing in E.

mod_table(M, Xs, Exp, Table):-	M0 is M-1,
	numlist(0, M0, V),
	all_instances(Xs, V, Exp, Exps),
	!,
	maplist(mod_arith(M), Exps, Table).
%
mod_arith(M, Exp, Exp=V):- V is Exp mod M.

%
unify_all([]).
unify_all([_]).
unify_all([X,Y|Z]):- zdd:(X=Y), unify_all([Y|Z]).

% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(X, Y))). % false
% ?- zdd((X<< {[a,b], [a,c], [a]},  Y<<{[a,b], [a]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b],  [a,c]}, Y<<{[a,b], [a]}, zdd_subfamily(Y, X))). % false
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(X, Y))).
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(Y, X))). % false

%%	zdd_subfamily(+X, +Y) is det.
% True if a ZDD X is a subfamily of a ZDD Y.

zdd_subfamily(X, Y):- shift(zdd_subfamily(X, Y)).
%
zdd_subfamily(X, Y, S):-zdd_shift(zdd_subfamily(X, Y, S)).

%
zdd_subfamily(X, X, _, _):-!.
zdd_subfamily(0, _, _, _):-!.
zdd_subfamily(_, 0, _, _):-!, fail.
zdd_subfamily(1, X, S, _):-!, zdd_has_1(X, S).
zdd_subfamily(X, X0, S, M):- memo(subfamily(X, X0)-B, M),
	(	nonvar(B) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(X0, t(A0, L0, R0), S),
		compare(C, A, A0),
		(	C = (=)
		->  zdd_subfamily(L, L0, S, M),
			zdd_subfamily(R, R0, S, M),
			B = true
		;	C = (>),
			zdd_subfamily(X, L0, S, M)
		)
	).

%%	zdd_appear(+A, +X) is det.
%	True if A is a member of a member of X.
% ?- zdd((X <<([a,b]+[a,c]), zdd_appear(e, X))). % false
% ?- zdd((X <<([a,b]+[a,c]), psa(X), zdd_appear(c, X))).
% ?- numlist(1, 10000, Ns), zdd((X<<pow(Ns), zdd_appear(10000, X))).

zdd_appear(A, X):- shift(zdd_appear(A, X)).
zdd_appear(A, X, S):- zdd_shift(zdd_appear(A, X, S)).

%
zdd_appear(A, X, S, M):- X > 1,
	memo(visited(X)-B, M),
	var(B),
	cofact(X, t(U,L,R), S),
	compare(C, A, U),
	(	C = (=)
	->  true
	;	C = (>),
	 	(	zdd_appear(A, L, S, M) -> true
		;	L < 2
		->  zdd_appear(A, R, S, M)
		; 	memo(visited(L)-true, M),  %  for earlier fail.
			zdd_appear(A, R, S, M)
		)
	).

%%	zdd_count_nodes(+X, -C) is det.
% C is unified with the number of intermediate nodes
% of a ZDD X.

% ?- N=10, time((numlist(1, N, L), zdd((X=pow(L), zdd_count_nodes(X, C))))).
zdd_count_nodes(X, C):- shift(zdd_count_nodes(X, C)).

zdd_count_nodes(X, C, S):- intermediate_nodes(X, [], L, S),
	length(L, C).
%
intermediate_nodes(I, X, Y, S):-integer(I), !,
	( memberchk(I, X) -> Z = X; Z = [I|X] ),
	cofact(I, if(_, L, R), S),
	intermediate_nodes(L, Z, Z0, S),
	intermediate_nodes(R, Z0, Y, S).
intermediate_nodes(_, X, X, _).

%
map_singleton([], [], _).
map_singleton([X|Xs], [Y|Ys], S):- zdd_singleton(X, Y, S),
	map_singleton(Xs, Ys, S).

%%	zdd_singleton(+X, -P, +G) is det.
% With an ROZDD bound to G, unify P with a sigleton ZDD for X.
% ?- listing(zdd_singleton).

zdd_singleton(X, P):- shift(zdd_singleton(X, P)).
%
zdd_singleton(X, P, S):- cofact(P, t(X, 0, 1), S).

%%	zdd_meet_t(+X, -Y, +G) is det.
% With an ROZDD bound to G, unify Y with the intersection of a ZDD X and true (= {{}})
%  as a family of sets.

zdd_meet_t(X, Y):- shift(zdd_meet_t_(X, Y)).
%
zdd_meet_t_(false, false, _):-!.
zdd_meet_t_(true, true, _):-!.
zdd_meet_t_(I, K, S):- cofact(I, if(_, _, V), S),
	zdd_meet_t_(V, K, S).

%%	zdd_meet_t(+X, +Y, -Z, +G) is det.
% With an ROZDD bound to G, unify Z with a ZDD such that the intersection
% of the truth set (the powerset) for the variables
% {X, X+1, ..., } and Y as a family of sets.

zdd_meet_t(X, Y, Z):- shift(zdd_meet_t(X, Y, Z)).
%
zdd_meet_t(_, false, false, _):-!.
zdd_meet_t(_, true, true, _):-!.
zdd_meet_t(H, I, K, S):- cofact(I, if(J, _, V), S),
	( H @< J ->  K = I
	; zdd_meet_t(H, V, K, S)
	).

%%	zdd_meet_p(+H, +X, -Y, +G) is det.
% Unify Y with the intersection of the
% powerset of {H, H+1, ...} without the empty set
% and X, as a familty of sets.

zdd_meet_p(X, Y, Z):- shift(zdd_meet_p(X, Y, Z)).
%
zdd_meet_p(_, true, false, _).		% assuming H is not max.
zdd_meet_p(_, false, false, _).
zdd_meet_p(H, X, Y, S):- cofact(X, if(I, L, R), S),
	(	H @=< I -> zdd_rm_1(R, R0, S),
				 cofact(Y, if(I, L, R0), S)
	;	zdd_meet_p(H, R, Y, S)
	).


%%	zdd_has_1(+X, +G) is det.
% With an ROZDD bound to G, true if a ZDD X has the empty set.

zdd_has_1(X):- shift(zdd_has_1(X)).
%
zdd_has_1(1, _):-!.
zdd_has_1(X, S):- X>1,
   cofact(X, t(_, L, _), S),
   zdd_has_1(L, S).


% ?- zdd((X<< set([a,b]), Y<<set([]), zdd_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<< set([a,c]), Y<< set([b, d, e]), zdd_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<< set([a,b,c]), Y<< set([d,e,f]), psa(X), psa(Y), zdd_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<set([a,d,c]), Y<<set([d,b,e]), zdd_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<set([c, a, d,c]), Y<<set([d,b,e, d]), zdd_merge_line(X, Y, Z), psa(Z))).

% special version of zdd_merge, only for linear ZDD.
zdd_merge_line(X, Y, Z):- shift(zdd_merge_line(X, Y, Z)).

zdd_merge_line(X, X, X, _):-!.   % idempotent (only for line version).
zdd_merge_line(0, _, 0, _):-!.
zdd_merge_line(_, 0, 0, _):-!.
zdd_merge_line(1, X, X, _):-!.
zdd_merge_line(X, 1, X, _):-!.
zdd_merge_line(X, Y, Z, S):- cofact(X, t(A, _, R), S),
	cofact(Y, t(A0, _, R0), S),
	compare(C, A, A0),
	(	C = (<)
	->	zdd_merge_line(R, A0, R0, Z0, S),
		cofact(Z, t(A, 0, Z0), S)
	;	C = (=)
	->	zdd_merge_line(R, R0, Z0, S),
		cofact(Z, t(A, 0, Z0), S)
	;	zdd_merge_line(R0, A, R, Z0, S),
		cofact(Z, t(A0, 0, Z0), S)
	).

%
zdd_merge_line(0, _, _, 0, _):-!.
zdd_merge_line(1, A, R, Z, S):-!, cofact(Z, t(A, 0, R), S).
zdd_merge_line(X, A, R, Z, S):- cofact(X, t(B, _, R0), S),
	compare(C, A, B),
	(	C = (<)
	->	zdd_merge_line(R, B, R0, Z0, S),
		cofact(Z, t(A, 0, Z0), S)
	;	C = (=)
	->	zdd_merge_line(R, R0, Z0, S),
		cofact(Z, t(A, 0, Z0), S)
	;	zdd_merge_line(R0, A, R, Z0, S),
		cofact(Z, t(B, 0, Z0), S)
	).


% ?- zdd((zdd_line([-a, b], X))).
% ?- zdd((X<<zdd(zdd_line([-a, b])), psa(X))).

% build line zdd for an unsorted list as it it.
zdd_line(X, Y):- shift(zdd_line(X, Y)).
%
zdd_line([], 1, _).
zdd_line([A|As], X, S):- zdd_line(As, X0, S),
	cofact(X, t(A, 0, X0), S).



% ?- zdd((U<<[x, y, z], P<<[a, b], zdd_subst(a, U, P, Q), psa(Q))).
% ?- zdd((U<<pow([a,b]), P<<pow([x,y]), zdd_subst(x, U, P, Q), psa(Q))).
zdd_subst(X, U, P, Q):- shift(zdd_subst(X, U, P, Q, [])).
%
zdd_subst(_, _, false, false, _, _):-!.
zdd_subst(_, _, true, true, _, _):-!.
zdd_subst(X, U, P, Q, As, S):-
	cofact(P, if(Y, L, R), S),
	(	X == Y
	->	zdd_merge(U, L, UL, S),
		zdd_join(UL, R, ULR, S),
		zdd_inserts(As, ULR, Q, S)
	;	zdd_subst(X, U, L, L0, [Y|As], S),
		zdd_subst(X, U, R, R0, As, S),
		zdd_join(L0, R0, Q, S)
	).
		/***********************
		*    product on zdd    *
		***********************/


% ?- zdd((X<<[a,b], Y<<[c,d], Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a], Y<<([c]+[e]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a], Y<<[c], Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X << 0, Y<<[c], Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a], Y<<0, Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a,b], Y<<([c]+[e]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a,b], Y<<([c,d]+[e,f]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<{[1, 2],[3,4]},  Y<<([c,d]+[e,f]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<[a], Y<<([c,d]+[e,f]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- zdd((X<<pow([a,b]), Y<<pow([c,d]), Z<<(X**Y), card(Z, C), sets(Z, S))).
% ?- numlist(1, 16, R), time(zdd((X<<pow(R), Y<<pow(R), Z<<(X**Y), card(Z, C))))
% ?- call_with_time_limit(100, (numlist(1, 10, R), time(zdd((X<<pow(R), Y<<pow(R), Z<<(X**Y), card(Z, C)))))).


% ?- call_with_time_limit(120, (numlist(1, 16, R),
%	time(zdd((X<<pow(R), Z<<(X**X), card(Z, C)))))).
%@ % 134,889,120 inferences, 9.751 CPU in 9.754 seconds (100% CPU, 13832978 Lips)

%@ X = 17,
%@ Z = 4369,
%@ C = 1048561.

zdd_product(X, Y, Z):- shift(zdd_product(X, Y, Z)).
%
zdd_product(X, Y, 0, _):- (X<2; Y<2), !.
zdd_product(X, Y, Z, S):- memo(prod(X, Y)-Z, S),
	(	nonvar(Z) -> true
	;	fold_product(X, Y, [], Lpath, 0, Z0, S),
		attach_handle(X, Lpath, Z1, S),
		zdd_join(Z0, Z1, Z, S)
	).
%
fold_product(_, X, P, P, Z, Z, _):- X<2, !.
fold_product(X, Y, P, Q, U, V, S):- cofact(Y, t(B, L, R), S),
	zdd_product(X, L, L0, S),
	zdd_join(L0, U, U0, S),
	fold_product(X, R, [B|P], Q, U0, V, S).
%
attach_handle(X, _, X, _):- X<2, !.
attach_handle(X, P, U, S):- cofact(X, t(A, L, R), S),
	attach_handle(R, P, R0, S),
	attach_handle(L, P, L0, S),
	pairs_insert(A, P, R0, R1, S),
	zdd_join(L0, R1, U, S).

%
pairs_insert(_, [], U, U, _):-!.
pairs_insert(A, [B|Bs], U, V, S):-
	zdd_insert((A,B), U, U0, S),
	pairs_insert(A, Bs, U0, V, S).

		/*************************
		*     Quotient on ZDD    *
		*************************/

% ?- zdd((X= {[1,2],[3,4]}, zdd_unify(X, [1=2, 3=4, 3=2], Y),  sets(Y, S))).
% ?- zdd((X= {[1,2],[2,3]}, zdd_unify(X, [3=2], Y),  sets(Y, S))).

%%	zdd_quotient(+X, +E, -Y, +G) is det.
% With an ROZDD bound to G,
% Unify Y with a quotient ZDD Y induced by clusters E.

zdd_quotient(X, Y, Z):- shift(zdd_quotient(X, Y, Z)).

zdd_quotient(X, _, X, _):- X<2, !.
zdd_quotient(X, [], X, _):- !.
zdd_quotient(X, Q, Y, S):- cofact(X, t(A, L, R), S),
		zdd_quotient(L, Q, L0, S),
		zdd_quotient(R, Q, R0, S),
		representative(A, Q, A0),
		zdd_insert(A0, L0, L1, S),
		zdd_join(L1, R0, Y, S).

% ?- representative(b, [[a,b]], R).
representative(A, Q, R):-
	(	member(B, Q),
		memberchk(A, B),
		B = [R|_]
	->	true
	;	R = A
	).

% ?- zdd((zdd_map([a], [0,1], M), sets(M, S))).
% ?- zdd((zdd_map([a,b], [0], M), sets(M, S))).
% ?- zdd((zdd_map([a,b], [0,1], M), sets(M, S))).
% ?- zdd((zdd_map([], [0,1], M), sets(M, S))).
% ?- zdd((zdd_map([a], [], M), sets(M, S))).
% ?- zdd((zdd_map([], [], M), sets(M, S))).
% ?- zdd((N = 3, numlist(1, N, L), time(map(L, L, M)))).
% ?- N = 1000, numlist(1, N, L), time(zdd(zdd_map(L, L, M))).

%%	zdd_map(+X, +Y, -Z, +G) is det.
% With an ROZDD bound to G,
% unify Z with the map spaces from X to Y.

zdd_map(X, Y, Z):- shift(zdd_map(X, Y, Z)).

zdd_map([], _, true, _).
zdd_map([X|Xs], R, Y, S):- zdd_map(Xs, R, Z, S),
	zdd_map(R, X, Z, Y, S).
%
zdd_map([], _, _, false, _).
zdd_map([V|Vs], X, Z, Y, S):-
	zdd_map(Vs, X, Z, Y0, S),
	cofact(Y, if(X-V, Z, Y0), S).

%%	zdd_insert(+A, +Y, -Z) is det.
% Insert A to each set in a ZDD Y, and unify Z
% with the augmented ZDD.
% ?- zdd((zdd_insert(a, true, X), zdd_insert(b, X, X1), prz(X1))).
% ?- zdd((X<<pow([a,b,c]), zdd_insert(x, X, X1), prz(X1))).
zdd_insert(X, Y, Z):- shift(zdd_insert(X, Y, Z)).

%
zdd_insert(A, 1, J, S):-!, zdd_singleton(A, J, S).
zdd_insert(_, 0, 0, _):-!.
zdd_insert(A, I, J, S):- memo(zdd_insert(A, I)-J, S),
	(	nonvar(J) ->  true
	;	cofact(I, t(X, L, R), S),
		compare(C, A, X),
		(	C = (=)
		->	zdd_join(L, R, K, S),
			cofact(J, t(X, 0, K), S)
		;	C = (<)
		->	cofact(J, t(A, 0, I), S)
		;  	zdd_insert(A, L, L0, S),
			zdd_insert(A, R, R0, S),
			cofact(J, t(X, L0, R0), S)
		)
	).

% ?- zdd((family([[a, b],[b]], X), sets(X, Sx),
%	zdd_insert_atoms([c, d], X, Y), sets(Y, Sy))).
%%	zdd_insert_atoms(+As:sorted list, +X:zdd, -Y:zdd) det.
%	Insert all atoms in As to X to get the result Y.

zdd_insert_atoms(As, X, Y):- shift(zdd_insert_atoms(As, X, Y)).
%
zdd_insert_atoms([], X, X, _):-!.
zdd_insert_atoms([A|As], X, Y, S):- zdd_insert_atoms(As, X, X0, S),
	zdd_insert(A, X0, Y, S).

% ?- zdd((ord_linear_zdd([c, d, e], Y), sets(Y, Sy))).
ord_linear_zdd(X, Y):- shift(ord_linear_zdd(X,Y)).
ord_linear_zdd([], 1, _).
ord_linear_zdd([A|As], X, S):- ord_linear_zdd(As, X0, S),
	cofact(X, t(A, 0, X0), S).
%
zdd_ord_filter(A, OrdList, F, S):-  zdd_ord_power(OrdList, P, S),
	zdd_insert(A, P, F, S).
%
zdd_meet_list([A], A, _):-!.
zdd_meet_list([A|As], B, S):- zdd_meet_list(As, A, B, S).
%
zdd_meet_list([], X, X, _).
zdd_meet_list([A|As], X, Y, S):- zdd_meet(A, X, X0, S),
	zdd_meet_list(As, X0, Y, S).
%
zdd_join_list([], X, X, _):-!.
zdd_join_list([A|As], X, Y, S):- zdd_join(A, X, X0, S),
	zdd_join_list(As, X0, Y, S).


% ?- zdd((X<<[a], zdd_disj_filter([b], X, Y), sets(Y, S))).
% ?- zdd((X<<([a]+[b]), zdd_disj_filter([-a,-b], X, Y), sets(Y, S))).
% ?- zdd((X<<pow([a,b,c]), zdd_disj_filter([a], X, Y), sets(Y, S))).
% ?- numlist(1, 10, Ns),
%	zdd((X<<pow(Ns), zdd_disj_filter([1,2,3], X, Y), card(Y, S))).
% ?- numlist(1, 1000, Ns), zdd((X<<pow(Ns),
%	time(zdd_disj_filter([1,2,3], X, Y)), card(Y, S))).
% ?- numlist(1, 3, Ns), zdd((X<<pow(Ns),
%	time(zdd_disj_filter([1,2,3], X, Y)), card(Y, S))).

zdd_disj_filter(L, X, Y):- shift(zdd_disj_filter(L, X, 0, Y)).
%
zdd_disj_filter([], _, Y, Y, _):-!.
zdd_disj_filter(_, 0, Y, Y, _):-!.
zdd_disj_filter([-A|As], X, Y0, Y, S):-!,
	zdd_proper_ideal(A, X, U, S),
	zdd_join(U, Y0, Y1, S),
	zdd_disj_filter(As, X, Y1, Y, S).
zdd_disj_filter([A|As], X, Y0, Y, S):-
	zdd_proper_filter(A, X, U, S),
	zdd_join(U, Y0, Y1, S),
	zdd_disj_filter(As, X, Y1, Y, S).

% ?- numlist(1, 1000, Ns), zdd((X<<pow(Ns),
%	zdd_proper_filter(1000, X, Y), zdd_proper_filter(1, X, Y0), card(Y, C), card(Y0, C0),
%	C=C0)).
%@
% ?- zdd((X << pow([a,b]), zdd_proper_filter(b, X, Y), card(Y, C))).
% ?- zdd((X << {[a]}, zdd_proper_filter(b, X, Y), card(Y, C))).

zdd_proper_filter(A, X, Y):- shift(zdd_proper_filter(A, X, Y)).
%
zdd_proper_filter(A, X, Y, S):-zdd_shift(zdd_proper_filter(A, X, Y, S)).

%
zdd_proper_filter(_, X, 0, _, _):- X<2, !.
zdd_proper_filter(A, X, Y, S, M):- memo(zdd_proper_filter(X)-Y, M),
	(	nonvar(Y)	-> true
	; 	cofact(X, t(B, L, R), S),
		compare(C, A, B),
		(	C = (=)
		->	cofact(Y, t(A, 0, R), S)
		;	C = (<)
		->	Y = 0
		;   zdd_proper_filter(A, L, L0, S, M),
			zdd_proper_filter(A, R, R0, S, M),
			cofact(Y, t(B, L0, R0), S)
		)
	).

% ?- zdd((X<<[a], zdd_proper_ideal([b], X, Y), sets(Y, S))).
% ?- zdd((X<<([a]+[b]), zdd_proper_ideal(a, X, Y), sets(Y, S))).
% ?- zdd((X<<pow([a,b,c]), zdd_proper_ideal(a, X, Y), sets(Y, S))).
% ?- numlist(1, 10, Ns),
%	zdd((X<<pow(Ns), zdd_proper_ideal(1, X, Y), card(Y, S))),
%	S is 2^9.
% ?- numlist(1, 1000, Ns), zdd((X<<pow(Ns),
%	time(zdd_proper_ideal(1000, X, Y)), card(Y, S))),
%	S is 2^999.

zdd_proper_ideal(A, X, Y):- shift(zdd_proper_ideal(A, X, Y)).
%
zdd_proper_ideal(A, X, Y, S):- zdd_shift(zdd_proper_ideal(A, X, Y, S)).

%
zdd_proper_ideal(_, X, X, _, _):- X<2, !.
zdd_proper_ideal(A, X, Y, S, M):- memo(zdd_proper_ideal(X)-Y, M),
	(	nonvar(Y)	-> true
	; 	cofact(X, t(B, L, R), S),
		compare(C, A, B),
		(	C = (=)
		->	Y = L
		;	C = (<)
		->	Y = X
		;   zdd_proper_ideal(A, L, L0, S, M),
			zdd_proper_ideal(A, R, R0, S, M),
			cofact(Y, t(B, L0, R0), S)
		)
	).

% ?- zdd((zdd_conj_filter([b], 0, Y), sets(Y, S))).
% ?- zdd((zdd_conj_filter([-b], 0, Y), sets(Y, S))).
% ?- zdd((X<<([a]+[b]), zdd_conj_filter([-a], X, Y), sets(Y, S))).
% ?- zdd((X<<([a]+[b]), zdd_conj_filter([-a, -b], X, Y), sets(Y, S))).
% ?- zdd((X<<pow([a,b,c]), zdd_conj_filter([a,b], X, Y), sets(Y, S))).
% ?- numlist(1, 10, Ns), zdd((X<<pow(Ns), zdd_conj_filter([1,2,3], X, Y), card(Y, S))), S =:= 2^(10-3).
% ?- numlist(1, 1000, Ns), zdd((X<<pow(Ns), zdd_conj_filter([1,2,3], X, Y), card(Y, S))), S =:= 2^(1000-3).
% ?- numlist(1, 1000, Ns), numlist(1, 998, L),
%	zdd((X<<pow(Ns), zdd_conj_filter(L, X, Y), card(Y, S))).

zdd_conj_filter(L, X, Y):- shift(zdd_conj_filter(L, X, Y)).
%
%
zdd_conj_filter([], X, X, _):-!.
zdd_conj_filter(_, 0, 0, _):-!.
zdd_conj_filter([-A|As], X, Y, S):- zdd_proper_ideal(A, X, X0, S),
	zdd_conj_filter(As, X0, Y, S).
zdd_conj_filter([A|As], X, Y, S):- zdd_proper_filter(A, X, X0, S),
	zdd_conj_filter(As, X0, Y, S).


		/*********************
		*     remove atom    *
		*********************/

% ?- zdd((family([[a,b,c],[a1,b,c1]], X), zdd_remove(b, X, Y),
%	sets(Y, Sy))).
%%	zdd_remove(+A, +X, -Y, +G) is det.
% Remove an atom A from each set in a ZDD X,
% and unify Y with the result.

zdd_remove(X, Y, Z):- shift(zdd_remove(X, Y, Z)).

zdd_remove(X, Y, Z, S):- zdd_shift(zdd_remove(X, Y, Z, S)).

%
zdd_remove(_, X, X,  _, _):- X<2, !.
zdd_remove(A, I, J, S, M):- memo(zdd_remove(I)-J, S),
	(	nonvar(J)	->  true
	; 	cofact(I, t(Ai, Li, Ri), S),
		compare(C, A, Ai),
		(	C = (<)  -> J = I
		;   C = (>)  -> zdd_remove(A, Li, Lia, S, M),
						zdd_remove(A, Ri, Ria, S, M),
						cofact(J, t(Ai, Lia, Ria), S)
		; 	zdd_join(Li, Ri, J, S)
		)
	).

% For developing.
dump_args(List):- nl,
	writeln('** Completed Options:'),
	maplist(writeln, List).
%
update_hit_count :-
	( nb_current(count_hit, C), integer(C)
	->	true
	;	C = 0
	),
	C0 is C + 1,
	b_setval(count_hit, C0).
%
print_hit_count	 :- (nb_current(count_hit, N) -> N0 = N; N0 = 0),
  format("~n~w", [count_hit = N0]).
%
newvars(Vars):- b_getval(boole_zip, Zip),
		length(Zip, N),
		I is N + 1,
		length(Vars, J),
		K is I + J - 1,
		(	K =<0 ->  Vars = []
		;	numlist(I, K, Vars)
		),
		zip(Vars, Vars, Zip0),
		append(Zip, Zip0, Zip1),
		b_setval(boole_zip, Zip1).
%
new_root_zdds(Zdds):- length(Zdds, N),
		length(Vars, N),
		newvars(Vars),
		maplist(var_to_zdd, Vars, Zdds).
%
var_to_zdd(I, X):- memo(truth(I)-X).

% ?- new_variables(1+2+A+B, [A], V).
new_variables(A, OldVs, NewVs):- term_variables(A, V0),
		subtract_by_memq(V0, OldVs, NewVs).

% ?- update_zip_by_term(A+B, [1-A], R), update_zip_by_term(A+B+C+D, R, S).
update_zip_by_term(A, OldZip, NewZip):- zip(_, Vs, OldZip),
	new_variables(A, Vs, NewVars),
	length(OldZip, Len),
	update_zip_by_term(NewVars, Len, OldZip, NewZip).
%
update_zip_by_term([], _, Z, Z).
update_zip_by_term([V|Vs], I, Z, NewZ):-
	I0 is I + 1,
	update_zip_by_term(Vs, I0, [I0-V|Z], NewZ).

% ?- subtract_by_memq([X, A, Y, B], [Y, X], U).
subtract_by_memq([], _, []).
subtract_by_memq([A|B], C, D):- memq(A, C), !,
	subtract_by_memq(B, C, D).
subtract_by_memq([A|B], C, [A|D]):-
	subtract_by_memq(B, C, D).
%
union_by_memq([], X, X).
union_by_memq([A|As], X, Y)	:- memq(A, X), !,
    union_by_memq(As, X, Y).
union_by_memq([A|As], X, [A|Y])	:-
	union_by_memq(As, X, Y).
%
union_by_memq([], []).
union_by_memq([A|B], C)	:- union_by_memq(B, C0),
				union_by_memq(A, C0, C).

% ?- solve([sat(a,X)]), zout(X).
zout_zip([]):-!.
zout_zip([I-Z|R]):-!, zdd_to_prop(Z, Z0),
		writeln(I-Z0),
		zout(R).
%
zout(X):- integer(X), !,
		zdd_to_prop(X, Y),
		writeln(node(X)=Y).
zout(X):- writeln(X).

%
zout(A, V):-!, zdd_to_prop(V, V0),
	   format("~w: ~w", [A, V0]).

%
zdd_to_prop(X, Y):- zdd_to_prop(X, Y, [slim_boole(true)]).
%
zdd_to_prop(X, Y, Options):- zdd_to_tree(X, T),
		tree_to_prop(T, Y0),
		( memberchk(slim_boole(true), Options)
		-> slim_boole(Y0, Y)
		;  Y=Y0
		).
%
ztree([]):-!.
ztree([[X-Y]|R]):- !, ztree(X, Y), 	ztree(R).
ztree(X):- zdd_to_tree(X, Y), writeln(Y).

%
ztree(X, Y):- zdd_to_tree(Y, Y0),
	  format("~w  = ~w\n",[X, Y0]).

%
zdd_to_tree(I, if(X, L, R)):- integer(I), !,
	cofact(I, if(X, L0, R0)),
	zdd_to_tree(L0,  L),
	zdd_to_tree(R0,  R).
zdd_to_tree(X, X).

% ?- tree_to_prop(if(1, truth(1), false), P, 1).
tree_to_prop(X, Y):- varnum(N), tree_to_prop(X, Y, N).
%
tree_to_prop(false, false, _):-!.
tree_to_prop(true, true, _):-!.
tree_to_prop(if(X, L, R), X*L0 + -(X)*R0, N):-
	tree_to_prop(L, TL, N),
	tree_to_prop(R, TR, N),
	depth(L, A, N),
	depth(R, B, N),
	X0 is X + 1,
	conj_of_negs(X0, A, TL, L0),
	conj_of_negs(X0, B, TR, R0).

%
depth(if(I,_,_), I, _):-!.
depth(_, N, N).

% ?- conj_of_negs(1, 3, A).
conj_of_negs(I, J, A):- I=<J, conj_of_negs(I, J, -(J), A).

% ?- conj_of_negs(3, 1, a, X).
% ?- conj_of_negs(1, 3, a, X).
conj_of_negs(I, J, A, A):- I>=J, !.
conj_of_negs(I, J, A, B):- J0 is J - 1,
						conj_of_negs(I, J0, -(J0)*A, B).

% ?- zdd((X<<pow([a,b]), card(X, C))).

%%	card(+I, -C) is det.
% unify C with the cardinality of the ZDD I
% as a family of sets.

card(X, Y):- shift(card(X, Y)).
%
card(X, Y, S):- zdd_shift(card(X, Y, S)).
%
card(I, I, _, _):- I < 2, !.
card(I, C, S, M):- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(_, L, R), S),
		card(R, Cr, S, M),
		card(L, Cl, S, M),
		C is Cl + Cr
	).

% ?- gf2_count([1,1,1], X).
% ?- gf2_count([2,2,2], X).
% ?- gf2_count([1, 2, 3, 4], X).
% ?- numlist(1, 100, Ns),  gf2_count(Ns, C).
gf2_count(X, Y):- gf2_count(X, 1, Y).

gf2_count([], X, X).
gf2_count([N|Ns], X, Y):- X0 is X*(2^N - 1),
	gf2_count(Ns, X0, Y).


% ?- sat_count(((a=b)=c)=(a=(b=c)), X).
% ?- zdd((even_odd_power([1-1], X, Y), gf2_sat_count([Y-1], 0, V))).
% ?- zdd((even_odd_power([1-1, 2-1], X, Y), gf2_sat_count([Y-1], 0, V))).
% ?- zdd((even_odd_power([1-1, 2-1, 3-1], X, Y), gf2_sat_count([Y-1], 0, V))).
% ?- zdd((even_odd_power([1-1, 2-2, 3-3], X, Y), gf2_sat_count([Y-1], 0, C))).

%
gf2_sat_count(X, Y, Z):- shift(gf2_sat_count(X, Y, Z)).
%
gf2_sat_count([], U, U, _):-!.
gf2_sat_count([P-C|As], U, V, S):-	gf2_sat_count(P, C, As, U, V, S).

%
gf2_sat_count(1, C, As, U, V, S):-!, U0 is U + C,
	gf2_sat_count(As, U0, V, S).
gf2_sat_count(0, _, As, U, V, S):-!, gf2_sat_count(As, U, V, S).
gf2_sat_count(P, C, As, U, V, S):- cofact(P, t(_-J, L, R), S),
	C0 is C * (2^J-1),
	gf2_sat_count(R, C0, [L-C|As], U, V, S).

% ?- zdd((even_odd_power([1,2,3], X, Y), psa(X), psa(Y))).
% ?- zdd((even_odd_power([1,2,3], X, Y), Z<<pow([1,2,3]),
%	zdd_join(X, Y, Z))).

even_odd_power(X, Y, Z):- shift(even_odd_power(X, Y, Z)).
%
even_odd_power([], 1, 0, _).
even_odd_power([X|Xs], Ev, Od, S):-
	even_odd_power(Xs, Ev0, Od0, S),
	zdd_insert(X, Ev0, X_Od, S),
	zdd_insert(X, Od0, X_Ev, S),
	zdd_join(X_Od, Od0, Od, S),
	zdd_join(X_Ev, Ev0, Ev, S).

% ?- zdd((X<<set([a,b]), card(X, D), even_odd_card(X, C))).
% ?- zdd((X<<set([a]), card(X, D), even_odd_card(X, C))).
% ?- zdd((X<<pow([a,b]), card(X, D), even_odd_card(X, C))).
% ?- zdd((X<<pow([a,b,c]), card(X, D), even_odd_card(X, C))).
% ?- zdd((X<<pow([a,b,c,d]), card(X, D), even_odd_card(X, C))).
% ?- zdd((X<<pow([a,b,c,d,e]), card(X, D), even_odd_card(X, C))).

even_odd_card(X, C):- shift(even_odd_card(X, C)).
%
even_odd_card(X, C, S):- zdd_shift(even_odd_card(X, C, S)).
%
even_odd_card(1, 1-0, _, _):-!.
even_odd_card(0, 0-0, _, _):-!.
even_odd_card(I, C, S, M):- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(_, L, R), S),
		even_odd_card(L, Cl, S, M),
		even_odd_card(R, Cr, S, M),
		Cr = (Er-Or),
		Cl = (El-Ol),
		E is El + Or,
		O is Ol + Er,
		C = (E-O)
	).

								%
zip([], [], []).
zip([X|Xs], [Y|Ys], [X-Y|Zs]) :- zip(Xs, Ys, Zs).
%
unify_zip([]).
unify_zip([X-X|R]):- unify_zip(R).

% For debugging.
zdd_snap(X):- zdd_to_prop(X, X0),
		  writeln(zdd_snap(X0)).


%
% ?- zdd((zdd_append([a,b,c], 1, X), cofact0(X, A, B), card(X, C), psa(X))).
% ?- zdd((zdd_append([a,b,c], 0, X), psa(X))).

%% zdd_append(+X, +Y, -Z [, G]) is det.
%  Insert all elements of X into Y. Z is the result.
%  Y is assumed to have no atom in X.
zdd_append(X, Y, Z):- shift(zdd_append(X, Y, Z)).
%
zdd_append(_, [], [], _S):-!.
zdd_append([], X, X, _S).
zdd_append([A|As], X, Y, S):- zdd_append(As, X, X0, S),
	cofact0(Y, A, X0, S).


		/******************************************
		*     Oprations on clauses of literals    *
		******************************************/

% ?- ltr_compare(C, a, -a).
% ?- ltr_compare(C, -(a), a).
% ?- ltr_compare(C, -(b), a).
% ?- ltr_compare(C, -(a), b).
% ?- ltr_compare(C, -(a), -(b)).
% ?- ltr_compare(C, -(a), -(a)).

%%  ltr_compare(-C, +X, +Y) is det.
%   Compare literals X and Y, and C is unified with
%	<, =, > as the standard compare/3.
%   ex. a<b, a < -a,  -a < b,   -a < -b.

ltr_compare(C, -(X), -(Y)):-!, compare(C, X, Y).
ltr_compare(C, X, -(Y)):-!, compare(C0, X, Y),
		(	C0 == (=) ->  C =(<)
		;	C0 = C
		).
ltr_compare(C, -(X), Y):-!, compare(C0, X, Y),
		(	C0 == (=) -> C = (>)
		;	C0 = C
		).
ltr_compare(C, X, Y):- compare(C, X, Y).

% ?- ltr_sort([b, b, a], X).
% ?- ltr_sort([-b, b,-b, -a], X).

%%	ltr_sort(+X, -Y) is det.
%   Y is unified with sorted X, w.r.t  ltr_compare.

ltr_sort(X, Y):- predsort(ltr_compare, X, Y).

% ?- zdd((X << ltr_pow([a, -b, c]), ltr_memberchk([a, -b], X))).
% ?- zdd((X << pow([a, -b, c]), ltr_has_set([a, b], X))).
% ?- zdd((X << ltr_power([a, -b, c]), ltr_has_set([a, b], X))).

ltr_memberchk(L, Z):- shift(ltr_memberchk(L, Z)).
%
ltr_memberchk(L, Z, S):- ltr_sort(L, L0),
	ltr_ord_memberchk(L0, Z, S).
%
ltr_ord_memberchk([], Z, S):- !,  zdd_has_1(Z, S).
ltr_ord_memberchk([A|As], Z, S):- Z > 1,
	cofact(Z, t(B, L, R), S),
	ltr_compare(C, A, B),
	(	C = (=) -> ltr_ord_memberchk(As, R, S)
	;	ltr_ord_memberchk([A|As], L, S)
	).

%%
% ?- zdd((a_big_cnf(1, X), ltr_card(1, X, C))).
% ?- zdd((a_big_cnf(2, X), ltr_card(2, X, C))).
% ?- zdd((a_big_cnf(3, X), ltr_card(3, X, C))).
% ?- zdd((a_big_cnf(10, X), ltr_card(10, X, C))).
% ?- zdd((a_big_cnf(20, X), ltr_card(20, X, C))).
% ?- time(zdd((a_big_cnf(100, X), ltr_card(100, X, C)))), C =:= 2^100.
%@ % 2,392,546 inferences, 0.290 CPU in 0.409 seconds (71% CPU, 8246121 Lips)
%@ X = 30001,
%@ C = 1267650600228229401496703205376 .
%
% ?- zdd((a_big_cnf(10, X), ltr_refute(X))).
% ?- time(zdd((a_big_cnf(20, X), ltr_refute(X)))).
% ?- call_with_time_limit(100, time(zdd((a_big_cnf(30, X), ltr_refute(X))))).

a_big_cnf(X, Y):- shift(a_big_cnf(X, Y)).
a_big_cnf(0, 1, _).
a_big_cnf(N, X, S):- N>0,
	N0 is N-1,
	a_big_cnf(N0, X0, S),
	ltr_insert(N, X0, R, S),
	ltr_insert(-N, X0, L, S),
	zdd_join(L, R, X, S).

% ?- numlist(1, 3, Ns), zdd((X<<(pow(Ns)-1), ltr_choices(X, Y), card(Y, C))).
% ?- time((numlist(1, 20, Ns), zdd((X<<(pow(Ns)-1), ltr_choices(X, Y), card(Y, C))))).
%@ % 143,146,953 inferences, 9.734 CPU in 9.756 seconds (100% CPU, 14705180 Lips)
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ X = 41,
%@ Y = 60,
%@ C = 1.
% ?- zdd((boole_to_dnf((a*b)+c, X), ltr_choices(X, Y), ltr_choices(Y, Z),
%	ltr_choices(Z, U), psa(X), nl, psa(Y), nl, psa(Z), nl, psa(U))), U=Y.
% ?- zdd((X<<{[a,b],[c,d]},  ltr_choices(X, Y), ltr_choices(Y, Z),
%	ltr_choices(Z, U), psa(X), nl, psa(Y), nl, psa(Z), nl, psa(U))), U=Y.
% ?- zdd((X<< ( [b,d]+[a,d]+[a,b,d] ), ltr_choices(X, Y),
%	ltr_choices(Y, Z), ltr_choices(Z, U),	ltr_choices(U, V),
%	 psa(X), nl, psa(Y), nl, psa(Z), nl, psa(U), nl, psa(V))).
% ?- zdd(( (X<< {[a,b]}), ltr_choices(X, Y), psa(X), nl, psa(Y))).
% ?- zdd(( (X<< {[a,b,c]}), ltr_choices(X, Y),	psa(X), nl, psa(Y))).
% ?- zdd(( (X<< {[a, b], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[], [a, b], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[a], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b], [c, d], [e,f]}, ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b]}, ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[]}, ltr_choices(X, Y), psa(Y))).

%
ltr_choices(X, Y):- shift(ltr_choices(X, Y)).

ltr_choices(1, 1, _):-!.
ltr_choices(0, 1, _):-!.

% ltr_choices(1, 0, _):-!.
% ltr_choices(0, 1, _):-!.

ltr_choices(X, Y, S):- memo(ltr_choices(X)-Y, S),
	(	nonvar(Y)	->	true
	;	ltr_choices(X, [], 1, Y, S)
	).
%

ltr_choices(0, _, Y, Y, _):-!.
ltr_choices(1, As, Y0, Y, S):-!, ltr_fold_insert(As, Y0, 0, Y, S).
ltr_choices(X, As, Y0, Y, S):- cofact(X, t(A, L, R), S),
	(	L = 0
	->	ltr_choices(R, [A|As], Y0, Y, S)
	; 	L = 1
	->	(	As = []
		->  Y = 0
		;	ltr_fold_insert(As, Y0, 0, Y1, S),
			ltr_choices(R, [A|As], Y1, Y, S)
		)
	;	ltr_choices(L, As, Y0, Y1, S),
		ltr_choices(R, [A|As], Y1, Y, S)
	).
%
ltr_fold_insert([], _, Y, Y, _).
ltr_fold_insert([A|As], Y0, U, Y, S):-
	ltr_insert(A, Y0, V, S),
	ltr_join(U, V, U0, S),
	ltr_fold_insert(As, Y0, U0, Y, S).

% ?- sat(-(true)).
%@ false.
% ?- sat(-(a + -a)).
%@ false.
% ?- sat(a).
%@ true.
% ?- sat(or(a, b)), sat_count(C).
%@ C = 3.
% ?- sat(xor(a, b)), sat_count(C).
%@ C = 1.
% ?- sat(exor(a, b)), sat_count(C).
%@ C = 2.
% ?- sat(e_x_o_r(a, b)), sat_count(C).
%@ C = 1.
% ?- sat(a), sat(-a).
%@ false.
% ?- sat(a), sat(a+b), sat_count(C).
%@ C = 2.
% ?- sat(a(1)+a(2)), sat_count(C).
%@ C = 3.
% ?- sat(A+a(2)), sat_count(C).
%@ C = 3.
% ?- sat(A + B), sat_count(C).
% ?- sat(a+b+c+d), sat_count(C).
%@ C = 15.
% ?- sat(a+b), sat(c+d),  sat_count(C).
%@ C = 9.
% ?-sat(a->b), sat(b->c), sat(c->d), sat(d->a), sat_count(C).
%@ C = 2.
% ?- sat(X=(a=b)), sat(Y=(b+c)), sat(X=Y), sat_count(C).
%@ C = 4.
% ?- sat(X=(a=b)), sat(Y=(b+c)), sat(X=Y), sat(X = (-Y)).
%@ false.
% ?-sat(a->b), sat(b->c), sat(c->d), sat(d->a), sat(-(d = b)).
%@ false.
% ?- E = (a=b)*(b=c)*(c=a), sat(E), sat_count(C).
%@ E =  (a=b)*(b=c)*(c=a),
%@ C = 2.
% ?- sat(A + B), sat_count(C).
%@ C = 3.
% ?- sat(A -> B), Vs = [A,B], sat(+([1|Vs])), sat_count(C).
%@ Vs = [A, B],
%@ C = 3 .
% ?- sat(A -> B), Vs = [A,B], sat_count(+([1|Vs]), C).
%@ Vs = [A, B],
%@ C = 4 .
% ?- sat(a=<(b=<a)), sat_count(Count).

sat(X):-
	(	nb_current(sat_state, S),  S \== [] -> true
	; 	open_state(S0, [extra([root-1, varnum-0, varzip-[]])]),
		b_setval(sat_state, S0)
	),
	b_getval(sat_state, State),
	get_key(root, Root, State),
	dnf(X, Y, State),
	ltr_merge(Y, Root, Root0, State),
	set_key(root, Root0, State),
	Root0 \== 0.


% ?- fos(a*b), fos_count(C).
% ?- fos(a), fos_count(C).
% ?- fos(pow([a, b])/a), fos_count.
% ?- fos(pow([a, b])//a), fos_count.
% ?- fos((a+b)*(x+Y)), fos_count.
% ?- fos((a+b)&(x+Y)), fos_count.
% ?- fos((a+b)&(a+c)), fos_count.

fos(X):-
	(	nb_current(fos_state, S),  S \== [] -> true
	; 	open_state(S0, [extra([root-0, varnum-0, varzip-[]])]),
		b_setval(fos_state, S0)
	),
	b_getval(fos_state, State),
	get_key(root, Root, State),
	fos(X, Y, State),
	zdd_join(Y, Root, Root0, State),
	set_key(root, Root0, State).
%
sat_count(C):- b_getval(sat_state, S),
	get_key(varnum, N, S),
	get_key(root, X, S),
	ltr_card(N, X, C, S).
%
fos_count(C):- b_getval(fos_state, S),
	get_key(root, R, S),
	card(R, C, S).
%
fos_count :- fos_count(C), writeln(C).

% ?- sat_count(a=<(b=<a), C).
% ?- sat_count(a->(b->a), C).
sat_count(F, C):- zdd((
	dnf(F, X),
	get_key(varnum, N),
	ltr_card(N, X, C))).

fos_count(F, C):- zdd((
	fos(F, X),
	card(X, C))).

sat0(X):-
	(	nb_current(sat_state, S),  S \== [] -> true
	; 	open_state(S0, [extra([root-1, varnum-0, varzip-[]])]),
		b_setval(sat_state, S0)
	),
	b_getval(sat_state, State),
	get_key(root, Root, State),
	dnf0(X, Y, State),
	ltr_merge(Y, Root, Root0, State),
	set_key(root, Root0, State),
	Root0 \== 0.
%
sat_count0(C):- b_getval(sat_state, S),
	get_key(root, R, S),
	get_key(varnum, N, S),
	ltr_card(N, R, C, S).
%
sat_count0(F, C):- zdd((
	dnf0(F, X),
	get_key(varnum, N),
	ltr_card(N, X, C))).


print_root:- b_getval(sat_state, S), get_key(root, R, S), psa(R, S).
%
sat_clear:- nb_setval(sat_state, []).


% ?- zdd((dnf(a->b, Y), psa(Y))).
% ?- zdd((dnf(a+b, X),  dnf(b+c, Y), ltr_merge(Y, X, Z), ltr_card(3, Z, C))).
% ?- zdd((dnf(a+b+c, X), ltr_card(3, X, Count))).
% ?- zdd((dnf(A+b+c, X), ltr_card(3, X, Count))).
% ?- zdd((dnf(A+B+C, X), ltr_card(3, X, Count))).
% ?- zdd((dnf(a+b, X), sat(b+c, Y), ltr_join(X, Y, Z), ltr_card(3, Z, C))).
% ?- zdd((dnf(a+b, X), dnf(b+c, Y), ltr_merge(X, Y, Z), ltr_card(3, Z, C))).
% ?- zdd((dnf((a+b)*(b+c), X), ltr_card(3, X, C))).
% ?- zdd((dnf(A->B, X),  dnf(B->C, Y), ltr_merge(Y, X, Z),
%	ltr_card(3, Z, Count))).
% ?- zdd((dnf((A->B)*(B->C), X),  ltr_card(3, X, Count))).
% ?- zdd((dnf(a=b, X), dnf(b=c, Y), ltr_merge(X, Y, Z), ltr_card(3, Z, Count))).
% ?- zdd((dnf(x+y, F), ltr_card(2, F, C), dnf(-(x), D), ltr_card(2, D, E))).
% ?- zdd((dnf(x, F), ltr_card(2, F, C))).
% ?- zdd((dnf(x, F), ltr_card(3, F, C))).
% ?- zdd((dnf(-x, F), ltr_card(3, F, C))).
% ?- zdd((dnf(true, F), ltr_card(3, F, C))).
% ?- zdd((dnf(false, F), ltr_card(3, F, C))).
% ?- zdd((dnf(x=y, F), ltr_card(2, F, C))).
% ?- zdd((dnf(x=y, F), ltr_card(3, F, C))).
% ?- zdd((dnf(x + y, F), ltr_card(2, F, C))).
% ?- zdd((dnf(-(x + y), F), ltr_card(2, F, C))).
% ?- zdd((dnf(x + y + z, F), ltr_card(3, F, C))).
% ?- zdd((dnf(-(x + y + z), F), ltr_card(3, F, C))).

% ?- zdd(prove_by_refutation((a->b)->(b->a))).
% ?- zdd(prove_by_refutation((a->b)->(a->b))).
% ?- zdd(prove_by_refutation((a->b)->(b->a))).
% ?- zdd(prove_by_refutation((a->b))).
% ?- zdd(prove_by_refutation((a))).
% ?- spy(boole_to_cnf), prove_by_refutation((a->b)->b).
% ?- prove_by_refutation(a->a).
% ?- prove_by_refutation(-(-(a->a))).
% ?- prove_by_refutation((-(a->a))).
% ?- prove_by_refutation((a)).
% ?- prove_by_refutation(-a).
% ?- N = 10, numlist(1, N, Ns), zdd((X<<(pow(Ns)), ltr_card(N, X, C))).
% ?- N = 1000, numlist(1, N, Ns), zdd((X<<(pow(Ns)-1), ltr_card(N, X, C))).

% ?- forall(valid_formula(F, N), (expand_boole_macro(F, F0),
%	zdd((boole_to_dnf(F0, D), psa(D),
%		ltr_card(N, D, C),  writeln(F), writeln(F0),
%	writeln(ltr_card(N, D, C)))))).

% valid_formula/2 with the number of variables.
valid_formula(-(-a)->a, 1).
valid_formula(a -> a, 1).
valid_formula(a -> (b -> a), 2).
valid_formula((a->b)->((b->c)->(a->c)), 3).
valid_formula(a->((a->b)->b), 2).
valid_formula((a*(a->b))->b, 2).
valid_formula((a->b)->(-b -> -a), 2).
valid_formula((a->b)->((b->c)->(a->c)), 3).

% ?- time(forall(valid_formula(A), prove_by_refutation(A))).
% valid_formula/1.v
valid_formula(-(-a)->a).
valid_formula(a -> a).
valid_formula(a -> (b -> a)).
valid_formula((a->b)->((b->c)->(a->c))).
valid_formula(a->((a->b)->b)).
valid_formula((a*(a->b))->b).
valid_formula((a->b)->(-b -> -a)).
valid_formula((a->b)->((b->c)->(a->c))).

% ?- time(forall(non_valid_formula(A), prove_by_refutation(A))).
non_valid_formula(a).
non_valid_formula((a->b)->a).
non_valid_formula((a->b)->(b->a)).
non_valid_formula(a -> (b -> c)).

% ?- zdd(prove_by_refutation((a->b)->(b->a))).
% ?- zdd(prove_by_refutation(a->b)).
% ?- zdd(prove_by_refutation(a)).
% ?- zdd(prove_by_refutation(a->a)).
% ?- zdd((ltr_complement_toplevel(0, X))).
% ?- zdd((ltr_complement_toplevel(1, X))).
% ?- zdd((ltr_choices(1, X))).
% ?- zdd((ltr_choices(0, X))).

%
prove_by_refutation(X):-					% writeln(X),
	zdd((cnf(X, X1),						% psa(X1),
		ltr_complement_toplevel(X1, X2),	% nl, nl, psa(X2),
		ltr_choices(X2, X3),				% nl, nl, psa(X3),
		(	ltr_refute(X3)
		->  writeln('VALID.')
		;	writeln('NOT VALID.')
		))).

% ?- spy(ltr_refute).
% ?- zdd((X<<dnf(((a + b)*(a + -c)*(-a + c)*(-a + -b))),
%	psa(X), ltr_refute(X))).
% ?- zdd((X<<cnf(((a + b)*(a + -c)*(-a + c)*(-a + -b))),
%	psa(X), ltr_refute(X))).

% ?- zdd((X<<dnf(((a + b)*(a + -c)*(-a + c)*(-a + -b))), psa(X))).
% ?- spy(ltr_refute).
% ?- zdd((X<<cnf(((a + b)*(a + -c)*(-a + c)*(-a + -b))),
%	psa(X), ltr_refute(X))).
% ?- zdd((X<< cnf(-a * a), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(-a * a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(-a + a,  X), psa(X), ltr_refute(X))).
% ?- zdd((cnf(-a + a,  X), psa(X), ltr_refute(X))).
% ?- zdd((cnf((-a )*(a),  X), psa(X), ltr_refute(X))).
% ?- zdd((dnf(-a + a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(a,  X), psa(X), ltr_refute(X))).

ltr_refute(X):- shift(ltr_refute(X)).
%
ltr_refute(X, S) :- zdd_has_1(X, S), !.
ltr_refute(X, S) :- X > 1,
	cofact(X, t(A, L, R), S),
	(	L = 0
	->	(  R > 2,
		   ltr_refute(R, S)				% pure literal
		)
	;	cofact(L, t(B, L0, R0), S),
		(	B = -(A)
		-> 	ltr_merge(R0, R, Y0, S),	% resolution
			ltr_join(L0, Y0, Y, S),
			ltr_refute(Y, S)
		;	R = 1
		->	ltr_refute(L, S)			% unit
		;   ltr_join(L, R, Y, S),		% pure literal
			ltr_refute(Y, S)
		)
	).

% ?- zdd((dnf(-a * a,  X), psa(X), ltr_refute(X))).
% ?- zdd((dnf(a->a,  X), psa(X), ltr_refute(X))).
% ?- zdd((dnf(-a * a,  X), psa(X), ltr_refute(X))). % 0
% ?- zdd((dnf( a + a,  X), psa(X), ltr_refute(X))).
% ?- zdd((dnf( a + a,  X), psa(X), ltr_card(1, X, C))).
% ?- zdd((dnf(-a * a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(a->a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_dnf(-a * a,  X), psa(X), ltr_refute(X))). % 0
% ?- zdd((boole_to_dnf( a + a,  X), psa(X), ltr_refute(X))).
% ?- zdd((dnf(1, X), ltr_card(1, X, C))).
% ?- zdd((dnf(0, X), ltr_card(1, X, C))).
% ?- zdd((dnf(a, X), ltr_card(1, X, C))).
% ?- zdd((dnf(a+b, X), ltr_card(2, X, C))).
% ?- zdd((dnf(-(a+b), X), ltr_card(2, X, C))).
% ?- zdd((dnf(-a * a,  X), psa(X))).
% ?- zdd((dnf(a, X), psa(X))).
% ?- zdd((dnf(-a, X), psa(X))).
% ?- zdd((dnf(a+b+(-c), X), psa(X))).
% ?- zdd((dnf(-a * a,  X), psa(X))).
% ?- zdd((dnf(+([2,3,4]), X), ltr_card(3, X, C))).
% ?- zdd((dnf(+([2,3,4]), X), ltr_card(X, C))).
% ?- zdd((dnf(*([2,3,4]), X), ltr_card(3, X, C))).
% ?- zdd((dnf(*([2,3,4]), X), ltr_card(X, C))).
% ?- zdd((dnf(*([2,3,4]), X), ltr_card(X, C))).
% ?- zdd((dnf(A+b, X), ltr_card(X, C))).
% ?- zdd((dnf(A+B+c+A+B, X), ltr_card(X, C))).
% ?- zdd((dnf(A+B+c+A+B, X),
% ?- sat_count(A+B+c+A+B, C).

dnf(X, Y):- shift(dnf(X,Y)).

dnf(X, Y, S):- intern(X, Z, S),
	boole_to_dnf(Z, Y, S).

dnf0(X, Y):- shift(dnf0(X,Y)).

dnf0(X, Y, S):- intern(X, Z, S),
	simple_boole(Z, Z0),
	boole_to_dnf(Z0, Y, S).


% ?- zdd((zdd:intern(A, B), extern(B, E))).
% ?- zdd((zdd:intern(A+B, X), extern(X, E))).

intern(X, Y):- shift(intern(X, Y)).

intern(X, Z, S):- memo(atomic_prop_memo - M, S),
		(	nonvar(M)-> true
		;   open_state(M, [])
		),
		getmemo(varnum-NewCount, OldCount, M),
		(	var(OldCount) -> OldCount=0; true ),
		count_atomic_prop(X, OldCount-Count, atomic_prop, M),
		term_variables(X, Vs),
		(	Vs = []
		->  Z  = X,
			Ws = Vs
		; 	copy_term(X-Vs, Z-Ws)
		),
		zip(Vs, Ws, Varzip),
		getmemo(varzip-NewVarzip, OldVarzip, M),
		(	var(OldVarzip) -> OldVarzip=[]; true),
		numbering_prop_var(Varzip, OldVarzip, OldVarzip,
					   NewVarzip, Count, NewCount),
		set_key(varnum, NewCount, S),
		set_key(varzip, NewVarzip, S).

%
qfind_key(K-V, [K0-V|_]):- K==K0, !.
qfind_key(P, [_|Z]):- qfind_key(P, Z).

%
numbering_prop_var([], _, Zip, Zip, C, C).
numbering_prop_var([P|Z], Zip0, Zip1, Zip2, C, D):-
	(	qfind_key(P, Zip0)
	->	numbering_prop_var(Z, Zip0, Zip1, Zip2, C, D)
	;	C0 is C + 1,
		P=(X-Y),
		Y= $(C0),
		numbering_prop_var(Z, Zip0, [X-Y|Zip1], Zip2, C0, D)
	).

		/*********************
		*     bool to dnf    *
		*********************/

% ?- simple_boole((c*b), X).
% ?- simple_boole((c*b)*a, X).
%@ X = c*(a*b).

% ?- simple_boole((c=:=b)=:=a, X).
simple_boole(X, Y):- boole_rule(X, X0), !,
	simple_boole(X0, Y).
simple_boole(X, X).


boole_rule(X, Y):-boole_AND(X, Y).
boole_rule(X, Y):-boole_OR(X, Y).
boole_rule(X, Y):-boole_IMPLY(X, Y).
boole_rule(X, Y):-boole_EQ(X, Y).
boole_rule(X, Y):-boole_NOT(X, Y).

%
boole_NOT(-(-X), X).
boole_NOT(-(1), 0).
boole_NOT(-(0), 1).
boole_NOT(-(X), -(Y)):-boole_rule(X, Y).

%
boole_AND(X*X, X).
boole_AND(0*_, 0).
boole_AND(_*0, 0).
boole_AND(1*X, X).
boole_AND(X*1, X).
boole_AND((X*Y)*Z, X*(Y*Z)).
boole_AND(X*Y, Y*X):-Y@<X.
boole_AND(X*(Y*Z), (Y*(X*Z))):-Y@<X.
boole_AND(X*Y, X0*Y):- boole_rule(X, X0).
boole_AND(X*Y, X*Y0):- boole_rule(Y, Y0).

%
boole_OR(X+X, X).
boole_OR(0+X, X).
boole_OR(X+0, X).
boole_OR(1+_, 1).
boole_OR(_+1, 1).
boole_OR((X+Y)+Z, X+(Y+Z)).
boole_OR(X+Y, Y+X):-Y@<X.
boole_OR(X+(Y+Z), (Y+(X+Z))):-Y@<X.
boole_OR(X+Y, X0+Y):- boole_rule(X, X0).
boole_OR(X+Y, X+Y0):- boole_rule(Y, Y0).

%
boole_EQ(X=:=X, 1).
boole_EQ(X=:=1, X).
boole_EQ(1=:=X, X).
boole_EQ(X=X, 1).
boole_EQ(1=X, X).
boole_EQ(X=1, X).
boole_EQ(X=:=Y, Y=:=X):-Y@<X.
boole_EQ(X=Y, Y=X):-Y@<X.
boole_EQ(X=:=Y, X0=:=Y):- boole_rule(X, X0).
boole_EQ(X=:=Y, X=:=Y0):- boole_rule(Y, Y0).
boole_EQ(X=Y, X0=Y):- boole_rule(X, X0).
boole_EQ(X=Y, X=Y0):- boole_rule(Y, Y0).

%
boole_IMPLY(1->Y, Y).
boole_IMPLY(0->_, 1).
boole_IMPLY(X->Y, X0->Y):- boole_rule(X, X0).
boole_IMPLY(X->Y, X->Y0):- boole_rule(Y, Y0).


% ?- zdd((boole_to_dnf(dnf([[a]]), Y), psa(Y))).
% ?- zdd((boole_to_dnf(dnf([[a, b]]), Y), psa(Y))).
% ?- zdd((boole_to_dnf(dnf([[a, a]]), Y), psa(Y))).
% ?- zdd((boole_to_dnf(dnf([[a, -a]]), Y), psa(Y))).
% ?- zdd((boole_to_dnf(dnf([[a, b], [a, c]]), Y), psa(Y))).
% ?- zdd((boole_to_dnf(x*dnf([[a, b], [a, c]]), Y), psa(Y))).

%@ true.
%@ zdd.pl qcompiled.

%
boole_to_dnf(X, Y):- simple_boole(X, X0),
	shift(boole_to_dnf(X0,Y)).
%
boole_to_dnf(X, Y, S):- make_boole_canonical(X, X0),
	canonical_boole_to_dnf(X0, Y, S).
%
canonical_boole_to_dnf(1, 1, _):-!.
canonical_boole_to_dnf(0, 0, _):-!.
canonical_boole_to_dnf(X * Y, Z, S):-!,
	boole_to_dnf(X, U, S),
	boole_to_dnf(Y, V, S),
	ltr_merge(U, V, Z, S).
canonical_boole_to_dnf(X + Y, Z, S):-!, boole_to_dnf(X, U, S),
	boole_to_dnf(Y, V, S),
	ltr_join(U, V, Z, S).
canonical_boole_to_dnf(+(V), Y, S):- fold_join_list(V, 0, Y, S).
canonical_boole_to_dnf(*(V), Y, S):-!, ltr_append(V, 1, Y, S).
canonical_boole_to_dnf(dnf(V), Y, S):-!, ltr_read_dnf(V, 0, Y, S).
canonical_boole_to_dnf(X, Y, S):-!, zdd_singleton(X, Y, S).
%
ltr_read_dnf(X, Y, Z):- shift(ltr_read_dnf(X, Y, Z)).
%
ltr_read_dnf([], Y, Y, _).
ltr_read_dnf([C|Cs], Y, Z, S):- ltr_append(C, 1, C0, S),
	ltr_join(C0, Y, Y0, S),
	ltr_read_dnf(Cs, Y0, Z, S).

%
fold_join_list([], Y, Y, _).
fold_join_list([X|Xs], Y0, Y, S):-
	fold_join_list(Xs, Y0, Y1, S),
	(	integer(X),
		X<2
	->	A = X
	; 	zdd_singleton(X, A, S)
	),
	ltr_join(A, Y1, Y, S).

%% ltr_join(+X, +Y, -Z) is det.
% Unify Z with a ZDD that represents the union of the ZDD X and Y
% as a family of sets of literals.
%
ltr_join(X, Y, Z):- shift(ltr_join(X, Y, Z)).
%
ltr_join(X, X, X, _):-!.
ltr_join(0, X, X, _):-!.
ltr_join(X, 0, X, _):-!.
ltr_join(1, _, 1, _):-!.
ltr_join(_, 1, 1, _):-!.
ltr_join(X, Y, Z, S):-
	(	X@<Y -> memo(ltr_join(X,Y)-Z, S)
	;	memo(ltr_join(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true %, write(.)
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		ltr_compare(C, A, A0),
		(	C = (<) -> 	ltr_join(L, Y, U, S),
						ltr_cofact(Z, t(A, U, R), S)
		;	C = (=)	->	ltr_join(R, R0, U, S),
						ltr_join(L, L0, V, S),
						ltr_cofact(Z, t(A, V, U), S)
		; 	ltr_join(L0, X, U, S),
			ltr_cofact(Z, t(A0, U, R0), S)
		)
	).

% ?- zdd((cnf(a, X), S<<sets(X))).
% ?- zdd((X<<cnf(-a), Y<<sets(X))).
% ?- zdd((X<<cnf(a+b), Y<<sets(X))).
% ?- zdd((X<<cnf(+([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<cnf(*([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<dnf(+([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<dnf(*([a,b,c])), Y<<sets(X))).
% ?- zdd((cnf(a, X), psa(X))).
% ?- zdd((cnf(-a, X), psa(X))).
% ?- zdd((cnf(a+b, X), psa(X))).
% ?- zdd((cnf(a+b+(-c), X), psa(X))).
% ?- zdd((cnf(-a * a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(a->a,  X), psa(X), ltr_refute(X))).
% ?- zdd((boole_to_cnf(-a * a,  X), psa(X))).
% ?- zdd((boole_to_cnf( a + a,  X), psa(X))).
% ?- zdd((cnf(-(a*b), X), psa(X))).
% ?- zdd((cnf(*([a,b,c]), X), psa(X))).
% ?- N = 10, numlist(2, N, Ns), zdd((cnf(*(Ns), X), psa(X))).
% ?- N = 100, numlist(2, N, Ns), zdd((cnf(*(Ns), X))).
% ?- N = 10000, numlist(2, N, Ns), zdd((cnf(*(Ns), X), ltr_card(10000, X, C))).
% ?- N = 10, numlist(2, N, Ns), zdd((cnf(*(Ns), X), ltr_card(9, X, C))).
% ?- N = 10, numlist(2, N, Ns), zdd((cnf(+(Ns), X), ltr_card(9, X, C))).
% ?- zdd((cnf(2+3+4+5, X), ltr_card(4, X, C))).
% ?- zdd((cnf(2+3+4+5, X), ltr_card(X, C))).

cnf(X, Y):- shift(cnf(X,Y)).

cnf(X, Y, S):- intern(X, Z, S),	boole_to_cnf(Z, Y, S).

cnf0(X, Y):- shift(cnf(X,Y)).

cnf0(X, Y, S):- intern(X, Z, S),
	simple_boole(Z, Z0),
	boole_to_cnf(Z0, Y, S).

%
boole_to_cnf(X, Y):- shift(boole_to_cnf(X,Y)).
%
boole_to_cnf(X, Y, S):- make_boole_canonical(X, X0),
	canonical_boole_to_cnf(X0, Y, S).

%
canonical_boole_to_cnf(1, 0, _):-!.
canonical_boole_to_cnf(0, 1, _):-!.
canonical_boole_to_cnf(X * Y, Z, S):-!,
	boole_to_cnf(X, U, S),
	boole_to_cnf(Y, V, S),
	ltr_join(U, V, Z, S).
canonical_boole_to_cnf(X + Y, Z, S):-!, boole_to_cnf(X, U, S),
	boole_to_cnf(Y, V, S),
	ltr_merge(U, V, Z, S).
canonical_boole_to_cnf(+(V), Y, S):- ltr_append(V, 1, Y, S).
canonical_boole_to_cnf(*(V), Y, S):-!, fold_join_list(V, 0, Y, S).
canonical_boole_to_cnf(cnf(V), Y, S):-!, ltr_read_cnf(V, 1, Y, S).
canonical_boole_to_cnf(X, Y, S):-!, zdd_singleton(X, Y, S).

ltr_read_cnf([], X, X, _).
ltr_read_cnf([X|Xs], Y, Z, S):- ltr_append(X, 1, X0, S),
	ltr_join(X0, Y, Y0, S),
	ltr_read_cnf(Xs, Y0, Z, S).

% ?- zdd((ltr_append([-b, a, -d, c], 1, X), psa(X))).
% ?- zdd((ltr_append([-b, 0, -d, c], 1, X), psa(X))).
% ?- zdd((ltr_append([-b, 1, -d, c], 1, X), psa(X))).

ltr_append(X, Y, Z):- shift(ltr_append(X, Y, Z)).
%
ltr_append([], X, X, _).
ltr_append([A|As], X, Y, S):- ltr_append(As, X, X0, S),
	(	A = 1 -> Y = X0
	;	A = 0 -> Y = 0
	; 	ltr_insert(A, X0, Y, S)
	).

% ?- zdd((X<<(ltr_set([a])+ltr_set([b])), psa(X))).
% ?- zdd((X<<(ltr_set([-a])+ltr_set([b])), psa(X))).
% ?- zdd((X<<(ltr_set([-a])+ltr_set([a])), psa(X))).

ltr_set(X, Y):- shift(ltr_set(X, Y)).
%
ltr_set(X, Y, S):- ltr_append(X, 1, Y, S).

%% ltr_complement(+X, -Y[, +G]) is det.
%  Y is the set of negated clauses C0 of C in X, where
%  C0 is the negated clause of C iff C0 is the set of all literals A0
%  such that the negation of A0 is in C.

% ?- zdd((X<<{[a, -b],[-b],[b, c]}, ltr_complement(X, Y), psa(Y))).
% ?- zdd((X<<{[-b],[-b],[b, c]}, psa(X), ltr_complement(X, Y),
%	nl, nl,  psa(Y))).

ltr_complement(X, Y):- shift(ltr_complement(X, Y)).
%
ltr_complement(X, X, _):- X<2, !.
ltr_complement(I, Y, S):- memo(ltr_complement(I)-Y, S),
	(	nonvar(Y) -> true
	; 	cofact(I, t(A, L, R), S),
		ltr_complement(L, L0, S),
		ltr_complement(R, R0, S),
		ltr_invert(A, NA),
		ltr_insert(NA, R0, R1, S),
		ltr_join(L0, R1, Y, S)
	).

%%	complementay(+A, -B) is det.
%    B is unfied with the complement of B.
ltr_invert(-A, A):-!.
ltr_invert(A, -A).

% ?- complementary(a, -a).
% ?- complementary(-a, b).

%%	complementay(+A, +B) is det.
%	True A and B are form a complementary to each other.
complementary(-(A), A):-!.
complementary(A, -(A)).

%% ltr_complement_toplevel(+X, -Y[, +G]) is det.

%  Change  0 and 1 at toplevel. Note that in CNF,
%  toplevel
%  0 ( set- theoretically empty {} ) is true as a set of clauses,
%  1 ( set- theoretically singleton {{}} ) is false as a set of clauses,
%      that is, unstisfiable.)
%  In DNF, toplevel 0 and 1 are consistent with false and true, respectively.

ltr_complement_toplevel(X, Y):- shift(ltr_complement_toplevel(X, Y)).
%
ltr_complement_toplevel(1, 0, _).
ltr_complement_toplevel(0, 1, _).
ltr_complement_toplevel(X, Y, S):- ltr_complement(X, Y, S).

% ?- zdd((boole_to_cnf(a*b, X), ltr_insert(c, X, Y), sets(Y, S))).
%@ X = 4,
%@ Y = 7,
%@ S = [[b, c], [a, c]].
% ?- zdd((boole_to_dnf((-a)*b, X), ltr_insert(a, X, Y), sets(Y, S))).
% ?- zdd((ltr_insert(a, 1, X), sets(X, S))).
% ?- zdd((ltr_insert(a, 1, X),
%	ltr_insert(b, X, Y), sets(Y, S))).
% ?- zdd((ltr_insert(a, 1, X),
%	ltr_insert(b, X, Y),
%	ltr_insert(-b, Y, Z),
%	sets(Z, S))).
% ?- zdd((ltr_insert(a, 1, X),
%	ltr_insert(b, X, Y),
%	ltr_insert(-b, Y, Z),
%	sets(Z, S))).

%%	ltr_insert(+X, +Y, -Z) is det.
% Insert a literal X into each set of literals
% in a ZDD Y, and unify Z with the result.
%
ltr_insert(X, Y, Z):- shift(ltr_insert(X, Y, Z)).
%
ltr_insert(_, 0, 0, _):-!.
ltr_insert(A, 1, J, S):-!, zdd_singleton(A, J, S).
ltr_insert(A, I, J, S):- memo(ltr_insert(A,I)-J, S),
	(	nonvar(J)	-> true
	;	cofact(I, t(X, L, R), S),
		ltr_compare(C, A, X),
		( 	C = (=)	 ->  zdd_join(L, R, U, S),
						 ltr_cofact(J, t(A, 0, U), S)
		;	C = (<)  ->  ltr_cofact(J, t(A, 0, I), S)
		;   C = (>)  ->  ltr_insert(A, L, L0, S),
						 ltr_insert(A, R, R0, S),
						 ltr_cofact(J, t(X, L0, R0), S)
		)
	).

% % simple version.
% ltr_merge(X, Y, Z):- shift(ltr_merge(X, Y, Z)).
% %
% ltr_merge(0, _, 0, _):-!.
% ltr_merge(_, 0, 0, _):-!.
% ltr_merge(1, X, X, _):-!.
% ltr_merge(X, 1, X, _):-!.
% ltr_merge(X, Y, Z, S):-
% 	(	X@<Y -> memo(ltr_merge(X,Y)-Z, S)
% 	;	memo(ltr_merge(Y,X)-Z, S)
% 	),
%  	(	nonvar(Z) -> true
% 	;	cofact(Y, t(A, L, R), S),
% 		ltr_merge(X, R, U, S),
% 		ltr_merge(X, L, V, S),
% 		ltr_insert(A, U, U0, S),
% 		ltr_join(V, U0, Z, S)
% 	).

%
% longer  version.  (which is more efficient longer or shorter version)

ltr_merge(X, Y, Z):- shift(ltr_merge(X, Y, Z)).

%
ltr_merge(0, _, 0, _):-!.
ltr_merge(_, 0, 0, _):-!.
ltr_merge(1, X, X, _):-!.
ltr_merge(X, 1, X, _):-!.
ltr_merge(X, Y, Z, S):-
	(	X@<Y -> memo(ltr_merge(X,Y)-Z, S)
	;	memo(ltr_merge(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		ltr_compare(C, A, A0),
		(	C = (=)
		->	ltr_merge(L, L0, L_L0, S),
			ltr_merge(L, R0, L_R0, S),
			ltr_merge(L0, R, L0_R, S),
			ltr_merge(R, R0, R_R0, S),
			ltr_join(L_R0, L0_R, U, S),
			ltr_join(R_R0, U, V, S),
			ltr_cofact(Z, t(A, L_L0, V), S)
		;   C = (<)
		->  ltr_merge(L, Y, L_Y, S),
			ltr_merge(R, Y, R_Y, S),
			ltr_cofact(Z, t(A, L_Y, R_Y), S)
		;   ltr_merge(L0, X, L0_X, S),
			ltr_merge(R0, X, R0_X, S),
			ltr_cofact(Z, t(A0, L0_X, R0_X), S)
		)
	).

% ?- zdd((X<<[a,c], Y<<[b, d, e], ltr_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<[a,b], Y<<[], ltr_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<[a,c], Y<<[b, d, e], ltr_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<[a,b], Y<<[-a, c], ltr_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((X<<[a,c], Y<<[b, d, e], ltr_merge_line(X, Y, Z), psa(Z))).
% ?- zdd((zdd_line([-a,b], X), Y<<[a, b, c], ltr_merge_line(X, Y, Z), psa(Z))).
% special version of ltre_merge.
ltr_merge_line(X, Y, Z):- shift(ltr_merge_line(X, Y, Z)).

ltr_merge_line(X, X, X, _):-!.		% idempotent
ltr_merge_line(0, _, 0, _):-!.
ltr_merge_line(_, 0, 0, _):-!.
ltr_merge_line(1, X, X, _):-!.
ltr_merge_line(X, 1, X, _):-!.
ltr_merge_line(X, Y, Z, S):- cofact(X, t(A, _, R), S),
	cofact(Y, t(A0, _, R0), S),
	(	complementary(A, A0)
	->	Z = 0
	;	ltr_compare(C, A, A0),
		(	C = (<)
		->	ltr_merge_line(R, A0, R0, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	C = (=)
		->	ltr_merge_line(R, R0, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	ltr_merge_line(R0, A, R, Z0, S),
			cofact(Z, t(A0, 0, Z0), S)
		)
	).

%
ltr_merge_line(0, _, _, 0, _):-!.
ltr_merge_line(1, A, R, Z, S):-!, cofact(Z, t(A, 0, R), S).
ltr_merge_line(X, A, R, Z, S):- cofact(X, t(B, _, R0), S),
	(	complementary(A, B)
	->	Z = 0
	; 	compare(C, A, B),
		(	C = (<)
		->	ltr_merge_line(R, B, R0, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	C = (=)
		->	ltr_merge_line(R, R0, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	ltr_merge_line(R0, A, R, Z0, S),
			cofact(Z, t(B, 0, Z0), S)
		)
	).


% ?- zdd((cofact(B, t(-b, 1, 1)),
%		cofact(C, t(a, 1, 1)),
%		ltr_cofact(A, t(-a, C, B)),
%		psa(C), psa(B),
%		psa(A))).
% ?- zdd((cofact(B, t(-b, 1, 1)),
%		cofact(C, t(-a, 1, 1)),
%		ltr_cofact(A, t(a, C, B)),
%		psa(C), psa(B),
%		psa(A))).

% Designed so that complentary pair does not appear in any clause.
ltr_cofact(X, Y):- shift(ltr_cofact(X, Y)).

%
ltr_cofact(Z, FOS, S):- nonvar(Z), !, cofact(Z, FOS, S).
ltr_cofact(Z, t(A, V, U), S):- U > 1, !,
	cofact(U, t(B, L, _), S),
	(	complementary(A, B)
	->  ltr_cofact(Z, t(A, V, L), S)
	;	cofact(Z, t(A, V, U), S)
	).
ltr_cofact(Z, T, S):- cofact(Z, T, S).

% ?- zdd((X<<[a], ltr_insert(b, X, Y), psa(Y))).
% ?- zdd((X<<[b], ltr_insert(a, X, Y), psa(Y))).
% ?- zdd((X<<[a], ltr_insert(-a, X, Y), psa(Y))).
% ?- zdd((X<<[-a], ltr_insert(a, X, Y), psa(Y))).
% ?- zdd((X<<([a,b,c]+[u,v]), ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<[a,b,c], ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b],[c]}, ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,-b],[c]}, ltr_insert(b, X, Y), psa(Y))).
% ?- zdd((zdd_append([-a, b],1, X), psa(X),  ltr_remove(-a, X, Y), psa(Y))).
% ?- zdd((X<<(set([-a, -b])+set([a, b, c])), psa(X),  ltr_remove(b, X, Y), psa(Y))).

ltr_remove(X, Y, Z):- shift(ltr_remove(X, Y, Z)).
%
ltr_remove(X, Y, Z, S):- zdd_shift(ltr_remove(X, Y, Z, S)).

%
ltr_remove(_, I, I, _, _):- I<2, !.
ltr_remove(A, I, J, S, M):- memo(ltr_remove(I)-J, M),
	(	nonvar(J) -> true
	;	cofact(I, t(X, L, R), S),
		ltr_compare(C, A, X),
		(	C = (=)
		->  ltr_join(L, R, J, S)
		;	C = (<)
		->  J = I
		;   ltr_remove(A, L, L0, S),
			ltr_remove(A, R, R0, S),
		    ltr_cofact(J, t(X, L0, R0), S)
		)
	).

% ?- zdd((X<<{[]}, ltr_card(1, X, C))).
% ?- zdd((X<<{[]}, ltr_card(2, X, C))).
% ?- zdd((X<<{[a]}, ltr_card(1, X, C))).
% ?- zdd((X<<{[a]}, ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + (-a), X), ltr_card(1, X, C))).
% ?- zdd((boole_to_dnf(a + -a, X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + b, X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + b + c, X), ltr_card(3, X, C))).
% ?- zdd((boole_to_dnf(a + b + c + d, X), ltr_card(4, X, C))).
% ?- zdd((boole_to_dnf(-(a + b + c + d), X), ltr_card(4, X, C))).
% ?- zdd((boole_to_dnf(a + b + c + -d, X), ltr_card(4, X, C))).
% ?- zdd((boole_to_dnf(a * b, X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a * b, X), ltr_card(3, X, C))).
% ?- zdd((boole_to_dnf(-(a * b), X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(-a * a, X), ltr_card(1, X, C))).
% ?- zdd((boole_to_dnf(-a * a, X), ltr_card(2, X, C))).

% problematic
% ?- zdd((boole_to_dnf(a + -a, X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + -a, X), ltr_card(X, C))).


% ?- zdd((boole_to_dnf(a + b + -a, X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + b + -a, X), psa(X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + b + -a + -b, X), psa(X), ltr_card(2, X, C))).
% ?- zdd((boole_to_dnf(a + b + c + -a + -b + -c, X), psa(X), ltr_card(3, X, C))).
% ?- zdd((boole_to_dnf(a + c + -a + -b + -c, X), psa(X), ltr_card(3, X, C))).

positive(-(_)):-!, fail.
positive(_).

% ?- zdd(( X<<pow([a,b]), ltr_card(X, C))).
% ?- zdd(( dnf(a+b, X), psa(X), varnum(N), ltr_card(X, C))).
% ?- zdd(( dnf(A+B, X), ltr_card(X, C))).
% ?- zdd(( dnf(a, X), psa(X), ltr_card(X, C))).

ltr_card(X, Y):- shift(ltr_card0(X, Y)).

ltr_card0(X, Y, S):- memo(atomic_prop_memo-M, S),
	memo(varnum-N, M),
	ltr_card(N, X, Y, S).


ltr_card(N, X, Y):- shift(ltr_card(N, X, Y)).
%
ltr_card(N, X, Y, S):- zdd_shift(ltr_card(N, X, Y, S)).
ltr_card(_, 0, 0, _,  _):-!.
ltr_card(N, 1, C, _,  _):-!, C is 2^N.
ltr_card(N, X, C, S, M):- memo(path_count(N, X)-C, M),
	(	nonvar(C) -> true
	;	cofact(X, t(A, L, R), S),
		N0 is N-1,
		(	L > 1,
			positive(A),
			cofact(L, t(B, L0, R0), S),
			complementary(A, B)
		->	ltr_join(R, L0, U, S),	% for not counting multple time.
			ltr_join(R0, L0, V, S),
			ltr_card(N0, U, C_U, S, M),
			ltr_card(N0, V, C_V, S, M),
			C is C_U + C_V
		;	ltr_card(N0, L, C_L, S, M),
			ltr_card(N0, R, C_R, S, M),
			C is C_L + C_R
		)
	).


% Remark.  % 1 is "true", 0 is "false", but 2 is a basic proposition.
% ?- zdd((dnf(X+2+1, A), psa(A), ltr_card(2, A, C))).
% ?- zdd((dnf(X+2+0, A), psa(A), ltr_card(2, A, C))).
% ?- zdd((dnf(a(1)+ a(2), A),ltr_card(2, A, C))).
% ?- zdd((dnf(a+b+true, A), psa(A), ltr_card(2, A, C))).
% ?- zdd((dnf(a+b+false, A), psa(A), ltr_card(2, A, C))).
% ?- zdd((dnf(nand(a,b)+ exor(a, b), A), psa(A), ltr_card(2, A, C))).
% ?- zdd((dnf(a_nand(a,b)+ exor(a, b), A), psa(A), ltr_card(3, A, C))).
% ?- zdd((dnf(a_nand(a,b)+ b_exor(a, b), A), psa(A), ltr_card(2, A, C))).

%%	atomic_prop(+X) is det.
% True if X is an atomic boolean form.
atomic_prop(X) :- var(X), !.
atomic_prop(X) :- integer(X), X>1, !.
atomic_prop($(_)) :-!.		% generated internally for prolog variable.
atomic_prop(X):- callable(X),
				 \+ reserved_boole_symbol(X),
				 functor(X, F, _),
				 atom_codes(F, [C|_]),
				 char_type(C, lower).
%
reserved_boole_symbol(1).
reserved_boole_symbol(0).
reserved_boole_symbol($(_)).
reserved_boole_symbol(true).
reserved_boole_symbol(false).
reserved_boole_symbol(not(_)).
reserved_boole_symbol(and(_,_)).
reserved_boole_symbol(or(_,_)).
reserved_boole_symbol(exor(_,_)).
reserved_boole_symbol(nand(_,_)).
reserved_boole_symbol(imply(_,_)).
reserved_boole_symbol(iff(_,_)).
reserved_boole_symbol(equiv(_,_)).

% ?- zdd((dnf(-a, X), psa(X))).
% ?- zdd((dnf(A+B, X), ltr_card(2, X, C))).

% ?- zdd((dnf(-(a+b), X), psa(X))).
% ?- zdd((dnf(-(a*b), X), psa(X))).

% ?- collect_basics(f(1, A, a(2), A, 1, 2), R).
% ?- collect_basics(9+ a(1)+b(I)*x(3), R).
%%	collect_basics(+A, -B) is det.
% Unify B with a list of basic boolean subterms appearing in A.
collect_basics(A, B):- collect_basics(A, B, atomic_prop).
%
collect_basics(A, B, Pred):- collect_basics(A, B0, [], Pred),
					   sort(B0, B).
%
collect_basics([], L, L, _):- !.
collect_basics(X, [X|L], L, Pred):- call(Pred, X), !.
collect_basics([X|Y], L, L0, Pred):- !,
	collect_basics(X, L, L1, Pred),
	collect_basics(Y, L1, L0, Pred).
collect_basics(X, L, L0, Pred):- compound(X), !,
	X =..[_|As],
	collect_basics(As, L, L0, Pred).
collect_basics(_, L, L, _).

% ?- make_boole_canonical(a->(b->c), R).
% ?- make_boole_canonical(a=b, R).
% ?- make_boole_canonical(+([a, b]), R).
% ?- make_boole_canonical(-(+([a, b])), R).
% ?- make_boole_canonical(-(*([a, b])), R).
% ?- make_boole_canonical(-(a+b), X).
% ?- make_boole_canonical(-(a+ -b), X).
% ?- make_boole_canonical(-(a = -b), X).
% ?- make_boole_canonical(-(-a = - -b), X).
% ?- make_boole_canonical(-(-(0) = - (-(0))), X).
% ?- make_boole_canonical(-(0), X).
% ?- make_boole_canonical(-(-(0)), X).
% ?- make_boole_canonical(p(1,1), X).
% ?- make_boole_canonical(true, X).
% ?- make_boole_canonical(false, X).
% ?- make_boole_canonical(-false, X).
% ?- make_boole_canonical(-true, X).
% ?- make_boole_canonical(- - -true, X).
% ?- make_boole_canonical(exor(a, b), X).
% ?- make_boole_canonical(nand(a, b), X).
% ?- make_boole_canonical(- nand(a, b), X).
% ?- make_boole_canonical(- #(a, b), X).

make_boole_canonical(X, Y):- make_boole_canonical(X, Y, boole_macro).
%
:- meta_predicate  make_boole_canonical(?, ?, :).
%
make_boole_canonical(-(-(X)),  Y, Pred):-!,
	make_boole_canonical(X, Y, Pred).
make_boole_canonical(-(X), Y, Pred):- make_boole_canonical(X, X0, Pred), !,
	(	de_morgan(X0, Y) -> true
	;	Y = -(X0)
	).
make_boole_canonical(X, Y, Pred):- default_boole_macro(X, X0),!,
	make_boole_canonical(X0, Y, Pred).
make_boole_canonical(X, Y, Pred):- call(Pred, X, X0), !,
	make_boole_canonical(X0, Y, Pred).
make_boole_canonical(X, X, _).

% ?- X= (a<-b).

%
default_boole_macro(X->Y,		-X + Y).
default_boole_macro(X=Y,	(X->Y)*(Y->X)).
default_boole_macro(nand(X,Y), -(X*Y)).
default_boole_macro(exor(X,Y),  (X + Y) * (-(X * Y))).
%
canonical_boole(X):- integer(X).
canonical_boole(+(_,_)).
canonical_boole(*(_,_)).
canonical_boole(-(_)).
canonical_boole(+(X)):- is_list(X).
canonical_boole(*(X)):- is_list(X).

%
boole_macro(true,		1).
boole_macro(false,		0).
boole_macro(X =:= Y,	X = Y).
boole_macro(X =\= Y,	-(X = Y)).
boole_macro(X <=> Y,	X = Y).
boole_macro(X \= Y,		-(X = Y)).
boole_macro(X =\= Y,	-(X = Y)).
boole_macro(X >= Y,	    Y -> X).		% in harmony arithmetic >
boole_macro(X =< Y,	    X -> Y).		% in harmony arithmetic <
boole_macro(X ~ Y,		X = Y).
boole_macro(X <-> Y,	X = Y).
boole_macro(\/(X,Y),	+(X,Y)).
boole_macro(\(X),		-(X)).
boole_macro(/\(X, Y),	*(X,Y)).
boole_macro((X; Y),		+(X,Y)).
boole_macro(\+(X),		-(X)).
boole_macro((X, Y),		*(X,Y)).
boole_macro(#(X, Y),	exor(X,Y)).
boole_macro(exor(X, Y),	(X + Y)*(-X + -Y)).
boole_macro(nand(X, Y),	-(X * Y)).
boole_macro(equiv(X, Y), X = Y).
boole_macro(imply(X, Y),	-(X) + Y).
boole_macro(not(X),		-(X)).
boole_macro(and(X, Y),	*(X,Y)).
boole_macro(or(X, Y),	+(X,Y)).

%
de_morgan(1,	0).
de_morgan(0,	1).
de_morgan(-(X), X).					% double negation.
de_morgan(*(X, Y), (-X) + (-Y)).	% de Morgan
de_morgan(+(X, Y), (-X) * (-Y)).	% de Morgan
de_morgan(*(X), +(Y)):- maplist(negate_macro, X, Y).
de_morgan(+(X), *(Y)):- maplist(negate_macro, X, Y).

%
pred_plus(A, U, A + U).
pred_mul(A, U, A * U).
%
negate_macro(X, -(X)).

% ?- univ_to_binary(+([X,Y,Z]), R).
% ?- univ_to_binary(*([X,Y,Z]), R).
univ_to_binary(U, B):- U=..[F, L], univ_to_binary(L, F, B).
%
univ_to_binary([X], _, X):-!.
univ_to_binary([X,Y|Z], F, B):- univ_to_binary([Y|Z], F, B0),
		B=..[F, X, B0].
